//--- BCM.c --- 10/11/07 ---(04/09/02)---
#define DeviceVers    16

// either CHAN_SS or BCM_RR should be defined by the compile command
//#define  BCM_RR   // for Reed-Relay  Front-End version (most BCM's)
//#define  BCM_SS   // for Solid-State Front-End version (only 1 device as of 9/04)

#define DevName       "Beam Cur Mon"

#ifdef BCM_SS
  #define DeviceType    0x19
#else
  #define DeviceType    0x09
#endif

//-----------------------------------------------------------------------------
//
//  Copyright (c) 2002-2006 National Superconducting Cyclotron Laboratory,
//            Michigan State University, and Mark Davis
//
//-----------------------------------------------------------------------------
//  Program to replace older ZWorld BL1500 (Micro-G) serial based controller in
//  the Beam Current Monitor (BCM) with a ZWorld BL2105 (SmartCat) ethernet
//  based one using the ethernet to communicate using simple Modbus commands
//  over TCP.
//
//  For notes on program structure, see the file:
//    ..\ZWORLD\NSCL_ZWORLD_LIBS\BASE_DLP_DLM_C.LIB
//-----------------------------------------------------------------------------
//
//  NOTE:
//    The external calibration technique in the previous version of this
//    code altered the calibration constants stored for converting the raw ADC
//    counts to a voltage value, but never actually used them (or the ZWorld
//    lib routine) to do this conversion, so this calibration had no effect on
//    the current values returned.
//
//    This version of the code now leaves the ADC calibration data put in the
//    Flash memory by ZWorld alone and stores calibration data for each range
//    in each channel which should result in much more accurate readings, as
//    well as making it simpler to modify the ranges.
//
//-----------------------------------------------------------------------------
//  !!!!!! DON'T FORGET TO CHANGE DeviceVers ABOVE !!!!!!
//-----------------------------------------------------------------------------
//
//  2007-10-11 - Mark Davis - Version 16
//               - Added ability to shift all the ranges by a factor of 10
//                 by ground digitalInput # 23
//
//  2006-08-29 - Mark Davis - Version 15
//               - Run-Time 0 calibration now starts with the range the
//                 channel is set to when the calibration process is started.
//               - 'T' now toggles Test Current, not the Filter...
//
//  2006-04-12 - Mark Davis - Version 14
//               - major restructure and cleaning up of code to make it simpler
//                 and conform to the behavior specified in the new Software
//                 Requirements Document.
//
//  2006-03-21 - Mark Davis - Version 13
//               - Use of the RangeOK status is now different.  Rather than
//                 setting this when a run-time zero calibration results in
//                 in a large offset, it is now cleared only to indicate that
//                 the bench-top calibration data is bad.  This can happen
//                 during the actual benchtop calibration, or it can happen
//                 when the BCM restarts and finds that the calibration data
//                 stored in flash is unusable.  NOTE: Only the benchtop
//                 calibration process or a user-typed command will save this
//                 value to flash.
//                 We now depend upon the overrange status that is updated
//                 continuously at run-time to to deal with large offsets that
//                 actually cause a problem (this allows large offsets to NOT
//                 cause a problem if the resulting measurement is still in
//                 the range of readable values).
//
//  2006-03-20 - Mark Davis - Version 12
//               - fixes to CHAN_ValidCalibData():
//                 - The value used to check for a corrupted RTOffset value
//                   was changed from 10 to 12 (10 was a value that can all
//                   to often be achieved based on input and range setting when
//                   a run-time calibration is done).  This caused the (valid)
//                   calibration data to be overwritten in too many cases.
//                 - The bench-top offset value (BCZeroOffset) is now checked
//                   for a corrupted value (10 is a safe "wild" value here)
//               - bad calibration data now sets the CHAN_STATUS_HW_ERR bit
//                 rather than reseting the calibration data to defaults.
//               - the default calibration for the odd #'d "software" ranges
//                 now equals the corresponding hardware range
//
//  2006-02-23 - Mark Davis - Version 11
//               - added CHAN_STATUS_RESET bit to chanStatus
//
//  2005-06-24 - Mark Davis - Version 10
//               - the calibration data in flash is now updated when the
//                 run-time offsets are reset to 0 (in CHAN_ClearOffsets)
//
//  2005-06-22 - Mark Davis - Version 9
//               - Bench Calibration offset now computed correctly (independent
//                 of RunTime zero-level offset) and the RunTime offsets are
//                 set to zero when calibration data for a channel is updated
//
//  2005-06-15 - Mark Davis - Version 8
//               - removed MaxVariance limitation from 0-level calibration
//
//  2005-06-10 - Mark Davis - Version 7
//               - variance is now computed relative to the range of possible
//                 values (rather than the current reading, which resulted in
//                 increasingly larger values near zero)
//               - added avgInterval register to provide # ms over which the
//                 averaging is being done
//               - increased MaxVariance to 0.20 (to avoid some ranges getting
//                 marked BAD when doing 0-offset calibration
//               - min 1 second must now elapse after last decrease in variance
//                 before 0-offset calibration cycle is completed for a range
//
//  2005-02-14 - Mark Davis - Version 6
//               - updated to use all latest base code changes
//               - new calibration data is no longer saved until the LAST range
//                 is calibrated
//               - added a new command to manuall save calib data for a channel
//
//  2004-06-28 - Mark Davis - Version 5
//               - updated to use new BlockList method of storing persistent
//                 configuration data
//
//  2004-01-22 - Mark Davis - Version 4
//               - updated to comply with newer/easier method of defining
//                 ModBus registers
//               - added command to stream data out to a telnet client
//               - recompiled to incorporate changes in base code like
//                 Interrupt driven I/O and processing of basic IP config pkts
//
//  2003-11-13 - Mark Davis (CHAN_03.bin)
//               - updated code with all applicable enhancements and interface
//                 standards done for the servo controller code.
//               - fixed some remote cmd handling and calibration/init bugs
//               - zero-level calibration now done with front-end setting
//                 (on/off) as-is (i.e it no longer determines noise offset
//                 for both front-end on and off)
//               - fixed Save/Load/Valid/ResetCalibData() functions
//
//  2003-05-22 - Mark Davis (CHAN_02.bin)
//               - xxx
//
//  2003-02-13 - Mark Davis - vt100 clear screen now sent for new telnet
//               connection.
//               - added support for setting the amount of time the Download
//               Manager waits for a connection (reboot due to power loss
//               restarts this code ASAP, while rebooting via the Reboot command
//               causes it to wait for 30 seconds).
//               (CHAN_01.bin)
//
//  2003-02-07 - Mark Davis - minor fixes to serial code to help avoid possible
//               memory access errors
//
//  2002-12-17 - Mark Davis - continued work on field-upgradable version
//
//  2002-09-17 - Mark Davis - made changes to use callback version of NSCL
//               lib that interfaces with the ZWorld RS485 functions.  This
//               drastically improves the maximum sampling rate.
//
//  2002-08-19 - Mark Davis - changed wiring for Activty LED and addr switch
//               to be consistent with other ZWorld controllers
//
//  2002-07-29 - Mark Davis - started modifictions for field-upgradable version
//
//  2002-07-08 - Mark Davis - added support for NSCL RS-485 serial protocol
//
//  2002-04-09 - Mark Davis - original version started based on Modbus test
//               code I wrote for the newer ZWorld BL2100 (SmartCat) based DC
//               Servo controller and the code John Priller wrote for the
//               ZWorld BL1500 (MicroG2) based Beam Current Monitor.
//
//-----------------------------------------------------------------------------
//
//  Operational and Programming NOTES:
//
//  - The Host expects the most significant bits at lower addresses while the
//    ZWorld keeps the most significant bits at the higher address
//
//  - This code assumes that, when recieving values via ModBus TCP protocol,
//    any values that span multiple registers (such as floating point values)
//    will be recieved via a Write Multiple Registers command rather than
//    multiple Write Single Register commands (the latter can cause anomolous
//    behavior).
//
//-----------------------------------------------------------------------------


//#define ModBusSerial


#define SERIAL_A_AVAIL
#if (_BOARD_TYPE_ != RCM3200)
  #define SERIAL_B_AVAIL
#endif
#define SERIAL_C_AVAIL
#define SERIAL_D_AVAIL
//#define SERIAL_E_AVAIL
//#define SERIAL_F_AVAIL


/*----- Ideas for next version -----

- rather than assume a linear transfer function for calibrating each range,
  modify the Windows calibration application and this controller code to use
  more than 2 points and interpolate.  This could significantly improve the
  accuracy (particularly at the most senstive range).

  Question:  ARE the values for the most sensitive range consistent over time?
             How accurate is our current source, and how much does IT vary?

------------------------------------*/

#class auto
#memmap xmem
#nodebug



#ifdef  COMPILE_PRIMARY_PROG
  #error "Cannot compile as primary program."
  #error "Remove COMPILE_PRIMARY_PROG macro in 'Options | Compiler | #defines' options"
  #error "Must Recompile BIOS also. Hit Ctrl-Y then recompile"
#endif

#ifdef  DevelopMode
  #ifdef  COMPILE_SECONDARY_PROG
  #error "DevelopMode set - must NOT compile as secondary program."
  #error "Remove COMPILE_SECONDAY_PROG macro in 'Options | Compiler | #defines' options"
  #error "Must Recompile BIOS also. Hit Ctrl-Y then recompile"
  #endif
#else
  #ifndef  COMPILE_SECONDARY_PROG
  #error "Must compile as secondary program or DevelopMode."
  #error "Put COMPILE_SECONDARY_PROG macro in 'Options Compiler | #defines' options"
  #error "Must Recompile BIOS also. Hit Ctrl-Y then recompile"
  #endif
#endif

#if (_BOARD_TYPE_ != BL2105)
  #error "Board/target type must be set for BL2105"
#endif

#define  ADCHistSize  100  // up to 1 second at 100Hz sample rate

#define  NumADC    2  // # of A/D inputs used
#define  NumDAC    1  // # of D/A outputs used

#define  NumChans  2  // number of channels this controller supports

#define DEBUG  false

#define  ChanMult     100  // multipler for chan number or'd to reg & coil #'s


//=============================================================================
//         Definitions the same or nearly so for all controllers
//=============================================================================




//=============================================================================
//          Common stuff but with Controller-specific variations
//=============================================================================

//----- bit usage masks for digitalIn -----
#define  BCM2_TEST_CUR_OFF    0x0040
#define  BCM1_TEST_CUR_OFF    0x0080
//#define  BaudIdx()  ((int)((~digitalIn >> 12) & 0x03))
#define  BaudIdx()  (1)  // BCM's are fixed at 19200
#define  AddrIdx()  ((int)((~digitalIn >> 8) & 0x0f))
#define  NotRangeX10     0x800000  // 0 ==> multiply FSCurrent's by 10

//----- bit usage masks for digitalOut -----
#define  ACTIVITY_LED    0x000001  // which bit is used for the activity LED

#define  BCM1_K1         0x000002  // Front End (extern input) On/Off
#define  BCM1_K2         0x000004  // Filter On/Off
#define  BCM1_K3         0x000008  // gain
#define  BCM1_K4         0x000010  // gain
#define  BCM1_K5         0x000020  // gain
#define  BCM1_K6         0x000040  //

#define  BCM1_TEST_CUR   0x000080  // enable test curr (SolidState FE vers only)

#define  BCM2_TEST_CUR   0x000200  // enable test curr (SolidState FE vers only)

#define  BCM2_K1         0x000400  // Front End (extern input (On/Off)
#define  BCM2_K2         0x000800  // Filter On/Off
#define  BCM2_K3         0x001000  // gain
#define  BCM2_K4         0x002000  // gain
#define  BCM2_K5         0x004000  // gain
#define  BCM2_K6         0x008000  //

//--- REAL I/O bit usage is not symetrical, so generic values that use the ---
//--- same bit patterns for both channels are derived from the real input ---
//--- values and used to set the real output values ---

//--- bit usage for bcm1_in, bcm2_in (created from real input bits) ---
#define TEST_SW_ON        0x01   // test current switch state

//--- bit usage for bcm1_out, bcm2_out (used to generate real output bits) ---
#define FRONTEND_BIT      0x01   // input connected relay  (K1)
#define RANGE_BIT         0x1E   // input gain relays      (K2-K5)
#define FILTER_BIT        0x20   // filter enabled relay   (K6) (0=On)
#define TESTCUR_BIT       0x40   // test-cur enabled relay (SS FrontEnd only)


//--- to make base code happy for now ---
typedef struct {
  int  msgID;
  int  seqNum;
} EventMsg;


//--- structure for Event messages ----
typedef struct  {
  int    event;     // which event?
  int    chanNum;   // for which channel?
  float  floatVal;
  int    intVal;
}  EventData;

//--- prototypes for functions to handle events ---
//--- !!!!! to be called ONLY within the ControlLoop() process !!!!! ---
void Init_(EventData *event);
void StartCalib_(EventData *event);
void AbortCalib_(EventData *event);
void FECalibPoint_(EventData *event);
void RangeUp_(EventData *event);
void RangeDown_(EventData *event);
void AvgIdxUp_(EventData *event);
void AvgIdxDown_(EventData *event);
void FrontEndOn_(EventData *event);
void FrontEndOff_(EventData *event);
void FilterOn_(EventData *event);
void FilterOff_(EventData *event);
void ClearOffsets_(EventData *event);
void TestCurOn_(EventData *event);
void TestCurOff_(EventData *event);
void SaveCalib_(EventData *event);
void NewRange_(EventData *event);
void NewAvgIdx_(EventData *event);
void DupRange_(EventData *event);

typedef void (*EventHandler)(/*EventData *event*/);

//--- values for EventData.event ---
const EventHandler  eventHandler[] = {
#define  MSG_Init          0
             Init_,
#define  MSG_StartCalib    1
             StartCalib_,
#define  MSG_AbortCalib    2
             AbortCalib_,
#define  MSG_FECalibPoint  3
             FECalibPoint_,
#define  MSG_RangeUp       4
             RangeUp_,
#define  MSG_RangeDown     5
             RangeDown_,
#define  MSG_AvgIdxUp      6
             AvgIdxUp_,
#define  MSG_AvgIdxDown    7
             AvgIdxDown_,
#define  MSG_FrontEndOn    8
             FrontEndOn_,
#define  MSG_FrontEndOff   9
             FrontEndOff_,
#define  MSG_FilterOn     10
             FilterOn_,
#define  MSG_FilterOff    11
             FilterOff_,
#define  MSG_ClearOffsets 12
             ClearOffsets_,
#define  MSG_TestCurOn    13
             TestCurOn_,
#define  MSG_TestCurOff   14
             TestCurOff_,
#define  MSG_SaveCalib    15
             SaveCalib_,
#define  MSG_NewRange     16
             NewRange_,
#define  MSG_NewAvgIdx    17
             NewAvgIdx_,
#define  MSG_DupRange     18
             DupRange_
};

#define  LastMsgNum  ( sizeof(eventHandler) / sizeof(eventHandler[0]) - 1 )


#use "Base_DLP_H.lib"

bool getMsg(EventData *eventData);
void putMsg(int event, EventData *eventData);

#define  EVENT(name)  putMsg(MSG_ ## name, &eventData)



//----- values we want to be PERSISTENT for each BCM channel -----
//-- !!!! Altering this structure will result loss of prev config data !!!! --
typedef struct {
  bool         rangeOK;        // range tested OK during BENCHTOP calibration
  bool         inputOn;        // state of Front-End when RTOffset determined
  float        RTOffset;       // Run-Time voltage offset
  float        BCAmpsPerVolt;  // Bench-Calibrated gain using known current
  float        BCZeroOffset;   // Bench-Calibrated offset from zero volts
} FrontEndCalibData;

#define NumRanges  16   // number of current ranges

//----- values we want to be PERSISTENT (i.e. stored in flash) -----
//--- (search for PERSISTENT for places code normally added to support them) ---
//-- !!!!! Adding to this struct is OK, but never touch existing items !!!! ---
typedef struct  {
  uint   vers;            // version of this structure
  uint   devType;         // device type this data came from
  FrontEndCalibData  calibData[NumChans][NumRanges];
}  ApplData;

#define  ApplDataVers   1

#define CALIB_DATA_SIZE   sizeof(applData.calibData[0])


//========================================
//===== Coil or ModBus bit registers =====
//========================================
//--- writing to a coil causes an action - values are not actually saved ---
//  See NewCoilValue() for details

//--- NOTE: 0-19 (00001-00020) are reserved for common values ---

#define  RangeUp      23  // [00024]
#define  RangeDown    24  // [00025]

#define  AvgIdxUp     25  // [00026]
#define  AvgIdxDown   26  // [00027]

#define  FrontEndOn   27  // [00028]
#define  FrontEndOff  28  // [00029]

#define  FilterOn     29  // [00030]
#define  FilterOff    30  // [00031]

#define  ClearOffsets 31  // [00032] - used to be 00013
#define  DoCalibrate  32  // [00033] - used to be 00003

#define  TestCurOn    33  // [00034] - Only for SolidState FrontEnd vers
#define  TestCurOff   34  // [00035] - Only for SolidState FrontEnd vers

#define  SaveCalib    35  // [00036] - Added 03/20/06
//                           =======
          #define  NUMCOILS      36


//======================================
//===== Read-Only ModBus registers =====
//======================================
//--- 30001-30050 (0-49) are reserved for generic/common values ---
//--- 30051-30099 are for controller specific values ---

//RO_REG_U16( roApplData1,   30051 )
//                           =====
       #define  NUM_READ_REGS   50


//=======================================
//===== Read/Write ModBus registers =====
//=======================================
//--- Note that read/write registers can be defined as "Read Only" ---
//--- This allows the client to send write commands for a register to ---
//--- initiate some action without actually changing the value of the ---
//--- register.  Also, devices that have read-only values but multiple ---
//--- logical devices need to use the R/W registers, as the Read-Only ---
//--- set does not support multiple devices, and generally is limited ---
//--- to values that apply to all logical devices handled by one CPU ---

//--- Special handling for chgs to reg values is done by DevNewRegValue() ---

//--- The ChanMult value determines how many registers can be defined ---
//--- per logical device.  A value of 100 means 40001-40100 are for ---
//--- device 0, 40101-40200 are for device 1, etc.  Note that only ---
//--- one set of register definitions is required (i.e. define 40001 ---
//--- but not 40101, 40201, etc).

RW_REG_F32( current,       40001, RO )
RW_REG_U16( rangeIdx,      40003, RW )
RW_REG_F32( fullScale,     40004, RO )  // upper end of range
RW_REG_U16( avgingSize,    40006, RO )  // # elements in averaging buffer
RW_REG_U16( avgingIdx,     40007, RW )  // # index into list of avg buffer sizes
RW_REG_U16( chanStatus,    40008, RW )  // See CHAN_STATUS_xxx values
RW_REG_F32( calibPoint1,   40009, RW )
RW_REG_F32( calibPoint2,   40011, RW )
RW_REG_U32( extra1,        40013, RO )  // long int "debug" data used by Test appl
RW_REG_U32( extra2,        40015, RO )
RW_REG_U32( extra3,        40017, RO )
RW_REG_U32( extra4,        40019, RO )
RW_REG_U32( extra5,        40021, RO )
RW_REG_U32( extra6,        40023, RO )
RW_REG_U32( extra7,        40025, RO )
RW_REG_F32( variance,      40027, RO )
RW_REG_F32( volts,         40029, RO )
RW_REG_U16( avgInterval,   40031, RO )  // # ms over which rdgs are averaged
//                         =====
     #define  NUMREGS         31   // don't forget _x32 types use 2 regs!

// search for _PERSISTENT_ to see where code normally added for saved values

RegInfo  rwRegs[NUMREGS+1];
RegInfo  roRegs[NUM_READ_REGS+1];

//=============================================================================
//              Definitions generally unique to this controller
//=============================================================================

#define ADC_FULL_SCALE  10.0  // nominal full scale value of ADC

#define  VarianceSecs  2  // # secs to monitor for variance computations


//
// Structure for storing ADC calibration values
//
typedef struct {
  float  Current;
  float  Voltage;
} RangeCalibData;

//
// ChanState structure
//
#define NUM_EXTRA 7
#define OLD_NUM_EXTRA  4


typedef struct {
  bool       calibState;           // channel is doing RTZero calibration
  bool       saveFrontEndOn;       // frontend on/off state before calib
  bool       saveFilterOn;         // filter on/off state before calib
  int        saveRange;            // range state before calib
  int        saveAvgSizeIdx;       // avgSizeIdx state before calib

  bool       saveRTZOffsets;       // save RTZero offsets for chan to flash
  bool       saveFECalib;          // save all FE calib data for chan to flash

  char       OverRange;            // over-range state (CHAN_OVERRANGE_xxx)
  bool       HWError;              // hardware error state (true or false)

  ulong      settlingStart;        // ms count for calibration settling
  float      variance;             // % variation in recent readings
  float      lowestVariance;       // for monitoring progress
  int        numSamples;           // # samples since last FrontEnd change

  int        minRaw;               // min raw ADC value over recent period
  int        maxRaw;               // max raw ADC value over recent period
  float      avgRaw;               // avgd rdg in float format (more accurate)

  float      avgCur;               // calc & offseted beam current
  float      avgVolts;             // average converted to volts
  float      adjVolts;             // adj volts: avgVolts less noise offset
  float      fracFSVolts;          // Fraction of Full-scale Voltage
  float      fracFSCur;            // Fraction of Full-scale Current

  char      *InputByte;            // pointer to input byte
  uint      *AdcWord;              // pointer to ADC input word
  char      *OutputByte;           // pointer to output byte

  bool       testCurOn;            // Test Current switch/relay is on
  bool       frontEndOn;           // Front End is connected to input
  bool       filterOn;             // Front End filter is on
  int        range;                // range # (0-15)
  int        avgSizeIdx;           // index into table of # samples to avg

  long       Extra[NUM_EXTRA];     // extra/debug data
  RangeCalibData   RangeCalib[2];  // Range calibration point data
} ChanState;


//--- values for ChanState.calibState ---
#define  CALIB_START         1
#define  CALIB_START_RANGE   2
#define  CALIB_WAIT_SETTLE   3
#define  CALIB_QUIT         99


//
// Masks for BCM setting message State byte
//
#define SETSTATE_FILTER              0x01   // filter bit
#define SETSTATE_AVG_B0              0x02   // averaging selection bit 0
#define SETSTATE_AVG_B1              0x04   // averaging selection bit 1
#define SETSTATE_AVG_B2              0x08   // averaging selection bit 2
#define SETSTATE_TESTCUR             0x40   // test cur bit (SS FrontEnd only!)
#define SETSTATE_FRONTEND            0x80   // frontEnd bit


//
// Range-related values
//
#define LOW_RANGE_START                 8   // start of low-current ranges
#define FILTER_RANGE_START             10   // start of filter-always-on ranges
#define VOLTAGE_OVERRANGE_LIMIT      9.00   // in volts
#define LIMIT_READY_OVERRANGE        1.05   // % of VOLTAGE_OVERRANGE_LIMIT
#define LIMIT_ERROR_OVERRANGE        1.10   // % of VOLTAGE_OVERRANGE_LIMIT
#define LIMIT_OVERRANGE_AUTODROP     0.75   // % of full-scale current
#define MAX_OFFSET_FFV               0.80   // max allowed % of VOLTAGE_OVERRANGE_LIMIT


//
// BCM status word masks
//
#define CHAN_STATUS_FILTER      0x0001
#define CHAN_STATUS_RDY_NORM    0x0002
#define CHAN_STATUS_RDY_OVER    0x0004
#define CHAN_STATUS_INCAL       0x0008
#define CHAN_STATUS_ERR_OVER    0x0010
#define CHAN_STATUS_HW_ERR      0x0020
#define CHAN_STATUS_TESTCURR    0x0040
#define CHAN_STATUS_FRONTEND    0x0080
#define CHAN_STATUS_RANGEBAD    0x0100
#define CHAN_STATUS_AUTORANGED  0x0200
#define CHAN_STATUS_AVG_B0      0x0400
#define CHAN_STATUS_AVG_B1      0x0800
#define CHAN_STATUS_AVG_B2      0x1000
// added 2006-02-23:
#define CHAN_STATUS_RESET       0x2000 // chan needs INIT cmd


//
// Possible BCM over-range states
//
#define CHAN_OVERRANGE_OK            0   // not overrange
#define CHAN_OVERRANGE_READY         1   // overrange, but still valid
#define CHAN_OVERRANGE_ERROR         2   // overrange, reading likely bad


typedef struct {
  float          FSCurrent;     // Full Scale Current
  float          FSVoltage;     // Full Scale Voltage (determines gain)
  char           RelayBits;     // what relays to set for this range
} RangeInfo;

#define  MinChgTime     500  // Min # ms to wait for chg to take effect
#define  MaxNoChgTime  3000  // Max # ms to wait for no improvement in rdg

//
// Range information
// Pre-initialized, ZWorld will place it in simulated ROM
//
// NOTE:  The hardware supports only 8 physical ranges, but the older
//        BCM's had 15 ranges, so the software ones where thrown in
//        to mimic the older devices
//
//  The FSVoltage values here are default values only
//
const RangeInfo RangeData[NumRanges] = {
/* FSCurrent, FSVoltage, RelayBits */
  {   1E-03,    7.558,     0x04  },  // RANGE 0    1mA       K3
  { 300E-06,    2.266,     0x04, },  // RANGE 1  300uA       K3

  { 100E-06,    9.000,     0x08, },  // RANGE 2  100uA       K4
  {  30E-06,    2.951,     0x08, },  // RANGE 3   30uA       K4

  {  10E-06,    7.306,     0x10, },  // RANGE 4   10uA       K5
  {   3E-06,    2.193,     0x10, },  // RANGE 5    3uA       K5

  {   1E-06,    7.964,     0x00, },  // RANGE 6    1uA   <none>
  { 300E-09,    2.388,     0x00, },  // RANGE 7  300nA   <none>

  { 100E-09,    7.931,     0x06, },  // RANGE 8  100nA   K2 + K3
  {  30E-09,    2.380,     0x06, },  // RANGE 9   30nA   K2 + K3

  {  10E-09,    8.647,     0x0A, },  // RANGE 10  10nA   K2 + K4
  {   3E-09,    2.595,     0x0A, },  // RANGE 11   3nA   K2 + K4

  {   1E-09,    7.290,     0x12, },  // RANGE 12   1nA   K2 + K5
  { 300E-12,    2.819,     0x12, },  // RANGE 13 300pA   K2 + K5

  { 100E-12,    7.940,     0x02, },  // RANGE 14 100pA   K2
  {  30E-12,    2.384,     0x02, }   // RANGE 15  30pA   K2
};

//--- The real full-scale currents.  Defaults to values in the RangeData ---
//--- table, but a jumper can be used to multiply those values by 10 ---
float fsCurrent[NumRanges];

//
// Averaging table
// (entries must be between 1 and MAX_AVG_CYC)
//
//  # show * 10ms => interval to average over
#define  NumAvgIdxs  8
const unsigned short AvgSelTable[NumAvgIdxs] =
{
  10,   // 0 (000)
   1,   // 1 (001)
   2,   // 2 (010)
   5,   // 3 (011)
  20,   // 4 (100)  default
  30,   // 5 (101)
  50,   // 6 (110)
 100    // 7 (111)
};

#define  DefaultAvgIdx  4

//===== function prototypes for controller-specific functions =====

void frontEndChg(int chanNum);

void CHAN_Initialize(char chanNum);
void CHAN_TakeReading(char chanNum);
int  CHAN_GetStatus(char chanNum);

void CHAN_LoadCalibData(char chanNum);
bool CHAN_ValidCalibData(char chanNum);
void CHAN_ResetCalibData(int chan, int range);
#define  ALL_RANGES  (-1)

void CHAN_RangeCalibrationPoint(char chanNum, int PointNum, float Current);

void CHAN_SaveRTZeroOffsets(char chanNum);
void CHAN_SaveFECalib(char chanNum);


void ShowCalibData(int chan);

void rtzCalib_tick(int chanNum);



//=============================================================================
//  GLOBAL variables
//=============================================================================

ApplData  applData;  // for storing application specific config data


//==================== device specific stuff ====================
uchar  bcm1_in, bcm1_out, bcm2_in, bcm2_out;
ChanState        CHAN[2];      // BCM channel data


xstring DevStateStr {
"------------ Chan 0 ------------        ------------ Chan 1 ------------",
"Range:    .. (.........) ...            Range:    .. (.........) ...",
"FrontEnd: ...                           FrontEnd: ...",
"Filter:   ..                            Filter:   ..",
"State:    .....                         State:    .....",
"Voltage: ##.###     (###.##%var)        Voltage: ##.###     (###.##%var)",
"Current: ##.###E-## (###.##%FS)         Current: ##.###E-## (###.##%FS)",
"Avging:   [#] ### samples (#### ms)     Avging:   [#] ### samples (#### ms)",
"Source:   ........                      Source:   ........",
""
};

const DigFmtInfo digInFmt[] = {
  { -1,-1, "",0, "",0 }
};

const DigFmtInfo digOutFmt[] = {
  { -1,-1, "",0, "",0 }
};

const AnalogFmtInfo  adcFmt[] = {
  { -1,-1, 0,0 }
};

const AnalogFmtInfo  dacFmt[] = {
  { -1,-1, 0,0 }
};



xstring ApplStrs  {
#define   HelpStr                 ApplStr(0)
 #ifdef BCM_SS
"0/1, +/-, A-dc, C-ancel, E-xt, F-lt, H-elp, I-nit, T-estcur, eX-tcmd",
 #else
"0/1, +/-, A=Adc, C=Cancel, E=Ext, F=Flt, H=Help, I=Init, X=eXtcmd",
 #endif

#define   ExtHelpStr              ApplStr(1)
  "Bye, Calib, Help, ID, IP, Reboot, Reg, Reset, Serial, Track, <reg>",

#define   LogParamsMsg            ApplStr(2)
  "Logging Enabled: Interval: %ldms, #Samples: %ld.  Press ESC to begin",

#define   ResetApplDataMsg        ApplStr(3)
  "Resetting Application data...",

#define   MsgErrForDestMsg        ApplStr(4)
  "msg err: %c for dest: %d",

#define   CmdParserErrMsg         ApplStr(5)
  "cmd parser error",

#define   ResetCalibDataMsg       ApplStr(6)
  "Reset calib data for chan %d",

#define   UnrecogCmdMsg           ApplStr(7)
  "Unrecognized command: %s",

#define   InvalidCalibOptionMsg   ApplStr(8)
  "Invalid calib option: %s",

#define   ClearOffsetsMsg       CtlString(9)
  "Clearning Runtime offsets for chan %d",

#define   InvalidChanNumMsg       ApplStr(10)
  "Invalid chan number: %s",

#define   CalibDataMsg            ApplStr(11)
  "BCM calibration data for channel %d:\r\n",

#define   CalibHeader1            ApplStr(12)
  "Run-time    ---Bench Calibration---",

#define   CalibHeader2            ApplStr(13)
  "Offset      Offset    Gain(A/V)  OK?",

#define   ESCToReturnMsg          ApplStr(14)
  "Press ESC to return to status screen",

#define   InvalidRangeMsg       CtlString(15)
  "Invalid range #: %d",

#define   CopiedRangeDataMsg    CtlString(16)
  "Copied Range %d data to all ranges",

#define   InvalidLogOptionMsg     ApplStr(17)
  "Invalid option for log command: %s",

#define   InvalidLogIntervalMsg   ApplStr(18)
  "Log interval must be > 0: %s",

#define   InvalidLogCountMsg      ApplStr(19)
  "Log count must be > 0: %s",

#define   LoggingDisabledMsg      ApplStr(20)
  "Logging is Disabled",

#define   IncompleteCmdMsg        ApplStr(21)
  "Incomplete command",

#define   ApplyCmdsToChanMsg      ApplStr(22)
  "Apply commands to chan %d",

#define   EndingCalibMsg          ApplStr(23)
  "Ending calibration for channel %d",

#define   CalibBadSuffix          ApplStr(24)
  ": Calib Data bad",

#define   SavingCalibDataMsg      ApplStr(25)
  "Saving calibration data",

#define   ErrOrRangeToSmallMsg    ApplStr(26)
  "Calibration error or Range too small for calibration points\n",

#define   CalibMsg1             CtlString(27)
  "%2d/%d: Var:%4.2f%%  Offset:%6.3fV/%sA, (%.2f%%FS)",

#define   CalibPointDataMsg     CtlString(28)
  "%2d/%d: CalibPoint %d: chan %d: (%7.3f volts, %s amps)",

#define   SavingRTZeroDataMsg     ApplStr(29)
  "Saving RunTime zero offsets",

#define   CalibHeader0            ApplStr(30)
  "-------------- In Use --------------   ---------- From Flash --------------",

#define   CalibHeader3            ApplStr(31)
  "------------------------------------",

#define   CurSettings           CtlString(32)
  "Chan %d:  Rng:%02d  Fltr:%s  FrEnd:%s  TestCur:%s  AvgIdx:%d",

#define   StartCalibMsg         CtlString(33)
  "--------- Starting Calibration of Channel %d --------",

""
};   //NewStr


xstring ApplHelpStrs {
  "Calib 1|2   0|1 amps         - Set FrontEnd calib point (do point 1 first)",
  "Calib Clear 0|1              - Set RT offsets to 0, mark all ranges good",
  "Calib Dup   0|1 range        - set all ranges calib data == to specified one",
  "Calib Reset 0|1            *** WARNING: Resets BENCHTOP calib for a chan!!!",
  "Calib Save  0|1              - save FrontEnd calib data for a chan to FLASH",
  "Calib Show  0|1              - show FrontEnd calib data by range for a chan",
  "Calib Zero  0|1              - Calib so each range reads 0 for current input",
  "Log Off/On [ms [count]]      - Disable/Enable streaming data output",
  ""
};


#use "Base_DLP_C.lib"


//=============================================================================
//         Standard functions with APPLICATION SPECIFIC variations
//=============================================================================

//-----------------------------------------------------------------------------
//  Initialization for processing of commands
//-----------------------------------------------------------------------------
int InitCmds(int chanNum)
{
  return 0;
}

//-----------------------------------------------------------------------------
//  Put device in state that is safe for controller to reboot
//-----------------------------------------------------------------------------
void PrepForReboot()
{
  ulong  lastWriteTime;


//... make changes to digitalOut, voltsOut[], etc ...

//--- wait for latest changes to take effect ---
  lastWriteTime = writeTime;  while (writeTime == lastWriteTime);
}


//-----------------------------------------------------------------------------
//  check validity of application/configuration data.  If any of it is not
//  usuable, change it to a reasonable default value and save any changes
//  to flash memory
//-----------------------------------------------------------------------------
void ValidateApplData(bool readFirst)
{
  bool  dataChgs;
  int  i, stat;
  char  *msg;


  msg = getBuf(90);

  dataChgs = false;

  if (readFirst)
    if ((stat = Block_Read(BID_APPL_DATA, 0, &applData, sizeof(applData))) < 2)
      sprintf(msg, ApplDataReadErrMsg, stat);

  if (applData.devType != DeviceType)  {
    LogMsg(ResetApplDataMsg);
    Block_SetSize(BID_APPL_DATA, sizeof(applData));
    memset(&applData, 0, sizeof(applData));  dataChgs = true;
    applData.vers = ApplDataVers;  applData.devType = DeviceType;
    for (i=0; i<NumChans; ++i)  CHAN_ResetCalibData(i, ALL_RANGES);
  }

#define CheckRange(param, minVal, maxVal, defValue)  \
  { if ((applData.param < (minVal)) OR (applData.param > (maxVal)))  {  \
    applData.param = (defValue);  dataChgs = true;  } }


//  CheckRange( param, minVal, maxVal, defaultVal );

//  if (applData.param not valid)  {
//    applData.param = defaultValue;
//    dataChgs = true;
//  }

//--- application specific checks (for BCM, the result should NOT alter what ---
//--- is in flash) ---
  for (i=0; i<NumChans; ++i)  CHAN_ValidCalibData(i);

// search for _PERSISTENT_ to see where code normally added for saved values


  freeTo(msg);

  if (dataChgs)  SaveApplData(true);
}

#define MAX_ARGS  8
//-----------------------------------------------------------------------------
//  parse a string into a list of arguments
//-----------------------------------------------------------------------------
void parseArgs(char *args, int *argc, char *argv[MAX_ARGS])
{
  char *ptr;
  short  i;


  ptr = args;  *argc = 0;  argv[0] = NULL;

//--- parse it into standard argc, argv ---
  for (*argc=0; *argc<MAX_ARGS; )  {
    while (*ptr AND (*ptr <= ' '))  ++ptr;
    if (! *ptr)  break;
    argv[*argc] = ptr;  ++(*argc);
    while (*ptr > ' ')  ++ptr;
    if (*ptr)  { *ptr = 0;  ++ptr; }
  }

}


//-----------------------------------------------------------------------------
//  process a multi-character command from the diagnostic terminal/console
//-----------------------------------------------------------------------------
void ProcessStrCmd(char *cmd, LinkProc *lp)
{
  char  *ptr;
  short  i, idx, state, chanNum;
  int  showChan, saveChan, range, argc;
  bool  validCmd, partialCmd;
  char  *msg, *temp, *args;
  char  *argv[MAX_ARGS];
  EventData  eventData;


  StackPtr();

  msg = getBuf(90 + 90 + 90);  temp = msg + 90;
  args = getBuf(strlen(cmd)+1);  strcpy(args, cmd);

  parseArgs(args, &argc, argv);


  validCmd = partialCmd = false;  strcpy(msg, CmdParserErrMsg);

  eventData.chanNum = cmdChan;

#define  arg(str)  (strcmpi(ptr, (str)) == 0)

  do  {
    if (argc < 2)  break;  // all cmds have at least 1 arument
    ptr = argv[0];


    if (arg("calib"))  {

    //--- make sure 3rd argument is a valid chan # ---
      if (argc < 3)  { partialCmd = true;  break; }
      ptr = argv[2];
      if (! (arg("0") OR arg("1")))  { strcpy(msg, "Invalid chan #");  break; }
      eventData.chanNum = chanNum = atoi(ptr);

      ptr = argv[1];

      if (arg("abort"))  {  // calib abort 0|1
        EVENT(AbortCalib);  validCmd = true;  *msg = 0;  break;
      }  else

      if (arg("clear"))  {  // calib clear 0|1
        EVENT(ClearOffsets);  validCmd = true;  *msg = 0;  break;
      }  else

      if (arg("dup"))  {  // calib dup 0|1 range
        if (argc < 4)  { partialCmd = true;  break; }
        eventData.intVal = atoi(argv[4]);  EVENT(DupRange);
        validCmd = true;  *msg = 0;  break;
      }  else

      if (arg("reset"))  {  // calib reset 0|1
        sprintf(msg, ResetCalibDataMsg, cmdChan);
        CHAN_ResetCalibData(chanNum, ALL_RANGES);  validCmd = true;  break;
      }  else

      if (arg("save"))  { // calib save 0|1
        EVENT(SaveCalib);  validCmd = true;  *msg = 0;  break;
      }  else

      if (arg("show"))  {  // calib show 0|1
        ShowCalibData(chanNum);  validCmd = true;  *msg = 0;  break;
      }  else

      if (arg("zero"))  {  // calib rtz 0|1
         EVENT(StartCalib);  validCmd = true;  *msg = 0;  break;
      }  else

      if (arg("1") OR arg("2"))  {  // calib 1|2  0|1  amps
        if (argc < 4)  { partialCmd = true;  break; }
        eventData.intVal = atoi(argv[1]);
        eventData.floatVal = atof(argv[3]);
        EVENT(FECalibPoint);
        validCmd = true;  *msg = 0;  break;
      }

    }  else


    if (arg("log"))  {  // log on|off
      if (argc < 2)  { // log cmd w/o any parameters
        if (logData)
          sprintf(msg, LogParamsMsg, logInterval*10L, logCount);
        else
          strcpy(msg, LoggingDisabledMsg);
        validCmd = true;  break;
      }
      ptr = argv[1];

      if (arg("off"))  {
        strcpy(msg, "Logging disabled.");
        logData = false;  validCmd = true;  break;
      }

      if (arg("on"))  { // log on [<ms> [<count>]]
        if (argc < 3)  {  // log on without any optional parameters
          sprintf(msg, LogParamsMsg, logInterval*10L, logCount);
          logData = true;  sampleCount = 0;  validCmd = true;  break;
        }
        logInterval = atol(argv[2])/10L;
        if (logInterval <= 0)  {
          sprintf(msg, InvalidLogIntervalMsg, argv[2]);  break; }
        if (argc > 3)  {
          logCount = atol(argv[3]);
          if (logCount < 0)  {
            sprintf(msg, InvalidLogCountMsg, argv[3]);  break;  }
          sprintf(msg, LogParamsMsg, logInterval*10L, logCount);
          logData = true;  sampleCount = 0;  validCmd = true;  break;
        }
      }
      strcpy(msg, "Invalid log option");   break;
    }

  }  while (false);

  strCmdState = false;


  if (partialCmd)  strcpy(msg, IncompleteCmdMsg);

  if (! validCmd)
    ProcessStrCmdCommon(cmd, msg, lp);  // not dev cmd - check generic ones
  else
    if (*msg)
      if (StreamDisp)
        ShowMsg(msg);
      else  {
        respShownTime = MS_TIMER - 5000;
        DispStr(0,23, msg, true);
        respShownTime = MS_TIMER;
      }

  freeTo(msg);
}


//-----------------------------------------------------------------------------
//  process single-char commands from the diagnostic console
//-----------------------------------------------------------------------------
void CharCmd(char cmd)
{
  char  origCmd;
  char  *msg;
  EventData  eventData;
  ChanState  *chan;


  msg = getBuf(90);

  origCmd = cmd;  cmd |= 0x20;

  eventData.chanNum = cmdChan;  chan = CHAN + cmdChan;

  switch (cmd)  {
    case '+':  EVENT(RangeUp);     break;
    case '-':  EVENT(RangeDown);   break;
    case 'c':  EVENT(AbortCalib);  break;
    case 'i':  EVENT(Init);        break;

    case '0':  case '1':
      cmdChan = cmd - '0';
      sprintf(msg, ApplyCmdsToChanMsg, cmdChan);  ShowMsg(msg);
      break;

    case 'e':  // toggle FrontEnd On/Off
      if (chan->frontEndOn)  EVENT(FrontEndOff);  else  EVENT(FrontEndOn);
      break;

    case 'f':  // toggle Filter On/Off
      if (chan->filterOn)  EVENT(FilterOff);  else  EVENT(FilterOn);
      break;

#ifdef BCM_SS
    case 't':  // toggle TestCurrent On/Off
      if (chan->testCurOn)  EVENT(TestCurOff);  else  EVENT(TestCurOn);
      break;
#endif
  }

  freeTo(msg);  CharCmdCommon(origCmd);
}


//-----------------------------------------------------------------------------
//  get next msg in event msg queue - returns true if a msg read, false if
//  no msg or an error occured
//-----------------------------------------------------------------------------
bool getMsg(EventData *eventData)
{
  if (Buf_Used(&eventMsgQueue) < 1) return false;

//  if (lastMsgProcessed == lastMsgToProcess)  return false;

  if (Buf_Remove(&eventMsgQueue, (char *)eventData, sizeof(EventData))
      != sizeof(EventData))  { LogErr(EventMsgErr);  return false;  }

//  lastMsgProcessed = eventData->seqNum;

  return true;
}

//-----------------------------------------------------------------------------
//  call to add an event msg to the queue
//-----------------------------------------------------------------------------
void putMsg(int event, EventData *eventData)
{
  eventData->event = event;

  if (Buf_Add(&eventMsgQueue, (char *)(eventData), sizeof(EventData))
      != sizeof(EventData))   LogErr("Error queing Event Msg");
}



//-----------------------------------------------------------------------------
//  Update data based on digital and analog input values and determine new
//  values for the digital and analog outputs
//
//  !!!!! WARNING !!!!!
//
//  This function is called by the timer-interrupt driven function that reads
//  and writes the Analog and Digital I/O values.  It should be limited to
//  processing current input values and updating output values that will become
//  the active values on the next interrupt (currently, this interuupt occurs
//  every 10ms, so this routine must take considerably less time than that to
//  do EVERYTHING it needs to).
//-----------------------------------------------------------------------------
nodebug void UpdateDeviceState()
{
  int  i, rate, numSamples;
  ulong  secTime, mask;
  char  *dataStr;
  EventData  eventData;


  bcm1_in = bcm2_in = 0;

  if (!(digitalIn & BCM1_TEST_CUR_OFF))  bcm1_in = TEST_SW_ON;
  if (!(digitalIn & BCM2_TEST_CUR_OFF))  bcm2_in = TEST_SW_ON;


//--- update readings and give any calibration process some time ---
  for (i=0; i<NumChans; ++i)  {
    CHAN_TakeReading(i);

    current[i] = CHAN[i].avgCur;
    rangeIdx[i] = CHAN[i].range;
    fullScale[i] = fsCurrent[CHAN[i].range];
    avgingIdx[i] = CHAN[i].avgSizeIdx;
    avgingSize[i] = AvgSelTable[avgingIdx[i]];
    avgInterval[i] = avgingSize[i] * 10;  // 10ms per sample
    chanStatus[i] = CHAN_GetStatus(i);
    extra1[i] = CHAN[i].Extra[0];
    extra2[i] = CHAN[i].Extra[1];
    extra3[i] = CHAN[i].Extra[2];
    extra4[i] = CHAN[i].Extra[3];
    extra5[i] = CHAN[i].Extra[4];
    extra6[i] = CHAN[i].Extra[5];
    extra7[i] = CHAN[i].Extra[6];
    variance[i] = CHAN[i].variance;
    volts[i] = voltsIn[i];

#ifdef BCM_RR
    CHAN[i].testCurOn = BOOL(*CHAN[i].InputByte & TEST_SW_ON);
#endif

    if (CHAN[i].calibState)  rtzCalib_tick(i);
  }

//--- simple "logging" of data - just spit out to the telnet client ---
  if (logData AND StreamDisp)  {
    if (++sampleCount == logInterval)  {
      sampleCount = 0;
      dataStr = getBuf(80);
      sprintf(dataStr, "DATA: %lu,0x%08lx,0x%08lx\r\n",
              MS_TIMER, CHAN[0].avgCur,  CHAN[1].avgCur );
      Buf_Add(&intBuf, dataStr, strlen(dataStr));
      freeTo(dataStr);
      if (logCount > 0)  if (--logCount <= 0)  logData = false;
    }
  }



//--- process all event msgs ---
  while (getMsg(&eventData))
    if ((eventData.event >= 0) AND (eventData.event <= LastMsgNum))
      if ((eventData.chanNum >= 0) AND (eventData.chanNum <= 1))
        (eventHandler[eventData.event])(&eventData);

//--- update # of samples being averaged ---
  for (i=0; i<NumChans; ++i)  {
    numSamples = AvgSelTable[CHAN[i].avgSizeIdx];
    if (adcAvgCount[i] != numSamples)  {
      adcAvgCount[i] = numSamples;  newAvgCount[i] = true;  }
  }


//--- update output values ---
  digitalOut = ( (ulong)bcm1_out << 1)
              | ((ulong)bcm2_out << 10)
              | (((ulong)(bcm2_out & TESTCUR_BIT)) << 3);
}

//-----------------------------------------------------------------------------
//  Do anything specific to this controller that is not time-ciritical.
//  This is called once during each pass of the ComLoop() function
//-----------------------------------------------------------------------------
void BackGroundProcess()
{
  int  i;


  if (recursiveBackGround)  return;
  recursiveBackGround = true;


  for (i=0; i<NumChans; ++i)  {
    if (CHAN[i].saveRTZOffsets)  CHAN_SaveRTZeroOffsets(i);
    if (CHAN[i].saveFECalib)  CHAN_SaveFECalib(i);
  }

  recursiveBackGround = false;
}


//-----------------------------------------------------------------------------
//  Do device specific formatting of current state data that cannot be
//  handled by entries in generic table for I/O values
//
//  Returns:
//      true:  last portion of data has been displayed
//      false:  more data to display during next pass
//-----------------------------------------------------------------------------
bool ShowDeviceState()
{
  char  *ptr;
  static int  i, ci, range, col;
  static float  cur, fs, percentFS;
  static bool  frontEndOn, filterOn, testCurOn, rangeOK, doingCalib;
  static ChanState  *chan;
  static RangeInfo  *rd;
  static int  showCount, devShowState, avgIdx, numSamples, calibState;
  char  *msg, *temp;


  msg = getBuf(90 + 32);  temp = msg + 90;

  switch (++devShowState)  {

    case 1:
      ci = showCount & 1;  col = 10 + ci*40;
      chan = CHAN + ci;
      range = chan->range;
      rd = RangeData + range;

      rangeOK = applData.calibData[ci][range].rangeOK;
      cur = chan->avgCur;
      percentFS = fabs(cur / fsCurrent[range] * 100.0);
      doingCalib = chan->calibState;
      frontEndOn = chan->frontEndOn;
      filterOn = chan->filterOn;
      testCurOn = chan->testCurOn;
      avgIdx = chan->avgSizeIdx;  numSamples = AvgSelTable[avgIdx];
      calibState = chan->calibState;
      break;

    case 2:
      sprintf(msg, "%02d (%s) %s",
         range, Engineering(fsCurrent[range], 4, temp), rangeOK ? "OK ":"BAD");
      DispStr(col,4, msg, false);
      break;

    case 3:
      DispStr(col,5, frontEndOn ? "ON " : (blinkOff ? "   " : "OFF"), false);
      DispStr(col,6, filterOn ? "ON " : "OFF", false);
      break;

    case 4:
      ptr = (doingCalib ? (blinkOff ? "     " : "CALIB") : "RUN  ");
      DispStr(col,7, ptr, false);

      DispStr(col-1,8, fmtFloat(voltsIn[ci],6,3,temp), false);
      DispStr(col+11,8, fmtFloat(chan->variance,6,2,temp), false);
      break;

    case 5:
      DispStr(col-1,9, Engineering(cur,4,temp), false);
      break;
    case 6:
      DispStr(col+11,9, fmtFloat(percentFS,6,2,temp), false);
      break;

    case 7:
      sprintf(msg, "[%d] %3d samples (%4d ms)",
                  avgIdx, numSamples, numSamples * 10 );
      DispStr(col,10, msg, false);

      *temp = 0;
      if (frontEndOn)  strcpy(temp, "EXTERN");
      if (testCurOn AND !blinkOff)  {
        if (*temp)  strcat(temp, "+");
        strcat(temp, "TEST");
      }
      if (! *temp)  if (!blinkOff)  strcpy(temp, "NONE");
      strcat(temp, "        ");
      DispStr(col,11, temp, false);

      DispStr(col,12,
         ((!calibState OR blinkOff) ? "           " : "CALIBRATING"), false);


      if (showCount & 1)  {
        sprintf(msg, "Cmd Chan: %d  ", cmdChan);
        DispStr(0,20, msg, false);
      }

      break;

  //--- to catch all non-active state values ---
    default:
      devShowState = 0;  break;
  }

  freeTo(msg);

  return ((devShowState == 0) AND !(++showCount & 1));
}


//-----------------------------------------------------------------------------
//  Determine what, if any, action should be taken as a result of one of the
//  coil values being changed
//-----------------------------------------------------------------------------
void NewCoilValue(int chanNum, int coilNum, int coilVal, int cmdRow)
{
  int  newRange, newIdx;
  char  *msg;
  EventData  eventData;


  StackPtr();

  msg = getBuf(90);

  sprintf(msg, "set coil [%05d][%d] = %d", coilNum+1, chanNum, coilVal);
  if (GenericDisp)  DispStr(CmdCol,cmdRow, msg, true);
  LogMsg(msg);

  if ((coilNum < 0) OR (coilNum >= NUMCOILS))  { freeTo(msg);  return; }
  if ((chanNum < 0) OR (chanNum >= NumChans))  { freeTo(msg);  return; }


  eventData.chanNum = chanNum;

  if (coilVal)  switch (coilNum)  {
    case DoReboot:     Reboot(10);           break;
    case DoInit:       EVENT(Init);          break;
    case AbortCalib:   EVENT(AbortCalib);    break;
    case DoCalibrate:  EVENT(StartCalib);    break;
    case RangeUp:      EVENT(RangeUp);       break;
    case RangeDown:    EVENT(RangeDown);     break;
    case AvgIdxUp:     EVENT(AvgIdxUp);      break;
    case AvgIdxDown:   EVENT(AvgIdxDown);    break;
    case FrontEndOn:   EVENT(FrontEndOn);    break;
    case FrontEndOff:  EVENT(FrontEndOff);   break;
    case FilterOn:     EVENT(FilterOn);      break;
    case FilterOff:    EVENT(FilterOff);     break;
    case ClearOffsets: EVENT(ClearOffsets);  break;
#ifdef BCM_SS
    case TestCurOn:    EVENT(TestCurOn);     break;
    case TestCurOff:   EVENT(TestCurOff);    break;
#endif
    case SaveCalib:    EVENT(SaveCalib);     break;  // added 03/20/06
  }

  coilVal = 0;

  freeTo(msg);

  ++criticalSection;
  ClearBit(coils[chanNum], coilNum);
  --criticalSection;
}

//-----------------------------------------------------------------------------
//  Determine what, if any, action should be taken as a result of one of the
//  registers being written to.
//
//  Called by the generic function NewRegValue() in the base code which is
//  responsible for actually copying the <regName>_New value to <regName>
//
//  Returns true if the new value should be used, false if it should not
//
//  If any value used during execution of the periodic ControlLoop() function
//  (besides the new register value) is altered here, you MUST increase the
//  criticalSection counter 1st (it will be restored to its previous value
//  after the new register value is in effect).  This is to avoid having the
//  ControlLoop() processes act on unsynchronized values.
//
//    Example:  DevNewRegValue() alters some non-register value in response
//    to a new register value.  Since the new register value does not go
//    into effect until DevNewRegValue() returns, any such values used within
//    ControlLoop() would be out of sync with the current register value.
//
//  NOTE:  Assumes that, when a register that is part of a larger data type
//         (e.g. a 32-bit float or long integer) that the lower numbered
//         registers will be written first.
//
//  Any string in *msg will be logged after returning from this func.
//-----------------------------------------------------------------------------
bool  DevNewRegValue(int chanNum, int regNum, int regVal,
                     int clientType, LinkProc *lp, char *msg)
{
  int  cmdRow, chStatus;
  EventData  eventData;
  bool  frontEndOn, filterOn, testCurOn;
  ChanState  *chan;


  cmdRow = CmdRow(LinkProc_Idx(lp));

  eventData.chanNum = chanNum;  chan = CHAN + chanNum;

//--- Check the New register value (or set of registers comprising a ---
//--- single value) and either reject it (in which case the new value ---
//--- will be ignored), or do whatever processing is necessary using the ---
//--- new value.  NOTE: The new value does not actually go in to effect ---
//--- until sometime after this function returns (one of the reasons that ---
//--- proper handling of the criticalSection counter is required). ---

  do  {

//  if (regNum == regName_Idx)  {
//    if (regName_New[chanNum] not usable)  {
//      sprintf(msg, "optional msg if default 'ignored' msg insufficient");
//      return false;
//    }
//    ... whatever else needs to be done with regName_New value ...
//    sprintf(msg, "...optional msg about action taken...");
//    (NOTE: regName[chanNum] = regName_New is handled by caller of this func
//    break;
//  }


    if (regNum == rangeIdx_Idx)  {
      eventData.intVal = rangeIdx_New[chanNum];  EVENT(NewRange);  break; }

    if (regNum == avgingIdx_Idx)  {
      eventData.intVal = avgingIdx_New[chanNum];  EVENT(NewAvgIdx);  break; }

  //--- writes to chanStatus used for commands but do not actually change ---
  //--- the value of chanStatus ---
    if (regNum == chanStatus_Idx)  {
      chStatus = chanStatus_New[chanNum];

      frontEndOn = (chStatus & SETSTATE_FRONTEND);
      filterOn = (chStatus & SETSTATE_FILTER);
      testCurOn = (chStatus & SETSTATE_TESTCUR);

      if (BOOL(frontEndOn) != BOOL(chan->frontEndOn))
        if (frontEndOn)  EVENT(FrontEndOn);  else  EVENT(FrontEndOff);

      if (BOOL(filterOn) != BOOL(chan->filterOn))
        if (filterOn)  EVENT(FilterOn);  else  EVENT(FilterOff);

#ifdef BCM_SS
      if (BOOL(testCurOn) != BOOL(chan->testCurOn))
        if (testCurOn)  EVENT(TestCurOn);  else  EVENT(TestCurOff);
#endif
      chanStatus_New[chanNum] = chanStatus[chanNum];
      break;
    }

    if (regNum == calibPoint1_Idx)  {
      eventData.intVal = 1;  eventData.floatVal = calibPoint1_New[chanNum];
      EVENT(FECalibPoint);  break;
    }

    if (regNum == calibPoint2_Idx)  {
      eventData.intVal = 2;  eventData.floatVal = calibPoint2_New[chanNum];
      EVENT(FECalibPoint);  break;
    }


// search for _PERSISTENT_ to see where code normally added for saved values

  } while (false);

  return true;
}

//---------------------------------------------------------------------------
//  Multiple calibration processes can be supported simply by having a
//  different starting state for each one.
//
//  The WaitState() macro (and skeleton CALIB_WAIT state) are supplied
//  for cases when you need to do an asynchronous "wait" for some event
//  (without stopping the main cooperative multitasking process loop)
//---------------------------------------------------------------------------
#define  WaitState(minMS, maxMS)  {  \
    returnState = calibState;  calibState = CALIB_WAIT;  \
    minWaitMS = (minMS);  maxWaitMS = (maxMS);  waitStart = MS_TIMER;  \
}

void calib_tick()  {  }  // dummy to satisfy older ComLoop() called method

//-------------------------------------
void rtzCalib_tick(int chanNum)
{
  int  i, baseRange;
  ulong  waited;
  ChanState  *chan;
  FrontEndCalibData  *fe;
  char  *msg, *temp;


  msg = getBuf(80+30);  temp = msg + 80;

  chan = CHAN + chanNum;

  switch (chan->calibState)  {

    case CALIB_START:  // prep for RunTime Zero-level offset calibration
      chan->saveRange = chan->range;
      chan->saveFilterOn = chan->filterOn;
      chan->saveAvgSizeIdx = chan->avgSizeIdx;

//      chan->range = 0;  chan->filterOn = false;
      chan->range = chan->range & 0xfe;  // ensure even-#'d range
      chan->avgSizeIdx = DefaultAvgIdx;
#ifdef BCM_SS
      chan->testCurOn = false;
#endif
      fe = applData.calibData[chanNum];
      for (i=0; i<NumRanges; ++i)  fe[i].RTOffset = 0.0;
      sprintf(msg, StartCalibMsg, chanNum);
      LogMsg(msg);
      chan->calibState = CALIB_START_RANGE;
      break;

    case CALIB_START_RANGE:
      frontEndChg(chanNum);
      chan->calibState = CALIB_WAIT_SETTLE;
      break;

    case CALIB_WAIT_SETTLE:
    //--- don't continue until some history for latest Front End settings ---
      if (elapsedMS(chan->settlingStart) < MinChgTime)  break;
      chan->calibState = CALIB_WAIT_SETTLE;

    //--- keep waiting as long as there is improvement ---
      if (chan->variance < chan->lowestVariance)  {
        chan->lowestVariance = chan->variance;
        chan->settlingStart = MS_TIMER;  break;
      }

    //--- wait at least 2 seconds from last improvement ---
      if (elapsedMS(chan->settlingStart) < 2000)  break;

    //--- reading did settle - update the run-time (RAM) copy for both ---
    //--- the hardware and software ranges (and save the on/off state of ---
    //--- the FrontEnd with it)---
      fe = applData.calibData[chanNum] + chan->range;
      fe->RTOffset = chan->avgVolts - fe->BCZeroOffset;
      fe->inputOn = chan->frontEndOn;
      fe[1].RTOffset = fe->RTOffset;  fe[1].inputOn = fe->inputOn;

      sprintf(msg, CalibMsg1,
              chan->range, chanNum, chan->variance, chan->avgVolts,
              Engineering(chan->avgCur, 4, temp),
              chan->fracFSVolts * 100.0 );
      LogMsg(msg);

    //--- move on to next range (or end calibration if no more) ---
      if ((chan->range += 2) > 15)  chan->range = 0;
      if (chan->range != (chan->saveRange & 0xfe))  {
        chan->calibState = CALIB_START_RANGE;  break; }

    //--- all done - save the results to flash and restore the range to ---
    //--- what it was before we started ---
      chan->saveRTZOffsets = true;
      chan->calibState = CALIB_QUIT;
      break;


  //--- finished or aborted - put back settings from before the calib ---
    case CALIB_QUIT:
      chan->range = chan->saveRange;
      chan->filterOn = chan->saveFilterOn;
      chan->avgSizeIdx = chan->saveAvgSizeIdx;
      chan->avgCur = 0.0;
      frontEndChg(chanNum);
      chan->calibState = CALIB_IDLE;
  }

  freeTo(msg);
}


//=============================================================================
//                 Event Handler Functions for this application
//=============================================================================
//  WARNING:  Do NOT call these functions directly from within the ComLoop()
//            process.  These functions are ONLY to be called by the
//            ControlLoop() process.  To execute one of these functions from
//            the ComLoop() process, send the corresponding msg to the
//            ControlLoop() process and IT will call the function when it is
//            safe to do so.
//=============================================================================

//-----------------------------------------------------------------------------
//  settings done for any change to state of Front End
//-----------------------------------------------------------------------------
void frontEndChg(int chanNum)
{
  char outByte;
  char  *msg, *s1, *s2, *s3;
  ChanState  *chan;


  chan = CHAN + chanNum;


  msg = getBuf(60);
  s1 = (chan->filterOn ?   "ON " : "OFF");
  s2 = (chan->frontEndOn ? "ON " : "OFF");
  s3 = (chan->testCurOn ?  "ON " : "OFF");
  sprintf(msg, CurSettings,
          chanNum, chan->range, s1, s2, s3, chan->avgSizeIdx);
  LogMsg(msg);  freeTo(msg);

//--- start with the range bits ---
  outByte = RangeData[chan->range].RelayBits;

//--- always turn filter ON if set range is >= FILTER_RANGE_START ---
//--- otherwise use user-requested filter setting ---
  if (chan->range >= FILTER_RANGE_START)  chan->filterOn = true;

//--- OR in bit if filter should be OFF (it's reversed, 1=OFF, 0=ON) ---
  if (! chan->filterOn)  SetBits(outByte, FILTER_BIT);

//--- OR in bit if Front End should be ON ---
  if (chan->frontEndOn)  SetBits(outByte, FRONTEND_BIT);

//--- OR in bit if Test Cur should be ON ---
#ifdef BCM_SS
  if (chan->testCurOn)  SetBits(outByte, TESTCUR_BIT);
#endif

//--- copy to value to be written back out to hardware ---
  (*(chan->OutputByte)) = outByte;

//--- reset parameters that affect other processing logic ---
  chan->settlingStart = MS_TIMER;
  chan->lowestVariance = chan->variance = 100.0;
  chan->numSamples = 0;
}


//-----------------------------------------------------------------------------
//  Do start-up initialization
//-----------------------------------------------------------------------------
void Init_(EventData *event)
{
  ChanState  *chan;

  if ((chan = CHAN + event->chanNum)->calibState)  return;

  CHAN_Initialize(event->chanNum);
}

//-----------------------------------------------------------------------------
//  Initiate a Run Time calibration of the zero-level offsets (i.e. determine
//  the voltage reading for each range for the current input and use that
//  as an offset so each range will read 0 for the existing input level)
//-----------------------------------------------------------------------------
void StartCalib_(EventData *event)
{
  ChanState  *chan;

  if ((chan = CHAN + event->chanNum)->calibState)  return;

  chan->calibState = CALIB_START;
}

//-----------------------------------------------------------------------------
//  Abort a calibration process started by the StartCalib() event
//-----------------------------------------------------------------------------
void AbortCalib_(EventData *event)
{
  if (CHAN[event->chanNum].calibState)
    CHAN[event->chanNum].calibState = CALIB_QUIT;
}

//-----------------------------------------------------------------------------
//  Process the given datapoint for the purpose of determining the gain and
//  offset for the specified range on one of the Front-End boards.
//-----------------------------------------------------------------------------
void FECalibPoint_(EventData *event)
{
  ChanState  *chan;

  if ((chan = CHAN + event->chanNum)->calibState)  return;

  CHAN_RangeCalibrationPoint(event->chanNum, event->intVal, event->floatVal);
}

//-----------------------------------------------------------------------------
//  Change the range to the next most sensitive one
//-----------------------------------------------------------------------------
void RangeUp_(EventData *event)
{
  ChanState  *chan;

  if ((chan = CHAN + event->chanNum)->calibState)  return;

  if (chan->range >= 15)  return;

  ++chan->range;  frontEndChg(event->chanNum);
}

//-----------------------------------------------------------------------------
//  Change the range to the next least sensitive one
//-----------------------------------------------------------------------------
void RangeDown_(EventData *event)
{
  ChanState  *chan;

  if ((chan = CHAN + event->chanNum)->calibState)  return;

  if (chan->range <= 0)  return;

  --chan->range;  frontEndChg(event->chanNum);
}

//-----------------------------------------------------------------------------
//  Change the range to specified one
//-----------------------------------------------------------------------------
void NewRange_(EventData *event)
{
  int  range;
  ChanState  *chan;


  if ((chan = CHAN + event->chanNum)->calibState)  return;

  range = event->intVal;
  if ((range < 0) OR (range > 15))  return;

  chan->range = range;  frontEndChg(event->chanNum);
}


//-----------------------------------------------------------------------------
//  Change the index into the table that controls the averaging interval
//-----------------------------------------------------------------------------
void AvgIdxUp_(EventData *event)
{
  ChanState  *chan;

  if ((chan = CHAN + event->chanNum)->calibState)  return;

  if (chan->avgSizeIdx+1 >= NumAvgIdxs)  return;

  ++chan->avgSizeIdx;  frontEndChg(event->chanNum);
}

//-----------------------------------------------------------------------------
//  Change the index into the table that controls the averaging interval
//-----------------------------------------------------------------------------
void AvgIdxDown_(EventData *event)
{
  ChanState  *chan;

  if ((chan = CHAN + event->chanNum)->calibState)  return;

  if (chan->avgSizeIdx <= 0)  return;

  --chan->avgSizeIdx;  frontEndChg(event->chanNum);
}

//-----------------------------------------------------------------------------
//  Change the index into the table that controls the averaging interval
//-----------------------------------------------------------------------------
void NewAvgIdx_(EventData *event)
{
  int  idx;
  ChanState  *chan;


  if ((chan = CHAN + event->chanNum)->calibState)  return;

  idx = event->intVal;
  if ((idx < 0) OR (idx >= NumAvgIdxs))  return;

  chan->avgSizeIdx = idx;  frontEndChg(event->chanNum);
}

//-----------------------------------------------------------------------------
//  Connect the Front-End amplifier board to the input
//-----------------------------------------------------------------------------
void FrontEndOn_(EventData *event)
{
  ChanState  *chan;

  if ((chan = CHAN + event->chanNum)->calibState)  return;

  chan->frontEndOn = true;  frontEndChg(event->chanNum);
}

//-----------------------------------------------------------------------------
//  Disconnect the Front-End amplifier board from the input
//-----------------------------------------------------------------------------
void FrontEndOff_(EventData *event)
{
  ChanState  *chan;

  if ((chan = CHAN + event->chanNum)->calibState)  return;

  chan->frontEndOn = false;  frontEndChg(event->chanNum);
}

//-----------------------------------------------------------------------------
//  Turn on the hardware filter on the front end board
//-----------------------------------------------------------------------------
void FilterOn_(EventData *event)
{
  ChanState  *chan;

  if ((chan = CHAN + event->chanNum)->calibState)  return;

  chan->filterOn = true;  frontEndChg(event->chanNum);
}

//-----------------------------------------------------------------------------
//  Turn off the hardware filter on the front end board
//-----------------------------------------------------------------------------
void FilterOff_(EventData *event)
{
  ChanState  *chan;

  if ((chan = CHAN + event->chanNum)->calibState)  return;

  chan->filterOn = false;  frontEndChg(event->chanNum);
}

//-----------------------------------------------------------------------------
//  Clear the run-time zero-level offsets for all the ranges for a channel
//-----------------------------------------------------------------------------
void ClearOffsets_(EventData *event)
{
  int i;
  ChanState  *chan;
  FrontEndCalibData  *fe;
  char  *msg;


  if ((chan = CHAN + event->chanNum)->calibState)  return;

  msg = getBuf(80);  sprintf(msg, ClearOffsetsMsg, event->chanNum);
  LogMsg(msg);  freeTo(msg);

  fe = applData.calibData[event->chanNum];
  for (i=0; i<NumRanges; ++i)  fe[i].RTOffset = 0.0;

  frontEndChg(event->chanNum);
}

//-----------------------------------------------------------------------------
//  For Front Ends that support it, turn on the test current
//-----------------------------------------------------------------------------
void TestCurOn_(EventData *event)
{
  ChanState  *chan;

  if ((chan = CHAN + event->chanNum)->calibState)  return;

#ifdef BCM_SS
  chan->testCurOn = true;
  frontEndChg(event->chanNum);
#endif
}

//-----------------------------------------------------------------------------
//  For Front Ends that support it, turn off the test current
//-----------------------------------------------------------------------------
void TestCurOff_(EventData *event)
{
  ChanState  *chan;

  if ((chan = CHAN + event->chanNum)->calibState)  return;

#ifdef BCM_SS
  chan->testCurOn = false;
  frontEndChg(event->chanNum);
#endif
}

//-----------------------------------------------------------------------------
//  Save the in-RAM values for the Front End board to FLASH memory
//-----------------------------------------------------------------------------
void SaveCalib_(EventData *event)
{
  ChanState  *chan;

  if ((chan = CHAN + event->chanNum)->calibState)  return;

  chan->saveFECalib = true;
}

//-----------------------------------------------------------------------------
//  Duplicate one range's FrontEnd calibration data across all ranges
//-----------------------------------------------------------------------------
void DupRange_(EventData *event)
{
  int  to, from, chanNum;
  ChanState  *chan;
  char  *msg;


  if ((chan = CHAN + event->chanNum)->calibState)  return;

  chanNum = event->chanNum;  from = event->intVal;

  msg = getBuf(80);

  if ((from < 0) OR (from > 15))  {
    sprintf(msg, InvalidRangeMsg, from);
    LogMsg(msg);  freeTo(msg);  return;
  }

  sprintf(msg, CopiedRangeDataMsg, from);  LogMsg(msg);

  freeTo(msg);

  for (to=0; to<16; ++to)  if (to != from)
    applData.calibData[chanNum][to] = applData.calibData[chanNum][from];
}



//=============================================================================
//                   Functions unique to this application
//=============================================================================


//-----------------------------------------------------------------------------
//  Show the in-use (in RAM) and in-flash Front-End calibration values
//-----------------------------------------------------------------------------
void ShowCalibData(int chanNum)
{
  FrontEndCalibData  *fe, *fd, *flashData;
  char  *msg, *temp1, *temp2;
  int  i, offset;
  bool  valuesMatch;


  msg = getBuf(90 + 32 + 32 + CALIB_DATA_SIZE);
  temp1 = msg + 90;  temp2 = temp1 + 32;
  flashData = (FrontEndCalibData *)(temp2 + 32);

//--- get a copy of what is in flash ---
  offset = (char *)(applData.calibData[chanNum]) - (char *)&applData;

  Block_Read(BID_APPL_DATA, offset, flashData, CALIB_DATA_SIZE);

//--- display both the in-use (in RAM) and flash copies ---
  SetStreamMode();

  sprintf(msg, CalibDataMsg, chanNum);  PutStr(msg);

  PutStr("    ");  PutStr(CalibHeader0);  PutStr("\r\n");

  PutStr("    ");  PutStr(CalibHeader1);
  PutStr("    ");  PutStr(CalibHeader1);  PutStr("\r\n");

  PutStr("    ");  PutStr(CalibHeader2);
  PutStr("   ");   PutStr(CalibHeader2);  PutStr("\r\n");

  PutStr("    ");  PutStr(CalibHeader3);
  PutStr("   ");   PutStr(CalibHeader3);  PutStr("\r\n");

  for (i=0; i<NumRanges; ++i)   {
    fe = applData.calibData[chanNum] + i;
    fd = flashData + i;

    valuesMatch = ( (fe->RTOffset == fd->RTOffset)
                AND (fe->BCZeroOffset == fd->BCZeroOffset)
                AND (fe->BCAmpsPerVolt == fd->BCAmpsPerVolt)
                AND (fe->rangeOK == fd->rangeOK) );

    sprintf(msg, "%2d: %7.3fV   %7.3fV  %s  %s %s %7.3fV   %7.3fV  %s  %s",
            i,

            fe->RTOffset,  fe->BCZeroOffset,
            Engineering(fe->BCAmpsPerVolt, 4, temp1),
            (fe->rangeOK ? "OK " : "BAD"),

            (valuesMatch ? " " : "*"),

            fd->RTOffset,  fd->BCZeroOffset,
            Engineering(fd->BCAmpsPerVolt, 4, temp2),
            (fd->rangeOK ? "OK " : "BAD")
         );
    ShowMsg(msg);
  }

  ShowMsg(ESCToReturnMsg);

  freeTo(msg);
}

//-----------------------------------------------------------------------------
//  Handle any device-specific device special file open requests
//-----------------------------------------------------------------------------
int DLP_DevFile_Open(LinkProc *lp, int argc, char *argv[])
{
  return -999;
}



//----------------------------------------------------------------------------
//  Determine min, max, and avg readings over the specified # samples
//  Returns the # of samples used to determine the value
//----------------------------------------------------------------------------
uint CHAN_MinMaxRdg(int chanNum, int numSamples)
{
  uint  minRaw, maxRaw, rdg;
  int  i, x;
  ChanState  *chan;
  uint  *rawHist;


  chan = CHAN + chanNum;

  if (numSamples > chan->numSamples)  numSamples = chan->numSamples;
  if (numSamples < 1)  return 0;

//--- determine starting index ---
  x = adcIdxNewest[chanNum] - numSamples + 1;
  if (x < 0)  x += ADCHistSize;

  rawHist = adcRawHist[chanNum];

  minRaw = maxRaw = rawHist[x];

  for (i=0; i<numSamples; ++i)  {
    rdg = rawHist[x];
    if (rdg < minRaw)  minRaw = rdg;
    if (rdg > maxRaw)  maxRaw = rdg;
    if (++x >= ADCHistSize)  x = 0;
  }

  chan->minRaw = minRaw;  chan->maxRaw = maxRaw;

  return numSamples;
}


//---------------------------------------------------------------------------
//  Save just the RunTime Zero offset data for one of the BCM channels
//
//  WARNING:  Writes to FLASH!!!  To be called ONLY from the ComLoop thread!!!!
//---------------------------------------------------------------------------
void CHAN_SaveRTZeroOffsets(char chanNum)
{
  int  i, offset;
  FrontEndCalibData  calibData[NumRanges];
  ChanState  *chan;


  chan = CHAN + chanNum;

  offset = (char *)(applData.calibData[chanNum]) - (char *)&applData;

  LogMsg(SavingRTZeroDataMsg);

//--- read calib data for the channel from flash into local buffer ---
  if (Block_Read(BID_APPL_DATA, offset, calibData, CALIB_DATA_SIZE) < 0)  {
    chan->HWError = true;  return; }

//--- replace the RTZeroOffset's in the copy from flash with what is ---
//--- currently in use ---
  for (i=0; i<NumRanges; ++i)  {
    calibData[i].inputOn = applData.calibData[chanNum][i].inputOn;
    calibData[i].RTOffset = applData.calibData[chanNum][i].RTOffset;
  }

//--- write the updated local copy back to flash ---
  if (Block_Write(BID_APPL_DATA, offset, calibData, CALIB_DATA_SIZE)
       < CALIB_DATA_SIZE)  chan->HWError = true;
  if (flashBad)  chan->HWError = true;

  chan->saveRTZOffsets = false;
}


//---------------------------------------------------------------------------
//  Save all the Front End calibration data for the specified chan to Flash
//
//  WARNING:  Writes to FLASH!!!  To be called ONLY from the ComLoop thread!!!!
//---------------------------------------------------------------------------
void CHAN_SaveFECalib(char chanNum)
{
  int  offset;
  ChanState  *chan;
  FrontEndCalibData  *fe;


  if ((chan = CHAN + chanNum)->calibState)  return;

  LogMsg(SavingCalibDataMsg);

  fe = applData.calibData[chanNum];
  offset = (char *)fe - (char *)&applData;

  Block_Write(BID_APPL_DATA, offset, fe, CALIB_DATA_SIZE);

  if (flashBad)  chan->HWError = true;

  chan->saveFECalib = false;
}


//-----------------------------------------------------------------------------
//  ReLoads the flashed calibration data for the given channel
//-----------------------------------------------------------------------------
void CHAN_LoadCalibData(char chanNum)
{
  int  offset;

  offset = (char *)(applData.calibData[chanNum]) - (char *)&applData;

  if (Block_Read(BID_APPL_DATA, offset,
              applData.calibData[chanNum], CALIB_DATA_SIZE) < 0)
    CHAN[chanNum].HWError = true;
}

//-----------------------------------------------------------------------------
//  Examines the in-memory copy of the saved calibration data for a channel
//  Sets rangeOK to false for any range with obviously corrupt calib data.
//  Returns false if ANY of the ranges have obviously corrupt calib data.
//-----------------------------------------------------------------------------
bool CHAN_ValidCalibData(char chanNum)
{
  int  i;
  FrontEndCalibData  *fe;
  bool  rangeValid, allValid;


//--- check values for wildly unreasonable values ---
  allValid = true;

  for (i=0; i<NumRanges; ++i)  {
    fe = applData.calibData[chanNum] + i;   rangeValid = true;

    if ( (fabs(fe->RTOffset) > 25.0) OR
         (fabs(fe->BCZeroOffset) > 10.0) OR
         (fe->BCAmpsPerVolt <= 0.0) OR
         (fe->BCAmpsPerVolt > 0.01) )  rangeValid = false;

    if (! rangeValid)  { CHAN_ResetCalibData(chanNum, i);  allValid = false; }
  }

  return allValid;
}

//-----------------------------------------------------------------------------
//  Reset the in-memory (in-use) calibration data to default values
//-----------------------------------------------------------------------------
void CHAN_ResetCalibData(int chan, int range)
{
  short  i;
  FrontEndCalibData  *fe;
  RangeInfo  *ri;


  for (i=0; i<NumRanges; ++i)
    if ((i == range) OR (range == ALL_RANGES))  {
      fe = applData.calibData[chan] + i;
      ri = RangeData + (i & 0xFE);  // use even #'d (hdw) range data for both
      fe->rangeOK = false;  // used to indicate bad or lack of benchtop calib
      fe->inputOn = false;
      fe->RTOffset = 0.0;
      fe->BCAmpsPerVolt = fsCurrent[i & 0xFE] / ri->FSVoltage;
      fe->BCZeroOffset = 0.0;
    }

}


//-----------------------------------------------------------------------------
//  Called on powerup and after receipt of INIT command
//-----------------------------------------------------------------------------
void CHAN_Initialize(char chanNum)
{
  ChanState  *chan;
  char  *msg;


  msg = getBuf(90);
  sprintf(msg, "Initializing Channel %d", chanNum);  LogMsg(msg);
  freeTo(msg);

  chan = CHAN + chanNum;

  chan->calibState = CALIB_IDLE;
  chan->OverRange = CHAN_OVERRANGE_OK;  // not overrange

  chan->saveRTZOffsets = false;

  chan->HWError = false;               // hardware okay

  chan->testCurOn = false;
  chan->frontEndOn = true;
  chan->filterOn = false;
  chan->range = 0;
  chan->avgSizeIdx = DefaultAvgIdx;

  chan->avgCur = 0.0;           // no reading yet

  CHAN_LoadCalibData(chanNum);
  CHAN_ValidCalibData(chanNum);

  frontEndChg(chanNum);
}


//-----------------------------------------------------------------------------
//  Update history of readings and calculate beam current and other values
//-----------------------------------------------------------------------------
void CHAN_TakeReading(char chanNum)
{
  float variation;
  int  curRange;
  ChanState  *chan;
  FrontEndCalibData  *fe;


  chan = CHAN + chanNum;

  curRange = chan->range;  fe = applData.calibData[chanNum] + curRange;

//--- update # samples of ADC history applies to current settings ---
//--- (chan->numSamples is set to 0 when settings are changed) ---
  if (chan->numSamples < ADCHistSize)  ++chan->numSamples;

//--- update min/max rdg values ---
  chan->avgRaw = adcAvgRaw[chanNum];
  CHAN_MinMaxRdg(chanNum, 50);  // for variation over last 0.50 sec

//--- compute variance for recent values ---
  if (chan->numSamples < 100)   // require at least 1.00 secs worth of data
    chan->variance = 100.0;
  else  {
    variation = fabs( ADCtoVolts(chanNum, chan->maxRaw)
                    - ADCtoVolts(chanNum, chan->minRaw) );
    chan->variance = variation / ADC_FULL_SCALE * 100.0;
  }

//--- convert averaged raw value to volts and subtract run-time zero offset ---
  chan->avgVolts = voltsIn[chanNum];
  chan->adjVolts = chan->avgVolts - fe->RTOffset;

//--- calculate beam current ---
  chan->avgCur = (chan->adjVolts - fe->BCZeroOffset) * fe->BCAmpsPerVolt;

//if (chanNum == 1)   //test
//  printf("Range: %d  Cur: %s  = A/V: %s *(%f-%f-%f)\n",   //test
//        curRange, Engineering(chan->avgCur,4,temp1),
//        Engineering(fe->BCAmpsPerVolt,4,temp2),
//        chan->avgVolts, fe->BCZeroOffset, fe->RTOffset);

//--- get fraction of full-scale voltage limit ---
  chan->fracFSVolts = (fabs(chan->avgVolts) / VOLTAGE_OVERRANGE_LIMIT);

//--- get fraction of full-scale current ---
//--- (can be negative for now) ---
  chan->fracFSCur = chan->avgCur / fsCurrent[curRange];

//--- now make fracFSCur absolute value for use elsewhere ---
  chan->fracFSCur = fabs(chan->fracFSCur);

//--- set overrange state based of fraction of fullscale voltage ---
  if (chan->fracFSVolts < LIMIT_READY_OVERRANGE)
    chan->OverRange = CHAN_OVERRANGE_OK;
  else if (chan->fracFSVolts < LIMIT_ERROR_OVERRANGE)
    chan->OverRange = CHAN_OVERRANGE_READY;
  else
    chan->OverRange = CHAN_OVERRANGE_ERROR;

//--- update "extra" data values ---
//--- Do NOT chg what data is put here, or Benchtop Calibration application ---
//--- will cease to work!!! ---
  chan->Extra[0] = curRange;
  chan->Extra[1] = (long) (fabs(fe->RTOffset / VOLTAGE_OVERRANGE_LIMIT) * 1000);
  chan->Extra[2] = (long) chan->avgRaw;
  chan->Extra[3] = (long) (fabs(chan->avgVolts) * 1000);
  chan->Extra[4] = (long) ((double)fe->BCAmpsPerVolt * 10000000000000.0);
  chan->Extra[5] = (long) ((double)fe->BCZeroOffset * 1000000.0);
  chan->Extra[6] = (long) ((double)fe->RTOffset * 1000000.0);
}


//-----------------------------------------------------------------------------
//  Constructs a device status byte for reporting back in poll messages
//-----------------------------------------------------------------------------
int CHAN_GetStatus(char chanNum)
{
  int retstat;
  ChanState  *chan;


  chan = CHAN + chanNum;

  retstat = 0;

  // filter state
  if (chan->filterOn)  retstat |= CHAN_STATUS_FILTER;

  // front-end state
  if (chan->frontEndOn)  retstat |= CHAN_STATUS_FRONTEND;

  // test-current switch state
  if (chan->testCurOn)  retstat |= CHAN_STATUS_TESTCURR;

  // hardware error state
  if (chan->HWError) retstat |= CHAN_STATUS_HW_ERR;

  // averaging table selection
  if (chan->avgSizeIdx & 0x01) retstat |= CHAN_STATUS_AVG_B0;
  if (chan->avgSizeIdx & 0x02) retstat |= CHAN_STATUS_AVG_B1;
  if (chan->avgSizeIdx & 0x04) retstat |= CHAN_STATUS_AVG_B2;

  // operational state
  if (chan->calibState)   retstat |= CHAN_STATUS_INCAL;

  // over-range state
  switch (chan->OverRange) {
    case CHAN_OVERRANGE_OK:       retstat |= CHAN_STATUS_RDY_NORM; break;
    case CHAN_OVERRANGE_READY:    retstat |= CHAN_STATUS_RDY_OVER; break;
    case CHAN_OVERRANGE_ERROR:    retstat |= CHAN_STATUS_ERR_OVER; break;
  }

  // operating range is bad
  if (! applData.calibData[chanNum][chan->range].rangeOK)
    retstat |= CHAN_STATUS_RANGEBAD;

  return(retstat);
}




//-----------------------------------------------------------------------------
//  External Calibration routine.
//
//  An external program will apply a current to our input, wait a bit (for
//  our averaging buffer to fill) and then send us a CMD_XCAL 1 or 2
//  message with the value of the applied current.  We save the average
//  ADC value and the provided current, and with two such points we can
//  determine appropriate values for determining the current for the specified
//  channel from the current voltage reading.
//-----------------------------------------------------------------------------
void CHAN_RangeCalibrationPoint(char chanNum, int PointNum, float Current)
{
  float ampsPerVolt, zeroOffset, deltaA, deltaV;
  ChanState *chan;
  int  i, range;
  FrontEndCalibData  *fe;
  char  *msg, *temp;
  bool  rangeOK;


  StackPtr();

  chan = CHAN + chanNum;

  range = chan->range;
  fe = applData.calibData[chanNum] + range;

  msg = getBuf(90 + 90);  temp = msg + 90;

  if ((PointNum < 1) OR (PointNum > 2))   {
    sprintf(msg, "Invalid sample point: %d", PointNum);  LogMsg(msg);
    freeTo(msg);  return;
  }


  //--- store the given current and measured corresponding voltage ---
  chan->RangeCalib[PointNum-1].Current = Current;
  chan->RangeCalib[PointNum-1].Voltage = ADCtoVolts(chanNum, chan->avgRaw);

  if (PointNum == 1)  LogMsg("===================================");
  sprintf(msg, CalibPointDataMsg,
          range, chanNum+1, PointNum, chanNum,
          chan->RangeCalib[PointNum-1].Voltage, Engineering(Current,4,temp));
  LogMsg(msg);

//--- if second point, calibrate the current range ---
  if (PointNum == 2) {
    calibPoint1[chanNum] = chan->RangeCalib[0].Current;
    calibPoint2[chanNum] = chan->RangeCalib[1].Current;

    deltaA = chan->RangeCalib[1].Current - chan->RangeCalib[0].Current;
    deltaV = chan->RangeCalib[1].Voltage - chan->RangeCalib[0].Voltage;

    ampsPerVolt = deltaA / deltaV;
    zeroOffset =
       chan->RangeCalib[0].Voltage - chan->RangeCalib[0].Current/ampsPerVolt;

    sprintf(msg, "Amps/Volt: %s   zeroOffset: %s Volts\n",
      Engineering(ampsPerVolt, 4, temp), Engineering(zeroOffset, 4, temp+16));
    LogMsg(msg);

    fe->RTOffset = 0.0;
    fe->BCZeroOffset = zeroOffset;
    fe->BCAmpsPerVolt = ampsPerVolt;

    rangeOK = true;
    if (fabs(deltaV) < 8.0)  rangeOK = false;
    if (fabs(zeroOffset) > 5.0)  rangeOK = false;
    fe->rangeOK = rangeOK;

    if (! (range & 1))  {  // save for software range (all odd #'s) as well
      ++fe;  ++range;
      fe->rangeOK = rangeOK;
      fe->RTOffset = 0.0;
      fe->BCZeroOffset = zeroOffset;
      fe->BCAmpsPerVolt = ampsPerVolt;
    }

    chanStatus[chanNum] = CHAN_GetStatus(chanNum);
  }


  freeTo(msg);
}



//============================================================================
main()
{
  char  *msg;
  int  i;  //tdebug
  float  rangeMult;


  firstTime = true;

  while (true)  {
    CommonInit();

    msg = getBuf(128);   // do NOT free up - "backup" in case of failed getBuf()

#if (_BOARD_TYPE_ != RCM3200)
    ioBoardType = SmartCat;
#else
    ioBoardType = NSCL16;
#endif

    bcm1_out = bcm2_out = RangeData[0].RelayBits;
    digOutBootState = ((ulong)bcm1_out << 1) | ((ulong)bcm2_out << 10);

    CommonHdwInit(digOutBootState);  DLPHdwInit();

//========== Device specific initialization needed ASAP =========

    DLPInit();

//========== Remaining Device-specific Initialization ==========

    rangeMult = (digitalIn & NotRangeX10) ? 1.0 : 10.0;
    for (i=0; i<NumRanges; ++i)
      fsCurrent[i] = RangeData[i].FSCurrent * rangeMult;


    bcm1_in = bcm2_in = 0;  logData = false;

    cmdChan = 0;  // which channel telnet commands are applied to

//--- clear and initialize ChanState structures ---
    memset(&CHAN[0],0,sizeof(ChanState));
    CHAN[0].OutputByte   = &bcm1_out;
    CHAN[0].InputByte    = &bcm1_in;
    CHAN[0].AdcWord      = adcRaw+0;
    CHAN_Initialize(0);

    memset(&CHAN[1],0,sizeof(ChanState));
    CHAN[1].OutputByte   = &bcm2_out;
    CHAN[1].InputByte    = &bcm2_in;
    CHAN[1].AdcWord      = adcRaw+1;
    CHAN_Initialize(1);

// search for _PERSISTENT_ to see where code normally added for saved values

    controlEnabled = true;  //initiate active control of the device

    while (ComLoop());

    freeTo(msg);  firstTime = false;
  }
}


