//----- DVRAppl.c ----- 06/20/12 -----(03/05/06)----

// EE Documents bundle # 045

// Sections that typically require application-specific code are marked with
// the comment:  //APPLCODE

//-----------------------------------------------------------------------------
//  Application specific code for the main application task that is run at
//  a fixed interval and application specific commands run on demand.
//
//APPLCODE
//  Application code to manage the D Voltage Regulator NIM module
//-----------------------------------------------------------------------------
// Update log (most recent first);
//
//
//  2012-06-20 - Mark Davis
//               - Reduced time since last pass that acts as the threshold for
//                 determining when the applic task should write log messages
//                 for changes in the state of the digital I/O signals.  This
//                 should reduce the occurance of confusing logs that contain
//                 a sequence of messages showing a change to one value but
//                 not the other (so it looks like it is telling you over and
//                 over the same value with no change).
//
//  2011-01-21 - Mark Davis
//               - Minor tweaks to support soft resets
//
//  2010-04-30 - Mark Davis
//               - Added hdwCfgVolts for diagnostic usage
//
//  2009-12-14 - Mark Davis
//               - Changed order of param files to deal with problems caused
//                 by the order in which they were restored at startup.
//
//  2009-06-23 - Mark Davis -
//               - Added log messages for the Limited and Error LED states.
//
//  2009-06-22 - Mark Davis -
//               - Reversed pressed/released messages for the INITIAL and
//                 FINAL setpoint buttons so they are correct.
//
//  2009-04-15 - Mark Davis -
//               - Corrected messages logged for RF ON/OFF digital input (they
//                 were backwards)
//
//  2009-03-20 - Mark Davis -
//               - Added digInDescr[] and digOutDescr[] to support the newer
//                 generic method for logging changes of interest in the state
//                 of the digital I/O signals.
//
//  2008-08-20 - Mark Davis -
//               - new values for rfOnSetpoint and rfOffSetpoint now ignored
//                 when in local mode
//
//  2008-03-06 - Mark Davis -
//               - split updateState() into standard 4 separate func calls
//                 (NO changes in the logic or actual code statements!)
//               - added check for out-of-range hardware jumper config (hdwCfg)
//                 values
//
//  2008-01-10 - Mark Davis -
//               - added (hopefully temp) method of setting gains based on
//                 on the hardware configuration.  The gains are currently
//                 forced to default values if the resistor in the header
//                 indicates a K500 or K1200 configuration.
//
//  2007-11-29 - Mark Davis -
//               - started work on DVR module code
//
//-----------------------------------------------------------------------------
//
//  The primary function of the controller is to monitor the ON/OFF status of
//  the RF via a digital input and maintain the cavity voltage at one of two
//  setpoints:  One for the OFF or INITIAL/cold state, and one for the ON or
//  FINAL/warm state.
//
//  Regulation of the cavity voltage is done by an analog feedback loop using
//  a setpoint voltage supplied by the control logic.  As long as the module
//  has power, one of the two setpoints (or a value in between) is applied to
//  the feedback loop.
//
//  To avoid surges due to unmatched conditions, particularly when the RF is
//  first turned on and still warming up, configurable delays and ramp rates
//  are used to gradually ramp the setpoint from one value to another.
//
//  The front panel contains a knob, a digital meter, and switches for viewing
//  the current operating voltage and making changes to the setpoints.  These
//  values and all the other software configurable parameters can also be
//  viewed and/or set via the control system software interface.
//
//-----------------------------------------------------------------------------
//  Run-time settings (values that change frequently - changes are
//  automatically saved):
//
//  rfOffSetpoint:    INITIAL/COLD setpoint.  The setpoint to apply when the RF
//                    is OFF.
//
//  rfOnSetpoint:     FINAL/WARM setpoint.  The setpoint to apply when the RF
//                    is ON.
//
//-----------------------------------------------------------------------------
//  Configuration values (normally set and left alone - changes are manually
//  saved using the command 'save devdata'):
//
//    coldRampRate:   The rate at which the active setpoint is ramped toward
//                    rfOnSetpoint after offToOnDelay has been observed.
//
//    warmRampRate:   The rate at which the active setpoint is ramped toward
//                    a new rfonSetpoint after the offToOnDelay has expired and
//                    the active setpoint has reached the previous rfOnSetpoint.
//
//    offToOnDelay:   The # of ms to wait after the RF changes from OFF to ON
//                    before the active setpoint begins ramping toward the
//                    rfOnSetpoint.
//
//    maxRegErr:      The maximum difference between the active setpoint and
//                    the cavity voltage readback that is not considered an
//                    error.
//
//    limiterActiveV: The maximum magnitude for the Limiter voltage that is
//                    not considered an error.
//
//    knobRate:       The scaling factor for applying knob movement to one of
//                    the setpoints
//
//    maxOffSetpoint: Maximum value for rfOffSetpoint
//
//    maxOnSetpoint:  Maximum value for rfOnSetpoint
//
//-----------------------------------------------------------------------------
//
//  Theory of operation:
//
//  The two main features of the module are an analog feedback loop that
//  controls the RF cavity voltage, and an embedded SBC (Single Board Computer)
//  whose primary tasks are to provide the setpoint for the feedback loop and
//  monitor the RF cavity voltage and ON/OFF status.
//
//  The controller maintains two setpoints:  rfOffSetpoint (INITIAL/cold) and
//  rfOnSetpoint (FINAL/warm).  These setpoints and the RF ON status determine
//  the activeSetpoint applied to the feedback loop as follows:
//
//  RF is OFF
//    (rampMode RAMP_OFF)
//    The active setpoint is set to rfOffSetpoint
//
//  RF changes from OFF to ON
//    (rampMode RAMP_WAIT):
//    The active setpoint remains at the rfOffSetpoint for offToOnDelay ms.
//
//  RF ON for at least offToOnDelay ms, active setpoint != rfOnSetpoint
//    (rampMode RAMP_COLD)
//    The active setpoint is ramped toward rfOnSetpoint using coldRampRate
//    (independent of any changes to rfOnSetpoint).
//
//  RF ON and active setpoint reaches rfOnSetpoint
//    (rampMode RAMP_WARM)
//    Any changes to the onSetPoint are applied using warmRampRate.
//
//
//  The front panel contains the following controls and displays:
//    - RF ON LED
//    - LIMITED LED
//    - ERROR LED
//    - DEE VOLTAGE (digital voltage meter)
//    - Optical encoder knob
//    - FINAL pushbutton
//    - INITIAL pushbutton
//    - OPEN/CLOSED LOOP toggle switch
//    - LOCAL/REMOTE toggle switch
//    - RF ON/OFF (BNC connector)
//    - RF IN (BNC connector) - to monitor Cavity Voltage
//    - PEAK DETECTOR IN (BNC connector)
//    - LIMITER (BNC connector)
//    - RF OUT (BNC connector)
//
//  The RF ON LED indicates the state of the RF.  It is controlled directly
//  by the RF ON/OFF input (i.e. the control logic cannot affect its state).
//
//  The LIMITED LED indicates whether or not the LIMITER voltage is within the
//  normal operating range.  It is turned on if the LIMITER voltage exceeds
//  the threshold defined by the Limiter Error parameter.
//
//  The ERROR LED indicates that the analog feedback loop is unable to
//  properly regulate the Cavity Voltage.  The control logic monitors the
//  difference between the Active Setpoint and Cavity Voltage and turns on
//  the LED if the difference exceeds the threshold set by the Regulation
//  Error parameter.
//
//  The DEE VOLTAGE digital voltage meter normally displays the RF Cavity
//  Voltage.  Whenever the INITIAL or FINAL pushbutton is being pressed, the
//  controller displays the RF-OFF or the RF-ON Setpoint, respectively.
//
//  The Optical encoder knob is used to change the RF-OFF Setpoint
//  (INITIAL/cold) or RF-ON Setpoint (FINAL/warm) values as follows:
//
//   - The RF-OFF Setpoint can be adjusted via the front panel by turning
//     the knob while pressing and holding the INITIAL button.  This operation
//     is independent of the position of the LOCAL/REMOTE switch.
//
//   - The RF-ON Setpoint can be adjusted via the front panel by setting
//     the LOCAL/REMOTE switch to the LOCAL position and turning the knob.
//     NOTE that the meter continues to display the RF Cavity Voltage unless
//     you are also pressing the FINAL button.
//
//  The OPEN/CLOSED LOOP toggle switch determines whether or not the analog
//  feedback loop is operational.  In the OPEN position, the feedback loop
//  is disabled and the difference between the Active Setpoint and the Cavity
//  Voltage is not used to regulate the Cavity Voltage.
//
//  The RF ON/OFF input enables/disables the RF OUT signal, controls the
//  state of the RF ON LED, and allows the control logic to monitor the
//  ON/OFF state of the RF, which it uses to decide which of the two
//  setpoints the Active Setpoint is ramped toward/set to.
//
//  The RF IN input provides the constant RF input signal.  This signal is
//  amplitude modulated by the feedback loop regulation circuit and feed to
//  the RF OUT connector when the RF is ON.  The amplitude of the resulting
//  RF OUT signal ultimately determines the RF Cavity Voltage.
//
//  The PEAK DETECTOR IN is used along with the Attenuator switches on the
//  front panel to derive the RF Cavity Voltage.
//
//-----------------------------------------------------------------------------

#include <common.h>

#include <unistd.h>
#include <sys/types.h>


//APPLCODE - change to match the name of this file
#ifndef  ApplicSource
  #define  ApplicSource  "DVRAppl.c"
#endif

#ifndef DidRegDefs
  #define  DidRegDefs
  #include <modbusRegs.c>
#endif
#ifndef RegDefs


//APPLCODE - change to the application name
const char far applName[] = { "DVR" };
const char ApplAbbrev[] = { "DVR" };

//APPLCODE - add ID to unistd.h and place it here
#define DeviceType  DT_DVR

//APPLCODE - set the # of input and outputs used
#define  NUM_APPL_AIN      4   // analog inputs used
#define  NUM_APPL_AOUT     2   // analog outputs used
#define  NUM_APPL_DIN      5   // digital inputs used
#define  NUM_APPL_DOUT     2   // digital outputs used

#define  ADC_MASK_APPL  0x0000000FL
#define  DAC_MASK_APPL  0x00000003L

//APPLCODE - set to state we want outputs in during initialization
#define InitialDigOutState  ( 0 )


#define SharedINC
#include <sharedAppl.c>



//=============================================================================
//  definitions related to use of the Digital and Analog I/O signals
//=============================================================================

//APPLCODE - define macros for each Digital bit and each Analog index

//=============== digitalIn bit usage ===============
#define  CLOSED_LOOP     0x00000001  // 0=Open Loop, 1=Closed Loop
#define  FINAL_BTN       0x00000002  // 1=SET button being pushed
#define  INITIAL_BTN     0x00000004  // 1=INITIAL button being pushed
#define  RF_IS_OFF       0x00000008  // 1=RF is OFF, 0=RF is ON
#define  LOC_REM_SW      0x00000010  // 1=Local, 0=Remote

const uchar  LogicOrderDIn[NUM_APPL_DIN] =  { 0, 1, 2, 3, 4 };

#define  ClosedLoop(din)        ( (bool)((din) & CLOSED_LOOP) )
#define  FinalBtnPushed(din)    ( (bool)((din) & FINAL_BTN) )
#define  InitialBtnPushed(din)  ( (bool)((din) & INITIAL_BTN) )
#define  RFisOff(din)           ( (bool)((din) & RF_IS_OFF) )
#define  LocalMode(din)         ( (bool)((din) & LOC_REM_SW) )

//--- msgs to log automatically when a digital input changes state ---
//--- use NULL for bits that we don't want to log changes for ---
const char far *digInDescr[NUM_APPL_DIN*2] = {
//--0 state descr--                  --1 state descr--
  "OPEN Loop",                       "CLOSED Loop",
  "FINAL setpoint btn released",     "FINAL setpoint btn pressed",
  "INITIAL setpoint btn released",   "INITIAL setpoint btn pressed",
  "RF ON",                           "RF OFF",
  "REMOTE mode",                     "LOCAL mode"
};


//=============== digitalOut bit usage ===============
#define  LIMITED_LED     0x00000001   // 1=On, 0=Off
#define  ERROR_LED       0x00000002   // 1=On, 0=Off
#define  LED_BITS        0x00000003   // bits for all LEDs

#define  ActivityLED     0x00000100L

const uchar  LogicOrderDOut[NUM_APPL_DOUT] =  { 0, 1 };

#define SetOffApplLEDs(dout)   ClearBits((dout), LED_BITS)

#define SetOnLimitLED(dout)   SetBits((dout), LIMITED_LED)
#define SetOnErrorLED(dout)   SetBits((dout), ERROR_LED)

#define  TurnOnActivityLED()    ClearBits(Next.digitalOut, ActivityLED)
#define  TurnOffActivityLED()   SetBits(Next.digitalOut, ActivityLED)

//--- msgs to log automatically when a digital output changes state ---
//--- use NULL for bits that we don't want to log changes for ---
const char far *digOutDescr[NUM_APPL_DOUT*2] = {
//--0 state descr--    --1 state descr--
  "Limited LED Off",   "Limited LED On",
  "Error LED Off",     "Error LED On"
};



//=============== Analog inputs ===============
#define  RF_CAV_VOLT      0  // RF cavity voltage
#define  LIMITER_VOLT     1  // Setpoint Limiter voltage
#define  UNUSED_V2        2
#define  HDW_CFG          3  // Fixed volt indicating HD1 component set

const uchar  LogicOrderAIn[NUM_APPL_AIN] = { 0, 1, 2, 3 };

//--- 5.07V divided in to 8 steps ---
#define  HdwConfig(ain)         ( (uint)(((5.07 - FABS(ain)) * 1.5779) + 0.5) )
/*
  Resistor bridge on HD1 that this reads:

  (0)  5.07V --   0K -- 100K -- Gnd  ==>  5.07V  (a short)
  (1)  5.07V --  15K -- 100K -- Gnd  ==>  4.4V
  (2)  5.07V --  33K -- 100K -- Gnd  ==>  3.8V
  (3)  5.07V --  62K -- 100K -- Gnd  ==>  3.13V
  (4)  5.07V -- 100K -- 100K -- Gnd  ==>  2.54V
  (5)  5.07V -- 167K -- 100K -- Gnd  ==>  1.90V
  (6)  5.07V -- 300K -- 100K -- Gnd  ==>  1.27V
  (7)  5.07V -- 700K -- 100K -- Gnd  ==>  0.63V
  (8)  5.07V -- open -- 100K -- Gnd  ==>  0.0V   (open)
*/
static char *HdwCfgDesc[9] = { "RSVD0", "K500 ", "K1200", "RSVD3",
                               "RSVD4", "RSVD5", "RSVD6", "RSVD7", "ERROR" };

//--- values for HdwConfig() ---
#define  RSVD0_CFG        0   // not used
#define  K500_CFG         1   // Header sets hdw gains for K500 use
#define  K1200_CFG        2   // Header sets hdw gains for K1200 use
#define  RSVD3_CFG        3   // reserved for future
#define  RSVD4_CFG        4   // reserved for future
#define  RSVD5_CFG        5   // reserved for future
#define  RSVD6_CFG        6   // reserved for future
#define  RSVD7_CFG        7   // reserved for future
#define  ERROR_CFG        8   // no resistor


//=============== Anlog outputs ===============
#define  SETPOINT         0  // Setpoint for Cav Voltage Regulator circuit
#define  FP_METER         1  // FrontPanel Meter if INITIAL or SET pressed

const uchar  LogicOrderAOut[NUM_APPL_AOUT] = { 0, 1 };


//=============================================================================
//  Application-specific data structures and definitions
//=============================================================================

//-----------------------------------------------------------------------------
//  Each value in the ApplState structure should be throught of as representing
//  a "signal" in a clocked logic circuit, with the "applic" task representing
//  the logic implemented by the circuit elements such as gates, latches, etc.
//
//  The main "applic" task is the highest priority task and executes the code
//  that implements the application logic on a real-time, periodic basis.
//
//                  ====== IMPORTANT NOTE on LOCKING ======
//
//  Remember to use a semaphore or some other form of locking mechanism to
//  protect all access to shared values by more than one task.  The only
//  exception to this is when the applic task needs to access a value protected
//  by other tasks using only a scheduler/kernel lock (the KLOCK and KUNLOCK
//  macros).  Because the applic task is the only task that runs at the highest
//  priority, no other task can preempt it, and if the scheduler is locked,
//  the applic task cannot preempt another task.  But scheduler locks must be
//  kept short to avoid creating large latency in the timely execution of other
//  tasks (the applic task in particular).
//
//-----------------------------------------------------------------------------
// Values that belong in the ApplState structure:
//
//   - Ones that are required for a "soft" reset to work.  A soft reset is
//     the same as a regular (hdw) reset except that it does not involve an
//     actual CPU reset.  This allows properly constructed code to avoid
//     changes to the analog and digital outputs during a reset, allowing
//     certain otherwise unrecoverable problems to be handled with a minimum
//     amount of disruption to operations.  The 3 instances of the ApplState
//     structure are saved and restored for all applications that support
//     soft resets.
//
//     The problem experienced at the NSCL this was created to address is
//     controllers in which the values write-protected portions of RAM can
//     change (without causing a write protection fault!).  As this problem
//     seems connected to the physical location of a controller (i.e.
//     controllers running the same code in other areas do not have the
//     problem), it is apparently being caused by some (as yet) unidentified
//     environmental condition (heat, power spikes, maybe even energetic
//     particles).
//
//   - Values for which we need to be able to track changes between cycles
//     (e.g. is curState.X different than prevState.X?)
//
//-----------------------------------------------------------------------------


// ***** WARNING ***** WARNING ***** WARNING ***** WARNING *****
//  Changing the applState structure means that a soft reset will fail!  A
//  change to the size will cause a soft restart to abort and do a CPU
//  restart instead.  And while a change that does NOT result in a new
//  size will not cause the soft restart to abort, it does mean that any
//  shift in the position of values can cause all kinds of mayhem.
// ***** WARNING ***** WARNING ***** WARNING ***** WARNING *****


typedef struct {
 //-----------------------------------------------------------------------
 // SOFTWARE SETTING INPUTS
 //
 //  These values represent Software Inputs to the application logic.
 //
 //  These values are set in the "Next" instance by tasks OTHER than the
 //  "applic" task to change some value used by the application logic.
 //  Remember that all tasks besides the main "applic" task must use a
 //  scheduler lock to protect reads and writes to these values.
 //
 //  Note that the application task may alter these values (e.g. to change
 //  an invalid value in to a valid one).
 //-----------------------------------------------------------------------


 //-----------------------------------------------------------------------
 // READINGS
 //
 //  These values represent "passive" inputs to the application logic.
 //
 //  These values are set in the "Next" instance by the applic task just
 //  after reading the state of the hardware inputs.
 //
 //  These values generally come from one of two sources:
 //    - The just-read state of analog and/or digital hardware inputs
 //    - Values maintained by other tasks
 //
 //  While some of these values are simply copies of "external" values,
 //  including them in the ApplState structure allows us to see how
 //  they change between cycles.  Also, some of these values are the result
 //  of applying logic to "external" values.  Applying that logic once at
 //  at the beginning of the applic loop and saving the result eliminates
 //  the need to repeat the same logic in multiple blocks of code later on,
 //  avoiding unnecessary redundancy.
 //-----------------------------------------------------------------------
  ulong  digitalIn;

  ulong  processTime;  // ms time applic loop started on applic-specific code

  bool   rfIsOn;       // RF ON digital input

  bool   localMode;    // LOCAL/REMOTE switch is in LOCAL position
  bool   closedLoop;   // OPEN/CLOSED LOOP switch is in CLOSED position

 //-----------------------------------------------------------------------
 // OUTPUTS from UN-CLOCKED (asynchronous, gate-level) LOGIC ELEMENTS
 //
 //  These values are outputs from one of the blocks of code in the applic
 //  task that represent a non-clocked logic element (such as a gate).
 //
 //  Each value here is set by one (and ONLY one) of the blocks of code in
 //  the applic task.  This group contains ONLY values that are set by a
 //  block of code that represents a NON-clocked logic element.
 //
 //  These values are set in the "Next" instance by the applic task just
 //  after reading the state of the hardware inputs.  These values must
 //  NOT be altered by other tasks (although they can read them as long
 //  as they use appropriate locking to do so).
 //
 //  Because these values represent the outputs of un-clocked logic elements,
 //  and such elements often need to be cascaded (the outputs used as inputs
 //  to other non-clocked elements), the result is often dependent on the
 //  order of execution of the blocks of code.
 //-----------------------------------------------------------------------


 //-----------------------------------------------------------------------
 // OUTPUTS from CLOCKED (latched inputs and outputs) LOGIC ELEMENTS
 //
 //  These values are outputs from one of the blocks of code in the applic
 //  task that represent a clocked logic element (such as latches).
 //
 //  Each value here is set by one (and ONLY one) of the blocks of code in
 //  the applic task.  This group contains only values that are set by a
 //  block of code that represents a CLOCKED logic element.
 //
 //  These values are set in the "Next" instance by the applic task AFTER
 //  all the other ApplState values are set.  Because they represent only
 //  the clocked logic elements, and each block of code reads only values
 //  from the Cur instance, the blocks can be executed in any order and the
 //  result will always be the same.
 //
 //  These values are used as inputs by other blocks and/or to set values
 //  used by other tasks (such as values read by the ModBus server task).
 //
 //  Again, these values must NOT be altered by other tasks (although they
 //  can read them as long as they use appropriate locking to do so).
 //-----------------------------------------------------------------------
  ulong  digitalOut;      // state for digital outputs

  float  activeSetpoint;  // value applied to regulation circuit

  float  rfOnSetpoint;    // warm/final setpoint
  float  rfOffSetpoint;   // cold/initial setpoint

} ApplState;


ApplState  prevState, curState, nextState;

#define  Prev  prevState
#define  Cur   curState
#define  Next  nextState

const uint ApplStateDataSize =  ( sizeof(ApplState)*3 + sizeof(setVal) );



int  knobTicks;

float  diff, maxChg;

ulong  msSincePrev;

bool  initialBtnPressed, finalBtnPressed, limited, regErr;

ulong  rfOnTime;   // time RF turned on (in milliseconds)

float  regulationError;  // diff between setpoint and readback


//----------------------------------------------------
// Values that

//=============================================================================
//  List of daemon processes.  The inetd process listens for connection
//  requests on each port in the list and creates a new instance of the
//  specified process to service each new connection.
//=============================================================================
#include <services.h>

//APPLCODE - add tasks that service network clients

//--- service funcs declared the same as standard command funcs:
//     e.g.:   void _task fastdatad(int argc, char **argv);

void _task tsh(int argc, char **argv);


ServiceInfo inetdConf[] = {
//--- add any application-specific inetd-type network services here ---
//--- (add port # names to services.h) ---
//  NAME          PORT           PROTOCOL  FUNC       PRI  -1  Config Params
//  ------------  -------------  --------  ---------- ---  --  ----------------

//--- standard ones defined in sharedAppl.c ---
//{ "console",    TTY0,          ASYNC,    tsh,        10, -1, &consCfg },
//{ "fastdatad",  FASTDATA_PORT, ProtoTCP, fastdatad,   5, -1, NULL },
//{ "modbusTCPd", MODBUS_PORT,   ProtoTCP, modbusTCPd,  6, -1, NULL },
//{ "tsh",        TELNET_PORT,   ProtoTCP, tsh,        10, -1, NULL }

  #define SharedServices
  #include <sharedAppl.c>
};

const int  NumServices = sizeof(inetdConf) / sizeof(ServiceInfo);





//=============================================================================
//  Defintions for variables that are accessable as ModBus registers
//=============================================================================
#else  // end of before-RegDefs portion

/*
  The purpose of this section is to define a set of C variables that are
  accessable as ModBus registers.  This approach avoids the need to use
  special functions or macros throughout the code when accessing register
  values, which keeps the application code fast and simple.

  The model used provides support for multiple logical devices.  The 2nd
  parameter in each register definition macro indicates if one instanace
  applies to all logical devices (specified by an "x" for "don't care"), or
  if each logical device requires its own instance (specified by an under-
  score "_" to signify "fill in the blank").  In the 1st case, the variable
  is defined as a scalar.  In the 2nd case, it is defined as an array.

  The definition macros support data types (floats and longs in particular)
  that require more than one register to store them.  To avoid the possibility
  of corrupted values, all reads of and writes to such values must be done
  using the multi-register commands and including all the registers that
  make up the value in the command.  Differences in register order between
  different client types is handled by a table that automatically swaps
  register values as needed based on the client type.

  The definition for each writable register also includes the function to
  call when the value is set via the ModBus interface.  Many registers can
  use the default setXXX() functions, but you can specify your own to provide
  any special logic needed (including ignoring the new setting if invalid).

  For registers, the format for the function definitions is:

    bool mySetFunc(SetRegFuncParams);

       where SetRegFuncParams is currently defined as:
              void *info, uchar chanNum, void *newVal, fd_t eventsfd

  For coils, the format for the function definitions is:

     bool (*SetCoilFunc)(SetCoilFuncParams);

         where SetCoilFuncParams is currently define as:
               void *info, uchar chanNum, uchar newVal, fd_t eventsfd

  The parameters to the call are as follows:
    *info:     A pointer to the Coil's/Register's CoilInfo/RegInfo structure
    chanNum:   The logical device # (0 is the 1st)
    *newVal:   A pointer to the new value requested in the Modbus packet
               (changed to native-byte order where needed)
    eventsfd:  The fd for the FIFO used to send msgs to the main application
               process

  The function returns true if the new value was accepted, false if not.

  The register numbers for multiple "channels" or logical devices are
  handled automatically based on the xxChanMult values.  These values specify
  the offset between a register value for one logical device and the "same"
  register for the next.

  Example:

    Assume the following definitions:
      #define  NumLogicalDevices  2
      #define  R4ChanMult  100
      REG4( motorPos,       _, 40050, F32, RW, setMotorPos )

    Then we would get:
      global variable definition:  float motorPos[2];
      register pair:  40050,40051  refers to:  motorPos[0]
      register pair:  40150.40151  refers to:  motorPos[1]

    If instead we had specified:
      REG4( motorPos,       x, 40050, F32, RW, setMotorPos )

    Then we would get:
      global variable definition:  float motorPos;
      register pairs:  40050,40051 & 40150,40151  BOTH refer to:  motorPos


  To keep the definitions as simple as possible, and avoid requiring multiple
  macros for each register definition, each macro is used more than once,
  with a different definition (#define) each time (a practice generally
  referred to as X-Macros).  While trying to follow the order of inclusion
  of files (including a recursive reference back to this file) can be rather
  confusing, the X-Macro method keeps the source code simple to read and
  maintain by doing everything with a single macro for each register.

  ==========================================================================
                             !!! WARNING !!!
  ==========================================================================
    Each sequence of register definitions MUST be contiguous!  There must
    NOT be any holes (i.e. no unused register #'s between used ones) and no
    overlap!
-----------------------------------------------------------------------------*/


//APPLCODE - set # of devices for which we want duplicate register sets
#ifndef NumLogicalDevices
  #define  NumLogicalDevices  1
#endif

//--------------------------------------------------------------
//APPLCODE - set max # of registers per set of coils
#ifndef NumC0Chans
  #define NumC0Chans_  NumLogicalDevices
  #define C0ChanMult_  100
#endif

#ifdef COIL0

COIL0( DoReboot,       x, 1, setCoil )
//COIL0( Coil00002,      _, 2, setCoil )
//COIL0( Coil00003,      _, 3, setCoil )

#undef COIL0
#endif


//--------------------------------------------------------------
//APPLCODE - set max # of registers per set of read-only registers
#ifndef NumR3Chans
  #define NumR3Chans_  NumLogicalDevices
  #define R3ChanMult_  100
#endif

#ifdef REG3

REG3( dlmVersTime,    x, 30001, U32 )
REG3( rtosVersTime,   x, 30003, U32 )
REG3( applVersTime,   x, 30005, U32 )
REG3( devType_,       x, 30007, U16 )
REG3( deviceID_,      x, 30008, U32 )
REG3( upSecs,         x, 30010, U32 )
REG3( percentIdle_,   x, 30012, U16 )

REG3( numChans,       x, 30013, U16 )
REG3( coilChanMult,   x, 30014, U16 )
REG3( reg3ChanMult,   x, 30015, U16 )
REG3( reg4ChanMult,   x, 30016, U16 )

REG3( digInBits,      x, 30017, U32 )

REG3( statusBitsStdA, x, 30019, U16 )
REG3( statusBitsStdB, x, 30020, U16 )
REG3( statusBitsAppl, x, 30021, U32 )

//APPLCODE - add registers for externally available read-only values
REG3( cavityVolts,     x, 30023, F32 )  /* RF Cavity voltage */
REG3( activeSetpoint,  x, 30025, F32 )  /* currently applied setpoint */
REG3( regulationError, x, 30027, F32 )  /* setpoint - reading error */
REG3( limiterVolts,    x, 30029, F32 )  /* setpoint limiter voltage */
REG3( rampMode,        x, 30031, U16 )  /* mode for activeSetpoint chgs */

REG3( hdwCfg,          x, 30032, U16 )  /* HD1 component set in use */

REG3( latency,         x, 30033, U16 )  // actual vs scheduled appl start in ms
REG3( tcpTickTime,     x, 30034, U16 )  // # ms inside tcp_tick() call
REG3( applLoopTime,    x, 30035, U16 )  // # ms used by applic task loop
REG3( kernelLoopTime,  x, 30036, U16 )  // # ms used by last kernel pass

REG3( hdwCfgVolts,     x, 30037, F32 )  // voltage indicating hdw cfg in use

#undef REG3
#endif


#ifdef RegMacros
//===== application-specific defintions related to register values =====

  #define  READING(idx)  REG_ARRAY(float, rdg_1, idx)

//--- values for rampMode ---
#define  RAMP_OFF   0     // RF is OFF
#define  RAMP_WAIT  1     // RF chgd from OFF to ON < offToOnDelay ms ago
#define  RAMP_COLD  2     // RF ON, setpoint ramps using coldRampRate
#define  RAMP_WARM  3     // RF ON, setpoint ramps using warmRampRate

const char *ModeDesc[4] = { "OFF ", "WAIT", "COLD", "WARM" };

//--- values for statusBitsAppl ---
#define  STAT_RF_ON    0x0001  /* RF on/off */
#define  STAT_LIMITED  0x0002  /* setpoint limited */
#define  STAT_REG_ERR  0x0004  /* regulation error */
#define  STAT_REMOTE   0x0008  /* local/remote mode */
#define  STAT_CLOSED   0x0010  /* open/closed loop */

#endif


//--------------------------------------------------------------
//APPLCODE - set max # of registers per set of read/write registers
#ifndef NumR4Chans
  #define NumR4Chans_  NumLogicalDevices
  #define R4ChanMult_  100
#endif

#ifdef REG4

REG4( timeZone_,      x, 40001, S16, RW, MB_timeZone )
REG4( curTime,        x, 40002, U32, RW, setU32 )

REG4( ipAddr,         x, 40004, U32, RW, setU32 )
REG4( netmask,        x, 40006, U32, RW, setU32 )
REG4( gateway,        x, 40008, U32, RW, setU32 )

REG4( setACoil,       _, 40010, U16, RW, setU16 )
REG4( resetACoil,     _, 40011, U16, RW, setU16 )

REG4( digOutBits,     x, 40012, U32, RW, setU32 )

REG4( cmdBitsStd,     x, 40014, U32, RW, MB_cmdBitsStd )
REG4( cmdBitsAppl,    x, 40016, U32, RW, MB_cmdBitsAppl )

//APPLCODE - add registers for externally available read/write values
REG4( rfOnSetpoint,   x, 40018, F32, RW, MB_rfOnSetpoint )
REG4( rfOffSetpoint,  x, 40020, F32, RW, MB_rfOffSetpoint )

REG4( coldRampRate,   x, 40022, F32, RW, setF32 )
REG4( warmRampRate,   x, 40024, F32, RW, setF32 )

REG4( offToOnDelay,   x, 40026, U16, RW, setU16 )

REG4( maxRegErr,      x, 40027, F32, RW, setF32 )
REG4( limiterActiveV, x, 40029, F32, RW, setF32 )

REG4( knobRate,       x, 40031, F32, RW, setF32 )

REG4( maxOffSetpoint, x, 40033, F32, RW, setF32 )
REG4( maxOnSetpoint,  x, 40035, F32, RW, setF32 )

#undef REG4
#endif


#ifdef RegMacros
//===== application-specific defintions related to register values =====

//  #define  ISET(idx)   REG_ARRAY(float, iSet_1, idx)

#endif

//--------------------------------------------------------------
#undef RegMacros

#endif
#ifndef RegDefs // start of after-RegDefs portion




//=============================================================================
//  defintions for "parameterized" data values - lists of values that we want
//  to be "persistent" (i.e. saved to a file and restored on startup)
//=============================================================================
//APPLCODE - settings/values we want saved to/restored from files


//--- parameter ID's for data in Device Data/calibration file ---
//--------------------------------------------------------------------------
// WARNING:  ID's can be added, but existing ID's (and the format of the
//           data they represent) should NEVER change!  NOTE:  Existing ones
//           can be "eliminated" by changing their name and no longer
//           using them, but never REUSE an old one for something new!!!
//--------------------------------------------------------------------------
//  NOTE:  ID's >= 1000 are rsvd for values shared by most applications
//--------------------------------------------------------------------------

/*--- examples from sharedAppl.c ---*
#define  AIN_LABELS     1000
#define  AOUT_LABELS    1001
#define  DIN_LABELS     1002
#define  DOUT_LABELS    1003

#define  TIME_ZONE      1004
#define  SRFS_IP_ADDR   1005

#define  READ_GAINS     1006
#define  READ_OFFSETS   1007
#define  SET_GAINS      1008
#define  SET_OFFSETS    1009
*-----------------------------------*/

const ParamData far  devDataParams[] = {
//APPLDATA
//                ID   ---value---  --restoredFunc--
  PERSISTENT_VAR(  1,  coldRampRate,    NULL ),
  PERSISTENT_VAR(  2,  warmRampRate,    NULL ),
  PERSISTENT_VAR(  3,  offToOnDelay,    NULL ),
  PERSISTENT_VAR(  4,  maxRegErr,       NULL ),
  PERSISTENT_VAR(  6,  limiterActiveV,  NULL ),
  PERSISTENT_VAR(  7,  knobRate,        NULL ),
//  PERSISTENT_VAR(  8,  rfOffSetpoint,   NULL ),  // now in settings
//  PERSISTENT_VAR(  9,  rfOnSetpoint,    NULL ),  // now in settings
  PERSISTENT_VAR( 10,  maxOnSetpoint,   NULL ),
  PERSISTENT_VAR( 11,  maxOffSetpoint,  NULL ),


  #define SharedDevParams
  #include <sharedAppl.c>
/* Examples from the SharedDevParams entries in sharedAppl.c:
  PERSISTENT_VAR( TIME_ZONE,     timeZone,   NULL ),
  PERSISTENT_VAR( SRFS_IP_ADDR,  srfsIPAddr, NULL ),

  PERSISTENT_TBL( AIN_LABELS,    ainLabel,  MAX_LABEL_LEN, NULL ),
  PERSISTENT_TBL( AOUT_LABELS,   aoutLabel, MAX_LABEL_LEN, NULL ),
  PERSISTENT_TBL( DIN_LABELS,    dinLabel,  MAX_LABEL_LEN, NULL ),
  PERSISTENT_TBL( DOUT_LABELS,   doutLabel, MAX_LABEL_LEN, NULL ),

  PERSISTENT_TBL( READ_GAINS,    readGain   ,0, NULL ),
  PERSISTENT_TBL( READ_OFFSETS,  readOffset ,0, NULL ),
  PERSISTENT_TBL( SET_GAINS,     setGain    ,0, NULL ),
  PERSISTENT_TBL( SET_OFFSETS,   setOffset  ,0, NULL ),
*/
  END_PARAM_DATA
};

//----------------------------------------------------
const ParamData far  settingsParams[] = {
//APPLDATA
//                ID   ---value---  --restoredFunc--
  PERSISTENT_VAR(  1,  rfOffSetpoint,   NULL ),
  PERSISTENT_VAR(  2,  rfOnSetpoint,    NULL ),

  END_PARAM_DATA
};


//-----------------------------------------------------------------
//  Application-specific values needed to restore the state of the
//  controller following a "soft" reset (i.e. a restart without a
//  CPU reset)
//-----------------------------------------------------------------
//RESTORE_FUNC(myApplStateVar)

const ParamData far  ctlrStateParams[] = {
// sample values from sharedAppl.c, common to all applications
//                   --ID--  ---value,granularity---  --restoreFunc--
//    PERSISTENT_VARx(  1000,  prevState,               NULL ),
//    PERSISTENT_VARx(  1001,  curState,                NULL ),
//    PERSISTENT_VARx(  1002,  nextState,               NULL ),
//    PERSISTENT_TBL (  1003,  setVal,4,                NULL ),
// // PERSISTENT_TBL (  1004,  shadowRegValues,4,       NULL ),

//APPLDATA
//    PERSISTENT_VAR (  1,  myApplStateVar,        PVRF(myApplStateVar) ),


  #define SharedCtlrStateParams
  #include <sharedAppl.c>

  END_PARAM_DATA
};


//-----------------------------------------------------------------------------
//  A list of parameterized "persistent" data.  Backup copies of such values
//  are stored in files and restored on startup.
//-----------------------------------------------------------------------------

//    Base File Name    Category  ParamData
//    ----------------  --------  ---------
//--- add any application-specific parameterized file defs here ---
//ParamFile  myAppCfgFile =
//  { "myFileName",     PDC_CTLR, myCfgParams     };

ParamFile  settingsFile =
    { "settings",       PDC_SET,  settingsParams  };


#define SharedParams
#include <sharedAppl.c>


ParamFile *paramFiles[] = {
  #define SharedParamsList
  #include <sharedAppl.c>
//--- add any application-specific ParamData references here ---
//  &myAppCfgFile,
  &settingsFile
};

const int NumParamFiles = sizeof(paramFiles) / sizeof(paramFiles[0]);


//=============================================================================
//  Definitions for event msgs.  Event msgs are the means that other processes
//  use to inform the application process of an event that it needs to handle.
//
//  In general, allowing multiple processes to concurrently modify the state
//  of the controller is a sure recipie for chaos.  By restricting all such
//  changes to the context of a single process, and sending messages to that
//  process to request changes, we eliminate all such concurrency issues (not
//  the least of which is keeping track of the order of events).
//
//  The most common event is a request for change of state that corresponds to
//  a message received from an external client via a protocol like ModBus.
//=============================================================================

//EVENT_DEF( myEventX )
//EVENT_DEF( myEventY )



//=============================================================================
//  This is where misc global variables should be defined
//=============================================================================
#define SharedGlobals
#include <sharedAppl.c>

//APPLCODE
//--- add definitions for application-specific global variables here ---



//=============================================================================
//  Functions referenced in the Modbus register definitions to handle writes to
//  specific global values defined as ModBus registers
//=============================================================================

//APPLCODE - add functions to process writes to ModBus register values

//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for "cmdBitsAppl"
//
//  Check the bits that were set and trigger the corresponding events as
//  appropriate (but do NOT save the new value - just use it to trigger
//  events)
//-----------------------------------------------------------------------------
bool _nearcall MB_cmdBitsAppl(SetRegFuncParams)
{
/*
  ulong  cmdBits;
  EventData  eventMsg;  // used by EVENT macro


  cmdBits = *((ulong *)newVal);

  if (cmdBits & x)  EVENT(MyEventX, chan, newVal, sizeof(int)); // w/ data
  if (cmdBits & y)  EVENT(MyEVentY, chan, NULL, 0);             // w/o data
*/

  return true;
}


//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for the rfOnSetpoint register
//
//  Check the new setting and alter output values as appropriate
//-----------------------------------------------------------------------------
bool MB_rfOnSetpoint(SetRegFuncParams)
{
  if (Cur.localMode)  return false;

  _EnterCritical_();  rfOnSetpoint = MB_VAL(float);  _ExitCritical_();

  return true;
}


//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for the rfOffSetpoint register
//
//  Check the new setting and alter output values as appropriate
//-----------------------------------------------------------------------------
bool MB_rfOffSetpoint(SetRegFuncParams)
{
  if (Cur.localMode)  return false;

  _EnterCritical_();  rfOffSetpoint = MB_VAL(float);  _ExitCritical_();

  return true;
}



//-----------------------------------------------------------------------------
#define SharedFuncs
#include <sharedAppl.c>


//=============================================================================
//  Misc Application-specific functions
//=============================================================================
//APPLCODE



//=============================================================================
//  Task functions to handle application-specific commands
//=============================================================================
//APPLCODE

//-----------------------------------------------------------------------------
//  example func to process a command entered via a user shell
//-----------------------------------------------------------------------------
/*
void _task miscTest(int argc, char **argv)
{
  int  idx;
  bool  validCmd;

//used by EVENT macro
  fd_t  eventsfd = -1;
  EventMsg  eventMsg;


  validCmd = false;

  for (;;)  {
    if (argc < 3)  break;
    if (! isdigit(argv[1][0]))  break;

    idx = atoi(argv[1]);
    if ((idx < 1) OR (idx > NumChan))  {
      printf("Invalid chan #\n");  Exit(-2); }

    --idx;  validCmd = true;

    if (StriEq(argv[2], "x"))     { EVENT(MyEventX, idx, NULL, 0);  break; }
    if (StriEq(argv[2], "y"))     { EVENT(MyEventY, idx, NULL, 0);  break; }

    validCmd = false;  break;
  }

  if (! validCmd)  { printf("Usage:  misc # x|y\r\n");  Exit(-1); }

  Exit(0);
}
*/

//=============================================================================
//  Generic structure to handle application-specific presentation of the
//  digital and analog input/output values.  Code to handle additional
//  formatting that cannot be handled in this generic manner should be
//  placed in the showState() task routine.
//=============================================================================
//APPLCODE - text and formatting info for application state summary

//-----------------------------------------------------------------------------
_nearcall char *fmtMode(DispElem far *elem, char *buf)
{
  return (char *) ModeDesc[ELEM_VAL(U16)];
}

//-----------------------------------------------------------------------------
_nearcall char *fmtHdwCfg(DispElem far *elem, char *buf)
{
  return (char *) HdwCfgDesc[ELEM_VAL(U16)];
}

//-----------------------------------------------------------------------------
  const char far DevStateStr[] = {
"Cavity Voltage: -###.### kV    Open/Closed: ######    HdwConfig: ##### \r\n"
"Active setpoint: ###.### kV    RF On/Off:   ### \r\n"
"Error:          -###.### kV    Loc/Rem:     ###### \r\n"
" \r\n"
"Limiter Voltage: -##.### V     Limited:     ### \r\n"
"                               Regulation:  ##### \r\n"
"RF OFF setpoint: ###.### kV \r\n"
"RF ON setpoint:  ###.### kV    Ramp Mode:   #### \r\n"
" \r\n"
};

const DispElem  far dispElems[] = {
  { 16, 3, VFPV(cavityVolts),     DARK_BLUE, fmtF32,    8,3 },
  { 17, 4, VFPV(activeSetpoint),  DARK_BLUE, fmtF32,    7,3 },
  { 16, 5, VFPV(regulationError), DARK_BLUE, fmtF32,    8,3 },

  { 17, 7, VFPV(limiterVolts),    DARK_BLUE, fmtF32,    7,3 },

  { 17, 9, VFPV(rfOffSetpoint),   DARK_BLUE, fmtF32,    7,3 },
  { 17,10, VFPV(rfOnSetpoint),    DARK_BLUE, fmtF32,    7,3 },

  { 44,10, VFPV(rampMode),        DARK_BLUE, fmtMode,   4,4 },

  { 65, 3, VFPV(hdwCfg),          DARK_BLUE, fmtHdwCfg, 5,5 },

  { 0,0, NULL, NULL, NULL, 0,0 }
};


const DigDispElem far digInElems[] = {
  { 44, 3,  CLOSED_LOOP,  "OPEN  ",DARK_RED,   "CLOSED",DARK_BLUE },
  { 44, 4,  RF_IS_OFF,    "ON ",DARK_BLUE,     "OFF",DARK_RED },
  { 44, 5,  LOC_REM_SW,   "REMOTE",DARK_BLUE,  "LOCAL ",DARK_BLUE },

  { 0,0, 0, NULL,NULL, NULL,NULL }
};

const DigDispElem far digOutElems[] = {
  { 44, 7,  LIMITED_LED,   "NO ",  DARK_BLUE,  "YES",  DARK_RED },
  { 44, 8,  ERROR_LED,     "OK   ",DARK_BLUE,  "ERROR",DARK_RED },

  { 0,0,  0,  NULL,NULL,  NULL,NULL }
};


//---- older style - do not use for newer applications -----
const DigFmtInfo digInFmt[] = {
//{ xx,yy, "0-str"  ,1,  "1-str" ,0 },  //  descr

  { -1,-1, "",0, "",0 }
};

const DigFmtInfo digOutFmt[] = {
//{ xx,yy, "0-str"  ,1,  "1-str" ,0 },  //  descr

  { -1,-1, "",0, "",0 }
};


const AnalogFmtInfo  adcFmt[] = {
  { -1,-1, 0,0 }
};

const AnalogFmtInfo  dacFmt[] = {
  { -1,-1, 0,0 }
};


//-----------------------------------------------------------------------------
//  "state":  Show continuously updated state of the system
//-----------------------------------------------------------------------------
void _task showState(int argc, char **argv)
{
  bool  blinkOff;
  char  temp[128];


  printf(clsCmd);  blinkOff = false;

  CursorTo(0, 3);  printf(DevStateStr);

  for (;;)  {
    CursorTo(0, 0);  blinkOff = !blinkOff;

    stdHeader();  displayElems((DispElem far *)dispElems, blinkOff, temp);

    displayDigElems(digitalIn, (DigDispElem far *)digInElems, blinkOff, temp);
    displayDigElems(digitalOut, (DigDispElem far *)digOutElems, blinkOff, temp);

    CursorTo(0,22);  fflush(stdout);  CheckETX();  sleepMS(500);
  }

}



//=============================================================================
//  Event-Handler functions called by the appl task to process event messages
//=============================================================================
//APPLCODE

//-----------------------------------------------------------------------------
//  Event-Handler for a request to ...
//-----------------------------------------------------------------------------
/*
void EVENT_MyEvent(EventMsg *event)
{
  uint  idx;
  float  setpoint;
  char  temp[16];


  idx = event->chanNum;

  if (error)  {
    fprintf(DiagLog, "error msg, chan:\n", idx+1);
    return;
  }

  setpoint = *((float *)event->data);

  fprintf(DiagLog, "New SetPoint for PS%d: %s\n",
     idx+1, fmtFloat(setpoint, 8,3, temp) );

  *((float *)((&iSet_1_RegInfo + idx)->regVal)) = setpoint;

  anaOut[idx] = ISETtoVolts(idx, setpoint);
}
*/


//=============================================================================
//  Update the the state of the application logic in well-defined phases
//=============================================================================
// IMPORTANT NOTES and RULES:
//
//  - The code to change ANY value must be in ONE place.  NEVER spread the
//    logic for updating a single value around.  There should be ONE place
//    to look to see how a value gets set.
//
//  - "cur" values are basically "inputs" - literally (they are set based
//    on the just-read state of a digital or analog input) and/or figuratively
//    (they hold a value set by some external source, such as a cmd from a
//    remote client or another user-interface task).
//
//  - "next" values are basically "outputs" - literally (they determine the
//    next state of a digital or analog output) and/or figuratively (they
//    can affect the state of the logic for the next pass).
//
//  - The "cur" OR the "next" instance of a value may be changed - NEVER
//    both of them (with the exception of the first step below that copies
//    all of cur to prev and next to cur).
//
//  - A value MUST be in the ApplState struct if one of the following applies:
//     - The application logic is dependent on if or how much the value has
//       changed between passes.
//     - The value is used as both input (used in in a conditional and/or in
//       an expression that can affect the result of an assignment) and output
//       (the target of an assignment).
//
//  - Variables that are NOT in the ApplState struct but are USED (in
//    conditionals or other expressions) by the application logic must be
//    treated just like "cur" values.  Variables SET (the target in an
//    assignment statement) by the application logic must be treated just
//    like "next" values.
//
//  - When the logic requires comparing an "input" to a related "output", be
//    sure to use the "cur" value for the "input", and the "prev" value for
//    the "output".
//
//    This is because during the processing of the loop, the "cur" values
//    represent the input values just read (what they are "now") and the
//    "prev" values represent the values the outputs were set to at the
//    beginning of the prev pass (i.e. they are the values that "caused"
//    the inputs to have their "cur" values).
//
//    During a pass the output values in "cur" represent what the outputs
//    where set to AFTER the "cur" inputs were updated, so comparing them
//    makes no sense.  Doing so is the equivalent of comparing a past response
//    to a condition that didn't exist at the time.
//=============================================================================


//=============================================================================
//  Definitions for functions that handle setting "next" values that serve
//  as inputs to the application control logic (run by the "applic" task in
//  a periodically executed loop).
//
//  While care must be taken in these functions to use locks to avoid the
//  possibility of being preempted in the middle of an operation, such locks
//  must be kept to as short a time as possible to avoid excessive latency
//  in the main control loop.
//
//  For REALLY short locks, using a critical section is best, because this
//  has the least overhead (but it has to be REALLY short to avoid causing
//  problems for interrupt handlers).
//
//  For longer locks (ones that involve anything but the most trivial
//  conditional logic), use either KLOCK() and KUNLOCK() (which disable the
//  scheduler) or a more specific global semaphore/mutex lock.
//=============================================================================

/*Examples
//-----------------------------------------------------------------------------
void onRequest(void)
{
  _EnterCritical_();
  Next.onRequest = true;  Next.offRequest = false;
  _ExitCritical_();

  fprintf(DiagLog, "ON request\n");
}

//-----------------------------------------------------------------------------
void offRequest(void)
{
  _EnterCritical_();
  Next.offRequest = true;  Next.onRequest = false;
  _ExitCritical_();

  fprintf(DiagLog, "OFF request\n");
}

//-----------------------------------------------------------------------------
void resetRequest(void)
{
  _EnterCritical_();  Next.resetRequest = true;  _ExitCritical_();

  fprintf(DiagLog, "RESET request\n");
}

//-----------------------------------------------------------------------------
void newSetpoint(float newAmps)
{
  char  temp[16];


  KLOCK();

  if (FABS(newAmps) > maxAmpSet)  newAmps = (float)SIGN(newAmps) * maxAmpSet;

  Next.ampSet = newAmps;

  KUNLOCK();

  fprintf(DiagLog, "New Setpoint: %s\n", fmtFloat(newAmps, 10,3, temp) );
}

//-----------------------------------------------------------------------------
*/



/*
=========== NOTES on Locking ============

The applic task runs at the highest priority (even above the kernel), so it
cannot be preempted by any other task.  As long as this is true, the applic
task does not need to use locks when accessing global variables.

However, all other tasks MUST disable the scheduler when accessing values
used by the applic task to avoid being preempted while doing so.  Care must
be taken to keep these times short to avoid latency issues.

For REALLY short times, a critical section can be used as this is much more
efficient than disabling and reenabling the scheduler.  But critical sections
disable all interrupts, so it is not practical in most cases.

*/

//=============================================================================
//  Update the "Cur"rent state of the controller
//
//  Be sure to read the "NOTES on Locking" section above....
//=============================================================================
//
//  Just before this function is called, readings were taken of the current
//  state of the hardware, and any changes to the output states specified by
//  the last pass were applied (just after taking the readings).  These values
//  are copied to Next as needed and constitute the Hardware Inputs.
//
//  Other tasks can affect the state of the control logic by changing values
//  in Next as well.  These values constitute Software Inputs.
//
//  This function is responsible for the Asynchronous logic that is applied
//  each pass.  It updates values in Next (or other global variables) based
//  on the Software and Hardware Input values in Next.  This portion of the
//  logic is analogous to unclocked gates in an electronic circuit.
//
//  After the Asychronous logic is applied, we set Prev = Cur and Cur = Next.
//  The result is that Prev contains the controller state as it was during the
//  previous pass, and Cur indicates the current state of the controller.
//
//  Also, see the "NOTES on Locking" section above....
//=============================================================================
void updateCurState(void)
{

  if (current != applTask)  return;  // must only be called by "applic" task

//-----------------------------------------------------------------
// The following conditions exist at this point:
// - The analog and digital hardware inputs have just been read
// - The outputs were just set to values computed during the last pass
// - "Cur" contains the state the inputs were in at the START of
//   of the previous pass, and the state the outputs were in DURING
//   the previous pass
// - "Next" contains the values that the outputs were just set
//   to and any "input" values that were changed by other tasks
//------------------------------------------------------------------

// Now we update the values in "Next" that are based solely on the
// state of the hardware inputs we just read and values maintained by
// by other tasks (basically, we deal with READINGS here: Values that
// we need but are not written to "Next" by another task)

  Next.processTime = GetMS();  curTime = time(NULL);

  msSincePrev = msSince(Cur.processTime);

  Next.digitalIn = digitalIn;


//APPLCODE
  initialBtnPressed = InitialBtnPushed(Next.digitalIn);
  finalBtnPressed = FinalBtnPushed(Next.digitalIn);
  hdwCfg = HdwConfig(readVal[HDW_CFG]);

  if (hdwCfg > 7)  hdwCfg = ERROR_CFG;

  // (ignore if both buttons being pressed)
  if (initialBtnPressed AND finalBtnPressed)
    initialBtnPressed = finalBtnPressed = false;


  cavityVolts = readVal[RF_CAV_VOLT];
  limiterVolts = readVal[LIMITER_VOLT];

  Next.rfIsOn = !RFisOff(Next.digitalIn);

  Next.localMode = LocalMode(Next.digitalIn);
  Next.closedLoop = ClosedLoop(Next.digitalIn);


  regulationError = cavityVolts - Prev.activeSetpoint;

  limited = (FABS(limiterVolts) > limiterActiveV);
  regErr = ((FABS(regulationError) > maxRegErr) AND Next.rfIsOn);

  _EnterCritical_();  knobTicks = qdCount;  qdCount = 0;  _ExitCritical_();



// Next we make a copy of the application state before we apply all the
// logic that updates the overall state.  This allows us to have a copy of
// all the relevant values that remain unchanged during the processing,
// avoiding all kinds of order-of-execution dependencies.

// We also make a copy of the state from last time before overwriting it
// so we can see if/how values have changed since the last pass

// Note that a copy is done (rather than just switching pointer values) to
// insure all values are set in both the source and destination.  If we
// simply kept pointers and swapped the pointer values, then EVERY value
// would have to be updated independently by the code, which is a pain to
// program, adds overhead and code, and is a VERY easy way to introduce odd
// bugs.  Overall, copying everything here is actually faster overall,
// simpler, and much safer (and is now required by the new softReset code).

  Prev = Cur;  Cur = Next;

//--- At this point the Cur copy contains what the inputs were at the start ---
//--- of this pass, and the Prev copy contains the output values from the ---
//--- the previous pass (i.e. the output values that resulted in the input ---
//--- values in Cur) ---

//--- log msgs for bit chgs only if not starving other processes for cycles ---
  if (applLoopTime < 6)  logDigIOChgs(false);




//--- At this point the Cur copy contains what the inputs were at the start ---
//--- of this pass, and the Prev copy contains the output values from the ---
//--- the previous pass (i.e. the output values that resulted in the input ---
//--- values in Cur) ---



//--------------------------------------------------------------------
// Next we update the rest of the "cur" values needed by the control
// logic (the ones that do NOT come from "next" values set by other
// tasks)
//
// NOTES:
//  - This portion of the logic has order-of-execution dependencies,
//    so be sure that all values referenced in an expression are ones
//    set before they are used.
//
//  - References to "prev" values are always safe since they are just
//    a copy of values from the last pass
//--------------------------------------------------------------------

 //------------------------------------------------------------------
 // rampMode, rfOnTime
 //    Description:
 //       Specifies how changes to Next.activeSetpoint should be
 //       handled.
 //
 //       If the RF is OFF, then it is set to rfOffSetpoint.
 //
 //       If the RF just changed from OFF to ON, then an interlval
 //       of offToOnDelay ms must pass before any changes are made.
 //
 //       If the RF is ON and offToOnDelay ms have passed, then
 //       it is ramped toward the rfOnSetpoint at the rate specified
 //       by coldRampRate.
 //
 //       If the RF is ON, the activeSetpoint has reached the
 //       rfOnSetpoint, then any difference between rfOnSetpoint and
 //       activeSetepoint is reduced at the rate specified by
 //       warmRampRate.
 //
 //    Inputs:
 //       offToOnDelay
 //       rampMode
 //       Cur.activeSetpoint
 //       Cur.rfOnSetpoint
 //       Prev.rfIsOff
 //       Cur.rfIsOff
 //------------------------------------------------------------------
  if (!Cur.rfIsOn)  rampMode = RAMP_OFF;

  switch (rampMode)  {
    case RAMP_OFF:  // waiting for RF ON
      if (Cur.rfIsOn)  {
        rampMode = RAMP_WAIT;  rfOnTime = GetMS();
        printf("Starting OFF-to-ON delay\r\n");
      }
      break;

    case RAMP_WAIT: // waiting for offToOnDelay to finish
      if (msSince(rfOnTime) >= offToOnDelay)  {
        rampMode = RAMP_COLD;
        printf("Ramping using COLD rate\r\n");
      }
      break;

    case RAMP_COLD:  // waiting for cold-mode ramping to finish
      if (Cur.activeSetpoint == Cur.rfOnSetpoint)  {
        rampMode = RAMP_WARM;
        printf("Using WARM rate for changes\r\n");
      }
      break;
  }

}


//=============================================================================
//  Now we update the values in "next" based on the "cur" and "prev"
//  values.
//=============================================================================
//  All values changed by updateNextState() are "outputs" from this func.
//  All other values referenced in this function are "inputs" to this func.
//
//  All the values in "next" provide information about the next logical state
//  for the controller:  Values to be applied during the next pass of the
//  control logic.  All "next" values fall in to one of 2 categories:
//
//    - They are set/updated ONLY by this function (which is executed only
//      by the "applic" task)
//
//    - They are set ONLY by other tasks (within a lock), and possibly
//      reset/cleared by the "applic" task in updateCurState() (after copying
//      "next" to "cur")
//
//
//  All values that act as input to this func must fall in to one of these
//  categories:
//
//    - "prev" values
//
//    - "cur" values updated directly by statements in updateCurState()
//
//    - "cur" values updated indirectly by updateCurState() due to the copy
//      of "next" to "cur", which includes the values in "next" that are
//      set by other tasks
//
//    - a shared/global value that is the equivalent of a "cur" value but
//      for which we don't need to track changes
//
//
//  All "outputs" from this function must be altered ONLY within this function.
//
//  All "inputs" to this function must NEVER be altered within this function.
//
//  Each output value (or VERY small "set" of ouput values) must be updated
//  in a single, self-contained block of code.  NO output may be referenced
//  OR altered in ANOTHER block of code:  This is a declarative style of
//  programming and has multiple advantages:
//
//     - There are no order-of-execution dependencies across nodes/blocks
//       within a pass.  They can be executed in any order and the result
//       will be the same (get used to it - this method will be the norm as
//       computer languages and hardware add support for atomic transactions
//       and computers evolve in to neural-net-like processor arrays - FPGAs
//       are just the beginning of this trend).
//
//     - Encapsulating all logic for an output within a single node/block
//       leaves just one place to start when tracking a problem backward from
//       the result (far simpler than guessing at the possible causes and
//       trying to work forward, looking for paths to the observed result).
//
//     - Drastically reduces the complexity of the code by greatly reducing
//       the number of possible paths through the code.  This simplifies
//       testing and verification processes.
//
//  Each node/block-of-code should be prefaced with the following information:
//   - The names of the values it modifies
//   - A descr of the meaning of the values and the logic used to set them
//   - A list of the values that serve as "inputs" to the node
//
//=============================================================================
void updateNextState(void)
{
//  char  temp[64];

//--------------------------------------------------------------------
// Next.activeSetpoint
//    Description:
//       Determines the setpoint to be applied to the regulation
//       circuit at the beginning of the next pass.
//       See rampMode for a description of how changes are handled
//    Inputs:
//       rampMode
//       msSincePrev
//       Cur.activeSetpoint
//       Cur.rfOffSetpoint
//       Cur.rfOnSetpoint
//       coldRampRate
//       warmRampRate
//----------------------------------------------------------------------
  switch (rampMode)  {
    case RAMP_OFF:  // RF is OFF - use rfOffSetpoint
      Next.activeSetpoint = Cur.rfOffSetpoint;  break;

    case RAMP_WAIT: // delay after recent OFF-to-ON
      Next.activeSetpoint = Cur.activeSetpoint;  break;

    case RAMP_COLD: // use the (presumably) slower cold rate
    case RAMP_WARM: // use the (presumably) faster warm rate
      if ((diff = Cur.rfOnSetpoint - Cur.activeSetpoint) == 0.0)  {
        Next.activeSetpoint = Cur.activeSetpoint;  break;  }
      maxChg =
        (rampMode == RAMP_COLD ? coldRampRate : warmRampRate);
      maxChg = maxChg * (float)msSincePrev / 1000.0;
      if (FABS(diff) > maxChg)  diff = (diff < 0.0 ? -maxChg : maxChg);
      Next.activeSetpoint = Cur.activeSetpoint + diff;
      break;
  }


//--------------------------------------------------------------------
// Next.rfOffSetpoint
//    Description:
//       The setpoint used whenever the RF is OFF.  Can be set by
//       remote clients or the Front Panel knob.
//    Inputs:
//       knobTicks
//       initialBtnPressed
//       rfOffSetpoint (Modbus register)
//----------------------------------------------------------------------
  Next.rfOffSetpoint = rfOffSetpoint;

  if (knobTicks AND initialBtnPressed)  {
    Next.rfOffSetpoint = Cur.rfOffSetpoint + (float) knobTicks * knobRate;
//    printf("Knob change to rfOffSetpoint\r\n");
  }

  if (Next.rfOffSetpoint < 0.0)
    Next.rfOffSetpoint = 0.0;
  if (Next.rfOffSetpoint > maxOffSetpoint)
    Next.rfOffSetpoint = maxOffSetpoint;

//  if (Next.rfOffSetpoint != rfOffSetpoint)
//    printf("rfOffSetpoint changed to %s\r\n",
//           fmtFloat(Next.rfOffSetpoint, 7, 3, temp));


//--------------------------------------------------------------------
// Next.rfOnSetpoint
//    Description:
//       The setpoint toward which the active setpoint is adjusted
//       whenever the RF is ON.
//    Inputs:
//       knobTicks
//       finalBtnPressed
//       Cur.localMode
//       Cur.rfOffSetpoint
//       maxOnSetpoint (Modbus register)
//       rfOnSetpoint (Modbus register)
//----------------------------------------------------------------------
  Next.rfOnSetpoint = rfOnSetpoint;

  if (knobTicks AND Cur.localMode)  {
    Next.rfOnSetpoint = Cur.rfOnSetpoint + (float) knobTicks * knobRate;
//    printf("Knob change to rfOnSetpoint\r\n");
  }

  if (Next.rfOnSetpoint < Cur.rfOffSetpoint)
    Next.rfOnSetpoint = Cur.rfOffSetpoint;
  if (Next.rfOnSetpoint > maxOnSetpoint)
    Next.rfOnSetpoint = maxOnSetpoint;

//  if (Next.rfOnSetpoint != rfOnSetpoint)
//    printf("rfOnSetpoint changed to %s\r\n",
//           fmtFloat(Next.rfOnSetpoint, 7, 3, temp));
}


//=============================================================================
//  Next we update output signal values to be written to the outputs at
//  the beginning of the next pass (just after the inputs are read).
//=============================================================================
void updateOutputState(void)
{

//--- update output values that are based on current conditions (values ---
//--- in "cur" or any values used in a conditional in the applic logic) ---
  // (insured earlier that only ONE of these can be true when setting them)
  if (initialBtnPressed)  setVal[FP_METER] = rfOffSetpoint;  else
  if (finalBtnPressed)    setVal[FP_METER] = rfOnSetpoint;  else
                          setVal[FP_METER] = cavityVolts;


  SetOffApplLEDs(Next.digitalOut);

  if (limited)  SetOnLimitLED(Next.digitalOut);
  if (regErr)   SetOnErrorLED(Next.digitalOut);


//--- update output values based on values that were set by the application ---
//--- logic (e.g. values in "next" or variables that are the target of an ---
//--- assignment statement in the application logic) ---
  setVal[SETPOINT] = Next.activeSetpoint;

}



//=============================================================================
// The last thing we do is update values read by other tasks
//=============================================================================
// IMPORTANT NOTES and WARNINGS:
//
//   To avoid concurrency issues, other tasks need to use locks when reading
//   values set here.  But be careful to keep such locks short to avoid
//   delaying the timing of the main application task.  If necessary, split
//   such code in to sections, getting and releasing the lock(s) in each
//   section to avoid increasing latency for the main application task.
//
//   Values in the "next" instance of the application state fall in to 2
//   categories:  Inputs and Outputs.
//
//   The "next" Outputs are those that may be changed in updateOutputState().
//   The reset are Inputs, and may be changed by other tasks within a lock.
//
//   Since each pass through the application logic starts by setting
//   cur = next, any "next" values set by other tasks become part of the
//   "cur" instance used by the rest of the applic logic.  Because the
//   "applic" task at a higher priority than any other real-time task, the
//   changes to "next" values by other tasks can ONLY happen between passes
//   through the applic logic loop.
//
//   NOTES:
//     - Some of these "next" input values values must be initialized in
//       updateCurState() in order to tell when a value is set by another
//       task.  For example, if we have a boolean flag that is set when we
//       get an ON command, then (after its "next" value is copied to "cur"),
//       we have to clear the flag in order to tell if another ON command is
//       received between passes.
//
//     - As with reads to these values, other tasks must use locks when
//       writing to the "next" input values to insure they are not preempted
//       by the "applic" task in the middle of a write.
//
//=============================================================================
void updateExternals(void)
{
//--- update settings based on values that were USED by the application ---
//--- logic (i.e. "cur" values that serve as "inputs" to the logic) ---
  activeSetpoint = Cur.activeSetpoint;

  statusBitsAppl = 0;
  if (Cur.rfIsOn)      statusBitsAppl |= STAT_RF_ON;
  if (limited)         statusBitsAppl |= STAT_LIMITED;
  if (regErr)          statusBitsAppl |= STAT_REG_ERR;
  if (!Cur.localMode)  statusBitsAppl |= STAT_REMOTE;
  if (Cur.closedLoop)  statusBitsAppl |= STAT_CLOSED;


//--- update settings based on values that were UPDATED by the application ---
//--- logic (i.e. "next" values that serve as "outputs" from the logic) ---
  rfOffSetpoint = Next.rfOffSetpoint;
  rfOnSetpoint = Next.rfOnSetpoint;


//--- temp method of setting read/set gains based on hardware config ---
  switch (hdwCfg)  {
    case K500_CFG:
      readGain[RF_CAV_VOLT] = 10.0;
      setGain[SETPOINT] = setGain[FP_METER] = 10.0;
      break;

    case K1200_CFG:
      readGain[RF_CAV_VOLT] = 20.0;
      setGain[SETPOINT] = setGain[FP_METER] = 20.0;
      break;
  }

  hdwCfgVolts = readVal[HDW_CFG];
}



//=============================================================================
_nearcall float readGetRaw(uint ui)  {
  return (ui < NumHdwADC ? anaIn[ui] : 0.0);  }

_nearcall float readMinRaw(uint ui)  {
  return (ui < NumHdwADC) ? (adcMinRaw(ui) - adcOffset[ui]) * adcGain[ui] : 0.0; }

_nearcall float readMaxRaw(uint ui)  {
  return (ui < NumHdwADC) ? (adcMaxRaw(ui) - adcOffset[ui]) * adcGain[ui] : 9.8; }

_nearcall void readDefaults(uint ui, float *gain, float *offset)  {
  *gain = ((ui == 0) ? 10.0 : 1.0);  *offset = 0.0; }

_nearcall void readCalibPoints(uint ui, float *low, float *high)  {
  *low = 0.0;  *high = 90.0; }

_nearcall char *readUnits(uint ui, char *buf)  { return "kV"; }


//=============================================================================
_nearcall float setGetRaw(uint ui)  {
  return (ui < NumHdwADC ? anaOut[ui] : 0.0);  }

_nearcall float setMinRaw(uint ui)  {
  return (ui < NumHdwDAC) ? (dacMinRaw(ui) - dacOffset[ui]) * dacGain[ui] : 0.0; }

_nearcall float setMaxRaw(uint ui)  {
  return (ui < NumHdwDAC) ? (dacMaxRaw(ui) - dacOffset[ui]) * dacGain[ui] : 9.8; }

_nearcall void setDefaults(uint ui, float *gain, float *offset)  {
  *gain = ((ui < 2) ? 10.0 : 1.0);  *offset = 0.0; }

_nearcall void setCalibPoints(uint ui, float *low, float *high)  {
  *low = 0.0;  *high = 90.0;  }

_nearcall char *setUnits(uint ui, char *buf)  { return "kV"; }


//=============================================================================
//  Set default labels for all the I/O signals
//=============================================================================
void setDefaultLabels(void)
{
  printf("Setting default labels for I/O signals\r\n");

//--- set application-appropriate default labels ---
//APPLCODE

//--- set default labels for the I/O signals ---
  fstrcpy(ainLabel[0], "RF_Cavity");
  fstrcpy(ainLabel[1], "Limiter");
  fstrcpy(ainLabel[2], "<unused>");
  fstrcpy(ainLabel[3], "HdwConfig");

  fstrcpy(aoutLabel[0], "Setpoint");
  fstrcpy(aoutLabel[1], "FP_Meter");

  fstrcpy(dinLabel[0], "Closed_Loop");
  fstrcpy(dinLabel[1], "FINAL_btn");
  fstrcpy(dinLabel[2], "INITIAL_btn");
  fstrcpy(dinLabel[3], "RF_OFF");
  fstrcpy(dinLabel[4], "Local_mode");

  fstrcpy(doutLabel[0], "Limited_LED");
  fstrcpy(doutLabel[1], "Error_LED");
  fstrcpy(doutLabel[8], "Activity_LED");
}


//=============================================================================
//  This function is called to force software configurable values to default
//  values.
//=============================================================================
void setDefaultConfig(void)
{
  printf("Setting in-RAM device config to default values\r\n");

//===== set default values for persistent data values =====
//APPLCODE

  coldRampRate =  50.0;  // in kV/sec
  warmRampRate = 100.0;  // in kV/sec

  offToOnDelay = 250;     // in milliseconds

  maxRegErr = 1.0;    // in kV

  hdwCfg = HdwConfig(digitalIn);

  switch (hdwCfg)  {
    case K500_CFG:
      limiterActiveV = 0.310;  // in kV
      maxOffSetpoint = 30.0;   // in kV
      maxOnSetpoint = 100.0;   // in kV
      break;

    case K1200_CFG:
      limiterActiveV = 0.310; // in kV
      maxOffSetpoint = 60.0;  // in kV
      maxOnSetpoint = 200.0;  // in kV
      break;

    default:
      limiterActiveV = 0.310; // in kV
      maxOffSetpoint = 60.0;  // in kV
      maxOnSetpoint = 200.0;  // in kV
  }

  knobRate = 0.05;
}


//=============================================================================
//  This function is called by the application loop just before a shutdown.
//  The appliation must do whatever is necessary for a clean shutdown and
//  then exit
//=============================================================================
void applicShutdown(void)
{
  printf("Application prepping for shutdown\r\n");  fflush(stdout);


  Exit(0);
}


//=============================================================================
//  This is the highest priority task, usually resumed at a regular interval
//  so that it can do what it needs and sleep again until the start of the
//  next interval.
//=============================================================================
void _task applic(int argc, char **argv)
{
  #define SharedApplLocals
  #include <sharedAppl.c>

//--- add local variable definitions here ---


  #define SharedApplLocals2
  #include <sharedAppl.c>

//--- add local array definitions here ---


//--- place code here to set initial default values ---
  allowSoftReset = true;


  #define SharedApplInit
  #include <sharedAppl.c>

//--- add application initialization code here ---


  #define SharedApplInit2
  #include <sharedAppl.c>


//--- The following loop is run at a periodic interval (normally 100 times ---
//--- per second, or every 10ms).  The application process normally ---
//--- consists of reading the current state of the digital and analog ---
//--- inputs, changing the state of the analog and digital outputs to the ---
//--- values computed during the previous pass, then computing new values ---
//--- for the outputs and updating values that outside clients can view ---
//--- (currently accessable as a set of ModBus registers) ---

  for (;;)  {
    #define SharedApplLoopStart
    #include <sharedAppl.c>

  //--- read the current state of the analog and digital Input signals, and ---
  //--- write the values computed at the end of the last pass ---
    digitalOut = Next.digitalOut;  rwIO();

  //--- process next event message (if any) ---
    processMsg(eventsfd, (EventMsg *)buf, sizeof(buf));

  //--- update the current state and determine the next state ---
    updateCurState();  updateNextState();
    updateOutputState();  updateExternals();


    #define SharedApplLoopEnd
    #include <sharedAppl.c>
  }

}


//=============================================================================
//  list of application-specific commands that are normally entered via a
//  user shell
//=============================================================================
//void _task nsclSeriald(int argc, char **argv);

//--- add any application-specific commands here ---
//APPLCODE
const CmdInfo far applCmds[] = {
//  { "misc",        miscTest,       "..." },
  { "state",     showState,   "Show state of controller" }
};

int const NumApplCmds = ( sizeof(applCmds) / sizeof(CmdInfo) );

//-----------------------------------------------------------------------------

#endif
