//----- ECROvenAppl.c ----- 10/01/12 -----(03/05/06)----

// EE Documents bundle # 100


// Sections that typically require application-specific code are marked with
// the comment:  //APPLCODE

//-----------------------------------------------------------------------------
//
//  Copyright (c) 2005-2010 National Superconducting Cyclotron Laboratory,
//            Michigan State University, and Mark Davis
//
//-----------------------------------------------------------------------------
//  Application specific code for the main application task that is run at
//  a fixed interval and application specific commands run on demand.
//
//APPLCODE
//  Program to control serveral power supplies that drive an induction heating
//  oven for the Electron Cyclotron Resonance (ECR) facility.
//-----------------------------------------------------------------------------
//
//  See NOTES section regarding issue of what happens when a bit is set in the
//  Alarm Status register.
//
//
//  2012-10-01 - Mark Davis
//               - Added support for sputter interlock input (including
//                 adding the SPTR_INTERLOCK bit to statusBitsAppl).
//
//  2012-09-26 - Mark Davis (*)
//               - Change ovenCom() so it uses blocking mode when writing to
//                 the fd for the serial port.  Prior to this some of the data
//                 might not have been written to the port's buffer causing
//                 the supply to get incomplete commands.
//
//  2012-09-11 - Mark Davis (*)
//               - Added support for the softReset mode.  While this keeps the
//                 outputs from changing state during a soft reset, it does
//                 NOT continue to generate pulses for the external watchdog
//                 timer, so any supplies that were on when the soft reset
//                 began will temporarily loose power during the reset (which
//                 should be a period of less than 30 seconds).  This change
//                 was done because the original ECROven box had its memory
//                 corrupted at least once.
//
//  2011-03-31 - Mark Davis
//               - Eliminated a typo that was causing the formatting of
//                 values obtained via the serial connections to the oven
//                 supplies to be messed up when they are logged because an
//                 error was reported by the PS.
//               - Modified the wording of the logged descriptions of the
//                 Alarm Status Register bits to agree with the wording in
//                 the manual for the Lambda (OVEN) supplies.
//
//  2011-02-16 - Mark Davis
//               - Changed the formatting of some of the numeric values from
//                 a fixed-point floating format to an "engineering" one
//                 because the values often would not fit in the allotted
//                 space for the fixed format.
//
//  2010-03-01 - Mark Davis
//               - Fixed a bug in ovenCom() that could cause it to write past
//                 the end of the buffer it uses to accumulate the response
//                 from the power supply.
//
//  2010-02-23 - Mark Davis
//               - The "state" command now shows the Sputter supply current
//                 in mA rather than Amps.
//               - Changed the logical order for displaying the state of the
//                 digital outputs (used by the "ios" command).
//
//  2010-02-16 - Mark Davis
//               - Changed ovenCom() so it will only send the OUT0/1 command
//                 if the output state is different than the desired one.  This
//                 should eliminate the problem where the Oven supply
//                 occasionally turns its output off for 1 second when the
//                 serial connection is active (this assumes that the reason
//                 for this is that the commands are somehow getting garbled
//                 once in a while).
//
//  2010-02-15 - Mark Davis
//               - Changed ovenCom() to use readBufCh() to avoid the overhead
//                 incured by using the fcntl layer to buffer single-char
//                 reads.  This changed allowed me to increase the rate of
//                 polling a little bit while still reducing the CPU usage.
//
//  2010-02-10 - Mark Davis
//               - Changed the rate of the pulses for the external watchdog
//                 timer to 10Hz (it was 5Hz)
//               - Removed the code that was setting the setpoints for all
//                 4 supplies to 0 AFTER their values were restored from the
//                 settings file.
//
//  2010-02-01 - Mark Davis
//               - Removed RF Gen supply, added 2nd Oven supply
//
//  2009-12-14 - Mark Davis
//               - Changed order of param files to deal with problems caused
//                 by the order in which they were restored at startup.
//
//  2009-07-15 - Mark Davis -
//               - Corrected count for writing :DCL; command to reset alarm
//                 states.  Basically harmless as the old count simply caused
//                 the terminating 0 byte to be written as well.
//
//  2009-06-11 - Mark Davis -
//               - Added checks to compare on/off commands to the current
//                 state and log msgs for on/off commands that have no effect.
//
//  2009-06-10 - Mark Davis -
//               - The power for the Oven supply is now handled as follows:
//                  - When turning ON, the power is enabled, and then (after
//                    a few seconds) the output is enabled.
//                  - When turning OFF, the output is disabled and the control
//                    voltage set to 0.  Then, after a (configurable) delay
//                    period has passed, the power is disabled.
//                  This is more like the old code, but avoids "yanking" the
//                  power while the supply has an active output.
//               - The on, off, and reset commands for each supply are now
//                 logged (independent of any msgs indicating whether or not
//                 they had any effect).
//
//  2009-06-08 - Mark Davis -
//               - "Invalid" and "Mangled" syslog entries for the Oven PS are
//                 now supressed unless there are several consequtive bad
//                 reponses from the supply (indicating more than just the
//                 occasional dropped char or string of chars).
//
//  2009-06-05 - Mark Davis -
//               - Reset cmd for the Oven supply is now ignored unless the
//                 output is off or an interlock exists.
//
//  2009-06-04 - Mark Davis -
//               - Stopped logging what appears to be unexpected msgs
//               - Added OVEN_OUT_INTLK bit to statusBitsAppl
//               - ovenOutErrInterlock is not longer reset if the error goes
//                 away (i.e. it is now a latched interlock)
//               - A xxxMaxErrMS value of 0 disables the comparison between
//                 the setting and the reading for the associated supply.
//
//  2009-06-01 - Mark Davis -
//               NOTE:  Most of this is to deal with the fact that the serial
//               driver has no hdw buffer so chars can often be "lost"
//               before the ISR can read them.
//               - Serial link to Oven PS is not considered "active" unless
//                 we are getting readable responses back (it was considered
//                 valid if we were getting pretty much any response).
//               - Increased max time between valid responses before flagging
//                 the serial link as inactive from 2 to 3 seconds.
//               - What appears to be unasked-for messages from the Oven
//                 supply are ignored unless preceeded by a valid reply.
//
//  2009-05-27 - Mark Davis -
//               - First installed version
//
//  2009-04-18 - Mark Davis -
//               - Began port to DEBROS platform
//
//-----------------------------------------------------------------------------
//  NOTES:
//
//    - The RF GEN Power Supply is external to the big metal box.  As of this
//      writing, it is either a HotShot 1-2kW or a NovaStar 3-5kW supply.
//
//    - A 2/4-wire RS-485 adapter is used to connect the 2-wire RS-485
//      interface on the ZWorld with the 4-wire RS-485 interface on the
//      RF Gen Power Supply.  The direction of data transfer on the 2-wire
//      portion of the 2/4-wire adapter is manually controlled by the
//      ZWorld using 2 digital outputs (one enable Tx mode the other Rx mode).
//
//    - The OVEN supply is a Lambda ZUP60-7 (60V, 7A) supply inside the box
//
//    - The SPUTTER and BIAS supplies are EMCO 4010N (-1kV, 10ma), inside the
//      box
//
//    - In addition to the usual software/hardware WatchDog timer on the
//      ZWorld, there is a stand-alone WatchDog timer that the ZWorld must
//      send a pulse to periodically or the power to the OVEN, SPUTTER, and
//      BIAS supplies will be cut off.
//
//-----------------------------------------------------------------------------
//  Original Oven box and code vs actual/planned changes:
//
//  Implemented changes:
//
//    - This code.  In other words, a new version of the basic ECROven
//      application that runs on the DEBROS RTOS platform.  This platform
//      provides far more flexibility, functionality, and smaller code size
//      than can be achieved with the older Dynamic-C, single-threaded model.
//      The shared code for all DEBROS-based controllers provides far more
//      in diagnostic and other capabilities, including a continuosly backed
//      up copy of the run-time log, as well as eliminating many of the issues
//      with the Dynamic-C code (such as the device effectively being "deaad"
//      for several seconds whenever it updates values in the flash memory).
//
//    - Added serial connection to Lambda ZUP (Oven) supply to provide some
//      actual feedback as to the state of the supply.  Readable state includes
//      setpoints and actual (readings) for the voltage and current, the
//      mode (constant voltage or constant current), and the state of various
//      error conditions (overvoltage, overcurrent, invalid setpoints, etc).
//
//      The serial connection also allows the software to control the supply
//      in ways that formerly required getting physical access to it (the
//      most prevalent of which was the occasional need to reenable the output
//      by pressing the "OUT" button on the front of the supply).
//
//
//  Suggested/planned changes:
//
//    - Change the method for enabling the output of the Oven supply to
//      the same method used for the Bias and Sputter supplies.
//
//      Currently, a relay (K1) controlled by the ZWorld is connected in
//      series with the dedicated watchdog timer module that shuts off all
//      the supplies if it stops getting reset pulses from the ZWorld.
//
//      This means the only method the ZWorld currently has of controlling
//      whether the Oven supply is providing voltage is to enable or disable
//      the AC power to the supply.  A "gentler" way of controlling the
//      output from the supply would be the same way it is done for the Bias
//      and Sputter supplies:  Connect this relay between the COM and On/Off
//      connections on the back of the supply.
//
//      This would would have no effect on the WatchDog timer's ability to
//      shut off the supply, but it would avoid the need to power-cycle the
//      supply just to enable/disable its output.
//
//      NOTE:  The output can also be enabled/disabled via commands sent
//             over the serial connection, and the new software currently
//             sends the appropriate commands via the serial link, but the
//             relay is a much more reliable and faster method.
//
//    - Whenever the serial link to the Lambda (Oven) supply is active,
//      compare the analog readbacks to the values read from the supply via
//      the serial link.  If they vary by more than some configurable amount
//      for more than a configurable period of time, treat it as an Interlock
//      condition and disable the output.
//
//    - For all supplies, compare the voltage setpoints to their readback
//      values.  If they vary by more than some configurable amount for more
//      than a configurable period of time, treat it as an Interlock condition
//      and disable the output of the offending supply.
//
//  NOTES:
//    - The RF Gen supply is apparently NEVER used with the ECROven box, so
//      the testing of the related portion of the code in this newer version
//      will not be completed.  In addition, the instructions for creating
//      the task responsible for communicating with the RF Gen supply via a
//      serial connection will be commented out to save on CPU cycles.
//
//    - Settings for numeric values apparently must be EXACTLY the right
//      number of digits.  For example, on the Lambda 60-7 we use, the command
//      to read the current limit (:CUR!;) returns the value in 4.3 format
//      (e.g. SA2.012).  So, setting the current limit only works if you
//      send the new value in the same format (e.g. :CUR3.000; works while
//      :CUR3.0; does NOT!)
//
//    - The oven supplys set various alarm bits when a condition occurs, but
//      does not provide a way to determine when the condition has cleared.  We
//      reset the value in the supply (using the :DCL; command) so we will
//      know if a new instance occurs, but there is no way to tell if/when
//      conditions indicated by non-zero bits in the Alarm Status Register
//      (saved in ovenAlarmStatus[]) are no longer true.
//
//      Is this because they are lactched conditions in the supply and the only
//      way to resume normal operations is to turn the supply off and (after 10
//      seconds) back on?  The manual actually only says that "... the front
//      panel control is disabled until the AC input is recycled by turning the
//      AC ON/OFF switch to OFF and after aprox. 10 sec. to ON".  This doesn't
//      make it clear whether or not it can still be operated via serial
//      commands without power-cycling it.  But expect not.
//
//
//-----------------------------------------------------------------------------

#include <common.h>

#include <unistd.h>
#include <sys/types.h>


//APPLCODE - change to match the name of this file
#ifndef  ApplicSource
  #define  ApplicSource  "ECROvenAppl.c"
#endif

#ifndef DidRegDefs
  #define  DidRegDefs
  #include <modbusRegs.c>
#endif
#ifndef RegDefs



//APPLCODE - change to the application name
const char far applName[] = { "ECROven" };
const char ApplAbbrev[] = { "ECROven" };

//APPLCODE - add ID to unistd.h and place it here
#define DeviceType  DT_ECROven

//APPLCODE - set the # of input and outputs used
#define  NUM_APPL_AIN      8   // analog inputs
#define  NUM_APPL_AOUT     4   // analog outputs
#define  NUM_APPL_DIN     22   // digital inputs
#define  NUM_APPL_DOUT    16   // digital outputs

#define  ADC_MASK_APPL  0x000000FFL
#define  DAC_MASK_APPL  0x0000000FL

#define SharedINC
#include <sharedAppl.c>


#define  RELAY_PULSE_LEN  500  // ms to energize relays for momentary contacts


//--- the 8 least significant bits are for appl/cmd-specific events ---
#define  PacketsBits       0x02L

#define  ShowPackets       (DiagFlags & PacketsBits)


//=============================================================================
//  definitions related to use of the Digital and Analog I/O signals
//=============================================================================

//APPLCODE - define macros for each Digital bit and each Analog index

//=============== digitalIn bit usage ===============
#define  UNUSED_DI_00       0x000001  // 00:
#define  UNUSED_DI_01       0x000002  // 01:
#define  UNUSED_DI_02       0x000004  // 02:
#define  UNUSED_DI_03       0x000008  // 03:
#define  UNUSED_DI_04       0x000010  // 04:
#define  UNUSED_DI_05       0x000020  // 05:
#define  UNUSED_DI_06       0x000040  // 06:
#define  UNUSED_DI_07       0x000080  // 07:
#define  ID_SW_00           0x000100  // 08: ID_SW_00
#define  ID_SW_01           0x000200  // 09: ID_SW_01
#define  ID_SW_02           0x000400  // 10: ID_SW_02
#define  ID_SW_03           0x000800  // 11: ID_SW_03
#define  ID_SW_04           0x001000  // 12: ID_SW_04
#define  ID_SW_05           0x002000  // 13: ID_SW_05
#define  ID_SW_06           0x004000  // 14: ID_SW_06
#define  ID_SW_07           0x008000  // 15: ID_SW_07
#define  SPTR_INTLK_OK      0x010000L // 16: 0 = interlock, 1 = NO interlock
#define  RF_PS_HotShot      0x020000L // 17: 0 = HotShot (2kW) Serial PS
#define  RF_PS_NovaStar     0x040000L // 18: 0 = NovaStar (3kW) Serial PS
#define  RF_PS_NOT_FAULT    0x080000L // 19: 0 = RF PS Fault (K7A)
#define  RF_PS_NOT_HEAT_ON  0x100000L // 20: 0 = RF PS Heat ON (K6A)
#define  RF_PS_NOT_READY    0x200000L // 21: 0 = RF PS Ready (K5A)



const uchar  LogicOrderDIn[NUM_APPL_DIN] =  {
   0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15,
  16, 17, 18, 19, 20, 21 };

/*EXAMPLE:
#define  ClosedLoop(din)        ( (bool)((din) & CLOSED_LOOP) )
#define  FinalBtnPushed(din)    ( (bool)((din) & FINAL_BTN) )
#define  InitialBtnPushed(din)  ( (bool)((din) & INITIAL_BTN) )
#define  RFisOff(din)           ( (bool)((din) & RF_IS_OFF) )
#define  LocalMode(din)         ( (bool)((din) & LOC_REM_SW) )
*/


//--- msgs to log automatically when a digital input changes state ---
//--- use NULL for bits that we don't want to log changes for ---
const char far *digInDescr[NUM_APPL_DIN*2] = {
//--0 state descr--           --1 state descr--
   NULL,                       NULL,                   // 00:
   NULL,                       NULL,                   // 01:
   NULL,                       NULL,                   // 02:
   NULL,                       NULL,                   // 03:
   NULL,                       NULL,                   // 04:
   NULL,                       NULL,                   // 05:
   NULL,                       NULL,                   // 06:
   NULL,                       NULL,                   // 07:
   NULL,                       NULL,                   // 08: ID_SW_00
   NULL,                       NULL,                   // 09: ID_SW_01
   NULL,                       NULL,                   // 10: ID_SW_02
   NULL,                       NULL,                   // 11: ID_SW_03
   NULL,                       NULL,                   // 12: ID_SW_04
   NULL,                       NULL,                   // 13: ID_SW_05
   NULL,                       NULL,                   // 14: ID_SW_06
   NULL,                       NULL,                   // 15: ID_SW_07
   "Sptr Interlock ASSERTED", "Sptr Interlock CLEARED", // 16: SPTR Interlock
   NULL,                       NULL,                   // 17:
//"HotShot 2kW RF PS",        "NOT HotShot 2kW RF PS", // 17: type of RF supply
   NULL,                       NULL,                   // 18:
//"NovaStar 3kW RF PS",       "NOT NovaStar 3kW RF PS",// 18: type of RF supply
   NULL,                       NULL,                   // 19:
//"RF PS Fault",              "RF PS Fault Cleared",   // 19: RF PS Fault
   NULL,                       NULL,                   // 20:
//"RF PS Heat ON",            "RF PS Heat OFF",        // 20: RF PS Heat
   NULL,                       NULL                    // 21:
//"RF PS Ready",              "RF PS NOT Ready"        // 21: RF PS Ready
};


//=============== digitalOut bit usage ===============
// driven by K1, connected to +24 VDC
#define  ActivityLED     0x000001  // 00: which bit is used for the activity LED
#define  UNUSED_DO_01    0x000002  // 01:
#define  UNUSED_DO_02    0x000004  // 02:
#define  OVEN2_OUT_ALT   0x000008  // 03: mirror of 05

#define  OVEN2_PS_OFF    0x000010  // 04: 0 = Oven2 PS ON (K?)
#define  OVEN2_OUT_OFF   0x000028  // 05 & 03: 0 = Enable Oven2 PS Output
//#define  OVEN2_OUT_OFF   0x000020  // 05: 0 = Enable Oven2 PS Output
//#define  EXT_485_RX      0x000040  // 06: enable Tx on extern 2/4-wire adapter
//#define  EXT_485_TX      0x000080  // 07: enable Rx on extern 2/4-wire adapter

// driven by bl2105 K2 input, connected to +24 VDC
#define  OVEN1_OUT_OFF   0x000100  // 08: 0 = Enable Oven PS Output
#define  RF_PS_RESET     0x000200  // 09: Reset pulse for some RF PwrSupply's
#define  RF_PS_STOP      0x000400  // 10: Stop pulse for some RF PwrSupply's
#define  EXT_WD_TIMER    0x000800  // 11: Extern Watchdog timer restart

#define  OVEN1_PS_OFF    0x001000  // 12: 0 = Oven1 PS ON (K1)
#define  BIAS_PS_OFF     0x002000  // 13: 0 = Bias PS ON (K2)
#define  SPTR_PS_OFF     0x004000  // 14: 0 = Sptr PS ON (K3)
#define  RF_PS_START     0x008000  // 15: 0 = RF Gen PS start Pulse active (K4)

const uchar  LogicOrderDOut[NUM_APPL_DOUT] =  {
  12, 8, 4, 5, 13, 14, 11, 1, 2, 3, 6, 7, 9, 10, 15, 0 };

//APPLCODE - set to state we want outputs in during initialization
#define InitialDigOutState  ( 0x00FF70L )


#define  TurnOnActivityLED()    ClearBits(Next.digitalOut, ActivityLED)
#define  TurnOffActivityLED()   SetBits(Next.digitalOut, ActivityLED)

#define  EXT_485_CTL     // explicity control Tx on external 2/4 wire adapter
#define  EXT_485_RX_TX      0x0000C0  // both Tx and Rx bits
#define  NOT_EXT_485_RX_TX  0x00003F  // all BUT Tx & Rx bits (for 8-bit value)


//--- msgs to log automatically when a digital output changes state ---
//--- use NULL for bits that we don't want to log changes for ---
const char far *digOutDescr[NUM_APPL_DOUT*2] = {
//--0 state descr--              --1 state descr--
   NULL,                          NULL,                          // 00:
   NULL,                          NULL,                          // 01:
   NULL,                          NULL,                          // 02:
   NULL,                          NULL,                          // 03:
  "Enable Power to Oven2 PS",    "Disable Power to Oven2 PS",    // 04:
  "Enable Oven2 PS Output",      "Disable Oven2 PS Output",      // 05:
   NULL,                          NULL,                          // 06:
   NULL,                          NULL,                          // 07:
  "Enable Oven1 PS Output",      "Disable Oven1 PS Output",      // 08:
   NULL,                          NULL,                          // 09:
//  "Begin RF Gen PS Reset Pulse", "End RF Gen PS Reset Pulse",    // 09:
  NULL,                          NULL,                           // 10:
//  "Begin RF Gen PS Stop Pulse",  "End RF Gen PS Stop Pulse",     // 10:
  NULL,                          NULL,                           // 11:
  "Enable Power to Oven1 PS",    "Disable Power to Oven1 PS",    // 12:
  "Assert Bias PS ON",           "Assert Bias PS OFF",           // 13:
  "Assert Sptr PS ON",           "Assert Sptr PS OFF",           // 14:
  NULL,                          NULL                            // 15:
//  "Begin RF Gen PS Start Pulse", "End RF Gen PS Start Pulse"     // 15:
 };



//=============== Analog inputs ===============
//--- indexes for anaIn[] and readVal[] ---
#define  OVEN1_READ_V     0  // Oven1 PS Voltage
#define  BIAS_READ_V      1  // Bias PS Voltage
#define  SPTR_READ_V      2  // Sputter PS Voltage
#define  SPTR_READ_A      3  // Sputter PS Current
#define  BIAS_READ_A      4  // Bias PS Current
#define  OVEN1_READ_A     5  // Oven1 PS Current
#define  OVEN2_READ_V     6  // Oven2 PS Voltage
#define  OVEN2_READ_A     7  // Oven2 PS Current


const uchar  LogicOrderAIn[NUM_APPL_AIN] = { 0, 5, 1, 4, 2, 3, 6, 7 };


//=============== Analog outputs ===============
//--- indexes for anaOut[] and setVal[] ---
#define  OVEN1_SET_V      0  // Oven PS Voltage control
#define  BIAS_SET_V       1  // Bias PS Voltage control
#define  SPTR_SET_V       2  // Sputter PS Voltage control
//#define  RFGEN_SET_V      3  // RF Gen PS Voltage control
#define  OVEN2_SET_V      3  // Oven2 PS Voltage control

const uchar  LogicOrderAOut[NUM_APPL_AOUT] = { 0, 1, 2, 3 };


//=============================================================================
//  Application-specific data structures and definitions
//=============================================================================

// Each value in the ApplState structure should be throught of as representing
// a "signal" in a clocked logic circuit, with the "applic" task representing
// the logic implemented by the circuit elements such as gates, latches, etc.

// The main "applic" task is the highest priority task and executes the code
// that implements the application logic on a real-time, periodic basis.

//                 ====== IMPORTANT NOTE on LOCKING ======
// Remember to use a semaphore or some other form of locking mechanism to
// protect all access to shared values by more than one task.  The only
// exception to this is when the applic task needs to access a value protected
// by other tasks using only a scheduler/kernel lock (the KLOCK and KUNLOCK
// macros).  Because the applic task is the only task that runs at the highest
// priority, no other task can preempt it, and if the scheduler is locked,
// the applic task cannot preempt another task.  But scheduler locks must be
// kept short to avoid creating large latency in the timely execution of other
// tasks (the applic task in particular).


// ***** WARNING ***** WARNING ***** WARNING ***** WARNING *****
//  Changing the applState structure means that a soft reset will fail!  A
//  change to the size will cause a soft restart to abort and do a CPU
//  restart instead.  And while a change that does NOT result in a new
//  size will not cause the soft restart to abort, it does mean that any
//  shift in the position of values can cause all kinds of mayhem.
// ***** WARNING ***** WARNING ***** WARNING ***** WARNING *****


typedef struct {
 //-----------------------------------------------------------------------
 // SOFTWARE SETTING INPUTS
 //
 //  These values represent Software Inputs to the application logic.
 //
 //  These values are set in the "Next" instance by tasks OTHER than the
 //  "applic" task to change some value used by the application logic.
 //  Remember that all tasks besides the main "applic" task must use a
 //  scheduler lock to protect reads and writes to these values.
 //
 //  Note that the application task may alter these values (e.g. to change
 //  an invalid value in to a valid one).
 //-----------------------------------------------------------------------
  float  ovenSetV[2];     // most recent setpoint for Oven PS
  float  biasSetV;        // most recent setpoint for Bias PS
  float  sptrSetV;        // most recent setpoint for Sputter PS
//  float  rfgenSetV;       // most recent setpoint for RF Gen PS

 //-----------------------------------------------------------------------
 // READINGS
 //
 //  These values represent "passive" inputs to the application logic.
 //
 //  These values are set in the "Next" instance by the applic task just
 //  after reading the state of the hardware inputs.
 //
 //  These values generally come from one of two sources:
 //    - The just-read state of analog and/or digital hardware inputs
 //    - Values maintained by other tasks
 //
 //  While some of these values are simply copies of "external" values,
 //  including them in the ApplState structure allows us to see how
 //  they change between cycles.  Also, some of these values are the result
 //  of applying logic to "external" values.  Applying that logic once at
 //  at the beginning of the applic loop and saving the result eliminates
 //  the need to repeat the same logic in multiple blocks of code later on,
 //  avoiding unnecessary redundancy.
 //-----------------------------------------------------------------------
  ulong  digitalIn;       // state of digital inputs

//  bool  rfgenFault;
//  bool  rfgenHeatOn;
//  bool  rfgenReady;

//  bool  rfgenInactive;    // RF PS not responding via RS-485 link
  bool  ovenInactive[2];   // Oven (Lambda) PS not responding via RS-232 link

  bool  ovenRuntimeErr[2]; // Error detected in Oven PS operation
  bool  biasRuntimeErr;    // Error detected in Bias PS operation
  bool  sptrRuntimeErr;    // Error detected in Sputter PS operation

//  float  rfgenReadV;
//  float  rfgenReadA;
//  float  rfgenReadW;

//  int  rfSupplyType;

  float  ovenReadV[2], serOvenReadV[2];
  float  ovenReadA[2], serOvenReadA[2];

  uint  ovenOpStatus[2], ovenAlarmStatus[2], ovenProgStatus[2];

  float  biasReadV;
  float  biasReadA;

  float  sptrReadV;
  float  sptrReadA;


 //-----------------------------------------------------------------------
 // OUTPUTS from UN-CLOCKED (asynchronous, gate-level) LOGIC ELEMENTS
 //
 //  These values are outputs from one of the blocks of code in the applic
 //  task that represent a non-clocked logic element (such as a gate).
 //
 //  Each value here is set by one (and ONLY one) of the blocks of code in
 //  the applic task.  This group contains ONLY values that are set by a
 //  block of code that represents a NON-clocked logic element.
 //
 //  These values are set in the "Next" instance of ApplState by the applic
 //  task just after reading the state of the hardware inputs.  These values
 //  must  NOT be altered by other tasks (although they can read them as long
 //  as they use appropriate locking to do so).
 //
 //  Because these values represent the outputs of un-clocked logic elements,
 //  and such elements often need to be cascaded (the outputs used as inputs
 //  to other non-clocked elements), the result is often dependent on the
 //  order of execution of the blocks of code.
 //-----------------------------------------------------------------------
  bool   ovenModeOn[2];           // Turn on power, enable output soon after
  bool   ovenOutErrInterlock[2];  // Oven PS Output state was wrong
  ulong  ovenOutErrStartTime[2];  // start of period of Output state error
  bool   sptrInterlock;           // Sputter PS power disabled by ext interlock

 //-----------------------------------------------------------------------
 // OUTPUTS from CLOCKED (latched inputs and outputs) LOGIC ELEMENTS
 //
 //  These values are outputs from one of the blocks of code in the applic
 //  task that represent a clocked logic element (such as latches).
 //
 //  Each value here is set by one (and ONLY one) of the blocks of code in
 //  the applic task.  This group contains only values that are set by a
 //  block of code that represents a CLOCKED logic element.
 //
 //  These values are set in the "Next" instance of ApplState by the applic
 //  task AFTER all the other ApplState values are set.  Because they represent
 //  only the clocked logic elements, and each block of code reads only values
 //  from the Cur instance, the blocks can be executed in any order and the
 //  result will always be the same.
 //
 //  These values are used as inputs by other blocks and/or to set values
 //  used by other tasks (such as values read by the ModBus server task).
 //
 //  Again, these values must NOT be altered by other tasks (although they
 //  can read them as long as they use appropriate locking to do so).
 //-----------------------------------------------------------------------
  ulong  digitalOut;      // state for digital outputs

  bool   enableOvenPower[2];   // enable power to the Oven PS except for reset
  bool   enableOvenOutput[2];  // enable the Oven PS output
  float  ovenSetVassert[2];    // actual in-use setpoint for Oven PS
  float  serOvenSetV[2];       // setpoint according to Oven PS
  float  serOvenSetA[2];       // setpoint according to Oven PS

  bool   assertOnBiasPS;  // enable the Bias PS
  float  biasSetVassert;  // actual in-use setpoint for Bias PS

  bool   assertOnSptrPS;  // enable the Sputter PS
  float  sptrSetVassert;  // actual in-use setpoint for Sputter PS

//  ulong  rfgenStartTime;  // start time (in ms) of RF Gen Start Pulse
//  ulong  rfgenStopTime;   // start time (in ms) of RF Gen Stop pulse
//  ulong  rfgenResetTime;  // start time (in ms) of RF Gen Reset pulse
//  float  rfgenSetVassert; // actual in-use setpoint for RF Gen PS

} ApplState;


//--- Event flags that are treated as "pulses" (i.e. some task sets the ---
//--- value of one or more of them (as an atomic operation) to indicate ---
//--- some event that the applic logic needs to deal with ---
bool    // rfgenStartCmd, rfgenStopCmd, rfgenResetCmd,
        biasOnCmd, biasOffCmd, sptrOnCmd, sptrOffCmd,
        ovenOnCmd[2], ovenOffCmd[2], ovenResetCmd[2], newOvenResp[2];

ulong  //rfgenLastRespTime,       // last time we got a resp from the RF Gen PS
       //rfgenReadErrStartTime,   // time existing read/set error first detected
       ovenLastRespTime[2],     // last time we got a valid resp from Oven PS
       ovenReadErrStartTime[2], // time existing read/set error first detected
       ovenOffCmdTime[2],       // time last active off cmd rcvd
       ovenPowerOnTime[2],      // time Oven PS power was enabled
       sptrReadErrStartTime,    // time existing read/set error first detected
       biasReadErrStartTime,    // time existing read/set error first detected
       wdPulseChanged;          // last time extern WatchDog timer pulse chgd state

ulong  ovenRespAge[2];      // # ms since last valid response from Oven PS

uint  postOvenResetMsgs[2]; // # of valid resps from Oven PS since a reset


ApplState  prevState, curState, nextState;

#define  Prev  prevState
#define  Cur   curState
#define  Next  nextState

const uint ApplStateDataSize =  ( sizeof(ApplState)*3 + sizeof(setVal) );


//--- bit values for ovenOpStatus ---
#define  OOS_ConstVolt      0x80
#define  OOS_FoldbackProt   0x40
#define  OOS_AutoRestart    0x20
#define  OOS_OutputOn       0x10
#define  OOS_FoldbackSRQ    0x08
#define  OOS_OverVoltSRQ    0x04
#define  OOS_OverTempSRQ    0x02
#define  OOS_AlarmBitSet    0x01


//=============================================================================
//  List of daemon processes.  The inetd process listens for connection
//  requests on each port in the list and creates a new instance of the
//  specified process to service each new connection.
//=============================================================================
#include <services.h>

//APPLCODE - add tasks that service network clients

//--- service funcs declared the same as standard command funcs:
//     e.g.:   void _task fastdatad(int argc, char **argv);

void _task tsh(int argc, char **argv);


ServiceInfo inetdConf[] = {
//--- add any application-specific inetd-type network services here ---
//--- (add port # names to services.h) ---
//  NAME          PORT           PROTOCOL  FUNC       PRI  -1  Config Params
//  ------------  -------------  --------  ---------- ---  --  ----------------

//--- standard ones defined in sharedAppl.c ---
//{ "console",    TTY0,          ASYNC,    tsh,        10, -1, &consCfg },
//{ "fastdatad",  FASTDATA_PORT, ProtoTCP, fastdatad,   5, -1, NULL },
//{ "modbusTCPd", MODBUS_PORT,   ProtoTCP, modbusTCPd,  6, -1, NULL },
//{ "tsh",        TELNET_PORT,   ProtoTCP, tsh,        10, -1, NULL }

  #define SharedServices
  #include <sharedAppl.c>
};

const int  NumServices = sizeof(inetdConf) / sizeof(ServiceInfo);



//=============================================================================
//  Defintions for variables that are accessable as ModBus registers
//=============================================================================
#else  // end of before-RegDefs portion

/*
  The purpose of this section is to define a set of C variables that are
  accessable as ModBus registers.  This approach avoids the need to use
  special functions or macros throughout the code when accessing register
  values, which keeps the application code fast and simple.

  The model used provides support for multiple logical devices.  The 2nd
  parameter in each register definition macro indicates if one instanace
  applies to all logical devices (specified by an "x" for "don't care"), or
  if each logical device requires its own instance (specified by an under-
  score "_" to signify "fill in the blank").  In the 1st case, the variable
  is defined as a scalar.  In the 2nd case, it is defined as an array.

  The definition macros support data types (floats and longs in particular)
  that require more than one register to store them.  To avoid the possibility
  of corrupted values, all reads of and writes to such values must be done
  using the multi-register commands and including all the registers that
  make up the value in the command.  Differences in register order between
  different client types is handled by a table that automatically swaps
  register values as needed based on the client type.

  The definition for each writable register also includes the function to
  call when the value is set via the ModBus interface.  Many registers can
  use the default setXXX() functions, but you can specify your own to provide
  any special logic needed (including ignoring the new setting if invalid).

  For registers, the format for the function definitions is:

    bool mySetFunc(SetRegFuncParams);

       where SetRegFuncParams is currently defined as:
              void *info, uchar chanNum, void *newVal, fd_t eventsfd

  For coils, the format for the function definitions is:

     bool (*SetCoilFunc)(SetCoilFuncParams);

         where SetCoilFuncParams is currently define as:
               void *info, uchar chanNum, uchar newVal, fd_t eventsfd

  The parameters to the call are as follows:
    *info:     A pointer to the Coil's/Register's CoilInfo/RegInfo structure
    chanNum:   The logical device # (0 is the 1st)
    *newVal:   A pointer to the new value requested in the Modbus packet
               (changed to native-byte order where needed)
    eventsfd:  The fd for the FIFO used to send msgs to the main application
               process

  The function returns true if the new value was accepted, false if not.

  The register numbers for multiple "channels" or logical devices are
  handled automatically based on the xxChanMult values.  These values specify
  the offset between a register value for one logical device and the "same"
  register for the Next.

  Example:

    Assume the following definitions:
      #define  NumLogicalDevices  2
      #define  R4ChanMult  100
      REG4( motorPos,       _, 40050, F32, RW, setMotorPos )

    Then we would get:
      global variable definition:  float motorPos[2];
      register pair:  40050,40051  refers to:  motorPos[0]
      register pair:  40150.40151  refers to:  motorPos[1]

    If instead we had specified:
      REG4( motorPos,       x, 40050, F32, RW, setMotorPos )

    Then we would get:
      global variable definition:  float motorPos;
      register pairs:  40050,40051 & 40150,40151  BOTH refer to:  motorPos


  To keep the definitions as simple as possible, and avoid requiring multiple
  macros for each register definition, each macro is used more than once,
  with a different definition (#define) each time (a practice generally
  referred to as X-Macros).  While trying to follow the order of inclusion
  of files (including a recursive reference back to this file) can be rather
  confusing, the X-Macro method keeps the source code simple to read and
  maintain by doing everything with a single macro for each register.

  ==========================================================================
                             !!! WARNING !!!
  ==========================================================================
    Each sequence of register definitions MUST be contiguous!  There must
    NOT be any holes (i.e. no unused register #'s between used ones) and no
    overlap!
-----------------------------------------------------------------------------*/


//APPLCODE - set # of devices for which we want duplicate register sets
#ifndef NumLogicalDevices
  #define  NumLogicalDevices  1
#endif

//--------------------------------------------------------------
//APPLCODE - set max # of registers per set of coils
#ifndef NumC0Chans
  #define NumC0Chans_  NumLogicalDevices
  #define C0ChanMult_  100
#endif

#ifdef COIL0

COIL0( DoReboot,       x, 1, setCoil )
//COIL0( Coil00002,      _, 2, setCoil )
//COIL0( Coil00003,      _, 3, setCoil )

#undef COIL0
#endif


//--------------------------------------------------------------
//APPLCODE - set max # of registers per set of read-only registers
#ifndef NumR3Chans
  #define NumR3Chans_  NumLogicalDevices
  #define R3ChanMult_  100
#endif

#ifdef REG3

REG3( dlmVersTime,    x, 30001, U32 )
REG3( rtosVersTime,   x, 30003, U32 )
REG3( applVersTime,   x, 30005, U32 )
REG3( devType_,       x, 30007, U16 )
REG3( deviceID_,      x, 30008, U32 )
REG3( upSecs,         x, 30010, U32 )
REG3( percentIdle_,   x, 30012, U16 )

REG3( numChans,       x, 30013, U16 )
REG3( coilChanMult,   x, 30014, U16 )
REG3( reg3ChanMult,   x, 30015, U16 )
REG3( reg4ChanMult,   x, 30016, U16 )

REG3( digInBits,      x, 30017, U32 )

REG3( statusBitsStdA, x, 30019, U16 )
REG3( statusBitsStdB, x, 30020, U16 )
REG3( statusBitsAppl, x, 30021, U32 )

//APPLCODE - add registers for externally available read-only values
REG3( rfSupplyType,   x, 30023, U16 )  // type of RF Power Supply in use

REG3( sptrReadV,      x, 30024, F32 )  // Sputter Disk PwrSupply Voltage Read
REG3( sptrReadA,      x, 30026, F32 )  // Sputter Disk PwrSupply Current Read

REG3( biasReadV,      x, 30028, F32 )  // Bias Disk PwrSupply Voltage Read
REG3( biasReadA,      x, 30030, F32 )  // Bias Disk PwrSupply Current Read

REG3( oven1ReadV,     x, 30032, F32 )  // Oven1 Disk PwrSupply Voltage Read
REG3( oven1ReadA,     x, 30034, F32 )  // Oven1 Disk PwrSupply Current Read

REG3( rfgenReadV,     x, 30036, F32 )  // RF Gen Disk PwrSupply Voltage Read
REG3( rfgenReadA,     x, 30038, F32 )  // RF Gen Disk PwrSupply Current Read
REG3( rfgenReadW,     x, 30040, F32 )  // RF Gen Disk PwrSupply Power Read

REG3( serOven1ReadV,  x, 30042, F32 )  // Oven1 PS Voltage Read via serial
REG3( serOven1ReadA,  x, 30044, F32 )  // Oven1 PS Current Read via serial

REG3( oven1Status,    x, 30046, U32 )  // bit-mapped Oven1 PS status

REG3( latency,        x, 30048, U16 )  // actual vs scheduled appl start in ms
REG3( tcpTickTime,    x, 30049, U16 )  // # ms inside tcp_tick() call
REG3( applLoopTime,   x, 30050, U16 )  // # ms used by applic task loop
REG3( nonKernelLocks, x, 30051, U16 )  // # times non-kernel tasks needed IP lock
REG3( failedLocks,    x, 30052, U16 )  // # times non-kernel tasks needed IP lock
REG3( duplicateLocks, x, 30053, U16 )  // # times non-kernel tasks needed IP lock
REG3( oven1NumMsgs,   x, 30054, U16 )  // # consecutive valid resps from Oven1 PS
REG3( oven1GoodMsgs,  x, 30055, U16 )  // # consecutive valid resps from Oven1 PS
REG3( oven1BadMsgs,   x, 30056, U16 )  // # consecutive valid resps from Oven1 PS

REG3( adc00,          x, 30057, U16 )
REG3( adc01,          x, 30058, U16 )
REG3( adc02,          x, 30059, U16 )
REG3( adc03,          x, 30060, U16 )
REG3( adc04,          x, 30061, U16 )
REG3( adc05,          x, 30062, U16 )
REG3( adc06,          x, 30063, U16 )
REG3( adc07,          x, 30064, U16 )

REG3( oven2ReadV,     x, 30065, F32 )  // Oven2 Disk PwrSupply Voltage Read
REG3( oven2ReadA,     x, 30067, F32 )  // Oven2 Disk PwrSupply Current Read

REG3( serOven2ReadV,  x, 30069, F32 )  // Oven1 PS Voltage Read via serial
REG3( serOven2ReadA,  x, 30071, F32 )  // Oven1 PS Current Read via serial

REG3( oven2Status,    x, 30073, U32 )  // bit-mapped Oven1 PS status

REG3( oven2NumMsgs,   x, 30075, U16 )  // # consecutive valid resps from Oven2 PS
REG3( oven2GoodMsgs,  x, 30076, U16 )  // # consecutive valid resps from Oven2 PS
REG3( oven2BadMsgs,   x, 30077, U16 )  // # consecutive valid resps from Oven2 PS

REG3( kernelLoopTime, x, 30078, U16 )  // # ms used by last kernel pass

#undef REG3
#endif


#ifdef RegMacros
//===== application-specific defintions related to register values =====

//--- bit usage for statusBitsAppl ---
//#define  RF_PS_FAULT        0x00000001
//#define  RF_PS_HEAT_ON      0x00000002
//#define  RF_PS_READY        0x00000004
#define  OVEN1_PS_ENABLED   0x00000008
#define  BIAS_PS_ENABLED    0x00000010
#define  SPTR_PS_ENABLED    0x00000020
//#define  RF_PS_INACTIVE     0x00000040   // RF Gen PS not responding via RS-485
#define  OVEN1_PS_INACTIVE  0x00000080   // Oven PS not responding via RS-232
#define  OVEN1_OUT_INTLK    0x00000100   // serial reports output state diff
#define  OVEN1_RESETTING    0x00000200   // power cycling oven1 supply
#define  OVEN2_PS_ENABLED   0x00000400
#define  OVEN2_PS_INACTIVE  0x00000800   // Oven PS not responding via RS-232
#define  OVEN2_OUT_INTLK    0x00001000   // serial reports output state diff
#define  OVEN2_RESETTING    0x00002000   // power cycling oven2 supply
#define  SPTR_INTERLOCK     0x00004000   // sputter supply interlock

//--- bit usage for rfSupplyType ---
#define  RF_TYPE_UNKNOWN   0  //
#define  RF_TYPE_NovaStar  1  // Ameritherm NovaStar (3kW) power supply
#define  RF_TYPE_HotShot   2  // Ameritherm HotShot (2kW) power supply

const char *SupplyTypeDesc[3] = { "Unknown ", "NovaStar", "HotShot " };


//--- bit usage for ovenStatus ---
  // Operational Status Register bits
#define  OS_ALARM_BIT     0x00000001   // Alarm Status register bit is set
#define  OS_OTP_SRQ       0x00000002   // Over Temp Protection SRQ enabled
#define  OS_OVP_SRQ       0x00000004   // Over Voltage Protection SRQ enabled
#define  OS_FP_SRQ        0x00000008   // Foldback protection SRQ enabled
#define  OS_OUT_ON        0x00000010   // Output is ON
#define  OS_AR_ON         0x00000020   // Auto-Restart ON
#define  OS_FPA           0x00000040   // Foldback Protection armed
#define  OS_CV_MODE       0x00000080   // Constant Voltage mode
  // Alarm Status Register bits
#define  OS_ERR_BIT       0x00000100   // Error Code register bit is set
#define  OS_FP_ACTIV      0x00000200   // Foldback Protection activated
#define  OS_IVS_FAIL      0x00000400   // Failure occured at Input Volt Supply
#define  OS_OTP_TRIP      0x00000800   // Over Temp Protection was tripped
#define  OS_OVT_TRIP      0x00001000   // Over Voltage Protection was tripped
  // Error Codes Register bits
#define  OS_INV_CUR       0x00010000   // Invalid Current setting
#define  OS_INV_VOLT      0x00020000   // Invalid Voltage setting
#define  OS_BUF_OVER      0x00040000   // Com buffer overflow
#define  OS_INV_CMD       0x00080000   // Invalid command rcvd


//  #define  READING(idx)  REG_ARRAY(float, rdg_1, idx)

#endif


//--------------------------------------------------------------
//APPLCODE - set max # of registers per set of read/write registers
#ifndef NumR4Chans
  #define NumR4Chans_  NumLogicalDevices
  #define R4ChanMult_  100
#endif

/*
  The "standard" functions for processing writes to Modbus register values:

    setS16, setU16, setS32, setU32, setF32

  These functions simply log a message for the change and update the
  value in a critical block (so the change cannot be interrupted).

  In many cases, just allowing any setting to be written to a register (or
  multiple registers comprising a larger value) is not acceptable.


  If any of the following apply, then a custom func must be used instead:

    - The order of changes is important.  For example, if you get ON,
      OFF, and RESET commands in rapid succession, you have to be sure
      they are processed in the order they were recieved.

    - Not all possible settings are acceptable.  The acceptable settings
      for a value are often a subset of the possible ones, and frequently
      are dependent upon the current state of the controller.


  IMPORTANT NOTES:

    The "applic" task periodically executes the code that makes up the
    bulk of the application logic.  Because it runs at the highest priority
    there is no chance that a lower priority task (like the daemon handling
    Modbus register writes) can change a value in the middle of this
    processing.

    However, the applic task CAN (and often does) preempt other processes,
    so any process that does read or make changes to the register values
    must lock the register values before doing so.  This is currently done
    by disabling the scheduler using the KLOCK() and KUNLOCK() macros.

    Use of such a lock insures that changes occur only between passes
    through the application logic, but there are still several issues
    that must be addressed when writing functions to handle writes to
    register values:

      - Priority inversion and latency:  Allowing a lower priority task
        to lock out higher ones (or, in this case, ALL tasks) allows
        the lower priority task to run when a higher-priority one should
        be running.  Care must be taken to keep such periods as short
        as possible to avoid disrupting the timing of the control logic.

      - Loss of encapsulation:  The last thing we want is to break the
        application logic in to small pieces spread through a bunch of
        functions executed by different tasks at different times.

        While the logic for processing writes to register values is often
        dependent upon the state of the control logic, functions outside
        the applic task must NOT directly alter the state of the application
        logic.  To achieve this, such functions must limit changes to
        values that act as input to the applic task.

        To this end, you need to treat the register values that act as
        inputs just like they were knobs, switches, analog inputs, or
        latches.  Like their hardware equivalent, the applic task logic
        reads their state at the beginning of each pass and uses this
        information when determining what the next state of the control
        logic should be.  And just like a hardware latch, values can be
        reset after reading them so the next write to the value can be
        treated as another instance of a request.

    The trickiest circumstances to handle are ones where the order of separate
    writes is important.  The most common example are commands for ON, OFF,
    and RESET.  If, for example, you had 3 separate flags, and simply set
    each once corresponding to which of the 3 commands you received, you
    end up with multiple flags set when the applic logic is processed, and
    no way to determine what the end result should be.  In these cases,
    the logic that controls the state of each flag should be placed in its
    own function and called during the processing of any event that could
    affect the flag's state.

Open issues:
    Ideally, the logic for setting each variable would be triggered each
    time an event occurs that could affect it:  A publish-subscribe type
    of mechanism.  One possibility to do this efficiently?  Use a bit-mask
    "array" to determine which functions to call for each register that is
    changed.  That way it doesn't take many bytes to determine which functions
    to call, and each function can update its corresponding register value
    based on its current state and the event that just occured.

    Or should I simply combine such values in to one (a struct, if necessary),
    and have a single function that is called for all the related values
    and updates the state of the single logical entity based on its
    current state and what event called it?  (But the # of combinations could
    get out of hand easily...)
*/

#ifdef REG4

REG4( timeZone_,      x, 40001, S16, RW, MB_timeZone )
REG4( curTime,        x, 40002, U32, RW, setU32 )

REG4( ipAddr,         x, 40004, U32, RW, setU32 )
REG4( netmask,        x, 40006, U32, RW, setU32 )
REG4( gateway,        x, 40008, U32, RW, setU32 )

REG4( setACoil,       _, 40010, U16, RW, setU16 )
REG4( resetACoil,     _, 40011, U16, RW, setU16 )

REG4( digOutBits,     x, 40012, U32, RW, setU32 )

REG4( cmdBitsStd,     x, 40014, U32, RW, MB_cmdBitsStd )
REG4( cmdBitsAppl,    x, 40016, U32, RW, MB_cmdBitsAppl )

//APPLCODE - add registers for externally available read/write values
REG4( sptrSetV,       x, 40018, F32, RW, setF32 )  // Sputter Disk PS Voltage Set
REG4( biasSetV,       x, 40020, F32, RW, setF32 )  // Bias Disk PS Voltage Set
REG4( oven1SetV,      x, 40022, F32, RW, setF32 )  // Oven1 Disk PS Voltage Set
REG4( rfgenSetV,      x, 40024, F32, RW, setF32 )  // RF Gen Disk PS Voltage Set

REG4( serOven1SetV,   x, 40026, F32, RW, setF32 )  // Oven1 PS V set via serial
REG4( serOven1SetA,   x, 40028, F32, RW, setF32 )  // Oven1 PS A set via serial

REG4( sptrMaxErrV,    x, 40030, F32, RW, setF32 )  // Max Set/Read diff
REG4( sptrMaxErrMS,   x, 40032, U32, RW, setU32 )  // Max error time in ms

REG4( biasMaxErrV,    x, 40034, F32, RW, setF32 )  // Max Set/Read diff
REG4( biasMaxErrMS,   x, 40036, U32, RW, setU32 )  // Max error time in ms

REG4( oven1MaxErrV,   x, 40038, F32, RW, setF32 )  // Max Set/Read diff
REG4( oven1MaxErrMS,  x, 40040, U32, RW, setU32 )  // Max error time in ms

REG4( forceOvenErr,   x, 40042, U16, RW, setU16 )  // to test Oven Output Interlocks

REG4( oven1PwrOffDelay, x, 40043, U16, RW, setU16 )  // power off delay in ms

REG4( oven2SetV,        x, 40044, F32, RW, setF32 )  // Oven2 Disk PS Voltage Set
REG4( serOven2SetV,     x, 40046, F32, RW, setF32 )  // Oven2 PS V set via serial
REG4( serOven2SetA,     x, 40048, F32, RW, setF32 )  // Oven2 PS A set via serial
REG4( oven2MaxErrV,     x, 40050, F32, RW, setF32 )  // Max Set/Read diff
REG4( oven2MaxErrMS,    x, 40052, U32, RW, setU32 )  // Max error time in ms
REG4( oven2PwrOffDelay, x, 40054, U16, RW, setU16 )  // power off delay in ms

#undef REG4
#endif


#ifdef RegMacros
//===== application-specific defintions related to register values =====

//  #define  ISET(idx)   REG_ARRAY(float, iSet_1, idx)

#endif

//--------------------------------------------------------------
#undef RegMacros

#endif
#ifndef RegDefs // start of after-RegDefs portion

//--- definitions for cmdBitsAppl ---
#define  CBA_DoReboot     0x00000001
#define  CBA_DoInit       0x00000002

#define  CBA_Oven1On      0x00000010
#define  CBA_Oven1Off     0x00000020
#define  CBA_BiasOn       0x00000040
#define  CBA_BiasOff      0x00000080
#define  CBA_SptrOn       0x00000100
#define  CBA_SptrOff      0x00000200
//#define  CBA_RFGenStart   0x00000400
//#define  CBA_RFGenStop    0x00000800
//#define  CBA_RFGenReset   0x00001000
#define  CBA_Oven1Reset   0x00002000
#define  CBA_Oven2On      0x00004000
#define  CBA_Oven2Off     0x00008000
#define  CBA_Oven2Reset   0x00010000L

//#define  CBA_RFGenCmds   (CBA_RFGenStart | CBA_RFGenStop | CBA_RFGenReset)


//=============================================================================
//  defintions for "parameterized" data values - lists of values that we want
//  to be "persistent" (i.e. saved to a file and restored on startup)
//=============================================================================
//APPLCODE - settings/values we want saved to/restored from files


//--- parameter ID's for data in Device Data/calibration file ---
//--------------------------------------------------------------------------
// WARNING:  ID's can be added, but existing ID's (and the format of the
//           data they represent) should NEVER change!  NOTE:  Existing ones
//           can be "eliminated" by changing their name and no longer
//           using them, but never REUSE an old one for something new!!!
//--------------------------------------------------------------------------
//  NOTE:  ID's >= 1000 are rsvd for values shared by most applications
//--------------------------------------------------------------------------

const ParamData far  devDataParams[] = {
//APPLDATA
//                ID   ---value---  --restoreFunc--
//PERSISTENT_VAR(  1,  myVarName,     NULL ),
//                ID   ---value---  granularity  --restoreFunc--
//PERSISTENT_TBL(  2   myArrayName, elemSize,     NULL ),

  PERSISTENT_VAR(  1,  oven1MaxErrV,       NULL ),
  PERSISTENT_VAR(  2,  oven1MaxErrMS,      NULL ),
  PERSISTENT_VAR(  3,  sptrMaxErrV,        NULL ),
  PERSISTENT_VAR(  4,  sptrMaxErrMS,       NULL ),
  PERSISTENT_VAR(  5,  biasMaxErrV,        NULL ),
  PERSISTENT_VAR(  6,  biasMaxErrMS,       NULL ),
  PERSISTENT_VAR(  7,  oven1PwrOffDelay,   NULL ),
  PERSISTENT_VAR(  8,  oven2MaxErrV,       NULL ),
  PERSISTENT_VAR(  9,  oven2MaxErrMS,      NULL ),
  PERSISTENT_VAR( 10,  oven2PwrOffDelay,   NULL ),

  #define SharedDevParams
  #include <sharedAppl.c>

  END_PARAM_DATA
};


//----------------------------------------------------
const ParamData far  settingsParams[] = {
//APPLDATA
//_nearcall void ampSetRestore(void far *curVal, char *newVal, uint count);
//
//                ID   ---value---  --restoreFunc--
  PERSISTENT_VAR(  1,  sptrSetV,     NULL ),
  PERSISTENT_VAR(  2,  biasSetV,     NULL ),
  PERSISTENT_VAR(  3,  oven1SetV,    NULL ),
//  PERSISTENT_VAR(  4,  rfgenSetV,    NULL ),
  PERSISTENT_VAR(  5,  oven2SetV,    NULL ),

  END_PARAM_DATA
};


//-----------------------------------------------------------------
//  Application-specific values needed to restore the state of the
//  controller following a "soft" reset (i.e. a restart without a
//  CPU reset)
//-----------------------------------------------------------------
const ParamData far  ctlrStateParams[] = {
// sample values from sharedAppl.c, common to all applications
//                   --ID--  ---value,granularity---  --restoreFunc--
//    PERSISTENT_VARx(  1000,  prevState,               NULL ),
//    PERSISTENT_VARx(  1001,  curState,                NULL ),
//    PERSISTENT_VARx(  1002,  nextState,               NULL ),
//    PERSISTENT_TBL (  1003,  setVal,4,                NULL ),
// // PERSISTENT_TBL (  1004,  shadowRegValues,4,       NULL ),

//APPLDATA
//    PERSISTENT_VAR(  1,  myApplStateVar,        NULL ),


  #define SharedCtlrStateParams
  #include <sharedAppl.c>

  END_PARAM_DATA
};


//-----------------------------------------------------------------------------
//  A list of parameterized "persistent" data.  Backup copies of such values
//  are stored in files and restored on startup.
//-----------------------------------------------------------------------------

//    Base File Name    Category  ParamData
//    ----------------  --------  ---------
//--- add any application-specific parameterized file defs here ---
//ParamFile  myAppCfgFile =
//  { "myFileName",     PDC_CTLR, myCfgParams,   0 };

ParamFile  settingsFile =
    { "settings",       PDC_SET,  settingsParams  };


#define SharedParams
#include <sharedAppl.c>


ParamFile *paramFiles[] = {
  #define SharedParamsList
  #include <sharedAppl.c>
//--- add any application-specific ParamData references here ---
//  &myAppCfgFile,
  &settingsFile
};

const int NumParamFiles = sizeof(paramFiles) / sizeof(paramFiles[0]);


//=============================================================================
//  Declarations for functions that handle setting "Next" values that serve
//  as inputs to the application control logic (run by the "applic" task in
//  a periodically executed loop).
//
//  While care must be taken in these functions to use locks to avoid the
//  possibility of being preempted in the middle of an operation, such locks
//  must be kept to as short a time as possible to avoid excessively latency
//  in the main control loop.
//=============================================================================

//void onCmd(void);
//void offCmd(void);
//void resetCmd(void);

//void newSetpoint(float newAmps);


//=============================================================================
//  This is where misc global variables should be defined
//=============================================================================
#define SharedGlobals
#include <sharedAppl.c>

//APPLCODE
//--- add definitions for application-specific global variables here ---
char  // rfgenCmd[32], rfgenResp[128],
      ovenCmd[2][32], ovenResp[2][64];


//=============================================================================
//  Functions referenced in the Modbus register definitions to handle writes to
//  specific global values defined as ModBus registers
//=============================================================================

//APPLCODE - add functions to process writes to ModBus register values

//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for "cmdBitsAppl"
//
//  Check the bits that where set and trigger the corresponding events as
//  appropriate (but do NOT save the new value - just use it to trigger
//  events)
//-----------------------------------------------------------------------------
bool _nearcall MB_cmdBitsAppl(SetRegFuncParams)
{
  ulong  cmdBits;


  cmdBits = *((ulong *)newVal);

  KLOCK();

  if (cmdBits & CBA_Oven1On)   { ovenOnCmd[0] = true;   ovenOffCmd[0] = false; }
  if (cmdBits & CBA_Oven1Off)  { ovenOnCmd[0] = false;  ovenOffCmd[0] = true;  }
  if (cmdBits & CBA_Oven1Reset) ovenResetCmd[0] = true;

  if (cmdBits & CBA_Oven2On)   { ovenOnCmd[1] = true;   ovenOffCmd[1] = false; }
  if (cmdBits & CBA_Oven2Off)  { ovenOnCmd[1] = false;  ovenOffCmd[1] = true;  }
  if (cmdBits & CBA_Oven2Reset) ovenResetCmd[1] = true;

  if (cmdBits & CBA_BiasOn)   { biasOnCmd = true;   biasOffCmd = false; }
  if (cmdBits & CBA_BiasOff)  { biasOnCmd = false;  biasOffCmd = true;  }

  if (cmdBits & CBA_SptrOn)   { sptrOnCmd = true;   sptrOffCmd = false; }
  if (cmdBits & CBA_SptrOff)  { sptrOnCmd = false;  sptrOffCmd = true;  }

//  if (cmdBits & CBA_RFGenStart) { rfgenStartCmd = true;  rfgenStopCmd = false; }
//  if (cmdBits & CBA_RFGenStop)  { rfgenStartCmd = false; rfgenStopCmd = true;  }
//  if (cmdBits & CBA_RFGenReset) rfgenResetCmd = true;

  cmdBitsAppl = 0;

  KUNLOCK();


  return true;
}


//-----------------------------------------------------------------------------
//  External ModBus client sent a new setpoint
//
//  Check the new setting and alter output values as appropriate
//-----------------------------------------------------------------------------
/*EXAMPLE
bool MB_ampSet(SetRegFuncParams)
{
  newSetpoint(MB_VAL(float));

  return true;
}
*/


//=============================================================================
//  functions to handle restoration of saved "persistent" values
//=============================================================================

/*
//-----------------------------------------------------------------------------
//  one of the setpoint values needs restoring from the settings file
//-----------------------------------------------------------------------------
_nearcall void myValRestore(void far *curVal, char *newVal, uint count)
{
  uint  u;
  float  fval;
  char  temp[32];

  for (u=0; u<NumPS; ++u)  if (curVal == (void far *) &ISET(u))  {
    fval = *(float far *)newVal;
    printf("restoring iSet_%u to: %s\r\n", u, fmtFloat(fval, 8,2, temp));
    newSetpoint(u, fval);
  }

}

//-----------------------------------------------------------------------------
//  restore an ampSet value from the settings file
//-----------------------------------------------------------------------------
_nearcall void ampSetRestore(void far *curVal, char *newVal, uint count)
{
  newSetpoint(NEW_VAL(float));
}


*/

//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
#define SharedFuncs
#include <sharedAppl.c>


//=============================================================================
//  Misc Application-specific functions
//=============================================================================
//APPLCODE



//=============================================================================
//  Task functions to handle application-specific commands
//=============================================================================
//APPLCODE

//-----------------------------------------------------------------------------
//  example func to process a command entered via a user shell
//-----------------------------------------------------------------------------
/*
void _task miscTest(int argc, char **argv)
{
  int  idx;
  bool  validCmd;

//used by EVENT macro
  fd_t  eventsfd = -1;
  EventMsg  eventMsg;


  validCmd = false;

  for (;;)  {
    if (argc < 3)  break;
    if (! isdigit(argv[1][0]))  break;

    idx = atoi(argv[1]);
    if ((idx < 1) OR (idx > NumChan))  {
      printf("Invalid chan #\n");  Exit(-2); }

    --idx;  validCmd = true;

    if (StriEq(argv[2], "x"))     { EVENT(MyEventX, idx, NULL, 0);  break; }
    if (StriEq(argv[2], "y"))     { EVENT(MyEventY, idx, NULL, 0);  break; }

    validCmd = false;  break;
  }

  if (! validCmd)  { printf("Usage:  misc # x|y\r\n");  Exit(-1); }

  Exit(0);
}
*/

//=============================================================================
//  Generic structure to handle application-specific presentation of the
//  digital and analog input/output values.  Code to handle additional
//  formatting that cannot be handled in this generic manner should be
//  placed in the showState() task routine.
//=============================================================================
//APPLCODE - text and formatting info for application state summary

#define  RedText  "\033[31m"
#define  BlueText "\033[34m"

//-----------------------------------------------------------------------------
_nearcall char *fmtSupType(DispElem far *elem, char *buf)
{
  return (char *) SupplyTypeDesc[ELEM_VAL(U16)];
}

//-----------------------------------------------------------------------------
_nearcall char *fmtInactive(DispElem far *elem, char *buf)
{
  return (ELEM_VAL(bool) ? RedText" NO" : BlueText"YES");
}

//-----------------------------------------------------------------------------
_nearcall char *fmtF32ser1(DispElem far *elem, char *buf)
{
  strcpy(buf, (Cur.ovenInactive[0] ? RedText : BlueText));
  fmtFloat(ELEM_VAL(float), elem->width, elem->prec, buf+5);
  return buf;
}
//---------------------------------------------------------------
_nearcall char *fmtF32ser2(DispElem far *elem, char *buf)
{
  strcpy(buf, (Cur.ovenInactive[1] ? RedText : BlueText));
  fmtFloat(ELEM_VAL(float), elem->width, elem->prec, buf+5);
  return buf;
}

//-----------------------------------------------------------------------------
_nearcall char *fmtBinary1(DispElem far *elem, char *buf)
{
  uint  val;

  strcpy(buf, (Cur.ovenInactive[0] ? RedText : BlueText));
  val = ELEM_VAL(uint);  val >>= elem->width;
  LongToBinMS1st((ulong)val, buf+5, elem->prec);
  return buf;
}
//--------------------------------------------------------------
_nearcall char *fmtBinary2(DispElem far *elem, char *buf)
{
  uint  val;

  strcpy(buf, (Cur.ovenInactive[1] ? RedText : BlueText));
  val = ELEM_VAL(uint);  val >>= elem->width;
  LongToBinMS1st((ulong)val, buf+5, elem->prec);
  return buf;
}

//-----------------------------------------------------------------------------
_nearcall char *fmtOvenState1(DispElem far *elem, char *buf)
{
  strcpy(buf, (Cur.enableOvenOutput[0] ? BlueText : RedText));

  if (Cur.ovenOutErrInterlock[0])
    strcat(buf, "INTLK");
  else
    strcat(buf, ((digitalOut & OVEN1_OUT_OFF) ? "  OFF" : "   ON") );

  return buf;
}
//--------------------------------------------------------------
_nearcall char *fmtOvenState2(DispElem far *elem, char *buf)
{
  strcpy(buf, (Cur.enableOvenOutput[1] ? BlueText : RedText));

  if (Cur.ovenOutErrInterlock[1])
    strcat(buf, "INTLK");
  else
    strcat(buf, ((digitalOut & OVEN2_OUT_OFF) ? "  OFF" : "   ON") );

  return buf;
}

//-----------------------------------------------------------------------------
float biasReadmA, sptrReadmA;

const char far DevStateStr[] = {
/*
"----- RF Gen Supply -----     ----- Bias Supply -----   ---- Oven Status ----\r\n"
" Readback:     -####.# V       Readback:  -####.## V     Const.Cur Mode:   #\r\n"
" Setpoint:     -####.# V       Setpoint:  -####.## V     Foldback Protect: #\r\n"
" Current:      -####.# A       Current:   -####.## A     Auto Restart:     #\r\n"
" Watts:        -####.# W       Assert ON:      ###       Output ON:        #\r\n"
" \r\n"
" HotShot PS:       ###        --- Sputter Disk PS ---    OverVolt Error:   #\r\n"
" NovaStar PS:      ###         Readback:  -####.## V     OverTemp Error:   #\r\n"
"                               Setpoint:  -####.## V     Input Volt Error: #\r\n"
" RF PS Fault:    #####         Current:   -####.## A     Foldback Error:   #\r\n"
" RF PS Heat:       ###         Assert ON:      ###       Prog/Com Stat:x####\r\n"
" RF PS Ready:      ### \r\n"
" Serial Com OK:    ###        ----- Oven Supply -----   --(via Serial)--\r\n"
"                               Readback:  -####.## V     -##.### V \r\n"
" Reset Pulse:      ###         Setpoint:  -####.## V     -##.### V,-##.### A \r\n"
" Stop Pulse        ###         Current:   -####.## A     -##.### A \r\n"
" Start Pulse:      ###         State:        ##### \r\n"
"                               Serial Com OK:  ### \r\n"
"\r\n"
*/

"----- Oven1/Oven2 Status -----       ----- Bias Supply ----- \r\n"
" Const.Cur Mode:      #      #        Readback:  -####.## V \r\n"
" Foldback Protect:    #      #        Setpoint:  -####.## V \r\n"
" Auto Restart:        #      #        Current:   -####.## mA \r\n"
" Output ON:           #      #        Assert ON:      ### \r\n"
" OverVolt Error:      #      # \r\n"
" OverTemp Error:      #      #       --- Sputter Disk PS --- \r\n"
" Input Volt Error:    #      #        Readback:  -####.## V \r\n"
" Foldback Error:      #      #        Setpoint:  -####.## V \r\n"
" Prog/Com Stat:   x####  x####        Current:   -####.## mA \r\n"
"                                      Ext. Interlock: ### \r\n"
" Serial Settings:-##.## -##.## V      Assert ON:      ### \r\n"
"                 -##.## -##.## A \r\n"
" Serial Readback:-##.## -##.## V \r\n"
"                 -##.## -##.## A \r\n"
"\r\n"
" Analog Setpoint:-##.## -##.## V \r\n"
" Analog Readback:-##.## -##.## V \r\n"
"                 -##.## -##.## A \r\n"
" State:           #####  ##### \r\n"
"\r\n"
};

const DispElem  far dispElems[] = {
//  { 15, 4, VFPV(rfgenReadV),    BLUE, fmtF32,    7,1 },
//  { 15, 5, VFPV(rfgenSetV),     BLUE, fmtF32,    7,1 },
//  { 15, 6, VFPV(rfgenReadA),    BLUE, fmtF32,    7,1 },
//  { 15, 7, VFPV(rfgenReadW),    BLUE, fmtF32,    7,1 },

//  { 19,15, VFPV(Cur.rfgenInactive), "", fmtInactive, 3,3 },

  { 22, 3, VFPV(Cur.ovenOpStatus[0]),     "", fmtBinary1, 7,1 },
  { 22, 4, VFPV(Cur.ovenOpStatus[0]),     "", fmtBinary1, 6,1 },
  { 22, 5, VFPV(Cur.ovenOpStatus[0]),     "", fmtBinary1, 5,1 },
  { 22, 6, VFPV(Cur.ovenOpStatus[0]),     "", fmtBinary1, 4,1 },
  { 22, 7, VFPV(Cur.ovenAlarmStatus[0]),  "", fmtBinary1, 4,1 },
  { 22, 8, VFPV(Cur.ovenAlarmStatus[0]),  "", fmtBinary1, 3,1 },
  { 22, 9, VFPV(Cur.ovenAlarmStatus[0]),  "", fmtBinary1, 2,1 },
  { 22,10, VFPV(Cur.ovenAlarmStatus[0]),  "", fmtBinary1, 1,1 },
  { 19,11, VFPV(Cur.ovenProgStatus[0]),   "", fmtBinary1, 0,4 },

  { 17,13, VFPV(serOven1SetV),   "", fmtF32ser1,   6,2 },
  { 17,14, VFPV(serOven1SetA),   "", fmtF32ser1,   6,2 },
  { 17,15, VFPV(serOven1ReadV),  "", fmtF32ser1,   6,2 },
  { 17,16, VFPV(serOven1ReadA),  "", fmtF32ser1,   6,2 },

  { 17,18, VFPV(oven1SetV),      BLUE, fmtF32,    6,2 },
  { 17,19, VFPV(oven1ReadV),     BLUE, fmtF32,    6,2 },
  { 17,20, VFPV(oven1ReadA),     BLUE, fmtF32,    6,2 },
  { 18,21, VFPV(digitalOut),        "", fmtOvenState1, 5,5 },
//  { 47,20, VFPV(Cur.ovenInactive[0]),  "", fmtInactive,  3,3 },

  { 29, 3, VFPV(Cur.ovenOpStatus[1]),     "", fmtBinary2, 7,1 },
  { 29, 4, VFPV(Cur.ovenOpStatus[1]),     "", fmtBinary2, 6,1 },
  { 29, 5, VFPV(Cur.ovenOpStatus[1]),     "", fmtBinary2, 5,1 },
  { 29, 6, VFPV(Cur.ovenOpStatus[1]),     "", fmtBinary2, 4,1 },
  { 29, 7, VFPV(Cur.ovenAlarmStatus[1]),  "", fmtBinary2, 4,1 },
  { 29, 8, VFPV(Cur.ovenAlarmStatus[1]),  "", fmtBinary2, 3,1 },
  { 29, 9, VFPV(Cur.ovenAlarmStatus[1]),  "", fmtBinary2, 2,1 },
  { 29,10, VFPV(Cur.ovenAlarmStatus[1]),  "", fmtBinary2, 1,1 },
  { 26,11, VFPV(Cur.ovenProgStatus[1]),   "", fmtBinary2, 0,4 },

  { 24,13, VFPV(serOven2SetV),   "", fmtF32ser2,   6,2 },
  { 24,14, VFPV(serOven2SetA),   "", fmtF32ser2,   6,2 },
  { 24,15, VFPV(serOven2ReadV),  "", fmtF32ser2,   6,2 },
  { 24,16, VFPV(serOven2ReadA),  "", fmtF32ser2,   6,2 },

  { 24,18, VFPV(oven2SetV),      BLUE, fmtF32,    6,2 },
  { 24,19, VFPV(oven2ReadV),     BLUE, fmtF32,    6,2 },
  { 24,20, VFPV(oven2ReadA),     BLUE, fmtF32,    6,2 },
  { 25,21, VFPV(digitalOut),        "", fmtOvenState2, 5,5 },

  { 49, 3, VFPV(biasReadV),     BLUE, fmtF32,    8,2 },
  { 49, 4, VFPV(biasSetV),      BLUE, fmtF32,    8,2 },
  { 49, 5, VFPV(biasReadmA),    BLUE, fmtF32,    8,2 },

  { 49, 9, VFPV(sptrReadV),     BLUE, fmtF32,    8,2 },
  { 49,10, VFPV(sptrSetV),      BLUE, fmtF32,    8,2 },
  { 49,11, VFPV(sptrReadmA),    BLUE, fmtF32,    8,2 },

  { 0,0, NULL, NULL, NULL, 0,0 }
};

//--- ATTR_NAME macros currently defined in fcntl.h ---
const DigDispElem far digInElems[] = {
//{ xx,yy,  BIT_MASK,          "0STR",ATTR_NAME,  "1STR",ATTR_NAME },
  { 54,12,  SPTR_INTLK_OK,     "YES",RED,         " NO",BLUE   },
//{ 19, 9,  RF_PS_HotShot,     "YES",BLUE,        " NO",RED    },
//{ 19,10,  RF_PS_NovaStar,    "YES" ,BLUE,       " NO",RED    },
//{ 17,12,  RF_PS_NOT_FAULT,   "FAULT",RED,       "   NO",BLUE },
//{ 19,13,  RF_PS_NOT_HEAT_ON, " ON",BLUE,        "OFF",RED    },
//{ 19,14,  RF_PS_NOT_READY,   "YES",BLUE,        " NO",RED    },

  { 0,0, 0, NULL,NULL, NULL,NULL }
};

const DigDispElem far digOutElems[] = {
//{ xx,yy,  BIT_MASK,      "0STR",ATTR_NAME,  "1STR",ATTR_NAME },
//{ 19,17,  RF_PS_RESET,   "YES",RED,     " NO",BLUE },
//{ 19,18,  RF_PS_STOP,    "YES",RED,     " NO",BLUE },
//{ 19,19,  RF_PS_START,   "YES",RED,     " NO",BLUE },
  { 54, 6,  BIAS_PS_OFF,   " ON",BLUE,    "OFF",RED  },
  { 54,13,  SPTR_PS_OFF,   " ON",BLUE,    "OFF",RED  },
//{ 47,19,  OVEN_OUT_OFF,  " ON",BLUE,    "OFF",RED  },

  { 0,0,  0,  NULL,NULL,  NULL,NULL }
};


//-----------------------------------------------------------------------------
//  "state":  Show continuously updated state of the system
//-----------------------------------------------------------------------------
void _task showState(int argc, char **argv)
{
  bool  blinkOff;
  char  temp[128];


  printf(clsCmd);  blinkOff = false;

  CursorTo(0, 2);  printf(DevStateStr);

  for (;;)  {
    CursorTo(0, 0);  blinkOff = !blinkOff;

    stdHeader();  displayElems((DispElem far *)dispElems, blinkOff, temp);

    biasReadmA = fmult(Next.biasReadA, 1000.0);
    sptrReadmA = fmult(Next.sptrReadA, 1000.0);

    displayDigElems(digitalIn, (DigDispElem far *)digInElems, blinkOff, temp);
    displayDigElems(digitalOut, (DigDispElem far *)digOutElems, blinkOff, temp);

    CursorTo(0,22);  fflush(stdout);  CheckETX();  sleepMS(500);
  }

}


//-----------------------------------------------------------------------------
//  Application-specific commands
//-----------------------------------------------------------------------------
#pragma warn -775

void _task cmd(int argc, char **argv)
{
  bool  validCmd;
  int  i;
  uint  idx;
  char  *psName, *cmdStr;
  float  *setV;


#define SetTrue(val)  { \
  _EnterCritical_();  val = true;  _ExitCritical_();  \
  validCmd = true;  break; \
}
#define SetVal(val)  { \
  KLOCK();  val = atof(argv[3]);  KUNLOCK();  \
  validCmd = true;  break; \
}


  validCmd = false;

  do  {
    if (argc < 3)  break;

    psName = argv[1];  cmdStr = argv[2];

    if (StriEq(cmdStr, "set") AND (argc < 4))  break;
/*
    if (StriEq(psName, "rfgen"))   {
      if (StriEq(cmdStr, "start"))  SetTrue(rfgenStartCmd);
      if (StriEq(cmdStr, "stop"))   SetTrue(rfgenStopCmd);
      if (StriEq(cmdStr, "reset"))  SetTrue(rfgenResetCmd);
      if (StriEq(cmdStr, "set"))    SetVal(rfgenSetV);
      if (StriEq(cmdStr, "send"))  if (argc > 3)  {
        KLOCK();
        *rfgenResp = 0;
        strncpy(rfgenCmd, argv[3], sizeof(rfgenCmd)-1);
        KUNLOCK();
        for (i=0; i<20; ++i)  {
          if (*rfgenResp)  { printf("[%.128s]\r\n", rfgenResp);  break; }
          sleepMS(100);
        }
        validCmd = true;
      }
    }
*/
    if (StriEq(psName, "bias"))   {
      if (StriEq(cmdStr, "on"))     SetTrue(biasOnCmd);
      if (StriEq(cmdStr, "off"))    SetTrue(biasOffCmd);
      if (StriEq(cmdStr, "set"))    SetVal(biasSetV);
    }

    if (StriEq(psName, "sptr"))   {
      if (StriEq(cmdStr, "on"))     SetTrue(sptrOnCmd);
      if (StriEq(cmdStr, "off"))    SetTrue(sptrOffCmd);
      if (StriEq(cmdStr, "set"))    SetVal(sptrSetV);
    }

    if (StriEq(psName, "oven1") OR StriEq(psName, "oven2"))   {
      idx = (uint)psName[4] - '1';
      setV = (idx ? &oven2SetV : &oven1SetV);
      if (StriEq(cmdStr, "on"))     SetTrue(ovenOnCmd[idx]);
      if (StriEq(cmdStr, "off"))    SetTrue(ovenOffCmd[idx]);
      if (StriEq(cmdStr, "reset"))  SetTrue(ovenResetCmd[idx]);
      if (StriEq(cmdStr, "set"))    SetVal(*setV);
      if (StriEq(cmdStr, "send"))  if (argc > 3)  {
        KLOCK();
        *ovenResp[idx] = 0;
        strncpy(ovenCmd[idx], argv[3], sizeof(ovenCmd[idx])-1);
        KUNLOCK();
        for (i=0; i<20; ++i)  {
          if (*ovenResp[idx])  {
            printf("[%.128s]\r\n", ovenResp[idx]);  break; }
          sleepMS(100);
        }
        validCmd = true;
      }
    }

  }  while(0);

  if (! validCmd)  {
    printf("\r\n"
           "Usage:  cmd bias|sptr| on|off|set #\r\n"
           "            oven1 on|off|reset|set #|send <str>\r\n"
           "            oven2 on|off|reset|set #|send <str>\r\n"
//           "            rfgen start|stop|reset|set #\r\n"
//           "            rfgen|oven send <str>\r\n"
          );
    Exit(-1);
  }

  Exit(0);
}

#pragma warn 775

//=============================================================================
//  Event-Handler functions called by the appl task to process event messages
//=============================================================================
//APPLCODE

//-----------------------------------------------------------------------------
//  Event-Handler for a request to ...
//-----------------------------------------------------------------------------
/*
void EVENT_MyEvent(EventMsg *event)
{
  uint  idx;
  float  setpoint;
  char  temp[16];


  idx = event->chanNum;

  if (error)  {
    fprintf(DiagLog, "error msg, chan:\n", idx+1);
    return;
  }

  setpoint = *((float *)event->data);

  fprintf(DiagLog, "New SetPoint for PS%d: %s\n",
     idx+1, fmtFloat(setpoint, 8,3, temp) );

  *((float *)((&iSet_1_RegInfo + idx)->regVal)) = setpoint;

  anaOut[idx] = ISETtoVolts(idx, setpoint);
}
*/


//=============================================================================
//  Update the the state of the application logic in well-defined phases
//=============================================================================
// IMPORTANT NOTES and RULES:
//
//  - The code to change ANY value must be in ONE place.  NEVER spread the
//    logic for updating a single value around.  There should be ONE place
//    to look to see how a value gets set.
//
//  - "Cur" values are basically "inputs" - literally (they are set based
//    on the just-read state of a digital or analog input) and/or figuratively
//    (they hold a value set by some external source, such as a cmd from a
//    remote client or another user-interface task).
//
//  - "Next" values are basically "outputs" - literally (they determine the
//    next state of a digital or analog output) and/or figuratively (they
//    can affect the state of the logic for the next pass).
//
//  - The "Cur" OR the "Next" instance of a value may be changed - NEVER
//    both of them (with the exception of the first step below that copies
//    all of cur to prev and next to cur).
//
//  - A value MUST be in the ApplState struct if one of the following applies:
//     - The application logic is dependent on if or how much the value has
//       changed between passes.
//     - The value is used as both input (used in in a conditional and/or in
//       an expression that can affect the result of an assignment) and output
//       (the target of an assignment).
//
//  - Variables that are NOT in the ApplState struct but are USED (in
//    conditionals or other expressions) by the application logic must be
//    treated just like "Cur" values.  Variables SET (the target in an
//    assignment statement) by the application logic must be treated just
//    like "Next" values.
//
//  - When the logic requires comparing an "input" to a related "output", be
//    sure to use the "Cur" value for the "input", and the "Prev" value for
//    the "output".
//
//    This is because during the processing of the loop, the "Cur" values
//    represent the input values just read (what they are "now") and the
//    "Prev" values represent the values the outputs were set to at the
//    beginning of the prev pass (i.e. they are the values that "caused"
//    the inputs to have their "Cur" values).
//
//    During a pass the output values in "Cur" represent what the outputs
//    where set to AFTER the "Cur" inputs were updated, so comparing them
//    makes no sense.  Doing so is the equivalent of comparing a past response
//    to a condition that didn't exist at the time.
//=============================================================================


//=============================================================================
//  Definitions for functions that handle setting "Next" values that serve
//  as inputs to the application control logic (run by the "applic" task in
//  a periodically executed loop).
//
//  While care must be taken in these functions to use locks to avoid the
//  possibility of being preempted in the middle of an operation, such locks
//  must be kept to as short a time as possible to avoid excessive latency
//  in the main control loop.
//
//  For REALLY short locks, using a critical section is best, because this
//  has the least overhead (but it has to be REALLY short to avoid causing
//  problems for interrupt handlers).
//
//  For longer locks (ones that involve anything but the most trivial
//  conditional logic), use either KLOCK() and KUNLOCK() (which disable the
//  scheduler) or a more specific global semaphore/mutex lock.
//=============================================================================

/*Examples
//-----------------------------------------------------------------------------
void onCmd(void)
{
  _EnterCritical_();
  Next.onCmd = true;  Next.offCmd = false;
  _ExitCritical_();

  fprintf(DiagLog, "ON Cmd\n");
}

//-----------------------------------------------------------------------------
void offCmd(void)
{
  _EnterCritical_();
  Next.offCmd = true;  Next.onCmd = false;
  _ExitCritical_();

  fprintf(DiagLog, "OFF Cmd\n");
}

//-----------------------------------------------------------------------------
void resetCmd(void)
{
  _EnterCritical_();  Next.resetCmd = true;  _ExitCritical_();

  fprintf(DiagLog, "RESET Cmd\n");
}

//-----------------------------------------------------------------------------
void newSetpoint(float newAmps)
{
  char  temp[16];


  KLOCK();

  if (FABS(newAmps) > maxAmpSet)  newAmps = (float)SIGN(newAmps) * maxAmpSet;

  Next.ampSet = newAmps;

  KUNLOCK();

  fprintf(DiagLog, "New Setpoint: %s\n", fmtFloat(newAmps, 10,3, temp) );
}

//-----------------------------------------------------------------------------
*/



//-----------------------------------------------------------------------------
//  Check the range of a new setVal[] value
//-----------------------------------------------------------------------------
void checkSetValRange(float *regVal, float *applVal, uint dacIdx)
{
  float  min, max;


  if (dacIdx >= NumHdwDAC)  return;

  min = setValMin[dacIdx];  if (*regVal < min)  *regVal = min;

  max = setValMax[dacIdx];  if (*regVal > max)  *regVal = max;

  *applVal = *regVal;
}


//-----------------------------------------------------------------------------
//  Check for read/set difference exceeding thresholds
//-----------------------------------------------------------------------------
void checkThresholds(char *name, ulong *readErrStartTime,
                     bool *runtimeErr, bool prevRuntimeErr,
                     float maxErrV, ulong maxErrMS,
                     float setV, float readV, float readA)
{
  float  err;
  char  temp[48];


  if (!maxErrMS)  return;

  err = setV - readV;
  if (FABS(err) <= maxErrV)  {
    *readErrStartTime = 0;  *runtimeErr = false;  }
  else  {
    if (!*readErrStartTime)  *readErrStartTime = GetMS();
    if (msSince(*readErrStartTime) > maxErrMS)  *runtimeErr = true;
  }

  if (*runtimeErr == prevRuntimeErr)  return;

  if (*runtimeErr)
    printf("%s Set/Read error exceeded thresholds:\r\n"
           "  Analog:  Volts: %.16s  SetV: %.16s  Amps: %.16s\r\n",
                  name, Engineering(readV, 6, temp),
                  Engineering(setV, 6, temp+16),
                  Engineering(readA, 6, temp+32)
          );
  else
    printf("%s Set/Read error now within thresholds\r\n", name);
}


//=============================================================================
//  Update the "Cur"rent state of the controller
//=============================================================================
//
//  Just before this function is called, readings were taken of the current
//  state of the hardware, and any changes to the output states specified by
//  the last pass were applied (just after taking the readings).  These values
//  are copied to Next as needed and constitute the Hardware Inputs.
//
//  Other tasks can affect the state of the control logic by changing values
//  in Next as well.  These values constitute Software Inputs.
//
//  This function is responsible for the Asynchronous logic that is applied
//  each pass.  It updates values in Next (or other global variables) based
//  on the Software and Hardware Input values in Next.  This portion of the
//  logic is analogous to unclocked gates in an electronic circuit.
//
//  After the Asychronous logic is applied, we set Prev = Cur and Cur = Next.
//  The result is that Prev contains the controller state as it was during the
//  previous pass, and Cur indicates the current state of the controller.
//
//=============================================================================
void updateCurState(void)
{
  uint  idx;


  if (current != applTask)  return;  // must only be called by "applic" task

//-----------------------------------------------------------------
// The following conditions exist at this point:
// - The analog and digital hardware inputs have just been read
// - The outputs were just set to values computed during the last pass
// - "Cur" contains the state the inputs were in at the START of
//   of the previous pass, and the state the outputs were in DURING
//   the previous pass
// - "Next" contains the values that the outputs were just set
//   to and any "input" values that were changed by other tasks
//------------------------------------------------------------------

// Now we update the values in "Next" that are based solely on the
// state of the hardware inputs we just read and values maintained by
// by other tasks (basically, we deal with READINGS here: Values that
// we need but are not written to "Next" by another task)

  Next.digitalIn = digitalIn;

//APPLCODE

  Next.sptrInterlock = !(digitalIn & SPTR_INTLK_OK);

//  if (!(digitalIn & RF_PS_HotShot))
//    Next.rfSupplyType = RF_TYPE_HotShot;   else
//  if (!(digitalIn & RF_PS_NovaStar))
//    Next.rfSupplyType = RF_TYPE_NovaStar;
//  else
//    Next.rfSupplyType = RF_TYPE_UNKNOWN;

//  Next.rfgenReadV = rfgenReadV;
//  Next.rfgenReadA = rfgenReadA;
//  Next.rfgenReadW = rfgenReadW;

//  Next.rfgenFault =  !(digitalIn & RF_PS_NOT_FAULT);
//  Next.rfgenHeatOn = !(digitalIn & RF_PS_NOT_HEAT_ON);
//  Next.rfgenReady =  !(digitalIn & RF_PS_NOT_READY);

  ovenRespAge[0] = msSince(ovenLastRespTime[0]);
  ovenRespAge[1] = msSince(ovenLastRespTime[1]);

  Next.ovenInactive[0] = (ovenRespAge[0] > 3000);
  Next.ovenInactive[1] = (ovenRespAge[1] > 3000);
//  Next.rfgenInactive = (msSince(rfgenLastRespTime) > 3000);

  Next.ovenReadV[0] = readVal[OVEN1_READ_V];
  Next.ovenReadA[0] = readVal[OVEN1_READ_A];

  Next.ovenReadV[1] = readVal[OVEN2_READ_V];
  Next.ovenReadA[1] = readVal[OVEN2_READ_A];

  Next.biasReadV = readVal[BIAS_READ_V];
  Next.biasReadA = readVal[BIAS_READ_A];

  Next.sptrReadV = readVal[SPTR_READ_V];
  Next.sptrReadA = readVal[SPTR_READ_A];

  Next.biasSetV = biasSetV;
  Next.sptrSetV = sptrSetV;
//  Next.rfgenSetV = rfgenSetV;


//--- apply range limits to new settings ---
  checkSetValRange(&oven1SetV,  &Next.ovenSetV[0],  OVEN1_SET_V);
  checkSetValRange(&oven2SetV,  &Next.ovenSetV[1],  OVEN2_SET_V);
  checkSetValRange(&biasSetV,  &Next.biasSetV,  BIAS_SET_V);
  checkSetValRange(&sptrSetV,  &Next.sptrSetV,  SPTR_SET_V);
//  checkSetValRange(&rfgenSetV, &Next.rfgenSetV, RFGEN_SET_V);


#define CheckThresholds(name, prefix)  \
  checkThresholds(name, \
                  &prefix##ReadErrStartTime,  \
                  &Next.prefix##RuntimeErr, \
                  Cur.prefix##RuntimeErr,  \
                  prefix##MaxErrV, \
                  prefix##MaxErrMS,  \
                  Cur.prefix##SetVassert, \
                  Next.prefix##ReadV, \
                  Next.prefix##ReadA)

#define CheckOvenThresholds(name, prefix, idx)  \
  checkThresholds(name, \
                  &ovenReadErrStartTime[idx],  \
                  &Next.ovenRuntimeErr[idx], \
                  Cur.ovenRuntimeErr[idx],  \
                  prefix##MaxErrV, \
                  prefix##MaxErrMS,  \
                  Cur.ovenSetVassert[idx], \
                  Next.ovenReadV[idx], \
                  Next.ovenReadA[idx])

//--- check set/read error against allowed thresholds ---
//--- (these checks are very time consuming at present) ---
  CheckOvenThresholds("Oven1", oven1, 0);
  CheckOvenThresholds("Oven2", oven2, 1);
  CheckThresholds("Bias", bias);
  CheckThresholds("Sputter", sptr);


// Next we do the asynchronous/gate-level type logic.  This logic often has
// order of execution dependencies so be careful!

//--- oven1PowerOnTime ---
  if (digitalOut & OVEN1_PS_OFF)
    ovenPowerOnTime[0] = 0;
  else
    if (!ovenPowerOnTime[0])  ovenPowerOnTime[0] = GetMS();
//--- oven2PowerOnTime ---
  if (digitalOut & OVEN2_PS_OFF)
    ovenPowerOnTime[1] = 0;
  else
    if (!ovenPowerOnTime[1])  ovenPowerOnTime[1] = GetMS();

  for (idx=0; idx<2; ++idx)  {
  //--- Next.ovenModeOn[] ---
    if (Next.ovenOutErrInterlock[idx])  {
      Next.ovenModeOn[idx] = false;
      if (ovenOnCmd[idx])
        printf("Oven%d ON cmd ignored while Interlock exists\n", idx+1);
    }
    else  {
      if (ovenOnCmd[idx])  Next.ovenModeOn[idx] = true;
      if (ovenOffCmd[idx])  Next.ovenModeOn[idx] = false;
    }
  //--- ovenOffCmdTime[] ---
    if (ovenOnCmd[idx])  ovenOffCmdTime[idx] = 0;
    if (ovenOffCmd[idx])
      if (!ovenOffCmdTime[idx])  ovenOffCmdTime[idx] = GetMS();
  }





// Next we make a copy of the application state before we apply all the
// logic that updates the overall state.  This allows us to have a copy of
// all the relevant values that remain unchanged during the processing,
// avoiding all kinds of order-of-execution dependencies.

// We also make a copy of the state from last time before overwriting it
// so we can see if/how values have changed since the last pass

// Note that a copy is done (rather than just switching pointer values) to
// insure all values are set in both the source and destination.  If we
// simply kept pointers and swapped the pointer values, then EVERY value
// would have to be updated independently by the code, which is a pain to
// program, adds overhead and code, and is a VERY easy way to introduce odd
// bugs.  Overall, copying everything here is actually faster overall,
// simpler, and much safer.

  Prev = Cur;  Cur = Next;

//--- At this point the Cur copy contains what the inputs were at the start ---
//--- of this pass, and the Prev copy contains the output values from the ---
//--- the previous pass (i.e. the output values that resulted in the input ---
//--- values in Cur) ---


//--- log msgs for bit chgs only if not starving other processes for cycles ---
  /* if (applLoopTime < 15) */  logDigIOChgs(false);

//  if (Cur.rfgenInactive != Prev.rfgenInactive)
//    printf("RF Gen PS %s responding\n",
//           Cur.rfgenInactive ? "NOT" : "started");

  for (idx=0; idx<2; ++idx)  {
    if (Cur.ovenInactive[idx] != Prev.ovenInactive[idx])
      printf("Oven%d PS %s responding\n",
             idx+1, Cur.ovenInactive[idx] ? "NOT" : "started");
    if (ovenOnCmd[idx])  {
      printf("Oven%d PS ON request\n", idx+1);
      if (Cur.enableOvenOutput[idx])  printf("Oven%d PS already ON\n", idx+1);
    }
    if (ovenOffCmd[idx])  {
      printf("Oven%d PS OFF request\n", idx+1);
      if (!Cur.enableOvenOutput[idx])  printf("Oven%d PS already OFF\n", idx+1);
    }
    if (ovenResetCmd[idx])  printf("Oven%d PS RESET request\n", idx+1);
  }

//  if (rfgenStartCmd)  {
//    printf("Start RF Gen PS request\n");
//    if (Cur.rfgenReady)  printf("RF Gen PS already ON\n");
//  }
//  if (rfgenStopCmd)  {
//    printf("Stop RF Gen PS request\n");
//    if (!Cur.rfgenHeatOn AND !Cur.rfgenReady)
//      printf("RF Gen PS already OFF\n");
//  }
//  if (rfgenResetCmd)  printf("Reset RF Gen PS request\n");

  if (biasOnCmd)  {
    printf("Bias PS ON request\n");
    if (Cur.assertOnBiasPS)  printf("Bias PS already ON\n");
  }
  if (biasOffCmd)  {
    printf("Bias PS OFF request\n");
    if (!Cur.assertOnBiasPS)  printf("Bias PS already OFF\n");
  }

  if (sptrOnCmd)  {
    printf("Sputter PS ON request\n");
    if (Cur.assertOnSptrPS)  printf("Sputter PS already ON\n");
    if (Cur.sptrInterlock)  printf("Ext interlock still asserted\n");
  }
  if (sptrOffCmd)  {
     printf("Sputter PS OFF request\n");
     if (!Cur.assertOnSptrPS)  printf("Sputter PS already OFF\n");
  }


}


//=============================================================================
//  Update the values in "Next" based on the "Cur" and "Prev" values.
//
//  Be sure to read the "NOTES on Locking" section above....
//=============================================================================
//
//  This function contains the Synchronous logic used to compute the next
//  state we want the controller to be in.
//
//  A few simple rules that apply to all the values used by this function:
//
//    - All values changed within this function are "outputs"
//
//    - Every value read by this function is an "input"
//
//    - There must be no overlap between inputs and outputs
//
//    - The output values from this function must not be altered outside of
//      this function.
//
//  Any global variable can be an input OR an output, but they must NEVER
//  serve both purposes.  Most inputs come from Prev or Cur, while most
//  outputs are in Next.  If the logic requires detecting a change in a
//  value between passes (edge-triggered logic), then the value should be
//  part of the ApplState structure so we can compare its Prev and Cur values.
//
//  Following these simple rules allows the code to be organized as small,
//  self-contained blocks, with well-defined inputs and outputs.  Furthermore,
//  there are no order-of-execution dependencies between blocks of code.  The
//  blocks can be executed in any order:  The result will always be the same.
//
//
//  Each output value (or VERY small "set" of output values) must be updated
//  in a single, self-contained block of code.  NO output may be referenced
//  OR altered in ANOTHER block of code:  This is a declarative, data-oriented
//  style of programming and has many advantages including:
//
//     - There are no order-of-execution dependencies across nodes/blocks
//       within a pass.  They can be executed in any order and the result
//       will be the same (get used to it - this method will be the norm as
//       computer languages and hardware add support for atomic transactions
//       and computers evolve in to neural-net-like processor arrays - FPGAs
//       are just the beginning of this trend).
//
//     - Encapsulating all logic for an output within a single node/block
//       leaves just one place to start when tracking a problem backward from
//       the result.  Far simpler than debugging code structured around the
//       traditional process/event-driven approach which can contain numerous
//       independent paths of execution leading to an observed result.
//
//  Each node/block-of-code should be prefaced with the following information:
//   - The name(s) of the value(s) it modifies
//   - A descr of the meaning of the value(s) and the logic used
//   - A list of the values that serve as "inputs" to the node
//
//=============================================================================
void updateNextState(void)
{
  uint  idx;


  if (current != applTask)  return;  // must only be called by "applic" task

//APPLCODE

 //--------------------------------------------------------------------
 // Next.enableOvenPower[]
 //    Description:
 //       Determines if the output that enables the AC power to the Oven
 //       supply should be asserted.  The power to the Oven supply is
 //       turned off immediately if an Interlock occurs or oven#PwrOffDelay
 //       ms after an off command is received.
 //    Inputs:
 //       ovenOnCmd[]
 //       ovenOffCmdTime[]
 //--------------------------------------------------------------------
  for (idx=0; idx<2; ++idx)  {
    uint offDelay = idx ? oven1PwrOffDelay : oven2PwrOffDelay;
    if (Cur.ovenOutErrInterlock[idx]
        OR (ovenOffCmdTime[idx] AND (msSince(ovenOffCmdTime[idx]) > offDelay)) )
      Next.enableOvenPower[idx] = false;
    else
      if (Cur.ovenModeOn[idx])  Next.enableOvenPower[idx] = true;
  }

 //--------------------------------------------------------------------
 // Next.enableOvenOutput[]
 //    Description:
 //       Determines if the output from the Oven PS should be enabled.
 //       The output is enabled a short time after the power to the
 //       supply is enabled.  The output is disabled if we get an Off
 //       command, the power is disabled, or an uncleared interlock
 //       exists.
 //    Inputs:
 //       ovenPowerOnTime
 //       Cur.ovenModeOn
 //       Cur.enableOvenPower
 //       Cur.ovenOutErrInterlock
 //--------------------------------------------------------------------
  for (idx=0; idx<2; ++idx)  {
    if (!Cur.ovenModeOn[idx] OR !Cur.enableOvenPower[idx]
        OR Cur.ovenOutErrInterlock[idx])
      Next.enableOvenOutput[idx] = false;
    else
      if (ovenPowerOnTime[idx])  if (msSince(ovenPowerOnTime[idx]) > 2000)
        Next.enableOvenOutput[idx] = true;
  }

 //--------------------------------------------------------------------
 // Next.ovenOutErrStartTime[0]
 //    Description:
 //       Keeps track of the beginning of a period during which the
 //       Oven PS reports that the state of its output is different
 //       than what we told it to be.
 //    Inputs:
 //       Cur.enableOvenOutput[]
 //       Cur.ovenOpStatus[]
 //       ovenResetCmd[]
 //--------------------------------------------------------------------
  {
    bool  shouldBeOn, isOn;

    for (idx=0; idx<2; ++idx)  {
      if (ovenResetCmd[idx])  {
        Next.ovenOutErrStartTime[idx] = 0;  postOvenResetMsgs[idx] = 0;
        forceOvenErr = 0;  //tdebug
      }
      else  {
        // NOTE: The newOvenResp[] and postOvenResetMsgs[] values are needed
        //       to insure the data read from the Oven PS is recent and
        //       not something that was "in the pipeline" at the time
        //       we got the reset command.
        if (newOvenResp[idx] AND (postOvenResetMsgs[idx] > 4))  {
          shouldBeOn = BOOL(Cur.enableOvenOutput[idx]);
          isOn = BOOL(Cur.ovenOpStatus[idx] & OOS_OutputOn);
          if (shouldBeOn == isOn)
            Next.ovenOutErrStartTime[idx] = 0;
          else
            if (!Next.ovenOutErrStartTime[idx])
              Next.ovenOutErrStartTime[idx] = GetMS();
        }
      }
    }
  }

 //--------------------------------------------------------------------
 // Next.ovenOutErrInterlock[]
 //    Description:
 //       Indicates that the Oven PS says that the on/off state of its
 //       output has been consistenly different (for several seconds)
 //       than what we been asserting via a hdw signal.
 //    Inputs:
 //       ovenResetCmd[]
 //       Cur.ovenOutErrStartTime[]
 //--------------------------------------------------------------------
  for (idx=0; idx<2; ++idx)  {
    if (ovenResetCmd[idx])
      Next.ovenOutErrInterlock[idx] = false;
    else
      if (Cur.ovenOutErrStartTime[idx])
        if (msSince(Cur.ovenOutErrStartTime[idx]) > 4000)
          Next.ovenOutErrInterlock[idx] = true;

    if (Next.ovenOutErrInterlock[idx] != Cur.ovenOutErrInterlock[idx])
      printf("Oven%d Output State error Interlock %s\n",
             idx+1, Next.ovenOutErrInterlock[idx] ? "Asserted" : "Cleared");
  }

 //--------------------------------------------------------------------
 // Next.ovenSetVassert[]
 //    Description:
 //       The setpoint to actually apply to the Oven supply
 //    Inputs:
 //       Cur.enableOvenOutput[]
 //       Cur.ovenSetV[]
 //--------------------------------------------------------------------
  for (idx=0; idx<2; ++idx)
    Next.ovenSetVassert[idx] =
      (Cur.enableOvenOutput[idx] ? Cur.ovenSetV[idx] : 0.0);



 //--------------------------------------------------------------------
 // Next.assertOnBiasPS
 //    Description:
 //       Determines if the output that enables the Bias supply should
 //       be asserted.
 //    Inputs:
 //       biasOnCmd
 //       biasOffCmd
 //--------------------------------------------------------------------
  if (biasOnCmd)   Next.assertOnBiasPS = true;
  if (biasOffCmd)  Next.assertOnBiasPS = false;

 //--------------------------------------------------------------------
 // Next.biasSetVassert
 //    Description:
 //       The setpoint to actually apply to the Bias supply
 //    Inputs:
 //       Cur.assertOnBiasPS
 //       Cur.biasSetV
 //--------------------------------------------------------------------
  Next.biasSetVassert = Cur.assertOnBiasPS ? Cur.biasSetV : 0.0;


 //--------------------------------------------------------------------
 // Next.assertOnSptrPS
 //    Description:
 //       Determines if the output that enables the Sputter supply should
 //       be asserted.
 //    Inputs:
 //       sptrOnCmd
 //       sptrOffCmd
 //--------------------------------------------------------------------
  if (sptrOnCmd AND !Cur.sptrInterlock)  Next.assertOnSptrPS = true;
  if (sptrOffCmd OR Cur.sptrInterlock)  Next.assertOnSptrPS = false;

 //--------------------------------------------------------------------
 // Next.sptrSetVassert
 //    Description:
 //       The setpoint to actually apply to the Sputter supply
 //    Inputs:
 //       Cur.assertOnSptrPS
 //       Cur.sptrSetV
 //--------------------------------------------------------------------
  Next.sptrSetVassert = Cur.assertOnSptrPS ? Cur.sptrSetV : 0.0;


 //--------------------------------------------------------------------
 // Next.rfgenStartTime
 //    Description:
 //       The time a pulse was started to turn ON the RF Gen supply.
 //    Inputs:
 //       rfgenStartCmd
 //--------------------------------------------------------------------
//  if (rfgenStartCmd)  Next.rfgenStartTime = GetMS();
//  if (msSince(Next.rfgenStartTime) >= 500)  Next.rfgenStartTime = 0;

 //--------------------------------------------------------------------
 // Next.rfgenStopTime
 //    Description:
 //       The time a pulse was started to turn OFF the RF Gen supply.
 //    Inputs:
 //       rgfenStopCmd
 //--------------------------------------------------------------------
//  if (rfgenStopCmd)  Next.rfgenStopTime = GetMS();
//  if (msSince(Next.rfgenStopTime) >= 500)  Next.rfgenStopTime = 0;

 //--------------------------------------------------------------------
 // Next.rfgenResetTime
 //    Description:
 //       The time a pulse was started to RESET the RF Gen supply.
 //    Inputs:
 //       rfgenResetCmd
 //--------------------------------------------------------------------
//  if (rfgenResetCmd)  Next.rfgenResetTime = GetMS();
//  if (msSince(Next.rfgenResetTime) >= 500)  Next.rfgenResetTime = 0;


 //--------------------------------------------------------------------
 // Next.rfgenSetVassert
 //    Description:
 //       The setpoint to actually apply to the RF Gen supply
 //    Inputs:
 //       Cur.rfgenReady
 //       Cur.rfgenSetV
 //--------------------------------------------------------------------
//  Next.rfgenSetVassert = Cur.rfgenReady ? Cur.rfgenSetV : 0.0;



//--- we are done with the event flags now so clear them all ---
//  rfgenStartCmd = rfgenStopCmd = rfgenResetCmd =
  biasOnCmd = biasOffCmd = sptrOnCmd = sptrOffCmd =
  ovenOnCmd[0] = ovenOffCmd[0] = ovenResetCmd[0] = newOvenResp[0] =
  ovenOnCmd[1] = ovenOffCmd[1] = ovenResetCmd[1] = newOvenResp[1] = false;
}



//=============================================================================
//  Next we update output signal values to be written to the outputs at
//  the beginning of the next pass (just after the inputs are read).
//
//  Be sure to read the "NOTES on Locking" section above....
//=============================================================================
void updateOutputState(void)
{
  if (current != applTask)  return;  // must only be called by "applic" task

//APPLCODE

//--- update output values that are based on current conditions (values ---
//--- in "Cur" or any values used in a conditional in the applic logic) ---

#define  Set(bits)    SetBits(Next.digitalOut, (bits))
#define  Clear(bits)  ClearBits(Next.digitalOut, (bits))

//  if (Next.rfgenStartTime)   Clear(RF_PS_START);   else  Set(RF_PS_START);
//  if (Next.rfgenStopTime)    Clear(RF_PS_STOP);    else  Set(RF_PS_STOP);
//  if (Next.rfgenResetTime)   Clear(RF_PS_RESET);   else  Set(RF_PS_RESET);

  if (Next.assertOnBiasPS)   Clear(BIAS_PS_OFF);   else  Set(BIAS_PS_OFF);

  if (Next.assertOnSptrPS)   Clear(SPTR_PS_OFF);   else  Set(SPTR_PS_OFF);

  if (Next.enableOvenOutput[0])  Clear(OVEN1_OUT_OFF); else  Set(OVEN1_OUT_OFF);
  if (Next.enableOvenPower[0])   Clear(OVEN1_PS_OFF);  else  Set(OVEN1_PS_OFF);

  if (Next.enableOvenOutput[1])  Clear(OVEN2_OUT_OFF); else  Set(OVEN2_OUT_OFF);
  if (Next.enableOvenPower[1])   Clear(OVEN2_PS_OFF);  else  Set(OVEN2_PS_OFF);

//--- toggle external WD timer pulse state every 50ms (for ~10Hz sq wave) ----
  if (msSince(wdPulseChanged) >= 50)  {
    wdPulseChanged = GetMS();
    if (Next.digitalOut & EXT_WD_TIMER)
      Clear(EXT_WD_TIMER);
    else
      Set(EXT_WD_TIMER);
  }

  setVal[OVEN1_SET_V] =  Next.ovenSetVassert[0];
  setVal[OVEN2_SET_V] =  Next.ovenSetVassert[1];
  setVal[BIAS_SET_V] =  Next.biasSetVassert;
  setVal[SPTR_SET_V] =  Next.sptrSetVassert;
//  setVal[RFGEN_SET_V] = Next.rfgenSetVassert;
}




//=============================================================================
// The last thing we do is update values read by other tasks
//
//  Be sure to read the "NOTES on Locking" section above....
//=============================================================================
void updateExternals(void)
{
  if (current != applTask)  return;  // must only be called by "applic" task

  digInBits = Cur.digitalIn;
  digOutBits = Cur.digitalOut;

//APPLCODE

  statusBitsAppl = 0;

//  if (Cur.rfgenFault)   statusBitsAppl |= RF_PS_FAULT;
//  if (Cur.rfgenHeatOn)  statusBitsAppl |= RF_PS_HEAT_ON;
//  if (Cur.rfgenReady)   statusBitsAppl |= RF_PS_READY;

//  if (Cur.rfgenInactive)  statusBitsAppl |= RF_PS_INACTIVE;
  if (Cur.ovenInactive[0])   statusBitsAppl |= OVEN1_PS_INACTIVE;
  if (Cur.ovenInactive[1])   statusBitsAppl |= OVEN2_PS_INACTIVE;

  if (Next.enableOvenOutput[0])  statusBitsAppl |= OVEN1_PS_ENABLED;
  if (Next.enableOvenOutput[1])  statusBitsAppl |= OVEN2_PS_ENABLED;
  if (Next.assertOnBiasPS)    statusBitsAppl |= BIAS_PS_ENABLED;
  if (Next.assertOnSptrPS)    statusBitsAppl |= SPTR_PS_ENABLED;

  if (Cur.ovenOutErrInterlock[0])  statusBitsAppl |= OVEN1_OUT_INTLK;
  if (Cur.ovenOutErrInterlock[1])  statusBitsAppl |= OVEN2_OUT_INTLK;

  if (Cur.sptrInterlock)  statusBitsAppl |= SPTR_INTERLOCK;

  sptrReadV = Cur.sptrReadV;
  sptrReadA = Cur.sptrReadA;

  biasReadV = Cur.biasReadV;
  biasReadA = Cur.biasReadA;

  oven1ReadV = Cur.ovenReadV[0];  oven2ReadV = Cur.ovenReadV[1];
  oven1ReadA = Cur.ovenReadA[0];  oven2ReadA = Cur.ovenReadA[1];

  serOven1SetV = Cur.serOvenSetV[0];    serOven2SetV = Cur.serOvenSetV[1];
  serOven1SetA = Cur.serOvenSetA[0];    serOven2SetA = Cur.serOvenSetA[1];
  serOven1ReadV = Cur.serOvenReadV[0];  serOven2ReadV = Cur.serOvenReadV[1];
  serOven1ReadA = Cur.serOvenReadA[0];  serOven2ReadA = Cur.serOvenReadA[1];

//  rfgenReadV = Cur.rfgenReadV;
//  rfgenReadA = Cur.rfgenReadA;
//  rfgenReadW = Cur.rfgenReadW;

//  rfSupplyType = Cur.rfSupplyType;

  oven1Status = (ulong) Cur.ovenOpStatus[0] |
                ((ulong) Cur.ovenAlarmStatus[0] << 8) |
                ((ulong) Cur.ovenProgStatus[0] << 16);

  oven2Status = (ulong) Cur.ovenOpStatus[1] |
                ((ulong) Cur.ovenAlarmStatus[1] << 8) |
                ((ulong) Cur.ovenProgStatus[1] << 16);

  adc00 = adcRaw[0];
  adc01 = adcRaw[1];
  adc02 = adcRaw[2];
  adc03 = adcRaw[3];
  adc04 = adcRaw[4];
  adc05 = adcRaw[5];
}



//-----------------------------------------------------------------------------
//  Parse the response from the Ameritherm Power Supply
//-----------------------------------------------------------------------------
/*
void Process_RF_PS_Resp(char *resp)
{
  char  *ptr;
  int  len;
  float  watts, amps, volts;


  KLOCK();
  watts = rfgenReadW;  amps = rfgenReadA;  volts = rfgenReadV;
  KUNLOCK();

  switch  (rfSupplyType)  {

  // sample response to command "1,data":
  //  (2 lines?)
    case RF_TYPE_HotShot:
      // scan bkwds from end for 1st space char
      for (ptr = resp+strlen(resp)-1; (ptr > resp) AND (*ptr > ' '); --ptr);

      if (strncmp(resp, "Power Out", 9) == 0)  watts = atof(ptr);  else
      if (strncmp(resp, "Amps Out", 8) == 0)  amps = atof(ptr);

      volts = (amps == 0.0) ? 0.0 : (watts / amps);

      break;


  // sample response to command "r001":
  //   0049%V 14%A 7%W 317 KHZ 0.0000 Sec HEAT ON GO 0005 CYC 38
    case RF_TYPE_NovaStar:
      ptr = strtok(resp, " ");
      while (ptr)  {
        if ((len = strlen(ptr)) < 1)  break;
        while (*ptr AND (*ptr == '0'))  { ++ptr;  --len; } // skip any leading 0's

        switch (ptr[len-1])  {
          case 'V':  volts = atof(ptr);  break;
          case 'A':  amps = atof(ptr);  break;
          case 'W':  watts = atof(ptr);  break;
        }

        if (ptr)  ptr = strtok(NULL, " ");
      }
      break;

  }

  KLOCK();
  rfgenReadW = watts;  rfgenReadA = amps;  rfgenReadV = volts;
  KUNLOCK();
}
*/

//-----------------------------------------------------------------------------
//  Process to manage comunications with the external RF Gen power supply
//-----------------------------------------------------------------------------
/*
void _task rfgenCom(int argc, char **argv)
{
  int  i, stat, psType, newType, fflags;
  fd_t  devfd;
  char  pktStartChar;
  ulong  baudRate, lastRcvTime;
  char  *nextCmd;
  char  buf[128];


  redirectStdio();

  devfd = 0;  psType = 0;  rfgenLastRespTime = 0;

  for (;;)  {
    _EnterCritical_();  newType = rfSupplyType;  _ExitCritical_();

    if (newType != psType)  {
      if (devfd)  { close(devfd);  devfd = 0; }
      psType = newType;
    }

    if (! psType)  { sleepMS(500);  continue; }

  //--- (re)open and initialize port for communicating with the RF Gen PS ---
    if (!devfd)  {
      if ((devfd = open("/dev/tty3", O_RDWR, 0)) < 0)  {
        printf("Unable to open port for RF PS\n");  Exit(-3); }

      fflags = fcntl(devfd, F_GETFL, 0);
      fcntl(devfd, F_SETFL, fflags | O_NONBLOCK);

      baudRate = (psType == RF_TYPE_HotShot) ? 38400 : 9600;
      ioctl(devfd, IOC_SET_BAUD, &baudRate);

    //--- set serial driver to use packet mode, waiting for a LF char to ---
    //--- indicate the start of each packet ---
      pktStartChar = 0x0a;  // PS packets start with a LF
      ioctl(devfd, IOC_SET_PKTMODE, &pktStartChar);
    }


  //--- send a command ---
  //HotShot cmd fmt:  #,<cmd>[,arg]
  //NovaStar cmd fmt:  <cmd>[###][,arg]
    if (*rfgenCmd)
      nextCmd = rfgenCmd;
    else
      nextCmd = (psType == RF_TYPE_HotShot) ? "1,data\r" : "r001\r";
    ioctl(devfd, IOC_START_PKT_TX, NULL);
    write(devfd, nextCmd, strlen(nextCmd));
    ioctl(devfd, IOC_END_PKT_TX, NULL);
    *rfgenCmd = 0;

  //--- wait for response packet ---
    lastRcvTime = GetMS();  *buf = 0;
    for (i=0; i<sizeof(buf)-1; )  {
      stat = read(devfd, buf+i, 1);
      if (stat > 0)  {
//        printf("<%02x>", buf[i]);  //tdebug
        buf[i] &= 0x7f;  lastRcvTime  = GetMS();
        if (buf[i] == 0x0D)  break;
        ++i;  continue;
      }
      if ((stat < 0) AND (errno != EAGAIN))  {
        perror("error reading serial packet");  break; }
      if (msSince(lastRcvTime) > 1000)  break;
      sleepMS(50);
    }
    ioctl(devfd, IOC_END_PKT_RX, NULL);

    if (i == 0)  continue;

    if (buf[i] != 0x0D)  {
      printf("incomplete packet: [%.128s]\r\n", buf);  continue; }

    buf[i] = 0;

  //--- process the response ---
    if (ShowPackets)  printf("[%.128s]\r\n", buf);
    _EnterCritical_();  rfgenLastRespTime = GetMS();  _ExitCritical_();
    Process_RF_PS_Resp(buf);  sleepMS(330);
  }

}
*/

//-----------------------------------------------------------------------------
//  Parse the response from the Ameritherm Power Supply
//-----------------------------------------------------------------------------
int Process_Oven_PS_Resp(char *resp, fd_t devfd, uint idx)
{
  char  *ptrAV, *ptrSV, *ptrAA, *ptrSA, *ptrOS, *ptrAL, *ptrPS,
        *ptrEnd, *newline, *msg;
  bool  resetAlarms, forceErr;
  float  actualVolts, setVolts, actualAmps, setAmps,
         anaVolts, anaSetVolts, anaAmps;
  uint  opStatus, alarmStatus, progStatus;
  char  temp[80];


// sample responses to ":STT?;" command:
//   AV00.000SV00.000AA00.000SA20.081OS00010001AL00101PS01100  (20-20)
//   AV10.26SV05.30AA1.999SA2.012OS11010000AL00100PS00000      (60-7?)

//--- Some basic sanity checks on the response ---
  ptrAV = strstr(resp, "AV");   if (ptrAV != resp)  return -1;
  ptrSV = strstr(ptrAV, "SV");  if (!ptrSV)  return -1;
  ptrAA = strstr(ptrSV, "AA");  if (!ptrAA)  return -2;
  ptrSA = strstr(ptrAA, "SA");  if (!ptrSA)  return -3;
  ptrOS = strstr(ptrSA, "OS");  if (!ptrOS)  return -4;
  ptrAL = strstr(ptrOS, "AL");  if (!ptrAL)  return -5;
  ptrPS = strstr(ptrAL, "PS");  if (!ptrPS)  return -6;

  ptrEnd = resp + strlen(resp);

  if (ptrEnd - ptrPS != 7)  return -7;
  if (ptrPS - ptrAL != 7)  return -8;
  if (ptrAL - ptrOS != 10)  return -9;

  KLOCK();

  anaVolts = Cur.ovenReadV[idx];
  anaSetVolts = Cur.ovenSetV[idx];
  anaAmps = Cur.ovenReadA[idx];

  KUNLOCK();


  actualVolts = atof(ptrAV+2);
  setVolts = atof(ptrSV+2);

  actualAmps = atof(ptrAA+2);
  setAmps = atof(ptrSA+2);

  opStatus = strtoi(ptrOS+2, NULL, 2);
  alarmStatus = strtoi(ptrAL+2, NULL, 2);
  progStatus = strtoi(ptrPS+2, NULL, 2);

//tdebug
  _EnterCritical_();  forceErr = (forceOvenErr == 0x0BAD);  _ExitCritical_();
  if (forceErr)  opStatus ^= OOS_OutputOn;

  resetAlarms = false;  newline = "\r\n";  msg = resp;

  if ((opStatus ^ Next.ovenOpStatus[idx]) & 0xFE)  {
    printf("%s%s%sOven%d OP Status: [%.8s]\r\n",
           newline, msg, newline, idx+1, LongToBinMS1st(opStatus, temp, 8));
    printf("  Constant Voltage: %d\r\n"
           "  Foldback Protection: %d\r\n"
           "  Auto-restart: %d\r\n"
           "  Output ON: %d\r\n"
           "  Foldback SRQ: %d\r\n"
           "  Over voltage SRQ: %d\r\n"
           "  Over temp SRQ: %d\r\n",
           BOOL(opStatus & OOS_ConstVolt),   BOOL(opStatus & OOS_FoldbackProt),
           BOOL(opStatus & OOS_AutoRestart), BOOL(opStatus & OOS_OutputOn),
           BOOL(opStatus & OOS_FoldbackSRQ), BOOL(opStatus & OOS_OverVoltSRQ),
           BOOL(opStatus & OOS_OverTempSRQ)
         );
    resetAlarms = true;  newline = msg = "";
  }

  if ((alarmStatus & 0x1E) AND
      ((alarmStatus ^ Next.ovenAlarmStatus[idx]) & 0x1E))  {
    printf("%s%s%sOven%d Alarm Status: [%.8s]\r\n",
           newline, msg, newline, idx+1, LongToBinMS1st(alarmStatus, temp, 5));
    if (alarmStatus & 0x10)  printf("  *** Overvoltage Protection tripped ***\r\n");
    if (alarmStatus & 0x08)  printf("  *** Overtemp Protection tripped ***\r\n");
    if (alarmStatus & 0x04)  printf("  *** Input Voltage Supply Failure ***\r\n");
    if (alarmStatus & 0x02)  printf("  *** Foldback Protection tripped ***\r\n");
    if (alarmStatus & 0x01)  printf("  *** Cmd Error ***\r\n");
    resetAlarms = true;  newline = msg = "";
    if (alarmStatus ^ Next.ovenAlarmStatus[idx])
      printf("  Analog:  Volts: %.16s  SetV: %.16s  Amps: %.16s\r\n"
             "  Serial:  Volts: %.16s  SetV: %.16s  Amps: %.16s\r\n",
             Engineering(anaVolts, 4, temp),
             Engineering(anaSetVolts, 4, temp+12),
             Engineering(anaAmps, 4, temp+24),
             Engineering(actualVolts, 4, temp+36),
             Engineering(setVolts, 4, temp+48),
             Engineering(actualAmps, 4, temp+60)
            );
  }

  if (progStatus & (progStatus != Next.ovenProgStatus[idx])) {
    printf("%s%s%sOven%d Serial Status: [%.8s]\r\n",
           newline, msg, newline, idx+1, LongToBinMS1st(progStatus, temp, 5));
    if (progStatus & 0x08)  printf("  Unrecognized command\r\n");
    if (progStatus & 0x04)  printf("  Buffer Overflow\r\n");
    if (progStatus & 0x02)  printf("  Req Voltage out of range\r\n");
    if (progStatus & 0x01)  printf("  Req Current out of range\r\n");
    resetAlarms = true;  newline = msg = "";
  }

  if (resetAlarms)  { write(devfd, ":DCL;", 5);  sleepMS(30);  }


  KLOCK();

  Next.serOvenReadV[idx] = actualVolts;
  Next.serOvenSetV[idx] = setVolts;
  Next.serOvenReadA[idx] = actualAmps;
  Next.serOvenSetA[idx] = setAmps;
  Next.ovenOpStatus[idx] = opStatus;
  Next.ovenAlarmStatus[idx] = alarmStatus;
  Next.ovenProgStatus[idx] = progStatus;

  KUNLOCK();

  return 0;
}

//-----------------------------------------------------------------------------
//  Process to manage comunications with the Lambda ZUP Power Supply that is
//  inside the big metal box
//-----------------------------------------------------------------------------
void _task ovenCom(int argc, char **argv)
{
  int  i, stat, fflags, ch;
  uint  idx;
  fd_t  devfd;
  bool  gotValidResp, shouldBeOn, isOn;
  U16  *numMsgs, *goodMsgs, *badMsgs;
  ulong  baudRate, lastRcvTime;
  char  *nextCmd;
  BufInfo  bufInfo;
  char  buf[128], readBuf[128];


  redirectStdio();

  if (argc < 2)  Exit(-1);

  idx = (uint)argv[1][0] - '1';  if (idx > 1)  Exit(-2);

  ovenLastRespTime[idx] = 0;  gotValidResp = false;
  if (!idx)  {
    numMsgs = &oven1NumMsgs;
    goodMsgs = &oven1GoodMsgs;
    badMsgs = &oven1BadMsgs;
  }
  else  {
    numMsgs = &oven2NumMsgs;
    goodMsgs = &oven2GoodMsgs;
    badMsgs = &oven2BadMsgs;
  }

  *numMsgs = *goodMsgs = *badMsgs = 0;


//--- (re)open and initialize port for communicating with the RF Gen PS ---
//--- idx == 0: oven1, /dev/tty2;  idx == 1: oven2, /dev/tty1 ---
  sprintf(buf, "/dev/tty%d", idx ? 1 : 2);
  if ((devfd = open(buf, O_RDWR, 0)) < 0)  {
    printf("Unable to open %s for Oven%d PS\r\n", buf, idx+1);  Exit(-3); }

  fflags = fcntl(devfd, F_GETFL, 0);
  fcntl(devfd, F_SETFL, fflags | O_NONBLOCK);

  baudRate = 9600;  ioctl(devfd, IOC_SET_BAUD, &baudRate);

  *buf = 0;

  bufInfo.buf = readBuf;  bufInfo.bufSize = sizeof(readBuf);
  bufInfo.fd = devfd;  bufInfo.nextOut = bufInfo.count = 0;


  for (;;)  {
    KLOCK();
    if (!*buf)
      *goodMsgs = *badMsgs = postOvenResetMsgs[idx] = 0;
    else  {
      ++*numMsgs;
      if (gotValidResp)  {
        ++*goodMsgs;  ++postOvenResetMsgs[idx];  *badMsgs = 0;  }
      else  {
        *goodMsgs = postOvenResetMsgs[idx] = 0;  ++*badMsgs;  }
    }
    KUNLOCK();

  //--- read any existing msg in the buffer ---
    for (i=0;;)  {
      if ((stat = readBufCh(&bufInfo)) < 0)  break;
      ch = stat & 0x7f;  if (ch == 0x0D)  break;
      if (ch > ' ')  ++i;
    }
    if (i) continue;

  //--- specify which supply we want to talk to ---
    write(devfd, ":ADR01;", 7);  sleepMS(50);
  //--- help insure Remote mode is not disabled by accident ---
    write(devfd, ":RMT2;", 6);  sleepMS(50);

  //--- set state of output ---
    shouldBeOn = BOOL(Cur.enableOvenOutput[idx]);
    isOn = BOOL(Cur.ovenOpStatus[idx] & OOS_OutputOn);
    if (shouldBeOn != isOn)  {
      write(devfd, (Cur.enableOvenOutput[idx] ? ":OUT1;" : ":OUT0;"), 6);
      sleepMS(50);
    }

  //--- request the complete status ---
    nextCmd = (*ovenCmd[idx] ? ovenCmd[idx] : ":STT?;");
    fcntl(devfd, F_SETFL, fflags);
    write(devfd, nextCmd, strlen(nextCmd));  sleepMS(100);
    fcntl(devfd, F_SETFL, fflags | O_NONBLOCK);
    *ovenCmd[idx] = 0;  gotValidResp = false;

  //--- wait for response ---
    lastRcvTime = GetMS();  *buf = 0;
    for (i=0; i<sizeof(buf)-2; )  {
      stat = readBufCh(&bufInfo);
      if (stat > 0)  {
//        printf("<%02x>", buf[i]);  //tdebug
        buf[i] = (uint)stat & 0x7f;  lastRcvTime  = GetMS();
        if (buf[i] == 0x0D)  break;
        ++i;  continue;
      }
      if ((stat < 0) AND (errno != EAGAIN))  {
        perror("error reading resp from Oven supply");  break; }
      if (msSince(lastRcvTime) > 1000)  break;
      sleepMS(50);
    }
    if (!i)  continue;

    if (buf[i] != 0x0D)  {
      buf[i+1] = 0;
      if (*badMsgs > 2)  printf("mangled response: [%.128s]\r\n", buf);
      continue;
    }

    buf[i] = 0;

  //--- process the response ---
    if (ShowPackets)  printf("[%.128s]\r\n", buf);

    if (nextCmd == ovenCmd[idx])  {
      KLOCK();  strncpy(ovenResp[idx], buf, sizeof(ovenResp[0])-1);  KUNLOCK();
      gotValidResp = true;
    }
    else
      if (Process_Oven_PS_Resp(buf, devfd, idx))  {
        if (*badMsgs > 2) printf("Invalid resp: [%.128s]\r\n", buf);  }
      else  {
        _EnterCritical_();
        ovenLastRespTime[idx] = GetMS();  newOvenResp[idx] = true;
        _ExitCritical_();
        gotValidResp = true;
      }

    sleepMS(200);
  }

}



//=============================================================================
_nearcall float readGetRaw(uint ui)  {
  return (ui < NumHdwADC ? anaIn[ui] : 0.0);  }

_nearcall float readMinRaw(uint ui)  {
  return (ui < NumHdwADC) ? (adcMinRaw(ui) - adcOffset[ui]) * adcGain[ui] : 0.0; }

_nearcall float readMaxRaw(uint ui)  {
  return (ui < NumHdwADC) ? (adcMaxRaw(ui) - adcOffset[ui]) * adcGain[ui] : 9.8; }

_nearcall void readDefaults(uint ui, float *gain, float *offset)  {
  *gain = 1.0;  *offset = 0.0; }

_nearcall void readCalibPoints(uint ui, float *low, float *high)  {
  *low = -9.5;  *high = 9.5; }

_nearcall char *readUnits(uint ui, char *buf)  { return "V"; }


//=============================================================================
_nearcall float setGetRaw(uint ui)  {
  return (ui < NumHdwADC ? anaOut[ui] : 0.0);  }

_nearcall float setMinRaw(uint ui)  {
  return (ui < NumHdwDAC) ? (dacMinRaw(ui) - dacOffset[ui]) * dacGain[ui] : 0.0; }

_nearcall float setMaxRaw(uint ui)  {
  return (ui < NumHdwDAC) ? (dacMaxRaw(ui) - dacOffset[ui]) * dacGain[ui] : 9.8; }

_nearcall void setDefaults(uint ui, float *gain, float *offset)  {
  *gain = 1.0;  *offset = 0.0;  }

_nearcall void setCalibPoints(uint ui, float *low, float *high)  {
  *low = -9.5;  *high = 9.5;  }

_nearcall char *setUnits(uint ui, char *buf)  { return "V"; }


//=============================================================================
//  Set default labels for all the I/O signals
//=============================================================================
void setDefaultLabels(void)
{
  uint  u;


  printf("Setting default labels for I/O signals\r\n");

//--- set application-appropriate default labels ---
//APPLCODE

//--- set default labels for the I/O signals ---
//--- Fixed size arrays are used so the values can be changed at run time ---
//--- (i.e. if pointers where used instead, and initialized to const string ---
//--- values, then we couldn't change them at run time) ---
/*
  fstrcpy(ainLabel[0], "ai0-Descr");
  fstrcpy(ainLabel[1], "ai1-Descr");

  fstrcpy(aoutLabel[0], "ao0-Descr");
  fstrcpy(aoutLabel[1], "ao1-Descr");

  fstrcpy(dinLabel[0], "di0-Descr");
  fstrcpy(dinLabel[1], "di1-Descr");

  fstrcpy(doutLabel[0], "do0-Descr");
  fstrcpy(doutLabel[1], "do1-Descr");
*/

  for (u=0; u<NUM_APPL_DIN; ++u)
    farsprintf((char far *)dinLabel[u], "unused DI-%02u", u);
  for (u=8; (u<NUM_APPL_DIN) AND (u<16); ++u)
    farsprintf((char far *)dinLabel[u], "ID_SW_%02u", u-8);
  for (u=0; u<NUM_APPL_DOUT; ++u)
    farsprintf((char far *)doutLabel[u], "unused DO-%02u", u);

  fstrcpy(dinLabel[16], "Sputter Interlock OK");
  fstrcpy(dinLabel[17], "Not HotShot RF PS");
  fstrcpy(dinLabel[18], "Not NovaStar RF PS");
//  fstrcpy(dinLabel[19], "RF PS Fault OK");
//  fstrcpy(dinLabel[20], "RF PS Heat OFF");
//  fstrcpy(dinLabel[21], "RF PS NOT Ready");


  fstrcpy(doutLabel[0], "Activity LED");

  fstrcpy(doutLabel[4], "Oven2 PS Power OFF");
  fstrcpy(doutLabel[5], "Oven2 Output Disabled");
//  fstrcpy(doutLabel[6], "Enable Ext RS-485 Rx");
//  fstrcpy(doutLabel[7], "Enable Ext RS-485 Tx");
  fstrcpy(doutLabel[8], "Oven1 Output Disabled");
//  fstrcpy(doutLabel[9], "RF PS Reset Pulse");
//  fstrcpy(doutLabel[10], "RF PS Stop Pulse");
  fstrcpy(doutLabel[11], "Extern Watchdog Timer");
  fstrcpy(doutLabel[12], "Oven1 PS Power OFF");
  fstrcpy(doutLabel[13], "Bias PS Power OFF");
  fstrcpy(doutLabel[14], "Sputter PS Power OFF");
//  fstrcpy(doutLabel[15], "RF Gen PS Start Pulse");


  fstrcpy(ainLabel[0], "Oven1 Read V");
  fstrcpy(ainLabel[1], "Bias Read V");
  fstrcpy(ainLabel[2], "Sputter Read V");
  fstrcpy(ainLabel[3], "Sputter Read A");
  fstrcpy(ainLabel[4], "Bias Read A");
  fstrcpy(ainLabel[5], "Oven1 Read A");
  fstrcpy(ainLabel[6], "Oven2 Read V");
  fstrcpy(ainLabel[7], "Oven2 Read A");

  fstrcpy(aoutLabel[0], "Oven1 Set V");
  fstrcpy(aoutLabel[1], "Bias Set V");
  fstrcpy(aoutLabel[2], "Sputter Set V");
//  fstrcpy(aoutLabel[3], "RF Gen Set V");
  fstrcpy(aoutLabel[3], "Oven2 Set V");
}




//=============================================================================
//  This function is called to force software configurable values to default
//  values.
//=============================================================================
void setDefaultConfig(void)
{
  printf("Setting in-RAM device config to default values\r\n");

//===== set default values for persistent data values =====
//APPLCODE

  sptrMaxErrV = 50.0;  // max diff between Sputter setpoint and readback
  sptrMaxErrMS = 5000;  // max time read/set err can exceed sptrMaxErrV

  biasMaxErrV = 50.0;  // max diff between Bias setpoint and readback
  biasMaxErrMS = 5000;  // max time read/set err can exceed biasMaxErrV

  oven1MaxErrV = 2.0;  // max diff between oven setpoint and readback
  oven1MaxErrMS = 3000;  // max time read/set err can exceed ovenMaxErrV
  oven1PwrOffDelay = 3000;  // # ms from off cmd to Oven PS AC power cutoff

  oven2MaxErrV = 2.0;  // max diff between oven setpoint and readback
  oven2MaxErrMS = 3000;  // max time read/set err can exceed ovenMaxErrV
  oven2PwrOffDelay = 3000;  // # ms from off cmd to Oven PS AC power cutoff
}


//=============================================================================
//  This function is called by the application loop just before a shutdown.
//  The appliation must do whatever is necessary for a clean shutdown and
//  then exit
//=============================================================================
void applicShutdown(void)
{
  printf("Application prepping for shutdown\r\n");  fflush(stdout);

//APPLCODE


  Exit(0);
}


//=============================================================================
//  This is the highest priority task, usually resumed at a regular interval
//  so that it can do what it needs and sleep again until the start of the
//  next interval.
//=============================================================================
void _task applic(int argc, char **argv)
{
  uint  idx;

  #define SharedApplLocals
  #include <sharedAppl.c>

//--- add local variable definitions here ---
//APPLCODE


  #define SharedApplLocals2
  #include <sharedAppl.c>



  #define SharedApplInit
  #include <sharedAppl.c>

  _EnterCritical_();  SchedMS = 20;  _ExitCritical_();

//--- At this point generic initialization has been done which includes ---
//--- the following: ---
//      - The Prev, Cur, and Next ApplState objects have been zeroed
//      -

//--- add application initialization code here ---
//APPLCODE
  allowSoftReset = true;

  //BZERO(rfgenCmd);  BZERO(rfgenResp);

//  rfgenLastRespTime = rfgenReadErrStartTime =
  sptrReadErrStartTime = biasReadErrStartTime = 0;

  BZERO((void*)ovenCmd);  BZERO((void *)ovenResp);

  for (idx=0; idx<2; ++idx)  {
    ovenRespAge[idx] = 0;  newOvenResp[idx] = false;
    postOvenResetMsgs[idx] = 0;
    ovenLastRespTime[idx] = ovenReadErrStartTime[idx] =
    ovenOffCmdTime[idx] = ovenPowerOnTime[idx] = 0;
  }

  wdPulseChanged = GetMS();


  #define SharedApplInit2
  #include <sharedAppl.c>

//--- start processes to communicate with supplies with a serial connection ---

//  argv_[0] = "rfgenCom";  argv_[1] = NULL;
//  createProc(rfgenCom, 4, 1, argv_, 0, 1, 2);

//--- task for serial com to oven1 (uses /dev/tty2) ---
  argv_[0] = "ovenCom";  argv_[1] = "1";  argv_[2] = NULL;
  createProc(ovenCom, 4, 2, argv_, 0, 1, 2);

//--- task for serial com to oven2 (uses /dev/tty1) ---
  argv_[0] = "ovenCom";  argv_[1] = "2";  argv_[2] = NULL;
  createProc(ovenCom, 4, 2, argv_, 0, 1, 2);


//--- The following loop is run at a periodic interval (normally 100 times ---
//--- per second, or every 10ms).  The application process normally ---
//--- consists of reading the current state of the digital and analog ---
//--- inputs, changing the state of the analog and digital outputs to the ---
//--- values computed during the previous pass, then computing new values ---
//--- for the outputs and updating values that outside clients can view ---
//--- (currently accessable as a set of ModBus registers) ---

  for (;;)  {
    #define SharedApplLoopStart
    #include <sharedAppl.c>


  //--- read the current state of the analog and digital Input signals, and ---
  //--- write the values computed at the end of the last pass ---
    digitalOut = Next.digitalOut;  rwIO();

  //--- process next event message (if any) ---
    processMsg(eventsfd, (EventMsg *)buf, sizeof(buf));

  //--- update the current state and determine the next state ---
    updateCurState();  updateNextState();
    updateOutputState();  updateExternals();


    #define SharedApplLoopEnd
    #include <sharedAppl.c>
  }

}


//=============================================================================
//  list of application-specific commands that are normally entered via a
//  user shell
//=============================================================================
//void _task nsclSeriald(int argc, char **argv);

//--- add any application-specific commands here ---
//APPLCODE
const CmdInfo far applCmds[] = {
//  { "misc",        miscTest,       "..." },
  { "cmd",       cmd,         "Application-specific commands" },
  { "ovenCom",   ovenCom,     "Lambda/Oven PS serial link" },
//  { "rfgenCom",  rfgenCom,    "RF Gen PS serial link" },
  { "state",     showState,   "Show state of controller" }
};

int const NumApplCmds = ( sizeof(applCmds) / sizeof(CmdInfo) );

//-----------------------------------------------------------------------------

#endif
