//----- LLRF2_Appl.c ----- 01/08/13 -----(03/05/06)----

//-----------------------------------------------------------------------------
//  Application code to measure and control phase of RF frequency signals.
//  Incorporates the following hardware:
//
//  - Single Board Computer (RCM3200: 44 MHz 8-bit CPU, 768K RAM, 512K flash)
//  - 2 Serial EEPROMs MicroChip SST 25VF016B-50 (2048K bytes)
//  - Xilinx Spartan-3A-DSP 3400 FPGA
//  - Texas Instruments TMS320C6748 DSP
//-----------------------------------------------------------------------------
//  Update log (most recent first):
//
//
//  2013-01-08 - Mark Davis
//               - updateCurState() now reads a waveform OR does the periodic
//                 I/O in a pass, but not both.
//
//  2012-11-16 - Mark Davis
//               - Cleaned up state display a bit, and added the data for and
//                 calls to the functions to read/write reg valus from/to the
//                 FPGA.
//
//  2012-11-15 - Mark Davis
//               - Brought up to date with changes to earlier version (the
//                 PhaseCtlrAppl.c) so it handles waveform data properly.
//
//  2012-11-14 - Mark Davis
//               - Commented out outdated readwave() function/task.  Waveform
//                 data is now read using DirectCA which uses direct CA TCP
//                 connections to the ca_tcpd task.
//               - Added typical and noise values to the read-only DSP register
//                 definitions to provide more reasonable values when testing
//                 on other hardware platforms.
//
//  2012-09-14 - Mark Davis
//               - Replaced getLock() and releaseLock() with the newer one from
//                 the PhaseCtlr appl that uses shared/global semaphores.
//                 Prior to this there were still instances where the locks
//                 would not be correctly released.
//
//  2012-01-16 - Mark Davis
//               - Changed argument for releaseLock() to uint.  This is so it
//                 is consistent with getLock() and to avoid any possible
//                 compiler errors (and excess code) converting between uchar
//                 and uint.
//               - Changed all getLock calls so a return value  of 0 is not
//                 treated as meaning that it got the lock.  This shouldn't be
//                 possible, but better safe than sorry.
//
//  2011-09-02 - Mark Davis
//               - Created using the PhaseCtlr code as a starting point.  This
//                 module incorporates the functionality of both the older
//                 (separate) LLRF and PhaseCtlr modules.
//                 In addition to a newer more capable FPGA, this version uses
//                 an 8-bit parallel read/write interface to communicate with
//                 the FPGA (the previous version used a serial SPI connection).
//
//-----------------------------------------------------------------------------

/*
===================
=== Terminology ===
===================

   Feedback Control Loop types:
      ADRC: Active Disturbance Rejection Control
      PID:  Proportional, Integral, Differential


============================
=== Operational Overview ===
============================

   The LLRF2 controller replaces the Digital Phase Controller and LLRF Turn On
   modules.  It contains a more powerful FPGA, DSP, and 2 EEPROMs.

   The Rabbit Core Module (RCM) and this application code is responsible for
   ongoing communication with the FPGA and DSP and for maintaining the config
   "files" stored in the 2 EEPROMS.  This is done using 2 SPI interfaces and
   one parallel interface.

   The connections to the SPI interfaces are as follows:

      SPI1 (the RCM's Serial Port B):
         Master:  RCM
         Slave:   DSP

      SPI2 (the RCM's Serial port C):
         Masters:  RCM, FPGA, DSP
         Slaves:   the EEPROMS

    16 of the RCMs I/O signals are used as a parallel I/O bus to communicate
    with the FPGA (8 for data, 6 for an address, 2 for R/W controls).


    The EEPROMS contain the configuration files for the FPGA and the DSP
    (one FPGA, one for the DSP).  The RCM provides the means to copy a file
    from a remote server to either of the EEPROMs and restart the FPGA and
    the DSP so they use the new configuration.

    When not updating one of the EEPROMs, the RCM's primary task is to
    continually update its copy of a set of values provided by the DSP and
    the FPGA, and to send them any changes to writable values made by
    external entities (e.g. the control system or a person typing a shell
    command).


==========================
=== Summary of signals ===
==========================

   ADDR0-ADDR11 (INPUTS):
      Used to read the state of the 12 DIP switches that determine the
      controller's ID.  The combination of the controller type and its ID must
      be unique for all controllers.


   /DSP_RESET (OUTPUT):
      A low level holds the DSP in RESET and stops it from using the SPI
      interfaces.  A low-to-high transition causes to reload its configuration
      from its EEPROM.

   PROG_B (OUTPUT):
      A low level holds the FPGA in RESET and stops it from using the SPI
      interfaces.  A low-to-high transition causes to reload its configuration
      from its EEPROM.


   Rab_SPI1_DSP_EN (OUTPUT):
      Effectively a Chip Select for the DSP's connection to the SPI1 bus.
      This is asserted to enable the RCM to communicate with the DSP via the
      SPI1 interface.


   Rab_SPI2.DSPen (Hi-Z state or OUTPUT driven low):
      The Chip Select for the EEPROM connected to the DSP.  A low level enables
      the EEPROM.  This can be asserted by the RCM or the DSP (but not both at
      once).  The RCM must hold this in a high impedance state when it is not
      using the SPI1 interface.  Before driving it low, it must drive both
      /DSP_RESET and PROG_B low to insure neither the DSP nor the FPGA is
      trying to use the SPI2 interface.

   Rab_SPI2.FPGAen (Hi-Z state or OUTPUT driven low):
      The Chip Select for the EEPROM connected to the FPGA.  A low level enables
      the EEPROM.  This can be asserted by the RCM or the FPGA (but not both at
      once).  The RCM must hold this in a high impedance state when it is not
      using the SPI1 interface.  Before driving it low, it must drive both
      /DSP_RESET and PROG_B low to insure neither the DSP nor the FPGA is
      trying to use the SPI2 interface.

   FPGA_DONE (INPUT):
      A low-level state indicates the FPGA is (re)loading its configuration
      from its EEPROM.  A high-level state indicates it is busy doing so and
      the Rabbit should wait until it is finished before attempting to
      communicate with the FPGA or its EEPROM.


   Rab_SPI1_CLK (Hi-Z state or OUTPUT):
      Clock for SPI1 interface.  When using the SPI1 interface, the RCM drives
      this high/low.  The rest of the time it holds it in a Hi-Z state.

   Rab_SPI1_MOSI (OUTPUT):
      Used to write data to the DSP (and possibly the FPGA in the future?).

      NOTE:  The RCM cannot set this to a Hi-Z state, so if something else
      needs to write to the SPI1 bus, something external to the Rabbit will
      have to control when the Rabbit is connected to this signal.

   Rab_SPI1_MISO (INPUT):
      Used to read data from the DSP (and possibly the FPGA in the future?)


   Rab_SPI2.CLK (Hi-Z or OUTPUT):
      Clock for SPI2 interface.  When using the SPI1 interface, the RCM drives
      this high/low.  The rest of the time it holds it in a Hi-Z state.

   Rab_SPI2.MOSI (OUTPUT):
      Used to write data to one of the EEPROMS.  Which EEPROM the data is
      written to is controlled by driving either the Rab_SPI2.FPGAen or the
      Rab_SPI2.DSPen signal low.

      NOTE:  The RCM cannot set this to a Hi-Z state, so something external to
      the RCM must disconnect the RCM when it is not using the SPI2 bus.  The
      easiest way to do this would be to connect the RCM to this signal ONLY
      when it is driving both the /DSP_RESET and PROG_B outputs low.

   Rab_SPI2.MISO (INPUT):
      Used to read from one of the EEPROMS.  Which EEPROM the data is read from
      is controlled by driving either the Rab_SPI2.FPGAen or the Rab_SPI2.DSPen
      signal low.


   CPU_RDY (OUTPUT):
      A high-level state means the Rabbit is currently running and is ready to
      communicate with the FPGA and DSP.

   ACTY (OUTPUT):
      Connected to an LED, the Rabbit pulses this to cause the LED to blink at
      various rates.  Used as a simple means of indicating the state of the
      controller to someone within view of the controller.


   FPGA.A0-FPGA.A5 (OUTPUTS):
      FPGA Address:  A 6-bit address that specifies which value the Rabbit
      wishes to read from or write to the FPGA.

   FPGA.D0-FPGA.D7 (I/O):
      FPGA Data:  An 8-bit bidirectional data bus for reading from and writing
      to the FPGA.  The Rabbit will leave these signals in a high impedance
      state except when it sets FPGA.WE high.  The FPGA will leave them in a
      high impedance state except when the Rabbit sets FPGA.OE high.

   FPGA.OE (OUTPUT):
      FPGA Output Enable:  Used to read an 8-bit value from the FPGA.  Normally
      in a low-level state, a value is read from the FPGA as follows:

        - Assert the address of the value to read on FPGA.A0-FPGA.A5
        - Set FPGA.OE high and wait at least 106ns
        - Read the value on FPGA.D0-FPGA.D7
        - Set FPFGA.OE low

   FPGA.WE (OUTPUT):
      FPGA Write Enable:  Used to write an 8-bit value to the FPGA.  Normally
      in a low-level state, a value is written to the FPGA as follows:

        - Assert the address of the value to read on FPGA.A0-FPGA.A5
        - Assert the value to be written on FPGA.D0-FPGA.D7
        - Pulse FPGA.WE high for at least 27ns

   FPGA.CLK (OUTPUT):
      FPGA Clock:  Not currently used.

      NOTE: This is connected to a Rabbit signal that CANNOT be configured as
      a clock signal for any of the Rabbit's serial ports.


=======================================
=== Outline of RCM responsibilities ===
=======================================

  Unlike most of the RCM-based controllers, the RCM code for this controller
  is not directly responsible for much in terms of actual real-time control
  of the device.  The primary role of the RCM is to manage the configuration
  data for and communication with the FPGA and DSP, which handle all the
  high-speed computational tasks needed to make the device work in real time.

  What the RCM IS responsible for:

   - With the noted exception of real-time control of the device(s) the
     controller is connected to, the RCM is responsible same things it is
     in ALL RCM-based controllers:
       - Communication with external entities such as the Control System
       - Managing persistent configuration values
       - Monitoring the state of the device
       - Run-time diagnostics

   - On startup and on demand:  Reset the FPGA and DSP and cause them to
     reload their configurations from their respective EEPROMs.

   - Copy the contents of a DSP or FPGA configuration file (usually located on
     a remote server) to one of the EEPROMS.

   - Continually R/W values from/to the FPGA and the DSP.

     NOTES:

       Communication with the FPGA is via an "ad-hoc" parallel bus (i.e. it
       uses general I/O signals for the data and address signals rather than
       the Rabbit's built-in I/O capability), so it is relatively slow.

       Comnunication with the DSP is the same as the earlier PhaseCtlr module,
       using the SPI1 bus.

       For values that do not trigger some action within the DSP or FPGA, the
       current value is normally read and/or written to the DSP or FPGA each
       time through the main application task's loop.  This ensures that the
       most recent value is always known/applied.

  ----------------------------------------------------------------------------

  This is the first controller where I will try adopting a new strategy in
  terms of the separation of the hardware and application logic layers.

  Prior to this, the separation was only as deep as the signal level.  The
  hardware layer managed reading the state of the available analaog and
  digital inputs and storing them in some global variables, and updating
  the state of the available analog and digital outputs using values stored
  in some global variables.

  This approach meant that the application layer did not have to worry about
  how the digital and analog signals where implemented by any particular
  hardware.  It just assumed that the signals existed and that, on request,
  any changes would be propogated in both directions by the hardware layer.


  While not as efficient as directly manipulating the hardware-specific
  registers and chips needed to control or read the signal states, it meant
  the application code could be built for and run on any platform that
  supported a very basic set of I/O signals without any changes to the code.

  But that model is built on the assumption that the application logic is
  tightly coupled to the state of the analog and digital inputs and outputs.

  In the broadest sense all the controller applications can be described as
  a continuous loop that reads the state of a set of inputs, computes what
  the next desired state should be, then updating the state of the outputs
  with the goal of bringing about the desired changes.

  Note that this holds true regardless of whether the inputs and outputs
  represent actual digital and analog inputs and outputs or simply values
  in memory that are accessible to other tasks (and, indirectly, logic
  external to the controllers).


  For most controllers, the current and next state of these hardware signals
  is the focus of the application logic, so it makes logical sense to place
  the hardware abstraction at that level.  Also, most applications only
  update the states of these signals once each pass, so the additional
  overhead of the abstraction layer is not significant.


  But the focus of the LLRF2 application task is communuication with the
  DSP, FPGA, and their EEPROMs.  It does not care what signals are needed to
  implement communication, or how they are used.  It only cares about when it
  NEEDS to communicate and when it CAN.

  Also, the communication requires manipulating the hardware signals dozens or
  even hundreds of time each pass, so the overhead of abstracting them can
  add up very quickly.


  Last but not least is the issue of testing the application logic.

  For many applications basic testing can be accomplished by manually
  manipulating the state of the inputs and monitoring the state of the outputs.
  But even in the cases where this is useful, it is not practical to simulate
  many of the behaviors of the real hardware.  In the case of the LLRF2
  application, there is no practical way to manually simulate much of anything
  the actual hardware does, so very little testing can be done on any other
  platform.

  A more practical approach in this case that may also be more practical for
  other applications is to move the level of abstraction closer to the
  application logic.  Rather than simply providing an abstraction of the analog
  and digital signals, the hardware layer API will include application
  specific functions that previously were implemented in the application layer.

  This has the following advantages over the simpler signal-abstraction
  approach:

    - More efficient:  They can manipulate the signals more directly so
      functions that have to manipulate the signals many times each call
      will be significantly faster.

    - Each platform the application is built for can provide its own version
      that operates the most efficiently for that platform.  For platforms
      that simply don't have the equivalent hardware, the function could
      provide a simulation of the hardware, providing a more realistic test
      on other platorms.

      NOTES:
         The ioctl() model could be used to make it possible to maintain the
         one advantage of the current model:  Build and run any application
         on any platform.  Of course, this approach simply amounts to run-time
         linking, and any platform that does not implement the (indirectly)
         specified function will simply return an error, so it is of dubious
         value.

         A better approach would be to provide a generic version of the
         functions that could be compiled for any platform and which would
         provide at least a minimal simulation of the hardware it was meant
         for.  This would also avoid the overhead of figuring out which
         function is being indirectly called each time.  And given that the
         function ID would have to be unique for ALL applications, that could
         turn in to a significant amount of overhead.

  ------------------------------------------------------------------------

  Questions/Issues:

    - Eliminate the functionality for writing to the DSP memory?  I assume
      this is superceeded by the DSP having its own dedicated EEPROM.

    - Keep in mind that some communication with the FPGA and/or DSP will
      require checking the value of a polled register value.  It is probably
      not practical to implent that level within the hardware abstraction
      layer (although such functions will still call hardware layer functions
      to actually read/write individual values or blocks of values, including
      the register being polled).

    - Should I still map at least some of the signals to the digitalIn and
      digitalOut values so I can monitor them while testing other functions?
      I COULD make the digitalOut read-only.

      I could have a separate function that returns the state of a bunch
      of the signals, but that would just require duplicating much of same
      functionality that already works for the digitalIn and digitalOut
      values.

    - Eliminate everything to do with the PLL?

    - DSP program number?  Is this concept still valid?  How is it handled
      now that the Rabbit writes directly to the EEPROM?  Just assume a
      fixed size and divide the EEPROM space in to equal size parts?

    - dspComErrors?


------------------------------------------------------------------------------
(from older DigPhaseCtlr:  needs to be revised for this version...)

  - The EEPROM contains the code for the FPGA.  On startup, one of the digital
    inputs to the FPGA (FPGA_SPI) is toggled, which causes it to take over the
    SPI connection to the EEPROM and load the binary image stored there.  A
    digital output from the FPGA is monitored to indicate when this has
    completed (it takes a little less than 2 seconds).

  - The code for the DSP is stored in the DSP itself.  Similar to how the
    DEBROS update command can be used to update the DEBROS boot image, the DSP
    code provides a mechanism for updating the copies of the DSP code stored
    within the DSP.  When the DSP is running a working copy of the DSP code,
    the progdsp command can be used to download a new copy of DSP code in to
    the DSP itself.

    So updating or storing the initial working copy of DSP code requires that
    a working copy of DSP code is already running in the DSP.  If a working
    copied is stored at the correct place within the DSP memory, then the DSP
    will load it and run it after being reset.  But if there isn't a working
    copy in the DSP memory, then getting it to run a working copy can only be
    accomplished by running the appropriate development software on a desktop
    PC and use of a special cable between the PC and the controller (again,
    very similar to how the DLM is downloaded to a Rabbit Core Module).


Communicating with the DSP:

  - Exchange of data with the DSP is done using a bi-directional (simultaneous
    send/recv) clocked SPI interface.  All data is exchanged in 32-bit (4-byte)
    packets.  While we are sending a 32-bit packet to the DSP, it is using the
    same clock pulses to simultaneously send a 32-bit packet in response.

    Each exchange of a 32-bit packet consists of setting the DSP /Enable signal
    low, writing (while simultaneously reading) 32-bits of data, and then
    setting the DSP /Enable signal high.

    Due to the way the DSP buffers are set up, all responses from the DSP are
    always 2 packets "behind" the ones that contained the commands that they
    are a response to.  We can write as many commands as we want in a
    sequence of packets, but the response to each one will always be received
    2 packets later.  Given that we generally want to read the response to
    a set of commands as soon as possible, we normally have to add some
    dummy or filler packets to the sequence to get the desired response
    packets.  For extended responses (like reading of waveform data), we must
    send lots of filler packets to get all the data.


Revamped register conventions (Jan 2010):

  - Communication with the DSP is done by reading and writing "register"
    values, each of which has an "address".  Even numbered addresses are for
    reading and odd numbered ones are for writing (i.e. the least significant
    bit specifies a read or write operation).

  - Reading a register value is done by sending the 4-byte address of the
    register and then reading (2 packets later) its 4 byte value.

  - Writing to a register is done by sending the 4-byte address and then the
    4-byte value.  The DSP responds (with the usual 2-packet delay) with the
    4-byte address and 4-byte value as a means of verifying that we are
    communicating with the DSP and that it received the correct values.

    The meaning of the 4-byte value is dependent on the register.  The most
    common representations used are 32-bit floats, 32 and 16-bit integers,
    and bit-mapped values.  The values are always in network byte order (most
    significant first).


  - Registers come in several varieties:

      Read-Only-and-Continuously:
         The value should be read each pass to keep track of the current state.

      Write-Only-and-Continuously:
         These represent settings that we want asserted all the time.  If the
         DSP resets, these values will be quickly re-asserted because they are
         written to the DSP each pass.

      Write-Once, Read-never:
         Writing to the value triggers some action so writing to them each
         pass is not an option.  These registers generally provide the means
         to write to one instance in a table of values maintained within the
         DSP itself, where the row written to is based on the current state
         of the DSP.

      Read/Write-continuously:
         Such registers have both a separate read and write register and often
         corresponding separate read and write ModBus values.  The usual case
         is that the write value represents a setpoint and read value
         represents the current state or readback value.

      Read-Trigger:
         Reading from the register triggers some event such as starting or
         stopping a calculation, so it is done only when we want to trigger
         the corresponding process or event.


  - Most the DSP registers have a corresponding ModBus register.  Tables are
    used to provide the information that links them together.  About 20 times
    per second the main application task uses these tables to read and write
    values from/to the DSP.

    The table for the read-only registers causes the ModBus register values
    that correspond to each DSP register to be updated.  The table for the
    writable registers is used to write the current value of each of the
    corresponding ModBus registers to the DSP.  This means that, if the DSP
    resets for any reason, the current state will be restored very quickly.

  - The dspread and dspwrite commands provide a mechanism to manually read
    values from and write values to the DSP.  Just be careful about using
    registers that trigger some event within the DSP as this can interfere
    with normal operation.

  - To facilitate testing, the normal periodic reading from and writing to the
    DSP registers can be disabled or enabled using the dspio command.


  - The register values provide information about the state of the DSP.
    One of the registers is a bit-mapped status value, with bits that
    indicate if new waveform data is available for reading.

    For each waveform we check to see if we have an open connection to a
    client for the waveform and if there is new data or a fixed interval
    of time has elapsed.  If these conditions are met, then we read the
    corresponding waveform data as follows:

      - Using the write/read 4-bytes methodology described above,
        read the # of 32-bit words of data available for the waveform.

      - Using the value 0x0010##00 for the 4 bytes we write each
        time (where ## indicates which waveform we want data for),
        use the write/read 4-bytes methodology to read the
        indicated # of 32-bit words of data.

    A couple of the waveforms take around 40 seconds to compute and the
    DSP must be told when to compute them.  For these an additional
    register provides the %complete the computation is so we can monitor
    the progress of the computation and determine when it is available
    for reading.


 - Some of the bit-mapped registers we can write to represent multiple values
   each represented by more than one bit.  When writing to these registers,
   all 0's for a given group means do not change the value of any of the
   bits in that group.

   To insure that the DSP state is properly restored if it is reset, the last
   value of each bit-group within such registers must be kept track of.  This
   minimizes problems caused by DSP resets.


 - The SPI interface used to communicate with the DSP is also used to
   communicate with the FPGA, the PLL, and the ADC's.  To allow multiple tasks
   to communicate with with these devices, a locking mechanism is provided to
   insure that only one task is using the shared SPI connection and related
   data values at any given time.


 - The DSP xxxCalib value registers (0x0010022A - 0x001000239) are used
   to write a given value in to a table of values based on the current state
   of another value (attenuations?), so they cannot be included in the set
   that is constantly being written to.  Writing to one of these registers
   is done just once when you need to write a value to the table stored in
   the DSP.  Saving the table is done by reading from the register at the
   end of that range (SaveCalibToDSP: 0x0010023A).

------------------------------------------------------------------------------*/


//-----------------------------------------------------------------------------
//  Work in progress:
//
//  - Need to deal with multiple sources of R/W requests for the DSP.  The
//    situation is not as simple as representing the DSP as a std streaming
//    device that reads() and writes() can be interleaved on.  The following
//    constraints exist:
//
//    - SPI port 1 is connected to the FPGA, PLL, DSP, and ADC's
//
//    - Different control bits and clock modes are required for different
//      devices
//
//    - Reads and Writes to the DSP are not independant.  In at least some
//      cases, you have to know how many bytes to read from the DSP at the
//      time you begin writing to it.
//
//    - Reads and Writes to the DSP always consist of an even number of bytes
//
//    - Reads and Writes to the DSP require a control bit be toggled between
//      each pair of bytes transferred
//
//
//  Events/processes that require use of the SPI connection:
//
//  - Writing commands to the ADC's
//
//  - Writing commands to the FPGA
//
//  - Writing commands to and reading data from the DSP (where some of the
//    responses can be up to 1024 bytes in length)
//
//  - Downloading a new config file to the FPGA (or the DSP?)
//
//  - A write to a Modbus register that mirrors a value in the DSP
//
//
//  Two options and issues regarding sharing this single resource that many
//  tasks need to use or affect:
//
//  - Extension of current approach:  Use a semaphore to limit access to a
//    single task at a time.
//
//  - Limit all access to a single task and send msgs to the task requesting
//    certain actions
//
//  The semaphore approach has one major drawback:  Tasks that cannot block
//  (like the main applic process) must either give up on an action or keep the
//  data needed to process the request and retry until successful or an error
//  occurs.  It has the advantage that it means less code for tasks that CAN
//  block without a problem.
//
//  The message approach has the advantage that a task can send a message about
//  the request and pretty much forget about it.  The disadvantages are that
//  their is no easy way to provide feedback to the requesting task, and that
//  all the work has to be done in one task.
//
//
//  Basic problem scenario:
//
//    A command is issued to download a new image to the FPGA or the DSP.
//    This can tie up the SPI for several seconds.  While this is going on,
//    several Modbus writes to registers mirroring DSP values happen.  These
//    writes cause event messages to be written to the main application's
//    event msg queue.  The application loop, which is still running while
//    the download to the FPGA is going on, cannot use the SPI line, so it
//    cannot read values from the DSP to update its state data, much less
//    process the messages to write new values to the DSP.  In order to
//    not loose the write events, it has to stop processing the event
//    messages, which means the queue could overflow, causing processes
//    that write to the queue to block.
//
//  I don't see any way around this.  The SPI line serves several purposes
//  and it is either in use or it is not.  If it is in use, a task that
//  wishes to use it can either block or not.  For processes that cannot
//  (or should not) block, they can either ignore/forget what they were
//  going to do with the SPI connection or keep track of what they were
//  about to do and keep polling periodically, waiting for their chance
//  to use the SPI connection.
//
//  This is really nothing but the basic limitation of any resource that
//  must be protected against concurrent usage.  A locking mechanism is
//  required that insures only one user at a time, and tasks contending for
//  the resource have 3 choices:
//
//  - block on the resource (and get woken up when it is available)
//
//  - poll in a loop, waiting until they get access, possibly doing other
//    useful stuff that is not dependent on the the event that cannot be
//    completed until the locked resource is available
//
//  - forget what they needed it for and continue on
//
//
//  So, when the main application process needs the SPI line, it will either
//  have to block on a wait queue, or simply suspend processing of event
//  messages.  The advantage of the blocking approach is that the order of
//  requests for the SPI line will be enforced by the wait queue.  On the
//  other hand, I might want to grant access based on priority as well, where
//  higher priority tasks get unblocked before lower priority ones.  But as
//  a general rule, ignoring the sequence of requests can lead to other
//  problems.
//
//-----------------------------------------------------------------------------
//
//  More notes on SPI port usage:
//
//  The model I am using for the SPI port is not right.  It should be treated
//  more like a disk drive or a network connection.
//
//  When a serial port is used as a point-to-point connection between 2 devices
//  it often makes sense to allow the application code to directly manage the
//  connection.  But when there are multiple devices connected via the same
//  communications medium, then the connection is a bus, and the connection
//  needs to be handled by a driver that provides a layer of abstraction for
//  communicating with the devices attached to the bus.
//
//  In this case, something more like a device special file per device attached
//  to the SPI port makes more sense, where the major device # is for the SPI
//  port and the minor specifies which device on the "bus".  Then individual
//  architecture modules can deal with them however I want (for devices that
//  a particular board does not have, it can either return errors for attempts
//  to access the device, or I could have a test mode where they mimic the
//  device to some degree, or simply return no errors).
//
//-----------------------------------------------------------------------------
// Keeping Modbus register values (global variables) in sync with the DSP
// settings vs reading waveform data after a settings
//
// - ampl and phase values in DSP and ModBus register (global variable) values
//   must be kept in sync
//
// - when we want to read the waveform data from the DSP to see the effect
//   of changing the ampl and/or phase value, we have to make the settings
//   and read the waveform data as a single transaction
//
// - one possibility is to read the waveform data EVERY time the ampl or
//   phase setpoints are altered.  Then there is no distinction between "just
//   setting" one of them vs setting-and-reading
//   BUT:  The problem with that is it doesn't allow for changing BOTH the
//   the phase and amplitude and reading the result.
//
//-----------------------------------------------------------------------------
//
//  "Pulsed" tuning test mode:
//
//    The DSP normally controls the Amplitude and Phase values using a standard
//    PID feedback-loop algorithm.  To facilitate the tuning of the PID weights
//    used by the feedback loop, the DSP and controller provide a "pulsed"
//    tuning/test mode where the DSP periodically cycles the Amplitude and/or
//    Phase setpoints between two settings and captures a portion of the
//    resulting outputs in a buffer that the controller's CPU can read and
//    send to a remote client for the purpose of providing a near real-time
//    plot of the resulting waveforms.
//
//
//  Basic operation:
//
//    The DSP contains all the logic and code for actually implementing the
//    pulsed test functionality.  It requires the following values to be set
//    before this mode is enabled:
//
//    - Nominal Amplitude and Phase setpoints:  These are the "normal"
//      setpoints, used when the pulsed mode is not active.
//
//    - Pulse Amplitude and Phase setpoints:  When the pulsed test mode is
//      activated, the DSP toggles the setpoints between the Nominal values
//      and Pulsed setpoints.
//
//    - Pulse Frequency:  Controls the # of times per second that the
//      Amplitude and/or Phase setpoints are toggled between the 2 setpoints
//
//    - Trigger position:  Specifies the # of samples that should be taken
//      before changing the setpoint(s).  This provides a simple means of
//      specifying a trigger delay, or a sliding window around the trigger
//      event.
//
//    The pulsed test mode is activated by setting one or both of the run bits
//    in the DSP's cmd word.  When activated, the DSP samples the outputs and
//    writes the values to one of two buffers.  One buffer is for the sample
//    window around the rising edge, and one is for the window around the
//    falling edge.  When the DSP has filled one of the buffers, it sets a bit
//    in the status word indicating the buffer contains new data to be read.
//    The DSP must then leave the contents of the buffer untouched until it
//    has been read by the controller CPU (see next paragraph).
//
//    The controller CPU periodically reads the status word from the DSP.
//    Whenever it sees that the new-data bit for one of the buffers is set,
//    it reads the data for that buffer from the DSP and then sets a bit in
//    the command word indicating it has read the contents of the buffer.
//    To avoid having the controller CPU attempt to read the same data from
//    the same buffer the next cycle, the DSP must respond by immeadiately
//    resetting the new-data bit in the status word.
//
//    Once the new-data bit has been reset, the DSP is free to write new
//    data to the buffer, repeating the above cycle.
//
//-----------------------------------------------------------------------------



#include <common.h>

#include <unistd.h>
#include <sys/types.h>

#include <linux/wait.h>

#include <semaphore.h>



#ifndef  ApplicSource
  #define  ApplicSource  "LLRF2_Appl.c"
#endif

#ifndef DidRegDefs
  #define  DidRegDefs
  #include <modbusRegs.c>
#endif
#ifndef RegDefs


#include <LLRF2_Appl.h>



//APPLCODE - change to the application name
const char far applName[] = { "Low Level RF Controller" };
const char ApplAbbrev[] = { "LLRF2" };

//APPLCODE - add ID to unistd.h and place it here
#define DeviceType  DT_LLRF2

//APPLCODE - set the # of input and outputs used
#define  NUM_APPL_AIN     1   // analog inputs
#define  NUM_APPL_AOUT    1   // analog outputs
#define  NUM_APPL_DIN     1   // digital inputs
#define  NUM_APPL_DOUT    2   // digital outputs

#define  ADC_MASK_APPL  0x00000000L
#define  DAC_MASK_APPL  0x00000000L

//APPLCODE - set to state we want outputs in during initialization
#define InitialDigOutState  ( 0 )


#define SharedINC
#include <sharedAppl.c>

#include <math.h>


#include <LLRF2.h>



//--- definitions for enabling diagnostic messages ---
#define  DSPErrorBits      0x80

#define  ShowDSPErrors     (DiagFlags & DSPErrorBits)


/*-------------------------------------------------
 NOTES:
   port B ("spi1")  connected to DSP
   port C ("spi2")  connected to the EEPROMs
---------------------------------------------------*/

//--- multipliers/divisors for converting between unsigned 16-bit int values ---
//--- used by the DSP and engineering units (volts and degress in this case) ---
#define  AmplConvFactor   ( 0.0000426233445431 )
//#define  PhaseConvFactor  ( 0.0054931640625 )  // 360 / 2^16
#define  PhaseConvFactor  ( 0.0000000838190317 )  // ~ 360 / 2^32
#define  FRACU16Divisor   ( 0.0000152590219 )  // div by this = mult by 65535.0
//#define  FreqConvFactor   (    0.0152590219 )  // scales 0-1000 -> 0-65535


#define  DSP_Prog_File   "/srfs/dspprog.bin"
#define  FPGA_Prog_File  "/srfs/llrf.bin"




#define  ReadDig()  { digitalIn = ReadDigInputs(); }
#define  WriteDig()  {  WriteDigOutputs(digitalOut); }



#define  PI_  3.1415926

//-----------------------------------------------------------------------------
// data types and function prototypes:
//-----------------------------------------------------------------------------

//long fltofp(float value);
//float fptofl(long value);



bool MB_DSP_REG(SetRegFuncParams);




//=============================================================================
//  definitions related to use of the Digital and Analog I/O signals
//=============================================================================

//======================= digitalIn usage ===========================
#define  FPGA_DONE    0x000001  // DI-00: FPGA loaded and running

const uchar  LogicOrderDIn[NUM_APPL_DIN] = { 0 };

//--- msgs to log automatically when a digital input changes state ---
//--- use NULL for bits that we don't want to log changes for ---
const char far *digInDescr[NUM_APPL_DIN*2] = {
//--0 state descr--    --1 state descr--
  NULL,                NULL
};



//====================== digitalOut usage ===========================

//------------------------------
#define  READY_LED    0x000001   // DO-00: CPU is running application
#define  ActivityLED  0x000002   // DO-01: Activity LED

const uchar  LogicOrderDOut[NUM_APPL_DOUT] = { 0, 1 };


#define  TurnOnActivityLED()    ClearBits(Next.digitalOut, ActivityLED)
#define  TurnOffActivityLED()   SetBits(Next.digitalOut, ActivityLED)

//--- msgs to log automatically when a digital output changes state ---
//--- use NULL for bits that we don't want to log changes for ---
const char far *digOutDescr[NUM_APPL_DOUT*2] = {
//--0 state descr--    --1 state descr--
  NULL,                NULL,
  NULL,                NULL
};


//=============== Analog input usage ===============
const uchar  LogicOrderAIn[NUM_APPL_AIN] = { 0 };


//=============== Analog output usage ===============
const uchar  LogicOrderAOut[NUM_APPL_AOUT] = { 0 };



//=============================================================================
//  ID's for values sent to or read from the DSP
//  OR'ing an ID with 0x8000 allows reading without writing
//=============================================================================

//--- DSP register numbers (without the 0x01000000 mask shared by all regs) ---


// Even # values used to read a register, Odd # values used to write to a reg

// P32 means the "PHASE" format, a 32-bit integer scaled to/from 0 to 360
// B32 means 32-bit uint bit-mapped value

//===== Read-Only registers =====
#define  FwdPhase           0x00100000L  // P32: Forward Phase
#define  RevPhase           0x00100002L  // P32: Reverse Phase
#define  CavPhase           0x00100004L  // P32: Cavity Phase
#define  Ref0Phase          0x00100006L  // P32: Reference 0 Phase

#define  Spare0Phase        0x00100008L  // P32: Spare RF 0 Phase
#define  Spare1Phase        0x0010000AL  // P32: Spare RF 1 Phase
#define  Spare2Phase        0x0010000CL  // P32: Spare RF 2 Phase
#define  Ref1Phase          0x0010000EL  // P32: Reference 1 Phase

#define  FwdAmpl            0x00100010L  // F32: Forward Amplitude
#define  RevAmpl            0x00100012L  // F32: Reverse Amplitude
#define  CavAmpl            0x00100014L  // F32: Cavity Amplitude
#define  Ref0Ampl           0x00100016L  // F32: Reference Amplitude

#define  Spare0Ampl         0x00100018L  // F32: Spare RF 0 Amplitude
#define  Spare1Ampl         0x0010001AL  // F32: Spare RF 1 Amplitude
#define  Spare2Ampl         0x0010001CL  // F32: Spare RF 2 Amplitude
#define  Ref1Ampl           0x0010001EL  // F32: Reference 1 Amplitude

#define  MaxFwdAmpl         0x00100020L  // F32: Max Forward Amplitude
#define  MaxRevAmpl         0x00100022L  // F32: Max Reverse Amplitude
#define  MaxCavAmpl         0x00100024L  // F32: Max Cavity Amplitude

#define  MaxSpare0Ampl      0x00100026L  // F32: Max Spare RF 0 Amplitude
#define  MaxSpare1Ampl      0x00100028L  // F32: Max Spare RF 1 Amplitude
#define  MaxSpare2Ampl      0x0010002AL  // F32: Max Spare RF 2 Amplitude
#define  MaxOutAmpl         0x0010002CL  // F32: Max Output Amplitude

#define  StatusBitsDSP      0x0010002EL  // B32: misc DSP status

#define  DSPLoad            0x00100030L  // F32: DSP load percentage
#define  PCBTemp            0x00100032L  // F32: PCB temperature
#define  Fan0Spd            0x00100034L  // U32: Fan0 Speed (in rpm)
#define  Fan1Spd            0x00100036L  // U32: Fan1 Speed (in rpm)


//===== R/W registers (the addresses shown are the read addresses) =====
#define  TrigPos            0x00100100L  // U32: trigger position (tuning plot)

#define  WarmLevel          0x00100102L  // F32: Warm-up Amplitude
#define  WarmTime           0x00100104L  // F32: time at warm-up Amplitude
#define  RampTime           0x00100106L  // F32: time to ramp to set-point

#define  FwdAtten           0x00100108L  // U32: Forward Attenuator setting
#define  RevAtten           0x0010010AL  // U32: Reverse Attenuator setting
#define  CavAtten           0x0010010CL  // U32: Cavity Attenuator setting

#define  Spare0Atten        0x0010010EL  // U32: Spare RF 0 Attenuator setting
#define  Spare1Atten        0x00100110L  // U32: Spare RF 1 Attenuator setting
#define  Spare2Atten        0x00100112L  // U32: Spare RF 2 Attenuator setting
#define  OutAtten           0x00100114L  // U32: Output Attenuator setting

#define  PhaseSet           0x00100116L  // P32: Phase setpoint
#define  AmplSet            0x00100118L  // F32: Ammplitude setpoint

#define  PulseFreq          0x0010011AL  // F32: Tuning Pulse Freq
#define  PulseDuty          0x0010011CL  // F32: Freq Search Amplitude setpoint

#define  RF_P               0x0010011EL  // F32: RF P value
#define  RF_I               0x00100120L  // F32: RF I value
#define  RF_D               0x00100122L  // F32: RF D value

#define  Tuner_P            0x00100124L  // F32: Tuner P value
#define  Tuner_I            0x00100126L  // F32: Tuner I value
#define  Tuner_D            0x00100128L  // F32: Tuner D value

#define  RF_B0              0x0010012AL  // F32: RF ADRC b0 value
#define  RF_WCtrl           0x0010012CL  // F32: RF ADCR control bandwidth
#define  RF_WObs            0x0010012EL  // F32: RF ADRC observer bandwidth

#define  Tuner_B0           0x00100130L  // F32: Tuner ADRC b0 value
#define  Tuner_WCtrl        0x00100132L  // F32: Tuner ADCR control bandwidth
#define  Tuner_WObs         0x00100134L  // F32: Tuner ADRC observer bandwidth

#define  PulseAmplSet       0x00100136L  // F32: Pulsed Amplitude setpoint
#define  PulsePhaseSet      0x00100138L  // P32: Pulsed Phase setpoint

#define  CavLowIlkLevel     0x0010013AL  // F32: Cavity Low Interlock Level
#define  CavLowIlkTime      0x0010013CL  // F32: Cavity Log Interlock Time

#define  DvDtIlkRate        0x0010013EL  // F32: -dV/dT Interlock Rate

#define  RevHighIlkRatio    0x00100140L  // F32: V+/V- Interlock Ratio
#define  RevHighIlkTime     0x00100142L  // F32: V+/V- Interlock Time

#define  FwdLowIlkRatio     0x00100144L  // F32: V+/V- Interlock Ratio
#define  FwdLowIlkTime      0x00100146L  // F32: V+/V- Interlock Time

#define  FwdHighIlkLevel    0x00100148L  // F32: V+/V- Interlock Level
#define  FwdHighIlkTime     0x0010014AL  // F32: V+/V- Interlock Time

#define  FwdCavPhase        0x0010014CL  // P32: Fwd-Cav Phase Calibration

#define  FFPhase            0x0010014EL  // P32: Feed Forward Phase
#define  FFGain             0x00100150L  // F32: Feed Fwd Gain

#define  FreqOffset         0x00100152L  // F32:  RF Frequency Offset


//===== RFQ Tuner values, some of which are forwarded by the control =====
//===== system to other devices/records =====
#define  RFQ_ManTuneSet     0x00100198L  // (RW) F32: manual tuner setpoint

#define  RFQ_T1_Offset      0x0010019AL  // (RW) F32: tuner 1 calib offset
#define  RFQ_T2_Offset      0x0010019CL  // (RW) F32: tuner 2 calib offset

#define  RFQ_T1_Output      0x0010019EL  // (RO) F32: tuner 1 output
#define  RFQ_T2_Output      0x001001A0L  // (RO) F32: tuner 2 output

#define  RFQ_T1_Feedback    0x001001A2L  // (RW) F32: tuner 1 feedback
#define  RFQ_T2_Feedback    0x001001A4L  // (RW) F32: tuner 2 feedback


//===== Special Registers =====
#define  SetDSPState        0x00100200L  // (WO) Bits: set/reset DSP state bits

#define  PiezoTunerPos      0x00100202L  // (RW) U32: piezo tuner position
#define  StepperPos         0x00100204L  // (RW) U32: stepper position

//#define  SetAFCsOnOff       0x00100206L  // (WO) Bits: turn AFC frequencies on/off

#define  Unused_208         0x00100208L  //

#define  Rd_L_Buf           0x0010020AL  // (RN) S16x2: read next tuning plot value (leading edge)
#define  Rd_T_Buf           0x0010020CL  // (RN) S16x2: read next tuning plot value (trailing edge)

#define  Reserved_20E       0x0010020EL  //
#define  Reserved_210       0x00100210L  //

//#define  WrMemBlock         0x00100212L  // (WO) U32: start prog mode for a memory block
//#define  WriteStatus        0x00100214L  // (RO) S32: (% done or -1 ==> busy)

//#define  CalcCTF            0x00100216L  // (**) N/A: begin calc of Cavity Transfer Func
//#define  PercentCTF         0x00100218L  // (RO) U32: percent done for CTF calculation
//#define  AbortCTF           0x0010021AL  // (**) N/A: abort CTF calc; == -1 if too late
//#define  SizeCTF            0x0010021CL  // (RO) U32: # of 32-bit words int plot data
//#define  PhaseCTF           0x0010021EL  // (RO) P32: read next Phase value from CTF plot
//#define  AmplCTF            0x00100220L  // (RO) F32: read next Ampl value from CTF plot

#define  PercentFFT         0x00100222L  // (RO) U32: percent done for FFT calculation
#define  SizeFFT            0x00100224L  // (RO) U32: # of 32-bit words of in plot data

#define  AmplFFT            0x00100226L  // (RN) F32: read next Ampl value from FFT plot

//#define  CTFCalcRate        0x00100228L  // (RW) F32: CTF calc rate

#define  FwdMagCalib        0x0010022AL  // (RW) F32: Forward Mag calib value
#define  FwdPhaseCalib      0x0010022CL  // (RW) P32: Forward Phase calib value

#define  RevMagCalib        0x0010022EL  // (RW) F32: Reverse Mag calib value
#define  RevPhaseCalib      0x00100230L  // (RW) P32: Reverse Phase calib value

#define  CavMagCalib        0x00100232L  // (RW) F32: Cavity Mag calib value
#define  CavPhaseCalib      0x00100234L  // (RW) P32: Cavity Phase calib value

#define  Spare0MagCalib     0x00100236L  // (RW) F32: Spare RF 0 Mag calib value
#define  Spare0PhaseCalib   0x00100238L  // (RW) P32: Spare RF 0 Phase calib value

#define  Spare1MagCalib     0x0010023AL  // (RW) F32: Spare RF 1 Mag calib value
#define  Spare1PhaseCalib   0x0010023CL  // (RW) P32: Spare RF 1 Phase calib value

#define  Spare2MagCalib     0x0010023EL  // (RW) F32: Spare RF 2 Mag calib value
#define  Spare2PhaseCalib   0x00100240L  // (RW) P32: Spare RF 2 Phase calib value

#define  OutMagCalib0       0x00100242L  // (RW) F32: Output Mag calib value DAC0
#define  OutPhaseCalib0     0x00100244L  // (RW) P32: Output Phase calib value DAC0

#define  OutMagCalib1       0x00100246L  // (RW) F32: Output Mag calib value DAC1
#define  OutPhaseCalib1     0x00100248L  // (RW) P32: Output Phase calib value DAC1

#define  SaveCalibToDSP     0x0010024AL  // (**) N/A: Save calib to DSP's ROM

#define  ROMProgErrs        0x0010024CL  // (RO) U32: ROM programming errors


//  (**) means a READ from the register triggers the stated functionality


//=============================================================================
//  List of daemon processes.  The inetd process listens for connection
//  requests on each port in the list and creates a new instance of the
//  specified process to service each new connection.
//=============================================================================
#include <services.h>

//--- service funcs declared the same as standard command funcs:
//     e.g.:   void _task fastdatad(int argc, char **argv);
void _task tsh(int argc, char **argv);

//void _task readWave(int argc, char **argv);


ServiceInfo inetdConf[] = {
//--- add any application-specific inetd-type network services here ---
//--- (add port # names to services.h) ---
//  NAME          PORT           PROTOCOL  FUNC        PRI -1  Config Params
//  ------------  -------------  --------  ----------- --- --  --------------
//  { "readWave",   READ_WAVE,     ProtoTCP, readWave,   12, -1, NULL },
/*
  NOTE:  The readWave approach is outdated as it is not a practical way
  way to support reading multiple waveforms concurrently (as is now done
  for EPICS, which has one waveform record for each of the waveforms).

  The new method uses the DEBROS EPICS CA (Channel Access) service (the
  ca_tcpd task, which handles the TCP virtual circuits, subscriptions, etc).

  The UDP portion of the CA spec is supported if needed (handled by the ca_udpd
  task), but because of the limited buffering that is practical on the current
  hardware (i.e it drops lots of the CA UDP broadcast packets), this service
  is not normally run.

  The EPICS waveform records use the DirectCA device support module which
  already knows the controller's IP address and opens a connection as needed
  to the ca_tcpd task and uses the standard CA protocol messages to create
  subscriptions for the waveforms and related values.
*/

//--- standard ones defined in sharedAppl.c ---
//{ "console",    TTY0,          ASYNC,    tsh,        10, -1, &consCfg },
//{ "fastdatad",  FASTDATA_PORT, ProtoTCP, fastdatad,   5, -1, NULL },
//{ "modbusTCPd", MODBUS_PORT,   ProtoTCP, modbusTCPd,  6, -1, NULL },
//{ "tsh",        TELNET_PORT,   ProtoTCP, tsh,        10, -1, NULL }

  #define SharedServices
  #include <sharedAppl.c>
};

const int  NumServices = sizeof(inetdConf) / sizeof(ServiceInfo);


//=============================================================================
//  Application-specific data structures and definitions
//=============================================================================

// values that can be different between prev/cur or cur/next during processing


// ***** WARNING ***** WARNING ***** WARNING ***** WARNING *****
//  Changing the applState structure means that a soft reset will fail!  A
//  change to the size will cause a soft restart to abort and do a CPU
//  restart instead.  And while a change that does NOT result in a new
//  size will not cause the soft restart to abort, it does mean that any
//  shift in the position of values can cause all kinds of mayhem.
// ***** WARNING ***** WARNING ***** WARNING ***** WARNING *****


typedef struct {
//--- values updated in "cur" instance at start of each pass ---
  ulong  digitalIn;       // state of digital inputs

//--- values updated in "next" instance at start of each pass ---
  ulong  digitalOut;      // state for digital outputs
} ApplState;


ApplState  prevState, curState, nextState;

#define  Prev  prevState
#define  Cur   curState
#define  Next  nextState

const uint ApplStateDataSize =  ( sizeof(ApplState)*3 + sizeof(setVal) );


//--- values and flags set by the 1st stage of the control loop when ---
//--- interpreting the state of the inputs and processing the next ---
//--- event msg ---

//ulong  calcStartTimeCTF;



//=============================================================================
//  Defintions for variables that are accessable as ModBus registers
//=============================================================================
#else  // end of before-RegDefs portion

/*
  The purpose of this section is to define a set of C variables that are
  accessable as ModBus registers.  This approach avoids the need to use
  special functions or macros throughout the code when accessing register
  values, which keeps the application code fast and simple.

  The model used provides support for multiple logical devices.  The 2nd
  parameter in each register definition macro indicates if one instanace
  applies to all logical devices (specified by an "x" for "don't care"), or
  if each logical device requires its own instance (specified by an under-
  score "_" to signify "fill in the blank").  In the 1st case, the variable
  is defined as a scalar.  In the 2nd case, it is defined as an array.

  The definition macros support data types (floats and longs in particular)
  that require more than one register to store them.  To avoid the possibility
  of corrupted values, all reads of and writes to such values must be done
  using the multi-register commands and including all the registers that
  make up the value in the command.  Differences in register order between
  different client types is handled by a table that automatically swaps
  register values as needed based on the client type.

  The definition for each writable register also includes the function to
  call when the value is set via the ModBus interface.  Many registers can
  use the default setXXX() functions, but you can specify your own to provide
  any special logic needed (including ignoring the new setting if invalid).

  For registers, the format for the function definitions is:

    bool mySetFunc(SetRegFuncParams);

       where SetRegFuncParams is currently defined as:
              void *info, uchar chanNum, void *newVal, fd_t eventsfd

  For coils, the format for the function definitions is:

     bool (*SetCoilFunc)(SetCoilFuncParams);

         where SetCoilFuncParams is currently define as:
               void *info, uchar chanNum, uchar newVal, fd_t eventsfd

  The parameters to the call are as follows:
    *info:     A pointer to the Coil's/Register's CoilInfo/RegInfo structure
    chanNum:   The logical device # (0 is the 1st)
    *newVal:   A pointer to the new value requested in the Modbus packet
               (changed to native-byte order where needed)
    eventsfd:  The fd for the FIFO used to send msgs to the main application
               process

  The function returns true if the new value was accepted, false if not.

  The register numbers for multiple "channels" or logical devices are
  handled automatically based on the xxChanMult values.  These values specify
  the offset between a register value for one logical device and the "same"
  register for the next.

  Example:

    Assume the following definitions:
      #define  NumLogicalDevices  2
      #define  R4ChanMult  100
      REG4( motorPos,       _, 40050, F32, RW, setMotorPos )

    Then we would get:
      global variable definition:  float motorPos[2];
      register pair:  40050,40051  refers to:  motorPos[0]
      register pair:  40150.40151  refers to:  motorPos[1]

    If instead we had specified:
      REG4( motorPos,       x, 40050, F32, RW, setMotorPos )

    Then we would get:
      global variable definition:  float motorPos;
      register pairs:  40050,40051 & 40150,40151  BOTH refer to:  motorPos


  To keep the definitions as simple as possible, and avoid requiring multiple
  macros for each register definition, each macro is used more than once,
  with a different definition (#define) each time (a practice generally
  referred to as X-Macros).  While trying to follow the order of inclusion
  of files (including a recursive reference back to this file) can be rather
  confusing, the X-Macro method keeps the source code simple to read and
  maintain by doing everything with a single macro for each register.

  ==========================================================================
                             !!! WARNING !!!
  ==========================================================================
    Each sequence of register definitions MUST be contiguous!  There must
    NOT be any holes (i.e. no unused register #'s between used ones) and no
    overlap!
-----------------------------------------------------------------------------*/


#ifndef NumLogicalDevices
  #define  NumLogicalDevices  1
#endif

//-----------------------------------------------------------------------------
#ifndef NumC0Chans
  #define NumC0Chans_  NumLogicalDevices
  #define C0ChanMult_  100
#endif

#ifdef COIL0

COIL0( DoNothing,       x, 1, setCoil )
//COIL0( Coil00002,      _, 2, setCoil )
//COIL0( Coil00003,      _, 3, setCoil )

#undef COIL0
#endif


//-----------------------------------------------------------------------------
#ifndef NumR3Chans
  #define NumR3Chans_  NumLogicalDevices
  #define R3ChanMult_  200
#endif

#ifdef REG3

REG3( dlmVersTime,     x, 30001, U32 )
REG3( rtosVersTime,    x, 30003, U32 )
REG3( applVersTime,    x, 30005, U32 )
REG3( devType_,        x, 30007, U16 )
REG3( deviceID_,       x, 30008, U32 )
REG3( upSecs,          x, 30010, U32 )
REG3( percentIdle_,    x, 30012, U16 )

REG3( numChans,        x, 30013, U16 )
REG3( coilChanMult,    x, 30014, U16 )
REG3( reg3ChanMult,    x, 30015, U16 )
REG3( reg4ChanMult,    x, 30016, U16 )

REG3( digInBits,       x, 30017, U32 )

REG3( statusBitsStdA,  x, 30019, U16 )
REG3( statusBitsStdB,  x, 30020, U16 )
REG3( statusBitsAppl,  x, 30021, U32 )

REG3( latency,         x, 30023, U16 )  // actual vs scheduled appl start in ms
REG3( tcpTickTime,     x, 30024, U16 )  // # ms inside tcp_tick() call
REG3( applLoopTime,    x, 30025, U16 )  // # ms used by applic task loop
REG3( kernelLoopTime,  x, 30026, U16 )  // # ms used by last kernel pass


//--- FPGA read-only values ---
REG3( statusBitsFPGA,  x, 30027, U16 )  // bit-mapped FPGA status
REG3( statusBitsTuner, x, 30028, U16 )  // bit-mapped Tunerstatus
REG3( statusBitsRF,    x, 30029, U16 )  // bit-ampped RF status
REG3( statusBitsILK,   x, 30030, U16 )  // bit-mapped Interlocks status

REG3( spark0_ctr,      x, 30031, U16 )  // # sparks detected on Chan 0
REG3( spark1_ctr,      x, 30032, U16 )  // # sparks detected on Chan 1
REG3( cavlow_ctr,      x, 30033, U16 )  // # times Cav_low interlock occured
REG3( rev_high_ctr,    x, 30034, U16 )  // # times Reverse power was too high
REG3( dvdt_ctr,        x, 30035, U16 )  // # rate chg cavity field sparks
REG3( fwd_low_ctr,     x, 30036, U16 )  // # amplifier faults
REG3( fwd_high_ctr,    x, 30037, U16 )  // # cav shut-offs to protect amplifier
REG3( PLC_ctr,         x, 30038, U16 )  // # PLC interlocks of cavity (while on)
REG3( safety_ctr,      x, 30039, U16 )  // # Safety intlks of cavity (while on)


//--- DSP read-only values ---
REG3( fwdPhase,        x, 30040, F32 )  // forward phase
REG3( revPhase,        x, 30042, F32 )  // reverse phase
REG3( cavPhase,        x, 30044, F32 )  // cavity phase
REG3( ref0Phase,       x, 30046, F32 )  // reference 0 phase

REG3( spare0Phase,     x, 30048, F32 )  // spare RF 0 phase
REG3( spare1Phase,     x, 30050, F32 )  // spare RF 1 phase
REG3( spare2Phase,     x, 30052, F32 )  // spare RF 2 phase
REG3( ref1Phase,       x, 30054, F32 )  // reference 1 phase

REG3( fwdAmpl,         x, 30056, F32 )  // forward amplitude
REG3( revAmpl,         x, 30058, F32 )  // reverse amplitude
REG3( cavAmpl,         x, 30060, F32 )  // cavity amplitude
REG3( ref0Ampl,        x, 30062, F32 )  // reference 0 amplitude

REG3( spare0Ampl,      x, 30064, F32 )  // spare RF 0 amplitude
REG3( spare1Ampl,      x, 30066, F32 )  // spare RF 1 amplitude
REG3( spare2Ampl,      x, 30068, F32 )  // spare RF 2 amplitude
REG3( ref1Ampl,        x, 30070, F32 )  // reference 1 amplitude

REG3( maxFwdAmpl,      x, 30072, F32 )  // max forward amplitude
REG3( maxRevAmpl,      x, 30074, F32 )  // max reverse amplitude
REG3( maxCavAmpl,      x, 30076, F32 )  // max cavity amplitude

REG3( maxSpare0Ampl,   x, 30078, F32 )  // max spare RF 0 amplitude
REG3( maxSpare1Ampl,   x, 30080, F32 )  // max spare RF 1 amplitude
REG3( maxSpare2Ampl,   x, 30082, F32 )  // max spare RF 2 amplitude
REG3( maxOutAmpl,      x, 30084, F32 )  // max output amplitude

REG3( statusBitsDSP,   x, 30086, U32 )  // bit-mapped status register

REG3( dspLoad,         x, 30088, F32 )  // DSP load percentage
REG3( pcbTemp,         x, 30090, F32 )  // PCB temperature
REG3( fan0Spd,         x, 30092, U16 )  // fan0 speed (in rpm)
REG3( fan1Spd,         x, 30093, U16 )  // fan1 speed (in rpm)


//--- RFQ Tuner read-only values (not used on any other cavities) ---
REG3( rfq_t1_output,   x, 30094, F32 )  // RFQ tuner 1 output
REG3( rfq_t2_output,   x, 30096, F32 )  // RFQ tuner 2 output


//--- Special DSP regsiter values ---
REG3( piezoReadPos,    x, 30098, S16 )  // current piezo pos read from DSP
REG3( stepperReadPos,  x, 30099, S16 )  // current stepper pos read from DSP

REG3( percentFFT,      x, 30100, U16 )  // percent done computing FFT waveform
REG3( sizeFFT,         x, 30101, U16 )  // # 32-bit words in FFT plot

REG3( romProgErrs,     x, 30102, U16 )  // # programming errors

#undef REG3
#endif


#ifdef RegMacros
//===== application-specific defintions related to register values =====

//  #define  READING(idx)  REG_ARRAY(float, rdg_1, idx)

#endif



//----- values for statusBitsDSP --- (DSP reg StatusBitsDSP 0x0010002E)
#define  DSP_ROM_BUSY        0x00000001  // DSP wirting calib data to EEPROM
#define  DSP_RS232_BUSY      0x00000002  // xmiting setpoint to RFQ module
#define  DSP_RFQ_OLD_SETPT   0x00000004  // RFQ tuner setpoint has not chgd
#define  DSP_RF_ON           0x00000008  // RF is ON

#define  DSP_RF_IPULSING     0x00000010  // RF pulsing using internal src
#define  DSP_RF_EPULSING     0x00000020  // RF pulsing using external src
#define  DSP_NEW_L_DATA      0x00000040  // new data in Rising/Leading Edge sample buf
#define  DSP_NEW_T_DATA      0x00000080  // new data in Falling/Trailing Edge sample buf

#define  DSP_FILL_FFT_BUF    0x00000100  // Filling buffer for FFT
#define  DSP_COMPUTING_FFT   0x00000200  // FFT being computed
#define  DSP_FFT_DATA_VALID  0x00000400  // FFT buffer has valid data
#define  DSP_FNDG_CAV_FREQ   0x00000800  // searching for resonance

#define  DSP_TUNING_CAV      0x00001000  // tuning cavity to ref freq
#define  DSP_RF_OUT_WARM     0x00002000  // RF output at warm level
#define  DSP_OUT_RAMPING     0x00004000  // RF output is ramping
#define  DSP_PHASE_AMPL_CTL  0x00008000  // in Phase/Ampl control mode

#define  DSP_CAV_LOCKED      0x00010000L // Cavity is locked
#define  DSP_MANUAL_CTL      0x00020000L // Manually config ctl mode enabled
#define  DSP_FREQ_OFFSET     0x00040000L // freqOffset enabled
#define  DSP_AMPL_FDBK_SRC   0x00080000L // Ampl Feedback from cav(0) or fwd(1)

#define  DSP_AMPL_CTL_ON     0x00100000L // Ampl Ctl loop enabled
#define  DSP_PHASE_MODE_0    0x00200000L // 00 = Open Loop  01 = Phase Ctl
#define  DSP_PHASE_MODE_1    0x00400000L // 10 = SEL Ctl    11 = invalid
#define  DSP_TUNER_CTL_ON    0x00800000L // Tuner Ctl loop enabled

#define  DSP_UNUSED_24       0x01000000L //
#define  DSP_UNUSED_25       0x02000000L //
#define  DSP_UNUSED_26       0x04000000L //
#define  DSP_UNUSED_27       0x08000000L //

#define  DSP_UNUSED_28       0x10000000L //
#define  DSP_UNUSED_29       0x20000000L //
#define  DSP_UNUSED_30       0x40000000L //
#define  DSP_UNUSED_31       0x80000000L //


#define  DSP_BUSY        (statusBitsDSP & (DSP_ROM_BUSY | DSP_RS232_BUSY)
#define  OLD_RFQ_SETNG   (statusBitsDSP & DSP_RFQ_OLD_SETPT)
#define  RF_ON           (statusBitsDSP & DSP_RF_ON)

#define  RF_IPULSING     (statusBitsDSP & DSP_RF_IPULSING)
#define  RF_EPULSING     (statusBitsDSP & DSP_RF_EPULSING)
#define  NEW_L_DATA      (statusBitsDSP & DSP_NEW_L_DATA)
#define  NEW_T_DATA      (statusBitsDSP & DSP_NEW_T_DATA)

#define  FILL_FFT_BUF    (statusBitsDSP & DSP_FILL_FFT_BUF)
#define  COMPUTING_FFT   (statusBitsDSP & DSP_COMPUTING_FFT)
#define  FFT_DATA_VALID  (statusBitsDSP & DSP_FFT_DATA_VALID)
#define  FNDG_CAV_FREQ   (statusBitsDSP & DSP_FNDG_CAV_FREQ)

#define  TUNING_CAV      (statusBitsDSP & DSP_TUNING_CAV)
#define  RF_OUT_WARM     (statusBitsDSP & DSP_RF_OUT_WARM)
#define  OUT_RAMPING     (statusBitsDSP & DSP_OUT_RAMPING)
#define  PHASE_AMPL_CTL  (statusBitsDSP & DSP_PHASE_AMPL_CTL)

#define  CAV_LOCKED      (statusBitsDSP & DSP_CAV_LOCKED)
#define  MANUAL_CTL      (statusBitsDSP & DSP_MANUAL_CTL)
#define  FREQ_OFFSET     (statusBitsDSP & DSP_FREQ_OFFSET)
#define  AMPL_FDBK_SRC   (statusBitsDSP & DSP_AMPL_FDBK_SRC)

#define  AMPL_CTL_ON     (statusBitsDSP & DSP_AMPL_CTL_ON)
#define  PHASE_MODE      ((statusBitsDSP >> 21) & 0x03)
#define  TUNER_CTL_ON    (statusBitsDSP & DSP_TUNER_CTL_ON)


//-----------------------------------------------------------------------------
#ifndef NumR4Chans
  #define NumR4Chans_  NumLogicalDevices
  #define R4ChanMult_  300
#endif

#ifdef REG4

REG4( timeZone_,       x, 40001, S16, RW, MB_timeZone )
REG4( curTime,         x, 40002, U32, RW, setU32 )

REG4( ipAddr,          x, 40004, U32, RW, setU32 )
REG4( netmask,         x, 40006, U32, RW, setU32 )
REG4( gateway,         x, 40008, U32, RW, setU32 )

REG4( setACoil,        x, 40010, U16, RW, setU16 )
REG4( resetACoil,      x, 40011, U16, RW, setU16 )

REG4( digOutBits,      x, 40012, U32, RW, setU32 )

REG4( cmdBitsStd,      x, 40014, U32, RW, MB_cmdBitsStd )
REG4( cmdBitsAppl,     x, 40016, U32, RW, MB_cmdBitsAppl )


//--- FPGA writable values ---
REG4( ADC_err_rst,     x, 40018, U16, RW, setU16 )  // bit-mapped ADC error resets
REG4( RF_soft_on,      x, 40019, U16, RW, MB_RF_soft_on)  // bit-0: enable RF output
REG4( RF_ilk_rst,      x, 40020, U16, RW, setU16 )  // bit-0: clr RF interlock
REG4( ilk_ctr_rst1,    x, 40021, U16, RW, setU16 )  // counter resets
REG4( ilk_ctr_rst2,    x, 40022, U16, RW, setU16 )  // counter resets


//--- DSP writable values ---
REG4( trigPos,         x, 40023, U32, RW, MB_DSP_REG   )

REG4( warmLevel,       x, 40025, F32, RW, MB_maxAmpl )  // warm-up amplitude
REG4( warmTime,        x, 40027, F32, RW, MB_DSP_REG )  // warm-up time (in sec)
REG4( rampTime,        x, 40029, F32, RW, MB_DSP_REG )  // ramp to setpoint (in sec)

REG4( fwdAtten,        x, 40031, U16, RW, MB_DSP_REG )  // fwd attenuation
REG4( revAtten,        x, 40032, U16, RW, MB_DSP_REG )  // rev attenuation
REG4( cavAtten,        x, 40033, U16, RW, MB_DSP_REG )  // cav attenuation

REG4( spare0Atten,     x, 40034, U16, RW, MB_DSP_REG )  // spare RF 0 atten
REG4( spare1Atten,     x, 40035, U16, RW, MB_DSP_REG )  // spare RF 1 atten
REG4( spare2Atten,     x, 40036, U16, RW, MB_DSP_REG )  // spare RF 2 atten
REG4( outAtten,        x, 40037, U16, RW, MB_DSP_REG )  // output attenuation

REG4( phaseSet,        x, 40038, F32, RW, MB_DSP_REG )  // phase setpoint
REG4( amplSet,         x, 40040, F32, RW, MB_maxAmpl )  // amplitude setpoint

REG4( pulseFreq,       x, 40042, F32, RW, MB_DSP_REG )  // pulse frequency
REG4( pulseDuty,       x, 40044, F32, RW, MB_DSP_REG )  // pulse duty cycle

REG4( rf_P,            x, 40046, F32, RW, MB_DSP_REG )  // RF PID P value
REG4( rf_I,            x, 40048, F32, RW, MB_DSP_REG )  // RF PID I value
REG4( rf_D,            x, 40050, F32, RW, MB_DSP_REG )  // RF PID D value

REG4( tuner_P,         x, 40052, F32, RW, MB_DSP_REG )  // tuner PID P value
REG4( tuner_I,         x, 40054, F32, RW, MB_DSP_REG )  // tuner PID I value
REG4( tuner_D,         x, 40056, F32, RW, MB_DSP_REG )  // tuner PID D value

REG4( rf_B0,           x, 40058, F32, RW, MB_DSP_REG )  // RF ADRC b0 value
REG4( rf_WCtrl,        x, 40060, F32, RW, MB_DSP_REG )  // RF ADRC ctl bandwidth
REG4( rf_WObs,         x, 40062, F32, RW, MB_DSP_REG )  // RF ADRC observ bwidth

REG4( tuner_B0,        x, 40064, F32, RW, MB_DSP_REG )  // tuner ADRC b0 value
REG4( tuner_WCtrl,     x, 40066, F32, RW, MB_DSP_REG )  // tuner ADRC ctl bwidth
REG4( tuner_WObs,      x, 40068, F32, RW, MB_DSP_REG )  // tuner ADRC observ bwth

REG4( pulseAmplSet,    x, 40070, F32, RW, MB_maxAmpl )  // pulsed ampl setpoint
REG4( pulsePhaseSet,   x, 40072, F32, RW, MB_DSP_REG )  // pulsed phase setpoint

REG4( cavLowIlkLevel,  x, 40074, F32, RW, MB_DSP_REG )  // cav low intlk level
REG4( cavLowIlkTime,   x, 40076, F32, RW, MB_DSP_REG )  // cav low intlk time

REG4( dvdtIlkRate,     x, 40078, F32, RW, MB_DSP_REG )  // -dV/dT intlk rate

REG4( revHighIlkRatio, x, 40080, F32, RW, MB_DSP_REG )  // V+/V- intlk ratio
REG4( revHighIlkTime,  x, 40082, F32, RW, MB_DSP_REG )  // V+/V- intlk time

REG4( fwdLowIlkRatio,  x, 40084, F32, RW, MB_DSP_REG )  // fwd low intlk ratio
REG4( fwdLowIlkTime,   x, 40086, F32, RW, MB_DSP_REG )  // fwd low intlk time

REG4( fwdHighIlkLevel, x, 40088, F32, RW, MB_DSP_REG )  // fwd high intlk level
REG4( fwdHighIlkTime,  x, 40090, F32, RW, MB_DSP_REG )  // fwd high intlk time

REG4( fwdCavPhase,     x, 40092, F32, RW, MB_DSP_REG )  // fwd-cav phase calibr

REG4( ffPhase,         x, 40094, F32, RW, MB_DSP_REG )  // feed-fwd phase calibr
REG4( ffGain,          x, 40096, F32, RW, MB_DSP_REG )  // feed-fwd gain calibr

REG4( freqOffset,      x, 40098, F32, RW, MB_DSP_REG )  // RF freq offset

//--- RFQ Tuner values (not used on any other cavities) ---
REG4( rfq_manTuneSet,  x, 40100, F32, RW, MB_DSP_REG )  // RFQ manual tuner setpoint

REG4( rfq_t1_offset,   x, 40102, F32, RW, MB_DSP_REG )  // RFQ tuner 1 calib offset
REG4( rfq_t2_offset,   x, 40104, F32, RW, MB_DSP_REG )  // RFQ tuner 2 calib offset

REG4( rfq_t1_feedback, x, 40106, F32, RW, MB_DSP_REG )  // RFQ tuner 1 feedback
REG4( rfq_t2_feedback, x, 40108, F32, RW, MB_DSP_REG )  // RFQ tuner 2 feedback


//--- Special DSP Register values ---
REG4( setDSPState,       x, 40110, U32, RW, MB_setDSPState )

REG4( piezoSetPos,       x, 40112, S16, RW, MB_DSP_REG )
REG4( stepperSetPos,     x, 40113, S16, RW, MB_DSP_REG )

REG4( fwdMagCalib,       x, 40114, F32, RW, MB_DSP_REG )
REG4( fwdPhaseCalib,     x, 40116, F32, RW, MB_DSP_REG )

REG4( revMagCalib,       x, 40118, F32, RW, MB_DSP_REG )
REG4( revPhaseCalib,     x, 40120, F32, RW, MB_DSP_REG )

REG4( cavMagCalib,       x, 40122, F32, RW, MB_DSP_REG )
REG4( cavPhaseCalib,     x, 40124, F32, RW, MB_DSP_REG )

REG4( spare0MagCalib,    x, 40126, F32, RW, MB_DSP_REG )
REG4( spare0PhaseCalib,  x, 40128, F32, RW, MB_DSP_REG )

REG4( spare1MagCalib,    x, 40130, F32, RW, MB_DSP_REG )
REG4( spare1PhaseCalib,  x, 40132, F32, RW, MB_DSP_REG )

REG4( spare2MagCalib,    x, 40134, F32, RW, MB_DSP_REG )
REG4( spare2PhaseCalib,  x, 40136, F32, RW, MB_DSP_REG )

REG4( outMagCalib0,      x, 40138, F32, RW, MB_DSP_REG )
REG4( outPhaseCalib0,    x, 40140, F32, RW, MB_DSP_REG )

REG4( outMagCalib1,      x, 40142, F32, RW, MB_DSP_REG )
REG4( outPhaseCalib1,    x, 40144, F32, RW, MB_DSP_REG )


//--- Gain and Offset settings for DSP values ---
REG4( fwdPhaseGain,      x, 40146, F32, RW, MB_setGain )
REG4( fwdPhaseOffset,    x, 40148, F32, RW, setF32 )
REG4( fwdAmplGain,       x, 40150, F32, RW, MB_setGain )
REG4( fwdAmplOffset,     x, 40152, F32, RW, setF32 )

REG4( revPhaseGain,      x, 40154, F32, RW, MB_setGain )
REG4( revPhaseOffset,    x, 40156, F32, RW, setF32 )
REG4( revAmplGain,       x, 40158, F32, RW, MB_setGain )
REG4( revAmplOffset,     x, 40160, F32, RW, setF32 )

REG4( cavPhaseGain,      x, 40162, F32, RW, MB_setGain )
REG4( cavPhaseOffset,    x, 40164, F32, RW, setF32 )
REG4( cavAmplGain,       x, 40166, F32, RW, MB_setGain )
REG4( cavAmplOffset,     x, 40168, F32, RW, setF32 )

REG4( ref0PhaseGain,     x, 40170, F32, RW, MB_setGain )
REG4( ref0PhaseOffset,   x, 40172, F32, RW, setF32 )
REG4( ref0AmplGain,      x, 40174, F32, RW, MB_setGain )
REG4( ref0AmplOffset,    x, 40176, F32, RW, setF32 )

REG4( spare0PhaseGain,   x, 40178, F32, RW, MB_setGain )
REG4( spare0PhaseOffset, x, 40180, F32, RW, setF32 )
REG4( spare0AmplGain,    x, 40182, F32, RW, MB_setGain )
REG4( spare0AmplOffset,  x, 40184, F32, RW, setF32 )

REG4( spare1PhaseGain,   x, 40186, F32, RW, MB_setGain )
REG4( spare1PhaseOffset, x, 40188, F32, RW, setF32 )
REG4( spare1AmplGain,    x, 40190, F32, RW, MB_setGain )
REG4( spare1AmplOffset,  x, 40192, F32, RW, setF32 )

REG4( spare2PhaseGain,   x, 40194, F32, RW, MB_setGain )
REG4( spare2PhaseOffset, x, 40196, F32, RW, setF32 )
REG4( spare2AmplGain,    x, 40198, F32, RW, MB_setGain )
REG4( spare2AmplOffset,  x, 40200, F32, RW, setF32 )

REG4( ref1PhaseGain,     x, 40202, F32, RW, MB_setGain )
REG4( ref1PhaseOffset,   x, 40204, F32, RW, setF32 )
REG4( ref1AmplGain,      x, 40206, F32, RW, MB_setGain )
REG4( ref1AmplOffset,    x, 40208, F32, RW, setF32 )

REG4( maxAmplSet,        x, 40210, F32, RW, setF32 )


#undef REG4
#endif


#ifdef RegMacros
//===== application-specific defintions related to register values =====

//  #define  ISET(idx)   REG_ARRAY(float, iSet_1, idx)


#endif

//--------------------------------------------------------------
#undef RegMacros


//-----------------------------------------------------------------------------
#ifdef PV_DECLS

char pvStrTest[] = { "....Initial Value...." };  //test

//-------------------------------
// NOTES:
//   - To allow slow clients to read a copy of waveform data without blocking
//     the application task for a significant amount of time, "double-buffering"
//     is used as follows:
//     - any access to waveformData[#] requires acquiring a global lock
//       (basically a simple semaphore with a wait queue)
//     - the application task always reads in to the even #'d rows
//     - the clients always read from the odd #'d rows
//     - when an even #'d row has newer data, both the even and odd rows
//       are locked, the even #'d row being locked just long enough to copy
//       from one to the other
//     - the client currently holding the lock for an odd #'d row can then
//       read from the odd #'d row without blocking the application task
//       (of course it can block other clients wanting to read the same row,
//       although support for multiple reader locks on the odd #'d rows vs
//       single writer locks could improve concurrency quite a bit).
//-----------------------------
#define  MaxDataPoints  1024  // WARNING: max value determined by DSP code
#define  NUM_WAVEFORMS     3

uint  numDataPoints[NUM_WAVEFORMS*2];
ulong far waveformData[NUM_WAVEFORMS*2][MaxDataPoints];
ulong  dataTimeMS[NUM_WAVEFORMS*2], dataTimeSecs[NUM_WAVEFORMS*2];
uint  lastWaveReadTime[NUM_WAVEFORMS*2];
bool   readerRdy[NUM_WAVEFORMS*2];
uint  lastReadPlotTime;

#define  WF_IDX(ptr)  \
  ((uint)(((ulong far *)(ptr) - waveformData[0]) / MaxDataPoints))


//--- indexes in to waveformData[] ---
#define  WAVE_AP_L_SRC    0   // Ampl-Phase Leading-edge Source
#define  WAVE_AP_L_CPY    1   // Ampl-Phase Leading-edge Copy
#define  WAVE_AP_T_SRC    2   // Ampl-Phase Trailing-edge Source
#define  WAVE_AP_T_CPY    3   // Ampl-Phase Trailing-edge Copy
#define  WAVE_FFT_SRC     4   // FFT Source
#define  WAVE_FFT_CPY     5   // FFT Copy
//#define  WAVE_CTF_P_SRC   6   // Cavity Transfer Func Phase Source
//#define  WAVE_CTF_P_CPY   7   // Cavity Transfer Func Phase Copy
//#define  WAVE_CTF_A_SRC   8   // Cavity Transfer Func Ampl Source
//#define  WAVE_CTF_A_CPY   9   // Cavity Transfer Func Ampl Copy


//ulong  startCTFWaitTime;

#undef PV_DECLS
#endif



#ifdef PV_DEFS

PV_STRING(pvStrTest, PV_RD, PV_UpdateVal, PV_LockVal, PV_SendVal, PV_RcvVal)

PV_ARRAY(AmplPhaseLead, waveformData[1], U32, PV_RD,
                          PV_UpdateValWF, PV_LockWave, PV_SendVal, PV_RcvVal)

PV_ARRAY(AmplPhaseTrail, waveformData[3], U32, PV_RD,
                          PV_UpdateValWF, PV_LockWave, PV_SendVal, PV_RcvVal)

PV_ARRAY(FFT, waveformData[5], F32, PV_RD,
                          PV_UpdateValWF, PV_LockWave, PV_SendVal, PV_RcvVal)

//PV_ARRAY(CTF_Phase, waveformData[7], U32, PV_RD,
//                          PV_UpdateValWF, PV_LockWave, PV_SendVal, PV_RcvVal)

//PV_ARRAY(CTF_Ampl, waveformData[9], F32, PV_RD,
//                          PV_UpdateValWF, PV_LockWave, PV_SendVal, PV_RcvVal)

#undef PV_DEFS
#endif


#ifdef PV_Macros


const uint  PV_WF_Idx[NUM_WAVEFORMS] = {
  AmplPhaseLead_IDX,
  AmplPhaseTrail_IDX,
  FFT_IDX   //,
//  CTF_Phase_IDX,
//  CTF_Ampl_IDX
};

#define  WAVE_PV(w)  (nonRegPVs + PV_WF_Idx[(w)/2])


#undef PV_Macros
#endif

//-----------------------------------------------------------------------------

#endif  // end of else portion of:  #ifndef RegDefs ... #else ... #endif

#ifndef RegDefs // start of after-RegDefs portion


//---- bit usage for setDSPState (DSP SetDSPState 0x00100201) ---
//---------------------------------------------------------------
#define  CMD_STOP_IPULSING      0x00000001  // stop pulsing using intern src
#define  CMD_START_IPULSING     0x00000002  // start pulsing using intern src
//       ---------------------
#define  CMD_IPULSING_BITS      0x00000003L

#define  CMD_DISABLE_EPULSING   0x00000004  // disable pulsing via extern trig
#define  CMD_ENABLE_EPULSING    0x00000008  // enable pulsing via extern trig
//       ---------------------
#define  CMD_EPULSING_BITS      0x0000000CL

#define  CMD_L_DATA_READ        0x00000010  // Done reading Rising edge plot
#define  CMD_T_DATA_READ        0x00000020  // Done reading Falling edge plot
#define  CMD_FFT_DATA_READ      0x00000040  // Done reading FFT plot

#define  CMD_UNUSED_7           0x00000080
//---------------------------------------------------------------
#define  CMD_CTL_MODE_AUTO      0x00000100  // set ctl to auto mode
#define  CMD_CTL_MODE_MANUAL    0x00000020  // set ctl to manual mode
//       -------------------
#define  CMD_CTL_MODE_BITS      0x00000300L

// The following are ignored when CMD_CTL_MODE_AUTO is set
#define  CMD_DISAB_FREQ_OFFSET  0x00000400  // disable RF frequency offset
#define  CMD_ENAB_FREQ_OFFSET   0x00000800  // enable RF frequency offset
//       ---------------------
#define  CMD_FREQ_OFFSET_BITS   0x00000C00L

#define  CMD_USE_CAV_CHAN       0x00001000  // use cav chan for RF ampl feedback
#define  CMD_USE_FWD_CHAN       0x00002000  // use fwd chan for RF ampl feedback
//       -----------------
#define  CMD_USE_CHAN_BITS      0x00003000L

#define  CMD_AMPL_CTL_OFF       0x00004000  // disable RF ampl control
#define  CMD_AMPL_CTL_ON        0x00008000  // enable RF ampl control
//       ------------------
#define  CMD_AMPL_CTL_BITS      0x0000C000L
//---------------------------------------------------------------
#define  CMD_PHCTL_OFF_SEL_OFF  0x00010000L  // disble RF phase ctl and SEL
#define  CMD_PHCTL_ON_SEL_OFF   0x00020000L  // enable RF phase ctl, disable SEL
#define  CMD_PHCTL_OFF_SEL_ON   0x00040000L  // disable RF phase ctl, enable SEL
//       --------------------
#define  CMD_PHCTL_SEL_BITS     0x00070000L

#define  CMD_UNUSED_19          0x00080000L

#define  CMD_TUNER_CTL_OFF      0x00100000L  // disable tuner control
#define  CMD_TUNER_CTL_ON       0x00200000L  // enable tuner control
//       --------------------
#define  CMD_TUNER_CTL_BITS     0x00300000L

#define  CMD_UNUSED_22          0x00400000
#define  CMD_UNUSED_23          0x00800000
//---------------------------------------------------------------
#define  CMD_UNUSED_24          0x01000000L
#define  CMD_UNUSED_25          0x02000000L
#define  CMD_UNUSED_26          0x04000000L
#define  CMD_UNUSED_27          0x08000000L

#define  CMD_UNUSED_28          0x10000000L
#define  CMD_UNUSED_29          0x20000000L
#define  CMD_UNUSED_30          0x40000000L
#define  CMD_UNUSED_31          0x80000000L
//---------------------------------------------------------------


//--- bit-masks to determine which groups have a bit set ---
const ulong SetDSPStateMasks[] = {
  CMD_IPULSING_BITS,
  CMD_EPULSING_BITS,
  CMD_CTL_MODE_BITS,
  CMD_FREQ_OFFSET_BITS,
  CMD_USE_CHAN_BITS,
  CMD_AMPL_CTL_BITS,
  CMD_PHCTL_SEL_BITS,
  CMD_TUNER_CTL_BITS,
  0  //indicates end of list
};



const char far *SetDSPStateDesc[32] = {
//------------------------------------------
  "Stop RF pulsing via intern src",
  "Start RF pulsing using intern src",

  "Disable RF pulsing via extern trig",
  "Enable RF pulsing via extern trig",

  "", "", "", "",
//------------------------------------------
  "Set Ctl to Auto Mode",
  "Set Ctl to Manual Mode",

  "Disable RF Freq Offset",
  "Enable RF Freq Offset",

  "Use cav chan for RF ampl feedback",
  "Use fwd chan for RF ampl feedback",

  "Disable RF Ampl ctl",
  "Enable RF Ampl ctl",
//------------------------------------------
  "Disable RF Phase Ctl and SEL",
  "Enable RF Phase Ctl, disable SEL",
  "Disable RF Phase Ctl, enable SEL",

  "",

  "Disable Tuner Ctl",
  "Enable Tuner Ctl",

  "", ""
//------------------------------------------
  "", "", "", "",
  "", "", "", ""
//------------------------------------------
};




//--- parameter ID's for data in LLRF2 Device Data file ---
//--------------------------------------------------------------------------
// WARNING:  ID's can be added, but existing ID's (and the format of the
//           data they represent) should NEVER change!  NOTE:  Existing ones
//           can be "eliminated" by changing their name and no longer
//           using them, but never REUSE an old one for something new!!!
//--------------------------------------------------------------------------
//  NOTE:  ID's >= 1000 are rsvd for values shared by most applications
//--------------------------------------------------------------------------
const ParamData far  devDataParams[] = {
//                ID   ---value---  --restoreFunc--
  PERSISTENT_VAR(  1, trigPos,         NULL ),

  PERSISTENT_VAR(  2, warmLevel,       NULL ),
  PERSISTENT_VAR(  3, warmTime,        NULL ),
  PERSISTENT_VAR(  4, rampTime,        NULL ),

  PERSISTENT_VAR(  5, fwdAtten,        NULL ),
  PERSISTENT_VAR(  6, revAtten,        NULL ),
  PERSISTENT_VAR(  7, cavAtten,        NULL ),

  PERSISTENT_VAR(  8, spare0Atten,     NULL ),
  PERSISTENT_VAR(  9, spare1Atten,     NULL ),
  PERSISTENT_VAR( 10, spare2Atten,     NULL ),

  PERSISTENT_VAR( 11, outAtten,        NULL ),

  PERSISTENT_VAR( 12, pulseFreq,       NULL ),
  PERSISTENT_VAR( 13, pulseDuty,       NULL ),

  PERSISTENT_VAR( 14, rf_P,            NULL ),
  PERSISTENT_VAR( 15, rf_I,            NULL ),
  PERSISTENT_VAR( 16, rf_D,            NULL ),

  PERSISTENT_VAR( 17, tuner_P,         NULL ),
  PERSISTENT_VAR( 18, tuner_I,         NULL ),
  PERSISTENT_VAR( 19, tuner_D,         NULL ),

  PERSISTENT_VAR( 20, rf_B0,           NULL ),
  PERSISTENT_VAR( 21, rf_WCtrl,        NULL ),
  PERSISTENT_VAR( 22, rf_WObs,         NULL ),

  PERSISTENT_VAR( 23, tuner_B0,        NULL ),
  PERSISTENT_VAR( 24, tuner_WCtrl,     NULL ),
  PERSISTENT_VAR( 25, tuner_WObs,      NULL ),

  PERSISTENT_VAR( 26, cavLowIlkLevel,  NULL ),
  PERSISTENT_VAR( 27, cavLowIlkTime,   NULL ),

  PERSISTENT_VAR( 28, dvdtIlkRate,     NULL ),

  PERSISTENT_VAR( 29, revHighIlkRatio, NULL ),
  PERSISTENT_VAR( 30, revHighIlkTime,  NULL ),

  PERSISTENT_VAR( 31, fwdLowIlkRatio,  NULL ),
  PERSISTENT_VAR( 32, fwdLowIlkTime,   NULL ),

  PERSISTENT_VAR( 33, fwdHighIlkLevel, NULL ),
  PERSISTENT_VAR( 34, fwdHighIlkTime,  NULL ),

  PERSISTENT_VAR( 35, fwdCavPhase,     NULL ),

  PERSISTENT_VAR( 36, ffPhase,         NULL ),
  PERSISTENT_VAR( 37, ffGain,          NULL ),

  PERSISTENT_VAR( 38, freqOffset,      NULL ),

  PERSISTENT_VAR( 39, rfq_t1_offset,   NULL ),
  PERSISTENT_VAR( 40, rfq_t2_offset,   NULL ),

  PERSISTENT_VAR( 41, fwdPhaseGain,      NULL ),
  PERSISTENT_VAR( 42, fwdPhaseOffset,    NULL ),
  PERSISTENT_VAR( 43, fwdAmplGain,       NULL ),
  PERSISTENT_VAR( 44, fwdAmplOffset,     NULL ),

  PERSISTENT_VAR( 45, revPhaseGain,      NULL ),
  PERSISTENT_VAR( 46, revPhaseOffset,    NULL ),
  PERSISTENT_VAR( 47, revAmplGain,       NULL ),
  PERSISTENT_VAR( 48, revAmplOffset,     NULL ),

  PERSISTENT_VAR( 49, cavPhaseGain,      NULL ),
  PERSISTENT_VAR( 50, cavPhaseOffset,    NULL ),
  PERSISTENT_VAR( 51, cavAmplGain,       NULL ),
  PERSISTENT_VAR( 52, cavAmplOffset,     NULL ),

  PERSISTENT_VAR( 53, ref0PhaseGain,     NULL ),
  PERSISTENT_VAR( 54, ref0PhaseOffset,   NULL ),
  PERSISTENT_VAR( 55, ref0AmplGain,      NULL ),
  PERSISTENT_VAR( 56, ref0AmplOffset,    NULL ),

  PERSISTENT_VAR( 57, spare0PhaseGain,   NULL ),
  PERSISTENT_VAR( 58, spare0PhaseOffset, NULL ),
  PERSISTENT_VAR( 59, spare0AmplGain,    NULL ),
  PERSISTENT_VAR( 60, spare0AmplOffset,  NULL ),

  PERSISTENT_VAR( 61, spare1PhaseGain,   NULL ),
  PERSISTENT_VAR( 62, spare1PhaseOffset, NULL ),
  PERSISTENT_VAR( 63, spare1AmplGain,    NULL ),
  PERSISTENT_VAR( 64, spare1AmplOffset,  NULL ),

  PERSISTENT_VAR( 65, spare2PhaseGain,   NULL ),
  PERSISTENT_VAR( 66, spare2PhaseOffset, NULL ),
  PERSISTENT_VAR( 67, spare2AmplGain,    NULL ),
  PERSISTENT_VAR( 68, spare2AmplOffset,  NULL ),

  PERSISTENT_VAR( 69, ref1PhaseGain,     NULL ),
  PERSISTENT_VAR( 70, ref1PhaseOffset,   NULL ),
  PERSISTENT_VAR( 71, ref1AmplGain,      NULL ),
  PERSISTENT_VAR( 72, ref1AmplOffset,    NULL ),

  PERSISTENT_VAR( 73, maxAmplSet,        NULL ),

  #define SharedDevParams
  #include <sharedAppl.c>

  END_PARAM_DATA
};

//----------------------------------------------------

//RESTORE_FUNC(ampSet);

const ParamData far  settingsParams[] = {
//APPLDATA
//_nearcall void xxxSetRestore(void far *curVal, char *newVal, uint count);
//
//                ID   ---value---  --restoreFunc--
//  PERSISTENT_VAR(  1,  xxxSet,     PVRF(xxxSet) ),
  PERSISTENT_VAR(  1, piezoSetPos,    NULL ),
  PERSISTENT_VAR(  2, stepperSetPos,  NULL ),

  PERSISTENT_VAR(  3, phaseSet,       NULL ),
  PERSISTENT_VAR(  4, amplSet,        NULL ),

  PERSISTENT_VAR(  5, pulseAmplSet,    NULL ),
  PERSISTENT_VAR(  6, pulsePhaseSet,   NULL ),

  PERSISTENT_VAR(  7, rfq_manTuneSet,  NULL ),

  END_PARAM_DATA
};


//-----------------------------------------------------------------
//  Application-specific values needed to restore the state of the
//  controller following a "soft" reset (i.e. a restart without a
//  CPU reset)
//-----------------------------------------------------------------
const ParamData far  ctlrStateParams[] = {
// sample values from sharedAppl.c, common to all applications
//                   --ID--  ---value,granularity---  --restoreFunc--
//    PERSISTENT_VARx(  1000,  prevState,               NULL ),
//    PERSISTENT_VARx(  1001,  curState,                NULL ),
//    PERSISTENT_VARx(  1002,  nextState,               NULL ),
//    PERSISTENT_TBL (  1003,  setVal,4,                NULL ),
// // PERSISTENT_TBL (  1004,  shadowRegValues,4,       NULL ),

//APPLDATA
//    PERSISTENT_VAR(  1,  myApplStateVar,        NULL ),


  #define SharedCtlrStateParams
  #include <sharedAppl.c>

  END_PARAM_DATA
};


//-----------------------------------------------------------------------------
//  A list of parameterized "persistent" data.  Backup copies of such values
//  are stored in files and restored on startup.
//-----------------------------------------------------------------------------

//    Base File Name    Category  ParamData
//    ----------------  --------  ---------
//--- add any application-specific parameterized file defs here ---
//const ParamFile far  myAppCfgFile =
//  { "myFileName",     PDC_CTLR, myCfgParams,   0 };

ParamFile  settingsFile =
    { "settings",       PDC_SET,  settingsParams  };


#define SharedParams
#include <sharedAppl.c>


ParamFile *paramFiles[] = {
  #define SharedParamsList
  #include <sharedAppl.c>
//--- add any application-specific ParamData references here ---
//  &myAppCfgFile
  &settingsFile
};

const int NumParamFiles = sizeof(paramFiles) / sizeof(paramFiles[0]);



float  ilkRatioGain, ilkRatioOffset, maxOutAmplGain, maxOutAmplOffset;



//===== FPGA registers that require special handling =====
#define  RF_SOFT_ON  0x0E



#define  DSP_RD( dspName, dspType, mbReg, gain, offset, typ, noise )  \
    { dspName, RT_##dspType, &mbReg##_RegInfo, gain, offset, typ, noise }

const RD_DSP_REG far rdDSPregs[] = {
  //      DSP Reg ID      DSP fmt ModBus Reg      Gain              Offset
  //      --------------  ------- --------------  --------------    ---------------
  DSP_RD( FwdPhase,       PHASE,  fwdPhase,       &fwdPhaseGain,    &fwdPhaseOffset,    0x8230D9E6L,  5 ),
  DSP_RD( RevPhase,       PHASE,  revPhase,       &revPhaseGain,    &revPhaseOffset,    0xE64288E8L,  5 ),
  DSP_RD( CavPhase,       PHASE,  cavPhase,       &cavPhaseGain,    &cavPhaseOffset,    0xDA93A22CL,  5 ),
  DSP_RD( Ref0Phase,      PHASE,  ref0Phase,      &ref0PhaseGain,   &ref0PhaseOffset,   0xEBF0809CL,  5 ),

  DSP_RD( Spare0Phase,    PHASE,  spare0Phase,    &spare0PhaseGain, &spare0PhaseOffset, 0x8230D9E6L,  5 ),
  DSP_RD( Spare1Phase,    PHASE,  spare1Phase,    &spare1PhaseGain, &spare1PhaseOffset, 0xE64288E8L,  5 ),
  DSP_RD( Spare2Phase,    PHASE,  spare2Phase,    &spare2PhaseGain, &spare2PhaseOffset, 0xDA93A22CL,  5 ),
  DSP_RD( Ref1Phase,      PHASE,  ref1Phase,      &ref1PhaseGain,   &ref1PhaseOffset,   0xEBF0809CL,  5 ),

  DSP_RD( FwdAmpl,        F32,    fwdAmpl,        &fwdAmplGain,     &fwdAmplOffset,     0x3F8D0BA5L,  5 ),
  DSP_RD( RevAmpl,        F32,    revAmpl,        &revAmplGain,     &revAmplOffset,     0x3E606060L,  5 ),
  DSP_RD( CavAmpl,        F32,    cavAmpl,        &cavAmplGain,     &cavAmplOffset,     0x4013A76BL,  5 ),
  DSP_RD( Ref0Ampl,       F32,    ref0Ampl,       &ref0AmplGain,    &ref0AmplOffset,    0x3EE4FE6EL,  5 ),

  DSP_RD( Spare0Ampl,     F32,    spare0Ampl,     &spare0AmplGain,  &spare0AmplOffset,  0x3F8D0BA5L,  5 ),
  DSP_RD( Spare1Ampl,     F32,    spare1Ampl,     &spare1AmplGain,  &spare1AmplOffset,  0x3E606060L,  5 ),
  DSP_RD( Spare2Ampl,     F32,    spare2Ampl,     &spare2AmplGain,  &spare2AmplOffset,  0x4013A76BL,  5 ),
  DSP_RD( Ref1Ampl,       F32,    ref1Ampl,       &ref1AmplGain,    &ref1AmplOffset,    0x3EE4FE6EL,  5 ),

  DSP_RD( MaxFwdAmpl,     F32,    maxFwdAmpl,     &fwdAmplGain,     &fwdAmplOffset,     0x407918DEL,  0 ),
  DSP_RD( MaxRevAmpl,     F32,    maxRevAmpl,     &revAmplGain,     &revAmplOffset,     0x3FE3C7EEL,  0 ),
  DSP_RD( MaxCavAmpl,     F32,    maxCavAmpl,     &cavAmplGain,     &cavAmplOffset,     0x40DF6F19L,  0 ),

  DSP_RD( MaxSpare0Ampl,  F32,    maxSpare0Ampl,  &spare0AmplGain,  &spare0AmplOffset,  0x407918DEL,  0 ),
  DSP_RD( MaxSpare1Ampl,  F32,    maxSpare1Ampl,  &spare1AmplGain,  &spare1AmplOffset,  0x3FE3C7EEL,  0 ),
  DSP_RD( MaxSpare2Ampl,  F32,    maxSpare2Ampl,  &spare2AmplGain,  &spare2AmplOffset,  0x40DF6F19L,  0 ),

  DSP_RD( MaxOutAmpl,     F32,    maxOutAmpl,     &maxOutAmplGain,  &maxOutAmplOffset,  0x3FCD6FCEL,  0 ),

  DSP_RD( StatusBitsDSP,  BITS,   statusBitsDSP,  NULL,             NULL,               0x00302088L,  0 ),

  DSP_RD( DSPLoad,        F32,    dspLoad,        NULL,             NULL,               0x4284B886L, 10 ),
  DSP_RD( PCBTemp,        F32,    pcbTemp,        NULL,             NULL,               0x4284B886L,  5 ),
  DSP_RD( Fan0Spd,        U32,    fan0Spd,        NULL,             NULL,                     1000,   5 ),
  DSP_RD( Fan1Spd,        U32,    fan1Spd,        NULL,             NULL,                     1000,   5 ),

  DSP_RD( RFQ_T1_Output,  F32,    rfq_t1_output,  NULL,             NULL,               0x00000000L,  0 ),
  DSP_RD( RFQ_T2_Output,  F32,    rfq_t2_output,  NULL,             NULL,               0x00000000L,  0 ),

//--- R/W registers that we want to read from each pass ---
  DSP_RD( PiezoTunerPos,  U32,    piezoReadPos,   NULL,             NULL,               0x00000000L,  0 ),
  DSP_RD( StepperPos,     U32,    stepperReadPos, NULL,             NULL,               0x00000000L,  0 ),

//--- RO registers in "special" registers group ---
  DSP_RD( PercentFFT,     U32,    percentFFT,     NULL,             NULL,               0x00000064L,  0 ),
  DSP_RD( SizeFFT,        U32,    sizeFFT,        NULL,             NULL,               0x000003E9L,  0 ),
  DSP_RD( ROMProgErrs,    U32,    romProgErrs,    NULL,             NULL,               0x00000000L,  0 ),

//--- R/W registers in "special" group that provide the current value from ---
//--- a table of values ---
  DSP_RD( FwdMagCalib,      F32,    fwdMagCalib,      NULL,         NULL,               0x311743D5L,  0 ),
  DSP_RD( FwdPhaseCalib,    PHASE,  fwdPhaseCalib,    NULL,         NULL,               0x8C7EAFCEL,  0 ),

  DSP_RD( RevMagCalib,      F32,    revMagCalib,      NULL,         NULL,               0x308A5217L,  0 ),
  DSP_RD( RevPhaseCalib,    PHASE,  revPhaseCalib,    NULL,         NULL,               0x96F5A5A2L,  0 ),

  DSP_RD( CavMagCalib,      F32,    cavMagCalib,      NULL,         NULL,               0x3187AE52L,  0 ),
  DSP_RD( CavPhaseCalib,    PHASE,  cavPhaseCalib,    NULL,         NULL,               0x8F7F91BCL,  0 ),

  DSP_RD( Spare0MagCalib,   F32,    spare0MagCalib,   NULL,         NULL,               0x311743D5L,  0 ),
  DSP_RD( Spare0PhaseCalib, PHASE,  spare0PhaseCalib, NULL,         NULL,               0x8C7EAFCEL,  0 ),

  DSP_RD( Spare1MagCalib,   F32,    spare1MagCalib,   NULL,         NULL,               0x308A5217L,  0 ),
  DSP_RD( Spare1PhaseCalib, PHASE,  spare1PhaseCalib, NULL,         NULL,               0x96F5A5A2L,  0 ),

  DSP_RD( Spare2MagCalib,   F32,    spare2MagCalib,   NULL,         NULL,               0x3187AE52L,  0 ),
  DSP_RD( Spare2PhaseCalib, PHASE,  spare2PhaseCalib, NULL,         NULL,               0x8C7EAFCEL,  0 ),

  DSP_RD( OutMagCalib0,     F32,    outMagCalib0,     NULL,         NULL,               0x30A93DD5L,  0 ),
  DSP_RD( OutPhaseCalib0,   PHASE,  outPhaseCalib0,   NULL,         NULL,               0x00000000L,  0 ),

  DSP_RD( OutMagCalib1,     F32,    outMagCalib1,     NULL,         NULL,               0x30A93DD5L,  0 ),
  DSP_RD( OutPhaseCalib1,   PHASE,  outPhaseCalib1,   NULL,         NULL,               0x00000000L,  0 )
};

#define  NUM_RD_DSP_REGS  (sizeof(rdDSPregs) / sizeof(RD_DSP_REG))

const uint  Num_RD_DSP_Regs = NUM_RD_DSP_REGS;


#define  DSP_WR( dspName, dspType, mbReg, min, max, mode, gain, offset )  \
    { dspName, RT_##dspType, &mbReg##_RegInfo, min, max, WM_##mode, gain, offset }

const WR_DSP_REG far wrDSPregs[] = {
  //      DSP Reg ID      DSP fmt ModBus Reg       Min    Max     Mode  Gain            Offset
  //      --------------  ------- --------------   ----   ------  ----  --------------  ---------------
  DSP_WR( TrigPos,        U32,    trigPos,         0,     1023,   EVERY, NULL,          NULL ),
  DSP_WR( WarmLevel,      F32,    warmLevel,       0.0,   0.0,    EVERY, &cavAmplGain,  &cavAmplOffset ),
  DSP_WR( WarmTime,       F32,    warmTime,        0.0,   32767,  EVERY, NULL,          NULL ),
  DSP_WR( RampTime,       F32,    rampTime,        0.0,   32767,  EVERY, NULL,          NULL ),

  DSP_WR( FwdAtten,       U32,    fwdAtten,        0,     63,     EVERY, NULL,          NULL ),
  DSP_WR( RevAtten,       U32,    revAtten,        0,     63,     EVERY, NULL,          NULL ),
  DSP_WR( CavAtten,       U32,    cavAtten,        0,     63,     EVERY, NULL,          NULL ),

  DSP_WR( Spare0Atten,    U32,    spare0Atten,     0,     63,     EVERY, NULL,          NULL ),
  DSP_WR( Spare1Atten,    U32,    spare1Atten,     0,     63,     EVERY, NULL,          NULL ),
  DSP_WR( Spare2Atten,    U32,    spare2Atten,     0,     63,     EVERY, NULL,          NULL ),

  DSP_WR( OutAtten,       U32,    outAtten,        0,     63,     EVERY, NULL,          NULL ),

  DSP_WR( PhaseSet,       PHASE,  phaseSet,        0.0,   0.0,    EVERY, &cavPhaseGain, &cavPhaseOffset ),
  DSP_WR( AmplSet,        F32,    amplSet,         0.0,   0.0,    EVERY, &cavAmplGain,  &cavAmplOffset ),

  DSP_WR( PulseFreq,      F32,    pulseFreq,       0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( PulseDuty,      F32,    pulseDuty,       0.0,   0.0,    EVERY, NULL,          NULL ),

  DSP_WR( RF_P,           F32,    rf_P,            0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( RF_I,           F32,    rf_I,            0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( RF_D,           F32,    rf_D,            0.0,   0.0,    EVERY, NULL,          NULL ),

  DSP_WR( Tuner_P,        F32,    tuner_P,         0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( Tuner_I,        F32,    tuner_I,         0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( Tuner_D,        F32,    tuner_D,         0.0,   0.0,    EVERY, NULL,          NULL ),

  DSP_WR( RF_B0,          F32,    rf_B0,           0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( RF_WCtrl,       F32,    rf_WCtrl,        0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( RF_WObs,        F32,    rf_WObs,         0.0,   0.0,    EVERY, NULL,          NULL ),

  DSP_WR( Tuner_B0,       F32,    tuner_B0,        0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( Tuner_WCtrl,    F32,    tuner_WCtrl,     0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( Tuner_WObs,     F32,    tuner_WObs,      0.0,   0.0,    EVERY, NULL,          NULL ),

  DSP_WR( PulseAmplSet,   F32,    pulseAmplSet,    0.0,   0.0,    EVERY, &cavAmplGain,  &cavAmplOffset ),
  DSP_WR( PulsePhaseSet,  PHASE,  pulsePhaseSet,   0.0,   0.0,    EVERY, &cavPhaseGain, &cavPhaseOffset ),

  DSP_WR( CavLowIlkLevel, F32,    cavLowIlkLevel,  0.0,   0.0,    EVERY, &cavAmplGain,  &cavAmplOffset ),
  DSP_WR( CavLowIlkTime,  F32,    cavLowIlkTime,   0.0,   0.0,    EVERY, NULL,          NULL ),

  DSP_WR( DvDtIlkRate,    F32,    dvdtIlkRate,     0.0,   0.0,    EVERY, &cavAmplGain,  &cavAmplOffset ),

  DSP_WR( RevHighIlkRatio,F32,    revHighIlkRatio, 0.0,   0.0,    EVERY, &ilkRatioGain, &ilkRatioOffset ),
  DSP_WR( RevHighIlkTime, F32,    revHighIlkTime,  0.0,   0.0,    EVERY, NULL,          NULL ),

  DSP_WR( FwdLowIlkRatio, F32,    fwdLowIlkRatio,  0.0,   0.0,    EVERY, &ilkRatioGain, &ilkRatioOffset ),
  DSP_WR( FwdLowIlkTime,  F32,    fwdLowIlkTime,   0.0,   0.0,    EVERY, NULL,          NULL ),

  DSP_WR( FwdHighIlkLevel,F32,    fwdHighIlkLevel, 0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( FwdHighIlkTime, F32,    fwdHighIlkTime,  0.0,   0.0,    EVERY, NULL,          NULL ),

  DSP_WR( FwdCavPhase,    PHASE,  fwdCavPhase,     0.0,   0.0,    EVERY, NULL,          NULL ),

  DSP_WR( FFPhase,        PHASE,  ffPhase,         0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( FFGain,         F32,    ffGain,          0.0,   0.0,    EVERY, NULL,          NULL ),

  DSP_WR( FreqOffset,     F32,    freqOffset,      0.0,   0.0,    EVERY, NULL,          NULL ),

  DSP_WR( RFQ_ManTuneSet, F32,    rfq_manTuneSet,  0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( RFQ_T1_Offset,  F32,    rfq_t1_offset,   0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( RFQ_T2_Offset,  F32,    rfq_t2_offset,   0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( RFQ_T1_Feedback,F32,    rfq_t1_feedback, 0.0,   0.0,    EVERY, NULL,          NULL ),
  DSP_WR( RFQ_T2_Feedback,F32,    rfq_t2_feedback, 0.0,   0.0,    EVERY, NULL,          NULL ),

//--- WO registers in the "special" registers group ---
  DSP_WR( SetDSPState,    U32,    setDSPState,     0,     0,      EVERY, NULL,          NULL ),

//--- R/W registers in the "special" registers group ---
  DSP_WR( PiezoTunerPos,  U32,    piezoSetPos,     0,     0,      EVERY, NULL,          NULL ),
  DSP_WR( StepperPos,     U32,    stepperSetPos,   0,     0,      EVERY, NULL,          NULL ),

//--- R/W registers in "special" group that write a value into a table ---
//--- and cannot be continuously written to ---
  DSP_WR( FwdMagCalib,      F32,    fwdMagCalib,     0.0,   0.0,     ONCE, NULL,          NULL ),
  DSP_WR( FwdPhaseCalib,    PHASE,  fwdPhaseCalib,   0.0,   0.0,     ONCE, NULL,          NULL ),

  DSP_WR( RevMagCalib,      F32,    revMagCalib,     0.0,   0.0,     ONCE, NULL,          NULL ),
  DSP_WR( RevPhaseCalib,    PHASE,  revPhaseCalib,   0.0,   0.0,     ONCE, NULL,          NULL ),

  DSP_WR( CavMagCalib,      F32,    cavMagCalib,     0.0,   0.0,     ONCE, NULL,          NULL ),
  DSP_WR( CavPhaseCalib,    PHASE,  cavPhaseCalib,   0.0,   0.0,     ONCE, NULL,          NULL ),

  DSP_WR( Spare0MagCalib,   F32,   spare0MagCalib,   0.0,   0.0,     ONCE, NULL,          NULL ),
  DSP_WR( Spare0PhaseCalib, PHASE, spare0PhaseCalib, 0.0,   0.0,     ONCE, NULL,          NULL ),

  DSP_WR( Spare1MagCalib,   F32,   spare1MagCalib,   0.0,   0.0,     ONCE, NULL,          NULL ),
  DSP_WR( Spare1PhaseCalib, PHASE, spare1PhaseCalib, 0.0,   0.0,     ONCE, NULL,          NULL ),

  DSP_WR( Spare2MagCalib,   F32,   spare2MagCalib,   0.0,   0.0,     ONCE, NULL,          NULL ),
  DSP_WR( Spare2PhaseCalib, PHASE, spare2PhaseCalib, 0.0,   0.0,     ONCE, NULL,          NULL ),

  DSP_WR( OutMagCalib0,     F32,   outMagCalib0,     0.0,   0.0,     ONCE, NULL,          NULL ),
  DSP_WR( OutPhaseCalib0,   PHASE, outPhaseCalib0,   0.0,   0.0,     ONCE, NULL,          NULL ),

  DSP_WR( OutMagCalib1,     F32,   outMagCalib1,     0.0,   0.0,     ONCE, NULL,          NULL ),
  DSP_WR( OutPhaseCalib1,   PHASE, outPhaseCalib1,   0.0,   0.0,     ONCE, NULL,          NULL )
};

#define  NUM_WR_DSP_REGS  (sizeof(wrDSPregs) / sizeof(WR_DSP_REG))

const uint  Num_WR_DSP_Regs = NUM_WR_DSP_REGS;



U16 *rdFPGAregs[] = {
  &statusBitsFPGA,
  &statusBitsTuner,
  &statusBitsRF,
  &statusBitsILK,

  &spark0_ctr,
  &spark1_ctr,
  &cavlow_ctr,
  &rev_high_ctr,
  &dvdt_ctr,
  &fwd_low_ctr,
  &fwd_high_ctr,
  &PLC_ctr,
  &safety_ctr
};

#define  NUM_RD_FPGA_REGS  (sizeof(rdFPGAregs) / sizeof(U16 *))


U16 *wrFPGAregs[] = {
  &ADC_err_rst,
  &RF_ilk_rst,
  &ilk_ctr_rst1,
  &ilk_ctr_rst2
};

#define  NUM_WR_FPGA_REGS  (sizeof(wrFPGAregs) / sizeof(U16 *))



//=============================================================================
//  Definitions for event msgs.  Event msgs are the means that other processes
//  use to inform the application process of an event that it needs to handle.
//
//  In general, allowing multiple processes to concurrently modify the state
//  of the controller is a sure recipie for chaos.  By restricting all such
//  changes to the context of a single process, and sending messages to that
//  process to request changes, we eliminate all such concurrency issues (not
//  the least of which is keeping track of the order of events).
//
//  The most common event is a request for change of state that corresponds to
//  a message received from an external client via a protocol like ModBus.
//=============================================================================

//EVENT_DEF( myEvent )


//=============================================================================
//  This is where misc global variables should be defined
//=============================================================================
#define SharedGlobals
#include <sharedAppl.c>

//--- add definitions for application-specific global variables here ---

LockData  locks[NUM_LOCKS];


bool  fpgaInitialized, pulseAmpl, pulsePhase;

uint  ADCVal;

bool  autoReadDSP, autoWriteDSP, autoReadFPGA, autoWriteFPGA, showFiller;

bool  rfIsOn, rfPulsing, tunerCtlisON, rfAtWarm, rfRamping, readWaves,
      cavLocked, amplCtlisON;

uint  curProgNum, phaseMode, lastIOTime;

const char *PhaseModeDesc[] = { "    OPEN", "PhaseCtl", "SEL Ctl ", "????????" };

//=============================================================================
//  Temporary functions to fill in for not-yet implemented system functions
//=============================================================================


#define LockOwnedBy(id)  ( locks[id].owner )
#define NotLockOwner(id)  (LockOwnedBy(id) != getpid())

//-----------------------------------------------------------------------------
//  Get the lock for the specified resource.
//
//  Returns:
//    If lock acquired:  depth of locks aquired by caller
//    On error:  < 0
//-----------------------------------------------------------------------------
int  getLock(uint id, ulong msTimeout)
{
  PID  mypid, owner;
  int  stat, count;
  task_t  *task;
  LockData  *lock;


  if (id >= NUM_LOCKS)  { errno = ENODEV;  return -1; }

  mypid = getpid();   lock = locks + id;

  _EnterCritical_();  owner = lock->owner;  _ExitCritical_();

//--- If we already own the lock, then just increment the count.  Since ---
//--- only the owner of the lock changes the count, we can safely incr ---
//--- and return the current value w/o worrying about preemption. ---
  if (owner == mypid)  { ++lock->count;  return lock->count; }

//--- If we aren't the owner, then try to get the lock.  If the previous ---
//--- owner died, then the LockData will be invalid but the semaphore for ---
//--- the lock will have been released so another task will be able to get ---
//--- it (and change the LockData values to indicate the new owner) ---
  stat = sem_wait_(&lock->sem, msToTicks(msTimeout));

  if (stat < 0)  {
    if (msTimeout > 100)  {
      _EnterCritical_();
      owner = lock->owner;  count = lock->count;
      _ExitCritical_();
      task = tasks + _activeTaskIdx_(owner);
      printf("Lock %d still held by %.15s[%d], count %d\r\n",
              id, task->argv[0], owner, count );
    }
    return stat;
  }


//--- Again, if we got here, we own the lock and only the owner will change ---
//--- the LockData values ---
  lock->owner = mypid;
  lock->count = 1;

//  fprintf(DiagLog, "%u now has lock %d\n", mypid, id);  //tdebug


  return lock->count;
}



//-----------------------------------------------------------------------------
//  If the calling process owns the specified lock, then reduce its lock count
//  (or release the lock if no more counts)
//
//  Returns:
//    On success:  New lock depth held by caller
//    On error:  < 0
//-----------------------------------------------------------------------------
int releaseLock(uint id)
{
  uint  owner;
  LockData  *lock;


  if (id >= NUM_LOCKS)  {
    fprintf(DiagLog, "*** releaseLock(%u): Invalid lock ID ***\n", id);
    return -1;
  }

  lock = locks + id;

  _EnterCritical_();  owner = lock->owner;  _ExitCritical_();

  if (owner != getpid())  {
    fprintf(DiagLog, "*** releaseLock(%u): Lock owned by: %u ***\n",
            id, owner);
    return -1;
  }

  // At this point, we know we own the lock, so no other task will make changes
  // to the values in the LockData structure until we release the semaphore

  if (lock->count < 1)  {
    fprintf(DiagLog, "*** releaseLock(%u) with count of %u ***\n",
            id, lock->count);
    return -1;
  }

  if (--lock->count == 0)  {
//    fprintf(DiagLog, "%u releasing lock %d\n",  getpid(), id);  //tdebug
    lock->owner = -1;  sem_post(&lock->sem);
  }

  return lock->count;
}



//=============================================================================


//-----------------------------------------------------------------------------
//  Convert a new ModBus register value to a floating point number
//
//  Note that the size of the buffer pointed to by newVal must match the
//  size of the specified ModBus register value.
//-----------------------------------------------------------------------------
float mb2float(void *newVal, RegInfo *mbRegInfo)
{
  uchar  type = mbRegInfo->dataType;

  switch (type)  {
    case DT_S16:  return (float) MB_VAL(S16);
    case DT_U16:  return (float) MB_VAL(U16);
    case DT_S32:  return (float) MB_VAL(S32);
    case DT_U32:  return (float) MB_VAL(U32);
    case DT_F32:  return         MB_VAL(F32);
  }

  return 0.0;
}

//-----------------------------------------------------------------------------
//  Convert a new ModBus register value to a signed long number
//
//  Note that the size of the buffer pointed to by newVal must match the
//  size of the specified ModBus register value.
//-----------------------------------------------------------------------------
S32 mb2S32(void *newVal, RegInfo *mbRegInfo)
{
  uchar  type = mbRegInfo->dataType;

  switch (type)  {
    case DT_S16:  return (S32) MB_VAL(S16);
    case DT_U16:  return (S32) MB_VAL(U16);
    case DT_S32:  return       MB_VAL(S32);
    case DT_U32:  return (S32) MB_VAL(U32);
    case DT_F32:  return (S32) MB_VAL(F32);
  }

  return 0;
}

//-----------------------------------------------------------------------------
//  Convert a new ModBus register value to an unsigned long number
//
//  Note that the size of the buffer pointed to by newVal must match the
//  size for the type of the specified ModBus register value.
//-----------------------------------------------------------------------------
U32 mb2U32(void *newVal, RegInfo *mbRegInfo)
{
  uchar  type = mbRegInfo->dataType;

  switch (type)  {
    case DT_S16:  return (U32) MB_VAL(S16);
    case DT_U16:  return (U32) MB_VAL(U16);
    case DT_S32:  return (U32) MB_VAL(S32);
    case DT_U32:  return       MB_VAL(U32);
    case DT_F32:  return (U32) MB_VAL(F32);
  }

  return 0;
}

//-----------------------------------------------------------------------------
//  Convert a floating value to the format of the specified ModBus register
//
//  Note that the size of the buffer pointed to by newVal must match the
//  size of the specified ModBus register value.
//-----------------------------------------------------------------------------
void float2mb(void *newVal, RegInfo *mbRegInfo, float fromVal)
{
  uchar  type = mbRegInfo->dataType;

  switch (type)  {
    case DT_S16:  MB_VAL(S16) = (int) fromVal;  break;
    case DT_U16:  MB_VAL(U16) = (uint) fromVal;  break;
    case DT_S32:  MB_VAL(S32) = (long) fromVal;  break;
    case DT_U32:  MB_VAL(U32) = (ulong) fromVal;  break;
    case DT_F32:  MB_VAL(float) = fromVal;  break;
  }
}

//-----------------------------------------------------------------------------
//  Convert a DSP register value to a format that is compatible with the
//  specified ModBus register (i.e. make sure it is a floating point or
//  an integer value as needed).
//-----------------------------------------------------------------------------
void dsp2mb(RegInfo *mbReg, S16 *mbVal,
            const RD_DSP_REG far *dspReg, ulong *dspVal)
{
  uchar  mbType = mbReg->dataType, dspType = dspReg->type;
  bool  offset0, gain1, scaleIt;
  float  gain, offset, fval;


//--- if DSP value is a scaled PHASE value, rescale it into a float ---
  if (dspType == RT_PHASE)  {
    *(float *)dspVal = (float)*dspVal * PhaseConvFactor;  dspType = RT_F32; }


//--- determine if the value needs to be scaled ---
  offset = (dspReg->offset ? *dspReg->offset : 0.0);  // NULL == no offset
  gain = (dspReg->gain ? *dspReg->gain : 1.0);  // NULL == no gain

  if (gain == 0.0)  gain = 1.0;

  offset0 = (offset == 0.0);  gain1 = (gain == 1.0);

  scaleIt = (!offset0 OR !gain1);

  if (scaleIt)  {
    switch (dspType)  {
      case RT_S32:  fval = (float)((S32)*dspVal);  break;
      case RT_U32:  fval = (float)((U32)*dspVal);  break;
      case RT_F32:  fval = *(float *)dspVal;  break;
    }
    if (!offset0)  fval -= offset;
    if (!gain1)  fval = fval * gain;
    *(float *)dspVal = fval;  dspType = RT_F32;
  }


//--- Make sure the DSP format is the same type as the ModBus format ---
//--- (i.e. both are 16 or 32-bit integers or both are 32-bit floats) ---
  if ((dspType == DT_F32) != (mbType == DT_F32))  {
    if (mbType == DT_F32)  {  // non-float to float
      if (dspType == DT_S32)  {  // signed or unsigned?
        *(float *)dspVal = (float)((S32)*dspVal);  dspType = DT_F32;  }
      else  {
        *(float *)dspVal = (float)((U32)*dspVal);  dspType = DT_F32;  }
    }
    else  {  // float to non-float
      fval = *(float *)dspVal;
      if ((mbType == DT_S16) OR (mbType == DT_S32)) { // signed or unsigned?
        *(S32 *)dspVal = (S32) fval;  dspType = DT_S32; }
      else  {
        *(U32 *)dspVal = (U32) fval;  dspType = DT_U32; }
    }
  }

  if (mbReg->numRegs < 2)  // ModBus is 16-bit int, DSP is 32-bit int
    *(uint *)mbVal = (uint)(*dspVal);
  else  // both are 32-bits (either float or long ints)
    memcpy(mbVal, dspVal, 4);
}


//-----------------------------------------------------------------------------
//  Convert a ModBus register value to the specified DSP register format
//-----------------------------------------------------------------------------
void mb2dsp(RegInfo *mbReg, S16 *mbVal,
            const WR_DSP_REG far *dspReg, ulong *dspVal)
{
  uchar  dspType = dspReg->type;
  bool  offset0, gain1, scaleIt;
  float  gain, offset, fval;


//--- determine if the value needs to be scaled ---
  offset = (dspReg->offset ? *dspReg->offset : 0.0);  // NULL == no offset
  gain = (dspReg->gain ? *dspReg->gain : 1.0);  // NULL == no gain

  if (gain == 0.0)  gain = 1.0;

  offset0 = (offset == 0.0);  gain1 = (gain == 1.0);

  scaleIt = (!offset0 OR !gain1);

  if (scaleIt)  {
    fval = mb2float(mbVal, mbReg);
    if (!gain1)  fval = fval / gain;
    if (!offset0)  fval += offset;
  }

//--- convert either the scaled or unscaled ModBus value to the DSP format ---
  switch (dspType)  {
    case RT_S32:
      if (scaleIt)
        *(S32 *)dspVal = (S32)fval;
      else
        *(S32 *)dspVal = mb2S32(mbVal, mbReg);
      break;

    case RT_U32:
      if (scaleIt)
        *(U32 *)dspVal = (U32)fval;
      else
        *(U32 *)dspVal = mb2U32(mbVal, mbReg);
      break;

    case RT_F32:
      if (scaleIt)
        *(float *)dspVal = fval;
      else
        *(float *)dspVal = mb2float(mbVal, mbReg);
      break;

    case RT_PHASE:
      if (scaleIt)
        *dspVal = (ulong)(fval / PhaseConvFactor);
      else
        *dspVal = (ulong)(mb2float(mbVal, mbReg) / PhaseConvFactor);
      break;
  }

}


//=============================================================================
//  Functions referenced in the Modbus register definitions to handle writes
//  to specific global values defined as ModBus registers
//=============================================================================

//-----------------------------------------------------------------------------
//  Got a new value for the register that enables/disables RF output.  This
//  register is different from the other writable FPGA registers in that the
//  value does not represents bits that need to be toggled to reset some other
//  value.
//-----------------------------------------------------------------------------
bool MB_RF_soft_on(SetRegFuncParams)
{
  writeFPGA(RF_SOFT_ON, MB_VAL(U16));

  return setU16(_SetRegFuncParams_);
}


//-----------------------------------------------------------------------------
//  Got a new value for one of the registers that mirrors one of the DSP's R/W
//  registers.  In response, we check the new value against the range of values
//  given for the DSP register and modify the ModBus register value if needed.
//
//  The writeMode specified for the DSP register determines when the new value
//  is written to the DSP as follows:
//
//  WM_ONCE:
//     The value must be written just once ASAP.  Writing to one of these DSP
//     registers triggers some action on the part of the DSP, so the register
//     must not be written to except in response to a write to the corresponding
//     ModBus register.  The most common use of this mode by the DSP is to
//     copy some related values to a table (possibly including the value just
//     written, or using the new value as an index in to the table?) as part
//     of a calibration process.
//
//  WM_EVERY:
//     The value represents some run-time setting and can be written to the DSP
//     repeatedly without causing a problem.  While writing such values to the
//     DSP just once would be sufficient if the the DSP never rebooted, such
//     values are not retained by the DSP across a reboot.  For that reason,
//     these values are resent to the DSP each pass through the application's
//     main loop (i.e. by the applic task).  The result is that if the DSP
//     restarts, or if a write should fail one time, the correct settings will
//     still be quickly (re)asserted.
//
//-----------------------------------------------------------------------------
bool MB_DSP_REG(SetRegFuncParams)
{
  uint  ui;
  RegInfo *mbRegInfo = (RegInfo *)info;
  ulong  dspVal;
  float  fval;
  const WR_DSP_REG far *dspReg = wrDSPregs;
  char  temp[32];

  for (ui=0; ui<NUM_WR_DSP_REGS; ++ui,++dspReg)  {
    if (dspReg->mbReg != mbRegInfo)  continue;

  //--- if a phase value, then make sure it is within the range 0 to 360 ---
    if (dspReg->type == RT_PHASE)  {
      fval = mb2float(newVal, mbRegInfo);
      if ((fval < 0.0) OR (fval >= 360.0))  {
        while (fval < 0.0)  fval += 360.0;
        while (fval >= 360.0)  fval -= 360.0;
        float2mb(newVal, mbRegInfo, fval);
      }
    }

  //--- check the new value against the specified range of allowed values ---
    if (dspReg->minVal != dspReg->maxVal)  {
      fval = mb2float(newVal, mbRegInfo);
      if (fval < dspReg->minVal)  float2mb(newVal, mbRegInfo, dspReg->minVal);
      if (fval > dspReg->maxVal)  float2mb(newVal, mbRegInfo, dspReg->maxVal);
    }

    if (mbRegInfo->flags & LogChanges)
      printf("new value for DSP reg: 0x%08lX: %.32s\r\n",
             dspReg->id, fmtRegVal(mbRegInfo, newVal, 0, temp) );

  //--- write the value to the DSP now? ---
    if (dspReg->writeMode == WM_ONCE)  {
      mb2dsp(mbRegInfo, newVal, dspReg, &dspVal);
      if (writeDSPReg(dspReg->id, dspVal, true) < 0)  return false;
    }

  //--- update the modbus register value using the new value ---
  // NOTE:  Can't use setMBReg() since that would just call this func again!
    switch (mbRegInfo->dataType)  {
      case DT_S16:  setS16(_SetRegFuncParams_);  break;
      case DT_U16:  setU16(_SetRegFuncParams_);  break;
      case DT_S32:  setS32(_SetRegFuncParams_);  break;
      case DT_U32:  setU32(_SetRegFuncParams_);  break;
      case DT_F32:  setF32(_SetRegFuncParams_);  break;
    }
    break;
  }

  return true;
}

//-----------------------------------------------------------------------------
//  process a new value for "cmdBitsAppl"
//-----------------------------------------------------------------------------
bool MB_cmdBitsAppl(SetRegFuncParams)
{
/*
  ulong newFmt = 0;

  cmdBitsAppl = MB_VAL(ulong);

//--- This is the older version of the bit-mapped cmds, so we translate ---
//--- the best we can in to the newer ones and pass them on ---

  if (cmdBitsAppl & OLDC_AMPL_OPEN_LOOP)  newFmt |= CMD_AMPL_CTL_OFF;
  if (cmdBitsAppl & OLDC_AMPL_CLOSED_LOOP)  newFmt |= CMD_AMPL_CTL_ON;

  if (cmdBitsAppl & OLDC_PHASE_OPEN_LOOP)  newFmt |= CMD_PHASE_CTL_OFF;
  if (cmdBitsAppl & OLDC_PHASE_CLOSED_LOOP)  newFmt |= CMD_PHASE_CTL_ON;

  if (cmdBitsAppl & OLDC_AMPL_TUNE_RUN)  newFmt |= CMD_RF_PULSING_ON;
  if (cmdBitsAppl & OLDC_AMPL_TUNE_STOP)  newFmt |= CMD_RF_PULSING_OFF;
  if (cmdBitsAppl & OLDC_PHASE_TUNE_RUN)  newFmt |= CMD_RF_PULSING_ON;
  if (cmdBitsAppl & OLDC_PHASE_TUNE_STOP)  newFmt |= CMD_RF_PULSING_OFF;

  SetMBReg(setDSPState, 0, U32, newFmt);
*/
  return true;
}

//-----------------------------------------------------------------------------
//  process a new value for "setDSPState"
//-----------------------------------------------------------------------------
bool MB_setDSPState(SetRegFuncParams)
{
  uint  i;
  ulong  mask, cmdBits;


  cmdBits = MB_VAL(ulong);

//--- setDSPState contains cmd bits in groups:  i.e. one bit is asserted to ---
//--- enable a state, the other to disable it.  None set within a group ---
//--- means do not change the current state for the group. ---

//--- Writing of values to the DSP is handled primarily by the application ---
//--- task.  To avoid slowing down the task that called this function ---
//--- waiting for access to the DSP interface, we just keep track of the ---
//--- most recent setting for each group of bits.  That way the actual ---
//--- writing of the bits (whenever it happens) will always use the last ---
//--- setting for each group, and we have a record of what state the DSP ---
//--- should be in any given time (and a way to re-assert that state if ---
//--- the DSP is reset) ---

  for (i=0; SetDSPStateMasks[i]; ++i)  {
    mask = SetDSPStateMasks[i];
    if (cmdBits & mask) setDSPState = (setDSPState & ~mask) | (cmdBits & mask);
  }

//--- Also, log a message showing what bits were set each time we get a new ---
//--- value ---
  for (i=0,mask=1; i<32; ++i,mask<<=1)  if (cmdBits & mask)
    fprintf(DiagLog, "setDSPstate: Set: %ls\n", SetDSPStateDesc[i]);


  return true;
}


//-----------------------------------------------------------------------------
//  process a new value for an amplitude value that needs to be checked
//  against maxAmplSet
//-----------------------------------------------------------------------------
bool MB_maxAmpl(SetRegFuncParams)
{
  if (REG->dataType != DT_F32)  return false;

  if (MB_VAL(F32) > maxAmplSet)  MB_VAL(F32) = maxAmplSet;

  return setF32(_SetRegFuncParams_);
}


//-----------------------------------------------------------------------------
//  process a new value for one of the xxxGain registers
//-----------------------------------------------------------------------------
bool MB_setGain(SetRegFuncParams)
{
  if (REG->dataType != DT_F32)  return false;

  if (MB_VAL(float) == 0.0)  return false;

  return setF32(_SetRegFuncParams_);
}


//-----------------------------------------------------------------------------
//  process a new value for a register with a phase value
//-----------------------------------------------------------------------------
bool MB_setPhase(SetRegFuncParams)
{
  if (REG->dataType != DT_F32)  return false;

  if (MB_VAL(float) < -720.0)  MB_VAL(float) = -360.0;
  if (MB_VAL(float) >  720.0)  MB_VAL(float) =  360.0;

  while (MB_VAL(float) < -360.0)  MB_VAL(float) += 360.0;
  while (MB_VAL(float) >  360.0)  MB_VAL(float) -= 360.0;

  return setF32(_SetRegFuncParams_);
}


//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
int _nearcall PV_UpdateVal(PV_OP_Params)
{
  //finish:  Need to update pv->newEvents and pv->val for any non-waveform
  //         PVs that don't do it elsewhere

//  printf("PV_UpdateVal(%ls)\r\n", pv->pvInfo.name);  //tdebug

  return -1;
}

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
int _nearcall PV_LockVal(PV_Lock_Params)
{
  //finish:  Implement locking for non-waveform values

//tdebug
//  printf("\r\nPV_LockVal(%ls, %d, %d)\r\n", pv->pvInfo.name, lockMode, lockType);

  return 0;
}

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
int _nearcall PV_SendVal(PV_RW_Params)
{
  PV_Info  pvInfo;


  //finish:  call (*copyFunc)(void *conInfo, void far *from, uint count)
  //         as needed to send PV value

  if (!pvGetInfo(id, &pvInfo))  pvInfo.name = "<invalid>";

  printf("PV_SendVal(%ls)\r\n", pvInfo.name);  //tdebug

  return -1;
}

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
int _nearcall PV_RcvVal(PV_RW_Params)
{
  PV_Info  pvInfo;


  //finish:  call (*copyFunc)(void *conInfo, void far *from, uint count)
  //         as needed to get new PV value
  //         (probably need more info so we know how big a payload is in
  //         the message that caused this call)

  if (!pvGetInfo(id, &pvInfo))  pvInfo.name = "<invalid>";

  printf("PV_RcvVal(%ls)\r\n", pvInfo.name);  //tdebug

  return -1;
}


//-----------------------------------------------------------------------------
//  If newer data available, update the copy of a PV's value that is sent to
//  external clients (and update the newEvents flag indicating what events the
//  new data applies to).
//-----------------------------------------------------------------------------
int _nearcall PV_UpdateValWF(PV_OP_Params)
{
  bool noChange;
  int  stat;
  uint  src, cpy, numBytes, srcLockID, cpyLockID;
  char far  *srcBuf, *cpyBuf;


  if ((src = WF_IDX(pv->val) - 1) >= NUM_WAVEFORMS*2)  return -1;
  cpy = src + 1;

//--- Don't even TRY to get the locks unless we KNOW there is new data ---
  _EnterCritical_();
  noChange = (dataTimeMS[cpy] == dataTimeMS[src]);
  _ExitCritical_();
  if (noChange)  return 0;

  if (ShowDebugMsgs)
    printf("New data avail for PV val: [%ls]\r\n", pv->pvInfo.name);


  srcBuf = (char far *) waveformData[src];    srcLockID = WAVE_DATA_0 + src;
  cpyBuf = (char far *) waveformData[cpy];    cpyLockID = WAVE_DATA_0 + cpy;


//--- Get a lock on the dest/cpy buffer so it won't change while we are ---
//--- accessing (and possibly updating) it ---

  if (getLock(cpyLockID, 0) <= 0)  return -1;  // try to get lock on dest buffer

//--- If the timestamps for the source and destination buffers are ---
//--- different, then update the contents and timestamp for the ---
//--- destination buffer.  NOTE that we don't really care about the ---
//--- possibility of the src date/time changing during the check as ---
//--- all we really care about is whether or not it IS different. ---

  stat = 0;
  if (dataTimeMS[cpy] != dataTimeMS[src])  do {
    if (getLock(srcLockID, 0) <= 0)  { stat = -1;  break; }  // try to get src buf lock

    numBytes = numDataPoints[src]<<2;

    fmemcpy(cpyBuf, srcBuf, numBytes);
    numDataPoints[cpy] = numDataPoints[src];
    dataTimeSecs[cpy] = dataTimeSecs[src];
    dataTimeMS[cpy] = dataTimeMS[src];

    releaseLock(srcLockID);  // release lock on src buffer

    pv->pvInfo.size = numBytes;
    pv->pvInfo.numElems = numDataPoints[src];

    pv->newEvents |= DBE_VALUE;

    if (ShowDebugMsgs)
      printf("Updated copy of PV val: [%ls]\r\n", pv->pvInfo.name);
  } while (0);

  releaseLock(cpyLockID);

  return stat;
}


//-----------------------------------------------------------------------------
//  Attempt to obtain or release a read or write lock on a waveform PV.
//
//  Returns the depth of the lock or < 0 if not successsful.
//
//  For lockMode LM_TryLock, a return value < 0 and errno == EAGAIN means
//  it could not get the lock w/o blocking.
//-----------------------------------------------------------------------------
int _nearcall PV_LockWave(PV_Lock_Params)
{
  uint  src, cpy, lockID;


//tdebug
//  printf("\r\nLock PV val: [%ls]  mode:[%d]  type:[%d]\r\n",
//         pv->pvInfo.name, lockMode, lockType);

  if ((src = WF_IDX(pv->val) - 1) >= NUM_WAVEFORMS*2)  return -1;
  cpy = src + 1;

  switch (lockType)  {
    case LT_READ:  lockID = WAVE_DATA_0 + cpy;  break;
    case LT_WRITE:  lockID = WAVE_DATA_0 + src;  break;
    default:  return -1;
  }

  switch (lockMode)  {
    case LM_Unlock:    return releaseLock(lockID);
    case LM_WaitLock:  return getLock(lockID, 5000);
    case LM_TryLock:   return getLock(lockID, 0);
  }

  return -1;
}


//-----------------------------------------------------------------------------


#define SharedFuncs
#include <sharedAppl.c>



//=============================================================================
//  Misc Application-specific functions
//=============================================================================


//-----------------------------------------------------------------------------
//  phaseDiff:  compute difference between 2 phases, returning a value between
//  -180.0 and +180.0
//-----------------------------------------------------------------------------
float  phaseDiff(float ph1, float ph2)
{
  float diff = ph1 - ph2;

  if (diff < -180.0)
    diff += 360.0;
  else
    if (diff > 180.0)  diff -= 360.0;

  return diff;
}


/*
//-----------------------------------------------------------------------------
//  "fptofl": Converts a decimal number in fixed point format to floating point
//-----------------------------------------------------------------------------
float fptofl(long value)
{
  float out;
  bool neg = value < 0;
  ulong *p = (ulong *)&value;

  if (neg)  value *= -1;

//finish
  out = (*p) & 0x0000FFFFL;
  out /= 65536.0;
  out += (float) ((*p) >> 16);

  if (neg)  out *= -1.0;

  return out;
}

//-----------------------------------------------------------------------------
//  "fltofp": Converts a decimal number in floating point format to fixed point
//-----------------------------------------------------------------------------
long fltofp(float value)
{
  long fixed;
  bool neg = (value < 0.0);

  if (neg)  value *= -1.0;

//finish
  fixed = (long) value;
  value -= (float) fixed;
  fixed <<= 16;
  fixed += (long) (65536.0 * value);

  if (neg)  fixed *= -1;

  return fixed;
}
*/

//=============================================================================
typedef struct {
  fd_t   fromfd;
  bool   eof;
}  ReadState;

#define RDSTATE ((ReadState *)rdState)

typedef struct {
  uint   progNum;     // 0 - 3
  uint   blkNum;      // 0 - 15
  ulong  ttlWritten;  // ttl bytes written to DSP mem for this program
  ulong  blkWritten;  // # bytes in current block written so far
  bool   showProgress;
}  WriteState;

#define WRSTATE ((WriteState *)wrState)

#define  DSP_BLK_SIZE  (0x10000L)


//-----------------------------------------------------------------------------
//  Read the sepcified # of bytes from a file descriptor.  If the end of the
//  file is reached, bytes with 0 value are returned instead.
//
//  Parameters:
//    *rdState:  Pointer to ReadState structure that contains the following:
//                 fromfd:  the file descriptor to read from
//                 eof:  flag to indicate that the End-Of-File was reached
//    *buf:      pointer to buffer to read the data in to
//    count:     # of bytes to read (or fill with zeros)
//
//  This function is designed to be called by the copyBytes() function which
//  provides a generic means of managing the transfer of a stream of bytes.
//
//  On return:
//    rdState values updated as needed
//    Returns the # of bytes successfully read or set to 0
//-----------------------------------------------------------------------------
int _nearcall readFromFile(void *rdState, char *buf, int count)
{
  int  stat;

  if (RDSTATE->eof)  { memset(buf, 0xff, count);  return count; }

  if ((stat = read(RDSTATE->fromfd, buf, count)) < 0)  return stat;

  if (stat < count)  { memset(buf+stat, 0, count-stat);  RDSTATE->eof = true; }

  return count;
}

//-----------------------------------------------------------------------------
//  initialize various default settings
//-----------------------------------------------------------------------------
void initDefaults(void)
{
  //finish
}


//=============================================================================
//  Event-Handler functions called by the appl task to process event messages
//=============================================================================

//-----------------------------------------------------------------------------
//  Event-Handler for a request to ...
//-----------------------------------------------------------------------------
/*
void EVENT_MyEvent(EventMsg *event)
{
  uint  idx;
  float  setpoint;
  char  temp[16];


  idx = event->chanNum;

  if (error)  {
    fprintf(DiagLog, "error msg, chan:\n", idx+1);
    return;
  }

  setpoint = *((float *)event->data);

  fprintf(DiagLog, "New SetPoint for PS%d: %s\n",
     idx+1, fmtFloat(setpoint, 8,3, temp) );

  *((float *)((&iSet_1_RegInfo + idx)->regVal)) = setpoint;

  anaOut[idx] = ISETtoVolts(idx, setpoint);
}
*/



//=============================================================================
//  Task functions to handle application-specific commands
//=============================================================================


//-----------------------------------------------------------------------------
//  example func to process a command entered via a user shell
//-----------------------------------------------------------------------------
/*
void _task miscTest(int argc, char **argv)
{
  int  idx;
  bool  validCmd;
  // used by EVENT macro
  fd_t  eventsfd = -1;
  EventMsg  eventMsg;


  validCmd = false;

  for (;;)  {
    if (argc < 3)  break;
    if (! isdigit(argv[1][0]))  break;

    idx = atoi(argv[1]);
    if ((idx < 1) OR (idx > NumChan))  {
      printf("Invalid chan #\n");  Exit(-2);  }

    --idx;  validCmd = true;

    if (StriEq(argv[2], "on"))     { EVENT(TurnOn,  idx, NULL, 0);  break; }
    if (StriEq(argv[2], "off"))    { EVENT(TurnOff, idx, NULL, 0);  break; }
    if (StriEq(argv[2], "reset"))  { EVENT(Reset,   idx, NULL, 0);  break; }

    validCmd = false;  break;
  }

  if (! validCmd)  { printf("Usage:  misc # on|off|reset\r\n");  Exit(-1); }

  Exit(0);
}
*/


//-----------------------------------------------------------------------------
//  auto-calibrate the fwd, rev, or cav atten value by cycling through
//  the range of atten values for a given pair of V-peak and phase values.
//-----------------------------------------------------------------------------
void _task autocal(int argc, char **argv)
{
  U16  u, curAtten;
  F32  vPeak, phase;
  byte  flags[3];
  RegInfo  *reg[3];


  if (showHelp(4, argc, argv))  {
    printf("Usage:  autocal fwd|rev|cav <V-peak> <Phase>\r\n");
    Exit(-1);
  }

  reg[0] = NULL;

  if (StriEq(argv[1], "fwd"))  {
    curAtten = fwdAtten;
    reg[0] = &fwdAtten_RegInfo;
    reg[1] = &fwdMagCalib_RegInfo;
    reg[2] = &fwdPhaseCalib_RegInfo;
  }  else
  if (StriEq(argv[1], "rev"))  {
    curAtten = revAtten;
    reg[0] = &revAtten_RegInfo;
    reg[1] = &revMagCalib_RegInfo;
    reg[2] = &revPhaseCalib_RegInfo;
  }  else
  if (StriEq(argv[1], "cav"))  {
    curAtten = cavAtten;
    reg[0] = &cavAtten_RegInfo;
    reg[1] = &cavMagCalib_RegInfo;
    reg[2] = &cavPhaseCalib_RegInfo;
  }

  if (! reg[0])  {
    printf("Invalid argument: %.32s\r\n", argv[1]);  Exit(-1); }

  vPeak = atof(argv[2]);  phase = atof(argv[3]);

//--- save the flags value for each register so we can restore it after ---
//--- temporarily changing it to avoid seeing the normal log messages on ---
//--- the terminal screen ---
  for (u=0; u<3; ++u)  flags[u] = reg[u]->flags;

//--- cycle the specified attenuation from 0 to 63, waiting 1 second after ---
//--- each change and then telling the DSP the v-peak and phase values ---
//--- that it should associate with the current conditions ---

  for (u=0; u<64; ++u)  {
    printf("%3d\r", u);  fflush(stdout);

    reg[0]->flags &= ~(byte)LogChanges;
    setRegU16(reg[0], 0, u);
    reg[0]->flags = flags[0];

    sleepMS(1000);

    reg[1]->flags &= ~(byte)LogChanges;
    setRegF32(reg[1], 0, vPeak);
    reg[1]->flags = flags[1];

    reg[2]->flags &= ~(byte)LogChanges;
    setRegF32(reg[2], 0, phase);
    reg[2]->flags = flags[2];

    CheckETX();
  }
  printf("\r\n");

  reg[0]->flags &= ~(byte)LogChanges;
  setRegU16(reg[0], 0, curAtten);
  reg[0]->flags = flags[0];


  Exit(0);
}


//-----------------------------------------------------------------------------
//  "cmd": set various DSP states
//-----------------------------------------------------------------------------
void _task cmd(int argc, char **argv)
{
  ulong  bits = 0;


  if (showHelp(3, argc, argv))  {
    printf("Usage:  cmd amplCtl on|off\r\n"
//           "            phaseCtl off|sel|auto\r\n"
//           "            rfPulse on|off\r\n"
           "            tunerCtl on|off\r\n"
          );
    Exit(-1);
  }

//finish
  do {
    if (StriEq(argv[1], "amplCtl"))  {
      if (StriEq(argv[2], "on"))  { bits = CMD_AMPL_CTL_ON;  break; }
      if (StriEq(argv[2], "off"))  { bits = CMD_AMPL_CTL_OFF;  break; }
    }

/*  //finish
    if (StriEq(argv[1], "phaseCtl"))  {
      if (StriEq(argv[2], "off"))  { bits = CMD_PHASE_CTL_OFF;  break; }
      if (StriEq(argv[2], "sel"))  { bits = CMD_PHASE_CTL_SEL;  break; }
      if (StriEq(argv[2], "auto"))  { bits = CMD_PHASE_CTL_AUTO;  break; }
    }

    if (StriEq(argv[1], "rfPulse"))  {
      if (StriEq(argv[2], "on"))  { bits = CMD_RF_PULSING_ON;  break; }
      if (StriEq(argv[2], "off"))  { bits = CMD_RF_PULSING_OFF;  break; }
    }
*/
    if (StriEq(argv[1], "tunerCtl"))  {
      if (StriEq(argv[2], "on"))  { bits = CMD_TUNER_CTL_ON;  break; }
      if (StriEq(argv[2], "off"))  { bits = CMD_TUNER_CTL_OFF;  break; }
    }

  } while (false);


  if (bits)
    SetMBReg(setDSPState, 0, U32, bits);
  else
    printf("Invalid parameters\r\n");

  Exit(0);
}



//-----------------------------------------------------------------------------
//  "defaults": initialize FPGA settings to default values
//-----------------------------------------------------------------------------
void _task defaults(int argc, char **argv)
{
  initDefaults();  Exit(0);
}



//-----------------------------------------------------------------------------
//  Reset the DSP and FPGA, causing them to reload their firmware from EEPROMs
//-----------------------------------------------------------------------------
void _task reset_DSP_FPGA(int argc, char **argv)
{
  int  stat;

  if (getLock(APPL_LOCK, 10000) <= 0)  Exit(-1);

  stop_DSP_FPGA();  stat = restart_DSP_FPGA(10000);

  releaseLock(APPL_LOCK);

  Exit(stat);
}


//-----------------------------------------------------------------------------
//  "setdspfirm": load a new DSP firmware file in to the serial EEPROM
//-----------------------------------------------------------------------------
void _task setDSPfirm(int argc, char **argv)
{
  int  i, fromfd;
  ulong  limit = 0x1F8000L;
  bool  verify = false;
  char  *fileName;


  if (showHelp(1, argc, argv))  {
    printf("\nusage:  setdspfirm [-v] [fileName] [-l<limit>]\r\n");
    Exit(-1);
  }

  fileName = DSP_Prog_File;
  for (i=1; i<argc; ++i)  {
    if (argv[i][0] != '-')  { fileName = argv[i];  continue; }
    if (StrEq(argv[i], "-v"))  verify = true;
    if (strncmp(argv[i], "-l", 2) == 0)  limit = ATOL(argv[i]+2);
  }

  if ((fromfd = open(fileName, O_RDONLY, 0)) < 0)  {
    perror(fileName);  Exit(-1); }

  printf("\r\nDownloading %.40s to EEPROM\r\n", fileName);

  updateEEPROM(DSP_EEPROM, fromfd, limit, true, verify);

  close(fromfd);  Exit(0);
}

//-----------------------------------------------------------------------------
//  "fpgaread": command to read one or more 8-bit values from the FPGA
//-----------------------------------------------------------------------------
void _task fpgaRead(int argc, char **argv)
{
  uint  u, regNum, count, maxCount;
  int  regVal;


  if (showHelp(3, argc, argv))  {
    printf("Usage:  fpgaread <reg> <count>\r\n"
           "                 auto on|off\r\n");
    Exit(-1);
  }

  if (StriEq(argv[1], "auto"))  {
    autoReadFPGA = StriEq(argv[2], "on");  Exit(0); }


  regNum = ATOU(argv[1]) & 0x3F;
  count = ATOU(argv[2]);

  maxCount = 64 - regNum;  if (count > maxCount)  count = maxCount;

  if (getLock(APPL_LOCK, 5000) <= 0)  { perror("getLock(...)");  Exit(-1); }

  for (u=0; u<count; ++u,++regNum)  {
    regVal = readFPGA(regNum);
    printf("  0x%02X:  ", regNum);
    if (regVal < 0)
      printf("<error>\r\n");
    else
      printf("0x%02X\r\n", regVal);
  }

  releaseLock(APPL_LOCK);

  Exit(0);
}


//-----------------------------------------------------------------------------
//  "fpgawrite": command to write a value to the FPGA
//-----------------------------------------------------------------------------
void _task fpgaWrite(int argc, char **argv)
{
  uint  regNum, setVal;


  if (showHelp(3, argc, argv))  {
    printf("Usage:  fpgawrite <reg> <value>\r\n"
           "                  auto on/off\r\n");
    Exit(-1);
  }

  if (StriEq(argv[1], "auto"))  {
    autoWriteFPGA = StriEq(argv[2], "on");  Exit(0); }

  regNum = ATOU(argv[1]) & 0x3F;
  setVal = ATOU(argv[2]) & 0xFF;

  printf("  write 0x%02X to 0x%02X\r\n", setVal, regNum);

  if (writeFPGA(regNum, setVal))  { perror("writeFPGA()");  Exit(-1); }

  Exit(0);
}


//-----------------------------------------------------------------------------
//  "setfpgafirm": load a new FPGA firmware file in to the serial EEPROM
//-----------------------------------------------------------------------------
void _task setFPGAfirm(int argc, char **argv)
{
  int  i, fromfd;
  ulong  limit = 0;
  bool  verify = false;
  char  *fileName;


  if (showHelp(1, argc, argv))  {
    printf("\nusage:  setfpgafirm [-v] [fileName] [-l<limit>]\r\n");
    Exit(-1);
  }

  fileName = FPGA_Prog_File;
  for (i=1; i<argc; ++i)  {
    if (argv[i][0] != '-')  { fileName = argv[i];  continue; }
    if (StrEq(argv[i], "-v"))  verify = true;
    if (strncmp(argv[i], "-l", 2) == 0)  limit = ATOL(argv[i]+2);
  }

  if ((fromfd = open(fileName, O_RDONLY, 0)) < 0)  {
    perror(fileName);  Exit(-1); }

  printf("\r\nDownloading %.40s to EEPROM\r\n", fileName);

  updateEEPROM(FPGA_EEPROM, fromfd, limit, true, verify);

  close(fromfd);  Exit(0);
}


//=============================================================================
//  Generic structure to handle application-specific presentation of the
//  digital and analog input/output values.  Code to handle additional
//  formatting that cannot be handled in this generic manner should be
//  placed in the showState() task routine.
//=============================================================================

//-----------------------------------------------------------------------------
_nearcall char *fmtPhaseMode(DispElem far *elem, char *buf)
{
  return (char *) PhaseModeDesc[ELEM_VAL(uint)];
}

//-----------------------------------------------------------------------------
_nearcall char *fmtWarming(DispElem far *elem, char *buf)
{
  return (ELEM_VAL(bool) ? " WARM  " : "WARMING");
}

//-----------------------------------------------------------------------------
_nearcall char *fmtRamping(DispElem far *elem, char *buf)
{
  return (ELEM_VAL(bool) ? "RAMPING" : "STEADY ");
}

//-----------------------------------------------------------------------------
_nearcall char *fmtLocked(DispElem far *elem, char *buf)
{
  return (ELEM_VAL(bool) ? "LOCKED  " : "UNLOCKED");
}
//-----------------------------------------------------------------------------
  const char far DevStateStr[] = {
"--INPUTS--  Max  Ampl   Phase Atten   -------- RF State --------   ---DSP---\r\n"
"Forward:  ##.## ##.## -###.##  ##     WarmLevel:##.##      ###     Load:###.#%%\r\n"
"Reverse:  ##.## ##.## -###.##  ##     WarmTime: ##.##    #######   Temp: ##.#C\r\n"
"Cavity:   ##.## ##.## -###.##  ##     RampTime: ##.##    #######   Prog ID: #\r\n"
"Reference:      ##.## -###.##                                      Errs:#####\r\n"
"Setpoints:##.## ##.## -###.##         CavLocked: ########\r\n"
"\r\n"
"                                -TUNE-  -Ampl-  -Phase-   Pulsing: ###\r\n"
"                                State:  ###### ########   Pulse Rate:###.##\r\n"
"                                Pulsed:   #.##  -###.##   Pulse Duty:###.##\r\n"
"                            \r\n"
"                                  --RF-- -Tuner-  ---- Tuner ----\r\n"
"                                P: #.###  #.###   State:   ######     \r\n"
"                                I: #.###  #.###   Piezo:   -#####     FFT:###%%\r\n"
"                                D: #.###  #.###   Stepper: -#####\r\n"
"                            \r\n"
"                            \r\n"
"                            \r\n"
"\r\n"
  };

const DispElem  far dispElems[] = {
  { 10, 3,  VFPV(maxFwdAmpl),    BLUE,  fmtF32, 5,2 },
  { 10, 4,  VFPV(maxRevAmpl),    BLUE,  fmtF32, 5,2 },
  { 10, 5,  VFPV(maxCavAmpl),    BLUE,  fmtF32, 5,2 },

  { 10, 7,  VFPV(maxOutAmpl), BLUE,  fmtF32, 5,2 },

  { 16, 3,  VFPV(fwdAmpl),  BLUE,  fmtF32, 5,2 },
  { 16, 4,  VFPV(revAmpl),  BLUE,  fmtF32, 5,2 },
  { 16, 5,  VFPV(cavAmpl),  BLUE,  fmtF32, 5,2 },
  { 16, 6,  VFPV(ref0Ampl), BLUE,  fmtF32, 5,2 },
  { 16, 7,  VFPV(amplSet),  BLUE,  fmtF32, 5,2 },

  { 22, 3,  VFPV(fwdPhase),  BLUE,  fmtF32, 7,2 },
  { 22, 4,  VFPV(revPhase),  BLUE,  fmtF32, 7,2 },
  { 22, 5,  VFPV(cavPhase),  BLUE,  fmtF32, 7,2 },
  { 22, 6,  VFPV(ref0Phase), BLUE,  fmtF32, 7,2 },
  { 22, 7,  VFPV(phaseSet),  BLUE,  fmtF32, 7,2 },

  { 31, 3,  VFPV(fwdAtten),   BLUE,  fmtU16, 2,2 },
  { 31, 4,  VFPV(revAtten),   BLUE,  fmtU16, 2,2 },
  { 31, 5,  VFPV(cavAtten),   BLUE,  fmtU16, 2,2 },

  { 59, 3,  VFPV(rfIsOn),     BLUE,  fmtBool, 3, FMT_ON_OFF },
  { 57, 4,  VFPV(rfAtWarm),   BLUE,  fmtWarming, 7,7 },
  { 57, 5,  VFPV(rfRamping),  BLUE,  fmtRamping, 7,7 },

  { 49, 7,  VFPV(cavLocked),  BLUE,  fmtLocked, 8,8 },

  { 48, 3,  VFPV(warmLevel),  BLUE,  fmtF32, 5,2 },
  { 48, 4,  VFPV(warmTime),   BLUE,  fmtF32, 5,2 },
  { 48, 5,  VFPV(rampTime),   BLUE,  fmtF32, 5,2 },

  { 72, 3,  VFPV(dspLoad),     BLUE,  fmtF32, 5,1 },
  { 72, 4,  VFPV(pcbTemp),     BLUE,  fmtF32, 5,1 },
  { 76, 5,  VFPV(curProgNum),  BLUE,  fmtU16, 1,1 },
  { 72, 6,  VFPV(romProgErrs), BLUE,  fmtU16, 5,5 },

  { 40,10,  VFPV(amplCtlisON),  BLUE,  fmtBool, 6, FMT_CLOSED_OPEN },
  { 42,11,  VFPV(pulseAmplSet), BLUE,  fmtF32, 4,2 },

  { 47,10,  VFPV(phaseMode),     BLUE,  fmtPhaseMode, 8,8 },
  { 48,11,  VFPV(pulsePhaseSet), BLUE,  fmtF32, 7,2 },

  { 67, 9,  VFPV(rfPulsing),  BLUE,  fmtBool, 3, FMT_ON_OFF },
  { 69,10,  VFPV(pulseFreq),  BLUE,  fmtF32, 6,2 },
  { 69,11,  VFPV(pulseDuty),  BLUE,  fmtF32, 6,2 },

  { 35,14,  VFPV(rf_P),  BLUE,  fmtF32, 5,3 },
  { 35,15,  VFPV(rf_I),  BLUE,  fmtF32, 5,3 },
  { 35,16,  VFPV(rf_D),  BLUE,  fmtF32, 5,3 },

  { 42,14,  VFPV(tuner_P),  BLUE,  fmtF32, 5,3 },
  { 42,15,  VFPV(tuner_I),  BLUE,  fmtF32, 5,3 },
  { 42,16,  VFPV(tuner_D),  BLUE,  fmtF32, 5,3 },

  { 59,14,  VFPV(tunerCtlisON),    BLUE,  fmtBool, 6, FMT_CLOSED_OPEN },
  { 59,15,  VFPV(piezoReadPos),    BLUE,  fmtS16, 6,6 },
  { 59,16,  VFPV(stepperReadPos),  BLUE,  fmtS16, 6,6 },

  { 74,15,  VFPV(percentFFT),   BLUE,  fmtU16, 3,3 },

  { 0,0, NULL, NULL, NULL, 0,0 }
};


//--- ATTR_NAME macros currently defined in fcntl.h ---
const DigDispElem far digInElems[] = {
//{ xx,yy,  BIT_MASK,          "0STR"  ,ATTR_NAME,  "1STR"  ,ATTR_NAME },

  { 0,0, 0, NULL,NULL, NULL,NULL }
};

const DigDispElem far digOutElems[] = {
//{ xx,yy,  BIT_MASK,      "0STR"  ,ATTR_NAME,  "1STR" ,ATTR_NAME },

  { 0,0,  0,  NULL,NULL,  NULL,NULL }
};


//-----------------------------------------------------------------------------
//  "state":  Show continuously updated state of the system
//-----------------------------------------------------------------------------
void _task showState(int argc, char **argv)
{
  bool  blinkOff;
  char  temp[128];


  printf(clsCmd);  blinkOff = false;

  CursorTo(0, 2);  printf(DevStateStr);

  for (;;)  {
    CursorTo(0, 0);  blinkOff = !blinkOff;

    stdHeader();  displayElems((DispElem far *)dispElems, blinkOff, temp);

    displayDigElems(digitalIn, (DigDispElem far *)digInElems, blinkOff, temp);
    displayDigElems(digitalOut, (DigDispElem far *)digOutElems, blinkOff, temp);

    CursorTo(0,22);  fflush(stdout);  CheckETX();  sleepMS(500);
  }

}

//-----------------------------------------------------------------------------
//  return a string with a human-readable format for the specified DSP value
//-----------------------------------------------------------------------------
char *dspVal2Str(ulong dspVal, uchar dspType, char *buf)
{
  float  fval;

  switch (dspType)  {
    case DT_S32:
      sprintf(buf, "%12ld", (S32)dspVal);  break;

    case DT_U32:
      sprintf(buf, "%12lu", (U32)dspVal);  break;

    case DT_F32:
      Engineering(*(float *)&dspVal, 6, buf);  break;

    case RT_PHASE:
      fval = (float)dspVal * PhaseConvFactor;
      Engineering(fval, 6, buf);  break;
  }


  return buf;
}

//-----------------------------------------------------------------------------
//  Check the list of known DSP registers against the specified register ID.
//  If it matches one of the known registers, then format the provided DSP
//  value based on the expected data format.  Otherwise, format the value as
//  a float.
//-----------------------------------------------------------------------------
char *fmtDSPval(ulong dspVal, ulong regNum, char *buf)
{
  uchar  dspType = DT_F32;
  uint  i, j;
  const RD_DSP_REG far *rdReg;
  const WR_DSP_REG far *wrReg;
  const char far *name = NULL;


  for (i=0,rdReg=rdDSPregs; i<NUM_RD_DSP_REGS; ++i,++rdReg)  {
    if (rdReg->id != regNum)  continue;
    name = rdReg->mbReg->name;  dspType = rdReg->type;  break;
  }

  if (! name)  for (i=0,wrReg=wrDSPregs; i<NUM_WR_DSP_REGS; ++i,++wrReg)  {
    if (wrReg->id != regNum)  continue;
    name = wrReg->mbReg->name;  dspType = wrReg->type;  break;
  }

  if (name)  {
    strcpyfn(buf, name);
    for (j=strlen(buf); j<16; ++j)  buf[j] = ' ';
    buf[15] = ' ';
    dspVal2Str(dspVal, dspType, buf+16);
  }
  else  {
    sprintf(buf, "%14ld  ", dspVal);
    dspVal2Str(dspVal, DT_F32, buf+16);
  }

  return buf;
}

//-----------------------------------------------------------------------------
//  "dspcalib": calibrate gain/offset for conversion to/from certain DSP
//  values and Engineering units.
//
//  The value to calibrate is specified by giving the name of a ModBus
//  register value.  If the register is referenced in one of the DSP register
//  tables, and there is a gain and offset value listed for it, then the
//  gain and offset can be calibrated by running the device at 2 different
//  states and providing the actual (measured) value for each state.
//
//  NOTE that some calibrations apply to more than one register, so changing
//  them for one can affect other readings/settings.
//-----------------------------------------------------------------------------
void _task dspcalib(int argc, char **argv)
{
  bool  foundReg;
  uint  ui;
  float  *gain, *offset;
  float  mbVal, egu[2], raw[2];
  const RD_DSP_REG far *rdReg;
  const WR_DSP_REG far *wrReg;
  RegInfo *mbReg;
  char  temp[64];


  if (showHelp(2, argc, argv))  {
    printf("\r\n Usage:  dspcalib <regName>\r\n");  Exit(-1); }


  foundReg = false;

//--- check RO DSP register list ---
  for (rdReg=rdDSPregs,ui=0; ui<NUM_RD_DSP_REGS; ++ui,++rdReg)  {
    mbReg = rdReg->mbReg;  strcpyfn(temp, mbReg->name);
    if (stricmp(temp, argv[1]))  continue;
    gain = rdReg->gain;  offset = rdReg->offset;
    foundReg = true;  break;
  }

//--- check RW DSP register list ---
  if (!foundReg)  for (wrReg=wrDSPregs,ui=0; ui<NUM_WR_DSP_REGS; ++ui,++wrReg) {
    mbReg = wrReg->mbReg;  strcpyfn(temp, mbReg->name);
    if (stricmp(temp, argv[1]))  continue;
    gain = wrReg->gain;  offset = wrReg->offset;
    foundReg = true;  break;
  }

  if (!foundReg)  {
    printf("\r\n Register not related to a DSP value\r\n");  Exit(-2); }

  if (!gain AND !offset)  {
    printf("\r\n Value does not require conversion\r\n");  Exit(-3); }


  printf("\r\n Existing:");
  if (gain)  printf("  Gain: %s", Engineering(*gain, 6, temp));
  if (offset)  printf("  Offset: %s", Engineering(*offset, 6, temp));
  printf("\r\n\n");

//--- get real EGU and DSP values for two test points ---
  for (ui=0; ui<2; ++ui)  {
    printf(" Test point %u:  Enter measured/actual value for %ls: ",
           ui+1, mbReg->name);
    GETS(temp);  if (!*temp)  Exit(-4);
    egu[ui] = atof(temp);  mbVal = mb2float(mbReg->regVal, mbReg);
    if (gain)  if (*gain != 0.0)  mbVal = mbVal / *gain;
    if (offset)  mbVal += *offset;
    raw[ui] = mbVal;
  }

  if ((raw[0] == raw[1]) OR (egu[0] == egu[1]))  {
    printf("\r\n Unusable calibration data\r\n");  Exit(-4); }

  KLOCK();
  if (gain)  *gain = (egu[1] - egu[0]) / (raw[1] - raw[0]);
  if (offset)  *offset = raw[0] - (gain ? egu[0]/(*gain) : raw[1]);
  KUNLOCK();

  printf("\r\n New:");
  if (gain)  printf("  Gain: %s", Engineering(*gain, 6,temp));
  if (offset)  printf("  Offset: %s", Engineering(*offset, 6, temp));
  printf("\r\n\n");

  Exit(0);
}

//-----------------------------------------------------------------------------
//  "dspio": command to enable/disable normal communication with the DSP
//-----------------------------------------------------------------------------
void _task dspio(int argc, char **argv)
{
  bool  val;


  if (showHelp(3, argc, argv))  {
    printf("Usage:  dspio <read|write|filler|waves> <on|off>\r\n");  Exit(-1); }

  if (StriEq("on", argv[2]))  val = true;  else
  if (StriEq("off", argv[2])) val = false;  else  {
    printf("invalid option\r\n");  Exit(-1);  }

  if (StriEq("read", argv[1]))    { autoReadDSP = val;   Exit(0); }
  if (StriEq("write", argv[1]))   { autoWriteDSP = val;  Exit(0); }
  if (StriEq("filler", argv[1]))  { showFiller = val;    Exit(0); }
  if (StriEq("waves", argv[1]))   { readWaves = val;     Exit(0); }

  printf("invalid option\r\n");  Exit(-1);
}


//-----------------------------------------------------------------------------
//  Function to read the data from one of the DSP's data buffers that contains
//  new waveform data for the Phase and Amplitude values surrounding a recent
//  change to the Phase and/or Amplitude setpoints.
//-----------------------------------------------------------------------------
int readWaveform(ulong bufID)
{
  static uint readCount;  //test
  ulong far  *buf;
  uint  w, lockID;
  U32  u32;
  float  fval;
  ulong  regNum, ackBits, val, count, countReg;


  ackBits = setDSPState;

  switch (bufID)  {
    case Rd_L_Buf:
      w = WAVE_AP_L_SRC;  ackBits |= CMD_L_DATA_READ;  countReg = 0;
      break;

    case Rd_T_Buf:
      w = WAVE_AP_T_SRC;  ackBits |= CMD_T_DATA_READ;  countReg = 0;
      break;

    case AmplFFT:
      w = WAVE_FFT_SRC;  ackBits |= CMD_FFT_DATA_READ;  countReg = SizeFFT;
      break;

//    case PhaseCTF:
//      if (hdwType == HT_PHASE_CTLR)  return -1;
//      w = WAVE_CTF_P_SRC;  ackBits = 0;  countReg = SizeCTF;
//      break;

//    case AmplCTF:
//      if (hdwType == HT_PHASE_CTLR)  return -1;
//      w = WAVE_CTF_A_SRC;  ackBits = 0;  countReg = SizeCTF;
//      break;

    default:
      return -1;
  }

  if (msSince_(lastWaveReadTime[w]) < 500)  return - 1;

  if (msSince_(lastReadPlotTime) < 200)  return -1;

  buf = waveformData[w];  lockID = WAVE_DATA_0 + w;

  regNum = bufID;


  if (getLock(lockID, 0) <= 0)  return -1;

//  if (ShowDebugMsgs)  fprintf(DiagLog, "readWaveform(w=%d)\n", w);  //tdebug

  if (readWaves)  {

  //--- get the # of long words in the waveform ---
    if (countReg)  {
      if (hdwType == HT_PHASE_CTLR)
        readDSPReg(countReg, &count);
      else  {
        switch (bufID)  {
          case Rd_L_Buf:  case Rd_T_Buf:  count = 1024;  break;
          case AmplFFT:                   count = 1001;  break;
//          case PhaseCTF:  case AmplCTF:   count = 1000;  break;
        }
      }
      if (count > MaxDataPoints)  count = MaxDataPoints;
      numDataPoints[w] = (uint) count;
    }

  //--- read specified DSP buffer with waveform data ---
    readPlot(regNum, buf, numDataPoints[w]);  lastReadPlotTime = GetMS_();

    if (hdwType != HT_PHASE_CTLR)  {  // generate a simple waveform
      readCount += 5;  val = readCount;
      for (uint u=0; u<numDataPoints[w]; ++u)  {
        val = ++val & 0xff;

        switch (bufID)  {
          case Rd_L_Buf:  // 2 16-bit LSB 1st values for Ampl and Phase
          case Rd_T_Buf:  // ampl: raw ADC rdg, phase: int mult of 360/2^16
            u32 = (val << 16) | (~val & 0xff);  buf[u] = htonl(u32);  break;

          case AmplFFT:  // Amplitude FFT (32-bit IEEE float)
            fval = (float) val / 256.0;  buf[u] = htonl(*((ulong *)&fval));  break;

//          case PhaseCTF:  // Phase CTF (expr as int mult of 360/2^32)
//            u32 = ~(val << 24);  buf[u] = htonl(u32);  break;

//          case AmplCTF:  // Ampl CTF (32-bit IEEE float) (?)
//  //          u32 = val << 24;  buf[u] = htonl(u32);  break;
//            fval = (float) val / 256.0;  buf[u] = htonl(*((ulong *)&fval));  break;
        }
      }
    }

  //--- let DSP know it is OK to write to the buffer again ---
    if (ackBits)  writeDSPRegQuick(SetDSPState, ackBits);

    _EnterCritical_();
    dataTimeSecs[w] = time(NULL);  dataTimeMS[w] = GetMS();
    _ExitCritical_();

    lastWaveReadTime[w] = GetMS_();
  }

  releaseLock(lockID);

  return 0;
}



//=============================================================================
//  Copy the "cur" state from last time to "prev", and update "cur"
//=============================================================================
//  Just before this function is called, readings were taken of the current
//  state of the hardware, and any changes to the output states specified by
//  the last pass were applied (just after taking the readings).
//
//  This func first copies "cur" to "prev" and then "next" to "cur" which does
//  2 things for us:
//
//    - We have a copy of the controller's logical state from the last pass
//      (so we can determine what has changed)
//
//    - We have a copy of values set by other tasks indicating events/requests
//      that happend/were made since the beginning of the last pass
//
//  We then reset the values (as needed) that are set by other tasks to
//  indicate events and requests, and update the current state based on the
//  readings just taken and the requests and events communicated by other
//  tasks.
//
//  All of this is in preperation for the updateNextState() function, which
//  implements the main "decision making" portion of the application logic,
//  determining what the next state should be based on the latest conditions
//  (as determined by the updateCurState() function).
//
//=============================================================================
void updateCurState(void)
{
  uint  x;
  ulong  maxWait1, maxWait2;
  static uint  check1st = 0;


  if (current != applTask)  return;

//-----------------------------------------------------------------
// The following conditions exist at this point:
// - The inputs have just been read
// - The outputs were just set to values computed during the last pass
// - "cur" contains the state the inputs were in at the START of
//   of the previous pass, and the state the outputs were in DURING
//   the previous pass
// - "next" contains the values that the outputs were just set
//   to and any "input" values that were changed by other tasks
//------------------------------------------------------------------

// Now we update the values in "Next" that are based solely on the
// state of the hardware inputs we just read and values maintained by
// by other tasks (basically, we deal with READINGS here: Values that
// we need but are not written to "Next" by another task)

//  Next.processTime = GetMS();  curTime = time(NULL);

//  msSincePrev = msSince(Cur.processTime);

  Next.digitalIn = digitalIn;


// Note that a copy is done (rather than just switching pointer values) to
// insure all values are set in both the source and destination.  If we
// simply kept pointers and swapped the pointer values, then EVERY value
// would have to be updated independently by the code, which is a pain to
// program, adds overhead and code, and is a VERY easy way to introduce odd
// bugs.  Overall, copying everything here is actually faster overall,
// simpler, and much safer.

  Prev = Cur;  Cur = Next;

//--- At this point the Cur copy contains what the inputs were at the start ---
//--- of this pass, and the Prev copy contains the output values from the ---
//--- the previous pass (i.e. the output values that resulted in the input ---
//--- values in Cur) ---


//--------------------------------------------------------------------
// Next we update the rest of the "cur" values needed by the control
// logic (the ones that do NOT come from "next" values set by other
// tasks)
//
// NOTES:
//  - This portion of the logic has order-of-execution dependencies,
//    so be sure that all values referenced in an expression are ones
//    set before they are used.
//
//  - References to "prev" values are always safe since they are just
//    a copy of values from the last pass
//--------------------------------------------------------------------

  //--- log msgs for bit chgs only if not starving other processes for cycles ---
  if (applLoopTime < 8)  logDigIOChgs(false);


//APPLCODE
//ignore for now:  if (!fpgaInitialized)  return;

  if (!autoReadDSP AND !autoWriteDSP)  return;

  if (getLockDSP(10) <= 0)  return;


//--- Amplitude and phase waveform data is read as a single stream of ---
//--- 32-bit values, where 16 of each 32 bits is an amplitude value and ---
//--- the other 16 is a phase value.  When operating in "pulsed" mode ---
//--- (where the Amplitude and/or Phase setpoints are being toggled ---
//--- between 2 values), the waveform data is updated based on when ---
//--- the change happens, and it provides data surrounding the "rising" ---
//--- edge and the "falling" edge.  When not operating in pulsed mode, ---
//--- the exact timing of the updates to the buffers is not known. ---

//--- The FFT waveform is 32-bits per value and, like the Ampl/Phase data ---
//--- is updated by the DSP whenever we tell it we have finished reading ---
//--- the last copy in its buffer ---

//--- In each case, status bits from the DSP indicate when the buffers ---
//--- have new data, and we need to read the data then tell the DSP that ---
//--- we have done so and that it may write new data to the buffers ---

  if (hdwType == HT_PHASE_CTLR)  {  //test
    maxWait1 = 2000;  maxWait2 = 10000;  }
  else  {
    maxWait1 = 500;  maxWait2 = 2000;  }


//--- Rotate the order we use to check for new waveform data and do not ---
//--- read more than one each pass ---
  for (x=0; x<3; ++x)  {
    if (++check1st >= 3)  check1st = 0;

    if (check1st == 0)  if (NEW_L_DATA
          OR (msSince_(lastWaveReadTime[WAVE_AP_L_SRC]) > maxWait1))
      if (!readWaveform(Rd_L_Buf))  { lastIOTime = GetMS_();  break; }

    if (check1st == 1)  if (NEW_T_DATA
          OR (msSince_(lastWaveReadTime[WAVE_AP_T_SRC]) > maxWait1))
      if (!readWaveform(Rd_T_Buf))  { lastIOTime = GetMS_();  break; }

    if (check1st == 2)  if (FFT_DATA_VALID
          OR (msSince_(lastWaveReadTime[WAVE_FFT_SRC]) > maxWait2))
      if (!readWaveform(AmplFFT))   { lastIOTime = GetMS_();  break; }
  }


/*
  Reading waveforms and doing the periodic I/O to/from the DSP are both
  lengthy processes, so we minimize latency on the rest of the system
  by not allowing them both to happen in the same pass.
*/
  if (msSince_(lastIOTime) > 20)  {

    ilkRatioGain = fwdAmplGain / revAmplGain;
    ilkRatioOffset = 0;  //finish - what to set this to?

    maxOutAmplGain = cavAmplGain * ffGain;
    maxOutAmplOffset = 0;  //finish - what to set this to?


    if (autoReadDSP)
      if (!read_DSP_Regs(rdDSPregs))   lastIOTime = GetMS_();
    if (autoWriteDSP)
      if (!write_DSP_Regs(wrDSPregs))  lastIOTime = GetMS_();
    if (autoReadFPGA)
      if (!read_FPGA_Regs(rdFPGAregs, 0, NUM_RD_FPGA_REGS))  lastIOTime = GetMS_();
    if (autoWriteFPGA)
      if (!write_FPGA_Regs(wrFPGAregs, 0, NUM_WR_FPGA_REGS))  lastIOTime = GetMS_();


    if (dspLoad > 999)  dspLoad = 999;
    if (pcbTemp > 99.9)  pcbTemp = 99.9;

    if (percentFFT > 999)  percentFFT = 999;

    rfIsOn = BOOL(RF_ON);
  //  rfPulsing = BOOL(RF_PULSES);  //finish
    tunerCtlisON = BOOL(TUNER_CTL_ON);
  //  curProgNum = (uint) CUR_PROG_NUM;  //finish
    rfAtWarm = BOOL(RF_OUT_WARM);
    rfRamping = BOOL(OUT_RAMPING);
    cavLocked = BOOL(CAV_LOCKED);
    amplCtlisON = BOOL(AMPL_CTL_ON);
    phaseMode = (uint) PHASE_MODE;

  }

  releaseLockDSP();
}


//=============================================================================
//  Now we update the values in "next" based on the "cur" and "prev"
//  values.
//=============================================================================
//  All values changed by updateNextState() are "outputs" from this func.
//  All other values referenced in this function are "inputs" to this func.
//
//  All the values in "next" provide information about the next logical state
//  for the controller:  Values to be applied during the next pass of the
//  control logic.  All "next" values fall in to one of 2 categories:
//
//    - They are set/updated ONLY by this function (which is executed only
//      by the "applic" task)
//
//    - They are set ONLY by other tasks (within a lock), and possibly
//      reset/cleared by the "applic" task in updateCurState() (after copying
//      "next" to "cur")
//
//
//  All values that act as input to this func must fall in to one of these
//  categories:
//
//    - "prev" values
//
//    - "cur" values updated directly by statements in updateCurState()
//
//    - "cur" values updated indirectly by updateCurState() due to the copy
//      of "next" to "cur", which includes the values in "next" that are
//      set by other tasks
//
//    - a shared/global value that is the equivalent of a "cur" value but
//      for which we don't need to track changes
//
//
//  All "outputs" from this function must be altered ONLY within this function.
//
//  All "inputs" to this function must NEVER be altered within this function.
//
//  Each output value (or VERY small "set" of ouput values) must be updated
//  in a single, self-contained block of code.  NO output may be referenced
//  OR altered in ANOTHER block of code:  This is a declarative style of
//  programming and has multiple advantages:
//
//     - There are no order-of-execution dependencies across nodes/blocks
//       within a pass.  They can be executed in any order and the result
//       will be the same (get used to it - this method will be the norm as
//       computer languages and hardware add support for atomic transactions
//       and computers evolve in to neural-net-like processor arrays - FPGAs
//       are just the beginning of this trend).
//
//     - Encapsulating all logic for an output within a single node/block
//       leaves just one place to start when tracking a problem backward from
//       the result (far simpler than guessing at the possible causes and
//       trying to work forward, looking for paths to the observed result).
//
//     - Drastically reduces the complexity of the code by greatly reducing
//       the number of possible paths through the code.  This simplifies
//       testing and verification processes.
//
//  Each node/block-of-code should be prefaced with the following information:
//   - The names of the values it modifies
//   - A descr of the meaning of the values and the logic used to set them
//   - A list of the values that serve as "inputs" to the node
//
//=============================================================================
void updateNextState(void)
{
  if (current != applTask)  return;

}


//=============================================================================
//  Next we update output signal values to be written to the outputs at
//  the beginning of the next pass (just after the inputs are read).
//=============================================================================
void updateOutputState(void)
{
//  digitalOut = 0;

//  if (rfIsOn)
//    SetBits(Next.digitalOut, ON_OFF_LED);
//  else
//    ClearBits(Next.digitalOut, ON_OFF_LED);

//APPLCODE

}


//=============================================================================
// The last thing we do is update values read by other tasks
//=============================================================================
// IMPORTANT NOTES and WARNINGS:
//
//   To avoid concurrency issues, other tasks need to use locks when reading
//   values set here.  But be careful to keep such locks short to avoid
//   delaying the timing of the main application task.  If necessary, split
//   such code in to sections, getting and releasing the lock(s) in each
//   section to avoid increasing latency for the main application task.
//
//   Values in the "next" instance of the application state fall in to 2
//   categories:  Inputs and Outputs.
//
//   The "next" Outputs are those that may be changed in updateOutputState().
//   The reset are Inputs, and may be changed by other tasks within a lock.
//
//   Since each pass through the application logic starts by setting
//   cur = next, any "next" values set by other tasks become part of the
//   "cur" instance used by the rest of the applic logic.  Because the
//   "applic" task at a higher priority than any other real-time task, the
//   changes to "next" values by other tasks can ONLY happen between passes
//   through the applic logic loop.
//
//   NOTES:
//     - Some of these "next" input values values must be initialized in
//       updateCurState() in order to tell when a value is set by another
//       task.  For example, if we have a boolean flag that is set when we
//       get an ON command, then (after its "next" value is copied to "cur"),
//       we have to clear the flag in order to tell if another ON command is
//       received between passes.
//
//     - As with reads to these values, other tasks must use locks when
//       writing to the "next" input values to insure they are not preempted
//       by the "applic" task in the middle of a write.
//
//=============================================================================
void updateExternals(void)
{
  if (current != applTask)  return;


  digInBits = Cur.digitalIn;
  digOutBits = Cur.digitalOut;

//--- set bits in older statusBitsAppl format ---
  statusBitsAppl = statusBitsDSP & 0x0000FFC8L;
//  if (RF_PULSES)  statusBitsAppl |= (OLD_PHASE_PULSES | OLD_AMPL_PULSES);

//APPLCODE
}


//=============================================================================
_nearcall float readGetRaw(uint ui)  {
  return (ui < NumHdwADC ? anaIn[ui] : 0.0);  }

_nearcall float readMinRaw(uint ui)  {
  return (ui < NumHdwADC) ? (adcMinRaw(ui) - adcOffset[ui]) * adcGain[ui] : 0.0; }

_nearcall float readMaxRaw(uint ui)  {
  return (ui < NumHdwADC) ? (adcMaxRaw(ui) - adcOffset[ui]) * adcGain[ui] : 9.8; }

_nearcall void readDefaults(uint ui, float *gain, float *offset)  {
  *gain = 1.0;  *offset = 0.0; }

_nearcall void readCalibPoints(uint ui, float *low, float *high)  {
  *low = -9.5;  *high = 9.5; }

_nearcall char *readUnits(uint ui, char *buf)  { return "V"; }


//=============================================================================
_nearcall float setGetRaw(uint ui)  {
  return (ui < NumHdwADC ? anaOut[ui] : 0.0);  }

_nearcall float setMinRaw(uint ui)  {
  return (ui < NumHdwDAC) ? (dacMinRaw(ui) - dacOffset[ui]) * dacGain[ui] : 0.0; }

_nearcall float setMaxRaw(uint ui)  {
  return (ui < NumHdwDAC) ? (dacMaxRaw(ui) - dacOffset[ui]) * dacGain[ui] : 9.8; }

_nearcall void setDefaults(uint ui, float *gain, float *offset)  {
  *gain = 1.0;  *offset = 0.0;  }

_nearcall void setCalibPoints(uint ui, float *low, float *high)  {
  *low = -9.5;  *high = 9.5;  }

_nearcall char *setUnits(uint ui, char *buf)  { return "V"; }


//=============================================================================
//  Set default labels for all the I/O signals
//=============================================================================
void setDefaultLabels(void)
{
  printf("Setting default labels for I/O signals\r\n");

//--- set application-appropriate default labels ---
//APPLCODE

//finish
  fstrcpy(dinLabel[0], "FPGA_DONE");

  fstrcpy(doutLabel[0], "CPU_RDY");
  fstrcpy(doutLabel[1], "ACTY");

}


//=============================================================================
//  This function is called to force software configurable values to default
//  values.
//=============================================================================
void setDefaultConfig(void)
{
  printf("Setting in-RAM device config to default values\r\n");

//===== set default values for persistent data values =====
//APPLCODE

//--- default values for application-specific data elements ---
  rf_P = 0.9;   tuner_P = 0.9;
  rf_I = 0.1;   tuner_I = 0.1;
  rf_D = 0.01;  tuner_D = 0.01;

  pulseFreq = 10;  trigPos = 128;

  fwdAtten = revAtten = cavAtten = 4;
  amplSet = 1.0;  phaseSet = 90;

  warmLevel = 1.0;
  warmTime = 1;
  rampTime = 2;

//  setAmplGain = setPhaseGain =
  ilkRatioGain = maxOutAmplGain =
  fwdAmplGain = fwdPhaseGain = revAmplGain = revPhaseGain =
  cavAmplGain = cavPhaseGain = ref0AmplGain = ref0PhaseGain =
  spare0PhaseGain = spare0AmplGain = spare1PhaseGain = spare1AmplGain  =
  spare2PhaseGain = spare2AmplGain = ref1PhaseGain = ref1AmplGain = 1.0;

//  setAmplOffset = setPhaseOffset =
  ilkRatioOffset = maxOutAmplOffset =
  fwdAmplOffset = fwdPhaseOffset = revAmplOffset = revPhaseOffset =
  cavAmplOffset = cavPhaseOffset = ref0AmplOffset = ref0PhaseOffset = 0.0;

  maxAmplSet = 999.0;
}

//=============================================================================
//  This function should be called to prepare for a DSP or ZWorld reset
//=============================================================================
void prepForReset(void)
{
//finish  writeSetDSPState(CMD_RF_OFF);

// do something more MB_setDSPState(...)
}

//=============================================================================
//  This function is called by the application loop just before a shutdown.
//  The appliation must do whatever is necessary for a clean shutdown and
//  then exit
//=============================================================================
void applicShutdown(void)
{
  printf("Application prepping for shutdown\r\n");  fflush(stdout);

  prepForReset();

  Exit(0);
}

//=============================================================================
//  This is the highest priority task, usually resumed at a regular interval
//  so that it can do what it needs and sleep again until the start of the
//  next interval.
//=============================================================================
void _task applic(int argc, char **argv)
{
  #define SharedApplLocals
  #include <sharedAppl.c>

//--- add local variable definitions here ---
//  fd_t  fpgaProgfd;
  LockData  *lock, *endLocks;


  #define SharedApplLocals2
  #include <sharedAppl.c>



  #define SharedApplInit
  #include <sharedAppl.c>

  _EnterCritical_();  SchedMS = 20;  _ExitCritical_();

//--- add application initialization code here ---

  autoReadDSP = autoWriteDSP = autoReadFPGA = autoWriteFPGA = readWaves = true;
  showFiller = false;

  rfq_t1_feedback_RegInfo.flags ^= LogChanges;
  rfq_t2_feedback_RegInfo.flags ^= LogChanges;

  rfq_t1_output_RegInfo.flags ^= LogChanges;
  rfq_t2_output_RegInfo.flags ^= LogChanges;


//--- initialize the locks ---
  endLocks = locks + NUM_LOCKS;
  for (lock=locks; lock<endLocks; ++lock)  {
    lock->count = 0;  lock->owner = 0;
    sem_init(&lock->sem, true, 1);
//    init_waitqueue_head(&lock->tasks);
  }


//  dspProgNum = 0;


  fpgaInitialized = pulseAmpl = pulsePhase = false;

//--- set defaults for # of data points in each waveform ---
  FBZERO(waveformData);  BZERO(readerRdy);
  BZERO(dataTimeSecs);  BZERO(dataTimeMS);
  for (ui=0; ui<NUM_WAVEFORMS*2; ++ui)  {
    numDataPoints[ui] = MaxDataPoints;
    lastWaveReadTime[ui] = GetMS_();
  }



//--- initialize ... ---
  initDefaults();

  SetBits(Next.digitalOut, READY_LED);

  if (getLock(APPL_LOCK, 10000) > 0)  {
    stop_DSP_FPGA();  stat = restart_DSP_FPGA(10000);
    releaseLock(APPL_LOCK);
  }


  #define SharedApplInit2
  #include <sharedAppl.c>

  lastIOTime = GetMS_();

//  calcStartTimeCTF = startCTFWaitTime = 0;

//  resetDSP();

//===================================================
//===== start loop that runs once per timer msg =====
//===================================================

//--- The following loop is run at a periodic interval (normally 100 times ---
//--- per second, or every 10ms).  The application process normally ---
//--- consists of reading the current state of the digital and analog ---
//--- inputs, changing the state of the analog and digital outputs to the ---
//--- values computed during the previous pass, then computing new values ---
//--- for the outputs and updating values that outside clients can view ---
//--- (currently accessable as a set of ModBus registers) ---

  for (;;)  {
    #define SharedApplLoopStart
    #include <sharedAppl.c>

  //--- read the current state of the analog and digital Input signals, and ---
  //--- write the values computed at the end of the last pass ---
    digitalOut = Next.digitalOut;  rwIO();

  //--- limit DSP/FPGA communication to aprox 15 times per second ---
    if (msSince_(lastIOTime) >= 50)  {

    //--- process next event message (if any) ---
      processMsg(eventsfd, (EventMsg *)buf, sizeof(buf));

    //--- update the current state and determine the next state ---
      updateCurState();  updateNextState();
      updateOutputState();  updateExternals();
    }

    #define SharedApplLoopEnd
    #include <sharedAppl.c>
  }

}


//=============================================================================

const CmdInfo far applCmds[] = {
  { "autocal",      autocal,        "calibrate fwd, rev, or cav Atten value" },
  { "cmd",          cmd,            "control DSP state" },
  { "defaults",     defaults,       "set FPGA params to default values" },
  { "dspcalib",     dspcalib,       "calibrate gain/offset for DSP/EGU conversion" },
  { "dspio",        dspio,          "enable/disable normal com with DSP" },
  { "dspread",      dspRead,        "read values from the DSP" },
  { "dspwrite",     dspWrite,       "write a value to the DSP" },
  { "fpgaread",     fpgaRead,       "read values from the FPGA" },
  { "fpgawrite",    fpgaWrite,      "write a value to the FPGA" },
  { "resetfpgadsp", reset_DSP_FPGA, "Reload the FPGA firmware from EEPROM" },
  { "setdspfirm",   setDSPfirm,     "Update and reload DSP firmware image" },
  { "setfpgafirm",  setFPGAfirm,    "Update and reload FPGA firmware image" },
  { "state",        showState,      "Show state of controller" }
};

int const NumApplCmds = ( sizeof(applCmds) / sizeof(CmdInfo) );

//-----------------------------------------------------------------------------
#endif

// progdsp /srfs/RFController2.0.bin
// setfpgafirm /srfs/llrf2.0.bin
