//--- SmartServo.c --- 04/14/07 ---(09/19/02)---

/*
2007-11-12

  New requirement from John Vincent:

  Step-wise movement toward cold setting:

  Current functionality:
  ----------------------
  - If RF-OK false for a specified period of time, change the setpoint to
    the cold setting.

  - Whenever RF-OK goes from false to true, use the current position as the
    set point.


  Revised functionality
  ---------------------
  - If RF lost for a specified amount of time (much smaller than current time)
    then move toward the cold setting in incremental steps (the size of the
    step and the time between steps being configurable amounts).

  - If RF-OK goes from false to true for at least the specified step interval
    amount of time, then stop stepping toward the cold setting and use the
    current position as the set point.
*/


  #define DeviceVers   3
  #define DevName      "RF Servo"
  #define DeviceType   0x2b  // finish - is this a previously unused #?
  #define MinPos       0.0
  #define MaxPos      10.0
//  #define  NSCLSerial   // for testing RS-485 only
  #if (_BOARD_TYPE_ != RCM3200)
    #error "Compiling for RF Servo but target board is not RCM3200!"
  #endif


  #define  PosVolts  (voltsIn[0])

//-----------------------------------------------------------------------------
//
//  Copyright (c) 2002-2005 National Superconducting Cyclotron Laboratory,
//            Michigan State University, and Mark Davis
//
//-----------------------------------------------------------------------------
//
//  Program to control a Servo motor using a ZWorld BL21xx single
//  board computer.
//
//  This version provides support for both the serial NSCL protocol and ModBus
//  over TCP/IP.
//
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//    !!!!! just ONE of the following must be defined when compiling !!!!!
//   (use the appropriate option for the Makefile ion the parent directory)
//-----------------------------------------------------------------------------
//#define BBServo   1
//#define DCServo   1
//#define RFServo   1

//-----------------------------------------------------------------------------
//  !!!!!! DON'T FORGET TO CHANGE DeviceVers ABOVE !!!!!!
//-----------------------------------------------------------------------------
//
//  See the following files common to all controller code for changes that
//  affect all versions:
//      BaseDefs.lib, BaseCode.lib, DLM_DLP_Defs.lib, DLM_DLP_Code.lib
//
//
//  2007-04-14 - Mark Davis
//               - Changed DC SErvo code to use 2nd analog input to deal with
//                 a controller that has lots of noise on the 1st input (both
//                 inputs are connected to the same thing).
//
//  2006-02-14 - Mark Davis
//               - changes for Phase-feedback lock-mode problem
//
//  2005-09-29 - Mark Davis
//               - IDLE mode is no longer entered when a limit is reached (this
//                 caused the DCServo to try to maintain a 0 ref voltage, but
//                 that can result in drift).  Disabling the amp is no good
//                 because that results in both limits going high so you can't
//                 tell what limit you are at).  Now it just changes the target
//                 motor pos to match the one corresponding to the limit pos,
//                 causing it to actively maintain that position (which still
//                 avoids buring out the motor if the switches don't properly
//                 disable the amp for the corresponding direction).
//
//  2005-06-16 - Mark Davis
//               - desiredPhase and phaseLocked values are now saved to flash.
//                 If saved phaseLocked value is less than 4 hours old when
//                 the controller reboots, then it is restored.
//
//  2005-06-14 - Mark Davis
//               - extended ignoreHdwErrs to ignore BOTH limits triggered for
//                 BBServo and DCServo drives
//
//  2005-06-10 - Mark Davis
//               - Put back that code that ignores new remote setting commands
//                 that are the same as the previous one (the main point of
//                 having this is the to INSURE that devices DO NOT MOVE when a
//                 crate is rebooted - EVEN IF THE DEVICE IS NOT AT THE
//                 REQUESTED POSITION!!!
//               - added ignoreHdwErrs flag
//
//  2005-05-26 - Mark Davis
//               - moved phaseAngle from 3xxxx to 4xxxx register group
//
//  2005-05-24 - Mark Davis
//               - testing/fixes/changes to Phase Feedback lock mode.
//
//  2005-05-20 - Mark Davis
//               - disabled code to ignore a new remote setting command that
//                 is the same as the previous one (this was causing settings
//                 to be ignored under some circumstances where it needed to
//                 take effect).  The newer EPICS functionality (updating the
//                 EPICS setpoint for local->remote change) negates the need
//                 for this functionality. (NOTE: put back on 2005-06-10)
//
//  2005-05-16 - Mark Davis
//               - removed startMargin for BBServo's - created separate plus
//                 and minus margins for stopping
//
//  2005-04-22 - Mark Davis
//               - inverted logic for Motor MOVE bits for RFServo control of
//                 of a Bang-Bang Servo motor
//
//  2005-03-23 - Mark Davis
//               - added "Correcting Position" message
//               - changed Jammed status to HDCTLERR to reflect the fact that
//                 this error encompasses a number of hardware and controller
//                 based problems (e.g. bad wiring, failing components,
//                 physical problem with the device, etc).  What it really means
//                 is the device did not respond as expected to a movement cmd.
//
//  2005-03-21 - Mark Davis
//               - finished 1st version of Phase Feedback lock mode.  Pressing
//                 the LOCK button (or sending equiv remote cmd) causes the
//                 controller to adjust the position as needed to maintain a
//                 constant phase value.
//
//  2005-02-22 - Mark Davis
//               - added "mirror" registers in the 4xxxx group for 0xxxx
//                 and 3xxxx values (so ModBus clients can get all the data
//                 they need in 1 poll rather than 3)
//
//  2004-11-11 - Mark Davis
//               - increased time to wait before saving updated noise data
//                 to one day (in the hope that making writes to flash even
//                 less frequent will make the controllers less susceptible to
//                 radiation-caused problems).
//
//  2004-10-29 - Mark Davis - (RFServo ver 2)
//               - altered interpretation of digital input bits used for
//                 switches so they make sense for how they are connected to
//                 the DIP switches and the way the switches are labeled.
//
//  2004-10-06 - Mark Davis - (DCS_13.bin)
//               - The gain on the interface board is computed as soon as
//                 possible.  Previously, if the gain could not be computed
//                 immeadiately after booting (e.g. the controller was in local
//                 mode) the gain was never recomputed (it just used the last
//                 one computed).
//               - fixes to the code that computes the gain (it was using the
//                 same ADC value rather than multiple consecutive ones as
//                 intended so the value was not as accurate as intended).
//
//  2004-06-28 - Mark Davis - (BBS_13.bin, DCS_12.bin)
//               - updated to use new BlockList method of storing persistent
//                 configuration data
//
//  2004-06-28 - Mark Davis - (BBS_12.bin, DCS_11.bin)
//               - updated to use new BlockList method of storing persistent
//                 configuration data
//
//  2004-04-20 - Mark Davis - (BBS_11.bin, DCS_10.bin)
//               - changes necessary to use latest updates to base code
//               - added support for NIM module PID-based servo controller
//
//  2004-01-12 - Mark Davis - (BBS_10.bin, DCS_09.bin)
//               - recompiled DC and BB versions to get latest BASE code chgs
//                 (incl. processing of basic IP config via broadcast pkts)
//               - made changes to work w/ interrupt driven digital/analog I/O
//               - reorganized code in preperation to extracting parts that
//                 are common to all controllers (2003-07-10)
//
//  2003-06-06 - Mark Davis - (BBS_09.bin, DCS_08.bin)
//               - fixed bugs in Motor History data capture and processing
//               - improved calibration code and added to it so it now
//                 determines the minimum ref voltage required to get the motor
//                 to move at around half the max speed.
//               - revamped positioning code for both DC and BB servos.  For
//                 DC Servoes, the ref voltage generated by the interface board
//                 is now used for more real-time response.  Also, the outputs
//                 that control motor movement are now updated ASAP after
//                 determining the motor should be stopped.
//               - reworked telnet interface commands and added commands to
//                 view and change the ADC and DAC calibration data (to allow
//                 a core module to be changed without the need to recalibrate
//                 the EPICS channel).
//               - added support for standard read-only registers that provide
//                 the version date and the device type values to ModBus clients
//               - added a "coil" that forces a drop of a telnet connection
//               - added registers to provide count of seconds since reboot
//                 and version # of Download Manager
//               - IP address for switch positions 0 OR F can be changed from
//                 the default, but not BOTH
//               - added support for netmask and gateway settings
//
//  2003-03-13 - Mark Davis - (BBS_08.bin, DCS_07.bin)
//               - a new position cmd no longer stops motor movement for
//                 BangBang Servos
//               - Init cmd now displays a status msg in telnet debug session
//               - Calibrate command now works properly when the Slop value
//                 is non-zero (DCServos only)
//               - Some fine tuning to calibration cmd and nudging (DCServos)
//               - Nudging is now done ONLY in the direction of movement caused
//                 by the last movement cmd (only applies to DCServos)
//
//  2003-02-18 - Mark Davis - a duplicate position cmd no longer stops a motor
//               movement in progress (BBS_07.bin, DCS_06.bin)
//
//  2003-02-13 - Mark Davis - added calibration command to adjust calibrations
//               for D/A channel 0 to so its output voltage matches readings
//               for A/D channel 0.  This avoids problems with motor movement
//               when rebooting the controller or switching between Local and
//               Remote modes.
//               - The amplifier is also disabled now when rebooting, which also
//               eliminates another unwanted motor movement.
//               - The slopPos value accessed via ModBus registers is now
//               properly updated when the slop is set to 0.0
//               - vt100 clear screen now sent for new telnet connection
//               - added support for setting the amount of time the Download
//               Manager waits for a connection (reboot due to power loss
//               restarts this code ASAP, while rebooting via the Reboot command
//               causes it to wait for 30 seconds).
//               (BBS_06.bin, DCS_05.bin)
//
//  2003-02-06 - Mark Davis - completely revamped local() and remote()
//               functions to simplify and hopefully get rid of the plague of
//               LOCAL/REMOTE behavioral anomolies once and for all.  Also
//               minor fix to serial code to avoid possible array boundary err
//               (BBS_05.bin)
//
//  2003-02-04 - Mark Davis - undid 2003-01-17 change, as that change caused
//               precisely the effect I was trying to avoid (the controller
//               moving the motor back to where it was before LOCAL mode was
//               used).  (in DCS3.bin)
//               - Added code to check for bad D/A calibibration data and use
//               default values if necessary (in DCS4.bin)
//               - Moved code that checks for bad calibration data (and uses
//               default values) to before the first readings are taken (this
//               stops the motor from being moved when the controller is 1st
//               turned on if using a Core Module with no calibration data).
//               (in DCS4.bin)
//
//  2003-01-22 - Mark Davis - chgd DCS device type from 11 (0x0b) to 27 (0x1b)
//               so PC-based test program will function (NOTE: this required
//               a new entry in the PC program's INI file as well)
//
//  2003-01-17 - Mark Davis - value of requested position (not just the set
//               position) now tracks the actual position in local mode.
//               (NOTE: undone on 2003-02-04 as this was WRONG!!!!!)
//
//  2003-01-14 - Mark Davis - elimiated overshoot method and changed so the
//               movement stops as soon as the motor reaches or PASSES the
//               specified position.  Also, the slop value now defaults to 0
//               rather than 0.001
//
//  2003-01-13 - Mark Davis - elimiated DeadZone for BangBang Servo positioning
//               and forced it to try to overshoot by 1 bit of A/D resolution
//
//  2002-12-20 - Mark Davis - changed code to not move motor after achieving
//               desired position unless a new position command is received
//
//  2002-12-19 - Mark Davis - changed BBS device type from 4 to 20 (0x14)
//               so PC-based test program will function (NOTE: this required
//               a new entry in the PC program's INI file as well)
//
//  2002-11-12 - Mark Davis - new register values are now shown as status msgs
//
//  2002-10-07 - Mark Davis - nudging mode now shuts itself off after the
//               specified position has been achieved.
//
//  2002-09-27 - Mark Davis - adding support for "slop" range and (for DCServos)
//               "nudging" to allow increased accuracy
//
//  2002-09-19 - Mark Davis - brand new version of ZWorld DC Servo code that
//               supports both NSCL RS-485 serial and ModBus/TCP clients
//
//----------------------------------------------------------------------------

#class auto
#memmap xmem
#nodebug


//#define ModBusSerial

#define SERIAL_A_AVAIL
#if (_BOARD_TYPE_ != RCM3200)
  #define SERIAL_B_AVAIL
#endif
#define SERIAL_C_AVAIL
#define SERIAL_D_AVAIL
//#define SERIAL_E_AVAIL
//#define SERIAL_F_AVAIL


//#define UseSpline



/*-----------------------------------------------------------------------------
Operational Notes:

- CARDINAL RULE:  Controllers DO NOT MOVE DEVICE if they get a position request
  from a REMOTE source that is the same as the LAST request from a remote
  source!  This is to INSURE the device DOES NOT MOVE when a crate is rebooted!
  It makes no difference if the device is not at the requested position!!!!
  WE DO NOT WANT A WORKING BEAM TO GET MESSED UP JUST BECAUSE A CRATE WAS
  REBOOTED!!!!!

- The Amplifier Enable/Disable switch must be in the Disable position when the
  module is powered up or down to avoid momentarily enabling the servo amp (and
  movement of the motor).  (A disable switch that disconnected the move bits -
  and set the +/- limit inputs high would be a handy addition and make it more
  like the other Servo Ctlrs).

- A Reference (error voltage) Calibration should be done whenever a new
  controller is set up (or swapped) in order to determine the minimum Reference
  (error) voltage necessary to reliably produce movement.  On the NIM module,
  this calibration can be started (or stopped) via the front panel by pressing
  both the Lock and Set buttons while in (panel) Local mode.

- Changes to CIB (Control Interface Board) on older DC Servo controllers:
  - cut or remove R19 (usually a 0-ohm "resistor" or wire) to eliminate the
    non-linear portion of the error amp's transfer function.  This eliminates
    massive chatering and oscillation.
  - Reducing the error amp gain by shorting R17 or R18 will reduce chatter
    and improve accuracy.  This effectively increases the resolution of the
    controller's output voltage and so is preferable to simply reducing the
    Gain setting on the Servo Amplifier or the PID loop parameters.
    (NOTE: Do NOT reduce CIB's error amp gain to less than 10!)
  - The controller checks the gain of the CIB error amp each time it is booted

- For devices with mechanical slop, the PID loop and the Servo Amp need to be
  tuned to eliminate as much overshoot as possible.  Slop means delays in
  changes to the calculated device position value which would cause a PID loop
  to oscillate endlessly (and the device position to overshoot back-and-forth
  endlessly), so the PID loop operates on the motor position, NOT the
  calculated device position.  This means that overshoot corrections are based
  on the motor position only and these corrections (i.e. reverse movement) are
  partly or wholely "absorbed" by the mechanical slop, resulting in little or
  no correction to the device's position (i.e. short of traversing the slop
  zone in reverse, there is no way to undo overshoot - and allowing this would
  cause it oscillate back and forth continuously, correcting overshoot back and
  forth, adding considerable wear to the drive mechanics and probably resulting
  in more slop - not to mention a device that never "settles").

- The Balance on the Servo Amplifier must be adjusted with a zero value
  reference/error voltage.  For the older DC Servos, this can be accomplished
  by simply switching them to LOCAL mode (which disconnects the controller's
  output to the amplifier within the controller box).  For NIM modules, the
  PID loop remains active and the result driving the amplifier, so you must
  use the LOCAL switch at the motor itself to disconnect the ref voltage.

- For the NIM module, the (+) and (-) Drive LED's both blink on and off if the
  device does not respond as expected (i.e. no movement when there should be,
  or movement in the wrong direction).

- The SET button controls what value the Meter ouput on the DB-25 connector
  has.  When the SET button is held closed, it will equal the current Set
  Point value (i.e. the position the device should be moving toward or the
  position the controller is trying to maintain).  When the SET button is
  not being held closed, it will be the current device position (NOT the
  same as the motor position if slop != 0)

- The DISABLE position of the ENABLE/DISABLE switch causes the controller
  to ignore new position request values.

- Pressing the Lock button will toggle the LOCK state of the controller.
  In the Locked state, the controller will change the motor position as
  needed to maintain a steady voltage on the Delta Phi input (the value
  maintained being the value of Delta Phi at the time the LOCK state was
  initiated).  If RF OK lost for more than the specified amount of time,
  the device position will be reset to what it was when the lock was engaged
  (the "cold" position).  NOTE that the locked state is not disabled and
  active corrections will resume when RF OK is reasserted.  Also, manual
  change commands (via the knob or the switch) will alter the value of the
  Delta Phi input the controller is trying to maintain, allowing the RF to
  be tuned while in Phase Feedback Lock mode.


-------------------------------------------------------------------------------
Pin usage for DB-25 connector on back of NIM module:

(NOTE that some inputs are inverted before reaching the CPU, so the logic
 may be inverted from what is listed for the digitalIn bits)

 1 - Shield
 2 - DCS Drive        - AOut: +/-10V ref/error voltage to DC Servo Amp
 3 - DCS Drive Common - AOut: Analog Common to Servo Amp
 4 - +10V Ref         - AOut: To (+) side of motor position feedback pot
 5 - Analog Common    - AOut: To (-) side of motor position feedback pot
 6 - Follow Pot Wiper - AIn:  To middle (wiper) of motor position feedback pot
 7 - Analog Common    -
 8 - (+) Limit        - DIn:  Gnd <= OK, Open or +5V <= (+) Limit reached
 9 - (-) Limit        - DIn:  Gnd <= OK, Open or +5V <= (-) Limit reached
10 - +5.25VDC         - AOut: (use as needed for digital logic inputs)
11 - Loc/Rem          - DIn:  Gnd <= Remote, Open or +5V <= Local
12 - Loc/Rem Common   - DIn:  Dig Common for Local/Remote switch
13 - Amp Enable       - DOut: Gnd => Enable Amp, Open or +5 => Disable Amp

14 - Amp Fault        - DIn:  Gnd <= Amp Operating, +5V <= Problem/Disabled
15 - Amp Reset        - DOut: Gnd => Reset Amp Fault, +5V => for normal oper.
16 - -10V Ref         - AOut: For external +/- drive switch (if needed)
17 - RF OK            - DIn:  Gnd. <= RF NOT OK, +5V <= RF OK
18 - Digital Common   - DIn:  Digital Common (for RF OK if needed)
19 - Meter (+)        - AOut: current read or (when SET pressed) setpoint
20 - Meter (-)        - AOut: Analog Common for Meter
21 - BBS Drive (+)    - DOut: Gnd => no (+) drive, +5V => Move motor in (+) dir
22 - BBS Drive (-)    - DOut: Gnd => no (-) drive, +5V => Move motor in (-) dir
23 - BBS Jumper       - DIn:  Gnd for Bang-Bang Servo mode
24 - Digital Common   - DIn:  Digital Common for BBS/DCS jumper
25 - DCS Jumper       - DIn:  Gnd for DC Servo mode

-------------------------------------------------------------------------------
Key things to remember when making programming changes:

- The ControlLoop interrupt driven process frequently interrupts the funcs
  called via the ComLoop process.  Any changes to shared variables must take
  this in to account to avoid cotention or coherency issues.

  To change values within ComLoop used by ControlLoop (generally due to a
  command from an external client) a "message" buffer is used.

  For access to values spanning multiple bytes (or related groups of values),
  the criticalSection counter is used to briefly disable the ControlLoop.

- The posReq and lastPosReq variables are used to keep track of values that
  came from an external command, and should NOT be altered except in response
  to an external command (or a process that wishes to be treated as an external
  client).

- motorPos is the current position of the motor - our only real readable
  feedback as to the state of the device.

- slopPos is the position of the center of the "slop zone" (the amount the
  motor can move without actually moving the position of the device).  Until
  the motor has moved from its startup position by an amount >= to the
  measured slop, this value can be off by as much as half the amount of slop.

- to avoid endless oscillation on devices with non-zero slop in the drive
  mechanism, the PID loop for DC servos must act on the actual motor position
  and the motor position needed to achieve the desired device position.  If,
  instead, it acted on the "real" (calculated) device position, then it would
  osciallate madly due to the delay in changes to the calculated position
  (caused by the motor movement changing direction - from pushing against one
  edge of the slop zone to traversing it, headed for the other edge).

*/

//-----------------------------------------------------------------------------
//  For next version:
//
// For BB Servo mode:
// - if alternative _dacCalib[0][x] value is reasonable, then use it even if
//   there are "good" factory calibration values (but NEVER overwrite the
//   factory data!)
//
// - support for a specified movement rate?   The velocity-mode PID loop
//   should make that easy (???)
//
//-----------------------------------------------------------------------------
//  Things to test after changes:
//
//  - Movement and status via serial commands
//
//  - Movement and status via ModBus/TCP
//
//  - Movement possible in opposite dir when limit switch active
//
//  - New position cmd received while movement for prev cmd in progress
//
//  - ability to ftp to device after rebooting from telnet interface
//
//  - "quick" reboot when power lost (i.e. minimal wait for ftp timeout)
//
//  - No movement unless CtlWord == 3 (DCServo)
//
//  - No movement if duplicate position command received
//
//  - Duplicate pos cmd does not stop movement for prev cmd with same pos value
//
//  - Amplifier reference volt < min move ref volt when idle (DC/PID modes)
//
//  - Calibration commands produces reasonable values
//
//  - Calibration commands put motor back where it was when done
//
//  - stored values survive reboots and power loss
//
//  - amp disabled when device does not move as expected (e.g. moves the wrong
//    direction or doesn't move when it should)
//
//  - proper recovery from "hdwCtlErr" state
//
//  - controller reboots when config switch settings change and uses new
//    settings (dedicated DC and BB modules with Hex switches on box)
//
//  - rebooting after updating code should not move things
//
//  - local, move, remote, reboot crate should not result in movement (i.e the
//    crate must save position of device when switching from local back to
//    remote
//
//  - approp movement/lack thereof when doing combinations of the following:
//    - switching between local and remote
//    - rebooting the controller (must be in local mode or have amp disabled
//      to avoid momentary movement)
//    - rebooting the crate (with controller in local or remote mode)
//    - disconnecting and reconnecting cables
//    - motor manually moved while in local mode
//
//
//-----------------------------------------------------------------------------
//
//  Operational and Programming NOTES:
//
//  - The Host expects the most significant bits at lower addresses while the
//    ZWorld keeps the most significant bits at the higher address
//
//  - This code assumes that, when recieving values via ModBus TCP protocol,
//    any values that span multiple registers (such as floating point values)
//    will be recieved via a Write Multiple Registers command rather than
//    multiple Write Single Register commands (the latter can cause anomolous
//    behavior).
//
// - The code assumes that the position pot that provides a position reading
//   (as a voltage on the 1st Analog input channel) is actually measuring the
//   position of the motor, NOT the device being controlled.  Because the
//   linkage between the motor and the device being moved is never perfect,
//   this code now supports a "slop" factor, which is the amount that the motor
//   can move without actually moving the device.
//
//   The code currently assumes the slop is symetrical and that the device does
//   not exert any appreciable amount of force on the motor once the desired
//   position is reached (see discussions below for appropriate improvements
//   that could be made in the future).  The code keeps track of where it
//   believes the motor position is relative to the "slop" zone, and adjusts
//   the needed motor position to achieve the specified device position, and
//   adjusts the read back position according to where it believes the device
//   really is.
//
//   Note that, until the motor is actually moved in one direction by an amount
//   equal to the slop factor, the code cannot know the actual position of the
//   device to an accuracy greater than the amount of slop in the linkage.
//

/*
------------------------------------------------------------------------------

  Food for thought regarding devices with "slop" in them:

    ActualPos() assumes that, if the motor position moves within the last
    known range of the slop zone, then the device itself did not move.
    But what if the new position of the motor is due to the device pushing
    it in that direction?  That is to say, if the motor isn't moving because
    we asked it to, then isn't it likely it is due to the device exerting a
    force on it?

    In short, if the motor position drifts when we are not exerting sufficient
    force to move it, then the drift is most likely due to the device pushing
    it, which means it must have drifted to the point where the motor is
    against one edge of the slop zone (and if that edge is the opposite one
    from what where we thought the motor was, then it drifted across the entire
    slop zone).  In any case, taking this in to account would result in more
    accurate device position information.

*/

//-----------------------------------------------------------------------------

#ifdef  COMPILE_PRIMARY_PROG
  #error "Cannot compile as primary program."
  #error "Remove COMPILE_PRIMARY_PROG macro from 'Options/Compiler/Defines'"
  #error "Must Recompile BIOS also. Hit Ctrl-Y then recompile"
#endif

#ifdef  DevelopMode
  #ifdef  COMPILE_SECONDARY_PROG
    #error "DevelopMode set - must NOT compile as secondary program."
    #error "Remove COMPILE_SECONDAY_PROG macro from 'Options/Compiler/Defines'"
    #error "Must Recompile BIOS also. Hit Ctrl-Y then recompile"
  #endif
#else
  #ifndef  COMPILE_SECONDARY_PROG
    #error "Must compile as secondary program or DevelopMode."
    #error "Add COMPILE_SECONDARY_PROG macro in 'Options/Compiler/Defines'"
    #error "Must Recompile BIOS also. Hit Ctrl-Y then recompile"
  #endif
#endif

#ifndef DCServo
#ifndef BBServo
#ifndef RFServo
  #error "Must define DCServo, BBServo, or RFServo"
#endif
#endif
#endif


#define PID_Servo     1


#define  NumADC    4  // # of A/D inputs used

#define  NumDAC    2  // # of D/A outputs used

#define  NumChans  1  // just one logical device

#define UseEncoder  // use optical encoder knob


//#define DEBUG  true


#define  ChanMult     100  // multipler for chan number or'd to reg & coil #'s


//=============================================================================
//         Definitions the same or nearly so for all controllers
//=============================================================================



//=============================================================================
//          Common stuff but with Controller-specific variations
//=============================================================================

//----- bit usage masks for digitalIn -----
  #define  LOCK_BUTTON      0x0001  // 0 = pushed
  #define  LIMIT_SW_POS     0x0002  // 1 = Normal, 0 = Limit Activated
  #define  LIMIT_SW_NEG     0x0004  // 1 = Normal, 0 = Limit Activated
  #define  SET_BUTTON       0x0008  // 0 = pushed
  #define  AMP_DISABLED     0x0010  // 0 = Amp Enabled, 1 = Amp Disabled
  #define  FAST_SW          0x0020  // 0 = closed/slow
  #define  LOC_REM_BUTTON   0x0040  // 0 = closed (toggle Local/Remote)
  #define  AMP_FAULT        0x0080  // 1 = Amp Fault
  #define  RF_OK            0x0100  // 0 = RF OK
  #define  BBS_JUMPER       0x0200  // 0 ==> BB Servo device
  #define  DCS_JUMPER       0x0400  // 0 ==> DC Servo device
  #define  EXTERN_LOCAL     0x0800  // Local sw at motor (1 for DC, 0 for BB)
  #define  BaudIdx()  (Baud_19200)
  #define  AddrIdx()  (0)

//----- bit usage masks for digitalOut -----
  #define  LOC_REM_LED      0x000001  // 0=LED ON, 1=LED OFF
  #define  DRIVE_MINUS_LED  0x000002  // 0=LED ON, 1=LED OFF
  #define  DRIVE_PLUS_LED   0x000004  // 0=LED ON, 1=LED OFF
  #define  AMP_DISABLE_BIT  0x000008  // 0=Enable Amp if Enable Sw on (DCServo)
  #define  LOCKED_LED       0x000010  // 0=LED ON, 1=LED OFF
  #define  FAULT_LED        0x000020  // 0=LED ON, 1=LED OFF
  #define  AMP_RESET_BIT    0x000040  // 0=Reset Amp (DC Servo)
  #define  MOVE_POS_BIT     0x000080  // 0=assert movement in pos dir (BBServo)
  #define  MOVE_NEG_BIT     0x000100  // 0=assert movement in net dir (BBServo)
  #define  CPU_BUSY_LED     0x000200  // 0=LED ON, 1=LED OFF
  #define  ACTIVITY_LED     0x000400  // which bit is used for the activity LED

#define  MOVE_BITS        (MOVE_POS_BIT | MOVE_NEG_BIT)


  const unsigned long LED_ErrPattern[6] =
     { LOCKED_LED, CPU_BUSY_LED,
       DRIVE_PLUS_LED, DRIVE_MINUS_LED,
       FAULT_LED, LOC_REM_LED };


//--- structure for Event messages ----
typedef struct  {
  int   msgID;          // which msg is it?
  unsigned int  seqNum;  // to identify last msg in seq of interdependent ones
  union {
    float newPos;       // new pos request from a diag cmd or external client
    char  enableAmp;    // soft wants amp enabled
  } msgData;
}  EventMsg;

//--- values for EventMsg.msgID ---
#define  MSG_NewPos        1
#define  MSG_EnableAmp     2



#use "Base_DLP_H.lib"


//----- values we want to be PERSISTENT (i.e. stored in flash) -----
//--- (search for PERSISTENT for places code normally added to support them) ---
//-- !!!!! Adding to this struct is OK, but NEVER TOUCH existing items !!!! ---
typedef struct  {
  uint   vers;            // version of this structure
  uint   devType;         // device type this data came from
  float  slop;            // how much motor can move w/o moving device
  float  minMoveRefVolt;  // min ref/velocity volt to amp to sustain movement
  float  pWeight;         // proportional portion of control voltage chgs
  float  iWeight;         // integral portion of control voltage chgs
  float  dWeight;         // differential portion of control voltage chgs
  float  knobRateF;       // scaling for fast adjustments via the knob
  float  knobRateS;       // scaling for slow adjustments via the knob
  float  activeNoise;     // normal readback noise during active control
  float  testAmpl;        // amplitude (in volts) of PID tuning/test sq. wave
  float  testOffset;      // lowest pos (in volts) of PID tuning/test sq. wave
  long   testPeriod;      // period (in ms) of PID tuning/test sq. wave
  float  cibGain;         // gain of error amp on Interface Bd (DC Servo only)
  float  setRate;         // scaling for manual chgs to set point
  float  switchRate;      // additional scaling for +/- Drive switch
  float  dcDamp;          // damping factor to use when adjusting ctl voltage
  uint   maxActiveSecs;   // max secs of active ctl after new setpoint
  uint   rfLockResetDelay; // max ms to pause Phase Lock w/o good RF
  float  maxPhaseErr;     // deadband for phaseLocked mode
  float  plusMargin;      // pos err below which (+) dir BB move cmds stop
  float  minusMargin;     // pos err above which (-) dir BB move cmds stop
  uint   ignoreHdwErrs;   // ignore hdw errors if non-zero
  float  desiredPhase;    // desired phase for PhaseLock mode
  bool   phaseLocked;     // phaseLocked state before last reboot
  ulong  lastUp;          // ZWorld adjusted UTC for when DLP was last running
}  ApplData;

// search for _PERSISTENT_ to see where code normally added for saved values

#define  ApplDataVers   7
// vers 1:  first version used for Block List storage method
// vers 2:  added dcDamp and maxActiveSecs
// vers 3:  added rfLockResetDelay
// vers 4:  added maxPhaseErr, plusMargin, minusMargin
// vers 5:  added ignoredHdwErrs (2005-06-10)
// vers 6:  added desiredPhase (2005-06-15)
// vers 7:  added phaseLocked and lastUp (2005-06-16)




//========================================
//===== Coil or ModBus bit registers =====
//========================================
//--- writing to a coil causes an action - values are not actually saved ---
//  See NewCoilValue() for details

//--- NOTE: 0-19 (00001-000020) are reserved for common values ---
#define  StartTest        20 // [00021]  // start square-wave output test
#define  StopTest         21 // [00022]  // end square-wave output test
#define  DoRefCalib       22 // [00023]  // start Ref volt calibration (DC,RF)
#define  DoGainCalib      23 // [00024]  // start CIB Gain calibration (DC only)
#define  DoDACCalib       24 // [00025]  // start DAC0 to ADC0 calib (DC only)
#define  UpdateDataPoint  25 // [00026]  // update y# for cur setting (see 4xxxx regs)
#define  EnableDriveLock  26 // [00027]  // enable driveLock - no movement allowed
#define  DisableDriveLock 27 // [00028]  // disable driveLock - movement allowed
#define  EnablePhaseLock  28 // [00029]  // enable Phase Feedback Lock
#define  DisablePhaseLock 29 // [00030]  // disable Phase Feedback Lock
//                              =======
              #define  NUMCOILS     30


//======================================
//===== Read-Only ModBus registers =====
//======================================
//--- 30001-30050 (0-49) are reserved for generic/common values ---
//--- 30051-3xxxx are for controller specific values ---

//RO_REG_F32( myReg,         30051 )
//                         =====
     #define  NUM_READ_REGS   50


//=======================================
//===== Read/Write ModBus registers =====
//=======================================
//--- Note that read/write registers can be defined as "Read Only" ---
//--- This allows the client to send write commands for a register to ---
//--- initiate some action without actually changing the value of the ---
//--- register.  Also, devices that have read-only values but multiple ---
//--- logical devices need to use the R/W registers, as the Read-Only ---
//--- set does not support multiple devices, and generally is limited ---
//--- to values that apply to all logical devices handled by one CPU ---

//--- Special handling for chgs to reg values is done by DevNewRegValue() ---

//--- The ChanMult value determines how many registers can be defined ---
//--- per logical device.  A value of 100 means 40001-40100 are for ---
//--- device 0, 40101-40200 are for device 1, etc.  Note that only ---
//--- one set of register definitions is required (i.e. define 40001 ---
//--- but not 40101, 40201, etc).

// NOTE: all position/mechanical values are in VOLTS!

RW_REG_F32( motorPos,      40001, RO ) // pos of motor driving device (via ADC0)
RW_REG_F32( motorPosB,     40003, RO ) // pos of motor (via ADC1)

RW_REG_F32( posSet,        40005, RO ) // cur pos value to reach/maintain
RW_REG_F32( posReq,        40007, RW ) // last new pos requested

RW_REG_U16( devStatus,     40009, RO ) // bit-mapped status bits
RW_REG_U16( ctlWord,       40010, RW ) // switch for client chgs to posReq
RW_REG_U32( versDate_,     40011, RO ) // mirror of versDate
RW_REG_U16( devTyp_,       40013, RO ) // mirror of versDate

RW_REG_F32( refVolt,       40014, RO ) // voltage actually controlling Servo Amp

RW_REG_U16( enableAmp,     40016, RW ) // soft enable/disable Servo Amp
RW_REG_F32( slopValue,     40017, RW ) // amount of mechanical slop in device
RW_REG_U16( oldSampleRate, 40019, RO ) // SampleRate moved to read-only regs

RW_REG_F32( realPos,       40020, RO ) // pos of device as opposed to motor
RW_REG_F32( slopPos,       40022, RO ) // same as realPos after move > slopValue

RW_REG_F32( pWeight,       40024, RW ) // weighting of Proportional factor
RW_REG_F32( iWeight,       40026, RW ) // weighting of Integral factor
RW_REG_F32( dWeight,       40028, RW ) // weighting of Differential factor

RW_REG_F32( pFactor,       40030, RO ) // current value of Proportional factor
RW_REG_F32( iFactor,       40032, RO ) // current value of Integral factor
RW_REG_F32( dFactor,       40034, RO ) // current value of Differential factor

RW_REG_F32( knobRateF,     40036, RW ) // scaling for fast knob adjustments
RW_REG_F32( knobRateS,     40038, RW ) // scaling for slow knob adjustments

RW_REG_F32( testAmpl,      40040, RW ) // amplitude of test-mode square wave
RW_REG_F32( testOffset,    40042, RW ) // lowest voltage of test-mode sq. wave
RW_REG_U32( testPeriod,    40044, RW ) // period of test-mode sq. wave in ms

RW_REG_F32( activeNoise,   40046, RW ) // normal pos noise during active ctl
RW_REG_F32( minMoveRefVolt,40048, RW ) // min ref volt required for movement

RW_REG_F32( ctlVolts,      40050, RO ) // voltage to Interface Board or Amp

RW_REG_F32( cibGain,       40052, RW ) // gain on modified CIB (Ctl.Inter.Brd.)
RW_REG_F32( setRefVolt,    40054, RO ) // desired refVolt value
RW_REG_F32( motorPosSet,   40056, RO ) // needed motor pos for cur posSet
RW_REG_F32( overShoot,     40058, RO ) // amount of overshoot during last move
RW_REG_F32( setRate,       40060, RW ) // scale factor for manual pos chgs

RW_REG_F32( switchRate,    40062, RW ) // scale factor for +/- switch
RW_REG_F32( dcDamp,        40064, RW ) // damping for DCServo ctl voltage chgs

RW_REG_U16( maxActiveSecs, 40066, RW ) // period of active ctl for new setpoint

RW_REG_U32( secCounter_,   40070, RO ) // mirror of secCounter
RW_REG_U16( dlmVersn_,     40072, RO ) // mirror of dlmVers
RW_REG_U16( comRate_,      40073, RO ) // mirror of comRate
RW_REG_U32( deviceID_,     40074, RO ) // mirror of deviceID
RW_REG_U32( coils31_00,    40076, RW ) // alt access to coils
RW_REG_U32( coils63_32,    40078, RW ) // alt access to coils

RW_REG_U16( rfLockResetDelay, 40080, RW ) // max ms to maintain Phase Lock w/o RF

RW_REG_F32( maxPhaseErr,   40081, RW ) // phaseLock deadband

RW_REG_F32( plusMargin,    40083, RW ) // err below which (+) BB move cmds stop
RW_REG_F32( minusMargin,   40085, RW ) // err below which (-) BB move cmds stop

RW_REG_F32( phaseAngle,    40087, RO ) // input on front of NIM module

RW_REG_U16( ignoreHdwErrs, 40089, RW ) // ignore hdw errs if set

RW_REG_F32( desiredPhase,  40090, RW ) // desired Phase value for PhaseLock mode
//                         =====
           #define  NUMREGS   93  // don't forget _x32 types use 2 regs!

// search for _PERSISTENT_ to see where code normally added for saved values

RegInfo  rwRegs[NUMREGS+1];
RegInfo  roRegs[NUM_READ_REGS+1];

//=============================================================================
//              Definitions generally unique to this controller
//=============================================================================

#define  MaxStickTime  5000  // max # ms to push with no measurable result

#define  MinCIBGain     1.0
#define  MaxCIBGain   500.0


//--- bit masks for devStatus [40009#U16] ---
#define  DS_DISABLE       0x0001  // Amplifier is Disabled
#define  DS_FAULT         0x0002  // Fault signal from amplifier
#define  DS_LOCAL         0x0004  // Local mode (soft or hdw)
#define  DS_PLIMIT        0x0008  // Positive Limit switch activated
#define  DS_NLIMIT        0x0010  // Negative Limit switch activated
#define  DS_PHASE         0x0020  // PHASE-FEEDBACK-IN SIGNAL RECEIVED
#define  DS_CALIBRATE     0x0040  // calibration or test sequence running
#define  DS_HDW_ERROR     0x0080  // Problem detected with hardware
#define  DS_RF_OK         0x0100  // RF OK
#define  DS_BBS_JUMPER    0x0200  // BB Servo device
#define  DS_DCS_JUMPER    0x0400  // DC Servo device
#define  DS_EXTERN_LOCAL  0x0800  // Local override at the motor
#define  DS_PHASELOCKED   0x1000  // Locked to external reference
#define  DS_DRIVELOCK     0x2000  // Drive Lock enabled - NO movement allowed
#define  DS_IDLE          0x4000  // device is "resting" (no active control)


//--- controller specific states for calibState ---
#define  CALIB_REF_INIT         100
#define  CALIB_REF_SETTLE       101
#define  CALIB_REF_START        102
#define  CALIB_REF_MOVE         103
#define  CALIB_REF_ABORT        190
#define  CALIB_REF_STOP         191

#define  CALIB_PID_START        400
#define  CALIB_PID_RUN          401
#define  CALIB_PID_STOP         490


//===== structures for serial cmd data and responses =====
// *** WARNING ***
// Remember that the host expects data in MSB 1st format while the Rabbit
// processor uses LSB 1st format for multi-byte formats
// *** WARNING ***

typedef struct {
  unsigned long rom_date;
  unsigned char dev_type;
} VersionInfo;

typedef struct PollMsgStruct {
  ushort  logicalStatus;
  ushort  digitalStatus;
  short   v1raw;
  short   v2raw;
}  PollMsgResp;

typedef struct SetMsgStruct {
  ushort  setInfo;
  short   setPos;
}  SetMsgData;

typedef struct ErrCountsStruct {
  unsigned char packedCounts[(NumErrTypes+1)/2];
}  ErrCountsResp;




//===== function prototypes for controller-specific functions =====

char *CtlStateDesc();

void NewPosCmd(float newPos);
void EnableAmpCmd(bool enableFlag);

bool  UpdateNoiseData();

void ComputeNextState();
void UpdateCtlValues();

float ActualPos(float motor_pos);
float RecentMovement(ushort msPeriod);
float AvgPos(ushort msStartBack, ushort msPeriod);

bool enablePhaseLock(bool remoteCmd);
bool disablePhaseLock(bool remoteCmd);


//=============================================================================
//  GLOBAL variables
//=============================================================================

ApplData  applData;  // for storing application specific config data


//==================== device specific stuff ====================

int  ctlState;
#define  CS_OVERRIDE      -1  // temporarily disables smart positioning code
#define  CS_IDLE           0
#define  CS_STEADY_REF     1
#define  CS_END_LOCAL_CHG  2
#define  CS_START_MOVE     3
#define  CS_ACTIVE_CTL     4

char  *runStateDesc;

ulong  lastNoiseChkTime, lastNoiseChgTime, lastMoveCmdTime, lastEnableTime,
       lastTimeRFOK, lastCorrectingMsgTime, newPosTime, lastCalibGainTime,
       lastPhaseChgTime, lockPressedTime;

long  knobTicks;

float  lastPosReq, halfSlopVal, reqMovement, lastAvgPos, newPos,
       halfADCRes, phaseErr;

bool  slopPosKnown, devInitialized, movePlusSw, moveMinusSw, motorLocal,
      localMode, plusLimit, minusLimit, ampFault, hdwCtlErr, fastAdjust,
      dcServo, bbServo, rfServo, bbMoveMinus, bbMovePlus, phaseLocked,
      localChgLastPass, preMotorLocalMode, rfOK, testingCIBGain, useNewPos,
      trackOverShoot, lockButton, setButton, ampIsDisabled, softDisabled,
      hardDisabled, gainComputed, oscillating, driveLock, idleRest,
      manualMoveCmd, ignoreSettings, reducePhaseErr, correctingPos,
      newCorrectingMsg, adjustPlusMargin, adjustMinusMargin, reachedGoal,
      newPhaseLockReq, phaseLockReqIgnored, restoreSavedPhaseLock,
      phaseLockActive, phaseLockPaused, phaseLockInactive, atALimit,
      ctlrLocal, wasLocalMode, chgdToLocal, chgdToRemote, newPhaseSetting;

int  stateAfterSettled, oscillCount;

const char *ModeDesc[] = { "*UNKNOWN*", "DC Servo ", "BB Servo ", "PID Servo" };

//--- for NIM Servo Controller - connected to a DC or BB servo device? ---
int  ctlMode;
#define  CM_Unknown          0
#define  CM_AnalogPosition   1
#define  CM_DigDirection     2
#define  CM_AnalogVelocity   3

#define  BBSCtlMode  (ctlMode == CM_DigDirection)

#define  RmtCmdsAllowed  \
   ( !localMode AND (((ctlWord & 0xff) == 3) OR BBSCtlMode) )


#define HistorySize  500   // five second's worth of data (at 0.01 sec res)
int  motorHistory[HistorySize];  // recent motor position history
int  lastHistIdx;



xstring DevStateStr {
"Lock Button:    ......      Local LED:     ...      Motor Pos:  -##.####",
"Positive Limit: ...         Drive (-) LED: ...      Delta Phi:  -##.####",
"Negative Limit: ...         Drive (+) LED: ...      +/- Drive:  -##.####",
"Set Button:     ......      Disable Amp:   ...",
"Amp is Enabled: ...         Locked LED:    ...      Ctl Output: -##.####",
"Fast/Slow Sw:   ....        Fault LED:     ...      Meter Out:  -##.####",
"Loc/Rem Button: ......      Reset Ampl:    ...",
"Ampl. Fault:    ...         BBS Move (+):  ...      P: (##.##)  -##.####",
"RF OK:          ...         BBS Move (-):  ...      I: (##.##)  -##.####",
"BBS Jumper:     ...         CPU Busy LED:  ...      D: (##.##)  -##.####",
"DCS Jumper:     ...         Activity LED:  ...",
"Motor Local Sw: ......                              Device Slop:  #.####",
"                            MinMoveRef: ##.####     Ext.Pos.Req:-##.####",
"DIP Switches: [....-....]   Pos. Noise: ##.####     Desired Pos:-##.####",
"                                                    Device Pos: -##.####",
"Control Mode: [.........]   Overshoot:  ##.####",
"[........][......][........]                        Movement:    -#.#### V/s",
""
};

const DigFmtInfo digInFmt[] = {
  { 16, 3, "CLOSED",1,  "OPEN  ",0 },  // Lock Button
  { 16, 4, "YES"   ,1,  "NO "   ,0 },  // Positive Limit Switch
  { 16, 5, "YES"   ,1,  "NO "   ,0 },  // Negative Limit Switch
  { 16, 6, "CLOSED",1,  "OPEN  ",0 },  // Set Button
  { 16, 7, "YES"   ,0,  "NO "   ,1 },  // Amp Enabled signal
  { 16, 8, "SLOW"  ,0,  "FAST"  ,0 },  // Fast/Slow Switch
  { 16, 9, "CLOSED",1,  "OPEN  ",0 },  // Local Button
  {  0, 0, ""      ,0,  ""      ,0 },  // Amplifier Fault
  { 16,11, "YES"   ,0,  "NO "   ,1 },  // RF OK
  { 16,12, "YES"   ,0,  "NO "   ,0 },  // Bang-Bang Servo jumper
  { 16,13, "YES"   ,0,  "NO "   ,0 },  // DC Servo jumper
//{ 16,14, "LOCAL ",1,  "REMOTE",0 },  // Ext Loc/Rem - dif for BB vs DC
  { -1,-1, "",0, "" ,0 }
};

const DigFmtInfo digOutFmt[] = {
  {  0, 0, ""    ,0,  ""    ,0 },  // Local/Remote LED
  { 43, 4, "ON " ,1,  "OFF" ,0 },  // Drive Minus LED
  { 43, 5, "ON " ,1,  "OFF" ,0 },  // Drive Plus LED
  { 43, 6, "NO " ,0,  "YES" ,1 },  // Amp Disable
  { 43, 7, "ON " ,0,  "OFF" ,0 },  // Locked LED
  { 43, 8, "ON " ,1,  "OFF" ,0 },  // Fault LED
  { 43, 9, "YES" ,1,  "NO " ,0 },  // Amp Reset
  { 43,10, "YES" ,1,  "NO " ,0 },  // Move Positive (BBS)
  { 43,11, "YES" ,1,  "NO " ,0 },  // Move Negative (BBS)
  { 43,12, "ON " ,1,  "OFF" ,0 },  // CPU Busy LED
  { 43,13, "ON " ,0,  "OFF" ,0 },  // Activity LED
  { -1,-1, "",0, "",0 }
};

const AnalogFmtInfo  adcFmt[] = {
  { 64, 3, 8,4 },
  { 64, 4, 8,4 },
  { 64, 5, 8,4 },
  { -1,-1, 0,0 }
};

const AnalogFmtInfo  dacFmt[] = {
  { 64, 7, 8,4 },
  { 64, 8, 8,4 },
  { -1,-1, 0,0 }
};



xstring ApplStrs  {
#define   HelpStr                 ApplStr(0)
"A=ADC fmt, C=Cancel, H=Help, L=Loc, R=Rem, S=Sock/Ser, X=eXt.cmd, Z=Zero errs",

#define   ExtHelpStr              ApplStr(1)
  "Bye, Calib, Coil, Help, ID, IP, Move, Reboot, Reg, Serial, Track, <reg>",

#define   ResetApplDataStr        ApplStr(2)
  "Resetting Application data...",

#define   CmdParseErrStr          ApplStr(3)
  "cmd parser error",

#define   CalibRefErrStr          ApplStr(4)
  "CALIB REF not applicable to BBServo mode",

#define  InProgressErrStr         ApplStr(5)
  "Test or calibration already in progress.",

#define  CalibLocalErrStr         ApplStr(6)
  "Cannot calibrate: Controller is in LOCAL mode.",

#define  LocalOnMsg               ApplStr(7)
  "Local mode override active",

#define  IncompleteCmdMsg         ApplStr(8)
  "Incomplete command",

#define  NewColdPosMsg            ApplStr(9)
  "Cold Pos chgd from %.3f to: %.3f",

#define  TelnetKilledMsg          ApplStr(10)
  "***** Telnet connection killed by remote command *****",

#define  MinRefCalibMsg           ApplStr(11)
  "Starting calibration of Minimum Reference Voltage",

#define  LocalModeMsg             ApplStr(12)
  "Controller in Local Mode",

#define  WaitSettleMsg            ApplStr(13)
  "  waiting for settled state...",

#define  MotorDisabledMsg         ApplStr(14)
  "Motor Disabled",

#define  WouldNotSettleMsg        ApplStr(15)
  "Device would not settle",

#define  NoiseLevelMsg            ApplStr(16)
  "  Position readback Noise level:",

#define  RefVoltMsg               ApplStr(17)
   "   slowly changing reference voltage...",

#define  MinusLimitMsg            ApplStr(18)
   "Hit the Minus limit!",

#define  MaxVoltMsg               ApplStr(19)
  "Max voltage exceeded",

#define  PlusLimitMsg             ApplStr(20)
  "   Hit the Plus limit!  Trying other direction...",

#define  ReturnPosMsg             ApplStr(21)
  "Returning to position prior to calibration...",

#define  CalibDACMsg              ApplStr(22)
  "Calibrating first DAC chan to first ADC chan...",

#define  SmallRangeMsg            ApplStr(23)
  "  To small a range of values for good calibration",

#define  ComputeGainMsg           ApplStr(24)
  "Determining gain of amp on Control Interface Board",

#define  NoResponseMsg            ApplStr(25)
  "No response from CIB amp",

#define  StateErrorMsg            ApplStr(26)
  "Error in calibration state machine",

#define  BackwardMsg              ApplStr(27)
  "Limit swiches seem to be wired backwards!!!!",

#define  HdwCtlErrMsg             ApplStr(28)
  "Hdw or Ctlr Error: Device not responding to cmds!!!",

#define  BadSplineDataMsg         ApplStr(29)
  "Cannot use altSetpoint due to unusable interpolation data",

#define  SaveDataPointsMsg        ApplStr(30)
  "Saving updated interpolation Data Points to flash",

#define  DevInitMsg               ApplStr(31)
  "Device Initialized...",

#define  ResetApplDataMsg         ApplStr(32)
  "Resetting Application data...",

#define  AbortingCalibMsg         ApplStr(33)
  ": Aborting Calibration...",

#define  TestStoppedMsg           ApplStr(34)
  ": Test Stopped",

#define  NotInSeqMsg              ApplStr(35)
  "new Data Point not in sequence",

#define  MovingToMinMsg           ApplStr(36)
  "  Moving to Min position",

#define  MovingToMaxMsg           ApplStr(37)
  "  Moving to Max position",

#define  InvalidAmplMsg           ApplStr(38)
  "Invalid test amplitude",

#define  InvalidTestOffsetMsg     ApplStr(39)
  "Invalid test offset",

#define  InvalidComboMsg          ApplStr(40)
  "Invalid test Offset + Amplituded",

#define  PeriodIncrMsg            ApplStr(41)
  "Test Period increased to 100ms",

#define  StartingTestMsg          ApplStr(42)
  "Starting Square Wave test run",

#define  SetpointNotInTableMsg    ApplStr(43)
  "Setpoint not in table",

#define  UpdatedDataPointMsg      ApplStr(44)
  "Updated DataPoint %d: %f,%f   SerialNum: %lu",

#define  OscillationErrMsg        ApplStr(45)
  "Ctl voltage oscillation (%s -> %s) - amp disabled",

#define  ResetDataPointsMsg       ApplStr(46)
  "---- Reset DataPoints values ----",

#define  DriveLockEnabledMsg      ApplStr(47)
  "DriveLock enabled - no movement allowed",

#define  DriveLockDisabledMsg     ApplStr(48)
  "DriveLock disabled - movement allowed",

#define  DriveLockOnMsg           ApplStr(49)
  "DriveLock override active",

#define  PhaseLockedMsg           ApplStr(50)
  "Phase Locked to: %.3f degrees (pos: %.4f)",

#define  PhaseUnlockedMsg         ApplStr(51)
  "Phase Lock disabled: returning %.4f",

#define  PhaseLockLostMsg         ApplStr(52)
  "Phase Lock lost due to loss of RF",

#define  ValidatingDatapointsMsg  ApplStr(53)
  "---- Validating dataPoints ----",

#define  NSCLSerialErrMsg         ApplStr(54)
  "msg err: %c for dest: %d",

#define  UnrecogCmdMsg            ApplStr(55)
  "Unrecognized command: %s",

#define  UserMoveCmdMsg           ApplStr(56)
  "User Command: Move to: %.3f",

#define  InvalidOptionMsg         ApplStr(57)
  "Invalid %s option/value: %s",

#define  SetCoilMsg               ApplStr(58)
  "set coil [%05d] = %d",

#define  MovingFromToMsg          ApplStr(59)
  "Moving from %.3f to: %.3f",

#define  CalibDataMsg1            ApplStr(60)
  "  (%d) ref: %.4f   pos: %.4f   rate: %.4f v/s   \r",

#define  MinVoltMsg               ApplStr(61)
  "Min. Voltage for sustained movement: %.4f",

#define  StartupPosMsg            ApplStr(62)
  "Pos at startup: %.4f",

#define  CalibPointMsg            ApplStr(63)
  "  calib point %d: ADC0: %7.3f  DAC0: %7.4f",

#define  DAC0CalibMsg             ApplStr(64)
  "New DAC0 calib:  Gain: %12.9f  Offset: %9.3f",

#define  CalibGainDataMsg         ApplStr(65)
  "Ref: %.4f  Ctl: %.4f  Pos: %.4f",

#define  UnsuableGainMsg          ApplStr(66)
  "Unusable CIB Gain: %.2f",

#define  CIBGainMsg               ApplStr(67)
  "CIB Gain: %7.3f",

#define  LimitSwitchMsg           ApplStr(68)
  "Limit switch triggered at %7.3f",

#define  NewPosViaSerialMsg       ApplStr(69)
  "New Pos req via NSCL serial: %7.3f",

#define  CannotLockMsg            ApplStr(70)
  "Cannot Lock Phase due to bad RF",

#define  PhaseNotLockedMsg        ApplStr(71)
  "Phase Feedback Lock NOT active",

#define  SettingIgnoredMsg        ApplStr(72)
  "Setting ignored because Phase Lock or DriveLock enabled",

#define  DriftCorrectionMsg       ApplStr(73)
  "Correcting Position due to drift",

#define  PhaseCorrectionMsg       ApplStr(74)
  "Correcting Position for Phase Lock",

#define  RemoteCmdsDisabledMsg    ApplStr(75)
  "Setting ignored - Remote commands disabled",

#define  ReturnToColdPosMsg       ApplStr(76)
  "RF off too long - returning to cold position: %f",

  ""
};   //NewStr


xstring ApplHelpStrs {
  "Calib ref                    - determines min ref voltage needed for movement",
  "DP [reset]                   - list or reset the interpolation DataPoints",
  "DriveLock [on|off]           - View/Enable/Disable DriveLock mode",
  "Move #                       - Move to new position (in volts)",
  "Phaselock [on|off]           - View/Enable/Disable Phase Feedback Lock",
  ""
};


#use "Base_DLP_C.lib"



#ifdef UseSpline
  SplineCoeffs  spline;
#endif

//=============================================================================
//          Standard functions with APPLICATION SPECIFIC variations
//=============================================================================


//-----------------------------------------------------------------------------
//  Initialization for processing of commands
//-----------------------------------------------------------------------------
int InitCmds(int chanNum)
{
  ctlWord = 0;

  devInitialized = true;
  LogMsg(DevInitMsg);

  return 0;
}


//-----------------------------------------------------------------------------
//  Put device in state that is safe for controller to reboot
//-----------------------------------------------------------------------------
void PrepForReboot()
{
  ulong  lastWriteTime;


  enableAmp = false;

//--- wait for latest changes to take effect ---
  lastWriteTime = writeTime;  while (writeTime == lastWriteTime);
}

//-----------------------------------------------------------------------------
//  check validity of application/configuration data.  If any of it is not
//  usuable, change it to a reasonable default value and save any changes
//  to flash memory
//-----------------------------------------------------------------------------
void ValidateApplData(bool readFirst)
{
  bool  dataChgs;
  int  stat;
  float  range, midPoint, defAmpl, defOffset;
  char  *msg;


  StackPtr();

  msg = getBuf(90);

  dataChgs = false;

  if (readFirst)
    if ((stat = Block_Read(BID_APPL_DATA, 0, &applData, sizeof(applData))) < 2)
      sprintf(msg, ApplDataReadErrMsg, stat);

  if (applData.devType != DeviceType)  {
    LogMsg(ResetApplDataMsg);
    memset(&applData, 0, sizeof(applData));
    applData.vers = ApplDataVers;  applData.devType = DeviceType;
    Block_SetSize(BID_APPL_DATA, sizeof(applData));  dataChgs = true;
  }

#define CheckRange(param, minVal, maxVal, defValue)  \
  { if ((applData.param < (minVal)) OR (applData.param > (maxVal)))  {  \
    applData.param = (defValue);  dataChgs = true;  } }


  CheckRange( slop, 0.0, 2.0, 0.0 );
  CheckRange( minMoveRefVolt, 0.0, DACMax, 0.0 );
  CheckRange( activeNoise, 0.0, 2.0, ADCRes * 1.5);

  range = MaxPos - MinPos;  midPoint = MinPos + range / 2.0;
  defAmpl = range / 20.0;  defOffset = midPoint - applData.testAmpl/2.0;

  CheckRange( testAmpl, DACRes, MaxPos, defAmpl );
  CheckRange( testOffset, MinPos, MaxPos, defOffset );
  if (applData.testOffset + applData.testAmpl > MaxPos)  {
    applData.testAmpl = defAmpl;  applData.testOffset = defOffset;
    dataChgs = true;
  }
  CheckRange( testPeriod, 0, 300000, 5000 );

  CheckRange( cibGain, 1.0, 1000.0, 500.0);

  CheckRange( knobRateF, 0.001, 1000.0, 1.0 );
  CheckRange( knobRateS, 0.0001, applData.knobRateF, 0.2 );
  CheckRange( setRate, 0.001, 1000.0, 1.0 );
  CheckRange( switchRate, 0.001, 1000.0, 1.0 );

  if ((applData.pWeight == 0.0) AND
      (applData.iWeight == 0.0) AND
      (applData.dWeight == 0.0))  {
    applData.pWeight = 10.0;
    applData.iWeight =  1.0;
    applData.dWeight =  0.01;
    dataChgs = true;
  }

  CheckRange( dcDamp, 0.001, 1000.0, 0.90 );
  CheckRange( maxActiveSecs, 0, 3600, 300 );

  CheckRange( rfLockResetDelay, 10, 32000, 3000 );

  CheckRange( maxPhaseErr, 0.0, 90.0, 2.0 );

  CheckRange( plusMargin, ADCRes, 1.0, ADCRes);
  CheckRange( minusMargin, ADCRes, 1.0, ADCRes);

  CheckRange( desiredPhase, -90.0, 90.0, 0.0);

// search for _PERSISTENT_ to see where code normally added for saved values


  freeTo(msg);

  if (dataChgs)  SaveApplData(true);
}


#ifdef UseSpline
//-----------------------------------------------------------------------------
//  Reset the data points used for interpolation
//-----------------------------------------------------------------------------
void ResetDataPoints()
{
  xySerialNum = 0;
  x1 = y1 = x2 = y2 = x3 = y3 = x4 = y4 = x5 = y5 = 0.0;
  x6 = y6 = x7 = y7 = x8 = y8 = x9 = y9 = x10 = y10 = 0.0;

  memset(&dataPoints, 0, sizeof(dataPoints));

  dataPoints.vers = DataPointsVers;
  dataPoints.numPoints = 10;
  dataPoints.devType = DeviceType;
  dataPoints.devId = deviceID;

  LogMsg(ResetDataPointsMsg);

  SaveDataPoints();
}

//-----------------------------------------------------------------------------
//  optionally (re)load the dataPoints from flash and verify that they are for
//  the current device (not left over from some previous ctlr usage)
//-----------------------------------------------------------------------------
void ValidateDataPoints(bool readFirst)
{
  bool  resetData;
  int  minBytes;


  resetData = false;

  minBytes = sizeof(dataPoints) - sizeof(dataPoints.xyCoords) + 8;

  LogMsg(ValidatingDatapointsMsg);  //tdebug
  if (readFirst)
    if (Block_Read(BID_DATA_POINTS, 0, &dataPoints, sizeof(dataPoints))
      < minBytes)  resetData = true;

  if (dataPoints.devType != DeviceType)  resetData = true;
  if (dataPoints.devId != deviceID)  resetData = true;

  if (dataPoints.numPoints < 2)  resetData = true;

  if (dataPoints.numPoints > MaxDataPoints)
    dataPoints.numPoints = MaxDataPoints;

  if (resetData)  ResetDataPoints();
}
#endif


//-----------------------------------------------------------------------------
//  process a multi-character command from the diagnostic terminal/console
//
//  If this function does not find the string to be a valid command, then it
//  calls ProcessStrCmdCommon() to see if it is a generic DLP command
//-----------------------------------------------------------------------------
void ProcessStrCmd(char *cmd, LinkProc *lp)
{
  char  *ptr;
  short  i, state, newCalibState;
  float  cur, newSlop, newPosReq, newWeight, newAmpl, newOffset, newPeriod;
  bool  validCmd;
  char  *msg, *temp, *cmdStr;


  StackPtr();

  msg = getBuf(90 + 90 + 90);  temp = msg + 90;  cmdStr = temp + 90;

  strcpy(cmdStr, cmd);  ptr = strtok(cmd, " ");
  state = 0;  validCmd = false;
  strcpy(msg, CmdParseErrStr);

  while (ptr AND (state < 900))  {

    switch (state)  {

      case 0:  // which command did they type?
        if (strcmpi(ptr, "calib") == 0)     { state = 101;  break; }
#ifdef UseSpline
        if (strcmpi(ptr, "dp") == 0)        { state = 110;  break; }
#endif
        if (strcmpi(ptr, "drivelock") == 0)   { state = 120;  break; }
        if (strcmpi(ptr, "move") == 0)      { state = 102;  break; }
        if (strcmpi(ptr, "phaselock") == 0) { state = 130;  break; }
        sprintf(msg, UnrecogCmdMsg, cmd);  state = 999;  break;


      case 101:  // DC Servo-specific form of calib command
        if (strcmpi(ptr, "ref") == 0)  {
          if (BBSCtlMode)  {
            strcpy(msg, CalibRefErrStr);  state = 999;  break; }
          validCmd = true;  newCalibState = CALIB_REF_INIT;
        }
        if (validCmd)  {
          if (calibrating)  {
            strcpy(msg, InProgressErrStr);  state = 999;  break; }
          if (localMode)  {
            strcpy(msg, CalibLocalErrStr);  state = 999;  break; }
          startCalibState = newCalibState;
          *msg = 0;  state = 999;  break;
        }
        state = 998;  break;


      case 102:  // move to new position
        validCmd = true;
        if (ignoreSettings)  {
          strcpy(msg, SettingIgnoredMsg);  state = 999;  break;  }
        newPosReq = atof(ptr);
        if ((newPosReq < MinPos) OR (newPosReq > MaxPos))  {
          state = 998;  break; }
        sprintf(msg, UserMoveCmdMsg, newPosReq);  LogMsg(msg);  *msg = 0;
        NewPosCmd(newPosReq);  state = 999;  break;


#ifdef UseSpline
      case 110:  // got "dp" (DataPoints) cmd
        validCmd = true;
        if (strcmpi(ptr, "reset") == 0)  {
          ResetDataPoints();  *msg = 0;  state = 999;  break; }
        state = 998;   break;
#endif

      case 120:  // got "DriveLock" cmd
        validCmd = true;
        if (strcmpi(ptr, "on") == 0)  {
          driveLock = true;  strcpy(msg, DriveLockEnabledMsg);
          state = 999;  break;
        }
        if (strcmpi(ptr, "off") == 0)  {
          driveLock = false;  strcpy(msg, DriveLockDisabledMsg);
          state = 999;  break;
        }
        state = 998;  break;

      case 130:  // got "phaselock" cmd
        validCmd = true;  *msg = 0;
        if (strcmpi(ptr, "on") == 0)  {
          enablePhaseLock(true);  state = 999;  break; }
        if (strcmpi(ptr, "off") == 0)  {
          disablePhaseLock(true);  state = 999;  break; }
        state = 998;  break;

    }

    if (ptr)  ptr = strtok(NULL, " ");
  }
  strCmdState = false;



//--- what state were we in when we ran out of cmd string to parse? ---
  switch (state)  {

#ifdef UseSpline
    case 110:  // got "dp" (DataPoints) cmd with no params
      ShowDataPoints(0);  *msg = 0;  validCmd = true;  break;
#endif

    case 120:  // "DriveLock" w/o parameters - show current state
      if (driveLock)
        strcpy(msg, DriveLockEnabledMsg);
      else
        strcpy(msg, DriveLockDisabledMsg);
      validCmd = true;  break;

    case 130: // "phaselock" w/o parameters - show current state
      if (phaseLocked)
        sprintf(msg, PhaseLockedMsg, desiredPhase, posSet);
      else
        strcpy(msg, PhaseNotLockedMsg);
      validCmd = true;  break;

    case 998:  // invalid option/value
      sprintf(msg, InvalidOptionMsg, cmd, ptr);  break;

    case 999:  // msg supplied by cmd parser state machine
      break;

    default:
      strcpy(msg, IncompleteCmdMsg);
  }

  if (! validCmd)
    ProcessStrCmdCommon(cmdStr, msg, lp);  // not dev cmd - check generic ones
  else
    if (*msg)
      if (StreamDisp)
        ShowMsg(msg);
      else  {
        respShownTime = MS_TIMER - 5000;
        DispStr(0,23, msg, true);
        respShownTime = MS_TIMER;
      }

  freeTo(msg);
}


//-----------------------------------------------------------------------------
//  process single-char commands from the diagnostic console
//-----------------------------------------------------------------------------
void CharCmd(char cmd)
{
  CharCmdCommon(cmd);

  cmd |= 0x20;


  if (cmd == 'l')  if (! motorLocal)  localMode = true;

  if (cmd == 'r')  if (! motorLocal)  localMode = false;

}


//-----------------------------------------------------------------------------
//  Update data based on digital and analog input values and determine new
//  values for the digital and analog outputs
//
//  !!!!! WARNING !!!!!
//
//  This function is called by the timer-interrupt driven function that reads
//  and writes the Analog and Digital I/O values.  It should be limited to
//  processing current input values and updating output values that will become
//  the active values on the next interrupt (currently, this interuupt occurs
//  every 10ms, so this routine must take considerably less time than that to
//  do EVERYTHING it needs to).  Note that formatting data for display (e.g.
//  sprintf calls) is very time consuming and should be avoided as much as
//  possible.
//-----------------------------------------------------------------------------
nodebug void UpdateDeviceState()
{
  int  i, idx;
  ulong  mask, switchBits;
  bool  ampWasEnabled;
  float  movement, scaleBy;
  char  *msg;


  msg = getBuf(90);

  ampWasEnabled = enableAmp;
  wasLocalMode = localMode;

  ctlMode = CM_Unknown;
  if (!(digitalIn & BBS_JUMPER))  ctlMode = CM_DigDirection;
  if (!(digitalIn & DCS_JUMPER))  ctlMode = CM_AnalogVelocity;


//===== update state based on most recent analog & digital input values =====

  motorLocal = BOOL(digitalIn & EXTERN_LOCAL);
  if (BBSCtlMode)  motorLocal = !motorLocal;

  if (motorLocal)  {
    if (digInChanges & EXTERN_LOCAL)  preMotorLocalMode = localMode;
    localMode = true;
  }
  else  {
    if (digInChanges & EXTERN_LOCAL)  localMode = preMotorLocalMode;
    if (digInChanges & LOC_REM_BUTTON)  {
      if (!(digitalIn & LOC_REM_BUTTON))  localMode = !localMode;  // 0 = pushed
      digInChanges &= ~LOC_REM_BUTTON;  // reset chgd-bit for Loc/Rem button
    }
  }
  digInChanges &= ~EXTERN_LOCAL;  // reset chgd-bit for External Loc/Rem state

  moveMinusSw = (voltsIn[2] < -1.5);  // +/- drive switch tied to ADC02
  movePlusSw = (voltsIn[2] > 1.5);

  plusLimit = !(digitalIn & LIMIT_SW_POS);
  minusLimit = !(digitalIn & LIMIT_SW_NEG);

  ampIsDisabled = BOOL(digitalIn & AMP_DISABLED);  // soft & hard set this bit

  if (BBSCtlMode)
    ampFault = (plusLimit AND minusLimit);  // requires wiring changes!
  else
    ampFault = BOOL(digitalIn & AMP_FAULT);

  rfOK = !(digitalIn & RF_OK);

  fastAdjust = BOOL(digitalIn & FAST_SW);
  lockButton = !(digitalIn & LOCK_BUTTON);
  setButton = !(digitalIn & SET_BUTTON);


  atALimit = (plusLimit OR minusLimit) AND (!plusLimit OR !minusLimit);

  ctlrLocal = (localMode AND !motorLocal);

  chgdToLocal = (!wasLocalMode AND localMode);
  chgdToRemote = (wasLocalMode AND !localMode);

//--- pressing and releasing LOCK btn requests toggle of Phase-feedback mode ---
//--- holding it for > 2 secs forces it to change setpoint to current ---
//--- phase reading (delta-phi) rather than using previous setpoint ---
  if (lockButton AND (! lockPressedTime))  lockPressedTime = MS_TIMER;

  if (!lockButton AND lockPressedTime)  {
    newPhaseLockReq = (!setButton AND ctlrLocal);
    newPhaseSetting = (elapsedMS(lockPressedTime) > 2000);
    lockPressedTime = 0;
  }

  if (! newPhaseLockReq)  phaseLockReqIgnored = newPhaseSetting = false;

  softDisabled = !enableAmp;
  hardDisabled = ampIsDisabled AND !softDisabled;

  if (softDisabled)  ampFault = false;
  if (elapsedMS(lastEnableTime) < 200)  ampIsDisabled = ampFault = false;

  if (rfOK)  lastTimeRFOK = MS_TIMER;

  usingCount = true;  knobTicks = qdCount;  qdCount = 0;  usingCount = false;

  manualMoveCmd = localMode AND (knobTicks OR movePlusSw OR moveMinusSw);

  ignoreSettings = driveLock OR phaseLocked;


//===== check for new events =====
  while (Buf_Used(&eventMsgQueue) > 0)  {
    if (Buf_Remove(&eventMsgQueue, (char *)(&eventMsg), sizeof(eventMsg))
        != sizeof(eventMsg))  LogErr(EventMsgErr);
    else   switch (eventMsg.msgID)  {
      case MSG_NewPos:
        if (driveLock)  break;
        newPos = eventMsg.msgData.newPos;  useNewPos = true;  break;
      case MSG_EnableAmp:
        if (driveLock)  break;
        enableAmp = (int) eventMsg.msgData.enableAmp;  break;
    }
  }


//--- check for manual/front panel initiation of minMoveRefVolt calibration ---
//--- for NIM module, pressing SET and LOCK buttons together while in local ---
//--- mode will either start a REF calibration or cancel cur calibration ---
  #define  LockAndSet  (LOCK_BUTTON | SET_BUTTON)
  if (lockButton AND setButton)  do  {
    if ((digInChanges & LockAndSet) != LockAndSet)  break;  // once per "press"
    digInChanges &= ~LockAndSet;  // reset chg bits for Lock and Set buttons

    if (ignoreSettings)  break;  // not while DriveLock or PhaseFeedback enabled
    if (!localMode OR motorLocal)  break;  // must be in ctlr local mode

    if (BBSCtlMode)  break;  // not applicable to BB Servos

    if (calibrating)
      cancelCalib = true;
    else
      startCalibState = CALIB_REF_INIT;
  }  while (false);


//--- Check for lock button alone pushed while RF OK ---
//--- This toggles the Phase Feedback Lock mode, where the controller ---
//--- attempts to maintain a steady Phase Feedback value ---
  if (newPhaseLockReq)  do  {
    newPhaseLockReq = false;  phaseLockReqIgnored = true;
    if (driveLock)  break;  // NO movement/changes allowed!
    if (phaseLocked)  {
      disablePhaseLock(false);  phaseLockReqIgnored = false;  break; }
    if (!rfOK)  break;  // must have RF to enable Phase Lock!
    if (! BBSCtlMode)  break;  // only for BB Servos right now
    phaseLockReqIgnored = false;  enablePhaseLock(false);
  }  while (false);


//--- if we lost RF for too long, return to "Cold Position" ---
  if (phaseLocked AND rfLockResetDelay AND !phaseLockInactive)
    if (elapsedMS(lastTimeRFOK) > rfLockResetDelay)  {
      sprintf(msg, ReturnToColdPosMsg, posSet);  LogMsg(msg);
      NewPosCmd(posSet);  phaseLockInactive = true;
    }

  phaseLockActive = (phaseLocked AND rfOK);
  if (phaseLockActive)  phaseLockInactive = false;
  phaseLockPaused = (phaseLocked AND !phaseLockActive AND !phaseLockInactive);



  devStatus = 0;
  if (ampIsDisabled)  devStatus |= DS_DISABLE;
  if (ampFault)       devStatus |= DS_FAULT;
  if (localMode)      devStatus |= DS_LOCAL;
  if (! softDisabled)  {
    if (plusLimit)      devStatus |= DS_PLIMIT;
    if (minusLimit)     devStatus |= DS_NLIMIT;
  }
  if (calibrating)    devStatus |= DS_CALIBRATE;
  if (flashBad)       devStatus |= DS_HDW_ERROR;
  if (oscillating)    devStatus |= DS_HDW_ERROR;
  if (hdwCtlErr)      devStatus |= DS_HDW_ERROR | DS_NLIMIT | DS_PLIMIT;
  if (rfOK)           devStatus |= DS_RF_OK;
  if (ctlMode == CM_DigDirection)    devStatus |= DS_BBS_JUMPER;
  if (ctlMode == CM_AnalogVelocity)  devStatus |= DS_DCS_JUMPER;
  if (motorLocal)     devStatus |= DS_EXTERN_LOCAL;
  if (driveLock)      devStatus |= DS_DRIVELOCK;
  if (idleRest)       devStatus |= DS_IDLE;
  if (phaseLocked)    devStatus |= DS_PHASELOCKED;


//--- update motor position history ---
  motorHistory[lastHistIdx] = adcRaw[0];
  if (++lastHistIdx == HistorySize)  lastHistIdx = 0;


//--- update register values ---
  motorPos = PosVolts;
  motorPosB = motorPos;
  refVolt = voltsOut[0];
  if (rfOK)  phaseAngle = voltsIn[1] * 10.0;  // 9V = 90 degrees


  realPos = ActualPos(motorPos);


//--- disable idleRest if new movement cmd ---
  if (idleRest AND !driveLock)  if (useNewPos OR manualMoveCmd) {
    // assume for one pass that fault was due ONLY to idleRest mode
    idleRest = softDisabled = ampIsDisabled = false;
    enableAmp = true;  lastMoveCmdTime = MS_TIMER;
  }

  if (phaseLocked)  {

  //--- apply manual chg cmds to desiredPhase while in Phase Feedback mode ---
    if (manualMoveCmd AND !driveLock AND localMode AND phaseLockActive)  {
      if (fastAdjust)  scaleBy = knobRateF;  else  scaleBy = knobRateS;
      if (movePlusSw)   knobTicks = (long) (switchRate *  10.0 / knobRateF);
      if (moveMinusSw)  knobTicks = (long) (switchRate * -10.0 / knobRateF);
      desiredPhase += (float) knobTicks * scaleBy / 100.0;
      lastPhaseChgTime = MS_TIMER;
    }

  //--- Has Phase Feedback value drifted outside deadzone? ---
  //--- NOTE: Once initiated, the phaseErr value needs to stay constant ---
  //--- until current correction has ended (to avoid stopping correction ---
  //--- before reaching the middle of the deadzone, or a never ending ---
  //--- series of corrections) ---
    if (!reducePhaseErr)  {
      phaseErr = desiredPhase - phaseAngle;
      if (fabs(phaseErr) > maxPhaseErr)  {
        reducePhaseErr = true;
      //--- ensure idleRest is disabled ---
        idleRest = softDisabled = ampIsDisabled = false;
        enableAmp = true;  lastMoveCmdTime = MS_TIMER;
      }
    }

  }
  if (!rfOK)  reducePhaseErr = false;


//===== compute next desired state =====
  ComputeNextState();


//===== update digital and analog output values based on next state =====

//--- update analog outputs ---
  voltsOut[0] = ctlVolts;
  if (setButton)
    voltsOut[1] = posSet;
  else
    voltsOut[1] = realPos;

//--- update digital outputs ---
#ifdef EXT_485_RX_TX
  digitalOut = (~ EXT_485_RX_TX) | (digitalOut & EXT_485_RX_TX);
#else
  digitalOut = 0xffffffff;
#endif


  movement = (AvgPos(100,100) - AvgPos(600,100)) * 2.0;

  if (motorLocal)  {
    if (MS_TIMER & 0x100)  ClearBits(digitalOut, LOC_REM_LED);  }
  else
    if (localMode)  ClearBits(digitalOut, LOC_REM_LED);

  if (hdwCtlErr)  {
    if (MS_TIMER & 0x100)  {
      ClearBits(digitalOut, DRIVE_MINUS_LED | DRIVE_PLUS_LED);  }
  }
  else  {
    if (movement < -activeNoise)  ClearBits(digitalOut, DRIVE_MINUS_LED);
    if (movement >  activeNoise)  ClearBits(digitalOut, DRIVE_PLUS_LED);
  }
//--- if attempting local move while locked, turn on drive LED's ---
  if (manualMoveCmd AND driveLock)  {
    ClearBits(digitalOut, DRIVE_MINUS_LED | DRIVE_PLUS_LED);
    usingCount = true;  qdCount = 0;  usingCount = false;
  }

  if (phaseLockReqIgnored)  { // rapid blink if ignored
    if (MS_TIMER & 0x40)  ClearBits(digitalOut, LOCKED_LED);  }
  else
    if (phaseLocked)  ClearBits(digitalOut, LOCKED_LED);

  if (calibrating)  if (MS_TIMER & 0x100)  ClearBits(digitalOut, CPU_BUSY_LED);

//  if (driveLock)  {  // drives Vincent nuts!
//    if (MS_TIMER & 0x80)  ClearBits(digitalOut, FAULT_LED);  }
//  else
    if (ampFault OR hardDisabled)  ClearBits(digitalOut, FAULT_LED);


  switch (ctlMode)  {

    case CM_AnalogPosition:
    case CM_AnalogVelocity:
      SetBits(digitalOut, MOVE_BITS); // just to be safe
      if (enableAmp)  ClearBits(digitalOut, AMP_DISABLE_BIT);
  //--- if amp fault when it should be enabled, try pulsing the RESET line ---
      if (ampFault AND enableAmp AND !ampIsDisabled)
        if (MS_TIMER & 0x200)  ClearBits(digitalOut, AMP_RESET_BIT);
      break;

    case CM_DigDirection:
      if (enableAmp)  ClearBits(digitalOut, AMP_DISABLE_BIT); //to read switch
      if (bbMovePlus)  ClearBits(digitalOut, MOVE_POS_BIT);
      if (bbMoveMinus)  ClearBits(digitalOut, MOVE_NEG_BIT);
      break;

    default:  // blink LEDS in pattern to show error
      for (i=0; i<6; ++i)  SetBits(digitalOut, LED_ErrPattern[i]);
      i = (int)((MS_TIMER/100L) % 6L);
      ClearBits(digitalOut, LED_ErrPattern[i]);
  }

  if (enableAmp AND !ampWasEnabled)  lastEnableTime = MS_TIMER;


  freeTo(msg);
}


//-----------------------------------------------------------------------------
//  Do anything specific to this controller that is not time-ciritical.
//  This is called once during each pass of the ComLoop() function
//-----------------------------------------------------------------------------
void BackGroundProcess()
{
  int  i;
  bool  correcting;


  if (recursiveBackGround)  return;
  recursiveBackGround = true;

  ++criticalSection;

//--- 3xxxx ModBus register values mirrored in the 4xxxx set ---
  versDate_ = versDate;
  devTyp_ = devTyp;
  secCounter_ = secCounter;
  dlmVersn_ = dlmVersn;
  comRate_ = comRate;
  deviceID_ = deviceID;

  correcting = correctingPos;  correctingPos = false;

  --criticalSection;


  if (!correcting)
    newCorrectingMsg = true;
  else  {
    if (newCorrectingMsg AND !atALimit
        AND (elapsedMS(lastCorrectingMsgTime) > 5000))  {
      if (phaseLocked)
        LogMsg(PhaseCorrectionMsg);
      else
        LogMsg(DriftCorrectionMsg);
      lastCorrectingMsgTime = MS_TIMER;
    }
    newCorrectingMsg = false;
  }

  ++criticalSection;
  if ((applData.desiredPhase != desiredPhase)
      AND (elapsedMS(lastPhaseChgTime) > 30000)) {
    applData.desiredPhase = desiredPhase;  ChgWait(1);
  }
  if (utcSec)  {
    if (restoreSavedPhaseLock)  {
      if ((utcSec - applData.lastUp) < 14400L)  // < than 4 hours old?
        if (applData.phaseLocked)  enablePhaseLock(false);
      restoreSavedPhaseLock = false;
    }
    if (applData.phaseLocked != phaseLocked)  {
      applData.phaseLocked = phaseLocked;  ChgWait(30);  }
  }
  --criticalSection;

  recursiveBackGround = false;
}


//-----------------------------------------------------------------------------
//  Do device specific formatting of current state data that cannot be
//  handled by entries in generic table for I/O values
//
//  Returns:
//      true:  last portion of data has been displayed
//      false:  more data to display during next pass
//-----------------------------------------------------------------------------
bool ShowDeviceState()
{
  char *ptr, *locRemDesc;
  static int  devShowState;
  char  *msg, *temp;



  StackPtr();

  msg = getBuf(90 + 90);  temp = msg + 90;

  locRemDesc = localMode ? \
    (blinkOff ? "      " : (motorLocal ? "MLOCAL": "LOCAL ")) : "REMOTE";

//--- The updating of the screen is done in stages to avoid taking too ---
//--- much time (all at once) away from other background & com tasks ---

  switch (++devShowState)  {

  //--- non-generic stuff in left column ----
    case 1:
      ptr = (motorLocal ? (blinkOff ? "      " : "LOCAL ") : "REMOTE");
      DispStr(16,14, ptr, false);

      if (BBSCtlMode)
        ptr = "N/A";
      else
        ptr = (ampFault ? (blinkOff ? "   " : "YES") : "NO ");
      DispStr(16,10, ptr, false);

      LongToBin(deviceID, temp, 8);
      sprintf(msg, "[%.4s-%.4s]", temp, temp+4);  DispStr(14,16, msg, false);

      DispStr(15,18, ModeDesc[ctlMode], false);

      sprintf(msg, "[%s][%s][%s]", runStateDesc, locRemDesc, CtlStateDesc());
      DispStr(0,19, msg, false);

      break;


  //--- non-generic stuff in middle column ----
    case 2:
      if (motorLocal)
        ptr = (blinkOff ? "     " : "BLINK");
      else
        ptr = (localMode ? "ON   " : "OFF  ");
      DispStr(43, 3, ptr, false);

      if (BBSCtlMode)
        DispStr(40,15, " N/A    ", false);
      else
        DispStr(40,15, fmtFloat(minMoveRefVolt,7,4,temp), false);
      break;
    case 3:
      DispStr(40,16, fmtFloat(activeNoise,7,4,temp), false);

      DispStr(40,18, fmtFloat(overShoot,7,4,temp), false);
      break;


  //--- non-generic stuff in right column ----
    case 4:
      DispStr(56,10, fmtFloat(pWeight,5,2,temp), false);
      DispStr(64,10, fmtFloat(pFactor,8,4,temp), false);
      break;
    case 5:
      DispStr(56,11, fmtFloat(iWeight,5,2,temp), false);
      DispStr(64,11, fmtFloat(iFactor,8,4,temp), false);
      break;
    case 6:
      DispStr(56,12, fmtFloat(dWeight,5,2,temp), false);
      DispStr(64,12, fmtFloat(dFactor,8,4,temp), false);
      break;

    case 7:
      DispStr(66,14, fmtFloat(slopValue,6,4,temp), false);
      DispStr(64,15, fmtFloat(posReq,8,4,temp), false);
      break;
    case 8:
      DispStr(64,16, fmtFloat(posSet,8,4,temp), false);
      DispStr(64,17, fmtFloat(realPos,8,4,temp), false);
      break;

    case 9:
      DispStr(64,19, fmtFloat(RecentMovement(1000),8,4,temp), false);
      break;


  //--- to catch all non-active state values ---
    default:
      devShowState = 0;  break;
  }

  freeTo(msg);

  return (devShowState == 0);
}


//-----------------------------------------------------------------------------
//  Determine what, if any, action should be taken as a result of one of the
//  coil values being changed
//-----------------------------------------------------------------------------
void NewCoilValue(int chanNum, int coilNum, int coilVal, int cmdRow)
{
  int  i, newRange, newIdx, stat;
  char  *msg;
  RegInfo  *reg;


  StackPtr();

  msg = getBuf(80);

  sprintf(msg, SetCoilMsg, coilNum+1, coilVal);
//  if (GenericDisp)  DispStr(CmdCol,cmdRow, msg, true);
  LogMsg(msg);

  if ((coilNum < 0) OR (coilNum >= NUMCOILS))  { freeTo(msg);  return; }

  if (coilVal)  switch (coilNum)  {

    case DoReboot:
      Reboot(10); break;

    case DoInit:
      InitCmds(0);  break;

    case AbortCalib:
      cancelCalib = true;   break;

    case StartTest:
      startCalibState = CALIB_PID_START;  break;

    case StopTest:
      startCalibState = CALIB_PID_STOP;  break;

    case DoRefCalib:
      if (! BBSCtlMode)  startCalibState = CALIB_REF_INIT;
      break;


#ifdef UseSpline
    case UpdateDataPoint:
      for (i=0; i<dataPoints.numPoints; ++i)
        if (altSetpoint == dataPoints.xyCoords[i][0])  break;
      if (i >= dataPoints.numPoints)  {
        LogErr(SetpointNotInTableMsg);  break;  }
      dataPoints.xyCoords[i][1] = realPos;   ++dataPoints.serialNum;
      reg = rwRegs + (y1_Idx + i*4);
      *((float *)reg->ptrCur) = realPos;  xySerialNum = dataPoints.serialNum;
      sprintf(msg, UpdatedDataPointMsg,
              i+1, dataPoints.xyCoords[i][0], dataPoints.xyCoords[i][1],
              xySerialNum );
      LogMsg(msg);  break;
#endif

    case EnableDriveLock:
      LogMsg(DriveLockEnabledMsg);   driveLock = true;  break;

    case DisableDriveLock:
      LogMsg(DriveLockDisabledMsg);   driveLock = false;  break;

    case EnablePhaseLock:
      enablePhaseLock(true);  break;

    case DisablePhaseLock:
      disablePhaseLock(true);  break;

  }
  coilVal = 0;

  freeTo(msg);

  ++criticalSection;
  ClearBit(coils[chanNum], coilNum);
  --criticalSection;
}


//-----------------------------------------------------------------------------
//  Determine what, if any, action should be taken as a result of one of the
//  registers being written to.
//
//  Called by the generic function NewRegValue() in the base code which is
//  responsible for actually copying the <regName>_New value to <regName>
//
//  Returns true if the new value should be used, false if it should not
//
//  If any value used during execution of the periodic ControlLoop() function
//  (besides the new register value) is altered here, you MUST increase the
//  criticalSection counter 1st (it will be restored to its previous value
//  after the new register value is in effect).  This is to avoid having the
//  ControlLoop() processes act on unsynchronized values.
//
//    Example:  DevNewRegValue() alters some non-register value in response
//    to a new register value.  Since the new register value does not go
//    into effect until DevNewRegValue() returns, any such values used within
//    ControlLoop() would be out of sync with the current register value.
//
//  NOTE:  Assumes that, when a register that is part of a larger data type
//         (e.g. a 32-bit float or long integer) that the lower numbered
//         registers will be written first.
//
//  Any string in *msg will be logged after returning from this func.
//-----------------------------------------------------------------------------
bool  DevNewRegValue(int chanNum, int regNum, int regVal,
                     int clientType, LinkProc *lp, char *msg)
{
  int  i, idx, pi, xy, cmdRow;
  ulong  mask;
  float  newPosReq;
  RegInfo  *reg;


  cmdRow = CmdRow(LinkProc_Idx(lp));

//--- Check the New register value (or set of registers comprising a ---
//--- single value) and either reject it (in which case the new value ---
//--- will be ignored), or do whatever processing is necessary using the ---
//--- new value.  NOTE: The new value does not actually go in to effect ---
//--- until sometime after this function returns (one of the reasons that ---
//--- proper handling of the criticalSection counter is required). ---

  do  {

//  if (regNum == regName_Idx)  {
//    if (regName_New not usable)  {
//      sprintf(msg, "optional msg if default 'ignored' msg insufficient");
//      return false;
//    }
//    ... whatever else needs to be done with regName_New value ...
//    sprintf(msg, "...optional msg about action taken...");
//    (NOTE: regName = regName_New is handled by caller of this function)
//    break;
//  }

    if (regNum == posReq_Idx)  {
      if (driveLock)  { strcpy(msg, SettingIgnoredMsg);  return false; }
      if (posReq_New != lastPosReq)  {
        if (!RmtCmdsAllowed) {
          strcpy(msg, RemoteCmdsDisabledMsg);  return false; }
        lastPosReq = posReq_New;
        respShownTime = MS_TIMER - 5000;
        if (! phaseLocked)
          sprintf(msg, MovingFromToMsg, realPos, posReq_New);
        else
          sprintf(msg, NewColdPosMsg, posSet, posReq_New);
        NewPosCmd(posReq_New);
      }
      break;
    }

#ifdef UseSpline
    if (regNum == altSetpoint_Idx)  {
      if (ignoreSettings)  { strcpy(msg, SettingIgnoredMsg);  return false; }
      ValidateDataPoints(false);
      if (ComputeSpline(dataPoints.xyCoords, dataPoints.numPoints, &spline))  {
        strcpy(msg, BadSplineDataMsg);  return false; }
      newPosReq = SplineFit(altSetpoint_New, dataPoints.xyCoords,
                            dataPoints.numPoints, &spline);
      if (newPosReq != lastPosReq)  {
        if (!RmtCmdsAllowed)  {
          strcpy(msg, RemoteCmdsDisabledMsg);  return false; }
        lastPosReq = newPosReq;
        respShownTime = MS_TIMER - 5000;
        sprintf(msg, MovingFromToMsg, realPos, newPosReq);
        NewPosCmd(newPosReq);
      }
      break;
    }
#endif

    if (regNum == slopValue_Idx)  {
      if ((slopValue_New < 0.0) OR (slopValue_New > 5.0)) return false;
      applData.slop = slopValue_New;  ChgWait(1);
      ++criticalSection;  // undone by caller after new reg value in effect
      halfSlopVal = slopValue_New / 2.0;  slopPosKnown = false;
      break;
    }

    if (regNum == pWeight_Idx)  {
      applData.pWeight = pWeight_New;  ChgWait(30);  break;  }

    if (regNum == iWeight_Idx)  {
      applData.iWeight = iWeight_New;  ChgWait(30);  break;  }

    if (regNum == dWeight_Idx)  {
      applData.dWeight = dWeight_New;  ChgWait(30);  break;  }

    if (regNum == knobRateF_Idx)  {
      if (knobRateF_New <= 0.0)  return false;
      applData.knobRateF = knobRateF_New;  ChgWait(1);  break;
    }

    if (regNum == knobRateS_Idx)  {
      if (knobRateS_New <= 0.0)  return false;
      applData.knobRateS = knobRateS_New;  ChgWait(1);  break;
    }

    if (regNum == setRate_Idx)  {
      if (setRate_New <= 0.0)  return false;
      applData.setRate = setRate_New;  ChgWait(1);  break;
    }

    if (regNum == switchRate_Idx)  {
      if (switchRate_New <= 0.0)  return false;
      applData.switchRate = switchRate_New;  ChgWait(1);  break;
    }

    if (regNum == dcDamp_Idx)  {
      if (dcDamp_New <= 0.0)  return false;
      applData.dcDamp = dcDamp_New;  ChgWait(1);  break;
    }

    if (regNum == maxActiveSecs_Idx)  {
      if (maxActiveSecs_New < 0)  return false;
      applData.maxActiveSecs = maxActiveSecs_New;  ChgWait(1);  break;
    }

    if (regNum == testAmpl_Idx)  {
      if (testAmpl_New <= 0.0)  return false;
      applData.testAmpl = testAmpl_New;  ChgWait(60);  break;
    }

    if (regNum == testOffset_Idx)  {
      if (testOffset_New < 0.0)  return false;
      applData.testOffset = testOffset_New;  ChgWait(60);  break;
    }

    if (regNum == testPeriod_Idx)  {
      if (testPeriod_New < 100)  return false;
      applData.testPeriod = testPeriod_New;  ChgWait(60);  break;
    }

    if (regNum == activeNoise_Idx)  {
      if ((activeNoise_New < 0.0) OR (activeNoise_New > 2.0))  return false;
      applData.activeNoise = activeNoise_New;  ChgWait(30);  break;
    }

    if (regNum == minMoveRefVolt_Idx)  {
      if ((minMoveRefVolt_New < 0.0) OR (minMoveRefVolt_New > DACMax))
        return false;
      applData.minMoveRefVolt = minMoveRefVolt_New;  ChgWait(30);  break;
    }

#ifdef UseSpline
    if ((regNum >= x1_Idx) AND (regNum <= y10_Idx))  {
      reg = rwRegs + regNum;  if (! LastRegInVal(reg))  break;
      idx = regNum - x1_Idx;  pi = idx / 4;  xy = (idx / 2) & 1;
      while (reg->wordIdx) --reg;  // get ptr to 1st reg in pair
      dataPoints.xyCoords[pi][xy] = *((float *)reg->ptrNew);
sprintf(msg, "xy[%d][%d] = %f", pi, xy, *((float *)reg->ptrNew) );  //tdebug
      if (pi >= dataPoints.numPoints)  dataPoints.numPoints = pi + 1;
      break;
    }

    if (regNum == xySerialNum_Idx)  {
      dataPoints.serialNum = xySerialNum_New;
      ValidateDataPoints(false);  SaveDataPoints();
    }
#endif

  //--- alternate means of setting modbus "coils" ---
    if (regNum == coils31_00_Idx)  {
      for (mask=1,i=0; i<32; ++i,mask<<=1)
        if (coils31_00_New & mask)  NewCoilValue(chanNum, i, 1, cmdRow);
      coils31_00_New = 0;  break;
    }
    if (regNum == coils63_32_Idx)  {
      for (mask=1,i=0; i<32; ++i,mask<<=1)
        if (coils63_32_New & mask)  NewCoilValue(chanNum, i+32, 1, cmdRow);
      coils63_32_New = 0;  break;
    }


    if (regNum == maxPhaseErr_Idx)  {
      if (maxPhaseErr_New < 0.0)  return false;
      applData.maxPhaseErr = maxPhaseErr_New;  ChgWait(1);  break;
    }

    if (regNum == rfLockResetDelay_Idx)  {
      if (rfLockResetDelay_New < 0.0)  return false;
      applData.rfLockResetDelay = rfLockResetDelay_New;  ChgWait(1);  break;
    }


    if (regNum == plusMargin_Idx)  {
      if (plusMargin_New < 0.0)  return false;
      applData.plusMargin = plusMargin_New;  ChgWait(1);  break;
    }
    if (regNum == minusMargin_Idx)  {
      if (minusMargin_New < 0.0)  return false;
      applData.minusMargin = minusMargin_New;  ChgWait(1);  break;
    }

    if (regNum == ignoreHdwErrs_Idx)  {
      applData.ignoreHdwErrs = ignoreHdwErrs_New;  ChgWait(1);  break;
    }

    if (regNum == desiredPhase_Idx)  {
      applData.desiredPhase = desiredPhase_New;  ChgWait(1);  break;
    }

// search for _PERSISTENT_ to see where code normally added for saved values

  } while (false);

  return true;
}


//---------------------------------------------------------------------------
//  Multiple calibration processes can be supported simply by having a
//  different starting state for each one.
//---------------------------------------------------------------------------
void calib_tick()
{
  int  i;
  float  avgPos, movement, avgRef;
  ulong  rdTime, waited, sampleTime;
  static int  nextState, refCount, cancelState;
  static ulong  startTime, waitStart, minWaitMS, maxWaitMS, lastReadTime,
                testStartTime, lastMoveCheckTime;
  static float  minInVolt, maxInVolt, saveSlopVal, refChg, saveCtlVolts,
                saveNoiseValue, returnToPos;
  static float  refVoltHistory[16];
  static int  minOutRaw, maxOutRaw, consecutiveMoves;
  static char  msg[90];


#define  StopBecause(str)  {  \
  strcpy(msg, (str));  calibState = cancelState;  break;  }


  if (recursiveCalib)  return;
  recursiveCalib = true;

//--- wait until we have new values to work with ---
  if (RdTime == lastReadTime)  { recursiveCalib = false;  return; }
  lastReadTime = rdTime;

//--- start a new calibration? ---
  if (startCalibState AND !calibrating)  {
    *msg = 0;  calibState = startCalibState;  cancelState = CALIB_IDLE; }
  startCalibState = CALIB_IDLE;

//--- external cancelation of current calibration? ---
  if (cancelCalib)  {
    if (calibrating)  { strcpy(msg, "Canceled");  calibState = cancelState; }
    cancelCalib = false;
  }


  switch (calibState)  {

    case CALIB_IDLE:
      runStateDesc = "READY   ";  UpdateNoiseData();  break;

  //--------------------------
  //  Determine the minimum reference voltage (voltage driving Servo Amp)
  //  that results in measurable motor movement.
  //--------------------------

    case CALIB_REF_INIT:
      if (BBSCtlMode)  { calibState = CALIB_IDLE;  break; }

      StreamDispMode(MinRefCalibMsg);
      runStateDesc = "MIN REF ";

      consecutiveMoves = 0;  refChg = 0.02;
      returnToPos = posSet;
      NewPosCmd(newPos);
      saveNoiseValue = activeNoise;
      cancelState = CALIB_REF_ABORT;  testStartTime = MS_TIMER;

    //--- make sure we have recent noise data 1st ---
      if (lastNoiseChgTime AND (ElapsedSec(lastNoiseChgTime) < 30))
        calibState = CALIB_REF_START;
      else  {
        ShowMsg(WaitSettleMsg);
        calibState = CALIB_REF_SETTLE;
        activeNoise = 0.0;
        lastNoiseChkTime = MS_TIMER;
        lastAvgPos = AvgPos(5000, 5000);
        calibState = CALIB_REF_SETTLE;
      }
      break;


    case CALIB_REF_SETTLE:  // wait for new noise data
      if (ampIsDisabled OR ampFault)  StopBecause(MotorDisabledMsg);
      if (ElapsedSec(testStartTime) > 60)  {
        strcpy(msg, WouldNotSettleMsg);  calibState = cancelState;  break; }
      if (UpdateNoiseData())  calibState = CALIB_REF_START;
      break;


    case CALIB_REF_START:  // prep for steadily increasing push
      sprintf(msg, "%s %.4f volts", NoiseLevelMsg, activeNoise);
      LogMsg(msg);
      ShowMsg(RefVoltMsg);
      ctlState = CS_STEADY_REF;  // tell ctl loop we want steady ref voltage
      setRefVolt = 0.0;  EnableAmpCmd(true);
      testStartTime = lastMoveCheckTime = MS_TIMER;
      calibState = CALIB_REF_MOVE;  break;


    case CALIB_REF_MOVE:  // check for recent movement
      if (ampIsDisabled OR ampFault)  StopBecause(MotorDisabledMsg);
      if (ElapsedSec(testStartTime) > 120)  StopBecause("Timeout");
      if ((refChg < 0.0) AND minusLimit)  StopBecause(MinusLimitMsg);
      if (setRefVolt > DACMax)  StopBecause(MaxVoltMsg);

      if ((refChg > 0.0) AND plusLimit)  {  // if hit +limit, try other dir
        LogMsg(PlusLimitMsg);
        refChg = -refChg;  calibState = CALIB_REF_START;  break;
      }

      if (elapsedMS(lastMoveCheckTime) < 500)  break;  // 500ms between checks
      lastMoveCheckTime = MS_TIMER;

      movement = (AvgPos(100,100) - AvgPos(600,100)) * 2.0;

      sprintf(msg, CalibDataMsg1,
              consecutiveMoves, setRefVolt, realPos, movement);
      ShowMsg(msg);

      if (fabs(movement) < activeNoise*0.5)  {  // too close to noise level
        setRefVolt += refChg;  consecutiveMoves = 0;  break;  }

    //--- seems we have definite movement ---
      if (++consecutiveMoves < 10)  break;  // has it been consistent?

      applData.minMoveRefVolt = minMoveRefVolt = fabs(setRefVolt);  // YES!!!!
      ChgWait(1);

      sprintf(msg, MinVoltMsg, minMoveRefVolt);  LogMsg(msg);

      calibState = CALIB_REF_STOP;  break;


    case CALIB_REF_ABORT:
      if (*msg)  {  strcat(msg, AbortingCalibMsg);  LogErr(msg); }
      activeNoise = saveNoiseValue;  calibState = CALIB_REF_STOP;

    case CALIB_REF_STOP:
      ShowMsg(ReturnPosMsg);
      NewPosCmd(returnToPos);
      calibState = cancelState = CALIB_IDLE;
      break;


  //--------------------------
  //  Test Mode for tuning PID parameters.  The desired position setting is
  //  controlled by a square wave whose amplitude, offset, and period can be
  //  set as desired.
  //--------------------------

    case CALIB_PID_START:  // prep for square-wave test mode
      cancelState = CALIB_PID_STOP;
      runStateDesc = "PID TUNE";

      if ((testAmpl <= 0.0) OR (testAmpl > 10.0))
        StopBecause(InvalidAmplMsg);
      if ((testOffset < 0.0) OR (testOffset > 10.0))
        StopBecause(InvalidTestOffsetMsg);
      if (testAmpl + testOffset > 10.0)
        StopBecause(InvalidComboMsg);

      if (testPeriod < 100)  {
        LogMsg(PeriodIncrMsg);
        applData.testPeriod = testPeriod = 100;
      }

      LogMsg(StartingTestMsg);
      returnToPos = posSet;  allowRegChgs = true;
      testStartTime = MS_TIMER;
      NewPosCmd(testOffset);
      calibState = CALIB_PID_RUN;  break;


    case CALIB_PID_RUN:  // switch between 2 positions to generate square wave
      if (localMode)  StopBecause(LocalModeMsg);
      if (elapsedMS(testStartTime) % testPeriod < testPeriod>>1)  {
        if (posSet != testOffset)  NewPosCmd(testOffset);
      }
      else
        if (posSet != testOffset + testAmpl)  NewPosCmd(testOffset + testAmpl);
      break;


    case CALIB_PID_STOP:  // end test - restore pos before test began
      if (*msg)  {  strcat(msg, TestStoppedMsg);  LogErr(msg); }
      allowRegChgs = false;
      NewPosCmd(returnToPos);
      calibState = cancelState = CALIB_IDLE;
      break;



    default:
      LogErr(StateErrorMsg);  calibState = CALIB_IDLE;
  }

  lastReadTime = RdTime;
  recursiveCalib = false;
}




//=============================================================================
//                   Functions unique to this application
//=============================================================================

// Use circular buffer to send messages/cmds to the interrupt task that need
// to be serialized (i.e. those for which the order of events is important,
// where simple flags may cause a problem if more than one is set at a time)

//-----------------------------------------------------------------------------
//  funcs called within ComLoop to send data to the ControlLoop process
//-----------------------------------------------------------------------------
void NewPosCmd(float newPos)
{
  eventMsg.msgID = MSG_NewPos;
  eventMsg.msgData.newPos = newPos;
  Buf_Add(&eventMsgQueue, (char *)(&eventMsg), sizeof(eventMsg));
}

//-----------------------------------------------------------------------------
void EnableAmpCmd(bool enableFlag)
{
  eventMsg.msgID = MSG_EnableAmp;
  eventMsg.msgData.enableAmp = enableFlag;
  Buf_Add(&eventMsgQueue, (char *)(&eventMsg), sizeof(eventMsg));
}


//-----------------------------------------------------------------------------
//  Enable Phase Feedback Lock mode, which adjusts the position as needed to
//  maintain a constant phase value
//  returns true on success, false if an error
//-----------------------------------------------------------------------------
bool enablePhaseLock(bool remoteCmd)
{
  char  *msg;

  if (phaseLocked)  return true;

  if (localMode AND remoteCmd)  { LogMsg(LocalOnMsg);  return false; }
  if (driveLock)  { LogMsg(DriveLockOnMsg);  return false; }
  if (!rfOK)  { LogMsg(CannotLockMsg);  return false; }

  ++criticalSection;
  if (newPhaseSetting)  desiredPhase = phaseAngle;
  phaseLocked = true;  restoreSavedPhaseLock = false;
  --criticalSection;

  msg = getBuf(90);
  sprintf(msg, PhaseLockedMsg, desiredPhase, posSet);  LogMsg(msg);
  freeTo(msg);

  return true;
}

//-----------------------------------------------------------------------------
//  disable the Phase Feedback Lock mode
//  returns true on success, false if an error
//-----------------------------------------------------------------------------
bool disablePhaseLock(bool remoteCmd)
{
  char  *msg;


  if (! phaseLocked)  return true;

  if (localMode AND remoteCmd)  { LogMsg(LocalOnMsg);  return false; }
  if (driveLock)  { LogMsg(DriveLockOnMsg);  return false; }

  ++criticalSection;
  phaseLocked = false;  restoreSavedPhaseLock = false;
  --criticalSection;

  msg = getBuf(90);
  sprintf(msg, PhaseUnlockedMsg, posSet);  LogMsg(msg);
  freeTo(msg);

  NewPosCmd(posSet);

  return true;
}


//---------------------------------------------------------------------------
//  Keep tabs on the amount of readback noise we get when the controller
//  is actively maintaining a fixed position.
//
//  There are two steps to computing the noise value:
//
//  - Wait for a period of negligable movement.  This is determined by
//    comparing the Average position over time.  Averaged values are
//    needed to reduce the noise to a point where we can be fairly certain
//    as to whether or not there has been significant movement.
//
//  - Find the minimum and maximum position values recorded over the period
//    checked.  The absolute difference between these is the noise.
//
//  Note that a badly tuned PID loop will cause considerable variation in
//  the motor position and is indistinguishable from the "real" noise.
//  If we could accurately model the precise effect of the PID loop on the
//  readback values, then we could use that model to improve the PID loop so
//  this wouldn't even be a factor!
//
//  Returns true if the noise value is updated.
//---------------------------------------------------------------------------
bool  UpdateNoiseData()
{
  float  avgPos, movement, newNoise, newMargin, err;
  float  *margin;
  bool  updateFlash;


//--- lastNoiseChkTime also set at start of move to insure changes are ---
//--- not made at the beginning of a movement ---
  if (ElapsedSec(lastNoiseChkTime) < 5)  return false;  // 5 sec between tests

//--- skip if not active ctl mode ---
  if (ctlState != CS_ACTIVE_CTL)  return false;

  lastNoiseChkTime = MS_TIMER;

  avgPos = AvgPos(5000, 5000);  // avg pos for last 5 seconds

  movement = fabs(avgPos - lastAvgPos);  lastAvgPos = avgPos;

  if (movement > ADCRes*1.5)  return false;  // too much movement


//--- use new value or do weighted avg with prev value ---
  newNoise = fabs(RecentMovement(5000));
  if (activeNoise == 0.0)
    activeNoise = newNoise;
  else
    activeNoise = (activeNoise*3.0 + newNoise) / 4.0;

//--- if major chg from value in flash, schedule update of flash ---
  if (RelativeDif(activeNoise, applData.activeNoise) > 0.50)  {
    applData.activeNoise = activeNoise;  ChgWait(ONE_HOUR);  }


//--- also adjust the "stop margins" for BB Servo movement ctl ---
  ++criticalSection;
  if ((adjustPlusMargin OR adjustMinusMargin)
       AND !(bbMoveMinus OR bbMovePlus))  {
    err = motorPosSet - avgPos;
    margin = adjustPlusMargin ? &plusMargin : &minusMargin;
    if (reqMovement > 0.0)
      newMargin = *margin - err;
    else
      newMargin = *margin + err;
    if ((newMargin >= 0.0) AND (newMargin < 1.0))
      *margin = (*margin * 3.0 + newMargin) / 4.0;
    adjustPlusMargin = adjustMinusMargin = false;
  }
  --criticalSection;

//--- if significant chg from value in flash, schedule update of flash ---
  if (RelativeDif(plusMargin, applData.plusMargin) > 0.30)  {
    applData.plusMargin = plusMargin;  ChgWait(ONE_HOUR);  }
  if (RelativeDif(minusMargin, applData.minusMargin) > 0.30)  {
    applData.minusMargin = minusMargin;  ChgWait(ONE_HOUR);  }


  lastNoiseChgTime = MS_TIMER;  return true;
}


//-----------------------------------------------------------------------------
//  return ptr to a string describing the current contol state
//-----------------------------------------------------------------------------
char *CtlStateDesc()
{
  char *stateStr;


  stateStr = NULL;

  if (hdwCtlErr)       stateStr = "HDCTLERR";  else
  if (testingCIBGain)  stateStr = "CIB-GAIN";  else
  if (calibrating)     stateStr = "CALIBRAT";  else
  if (driveLock)       stateStr = "DRVLOCKD";  else
  if (phaseLocked)     stateStr = "PHA_LOCK";  else
  if (oscillating)     stateStr = "OSCLTERR";  else
  if (ampIsDisabled OR ampFault)
    stateStr = (hardDisabled ? "HARD_DIS" : "SOFT_DIS");  else
  if (idleRest)        stateStr = "IDLEREST";

  if (stateStr)  {  // one of the above conditions matched...
    if (blinkOff)  stateStr = "        ";
    return stateStr;
  }

//--- no higher priority msg to show, so just show the normal state ---
  switch (ctlState)  {
    case CS_IDLE:        stateStr = "IDLE    ";  break;
    case CS_STEADY_REF:  stateStr = "FIXEDREF";  break;
    case CS_START_MOVE:  stateStr = "NEW POS ";  break;
    case CS_ACTIVE_CTL:  stateStr = "ACTIVE  ";  break;
    default:  stateStr = blinkOff ? "        " : "UNKNOWN ";
  }

  return stateStr;
}


//----------------------------------------------------------------------------
//  Determine the actual position of the device being controlled based on the
//  current position of the motor and the slop zone the motor moves within
//  and update the value of slopPos.
//
//  NOTE:  The "actual" position of the device is defined as the center of
//         the slop zone.
//----------------------------------------------------------------------------
float ActualPos(float motor_pos)
{
  if (slopValue == 0.0)  {
    slopPosKnown = true;  return (slopPos = motor_pos);  }


//  NOTE:  While it is true that, in reality, the REAL slopPos (center of
//         the of slop zone) and motorPos values cannot differ more than
//         halfSlopVal, it is possible for the assumed value of slopPos
//         (= motorPos at startup) and the real motorPos values to vary
//         by twice that (e.g. assume the motor is against one edge of the
//         slop zone at startup.  Then it has to move a full slopValue amount
//         from its original position - the assumed slopPos value, starting
//         off wrong by halfSlopVal - before we know for certain the true
//         value of slopPos).

//--- check to see if we actually moved the device (for sure) yet ---
  if (!slopPosKnown)  {
  //--- touch negative edge (for sure) yet? ---
    if ((motor_pos < slopPos) AND (slopPos-motor_pos > slopValue)) {
      slopPos = motor_pos + halfSlopVal;  slopPosKnown = true;  }
  //--- touch positive edge (for sure) yet? ---
    if ((motor_pos > slopPos) AND (motor_pos-slopPos > slopValue)) {
      slopPos = motor_pos - halfSlopVal;  slopPosKnown = true;  }
  }

//--- if still don't know slop position, then just return orig motor pos ---
  if (!slopPosKnown)  return slopPos;

//--- did we actually move the device in the neg dir? ----
  if (motor_pos < slopPos-halfSlopVal)
    return (slopPos = motor_pos + halfSlopVal);

//--- did we actually move the device in the pos dir? ---
  if (motor_pos > slopPos+halfSlopVal)
    return (slopPos = motor_pos - halfSlopVal);

//--- just moved within the slop zone, so return last know position ---
  return slopPos;
}

//----------------------------------------------------------------------------
//  based on current position of the motor and the slop zone, determine the
//  position the motor needs to be moved to in order to move the device to
//  the specified position
//
//  NOTE:  The "actual" position of the device is defined as the center of
//         the slop zone.
//----------------------------------------------------------------------------
float NeededPos(float desiredPos)
{
  if (slopValue == 0.0)  return desiredPos;

//--- If we don't know where the motor position is within the slop zone, ---
//--- and the max needed movement is less than the amount of slop, then we ---
//--- don't really know how far to move the motor, so just say to move it ---
//--- to the specified position, which will result in an error no larger ---
//--- than half the slopValue ---
  if (!slopPosKnown AND (fabs(desiredPos - slopPos) < slopValue))
    return desiredPos;

//--- We know where the device is and the motor pos relative to the slop ---
//--- zone so we can determine how much to move the motor and in which ---
//--- direction in order to push the device to the desired position ---

//--- move device in neg dir? ---
  if (desiredPos < slopPos)  return desiredPos - halfSlopVal;

//--- move device in pos dir? ---
  if (desiredPos > slopPos)  return desiredPos + halfSlopVal;

//--- no change so just stay at last read position ---
  return motorPos;
}

//----------------------------------------------------------------------------
//  Return the amount of recent motor movement.  The motor position history
//  stores a new value every 10ms so that is the max resolution.
//----------------------------------------------------------------------------
float RecentMovement(ushort msPeriod)
{
  int  i, x, minIdx, maxIdx, numSamples;
  uint  minPos, maxPos;
  float  voltChg;


  numSamples = msPeriod / 10;
  if (numSamples > HistorySize)  numSamples = HistorySize;

//--- determine starting index ---
  x = lastHistIdx - numSamples;
  if (x < 0) x += HistorySize;

  minPos = maxPos = (uint)motorHistory[x];
  minIdx = maxIdx = 0;

  for (i=0; i<numSamples; ++i)  {
    if ((uint)motorHistory[x] < minPos)  {
      minPos = (uint)motorHistory[x];  minIdx = i;  }
    if ((uint)motorHistory[x] > maxPos)  {
      maxPos = (uint)motorHistory[x];  maxIdx = i;  }
    if (++x >= HistorySize)  x = 0;
  }

  voltChg = ADCtoVolts(0, maxPos) - ADCtoVolts(0, minPos);
  if (maxIdx < minIdx)  voltChg = 0.0 - voltChg;

  return voltChg;
}


//----------------------------------------------------------------------------
//  Return the average pos value for the motor.  The motor position history
//  stores a new value every 10ms so that is the max resolution.
//----------------------------------------------------------------------------
float AvgPos(ushort msStartBack, ushort msPeriod)
{
  int  i, x, minIdx, maxIdx, numSamples, startIdx;
  ulong  posSum;
  float  avgPos;


//--- determine # of samples ---
  numSamples = msPeriod / 10;
  if (numSamples < 2)  return ADCtoVolts(0, motorHistory[lastHistIdx]);

  if (numSamples > HistorySize)  numSamples = HistorySize;

//--- determine starting index ---
  x = msStartBack / 10;
  if (x > HistorySize)  x = HistorySize;
  if (x > numSamples)  numSamples = x;
  x = lastHistIdx - x;
  if (x < 0) x += HistorySize;

  for (i=0,posSum=0; i<numSamples; ++i)  {
    posSum += (ulong)motorHistory[x];
    if ((++x) >= HistorySize)  x = 0;
  }

  avgPos = (uint) (posSum / numSamples);

  return ADCtoVolts(0, avgPos);
}


//-----------------------------------------------------------------------------
//  Update ctlVolts and other values used by the caller to set the next state
//  of the device.
//
//  !!!!! WARNING !!!!!
//
//  This function is called by the timer-interrupt driven function that reads
//  and writes the Analog and Digital I/O values.  It should be limited to
//  processing current input values and updating output values that will become
//  the active values on the next interrupt (currently, this interuupt occurs
//  every 10ms, so this routine must take considerably less time than that to
//  do EVERYTHING it needs to).  Note that formatting data for display (e.g.
//  sprintf calls) is very time consuming and should be avoided as much as
//  possible.
//
// NOTES:
// (*)  It is important to ignore duplicate requests to stop the motor from
//      being moved when switching from REMOTE mode to LOCAL and back.
//-----------------------------------------------------------------------------
void ComputeNextState()
{
  float  scaleBy, posChg, curErr;


#define  AvgRealPos  (ActualPos(AvgPos(1000,1000)))


  if (driveLock OR idleRest OR chgdToRemote)  {
    if (driveLock)  { newPos = posSet = AvgRealPos;  useNewPos = true; }
    if (driveLock OR idleRest)  { ctlState = CS_IDLE;  enableAmp = false; }
    bbMovePlus = bbMoveMinus = false;  reachedGoal = true;
    pFactor = iFactor = dFactor = setRefVolt = 0.0;
    ctlVolts = 0.0;
    return;
  }

  if (testingCIBGain)  return;

  if (chgdToLocal)  {  // just switched to local mode?
    enableAmp = true;  hdwCtlErr = idleRest = oscillating = false;
    oscillCount = 0;
  }


//--- if motor under manual ctl at device, reset ctl values ---
//--- NOTE:  The RF Bang-Bang Servos don't HAVE a motor local switch so we ---
//---        can't tell the difference between drift and movement via the ---
//---        switches at the motor control box. ---
//---        To allow for this, the NIM module should be set to Disabled ---
//---        before moving the motor via the motor control box.  If this ---
//---        is not done, then posSet will NOT follow the current position ---
//---        and the control system will act based on a setpoint that no ---
//---        longer matches the current position (which will result in the ---
//---        motor moving back to the old set point when the controller is ---
//---        taken out of local mode) ---
  if (motorLocal OR (BBSCtlMode AND (ampIsDisabled OR ampFault)))  {
    bbMovePlus = bbMoveMinus = false;
    pFactor = iFactor = dFactor = setRefVolt = 0.0;
    posSet = AvgRealPos;  motorPosSet = NeededPos(posSet);
  }

//--- if new goal position, make sure amp isn't disabled due to previous ---
//--- detection of a hardware/controller error condition ---
  if (useNewPos)  enableAmp = true;



//===== RF Servo controller connected to DC or BB Servo =====
//  When connected to a DC Servo, we have direct control of the ref/velocity
//  voltage fed to the amplifier that drives the motor and can read the state
//  of the amplifier.
//  When connected to a Bang-Bang servo, we are limited to the 2 movement bits.
//  In both cases, there are 2 local modes:  A software one invoked at the NIM
//  module, where we retain direct control over movement of the motor, and
//  a hardware one (at or near the motor), where we can only monitor what the
//  motor is doing.
//=====================================

  do  {
    if (hardDisabled OR ampFault OR motorLocal)  break;  // ctl vals reset above
    if (softDisabled AND !useNewPos)  break;

  //--- Phase Feedback mode active - move motor in dir needed ---
    if (phaseLockActive OR phaseLockPaused)  {
      bbMovePlus = bbMoveMinus = false;
      pFactor = iFactor = dFactor = setRefVolt = 0.0;

      if (useNewPos)  { posSet = newPos;  useNewPos = false; }

      if (phaseLockActive)  {
      //--- stop correcting once we pass the center of the deadzone ---
        curErr = desiredPhase - phaseAngle;
        if (SIGN(curErr) != SIGN(phaseErr))  reducePhaseErr = false;

        if (! reducePhaseErr)  break;

        if (fastAdjust)  scaleBy = knobRateF;  else  scaleBy = knobRateS;
        if (BBSCtlMode)  {
          if (phaseErr > 0.0)   bbMovePlus = true;   else
          if (phaseErr < 0.0)  bbMoveMinus = true;
          if (bbMovePlus OR bbMoveMinus)  correctingPos = true;
        }
        else  {  // finish - need smarter algorithm to relate error to velocity
          setRefVolt = phaseErr * knobRateS;
        }
      }

      ctlState = CS_STEADY_REF;  break;
    }


  //--- Remote mode: achieve/maintain current set point ---
    if (!localMode)  { UpdateCtlValues();  break; }

  //--- Local mode, but no changes via the NIM front panel this pass ---
  //--- In this case, reset ctl values and prep to maintain new pos ---
    if (! manualMoveCmd)  {
      if (!localChgLastPass)
        updateCtlValues();
      else {
        bbMovePlus = bbMoveMinus = false;
        pFactor = iFactor = dFactor = setRefVolt = 0.0;
        ctlState = CS_END_LOCAL_CHG;  localChgLastPass = false;
      }
      break;
    }


  //===== Local mode and changes via the NIM front panel =====

  //--- if SET button is pressed, apply changes to set point ---
    if (setButton)  {
      if (fastAdjust)  scaleBy = knobRateF;  else  scaleBy = knobRateS;

      if (movePlusSw)   posChg = scaleBy * switchRate * setRate / 100.0;
      if (moveMinusSw)  posChg = scaleBy * switchRate * setRate / -100.0;

      if (knobTicks)  posChg = (float) knobTicks * scaleBy * setRate / 100.0;

      newPos = posSet + posChg;  useNewPos = true;

      UpdateCtlValues();  break;
    }


  //--- SET button not pressed, so apply changes to motor velocity ---
    bbMovePlus = bbMoveMinus = false;
    pFactor = iFactor = dFactor = setRefVolt = 0.0;

    if (fastAdjust)  scaleBy = knobRateF;  else  scaleBy = knobRateS;
    if (BBSCtlMode)  {
      if (movePlusSw)   bbMovePlus = true;   else
      if (moveMinusSw)  bbMoveMinus = true;  else
      if (knobTicks > 0)  bbMovePlus = true;   else
      if (knobTicks < 0)  bbMoveMinus = true;
    }
    else  {
      setRefVolt = (float) knobTicks * scaleBy;
      if (movePlusSw)   setRefVolt =  10.0 * scaleBy / knobRateF * switchRate;
      if (moveMinusSw)  setRefVolt = -10.0 * scaleBy / knobRateF * switchRate;
    }

    localChgLastPass = true;  ctlState = CS_STEADY_REF;

  } while (false);


  ctlVolts = setRefVolt;
  if (ctlVolts > DACMax)  ctlVolts = DACMax;
  if (ctlVolts < DACMin)  ctlVolts = DACMin;
}


//-----------------------------------------------------------------------------
//  Update the values that control the motor movement.
//
//  For the PID (Proportional-Integral-Derivative) algorithm, the voltage
//  feed to the Servo Amplifier is computed.  The control (or reference)
//  voltage is basically a velocity vector, specifying direction (by +/-
//  polairty) and speed (via amplitude)
//
//  For Bang-Bang Servos, all movement is controlled by two drive bits, each
//  one controlling motor movement in one direction.
//
//  !!!!! WARNING !!!!!
//
//  This function is called by the timer-interrupt driven function that reads
//  and writes the Analog and Digital I/O values.  It should be limited to
//  processing current input values and updating output values that will become
//  the active values on the next interrupt (currently, this interuupt occurs
//  every 10ms, so this routine must take considerably less time than that to
//  do EVERYTHING it needs to).  Note that formatting data for display (e.g.
//  sprintf calls) is very time consuming and should be avoided as much as
//  possible.
//
//  Also, any diagnostic output that is generated must be kept to a minimum or
//  it will quickly flood the output stream.
//-----------------------------------------------------------------------------
void UpdateCtlValues()
{
  float  posDiff, absDiff, lastIFactor, absRefVolts, margin;
  bool  sufficientForce, reversedForce, gotCloser;
  char  *msg;
  static bool  passedGoal, lastMoveMinus, lastMovePlus;
  static ulong  lastTimeCondsOK;
  static float  posErr0, posErr1, posErr2, ctlVal0, ctlVal1, lastRefVolt,
                closestSoFar;


  if (useNewPos)  ctlState = CS_START_MOVE;

  msg = getBuf(90);


  switch (ctlState)  {

    case CS_IDLE:  // nothing to do for now....
      reqMovement = 0.0;  trackOverShoot = false;
      bbMovePlus = bbMoveMinus = false;
      setRefVolt = 0.0;  pFactor = iFactor = dFactor = 0.0;
      reachedGoal = true;
      break;


    case CS_STEADY_REF:  // maintain steady ref voltage
      trackOverShoot = false;
      break;


    case CS_END_LOCAL_CHG:  // local mode just after chgs stopped
      newPos = realPos;

    case CS_START_MOVE:  // new position requested
      if (newPos < MinPos)  newPos = MinPos;
      if (newPos > MaxPos)  newPos = MaxPos;

      posSet = newPos;  useNewPos = false;

      motorPosSet = NeededPos(posSet);
      posDiff = reqMovement = motorPosSet - motorPos;

      if (BBSCtlMode)  {
        bbMoveMinus =  BOOL(posDiff < -minusMargin);
        bbMovePlus =   BOOL(posDiff >  plusMargin);
        lastMoveMinus = lastMovePlus = false;
      }
      else  {
        posErr0 = posDiff;
        posErr1 = 0.0;

        pFactor = pWeight * posErr0;
        iFactor = 0.0;
        dFactor = 0.0;
        setRefVolt = lastRefVolt = ctlVal0 = pFactor;
      }

      enableAmp = true;
      oscillating = hdwCtlErr = idleRest = passedGoal = reachedGoal = false;
      oscillCount = 0;

      if (BBSCtlMode)
        sufficientForce = (bbMoveMinus OR bbMovePlus);
      else
        sufficientForce =
           ((fabs(setRefVolt) > minMoveRefVolt) AND (minMoveRefVolt > 0.05));

      closestSoFar = fabs(posDiff);

      lastNoiseChkTime = lastTimeCondsOK = lastMoveCmdTime = MS_TIMER;

      overShoot = 0.0;  trackOverShoot = true;

      if (slopPosKnown)
        if (reqMovement < 0)
          adjustMinusMargin = true;
        else
          adjustPlusMargin = true;

      ctlState = CS_ACTIVE_CTL;
      break;


    case CS_ACTIVE_CTL:  // update control voltage based on new state
      if (ampIsDisabled OR ampFault)  {
        bbMovePlus = bbMoveMinus = adjustPlusMargin = adjustMinusMargin = false;
        setRefVolt = 0.0;  pFactor = iFactor = dFactor = 0.0;
        ctlState = CS_IDLE;  break;
      }


    //--- hit a limit switch appropriate for the dir of movement? ---
      if ( ((reqMovement > 0.0) AND plusLimit) OR
           ((reqMovement < 0.0) AND minusLimit) )  {
        adjustPlusMargin = adjustMinusMargin = false;
        if (! ignoreHdwErrs)  {
          sprintf(msg, LimitSwitchMsg, realPos);  LogMsg(msg);
          motorPosSet = motorPos;  reqMovement = 0.0;  reachedGoal = true;
        }
      }

      posDiff = motorPosSet - motorPos;  absDiff = fabs(posDiff);

      if (BBSCtlMode)  {
        sufficientForce = (bbMoveMinus OR bbMovePlus);
        reversedForce =
           (bbMoveMinus AND lastMovePlus) OR (bbMovePlus AND lastMoveMinus);
        lastMoveMinus = bbMoveMinus;  lastMovePlus = bbMovePlus;
      }
      else  {
        sufficientForce =
           ((fabs(setRefVolt) > minMoveRefVolt) AND (minMoveRefVolt > 0.05));
        reversedForce = (SIGN(setRefVolt) != SIGN(lastRefVolt));
        lastRefVolt = setRefVolt;
        ctlVal1 = ctlVal0;
        posErr2 = posErr1;  posErr1 = posErr0;  posErr0 = posDiff;
      }
      gotCloser = (absDiff < closestSoFar);


    //--- if trying to move AFTER having reached our goal, then it must be ---
    //--- due to a need to correct for drift or overshoot ---
      if (reachedGoal AND sufficientForce)  correctingPos = true;

    //--- determine if we have passed and/or reached our goal ---
      if (SIGN(posDiff) != SIGN(reqMovement))  passedGoal = true;
      if ((absDiff <= halfADCRes) OR passedGoal)  if (! reachedGoal)  {
        lastCorrectingMsgTime = MS_TIMER;  reachedGoal = true;  }

    //--- keep track of most recent time nominal conditions existed ---
      if (gotCloser OR reversedForce OR !sufficientForce OR reachedGoal)  {
        closestSoFar = absDiff;  lastTimeCondsOK = MS_TIMER;  }

    //--- keep track of how far we overshoot or wander from goal pos ---
      if (trackOverShoot AND passedGoal)
        if (absDiff > overShoot)  overShoot = (overShoot + absDiff) / 2.0;

    //--- if device not responding as expected, stop trying until new pos ---
    //--- command received ---
    //--- NOTE this is also meant to catch miswired or bad limit switches ---
      if ((elapsedMS(lastTimeCondsOK) > MaxStickTime)
          AND (calibState != CALIB_REF_SETTLE))  {
        adjustPlusMargin = adjustMinusMargin = false;
        if (! ignoreHdwErrs)  {
          if ((plusLimit OR minusLimit) AND !reachedGoal) // (hitting expected limit handled above)
            LogErr(BackwardMsg);
          else
            LogErr(HdwCtlErrMsg);
          enableAmp = false;  hdwCtlErr = true;  ctlState = CS_IDLE;  break;
        }
      }

  //--- limit time we actively control motor after getting a new setpoint ---
    if (maxActiveSecs > 0)
      idleRest = (ElapsedSec(lastMoveCmdTime) > maxActiveSecs);


    //--- update values that control motor movement ---

      if (BBSCtlMode)  {
        if (reachedGoal)
          bbMoveMinus = bbMovePlus = false;
        else  {
          bbMoveMinus = BOOL(posDiff < -minusMargin);
          bbMovePlus =  BOOL(posDiff >  plusMargin);
        }
      }
      else  {
      /*--- velocity-based algorithm (does not suffer from Integral windup...) ---
        pFactor = pWeight * posErr0;
        iFactor = iWeight * (posErr0 - 2.0*posErr1 + posErr2) * 0.01;
        dFactor = dWeight * (posErr0 - posErr1) / 0.01;

        ctlVal0 = ctlVal1 + pFactor + iFactor + dFactor;

        if (ctlVal0 < -10.0)  ctlVal0 = -10.0;
        if (ctlVal0 > 10.0)   ctlVal0 =  10.0;
      ------------*/

      /*--- positional-based algorithm (requires Integral windup check) ---*/
        if (iWeight == 0.0)  iFactor = 0.0;
        lastIFactor = iFactor;

        pFactor = pWeight * posErr0;
        iFactor = iFactor + iWeight * posErr0 * 0.01;
        dFactor = dWeight * (posErr0 - posErr1) / 0.01;

        ctlVal0 = pFactor + iFactor + dFactor;

        if (ctlVal0 < -10.0)  { ctlVal0 = -10.0;  iFactor = lastIFactor; }
        if (ctlVal0 >  10.0)  { ctlVal0 =  10.0;  iFactor = lastIFactor; }

        setRefVolt = ctlVal0;
      }

      break;


    default:
      break;
  }

  freeTo(msg);
}


//-----------------------------------------------------------------------------
//  Handle any device-specific device special file open requests
//-----------------------------------------------------------------------------
int DLP_DevFile_Open(LinkProc *lp, int argc, char *argv[])
{
  return -999;
}




//============================================================================
main()
{
  int  i;
  char  *msg;


  firstTime = true;

  while (true)  {
    CommonInit();

    msg = getBuf(128);   // do NOT free up - "backup" in case of failed getBuf()

xx    #if (_BOARD_TYPE_ != RCM3200)
xx      ioBoardType = SmartCat;
xx    #else
xx      ioBoardType = NSCL_NIM_RFServo;
xx    #endif


xx    digOutBootState = AMP_DISABLE_BIT | MOVE_BITS;


xx    CommonHdwInit(digOutBootState);  DLPHdwInit();

    ctlMode = CM_Unknown;
    if (!(digitalIn & BBS_JUMPER))  ctlMode = CM_DigDirection;
    if (!(digitalIn & DCS_JUMPER))  ctlMode = CM_AnalogVelocity;

    runStateDesc = "READY   ";


xx    DLPInit();

  //========== Remaining Device-specific Initialization ==========


    ctlState = CS_IDLE;  reqMovement = 0.0;
    devInitialized = false;


  //--- read current position and set requested pos to the same value ---
    for (i=0; i<NumDAC; ++i)  voltsOut[i] = 0.0;
    realPos = PosVolts;
    voltsOut[0] = 0.0;  // so motor won't immead move
    hdwCtlErr = idleRest = oscillating = false;
    oscillCount = 0;

    cibGain = 200.0;

    slopPos = realPos = PosVolts;  slopPosKnown = false;

    for (i=0; i<HistorySize; ++i)  motorHistory[i] = adcRaw[0];
    lastHistIdx = (int)((MS_TIMER/10L) % HistorySize);

    overShoot = 0.0;  halfADCRes = ADCRes / 2.0;

  #define RestoreReg(regName)  regName = applData.regName

  //--- set initial register values from values read from flash ---
    slopValue = applData.slop;  halfSlopVal = slopValue / 2.0;

    RestoreReg( pWeight );
    RestoreReg( iWeight );
    RestoreReg( dWeight );

    RestoreReg( knobRateF );
    RestoreReg( knobRateS );
    RestoreReg( setRate );
    RestoreReg( switchRate );
    RestoreReg( dcDamp );
    RestoreReg( maxActiveSecs );

    RestoreReg( testAmpl );
    RestoreReg( testOffset );
    RestoreReg( testPeriod );

    RestoreReg( cibGain );

    RestoreReg( activeNoise );
    RestoreReg( minMoveRefVolt );

    RestoreReg( rfLockResetDelay );

    RestoreReg( maxPhaseErr );

    RestoreReg( plusMargin );
    RestoreReg( minusMargin );

    RestoreReg( ignoreHdwErrs );

    RestoreReg( desiredPhase );

  // search for _PERSISTENT_ to see where code normally added for saved values

    #ifdef UseSpline
       xySerialNum = dataPoints.serialNum;
       x1 = dataPoints.xyCoords[0][0];   y1 = dataPoints.xyCoords[0][1];
       x2 = dataPoints.xyCoords[1][0];   y2 = dataPoints.xyCoords[1][1];
       x3 = dataPoints.xyCoords[2][0];   y3 = dataPoints.xyCoords[2][1];
       x4 = dataPoints.xyCoords[3][0];   y4 = dataPoints.xyCoords[3][1];
       x5 = dataPoints.xyCoords[4][0];   y5 = dataPoints.xyCoords[4][1];
       x6 = dataPoints.xyCoords[5][0];   y6 = dataPoints.xyCoords[5][1];
       x7 = dataPoints.xyCoords[6][0];   y7 = dataPoints.xyCoords[6][1];
       x8 = dataPoints.xyCoords[7][0];   y8 = dataPoints.xyCoords[7][1];
       x9 = dataPoints.xyCoords[8][0];   y9 = dataPoints.xyCoords[8][1];
      x10 = dataPoints.xyCoords[9][0];  y10 = dataPoints.xyCoords[9][1];
    #endif

    lastNoiseChgTime = lastCorrectingMsgTime = 0;
    lastNoiseChkTime = lastMoveCmdTime = lastEnableTime = lastTimeRFOK = MS_TIMER;
    lastCalibGainTime = 0;
    stateAfterSettled = CALIB_IDLE;

    lockPressedTime = 0;  newPhaseSetting = false;

    motorPos = PosVolts;  realPos = ActualPos(motorPos);
    sprintf(msg, StartupPosMsg, realPos);  LogMsg(msg);

    posReq = lastPosReq = posSet = motorPosSet = lastAvgPos = realPos;

    driveLock = phaseLocked = phaseLockInactive = reducePhaseErr = false;
    movePlusSw = moveMinusSw = bbMovePlus = bbMoveMinus = correctingPos = false;
    localMode = ctlrLocal = motorLocal = wasLocalMode = false;
    chgdToLocal = chgdToRemote = newPhaseLockReq = false;
    plusLimit = minusLimit = atALimit = ampFault = localChgLastPass = false;
    adjustPlusMargin = adjustMinusMargin = ignoreSettings = false;

    lastPhaseChgTime = MS_TIMER;

    newCorrectingMsg = reachedGoal = true;

    ctlVolts = voltsOut[0];

    testingCIBGain = gainComputed = false;

    newPos = realPos;  useNewPos = enableAmp = true;

    if (applData.lastUp)  restoreSavedPhaseLock = true;

    minStackPtr = 0xffff;  maxStackUsed = 0;

    controlEnabled = true;  //initiate active control of the device


    while (ComLoop());

    freeTo(msg);  firstTime = false;
  }
}

