//----- testAppl.c ----- 08/11/10 -----(03/05/06)----

// Sections that typically require application-specific code are marked with
// the comment:  //APPLCODE

//-----------------------------------------------------------------------------
//  Application specific code for the main application task that is run at
//  a fixed interval and application specific commands run on demand.
//
//APPLCODE
//  Application code to ...
//-----------------------------------------------------------------------------
//
//  2010-08-11 - Mark Davis
//               - Created test "application" to use as a place to put code
//                 that is only needed to test other modules (i.e. commands
//                 that sever little or no purpose on real device controllers
//                 and just take up space).
//
//  2009-12-14 - Mark Davis
//               - Changed order of param files to deal with problems caused
//                 by the order in which they were restored at startup.
//
//  2007-03-14 - Mark Davis -
//               - added call to rwIO() (replaces "invisible" code in the
//                 sharedAppl.c file that did this for all applications).
//
//  2007-03-04 - Mark Davis -
//               - reworked X-Macro usage to allow application-specific
//                 register definitions to reside in the same file as the
//                 rest of the application-specific code.
//
//  2006-11-14 - Mark Davis -
//               - original version started based on ps8rack code
//
//-----------------------------------------------------------------------------
/*
  Checklist for adding apaplication-specific functionality:

  Sections that application-specific code goes is generally marked with a
  comment line like this:

  //APPLCODE - ... descr of what to do ...


  Print this section out and remove it after all items completed and tested:

  __ Write Software Requirements document

  __ Add summary of functionality and operational overview to top of source
     code file (just after the update log)

  __ Set applName[] and ApplAbbrev[] values near top of file

  __ Set NUM_APPL_AIN, _AOUT, _DIN, _DOUT values

  __ Define digital inputs/outputs:
       - #define name for each bit
       - LogicOrderXXX[] for the bits to group them in the most readable order
         in the ios command
       - DigXXXDescr[] 0 and 1 state description for each bit whose value
         changes we want logged
       - set InitialDigOutState value to specify state the digitalOut bits
         should be set to on startup

  __ Define analog inputs/outputs:
       - #define name for each signal used
       - LogicOrderXXX[] for the bits to group them in the most readable order
         in the ios command

  __ Define ApplState structure such that it contains all the data
     needed to define the current state of the device and the control logic.
     The code uses 3 instances of this structure:  Prev, Cur, Next (see
     description below on how these are used)

  __ Add #defines for bit-mapped ApplState values

  __ Add COIL0(), REG3(), and REG4() macros to define global variables that
     need to be accessible via the ModBus interface.  Even if a value does
     not need to accessed by an external client, this is often the easiest
     way to define values that can be examined and changed at run time.

  __ Add #defines for bit-mapped register values

  __ Set default values for labels (short descirptions) for each of the
     digital and analog inputs and outputs used (in setDefaultLabels() func)

  __ Set the DevStateStr[] text for showing the basic state of the controller
     (via the "state" command), and add entries to the xxxElems[] "tables" to
     manage basic formatting of digital and analog values.  Add code as needed
     to the showState() function to handle any formatting that the cannot be
     handled by the basic xxxElems[] table entries.


This is a good point at which to test what you have done so far.  Compile the
code and load it on a controller and make sure the state and ios commands
show the right values and that the descriptions and labels agrees with the
state of each input, that the correct messages are logged, etc.


  __ Add devDataParams[] for configuration values that need to be persistent
     (i.e. those that are configurable without changing the code and need to
     be restored at startup).  Values in this structure are saved as a group
     to a file (on command) to the flash file system on the controller, and
     that file is automatically backed up to an external server.

  __ Add setingsParams[] for run-time values that may be frequently changed
     at run time and that we want to be restored on startup.  Changes to
     the values in this structure are automatically saved to the flash file
     system on occasion (the rate varies based on how often changes occur to
     minimize wear on the flash memory).  This file is NOT (currently) backed
     up, but it probably will be in the future.

  __ Add any global variables needed to the SharedGlobals section.  Remember
     that any global values accessed by multiple tasks will need to use some
     form of locking to avoid concurrency issues.

  __ Write MB_regName() functions for any values defined as ModBus registers
     that require special handling when they receive a new value.  The usual
     functions for handling register value changes (e.g. setU16, setF32, etc)
     only insure that the value is locked before changing it.  If a new value
     requires ANYTHING else (e.g. a range check, or possibly comparing to
     or altering other values), then a custom function must be written to
     handle it (and don't forget to include locking for any global values
     referenced in the function - the usual methods being KLOCK()/KUNLOCK()
     or - for REALLY short pieces of code - _EnterCritical_()/_ExitCritical_() ).

  __ Write xxxRestore() functions for any "persistent" values that need
     special handling when they are restored from a file.  This can happen
     at run time if a controller has no copy of the file used to save the
     values or a backup version of a file is newer.  Again, don't forget to
     lock values against concurrent access as needed.

  __ For global values that may be changed by more than one task (e.g. the
     applic and the modbusTCPd tasks), it often makes the most sense to
     write a single function that can be called by any task to safely change
     the value.  For register values, these functions would be called by
     the MB_regName() functions.

     Write functions as needed for such values and place them just before the
     updateCurState() function and add forward declarations near the top of
     the source file (to satisfy the need for prototypes in the MB_regName()
     functions).

  Much more details in the comments preceeding the following functions, but
  the basic idea is that the applic task runs every 10ms and calls the
  following functions:
    rwIO():
       - get current state of digital and analog input signals
       - update state of output signals using values computed in previous pass
    updateCurState():
       - interpret the meaning of the latest values
    updateNextState():
       - decide what changes need to be made
    updateOutputState():
       - apply changes to values used to set output signals in rwIO()
    updateExternals():
       - update state of values accessed by other tasks (in particular, the
         values accessed via external clients through processes such as
         modbusTCPd)

  __ Add code as needed to implement the application login in updateCurState(),
     updateNextState(), updateOutputState(), and updateExternals()

  __ Write "_task" functions as needed to implement applications-specific
     commands that can be executed from a login shell and add them to the
     applCmds[] table at the end of the source file.

     Many applications use a single "cmd" task and various parameters to
     handle most or all the application-specific commands.  Following this
     standard helps make controllers easier to use for everyone.

*/
//-----------------------------------------------------------------------------

#include <common.h>

#include <unistd.h>
#include <sys/types.h>

#include <semaphore.h>
#include <sys/time.h>


//APPLCODE - change to match the name of this file
#ifndef  ApplicSource
  #define  ApplicSource  "testAppl.c"
#endif

#ifndef DidRegDefs
  #define  DidRegDefs
  #include <modbusRegs.c>
#endif
#ifndef RegDefs


//APPLCODE - change to the application name
const char far applName[] = { "Test" };
const char ApplAbbrev[] = { "Test" };


//APPLCODE - add ID to unistd.h and place it here
#define DeviceType  DT_Test

//APPLCODE - set the # of input and outputs used
#define  NUM_APPL_AIN      8   // analog inputs
#define  NUM_APPL_AOUT     4   // analog outputs
#define  NUM_APPL_DIN     16   // digital inputs
#define  NUM_APPL_DOUT    16   // digital outputs

#define  ADC_MASK_APPL  0x000000FFL
#define  DAC_MASK_APPL  0x0000000FL

#define SharedINC
#include <sharedAppl.c>



//=============================================================================
//  definitions related to use of the Digital and Analog I/O signals
//=============================================================================

//APPLCODE - define macros for each Digital bit and each Analog index

//=============== digitalIn bit usage ===============
/*EXAMPLE:
#define  CLOSED_LOOP     0x00000001  // 00: 0=Open Loop, 1=Closed Loop
#define  FINAL_BTN       0x00000002  // 01: 1=SET button being pushed
#define  INITIAL_BTN     0x00000004  // 02: 1=INITIAL button being pushed
#define  RF_IS_OFF       0x00000008  // 03: 1=RF is OFF, 0=RF is ON
#define  LOC_REM_SW      0x00000010  // 04: 1=Local, 0=Remote
*/

const uchar  LogicOrderDIn[NUM_APPL_DIN] =  {
   0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15 };

/*EXAMPLE:
#define  ClosedLoop(din)        ( (bool)((din) & CLOSED_LOOP) )
#define  FinalBtnPushed(din)    ( (bool)((din) & FINAL_BTN) )
#define  InitialBtnPushed(din)  ( (bool)((din) & INITIAL_BTN) )
#define  RFisOff(din)           ( (bool)((din) & RF_IS_OFF) )
#define  LocalMode(din)         ( (bool)((din) & LOC_REM_SW) )
*/


//--- msgs to log automatically when a digital input changes state ---
//--- use NULL for bits that we don't want to log changes for ---
const char far *digInDescr[NUM_APPL_DIN*2] = {
//--0 state descr--    --1 state descr--
  "DI-00 now 0",       "DI-00 now 1",
  "DI-01 now 0",       "DI-01 now 1",
  "DI-02 now 0",       "DI-02 now 1",
  "DI-03 now 0",       "DI-03 now 1",
  "DI-04 now 0",       "DI-04 now 1",
  "DI-05 now 0",       "DI-05 now 1",
  "DI-06 now 0",       "DI-06 now 1",
  "DI-07 now 0",       "DI-07 now 1",
  "DI-08 now 0",       "DI-08 now 1",
  "DI-09 now 0",       "DI-09 now 1",
  "DI-10 now 0",       "DI-10 now 1",
  "DI-11 now 0",       "DI-11 now 1",
  "DI-12 now 0",       "DI-12 now 1",
  "DI-13 now 0",       "DI-13 now 1",
  "DI-14 now 0",       "DI-14 now 1",
  "DI-15 now 0",       "DI-15 now 1"
};


//=============== digitalOut bit usage ===============
/*EXAMPLE:
#define  LIMITED_LED     0x00000001   // 00: 1=On, 0=Off
#define  ERROR_LED       0x00000002   // 01: 1=On, 0=Off
#define  LED_BITS        0x00000003   // 02: bits for all LEDs
*/

#define  ActivityLED     0x00000100L

const uchar  LogicOrderDOut[NUM_APPL_DOUT] =  {
   0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15 };

//APPLCODE - set to state we want outputs in during initialization
#define InitialDigOutState  ( 0 )


#define  TurnOnActivityLED()    ClearBits(Next.digitalOut, ActivityLED)
#define  TurnOffActivityLED()   SetBits(Next.digitalOut, ActivityLED)

//--- msgs to log automatically when a digital output changes state ---
//--- use NULL for bits that we don't want to log changes for ---
const char far *digOutDescr[NUM_APPL_DOUT*2] = {
//--0 state descr--    --1 state descr--
  "DO-00 now 0",       "DO-00 now 1",
  "DO-01 now 0",       "DO-01 now 1",
  "DO-02 now 0",       "DO-02 now 1",
  "DO-03 now 0",       "DO-03 now 1",
  "DO-04 now 0",       "DO-04 now 1",
  "DO-05 now 0",       "DO-05 now 1",
  "DO-06 now 0",       "DO-06 now 1",
  "DO-07 now 0",       "DO-07 now 1",
   NULL,                NULL,
  "DO-09 now 0",       "DO-09 now 1",
  "DO-10 now 0",       "DO-10 now 1",
  "DO-11 now 0",       "DO-11 now 1",
  "DO-12 now 0",       "DO-12 now 1",
  "DO-13 now 0",       "DO-13 now 1",
  "DO-14 now 0",       "DO-14 now 1",
  "DO-15 now 0",       "DO-15 now 1"
};




//=============== Analog inputs ===============
//--- indexes for anaIn[] and readVal[] ---
/*EXAMPLE:
#define  RF_CAV_VOLT      0  // RF cavity voltage
#define  LIMITER_VOLT     1  // Setpoint Limiter voltage
#define  HDW_CFG          2  // Fixed volt indicating HD1 component set
*/

const uchar  LogicOrderAIn[NUM_APPL_AIN] = { 0, 1, 2, 3, 4, 5, 6, 7 };


//=============== Anlog outputs ===============
//--- indexes for anaOut[] and setVal[] ---
/*EXAMPLE:
#define  SETPOINT         0  // Setpoint for Cav Voltage Regulator circuit
#define  FP_METER         1  // FrontPanel Meter if INITIAL or SET pressed
*/

const uchar  LogicOrderAOut[NUM_APPL_AOUT] = { 0, 1, 2, 3 };


//=============================================================================
//  Application-specific data structures and definitions
//=============================================================================

// Each value in the ApplState structure should be throught of as representing
// a "signal" in a clocked logic circuit, with the "applic" task representing
// the logic implemented by the circuit elements such as gates, latches, etc.

// The main "applic" task is the highest priority task and executes the code
// that implements the application logic on a real-time, periodic basis.

//                 ====== IMPORTANT NOTE on LOCKING ======
// Remember to use a semaphore or some other form of locking mechanism to
// protect all access to shared values by more than one task.  The only
// exception to this is when the applic task needs to access a value protected
// by other tasks using only a scheduler/kernel lock (the KLOCK and KUNLOCK
// macros).  Because the applic task is the only task that runs at the highest
// priority, no other task can preempt it, and if the scheduler is locked,
// the applic task cannot preempt another task.  But scheduler locks must be
// kept short to avoid creating large latency in the timely execution of other
// tasks (the applic task in particular).


// ***** WARNING ***** WARNING ***** WARNING ***** WARNING *****
//  Changing the applState structure means that a soft reset will fail!  A
//  change to the size will cause a soft restart to abort and do a CPU
//  restart instead.  And while a change that does NOT result in a new
//  size will not cause the soft restart to abort, it does mean that any
//  shift in the position of values can cause all kinds of mayhem.
// ***** WARNING ***** WARNING ***** WARNING ***** WARNING *****


typedef struct {
 //-----------------------------------------------------------------------
 // SOFTWARE SETTING INPUTS
 //
 //  These values represent Software Inputs to the application logic.
 //
 //  These values are set in the "Next" instance by tasks OTHER than the
 //  "applic" task to change some value used by the application logic.
 //  Remember that all tasks besides the main "applic" task must use a
 //  scheduler lock to protect reads and writes to these values.
 //
 //  Note that the application task may alter these values (e.g. to change
 //  an invalid value in to a valid one).
 //-----------------------------------------------------------------------
//  bool   onRequest;
//  bool   offRequest;
//  bool   resetRequest;

//  float  ampSet;      // most recent setpoint


 //-----------------------------------------------------------------------
 // READINGS
 //
 //  These values represent "passive" inputs to the application logic.
 //
 //  These values are set in the "Next" instance of ApplState by the applic
 //  task just after reading the state of the hardware inputs.
 //
 //  These values generally come from one of two sources:
 //    - The just-read state of analog and/or digital hardware inputs
 //    - Values maintained by other tasks
 //
 //  While some of these values are simply copies of "external" values,
 //  including them in the ApplState structure allows us to see how
 //  they change between cycles.  Also, some of these values are the result
 //  of applying logic to "external" values.  Applying that logic once at
 //  at the beginning of the applic loop and saving the result eliminates
 //  the need to repeat the same logic in multiple blocks of code later on,
 //  avoiding unnecessary redundancy.
 //-----------------------------------------------------------------------
//--- values updated in "cur" instance at start of each pass ---
  ulong  processTime;     // time (in ms) when processing was started
//  bool   rfIsOn;
  ulong  digitalIn;       // state of digital inputs


 //-----------------------------------------------------------------------
 // OUTPUTS from UN-CLOCKED (asynchronous, gate-level) LOGIC ELEMENTS
 //
 //  These values are outputs from one of the blocks of code in the applic
 //  task that represent a non-clocked logic element (such as a gate).
 //
 //  Each value here is set by one (and ONLY one) of the blocks of code in
 //  the applic task.  This group contains ONLY values that are set by a
 //  block of code that represents a NON-clocked logic element.
 //
 //  These values are set in the "Next" instance of ApplState by the applic
 //  task just after reading the state of the hardware inputs.  These values
 //  must  NOT be altered by other tasks (although they can read them as long
 //  as they use appropriate locking to do so).
 //
 //  Because these values represent the outputs of un-clocked logic elements,
 //  and such elements often need to be cascaded (the outputs used as inputs
 //  to other non-clocked elements), the result is often dependent on the
 //  order of execution of the blocks of code.
 //-----------------------------------------------------------------------
//  bool   ovenModeOn;           // Turn on power, enable output soon after
//  bool   ovenOutErrInterlock;  // Oven PS Output state was wrong
//  ulong  ovenOutErrStartTime;  // start of period of Output state error


 //-----------------------------------------------------------------------
 // OUTPUTS from CLOCKED (latched inputs and outputs) LOGIC ELEMENTS
 //
 //  These values are outputs from one of the blocks of code in the applic
 //  task that represent a clocked logic element (i.e. it acts like a
 //  clocked latch but is in software and the clock is one pass through
 //  the main application loop).
 //
 //  Each value here is set by one (and ONLY one) of the blocks of code in
 //  the applic task.  This group contains only values that are set by a
 //  block of code that represents a CLOCKED logic element.
 //
 //  These values are set in the "Next" instance of ApplState by the applic
 //  task AFTER all the other ApplState values are set.  Because they represent
 //  only the clocked logic elements, and each block of code reads only values
 //  from the Cur instance, the blocks can be executed in any order and the
 //  result will always be the same.
 //
 //  These values are used as inputs by other blocks and/or to set values
 //  used by other tasks (such as values read by the ModBus server task).
 //
 //  Again, these values must NOT be altered by other tasks (although they
 //  can read them as long as they use appropriate locking to do so).
 //-----------------------------------------------------------------------
  ulong  digitalOut;      // state for digital outputs

//  float  activeSetpoint;  // value applied to regulation circuit

//  float  rfOnSetpoint;    // warm/final setpoint
//  float  rfOffSetpoint;   // cold/initial setpoint

} ApplState;


//bool  localMode, closedLoop;

ApplState  prevState, curState, nextState;

#define  Prev  prevState
#define  Cur   curState
#define  Next  nextState


//ulong  rfOnTime;   // time RF turned on (in milliseconds)

//float  regulationError;  // diff between setpoint and readback

const uint ApplStateDataSize =  ( sizeof(ApplState)*3 + sizeof(setVal) );




//=============================================================================
//  List of daemon processes.  The inetd process listens for connection
//  requests on each port in the list and creates a new instance of the
//  specified process to service each new connection.
//=============================================================================
#include <services.h>

//APPLCODE - add tasks that service network clients

//--- service funcs declared the same as standard command funcs:
//     e.g.:   void _task fastdatad(int argc, char **argv);

void _task tsh(int argc, char **argv);


ServiceInfo inetdConf[] = {
//--- add any application-specific inetd-type network services here ---
//--- (add port # names to services.h) ---
//  NAME          PORT           PROTOCOL  FUNC       PRI  -1  Config Params
//  ------------  -------------  --------  ---------- ---  --  ----------------

//--- standard ones defined in sharedAppl.c ---
//{ "console",    TTY0,          ASYNC,    tsh,        10, -1, &consCfg },
//{ "fastdatad",  FASTDATA_PORT, ProtoTCP, fastdatad,   5, -1, NULL },
//{ "modbusTCPd", MODBUS_PORT,   ProtoTCP, modbusTCPd,  6, -1, NULL },
//{ "tsh",        TELNET_PORT,   ProtoTCP, tsh,        10, -1, NULL }

  #define SharedServices
  #include <sharedAppl.c>
};

const int  NumServices = sizeof(inetdConf) / sizeof(ServiceInfo);



//=============================================================================
//  Defintions for variables that are accessable as ModBus registers
//=============================================================================
#else  // end of before-RegDefs portion

/*
  The purpose of this section is to define a set of C variables that are
  accessable as ModBus registers.  This approach avoids the need to use
  special functions or macros throughout the code when accessing register
  values, which keeps the application code fast and simple.

  The model used provides support for multiple logical devices.  The 2nd
  parameter in each register definition macro indicates if one instanace
  applies to all logical devices (specified by an "x" for "don't care"), or
  if each logical device requires its own instance (specified by an under-
  score "_" to signify "fill in the blank").  In the 1st case, the variable
  is defined as a scalar.  In the 2nd case, it is defined as an array.

  The definition macros support data types (floats and longs in particular)
  that require more than one register to store them.  To avoid the possibility
  of corrupted values, all reads of and writes to such values must be done
  using the multi-register commands and including all the registers that
  make up the value in the command.  Differences in register order between
  different client types is handled by a table that automatically swaps
  register values as needed based on the client type.

  The definition for each writable register also includes the function to
  call when the value is set via the ModBus interface.  Many registers can
  use the default setXXX() functions, but you can specify your own to provide
  any special logic needed (including ignoring the new setting if invalid).

  For registers, the format for the function definitions is:

    bool mySetFunc(SetRegFuncParams);

       where SetRegFuncParams is currently defined as:
              void *info, uchar chanNum, void *newVal, fd_t eventsfd

  For coils, the format for the function definitions is:

     bool (*SetCoilFunc)(SetCoilFuncParams);

         where SetCoilFuncParams is currently define as:
               void *info, uchar chanNum, uchar newVal, fd_t eventsfd

  The parameters to the call are as follows:
    *info:     A pointer to the Coil's/Register's CoilInfo/RegInfo structure
    chanNum:   The logical device # (0 is the 1st)
    *newVal:   A pointer to the new value requested in the Modbus packet
               (changed to native-byte order where needed)
    eventsfd:  The fd for the FIFO used to send msgs to the main application
               process

  The function returns true if the new value was accepted, false if not.

  The register numbers for multiple "channels" or logical devices are
  handled automatically based on the xxChanMult values.  These values specify
  the offset between a register value for one logical device and the "same"
  register for the Next.

  Example:

    Assume the following definitions:
      #define  NumLogicalDevices  2
      #define  R4ChanMult  100
      REG4( motorPos,       _, 40050, F32, RW, setMotorPos )

    Then we would get:
      global variable definition:  float motorPos[2];
      register pair:  40050,40051  refers to:  motorPos[0]
      register pair:  40150.40151  refers to:  motorPos[1]

    If instead we had specified:
      REG4( motorPos,       x, 40050, F32, RW, setMotorPos )

    Then we would get:
      global variable definition:  float motorPos;
      register pairs:  40050,40051 & 40150,40151  BOTH refer to:  motorPos


  To keep the definitions as simple as possible, and avoid requiring multiple
  macros for each register definition, each macro is used more than once,
  with a different definition (#define) each time (a practice generally
  referred to as X-Macros).  While trying to follow the order of inclusion
  of files (including a recursive reference back to this file) can be rather
  confusing, the X-Macro method keeps the source code simple to read and
  maintain by doing everything with a single macro for each register.

  ==========================================================================
                             !!! WARNING !!!
  ==========================================================================
    Each sequence of register definitions MUST be contiguous!  There must
    NOT be any holes (i.e. no unused register #'s between used ones) and no
    overlap!
-----------------------------------------------------------------------------*/


//APPLCODE - set # of devices for which we want duplicate register sets
#ifndef NumLogicalDevices
  #define  NumLogicalDevices  1
#endif

//--------------------------------------------------------------
//APPLCODE - set max # of registers per set of coils
#ifndef NumC0Chans
  #define NumC0Chans_  NumLogicalDevices
  #define C0ChanMult_  100
#endif

#ifdef COIL0

COIL0( DoReboot,       x, 1, setCoil )
//COIL0( Coil00002,      _, 2, setCoil )
//COIL0( Coil00003,      _, 3, setCoil )

#undef COIL0
#endif


//--------------------------------------------------------------
//APPLCODE - set max # of registers per set of read-only registers
#ifndef NumR3Chans
  #define NumR3Chans_  NumLogicalDevices
  #define R3ChanMult_  100
#endif

#ifdef REG3

REG3( dlmVersTime,    x, 30001, U32 )
REG3( rtosVersTime,   x, 30003, U32 )
REG3( applVersTime,   x, 30005, U32 )
REG3( devType_,       x, 30007, U16 )
REG3( deviceID_,      x, 30008, U32 )
REG3( upSecs,         x, 30010, U32 )
REG3( percentIdle_,   x, 30012, U16 )

REG3( numChans,       x, 30013, U16 )
REG3( coilChanMult,   x, 30014, U16 )
REG3( reg3ChanMult,   x, 30015, U16 )
REG3( reg4ChanMult,   x, 30016, U16 )

REG3( digInBits,      x, 30017, U32 )

REG3( statusBitsStdA, x, 30019, U16 )
REG3( statusBitsStdB, x, 30020, U16 )
REG3( statusBitsAppl, x, 30021, U32 )

REG3( latency,        x, 30023, U16 )  // actual vs scheduled appl start in ms
REG3( tcpTickTime,    x, 30024, U16 )  // # ms inside tcp_tick() call
REG3( applLoopTime,   x, 30025, U16 )  // # ms used by applic task loop
REG3( kernelLoopTime, x, 30026, U16 )  // # ms used by last kernel pass

//APPLCODE - add registers for externally available read-only values
REG3( rdg_1,          x, 30027, F32 )  /* scaled reading for chan 1 */
REG3( rdg_2,          x, 30029, F32 )  /* scaled reading for chan 2 */
REG3( rdg_3,          x, 30031, F32 )  /* scaled reading for chan 3 */
REG3( rdg_4,          x, 30033, F32 )  /* scaled reading for chan 4 */
REG3( rdg_5,          x, 30035, F32 )  /* scaled reading for chan 5 */
REG3( rdg_6,          x, 30037, F32 )  /* scaled reading for chan 6 */
REG3( rdg_7,          x, 30039, F32 )  /* scaled reading for chan 7 */
REG3( rdg_8,          x, 30041, F32 )  /* scaled reading for chan 8 */

#undef REG3
#endif


#ifdef RegMacros
//===== application-specific defintions related to register values =====

  #define  READING(idx)  REG_ARRAY(float, rdg_1, idx)

#endif


//--------------------------------------------------------------
//APPLCODE - set max # of registers per set of read/write registers
#ifndef NumR4Chans
  #define NumR4Chans_  NumLogicalDevices
  #define R4ChanMult_  100
#endif

/*
  The "standard" functions for processing writes to Modbus register values:

    setS16, setU16, setS32, setU32, setF32

  These functions simply log a message for the change and update the
  value in a critical block (so the change cannot be interrupted).

  In many cases, just allowing any setting to be written to a register (or
  multiple registers comprising a larger value) is not acceptable.


  If any of the following apply, then a custom func must be used instead:

    - The order of changes is important.  For example, if you get ON,
      OFF, and RESET commands in rapid succession, you have to be sure
      they are processed in the order they were recieved.

    - Not all possible settings are acceptable.  The acceptable settings
      for a value are often a subset of the possible ones, and frequently
      are dependent upon the current state of the controller.


  IMPORTANT NOTES:

    The "applic" task periodically executes the code that makes up the
    bulk of the application logic.  Because it runs at the highest priority
    there is no chance that a lower priority task (like the daemon handling
    Modbus register writes) can change a value in the middle of this
    processing.

    However, the applic task CAN (and often does) preempt other processes,
    so any process that does read or make changes to the register values
    must lock the register values before doing so.  This is currently done
    by disabling the scheduler using the KLOCK() and KUNLOCK() macros.

    Use of such a lock insures that changes occur only between passes
    through the application logic, but there are still several issues
    that must be addressed when writing functions to handle writes to
    register values:

      - Priority inversion and latency:  Allowing a lower priority task
        to lock out higher ones (or, in this case, ALL tasks) allows
        the lower priority task to run when a higher-priority one should
        be running.  Care must be taken to keep such periods as short
        as possible to avoid disrupting the timing of the control logic.

      - Loss of encapsulation:  The last thing we want is to break the
        application logic in to small pieces spread through a bunch of
        functions executed by different tasks at different times.

        While the logic for processing writes to register values is often
        dependent upon the state of the control logic, functions outside
        the applic task must NOT directly alter the state of the application
        logic.  To achieve this, such functions must limit changes to
        values that act as input to the applic task.

        To this end, you need to treat the register values that act as
        inputs just like they were knobs, switches, analog inputs, or
        latches.  Like their hardware equivalent, the applic task logic
        reads their state at the beginning of each pass and uses this
        information when determining what the next state of the control
        logic should be.  And just like a hardware latch, values can be
        reset after reading them so the next write to the value can be
        treated as another instance of a request.

    The trickiest circumstances to handle are ones where the order of separate
    writes is important.  The most common example are commands for ON, OFF,
    and RESET.  If, for example, you had 3 separate flags, and simply set
    each once corresponding to which of the 3 commands you received, you
    end up with multiple flags set when the applic logic is processed, and
    no way to determine what the end result should be.  In these cases,
    the logic that controls the state of each flag should be placed in its
    own function and called during the processing of any event that could
    affect the flag's state.

    Ideally, the logic for setting each variable would be triggered each
    time an event occurs that could affect it:  A publish-subscribe type
    of mechanism.  One possibility to do this efficiently?  Use a bit-mask
    "array" to determine which functions to call for each register that is
    changed.  That way it doesn't take many bytes to determine which functions
    to call, and each function can update its corresponding register value
    based on its current state and the event that just occured.

    Or should I simply combine such values in to one (a struct, if necessary),
    and have a single function that is called for all the related values
    and updates the state of the single logical entity based on its
    current state and what event called it?  (But the # of combinations could
    get out of hand easily...)
*/

#ifdef REG4

REG4( timeZone_,      x, 40001, S16, RW, MB_timeZone )
REG4( curTime,        x, 40002, U32, RW, setU32 )

REG4( ipAddr,         x, 40004, U32, RW, setU32 )
REG4( netmask,        x, 40006, U32, RW, setU32 )
REG4( gateway,        x, 40008, U32, RW, setU32 )

REG4( setACoil,       _, 40010, U16, RW, setU16 )
REG4( resetACoil,     _, 40011, U16, RW, setU16 )

REG4( digOutBits,     x, 40012, U32, RW, setU32 )

REG4( cmdBitsStd,     x, 40014, U32, RW, MB_cmdBitsStd )
REG4( cmdBitsAppl,    x, 40016, U32, RW, MB_cmdBitsAppl )

//APPLCODE - add registers for externally available read/write values
//REG4( ampSet,           x, 40018, F32, RW, MB_ampSet )  /* desired output */

#undef REG4
#endif


#ifdef RegMacros
//===== application-specific defintions related to register values =====

//  #define  ISET(idx)   REG_ARRAY(float, iSet_1, idx)

#endif

//--------------------------------------------------------------
#undef RegMacros

#endif
#ifndef RegDefs // start of after-RegDefs portion




//=============================================================================
//  defintions for "parameterized" data values - lists of values that we want
//  to be "persistent" (i.e. saved to a file and restored on startup)
//=============================================================================
//APPLCODE - settings/values we want saved to/restored from files


//--- parameter ID's for data in Device Data/calibration file ---
//--------------------------------------------------------------------------
// WARNING:  ID's can be added, but existing ID's (and the format of the
//           data they represent) should NEVER change!  NOTE:  Existing ones
//           can be "eliminated" by changing their name and no longer
//           using them, but never REUSE an old one for something new!!!
//--------------------------------------------------------------------------
//  NOTE:  ID's >= 1000 are rsvd for values shared by most applications
//--------------------------------------------------------------------------

//RESTORE_FUNC(myVar);

const ParamData far  devDataParams[] = {
//APPLDATA
//                ID   ---value---  --RESTORE_FUNC--
//PERSISTENT_VAR(  1,  myVarName,     PVRF(myVar) ),
//                ID   ---value---  granularity  --RESTORE_FUNC--
//PERSISTENT_TBL(  2   myArrayName, elemSize,     NULL ),

  #define SharedDevParams
  #include <sharedAppl.c>

  END_PARAM_DATA
};


//----------------------------------------------------

//RESTORE_FUNC(ampSet);

const ParamData far  settingsParams[] = {
//APPLDATA
//_nearcall void ampSetRestore(void far *curVal, char *newVal, uint count);
//
//                ID   ---value---  --RESTORE_FUNC--
//  PERSISTENT_VAR(  1,  ampSet,     PVRF(ampSet) ),
//  PERSISTENT_VAR(  2,  rfOnSetpoint,    NULL ),

  END_PARAM_DATA
};


//-----------------------------------------------------------------
//  Application-specific values needed to restore the state of the
//  controller following a "soft" reset (i.e. a restart without a
//  CPU reset)
//-----------------------------------------------------------------
const ParamData far  ctlrStateParams[] = {
// sample values from sharedAppl.c, common to all applications
//                   --ID--  ---value,granularity---  --RESTORE_FUNC--
//    PERSISTENT_VARx(  1000,  prevState,               NULL ),
//    PERSISTENT_VARx(  1001,  curState,                NULL ),
//    PERSISTENT_VARx(  1002,  nextState,               NULL ),
//    PERSISTENT_TBL (  1003,  setVal,4,                NULL ),
// // PERSISTENT_TBL (  1004,  shadowRegValues,4,       NULL ),

//APPLDATA
//    PERSISTENT_VAR(  1,  myApplStateVar,        NULL ),


  #define SharedCtlrStateParams
  #include <sharedAppl.c>

  END_PARAM_DATA
};


//-----------------------------------------------------------------------------
//  A list of parameterized "persistent" data.  Backup copies of such values
//  are stored in files and restored on startup.
//-----------------------------------------------------------------------------

//    Base File Name    Category  ParamData
//    ----------------  --------  ---------
//--- add any application-specific parameterized file defs here ---
//ParamFile  myAppCfgFile =
//  { "myFileName",     PDC_CTLR, myCfgParams,   0 };

ParamFile  settingsFile =
    { "settings",       PDC_SET,  settingsParams  };


#define SharedParams
#include <sharedAppl.c>


ParamFile *paramFiles[] = {
  #define SharedParamsList
  #include <sharedAppl.c>
//--- add any application-specific ParamData references here ---
//  &myAppCfgFile,
  &settingsFile
};

const int NumParamFiles = sizeof(paramFiles) / sizeof(paramFiles[0]);


//=============================================================================
//  Declarations for functions that handle setting "Next" values that serve
//  as inputs to the application control logic (run by the "applic" task in
//  a periodically executed loop).
//
//  While care must be taken in these functions to use locks to avoid the
//  possibility of being preempted in the middle of an operation, such locks
//  must be kept to as short a time as possible to avoid excessively latency
//  in the main control loop.
//=============================================================================

//void onRequest(void);
//void offRequest(void);
//void resetRequest(void);

//void newSetpoint(float newAmps);


//=============================================================================
//  This is where misc global variables should be defined
//=============================================================================
#define SharedGlobals
#include <sharedAppl.c>

//APPLCODE
//--- add definitions for application-specific global variables here ---



//=============================================================================
//  Functions referenced in the Modbus register definitions to handle writes to
//  specific global values defined as ModBus registers
//=============================================================================

//APPLCODE - add functions to process writes to ModBus register values

//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for "cmdBitsAppl"
//
//  Check the bits that where set and trigger the corresponding events as
//  appropriate (but do NOT save the new value - just use it to trigger
//  events)
//-----------------------------------------------------------------------------
bool _nearcall MB_cmdBitsAppl(SetRegFuncParams)
{
/*
  ulong  cmdBits;
  EventData  eventMsg;  // used by EVENT macro


  cmdBits = *((ulong *)newVal);

  if (cmdBits & x)  EVENT(MyEventX, chan, newVal, sizeof(int)); // w/ data
  if (cmdBits & y)  EVENT(MyEVentY, chan, NULL, 0);             // w/o data
*/

  return true;
}

//-----------------------------------------------------------------------------
//  External ModBus client sent a new setpoint
//
//  Check the new setting and alter output values as appropriate
//-----------------------------------------------------------------------------
/*EXAMPLE
bool MB_ampSet(SetRegFuncParams)
{
  newSetpoint(MB_VAL(float));

  return true;
}
*/


//=============================================================================
//  functions to handle restoration of saved "persistent" values
//=============================================================================

/*
//-----------------------------------------------------------------------------
//  one of the setpoint values needs restoring from the settings file
//-----------------------------------------------------------------------------
RESTORE_FUNC(myVal)
{
  uint  u;
  float  fval;
  char  temp[32];

  for (u=0; u<NumPS; ++u)  if (curVal == (void far *) &ISET(u))  {
    fval = *(float far *)newVal;
    printf("restoring iSet_%u to: %s\r\n", u, fmtFloat(fval, 8,2, temp));
    newSetpoint(u, fval);
  }

}

//-----------------------------------------------------------------------------
//  restore an ampSet value from the settings file
//-----------------------------------------------------------------------------
RESTORE_FUNC(ampSet)
{
  newSetpoint(NEW_VAL(float));
}


*/

//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
#define SharedFuncs
#include <sharedAppl.c>


//=============================================================================
//  Misc Application-specific functions
//=============================================================================
//APPLCODE



//=============================================================================
//  Task functions to handle application-specific commands
//=============================================================================
//APPLCODE

//-----------------------------------------------------------------------------
//  example func to process a command entered via a user shell
//-----------------------------------------------------------------------------
/*
void _task miscTest(int argc, char **argv)
{
  int  idx;
  bool  validCmd;

//used by EVENT macro
  fd_t  eventsfd = -1;
  EventMsg  eventMsg;


  validCmd = false;

  for (;;)  {
    if (argc < 3)  break;
    if (! isdigit(argv[1][0]))  break;

    idx = atoi(argv[1]);
    if ((idx < 1) OR (idx > NumChan))  {
      printf("Invalid chan #\n");  Exit(-2); }

    --idx;  validCmd = true;

    if (StriEq(argv[2], "x"))     { EVENT(MyEventX, idx, NULL, 0);  break; }
    if (StriEq(argv[2], "y"))     { EVENT(MyEventY, idx, NULL, 0);  break; }

    validCmd = false;  break;
  }

  if (! validCmd)  { printf("Usage:  misc # x|y\r\n");  Exit(-1); }

  Exit(0);
}
*/

//=============================================================================
//  Generic structure to handle application-specific presentation of the
//  digital and analog input/output values.  Code to handle additional
//  formatting that cannot be handled in this generic manner should be
//  placed in the showState() task routine.
//=============================================================================
//APPLCODE - text and formatting info for application state summary

/*EXAMPLE:
//-----------------------------------------------------------------------------
_nearcall char *fmtMode(DispElem far *elem, char *buf)
{
  return (char *) ModeDesc[ELEM_VAL(U16)];
}

//-----------------------------------------------------------------------------
_nearcall char *fmtHdwCfg(DispElem far *elem, char *buf)
{
  return (char *) HdwCfgDesc[ELEM_VAL(U16)];
}
*/

//-----------------------------------------------------------------------------
  const char far DevStateStr[] = {
"---- Nothing defined yet ---\r\n"
/*EXAMPLE:
"Cavity Voltage:  ###.## kV    Open/Closed: ######    HdwConfig: ##### \r\n"
"Active setpoint: ###.## kV    RF On/Off:   ### \r\n"
"Error:          -###.## kV    Loc/Rem:     ###### \r\n"
" \r\n"
"Limiter Voltage: -##.### V    Limited:     ### \r\n"
"                              Regulation:  ##### \r\n"
"RF OFF setpoint: ###.## kV \r\n"
"RF ON setpoint:  ###.## kV    Ramp Mode:   #### \r\n"
" \r\n"
*/
};

const DispElem  far dispElems[] = {
/*EXAMPLE:
  { 17, 3, VFPV(cavityVolts),     DARK_BLUE, fmtF32,    6,2 },
  { 17, 4, VFPV(activeSetpoint),  DARK_BLUE, fmtF32,    6,2 },
  { 16, 5, VFPV(regulationError), DARK_BLUE, fmtF32,    7,2 },

  { 17, 7, VFPV(limiterVolts),    DARK_BLUE, fmtF32,    7,3 },

  { 17, 9, VFPV(rfOffSetpoint),   DARK_BLUE, fmtF32,    6,2 },
  { 17,10, VFPV(rfOnSetpoint),    DARK_BLUE, fmtF32,    6,2 },

  { 43,10, VFPV(rampMode),        DARK_BLUE, fmtMode,   4,4 },

  { 64, 3, VFPV(hdwCfg),          DARK_BLUE, fmtHdwCfg, 5,5 },
*/
  { 0,0, NULL, NULL, NULL, 0,0 }
};

//--- ATTR_NAME macros currently defined in fcntl.h ---
const DigDispElem far digInElems[] = {
/*EXAMPLE:
//{ xx,yy,  BIT_MASK,     "0STR"  ,ATTR_NAME,  "1STR"  ,ATTR_NAME },
  { 43, 3,  CLOSED_LOOP,  "OPEN  ",DARK_RED,   "CLOSED",DARK_BLUE },
  { 43, 4,  RF_IS_OFF,    "ON "   ,DARK_BLUE,  "OFF"   ,DARK_RED },
  { 43, 5,  LOC_REM_SW,   "REMOTE",DARK_BLUE,  "LOCAL ",DARK_BLUE },
*/
  { 0,0, 0, NULL,NULL, NULL,NULL }
};

const DigDispElem far digOutElems[] = {
/*EXAMPLE:
//{ xx,yy,  BIT_MASK,     "0STR"  ,ATTR_NAME,  "1STR" ,ATTR_NAME },
  { 43, 7,  LIMITED_LED,   "NO "  ,DARK_BLUE,  "YES"  ,DARK_RED },
  { 43, 8,  ERROR_LED,     "OK   ",DARK_BLUE,  "ERROR",DARK_RED },
*/
  { 0,0,  0,  NULL,NULL,  NULL,NULL }
};

//---------------------------------------------------------------
//---- older style - do not use for newer applications -----
const DigFmtInfo digInFmt[] = {
//{ xx,yy, "0-str"  ,1,  "1-str" ,0 },  //  descr
  { -1,-1, "",0, "",0 }
};

const DigFmtInfo digOutFmt[] = {
//{ xx,yy, "0-str"  ,1,  "1-str" ,0 },  //  descr
  { -1,-1, "",0, "",0 }
};


const AnalogFmtInfo  adcFmt[] = {
  { -1,-1, 0,0 }
};

const AnalogFmtInfo  dacFmt[] = {
  { -1,-1, 0,0 }
};


//-----------------------------------------------------------------------------
//  "state":  Show continuously updated state of the system
//-----------------------------------------------------------------------------
void _task showState(int argc, char **argv)
{
  bool  blinkOff;
  char  temp[128];


  printf(clsCmd);  blinkOff = false;

  CursorTo(0, 3);  printf(DevStateStr);

  for (;;)  {
    CursorTo(0, 0);  blinkOff = !blinkOff;

    stdHeader();  displayElems((DispElem far *)dispElems, blinkOff, temp);

    displayDigElems(digitalIn, (DigDispElem far *)digInElems, blinkOff, temp);
    displayDigElems(digitalOut, (DigDispElem far *)digOutElems, blinkOff, temp);

    CursorTo(0,22);  fflush(stdout);  CheckETX();  sleepMS(500);
  }

}

//-----------------------------------------------------------------------------
//  Application-specific commands
//-----------------------------------------------------------------------------
void _task cmd(int argc, char **argv)
{
/*
  bool  validCmd, allChans, validChan;
  int  chan;
  char  *cmdStr;


  validCmd = allChans = validChan = false;

  do  {
    if (argc < 2)  break;

    cmdStr = argv[1];

    if (argc > 2)  {
      if (StriEq(argv[2], "all"))
        allChans = validChan = true;
      else  {
        if (isdigit(*argv[2]))  chan = atoi(argv[2]) - 1;
        if ((chan >= 0) AND (chan < 4))  validChan = true;
      }
    }

    if (StriEq(cmdStr, "enable"))  {
      if (!validChan)  { printf("Invalid channel\r\n");  break; }
      KLOCK();
      if (allChans)
        cmdBitsAppl |= cmdEnableAllBits;
      else
        cmdBitsAppl |= cmdEnableBits[chan];
      KUNLOCK();  validCmd = true;
    }

    if (StriEq(cmdStr, "disable"))  {
      if (!validChan)  { printf("Invalid channel\r\n");  break; }
      KLOCK();
      if (allChans)
        cmdBitsAppl |= cmdDisableAllBits;
      else
        cmdBitsAppl |= cmdDisableBits[chan];
      KUNLOCK();  validCmd = true;
    }

    if (StriEq(cmdStr, "clear"))  {
      if (!validChan)  { printf("Invalid channel\r\n");  break; }
      KLOCK();
      if (allChans)
        cmdBitsAppl |= cmdClearAllBits;
      else
        cmdBitsAppl |= cmdClearBits[chan];
      KUNLOCK();  validCmd = true;
    }

//    if (StriEq(cmdStr, "reset"))  {
//      KLOCK();  cmdBitsAppl |= CBA_RESET_FAULT;  KUNLOCK();  validCmd = true; }

  }  while(0);


  if (! validCmd)  {
    printf("\r\n"
           "Usage:  cmd enable #|all   (enable one or all channels)\r\n"
           "            disable #|all  (disable one or all channels)\r\n"
           "            clear #|all    (clear count for one or all channels)\r\n"
           "            reset          (reset current spark fault)\r\n" );
    Exit(-1);
  }
*/

  Exit(0);
}

static int  lastpid = 0;

//-----------------------------------------------------------------------------
void semErr(sem_t *sem, const char far *funcName)
{
  ulong  msTime = GetMS();

  printf("%s  %5u.%03u [%05d]: sem_%ls(): %ls\r\n",
         (getpid() == lastpid) ? "" : "\r\n",
         (uint)(msTime/1000), (uint)(msTime % 1000), getpid(),
         funcName, strerror(errno) );

  lastpid = getpid();
}

//-----------------------------------------------------------------------------
void semMsg(sem_t *sem, const char far *msg)
{
  ulong  msTime = GetMS();

  printf("%s  %5u.%03u [%05d]: %ls\r\n",
         (getpid() == lastpid) ? "" : "\r\n",
         (uint)(msTime/1000), (uint)(msTime % 1000), getpid(), msg);

  lastpid = getpid();
}

//-----------------------------------------------------------------------------
//  Task to test use of POSIX-style semaphores
//-----------------------------------------------------------------------------
void _task sem_user(int argc, char **argv)
{
  sem_t  *sem = NULL;
  int  semValue, stat;
  uint  u;
  long  waitSecs, delay;
  char  *name;
  char  func;
  struct timespec endWaitTOD;
  char  msg[128];


#define  SemName  "test-sem"


  delay = (getpid() % 4) * 500 + 1000;

  waitSecs = atol(argv[1]);

  for (u=0; (func=argv[2][u])!=0; ++u)  {
    switch (func)  {

      case 'o':
        sem = sem_open(SemName, 0, 0, 0);
        if (!sem)  { semErr(sem, "open");  semMsg(sem, "exiting");  Exit(-1);  }
        semMsg(sem, "opened");
        sem_getvalue(sem, &semValue);
        sprintf(msg, "cur value: %d", semValue);  semMsg(sem, msg);
        break;

      case 'w':
        if (waitSecs < 0)  {
          name = "wait";  stat = sem_wait(sem);  }
        else if (waitSecs == 0)  {
          name = "trywait";  stat = sem_trywait(sem);  }
        else  {
          endWaitTOD.tv_sec = time(NULL) + waitSecs;  endWaitTOD.tv_nsec = 0;
          name = "timedwait";  stat = sem_timedwait(sem, &endWaitTOD);
        }
        if (stat)  semErr(sem, name);  else  semMsg(sem, "got lock");
        break;

      case 'p':
        if (sem_post(sem))
          semErr(sem, "post");
        else
          semMsg(sem, "released lock");
        break;

      case 'c':
        if (sem_close(sem))
          perror("close");
        else
          semMsg(sem, "closed");
        break;

      case 'u':
        if (sem_unlink(SemName))
          semErr(sem, "unlink");
        else
          semMsg(sem, "unlinked");
        break;

      default:
        sprintf(msg, "unrecognized func specifier: '%c'", func);
        semMsg(sem, msg);
    }
    sleepMS(delay);
  }

  semMsg(sem, "exiting");  Exit(0);
}


//-----------------------------------------------------------------------------
//  Test the POSIX-style semaphore functions
//-----------------------------------------------------------------------------
void _task sem_test(int argc, char **argv)
{
  #define  MaxTasks  8
  uint  u, numTasks;
  int  stat;
  sem_t  *sem;
  int  pids[8];
  char  *argv_[4];


  if (showHelp(3, argc, argv))  {
    printf("\n"
           "usage: sem_test <waitSec> <testMask> [...testMask]\r\n"
           "\r\n"
           "          waitSec:  > 0:  use sem_timedwait() (block w/timeout)\r\n"
           "                      0:  use sem_trywait()   (don't block)\r\n"
           "                    < 0:  use sem_wait()      (block)\r\n"
           "          testMask str indicating funcs to call & order:\r\n"
           "                    o:   sem_open()\r\n"
           "                    w:   sem_???wait()\r\n"
           "                    p:   sem_post()\r\n"
           "                    c:   sem_close()\r\n"
           "                    u:   sem_unlink()\r\n"
           "          # of testMasks = # of concurrent tasks\r\n"
          );

    Exit(-1);
  }

  sem = sem_open("test-sem", O_CREAT, 0666, 1);

  if (!sem)  { perror("sem_open");  Exit(-1);  }

  semMsg(sem, "Created");

  argv_[0] = "sem_user";  argv_[1] = argv[1];  argv_[3] = NULL;

  numTasks = 0;
  for (u=2; u<argc; ++u)  {
    argv_[2] = argv[u];
    pids[numTasks++] = createProc(sem_user, 9, 3, argv_, 0, 1, 2);
    if (numTasks == MaxTasks)  break;
  }

  sem_close(sem);  sem_unlink("test-sem");

  for (u=0; u<numTasks; ++u)  waitpid(pids[u], &stat, 0);

  printf("\r\nall child tasks exited\r\n");

  Exit(0);
}


//=============================================================================
//  Event-Handler functions called by the appl task to process event messages
//=============================================================================
//APPLCODE

//-----------------------------------------------------------------------------
//  Event-Handler for a request to ...
//-----------------------------------------------------------------------------
/*
void EVENT_MyEvent(EventMsg *event)
{
  uint  idx;
  float  setpoint;
  char  temp[16];


  idx = event->chanNum;

  if (error)  {
    fprintf(DiagLog, "error msg, chan:\n", idx+1);
    return;
  }

  setpoint = *((float *)event->data);

  fprintf(DiagLog, "New SetPoint for PS%d: %s\n",
     idx+1, fmtFloat(setpoint, 8,3, temp) );

  *((float *)((&iSet_1_RegInfo + idx)->regVal)) = setpoint;

  anaOut[idx] = ISETtoVolts(idx, setpoint);
}
*/


//=============================================================================
//  Update the the state of the application logic in well-defined phases
//=============================================================================
// IMPORTANT NOTES and RULES:
//
//  - The code to change ANY value must be in ONE place.  NEVER spread the
//    logic for updating a single value around.  There should be ONE place
//    to look to see how a value gets set.
//
//  - "Cur" values are basically "inputs" - literally (they are set based
//    on the just-read state of a digital or analog input) and/or figuratively
//    (they hold a value set by some external source, such as a cmd from a
//    remote client or another user-interface task).
//
//  - "Next" values are basically "outputs" - literally (they determine the
//    next state of a digital or analog output) and/or figuratively (they
//    can affect the state of the logic for the next pass).
//
//  - The "Cur" OR the "Next" instance of a value may be changed - NEVER
//    both of them (with the exception of the first step below that copies
//    all of cur to prev and next to cur).
//
//  - A value MUST be in the ApplState struct if one of the following applies:
//     - The application logic is dependent on if or how much the value has
//       changed between passes.
//     - The value is used as both input (used in in a conditional and/or in
//       an expression that can affect the result of an assignment) and output
//       (the target of an assignment).
//
//  - Variables that are NOT in the ApplState struct but are USED (in
//    conditionals or other expressions) by the application logic must be
//    treated just like "Cur" values.  Variables SET (the target in an
//    assignment statement) by the application logic must be treated just
//    like "Next" values.
//
//  - When the logic requires comparing an "input" to a related "output", be
//    sure to use the "Cur" value for the "input", and the "Prev" value for
//    the "output".
//
//    This is because during the processing of the loop, the "Cur" values
//    represent the input values just read (what they are "now") and the
//    "Prev" values represent the values the outputs were set to at the
//    beginning of the prev pass (i.e. they are the values that "caused"
//    the inputs to have their "Cur" values).
//
//    During a pass the output values in "Cur" represent what the outputs
//    where set to AFTER the "Cur" inputs were updated, so comparing them
//    makes no sense.  Doing so is the equivalent of comparing a past response
//    to a condition that didn't exist at the time.
//=============================================================================


//=============================================================================
//  Definitions for functions that handle setting "Next" values that serve
//  as inputs to the application control logic (run by the "applic" task in
//  a periodically executed loop).
//
//  While care must be taken in these functions to use locks to avoid the
//  possibility of being preempted in the middle of an operation, such locks
//  must be kept to as short a time as possible to avoid excessive latency
//  in the main control loop.
//
//  For REALLY short locks, using a critical section is best, because this
//  has the least overhead (but it has to be REALLY short to avoid causing
//  problems for interrupt handlers).
//
//  For longer locks (ones that involve anything but the most trivial
//  conditional logic), use either KLOCK() and KUNLOCK() (which disable the
//  scheduler) or a more specific global semaphore/mutex lock.
//=============================================================================

/*Examples
//-----------------------------------------------------------------------------
void onRequest(void)
{
  _EnterCritical_();
  Next.onRequest = true;  Next.offRequest = false;
  _ExitCritical_();

  fprintf(DiagLog, "ON request\n");
}

//-----------------------------------------------------------------------------
void offRequest(void)
{
  _EnterCritical_();
  Next.offRequest = true;  Next.onRequest = false;
  _ExitCritical_();

  fprintf(DiagLog, "OFF request\n");
}

//-----------------------------------------------------------------------------
void resetRequest(void)
{
  _EnterCritical_();  Next.resetRequest = true;  _ExitCritical_();

  fprintf(DiagLog, "RESET request\n");
}

//-----------------------------------------------------------------------------
void newSetpoint(float newAmps)
{
  char  temp[16];


  KLOCK();

  if (FABS(newAmps) > maxAmpSet)  newAmps = (float)SIGN(newAmps) * maxAmpSet;

  Next.ampSet = newAmps;

  KUNLOCK();

  fprintf(DiagLog, "New Setpoint: %s\n", fmtFloat(newAmps, 10,3, temp) );
}

//-----------------------------------------------------------------------------
*/

/*
=========== NOTES on Locking ============

The applic task runs at the highest priority (even above the kernel), so it
cannot be preempted by any other task.  As long as this is true, the applic
task does not need to use locks when accessing global variables.

However, all other tasks MUST disable the scheduler when accessing values
used by the applic task to avoid being preempted while doing so.  Care must
be taken to keep these times short to avoid latency issues.

For REALLY short times, a critical section can be used as this is much more
efficient than disabling and reenabling the scheduler.  But critical sections
disable all interrupts, so it is not practical in most cases.

*/

//=============================================================================
//  Update the "Cur"rent state of the controller
//
//  Be sure to read the "NOTES on Locking" section above....
//=============================================================================
//
//  Just before this function is called, readings were taken of the current
//  state of the hardware, and any changes to the output states specified by
//  the last pass were applied (just after taking the readings).  These values
//  are copied to Next as needed and constitute the Hardware Inputs.
//
//  Other tasks can affect the state of the control logic by changing values
//  in Next as well.  These values constitute Software Inputs.
//
//  This function is responsible for the Asynchronous logic that is applied
//  each pass.  It updates values in Next (or other global variables) based
//  on the Software and Hardware Input values in Next.  This portion of the
//  logic is analogous to unclocked gates in an electronic circuit.
//
//  After the Asychronous logic is applied, we set Prev = Cur and Cur = Next.
//  The result is that Prev contains the controller state as it was during the
//  previous pass, and Cur indicates the current state of the controller.
//
//  Also, see the "NOTES on Locking" section above....
//=============================================================================
void updateCurState(void)
{
  if (current != applTask)  return;  // must only be called by "applic" task

//-----------------------------------------------------------------
// The following conditions exist at this point:
// - The analog and digital hardware inputs have just been read
// - The outputs were just set to values computed during the last pass
// - "Cur" contains the state the inputs were in at the START of
//   of the previous pass, and the state the outputs were in DURING
//   the previous pass
// - "Next" contains the values that the outputs were just set
//   to and any "input" values that were changed by other tasks
//------------------------------------------------------------------

// Now we update the values in "Next" that are based solely on the
// state of the hardware inputs we just read and values maintained by
// by other tasks (basically, we deal with READINGS here: Values that
// we need but are not written to "Next" by another task)

  Next.digitalIn = digitalIn;

//APPLCODE

/*EXAMPLES:
  Next.rfgenFault =  !(digitalIn & RF_PS_NOT_FAULT);
  Next.rfgenHeatOn = !(digitalIn & RF_PS_NOT_HEAT_ON);
  Next.rfgenReady =  !(digitalIn & RF_PS_NOT_READY);

  ovenRespAge = msSince(ovenLastRespTime);

  Next.ovenInactive = (ovenRespAge > 3000);
  Next.rfgenInactive = (msSince(rfgenLastRespTime) > 3000);

  Next.ovenReadV = readVal[OVEN_READ_V];
  Next.ovenReadA = readVal[OVEN_READ_A];

  Next.biasReadV = readVal[BIAS_READ_V];
  Next.biasReadA = readVal[BIAS_READ_A];

  Next.sptrReadV = readVal[SPTR_READ_V];
  Next.sptrReadA = readVal[SPTR_READ_A];

  Next.biasSetV = biasSetV;
  Next.sptrSetV = sptrSetV;
  Next.rfgenSetV = rfgenSetV;


//--- apply range limits to new settings ---
  checkSetValRange(&ovenSetV,  &Next.ovenSetV,  OVEN_SET_V);
  checkSetValRange(&biasSetV,  &Next.biasSetV,  BIAS_SET_V);
  checkSetValRange(&sptrSetV,  &Next.sptrSetV,  SPTR_SET_V);
  checkSetValRange(&rfgenSetV, &Next.rfgenSetV, RFGEN_SET_V);


#define CheckThresholds(name, prefix)  \
  checkThresholds(name, &prefix##ReadErrStartTime,  \
                  &Next.prefix##RuntimeErr, Cur.prefix##RuntimeErr,  \
                  prefix##MaxErrV, prefix##MaxErrMS,  \
                  Cur.prefix##SetVassert, Next.prefix##ReadV, Next.prefix##ReadA)

//--- check set/read error against allowed thresholds ---
//--- (these checks are very time consuming at present) ---
  CheckThresholds("Oven", oven);
  CheckThresholds("Bias", bias);
  CheckThresholds("Sputter", sptr);
*/


// Next we do the asynchronous/gate-level type logic.  This logic often has
// order of execution dependencies so be careful!

/*EXAMPLES:
//--- ovenPowerOnTime ---
  if (digitalOut & OVEN_PS_OFF)
    ovenPowerOnTime = 0;
  else
    if (!ovenPowerOnTime)  ovenPowerOnTime = GetMS();

//--- Next.ovenModeOn ---
  if (Next.ovenOutErrInterlock)  {
    Next.ovenModeOn = false;
    if (ovenOnCmd)  printf("Oven ON cmd ignored while Interlock exists\r\n");
  }
  else  {
    if (ovenOnCmd)  Next.ovenModeOn = true;
    if (ovenOffCmd)  Next.ovenModeOn = false;
  }

//--- ovenOffCmdTime ---
  if (ovenOnCmd)  ovenOffCmdTime = 0;
  if (ovenOffCmd)  if (!ovenOffCmdTime)  ovenOffCmdTime = GetMS();
*/



// Next we make a copy of the application state before we apply all the
// logic that updates the overall state.  This allows us to have a copy of
// all the relevant values that remain unchanged during the processing,
// avoiding all kinds of order-of-execution dependencies.

// We also make a copy of the state from last time before overwriting it
// so we can see if/how values have changed since the last pass

// Note that a copy is done (rather than just switching pointer values) to
// insure all values are set in both the source and destination.  If we
// simply kept pointers and swapped the pointer values, then EVERY value
// would have to be updated independently by the code, which is a pain to
// program, adds overhead and code, and is a VERY easy way to introduce odd
// bugs.  Use of pointers would also result in larger, slower code for every
// reference to Prev, Cur, and Next.  Overall, copying everything here is
// actually faster overall, simpler, and much safer.

  Prev = Cur;  Cur = Next;


//--- log msgs for bit chgs only if not starving other processes for cycles ---
  if (applLoopTime < 14)  logDigIOChgs(false);

/*EXAMPLES:
  if (Cur.rfgenInactive != Prev.rfgenInactive)
    printf("RF Gen PS %s responding\r\n",
           Cur.rfgenInactive ? "NOT" : "started");

  if (Cur.ovenInactive != Prev.ovenInactive)
    printf("Oven PS %s responding\r\n",
           Cur.ovenInactive ? "NOT" : "started");


  if (rfgenStartCmd)  {
    printf("Start RF Gen PS request\r\n");
    if (Cur.rfgenReady)  printf("RF Gen PS already ON\r\n");
  }
  if (rfgenStopCmd)  {
    printf("Stop RF Gen PS request\r\n");
    if (!Cur.rfgenHeatOn AND !Cur.rfgenReady)
      printf("RF Gen PS already OFF\r\n");
  }
  if (rfgenResetCmd)  printf("Reset RF Gen PS request\r\n");
*/

}


//=============================================================================
//  Update the values in "Next" based on the "Cur" and "Prev" values.
//
//  Be sure to read the "NOTES on Locking" section above....
//=============================================================================
//
//  This function contains the Synchronous logic used to compute the next
//  state we want the controller to be in.
//
//  A few simple rules that apply to all the values used by this function:
//
//    - All values changed within this function are "outputs"
//
//    - Every value read by this function is an "input"
//
//    - There must be no overlap between inputs and outputs
//
//    - The output values from this function must not be altered outside of
//      this function.
//
//  Any global variable can be an input OR an output, but they must NEVER
//  serve both purposes.  Most inputs come from Prev or Cur, while most
//  outputs are in Next.  If the logic requires detecting a change in a
//  value between passes (edge-triggered logic), then the value should be
//  part of the ApplState structure so we can compare its Prev and Cur values.
//
//  Following these simple rules allows the code to be organized as small,
//  self-contained blocks, with well-defined inputs and outputs.  Furthermore,
//  there are no order-of-execution dependencies between blocks of code.  The
//  blocks can be executed in any order:  The result will always be the same.
//
//
//  Each output value (or VERY small "set" of output values) must be updated
//  in a single, self-contained block of code.  NO output may be referenced
//  OR altered in ANOTHER block of code:  This is a declarative, data-oriented
//  style of programming and has many advantages including:
//
//     - There are no order-of-execution dependencies across nodes/blocks
//       within a pass.  They can be executed in any order and the result
//       will be the same (get used to it - this method will be the norm as
//       computer languages and hardware add support for atomic transactions
//       and computers evolve in to neural-net-like processor arrays - FPGAs
//       are just the beginning of this trend).
//
//     - Encapsulating all logic for an output within a single node/block
//       leaves just one place to start when tracking a problem backward from
//       the result.  Far simpler than debugging code structured around the
//       traditional process/event-driven approach which can contain numerous
//       independent paths of execution leading to an observed result.
//
//  Each node/block-of-code should be prefaced with the following information:
//   - The name(s) of the value(s) it modifies
//   - A descr of the meaning of the value(s) and the logic used
//   - A list of the values that serve as "inputs" to the node
//
//=============================================================================
void updateNextState(void)
{

  if (current != applTask)  return;  // must only be called by "applic" task

//APPLCODE

/* ===== EXAMPLES =====

====== ON/OFF/Reset state =======
//--------------------------------------------------------------------
// Next.state
// Next.onRequest, Next.offRequest, Next.resetRequest (resets)
//    Description:
//       Determines the next state of for the application logic
//       (STATE_ON, STATE_OFF, STATE_INTLK)
//    Inputs:
//       Cur.onRequest
//       Cur.offRequest
//       Cur.resetRequest
//       Cur.externIntlk
//       Cur.bulkSup[]
//       holecOK
//       psReady
//----------------------------------------------------------------------
//--- chg state from OFF to ON? ---
  if (Cur.onRequest)  do  {
    Next.onRequest = false;
    if (!supplyType)  {
      printf("Must set supplyType 1st!\n");  break;  }
    if ((Cur.state == STATE_INTLK) OR Cur.externIntlk)  {
      printf("Must clear Interlock first\n\n");  break; }
    if (Cur.state == STATE_ON)  { printf("Already ON\n\n");  break; }
    Next.state = STATE_ON;  printf("Turning ON\n\n");
  } while(0);

//--- chg state from ON to OFF? ---
  if (Cur.offRequest)  do {
    Next.offRequest = false;
    if ((Cur.state == STATE_INTLK) OR Cur.externIntlk)  {
      printf("Must clear Interlock first\n\n");  break; }
    if (Cur.state == STATE_OFF)  { printf("Already OFF\n\n");  break; }
    Next.state = STATE_OFF;  printf("Turning OFF\n\n");
  } while(0);

//--- chg state from INTLK to OFF? ---
  if (Cur.resetRequest)  do {
    Next.resetRequest = false;
    if (Cur.state != STATE_INTLK)  {
      printf("No Interlock to clear\n\n");  break;  }
    if (Cur.externIntlk)  {
      printf("External Interlock still exists\n\n");  break; }
    Next.state = STATE_OFF;  printf("Interlock cleared\n\n");
  } while(0);

//--- chg state to INTLK due to external interlock? ---
  if (Cur.externIntlk OR !holecOK)  do {
    if (Cur.state == STATE_INTLK)  break;
    Next.state = STATE_INTLK;
  } while(0);
//--- change state to INTLK due to a bulk supply error? ---
  for (ui=0; ui<NumPS; ++ui)  if (Cur.bs[ui].error)  Next.state = STATE_INTLK;


====== Pulse a digital output =====
//--------------------------------------------------------------------
// Next.rfgenResetTime
//    Description:
//       The time a pulse was started to RESET the RF Gen supply.
//    Inputs:
//       cmdBitsAppl
//--------------------------------------------------------------------
  if (cmdBitsAppl & CBA_RFGenReset)  Next.rfgenResetTime = GetMS();
  if (msSince(Next.rfgenResetTime) >= 500)  Next.rfgenResetTime = 0;

*/


}



//=============================================================================
//  Next we update output signal values to be written to the outputs at
//  the beginning of the next pass (just after the inputs are read).
//
//  Be sure to read the "NOTES on Locking" section above....
//=============================================================================
void updateOutputState(void)
{
  if (current != applTask)  return;  // must only be called by "applic" task

//APPLCODE

//--- update output values that are based on current conditions (values ---
//--- in "Cur" or any values used in a conditional in the applic logic) ---

// NOTE:  Do NOT clear all the bits of a bit-mapped value unless you are
//        going to check the state of every bit every time (i.e. clearing
//        every bit and then forgetting about some macro or piece of code
//        shared by multiple applications to manage a bit is an easy way
//        to mess things up - so do NOT touch bits that aren't specifically
//        handled by the function that clears them!

/* ===== EXAMPLES =====

  // (insured earlier that only ONE of these can be true when setting them)
  if (initialBtnPressed)  setVal[FP_METER] = rfOffSetpoint;  else
  if (finalBtnPressed)    setVal[FP_METER] = rfOnSetpoint;  else
                          setVal[FP_METER] = cavityVolts;


  SetOffApplLEDs(Next.digitalOut);

  if (limited)  SetOnLimitLED(Next.digitalOut);
  if (regErr)   SetOnErrorLED(Next.digitalOut);


//--- update output values based on values that were set by the application ---
//--- logic (e.g. values in "Next" or variables that are the target of an ---
//--- assignment statement in the application logic) ---

  setVal[CTL_VOLTS] = Next.ctlVolts;  // ctl voltage to DC Servo amplifier

  setVal[EXTERN_METER] = Next.meterValue;
*/

}




//=============================================================================
// The last thing we do is update values read by other tasks
//
//  Be sure to read the "NOTES on Locking" section above....
//=============================================================================
void updateExternals(void)
{
  if (current != applTask)  return;  // must only be called by "applic" task

  digInBits = Cur.digitalIn;
  digOutBits = Cur.digitalOut;

//APPLCODE

/*Example
  iset = Cur.iset;
  polCmd = Cur.polCmd;
  ampApply = Cur.ampApply;

  state = Cur.state;

  ampSet = Cur.ampSet;


  vctrlStatus = (Cur.regPol ? 1 : 0);
  vctrlStatus |= (Cur.hystRegPol ? 2 : 0);

  digInBits = Cur.digitalIn;

  psCount = Cur.psCount;

  ifdbk = Cur.ifdbk;
  vout = Cur.vout;
  vctrl = Cur.vctrl;
  vlim = Cur.vlim;

  outMode = Cur.outMode;


  for (ui=0; ui<NumPS;  ++ui)  {
    bsCur = Cur.bs + ui;

    *bsvX[ui] = Cur.biasVolts;
    *ampsX[ui] = Cur.current;

    bsState = 0;

    if (Cur.isOn)      bsState |= PS_IS_ON;
    if (Cur.isAvail)   bsState |= PS_AVAIL;
    if (Cur.tempOK)    bsState |= PS_TEMP_OK;
    if (!Cur.biasErr)  bsState |= PS_BIAS_OK;

    if (Cur.setOn)     bsState |= PS_SET_ON;
    if (Cur.enabled)   bsState |= PS_ENABLED;

    PS_STATE(ui) = bsState;
  }
*/

}



//=============================================================================
_nearcall float readGetRaw(uint ui)  {
  return (ui < NumHdwADC ? anaIn[ui] : 0.0);  }

_nearcall float readMinRaw(uint ui)  {
  return (ui < NumHdwADC) ? (adcMinRaw(ui) - adcOffset[ui]) * adcGain[ui] : 0.0; }

_nearcall float readMaxRaw(uint ui)  {
  return (ui < NumHdwADC) ? (adcMaxRaw(ui) - adcOffset[ui]) * adcGain[ui] : 9.8; }

_nearcall void readDefaults(uint ui, float *gain, float *offset)  {
  *gain = 1.0;  *offset = 0.0; }

_nearcall void readCalibPoints(uint ui, float *low, float *high)  {
  *low = -9.5;  *high = 9.5; }

_nearcall char *readUnits(uint ui, char *buf)  { return "V"; }


//=============================================================================
_nearcall float setGetRaw(uint ui)  {
  return (ui < NumHdwADC ? anaOut[ui] : 0.0);  }

_nearcall float setMinRaw(uint ui)  {
  return (ui < NumHdwDAC) ? (dacMinRaw(ui) - dacOffset[ui]) * dacGain[ui] : 0.0; }

_nearcall float setMaxRaw(uint ui)  {
  return (ui < NumHdwDAC) ? (dacMaxRaw(ui) - dacOffset[ui]) * dacGain[ui] : 9.8; }

_nearcall void setDefaults(uint ui, float *gain, float *offset)  {
  *gain = 1.0;  *offset = 0.0;  }

_nearcall void setCalibPoints(uint ui, float *low, float *high)  {
  *low = -9.5;  *high = 9.5;  }

_nearcall char *setUnits(uint ui, char *buf)  { return "V"; }


//=============================================================================
//  Set default labels for all the I/O signals
//=============================================================================
void setDefaultLabels(void)
{
  printf("Setting default labels for I/O signals\r\n");

//--- set application-appropriate default labels ---
//APPLCODE

//--- set default labels for the I/O signals ---
//--- Fixed size arrays are used so the values can be changed at run time ---
//--- (i.e. if pointers where used instead, and initialized to const string ---
//--- values, then we couldn't change them at run time) ---
/*
  fstrcpy(ainLabel[0], "ai0-Descr");
  fstrcpy(ainLabel[1], "ai1-Descr");

  fstrcpy(aoutLabel[0], "ao0-Descr");
  fstrcpy(aoutLabel[1], "ao1-Descr");

  fstrcpy(dinLabel[0], "di0-Descr");
  fstrcpy(dinLabel[1], "di1-Descr");

  fstrcpy(doutLabel[0], "do0-Descr");
  fstrcpy(doutLabel[1], "do1-Descr");
*/
}


//=============================================================================
//  This function is called to force software configurable values to default
//  values.
//=============================================================================
void setDefaultConfig(void)
{
  printf("Setting in-RAM device config to default values\r\n");

//===== set default values for persistent data values =====
//APPLCODE

}


//=============================================================================
//  This function is called by the application loop just before a shutdown.
//  The appliation must do whatever is necessary for a clean shutdown and
//  then exit
//=============================================================================
void applicShutdown(void)
{
  printf("Application prepping for shutdown\r\n");  fflush(stdout);

//APPLCODE


  Exit(0);
}


//=============================================================================
//  This is the highest priority task, usually resumed at a regular interval
//  so that it can do what it needs and sleep again until the start of the
//  next interval.
//=============================================================================
void _task applic(int argc, char **argv)
{
  #define SharedApplLocals
  #include <sharedAppl.c>

//--- add local variable definitions here ---
//APPLCODE


  #define SharedApplLocals2
  #include <sharedAppl.c>



  #define SharedApplInit
  #include <sharedAppl.c>

//--- At this point generic initialization has been done which includes ---
//--- the following: ---
//      - The Prev, Cur, and Next ApplState objects have been zeroed
//      -


//--- add application initialization code here ---
//APPLCODE


  #define SharedApplInit2
  #include <sharedAppl.c>


//--- The following loop is run at a periodic interval (normally 100 times ---
//--- per second, or every 10ms).  The application process normally ---
//--- consists of reading the current state of the digital and analog ---
//--- inputs, changing the state of the analog and digital outputs to the ---
//--- values computed during the previous pass, then computing new values ---
//--- for the outputs and updating values that outside clients can view ---
//--- (currently accessable as a set of ModBus registers) ---

  for (;;)  {
    #define SharedApplLoopStart
    #include <sharedAppl.c>


  //--- read the current state of the analog and digital Input signals, and ---
  //--- write the values computed at the end of the last pass ---
    digitalOut = Next.digitalOut;  rwIO();

  //--- process next event message (if any) ---
    processMsg(eventsfd, (EventMsg *)buf, sizeof(buf));

  //--- update the current state and determine the next state ---
    updateCurState();  updateNextState();
    updateOutputState();  updateExternals();


    #define SharedApplLoopEnd
    #include <sharedAppl.c>
  }

}


//=============================================================================
//  list of application-specific commands that are normally entered via a
//  user shell
//=============================================================================
//void _task nsclSeriald(int argc, char **argv);

//--- add any application-specific commands here ---
//APPLCODE
const CmdInfo far applCmds[] = {
//  { "misc",        miscTest,       "..." },
  { "cmd",       cmd,         "Application-specific commands" },
  { "sem_test",  sem_test,    "Test POSIX-style semaphore funcs" },
  { "state",     showState,   "Show state of controller" }
};

int const NumApplCmds = ( sizeof(applCmds) / sizeof(CmdInfo) );

//-----------------------------------------------------------------------------

#endif
