//----- spi.c ----- 01/08/13 -----(05/05/05)-----

//-----------------------------------------------------------------------------
//  NOTE:  While multiple serial interfaces can be configured for SPI use,
//         only ONE can be IN use (transmitting or receiving) at a time!
//-----------------------------------------------------------------------------
//  Update log (most recent first)
//
//
//  2013-01-08 - Mark Davis (*)
//               - Added .cseg directive to SPITxRx assembly func.  Prior to
//                 this, it was being located in FARCODE, so the calls to it
//                 (which are all local ones) would work only as long as the
//                 call came from within the same 4K memory block.
//
//  2012-02-08 - Mark Davis
//               - errno is now set to EBADF rather than ENXIO when a function
//                 is passed an invalid device pointer.
//
//  2011-12-09 - Mark Davis
//               - Changed macro name, modified comments
//
//  2010-03-09 - Mark Davis
//               - The "invalid SPI port" msg is now logged only if the
//                 DebugBits are set (i.e. ShowDebugMsgs is true)
//
//  2007-03-26 - Mark Davis
//               - added check for non-NULL devio pointer in most funcs
//
//  2006-05-16 - Mark Davis -
//               -  ported from earlier ZWorld compiler compatible version
//
//-----------------------------------------------------------------------------
//  To-Do:
//
//    - Rabbit 3000 manual has two references to the ability to set the order
//      of the bits (just search on the word "order").  Supposedly this is
//      controlled via the Serial Port Extended Registers for ports A-D (SxER),
//      but the description of that register has a description for each bit
//      and none of them mentions anything about the bit order.
//
//
//    - Revisions A-C of the Rabbit 2000 include the "transceive" feature
//      that allows a transmit and receive (full-duplex) operation to be
//      started at the same time without having to issue the receive command
//      within one-half bit time after issuing a transmit command.  This could
//      be used to free up a handful of CPU cycles per char (a tiny bit, but,
//      it also avoids disabling interrupts to guarantee the timing of
//      issuing one then the other).
//
//
//    - Add code for all the Rabbit 3000 features and what ports support
//      them, then use the CPU ID and revision values to determine if/when
//      they can be used.
//
//
//    - It might speed things up if I added support for interrupt driven I/O.
//
//      Like in async mode, the ports each have a 1-byte buffer that can be
//      read or written to while the hdw is busy reading a byte into the rcv
//      shift register or writing one from the xmit shift include.
//
//      The possible gain comes from the fact that even at the current max
//      clock freq of 2MHz, the non-interrupt driven method ends up looping
//      several times waiting for an operation to finish:  CPU cycles that
//      could be used to do something useful.
//
//      NOTE:  I should check this for the read, write, and transceive (w/r)
//             modes to see just how long it is a different rates.  If it IS
//             long enough, then the added overhead of the ISR (very little)
//             and copying data to/from buffers would still leave enough for
//             the applic code to do processing during what was wasted CPU
//             cycles just waiting for the current SPI operation to complete.
//
//             It might also allow each port to be running simultaneously
//             at different clock rates.  As long as the Rabbit is the one
//             clock, then it can always read the current byte in the
//             1-byte rcv buffer before it initiates the next rcv operation
//             (which generates the clock pulses that cause the sending
//             device to send the next byte).
//
//      Of course it gets a bit trickier when additional handshake signals
//      are needed.  For example, the PhaseCtlr has to communicate with the
//      DSP usign 4 byte "packets" and toggle an output bit to indicate the
//      start and end of each packet of 4 bytes (and it has use the
//      transceive mode where it is simultaneously sending and receiving a
//      byte.  For this, I would have to have a way to supply an ISR-callable
//      func (possibly even two different funcs) to manage the output and
//      some whay to tell it WHEN to call the function(s).
//
//-----------------------------------------------------------------------------

#include <common.h>

#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>

#include <errno.h>

#include <sys/serial.h>
#include <sys/spi.h>

#include <Rabbit.h>
#include <dcdefs.h>

//#include <sys/farprintf.h>



#define SPI_ICLK  0x0C             // clocked mode using internal clock
#define SPI_RX    0x40 | SPI_ICLK  // start clocked byte receive
#define SPI_TX    0x80 | SPI_ICLK  // start clocked byte transmit

//char  SPIxor;
//char  SPITXxor;

//char  spiRXxor[4];    // flags for inverting the received byte
//char  spiTXxor[4];    // flags for inverting the transmitted byte

uchar  SPI_DR, SPI_CR, SPI_SR, SPI_ER;

const uchar  serDataReg[4] = { SADR, SBDR, SCDR, SDDR };
const uchar  serCtlReg[4]  = { SACR, SBCR, SCCR, SDCR };
const uchar  serStatReg[4] = { SASR, SBSR, SCSR, SDSR };
const uchar  serExtReg[4]  = { SAER, SBER, SCER, SDER };
const uchar *pShExtReg[4]  =
                       { &SAERShadow, &SBERShadow, &SCERShadow, &SDERShadow };

const uchar  TATReg[4] = { TAT4R, TAT5R, TAT6R, TAT7R };

const uchar *pTATshReg[4] =
                    { &TAT4RShadow, &TAT5RShadow, &TAT6RShadow, &TAT7RShadow };

IODEV_FUNCS(spi)

bool  spiInUse;

const uchar RevBitTable[] = {
  0x00, 0x80, 0x40, 0xC0, 0x20, 0xA0, 0x60, 0xE0, 0x10, 0x90, 0x50, 0xD0, 0x30, 0xB0, 0x70, 0xF0,
  0x08, 0x88, 0x48, 0xC8, 0x28, 0xA8, 0x68, 0xE8, 0x18, 0x98, 0x58, 0xD8, 0x38, 0xB8, 0x78, 0xF8,
  0x04, 0x84, 0x44, 0xC4, 0x24, 0xA4, 0x64, 0xE4, 0x14, 0x94, 0x54, 0xD4, 0x34, 0xB4, 0x74, 0xF4,
  0x0C, 0x8C, 0x4C, 0xCC, 0x2C, 0xAC, 0x6C, 0xEC, 0x1C, 0x9C, 0x5C, 0xDC, 0x3C, 0xBC, 0x7C, 0xFC,
  0x02, 0x82, 0x42, 0xC2, 0x22, 0xA2, 0x62, 0xE2, 0x12, 0x92, 0x52, 0xD2, 0x32, 0xB2, 0x72, 0xF2,
  0x0A, 0x8A, 0x4A, 0xCA, 0x2A, 0xAA, 0x6A, 0xEA, 0x1A, 0x9A, 0x5A, 0xDA, 0x3A, 0xBA, 0x7A, 0xFA,
  0x06, 0x86, 0x46, 0xC6, 0x26, 0xA6, 0x66, 0xE6, 0x16, 0x96, 0x56, 0xD6, 0x36, 0xB6, 0x76, 0xF6,
  0x0E, 0x8E, 0x4E, 0xCE, 0x2E, 0xAE, 0x6E, 0xEE, 0x1E, 0x9E, 0x5E, 0xDE, 0x3E, 0xBE, 0x7E, 0xFE,
  0x01, 0x81, 0x41, 0xC1, 0x21, 0xA1, 0x61, 0xE1, 0x11, 0x91, 0x51, 0xD1, 0x31, 0xB1, 0x71, 0xF1,
  0x09, 0x89, 0x49, 0xC9, 0x29, 0xA9, 0x69, 0xE9, 0x19, 0x99, 0x59, 0xD9, 0x39, 0xB9, 0x79, 0xF9,
  0x05, 0x85, 0x45, 0xC5, 0x25, 0xA5, 0x65, 0xE5, 0x15, 0x95, 0x55, 0xD5, 0x35, 0xB5, 0x75, 0xF5,
  0x0D, 0x8D, 0x4D, 0xCD, 0x2D, 0xAD, 0x6D, 0xED, 0x1D, 0x9D, 0x5D, 0xDD, 0x3D, 0xBD, 0x7D, 0xFD,
  0x03, 0x83, 0x43, 0xC3, 0x23, 0xA3, 0x63, 0xE3, 0x13, 0x93, 0x53, 0xD3, 0x33, 0xB3, 0x73, 0xF3,
  0x0B, 0x8B, 0x4B, 0xCB, 0x2B, 0xAB, 0x6B, 0xEB, 0x1B, 0x9B, 0x5B, 0xDB, 0x3B, 0xBB, 0x7B, 0xFB,
  0x07, 0x87, 0x47, 0xC7, 0x27, 0xA7, 0x67, 0xE7, 0x17, 0x97, 0x57, 0xD7, 0x37, 0xB7, 0x77, 0xF7,
  0x0F, 0x8F, 0x4F, 0xCF, 0x2F, 0xAF, 0x6F, 0xEF, 0x1F, 0x9F, 0x5F, 0xDF, 0x3F, 0xBF, 0x7F, 0xFF
};

#define SPI  ((SerialInfo *)ioDev)

//-----------------------------------------------------------------------------
//  Kernel function to initialize a serial port for SPI use
//
//  Intended to be called ONLY from serial.c's _serial_() func!!!!
//
//  Attributes:
//     port:         serial port # (0-3)
//     clkDiv        clock divider (>= 5) (see spi.h)
//     clkMode:      SPI clode mode (0-3)  (see spi.h)
//-----------------------------------------------------------------------------
bool  _spiInit_(void *ioDev, uchar clkDiv, uchar clkMode)
{
  uchar  port;

//KLOCK()/KUNLOCK() done by _serial_() function (see serial.c)

  port = SPI->port;

  SPI->clkDiv = clkDiv;  SPI->clkMode = clkMode;

  SPI->serDataReg = serDataReg[port];  // serial Data Register
  SPI->serCtlReg  = serCtlReg[port];   // serial Control Register
  SPI->serStatReg = serStatReg[port];  // serial Status Register
  SPI->serExtReg  = serExtReg[port];   // serial Extended Register
  SPI->pShExtReg  = pShExtReg[port];   // ptr to "shadow" for Extended Register
  SPI->TATReg     = TATReg[port];      // Timer A Time const reg (timer divider)
  SPI->pTATshReg  = pTATshReg[port];   // ptr to "shadow" for Time constant reg

  if (port == 0)
    BitWrPortI ( PCFR, &PCFRShadow, 1, 6 );   // enable serial A UART output
  else if (port == 1)
    BitWrPortI ( PCFR, &PCFRShadow, 1, 4 );   // PC4 = serial B UART output
#if (_CPU_ID_ >= R3000)
  else  if (port == 2)  {
    BitWrPortI ( PCFR, &PCFRShadow, 1, 2 );   // PC2 = serial C UART output
    BitWrPortI ( PFDDR, &PFDDRShadow, 1, 1 ); // set PF1 as output
    BitWrPortI ( PFFR, &PFFRShadow, 1, 1 );   // set PF1 as serial clock
  }
  else  if (port == 3)  {
    BitWrPortI ( PCFR, &PCFRShadow, 1, 0 );   // PC0 = serial D UART output
    BitWrPortI ( PFDDR, &PFDDRShadow, 1, 0 ); // set PF0 as output
    BitWrPortI ( PFFR, &PFFRShadow, 1, 0 );   // set PF0 as serial clock
  }
#endif

//--- Read from PC#; internal clock; Disable interrupts ---
  WrPortI ( serCtlReg[port], NULL, SPI_ICLK );

  _spiConfig_(SPI, clkDiv, clkMode);

  RdPortI ( serDataReg[port] );  // dummy read to empty the buffer


  return true;
}


//-----------------------------------------------------------------------------
//  Returns true if provided pointer is a valid, in-use SPI Serial device.
//  Returns false if is not.
//-----------------------------------------------------------------------------
bool  validSPI(void *ioDev)
{
  if (ioDev)
    if (SPI == serPort + (int)(SPI - serPort))
      if (SPI->inUse)  return true;

  if (ShowDebugMsgs)
    fprintf(DiagLog, "***** invalid SPI port: [%04X] *****\n", (uint)ioDev);

  return false;
}

//-----------------------------------------------------------------------------
int _spiGetQueues_(void *ioDev, WaitQueues far *queues)
{
  queues->readQueue = queues->writeQueue = &sleepQueue;
  return 0;
}

//----------------------------------------------------------------------------
//  Attempt to read bytes from an SPI port
//
//  Should only be called from the fcntl.c/read() function!
//
//  RETURN VALUEs:
//  - if bytes read: # of bytes read
//  - if no bytes read:
//    - if more possible in the future: -1 and ERRNO == EAGAIN
//    - if no more possible (EOF):  0
//----------------------------------------------------------------------------
#pragma offset_labels on

int _spiRead_(void *ioDev, char *buf, int numBytes, bool dontBlock)
{
  if (!validSPI(ioDev))  { ERRNO = EBADF;  return -1; }

//--- SPI I/O not reentrant yet, so use a global lock for all SPI I/O ---

  _EnterCritical_();
  if (spiInUse)  {
    _ExitCritical_();
    fprintf(DiagLog, "***** nested _spiRead_() *****");  //tdebug
    ERRNO = EAGAIN;  return -1;
  }
  spiInUse = true;
  _ExitCritical_();


//--- global vars:  is this the MAIN reason the SPI lib is not reentrant? ---
//finish: If I passed these to other funcs, could all this be reentrant?
//  SPIxor = SPI->spiRXxor;
  SPI_DR = SPI->serDataReg;
  SPI_CR = SPI->serCtlReg;
  SPI_SR = SPI->serStatReg;
  SPI_ER = SPI->serExtReg;

#asm

; set up the registers
      push  ix            ; preserve

;;--- load DE with 0xFF for use in testing for the end of loop condition ---
      ld    a, 0xFF       ; set up for testing end of loop
      ld    d, a
      ld    e, a

;;--- get addr of buffer and the number of bytes to read ---
      ld    ix, (sp+2+.buf)       ; get the destination address
      ld    iy, (sp+2+.numBytes)  ; get the number of bytes

; de = byte count mask
; ix = destination address
; iy = byte count

      ld    bc, (_SPI_DR)
  ioi ld    a, (bc)       ; dummy read to empty any "junk"

;;--- start a read cycle ---
SPI_ReadA:
      ld    a, SPI_RX     ; set up RX control value
      ld    bc, (_SPI_CR)
  ioi ld    (bc), a       ; load RX control value: read a byte

;;--- wait for the receiver to complete ---
      ld    bc,(_SPI_SR)
Rx0:
  ioi ld    a, (bc)       ; read the status register
      and   0x80          ; test receiver bit
      jr    z, Rx0        ; jump if not done yet

;;--- get the byte (without inverting the bits) ---
      ld    bc, (_SPI_DR)
  ioi ld    a, (bc)       ; get the byte
      ld    c, a

;;--- alternative: supports the inverting of the bits ---
;;      ld    a, (SPIxor) ; get the byte invert value
;;      ld    L, a        ; save for later
;;      ld    bc, (_SPI_DR)
;;  ioi ld    a, (bc)     ; get the byte
;;      xor   L           ; invert the bits
;;      ld    c, a        ; c = byte with bits inverted

;--- reverse the order of the bits ---
      xor   a                ; a = 0
      ld    b, a             ; bc = offset into BitRev table
      ld    HL, _RevBitTable ; get address of table
      add   HL, bc           ; calculate address of bit reversed byte
      ld    L, (HL)          ; get new byte

;;--- store the byte in the buffer and check the count ---
      ld    (ix), l       ; store the byte
      inc   ix            ; point to next byte
      dec   iy            ; calculate bytes remaining
      and   iy, de        ; check if done: z=1 if done
      jr    nz, SPI_ReadA ; jump if not done

      pop ix

#endasm

  spiInUse = false;

  return numBytes;
}


//----------------------------------------------------------------------------
//  Attempt to write bytes to an SPI port
//
//  Should only be called from the fcntl.c/write() function!
//
//  RETURN VALUEs:
//  - if bytes written: # of bytes written
//  - if no bytes written:
//    - if more possible in the future: -1 and ERRNO == EAGAIN
//    - if no more possible (EOF):  0
//----------------------------------------------------------------------------
int _spiWrite_(void *ioDev, char *buf, int numBytes, bool dontBlock)
{
  if (!validSPI(ioDev))  { ERRNO = EBADF;  return -1; }

  _EnterCritical_();
  if (spiInUse)  { _ExitCritical_();  ERRNO = EAGAIN;  return 0; }
  spiInUse = true;
  _ExitCritical_();

//  SPITXxor = SPI->spiTXxor;
  SPI_DR = SPI->serDataReg;
  SPI_CR = SPI->serCtlReg;
  SPI_SR = SPI->serStatReg;
  SPI_ER = SPI->serExtReg;

#asm

; set up the registers
      push  ix            ; preserve

      ld    a, 0xFF       ; set up for testing end of loop
      ld    d, a
      ld    e, a          ; de = 0xFFFF

      ld    ix, (sp+2+.buf)      ; get the source address
      ld    iy, (sp+2+.numBytes)     ; get the number of bytes

; send the data bytes
; de = byte count mask
; ix = source address
; iy = byte count

SPI_WriteA:
//      ld    a, (SPITXxor)   ; get bit inversion value
      ld    l, (ix)     ; get a byte
//      xor l             ; invert if necessary
//      ld    l, a        ; restore for call

;;    call  SPI_Xmit    ; send it
;;--- swamp order of bits ---
      ld    c, L            ; copy the byte value
      xor   a               ; a = 0
      ld    b, a            ; bc = offset into BitRev table
      ld    HL, _RevBitTable  ; get address of table
      add   HL, bc          ; calculate address of bit reversed byte
      ld    a, (HL)         ; get bit-reversed version of byte value

      ld    bc,(_SPI_DR)
  ioi ld    (bc), a         ; load it
      ld    a, SPI_TX       ; get TX control value
      ld    bc,(_SPI_CR)
  ioi ld    (bc), a         ; load TX control value - send the byte

; wait for the buffer to be available
      ld    bc,(_SPI_CR)    ; get addr of Control Register
Tx0:
  ioi ld    a, (bc)         ; read the Control Register
      and   0x08            ; test buffer busy bit
      jr    nz, Tx0         ; repeat if not done

      inc   ix              ; point to next byte
      dec   iy              ; update bytes remaining
      and   iy, de          ; check if done (iy = 0): z=1 if done
      jr    nz, SPI_WriteA  ; jump if not done

;;--- wait for last byte to complete ---
      ld    bc,(_SPI_SR)     ; get addr of Status Register
SPI_WaitXmit:
  ioi ld    a, (bc)          ; read the Status Register
      and   0x04             ; test transmitter busy bit
      jr    nz, SPI_WaitXmit ; repeat if not done

      pop ix
#endasm

  spiInUse = false;

  return numBytes;
}

#pragma offset_labels off


//-----------------------------------------------------------------------------
//  Transmit and Receive a byte via a serial port set up in synchronous mode
//
//  Input parameter: the byte value to transmit: in L
//  Return value: the received byte in L
//
//  Uses: A, HL, BC
//-----------------------------------------------------------------------------
#asm nodebug root
     .cseg

;; NOTE:  16-bit I/O instrs prefixed with ioi use only the lower byte of the
;;        16-bit address (e.g. ioi ld (hl),a  uses only L for the I/O address)

SPITxRx::
;;--- swap the bit order ---
      ld     c, l             ; c = byte value to xmit
      xor    a                ; a = 0
      ld     b, a             ; bc = offset into BitRev table
      ld     hl, _RevBitTable ; get address of table
      add    hl, bc           ; calculate address of bit reversed byte
      ld     a, (hl)          ; get bit-reversed vers of byte to send

      ld     hl,(_SPI_DR)
  ioi ld     (hl), a          ; load value in to the SPI Data Register

//#if (_CPU_ID_ >= R3000)       ; 2.10
//      ld     a, SPI_RX|SPI_TX ; get transceive value
//      ld     hl,(_SPI_CR)
//  ioi ld     (hl), a          ; initiate transceive operation
//#else
; the next 3 instructions minimize the time at IPL = 3
      ld     hl,(_SPI_CR)     ; get address of Control Register
      ld     c, SPI_RX        ; set up RX control value
      ld     a, SPI_TX        ; get TX control value
      ipset  3                ; disable interrupts
  ioi ld     (hl), a          ; load TX control value
      nop
  ioi ld     (hl), c          ; load RX control value
      nop
      ipres                   ; restore prev Interrupt Priority Level
//#endif

; wait for rcv to complete - (completes AFTER the xmt)
      ld     hl, (_SPI_SR)    ; get addr of status register
SPITxRx0:
  ioi ld     a, (hl)          ; read the status register
      and    0x80             ; test receiver done bit
      jr     z, SPITxRx0      ; repeat if not done

      ld     hl, (_SPI_DR)    ; get the received byte
  ioi ld     a, (hl)

;;--- reverse the order of the bits ---
      ld    c, a              ; copy the byte value
      xor   a                 ; a = 0
      ld    b, a              ; bc = offset into BitRev table
      ld    hl, _RevBitTable  ; get address of table
      add   hl, bc            ; calculate address of bit reversed byte
      ld    l, (hl)           ; get reversed-bit version


      ret  ;SPITxRx

#endasm



//----------------------------------------------------------------------------
//  Send AND Receive the specified # of bytes to/from an SPI port
//
//  Should normally only be called from the fcntl.c/read() function!
//
//  RETURN VALUEs:
//  - if bytes read: # of bytes read
//  - if no bytes read:
//    - if more possible in the future: -1 and ERRNO == EAGAIN
//    - if no more possible (EOF):  0
//----------------------------------------------------------------------------
#pragma offset_labels on

int _spiWrRd_(void *ioDev, char *xmitBuf, char *rcvBuf, int numBytes)
{
  if (!validSPI(ioDev))  { ERRNO = EBADF;  return -1; }

  _EnterCritical_();
  if (spiInUse)  { _ExitCritical_();  ERRNO = EAGAIN;  return -1; }
  spiInUse = true;
  _ExitCritical_();

//  SPIxor = SPI->spiRXxor;
  SPI_DR = SPI->serDataReg;
  SPI_CR = SPI->serCtlReg;
  SPI_SR = SPI->serStatReg;
  SPI_ER = SPI->serExtReg;

#asm

; set up the registers
      push   ix                    ; preserve ix
      ld     ix, (sp+2+.xmitBuf)   ; get the source address
      ld     iy, (sp+2+.rcvBuf)    ; get the destination address
      ld     hl, (sp+2+.numBytes)  ; and byte count
      ld     e, L             ; save byte count for processing

; send and receive the data bytes
SPIWrRdA:
      ld     L, (ix)          ; get a byte to transmit
      call   SPITxRx          ; send and receive a byte

      ld     (iy), L          ; save received byte
      inc    ix               ; point to next transmit byte
      inc    iy               ; point to next receive byte
      dec    e                ; update bytes remaining
      jr     nz, SPIWrRdA     ; jump if not done

      pop    ix

#endasm

  spiInUse = false;

  return numBytes;
}

#pragma offset_labels off


//----------------------------------------------------------------------------
//  Get the global SPI lock and set values used by the _xxxFast_() functions.
//
//  NOTE:  Nested locking/unlocking is NOT supported, and _spiUnlock_()
//         MUST be called to release the global SPI lock.
//
//  Returns 0 if the lock is obtained and the _xxxFast_() functions can now
//  be used by the caller, < 0 if an error.
//----------------------------------------------------------------------------
int _spiLock_(void *ioDev)
{
  if (!validSPI(ioDev))  { ERRNO = EBADF;  return -1; }

  _EnterCritical_();
  if (spiInUse)  { _ExitCritical_();  ERRNO = EBUSY;  return -1; }
  spiInUse = true;
  _ExitCritical_();

//  SPIxor = SPI->spiRXxor;
  SPI_DR = SPI->serDataReg;
  SPI_CR = SPI->serCtlReg;
  SPI_SR = SPI->serStatReg;
  SPI_ER = SPI->serExtReg;

  return 0;
}

//----------------------------------------------------------------------------
//  Release the global SPI lock
//----------------------------------------------------------------------------
int _spiUnlock_(void *ioDev)
{
//  if (!validSPI(ioDev))  { ERRNO = EBADF;  return -1; }

  _EnterCritical_();
  if (!spiInUse)  { _ExitCritical_();  ERRNO = EACCES;  return -1; }
  spiInUse = false;
  _ExitCritical_();

  return 0;
}


//----------------------------------------------------------------------------
//  Send AND Receive the specified # of bytes to/from an SPI port.
//
//  NOTE:  _spiLock_() must be called before calling this function to get the
//         global SPI lock and initialize values used to read/write from/to
//         the port, and _spiUnlock_() must be called later to release the
//         global SPI lock.
//
//  RETURN VALUE:  # bytes written/read
//----------------------------------------------------------------------------
// NOTES on speed:
//
//   Possible use of a far * for the rcvBuf:
//
//     All reads and writes from/to a far address operate on 2 bytes at once.
//     If the address is at the end of a 64K boundary (0x?FFFF), the R/W
//     operation will wrap back around to the beginning of the 64K boundary.
//     And even if it doesn't cross a 64K boundary, crossing a 4K boundary
//     could result in a write to a write-protected 4K block of memory.
//
//     But my tests with having the caller use a larger near buffer for
//     calls to this func within the loop (and then copying the whole
//     near buffer to a far, a MUCH faster process then all the additional
//     code that goes in to using and upating a far pointer to a long after
//     each call to this func for the next 4 bytes).
//
//
//   What uses most the time?
//
//      The majority of the time spent is in the actual xmit/rcv process in
//      SPITxRx.  If I eliminate the bit reversing, there is only a tiny
//      improvement.  If I simply don't wait for the rcv done bit in the
//      status register, that cuts the time way down (but of course it
//      probably means I get garbage).
//
//
//   Bottom line:
//
//      The main limitation seems to be the actual xmit/rcv step in SPITxRx.
//      All the other things I tried, expecting to make a sizeable impact on
//      the overall speed each had only minor effects.  All added together
//      they might add up to a small boost, but the max clock speed for
//      synchronous transfers seems to be the major limitation at this time.
//----------------------------------------------------------------------------
#pragma offset_labels on

int _spiWrRdFast_(char *xmitBuf, char *rcvBuf, int numBytes)
{
;

#asm

; set up the registers
      push   ix                    ; preserve ix

      ld     ix, (sp+2+.xmitBuf)   ; get the source address
      ld     iy, (sp+2+.rcvBuf)    ; get the destination address
      ld     hl, (sp+2+.numBytes)  ; and byte count
      ld     e, L             ; save byte count for processing

; send and receive the data bytes
SPIWrRdAF:
      ld     L, (ix)          ; get a byte to transmit
      call   SPITxRx          ; send and receive a byte
      ld     (iy), L          ; save received byte

      inc    ix               ; point to next transmit byte
      inc    iy               ; point to next receive byte

      dec    e                ; update bytes remaining
      jr     nz, SPIWrRdAF    ; jump if not done

      pop    ix

#endasm
;

  return numBytes;
}

#pragma offset_labels off



//----------------------------------------------------------------------------
//  change the configuration of an SPI serial connection
//----------------------------------------------------------------------------
int  _spiConfig_(void *ioDev, uchar clkDiv, uchar clkMode)
{
  if (!validSPI(ioDev))  { ERRNO = EBADF;  return -1; }

  if (clkDiv < 5)  clkDiv = 5;  // min divisor for max stable rate

//--- set clock divider for timer associated with clocked serial port output ---
  WrPortI( SPI->TATReg, SPI->pTATshReg, clkDiv);
  SPI->clkDiv = clkDiv;

//--- for Rabbit 3000 processors, set clock mode (inactive level and phase) ---
#if (_CPU_ID_ >= R3000)
  WrPortI( SPI->serExtReg, SPI->pShExtReg, ((clkMode+1) & 3) << 4 );
  SPI->clkMode = clkMode;
#endif

  return 0;
}

//-----------------------------------------------------------------------------
//  release an SPI port for use by another process
//-----------------------------------------------------------------------------
int _spiClose_(void *ioDev)
{
  if (!validSPI(ioDev))  { ERRNO = EBADF;  return -1; }

  KLOCK();  SPI->inUse = false;  KUNLOCK();

  return 0;
}


//-----------------------------------------------------------------------------
//  Kernel-space function to reposition a file offset
//
//  WARNING!  To be called ONLY by KERNELFUNC functions (or functions called BY
//            KERNELFUNC functions) !!!!
//-----------------------------------------------------------------------------
off_t  _spiLseek_(void *ioDev, off_t offset, int whence)
{
  if (!validSPI(ioDev))  { ERRNO = EBADF;  return -1; }

  ERRNO = ESPIPE;  return (off_t)-1;
}


//-----------------------------------------------------------------------------
//  Get information about ...
//-----------------------------------------------------------------------------
int  _spiFstat_(void *ioDev, struct stat *fsbuf)
{
  if (!validSPI(ioDev))  { ERRNO = EBADF;  return -1; }

  fsbuf->st_mode = S_IFCHR | 00777;

  return 0;
}


//-----------------------------------------------------------------------------
//  Kernel-space function to handle ioctl calls for SPI serial ports
//-----------------------------------------------------------------------------
int _spiioctl_(void *ioDev, int request, void *arg)
{
  int  stat;


  if (!validSPI(ioDev))  { ERRNO = EBADF;  return -1; }

  stat = -1;  KLOCK();

  switch (request)  {
    case IOC_SET_CLK_RATE:
      _spiConfig_(ioDev, *((uchar *)arg), SPI->clkMode);  stat = 0;  break;

#if (_CPU_ID_ >= R3000)
    case IOC_SET_CLK_MODE:
      _spiConfig_(ioDev, SPI->clkDiv, *((uchar *)arg));  stat = 0;  break;
#endif

    default:
      ERRNO = EINVAL;
  }

  KUNLOCK();  return stat;
}


//-----------------------------------------------------------------------------
//  Kernel-space function to return a description of the serial port
//-----------------------------------------------------------------------------
char *_spiDescr_(void *ioDev, char *buf, int bufSize)
{
  char  *ptr;

  sprintf(buf, "spi 0x%04lX ", (ulong)ioDev);

  ptr = buf + strlen(buf);  strcpy(ptr, "bad bad");

//finish: limit output to bufSize

  if (validSPI(ioDev))
    sprintf(ptr, "ser[%d] /dev/spi%d mode:%d rate:%ld",
                 (uint)(SPI - serPort), SPI->port,
                 SPI->clkMode, (ulong) 10000000UL / SPI->clkDiv);

  return buf;
}


//-----------------------------------------------------------------------------
