//----- starterFS.c ----- 02/08/12 -----(05/22/06)-----

/*---------------------------------------------------------------------------*
 * To create a new driver:
 *  - make a copy of this file and change all occurances of "starterFS" to
 *    name for new driver
 *  - copy and modify include/sys/starterFS.h using name of new driver
 *---------------------------------------------------------------------------*/

//-----------------------------------------------------------------------------
//  Standard DEBROS virtual file/device I/O API for reading and writing to
//  ...
//-----------------------------------------------------------------------------
//  Update log (most recent first)
//
//
//  20xx-xx-xx - xxxx xxxxx
//               - original version
//
//-----------------------------------------------------------------------------

#include <common.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include <errno.h>
#include <common.h>

#include <string.h>

#include <sys/starterFS.h>

#include <sys/farprintf.h>

#include <linux/limits.h>


#define  MaxFDs   2   // max # of open file descriptors for this device

#define  DevDescr     "starterFS"
#define  NodeDescr    "node"


typedef struct  {
  int         openCount;   // # of FDEV's that point to this entry
  char far   *name;
  bool        rdOK;        // open for reading
  bool        wrOK;        // open for writing
  uint        unreadBytes; // # of bytes available for reading
  uint        freeBytes;   // # of unused bytes in output buffer
  ulong       offset;      // offset for next r/w operation
  ulong       diagFlags;
} FileInfo;


static FileInfo  openFiles[MaxFDs];

char far  fileNames[MaxFDs][NAME_MAX];   // name associated with the xxx


FILESYS_FUNCS(starterFS)

IODEV_FUNCS(starterFS)

#define FileRef  ((FileInfo *)ioDev)


//-----------------------------------------------------------------------------
//  Initialize data structures for managing named pipes (starterFS streams)
//-----------------------------------------------------------------------------
void starterFSInit_(void)
{
  uint  u;

  BZERO(openFiles);
  for (u=0; u<MaxFDs; ++u)  {
    openFiles[u].name = fileNames[u];
  }

  starterFSFSysFuncs.tick = NULL;

  registerFileSys(&starterFSFSysFuncs);
}

//-----------------------------------------------------------------------------
//  Returns true if pointer is a valid, in-use file ref for this device
//  Returns false if is not.
//-----------------------------------------------------------------------------
bool  validFileRef(void *ioDev)
{
  if (ioDev)
    if (FileRef == openFiles + (int)(FileRef-openFiles))
      if (FileRef->openCount)  return true;

  return false;
}

//-----------------------------------------------------------------------------
//  Return ptr into table of existing files, or NULL if name not found
//-----------------------------------------------------------------------------
FileInfo *starterPtr(const char *path)
{
  uint  u;
  FileInfo  *fileRef;

//--- look for the specified name in the list of files's ---
  for (u=0; u<MaxFDs; ++u)  {
    fileRef = openFiles + u;
    if (fstrcmp(fileRef->name, path) == 0)  return fileRef;
  }

  return NULL;
}


//-----------------------------------------------------------------------------
//  return a file descriptor for a file stored on this device
//
//  To be called ONLY by fcntl functions
//-----------------------------------------------------------------------------
fd_t _starterFSOpen_(const char *path, int flags, int mode)
{
  uint  u;
  fd_t  fd;
  FileInfo *fileRef;


//--- ignore names that don't start with "xxxx" ---
  if (strncmp(path, "xxxx", 4))  { errno = ENXIO;  return -1; }


  KLOCK();

//--- look for the specified name in the device's list of files ---
  for (u=0; u<MaxFDs; ++u)  {
    fileRef = openFiles + u;
    if (fileRef->openCount)  continue;
    //... do whatever other checks are needed here ...
    break;
  }
  if (u == MaxFDs)  { KUNLOCK();  ERRNO = ENFILE;  return -1; }

//--- allocate a file descriptor to reference the device file ---
  if ((fd = _allocFile_(3, &starterFSIOFuncs, fileRef)) < 0)  {
    KUNLOCK();  return -1;  }

  fileRef->offset = 0;
  fileRef->openCount = 1;
  //... do whatever other initiailization is needed here ...

  KUNLOCK();

//--- everything OK - return the descriptor assigned for this process ---
  return fd;
}

//-----------------------------------------------------------------------------
//  func called each kernel tick to provide file system with CPU cycles
//  Should ONLY be called by the kernel task!
//-----------------------------------------------------------------------------
void _starterFSTick_(void)
{
  //  if this func needs to actually be called in the future, the line in
  //  starterFSInit_() that sets the pointer to this func to NULL must be
  //  commented out...
}


//-----------------------------------------------------------------------------
//  Return the state of the specified ioDev entry or NULL if not a valid
//  entry #.
//
//  For valid entrys, the entryState struct is filled in
//-----------------------------------------------------------------------------
void *_starterEntry_(uint idx, EntryState *entryState)
{
  FileInfo *fileRef;

  if (idx >= MaxFDs)  return NULL;

  fileRef = openFiles + idx;

  entryState->openCount = fileRef->openCount;
  entryState->unused = (entryState->openCount == 0);

  return fileRef;
}

//-----------------------------------------------------------------------------
//  Delete a xxx file
//
//  To be called ONLY by fcntl functions
//-----------------------------------------------------------------------------
int _starterUnlink_(char *path)
{
  FileInfo  *fileRef;


  KLOCK();

  fileRef = filePtr(path);
  if (!fileRef)  { KUNLOCK();  ERRNO = ENOENT;  return -1; }

//--- don't delete it if it is in use ---
  if (fileRef->openCount > 0)  { KUNLOCK();  ERRNO = EBUSY;  return -1; }

  fileRef->name[0] = 0;

  KUNLOCK();

  return 0;
}


//-----------------------------------------------------------------------------
int _starterFSGetQueues_(void *ioDev, WaitQueues far *queues)
{
//--- use NULL rather than sleepQueue if a request cannot fail one time ---
//--- and succeed later on ---
  queues->readQueue = queues->writeQueue = &sleepQueue;
  return 0;
}


//-----------------------------------------------------------------------------
//  Read bytes from an open file
//
//  Should only be called from the fcntl.c/read() function!
//
//  RETURN VALUEs:
//  - if bytes read: # of bytes read
//  - if no bytes read:
//    - if more possible in the future: -1 and ERRNO == EAGAIN
//    - if no more possible (EOF):  0
//-----------------------------------------------------------------------------
int  _starterFSRead_(void *ioDev, char *buf, int numBytes, bool dontBlock)
{
  int  i;


  KLOCK();

  if (! validFileRef(ioDev))  { KUNLOCK();  ERRNO = EBADF;  return -1; }

  if (numBytes <= 0)  { KUNLOCK();  return 0; }

//--- replace with code to do a real read ---
  for (i=0; i<numBytes; ++i)  {
    buf[i] = (char)(FileRef->offset & 0xff);
    ++FileRef->offset;
  }

  KUNLOCK();

  return numBytes;
}

//-----------------------------------------------------------------------------
//  Write bytes to an open file
//
//  Should only be called from the fcntl.c/write() function!
//
//  RETURN VALUEs:
//  - if bytes written: # of bytes written
//  - if no bytes written:
//    - if more possible in the future: -1 and ERRNO == EAGAIN
//    - if no more possible (EOF):  0
//-----------------------------------------------------------------------------
int  _starterFSWrite_(void *ioDev, char *buf, int numBytes, bool dontBlock)
{
  KLOCK();

  if (! validFileRef(ioDev))  { KUNLOCK();  ERRNO = EBADF;  return -1; }

  if (numBytes <= 0)  { KUNLOCK();  return 0; }

//--- add code to do an actual write and return a valid count ---

  KUNLOCK();

  return numBytes;
}


//-----------------------------------------------------------------------------
//  close a starterFS
//-----------------------------------------------------------------------------
int  _starterFSClose_(void *ioDev)
{
  KLOCK();

  if (! validFileRef(ioDev))  { KUNLOCK();  ERRNO = EBADF;  return -1; }

  if (FileRef->openCount > 0)  --FileRef->openCount;

  KUNLOCK();

  return 0;
}


//-----------------------------------------------------------------------------
//  Kernel-space function to reposition a file offset
//
//  To be called ONLY by fcntl functions
//-----------------------------------------------------------------------------
off_t  _starterFSLseek_(void *ioDev, off_t offset, int whence)
{
  KLOCK();

  if (! validFileRef(ioDev))  { ERRNO = EBADF;  return -1; }

  FileRef->offset = offset;

  KUNLOCK();

  return offset;
}


//-----------------------------------------------------------------------------
//  Get information about an open device file
//
//  To be called ONLY by fcntl functions
//-----------------------------------------------------------------------------
int  _starterFSFstat_(void *ioDev, struct stat *fsbuf)
{
  KLOCK();

  if (! validFileRef(ioDev))  { KUNLOCK();  ERRNO = EBADF;  return -1; }

//finish
  fsbuf->st_size  = 12345;         // total size in bytes
  fsbuf->st_atime = 1207594478L;   // time of last access
  fsbuf->st_mtime = 1207594478L;   // time of last modification
  fsbuf->st_ctime = 1207594478L;   // time of last status change

  fsbuf->st_mode = S_IFREG | 00777;

  KUNLOCK();

  return 0;
}

//-----------------------------------------------------------------------------
//  Kernel-space function to handle ioctl calls for this device's files
//
//  To be called ONLY by fcntl functions
//-----------------------------------------------------------------------------
int _starterFSioctl_(void *ioDev, int request, void *arg)
{
  int  stat;


#define CHAR_ARG   (*((char *)arg))
#define INT_ARG    (*((int *)arg))
#define UINT_ARG   (*((uint *)arg))
#define ULONG_ARG  (*((ulong *)arg))


  if (! validFileRef(ioDev))  { ERRNO = EBADF;  return -1; }

  FileRef->diagFlags = current->diagFlags;

  stat = -1;

  switch (request)  {
    case FIONREAD:  // return # of unread bytes
      if (!FileRef->rdOK)  { INT_ARG = -1;  errno = ENOLINK;  break; }
      KLOCK();  INT_ARG = FileRef->unreadBytes;  KUNLOCK();
      stat = 0;  break;

    case FIONSPACE:  // return # of free bytes in the xmit buffer
      if (!SI->wrOK)  { INT_ARG = -1;  errno = ENOLINK;  break; }
      KLOCK();  INT_ARG = FileRef->freeBytes;  KUNLOCK();
      stat = 0;  break;

    default:
      errno = EINVAL;
  }

  return stat;
}


//-----------------------------------------------------------------------------
//  Kernel-space function to get a string description of a device file
//
//  To be called ONLY by fcntl functions
//-----------------------------------------------------------------------------
char *_starterFSDescr_(void *ioDev, char *buf, int bufSize)
{
  char  *ptr;

  sprintf(buf, "%s 0x%04lX ", DevDescr, (ulong)ioDev);

  ptr = buf + strlen(buf);  strcpy(ptr, "bad bad");

//finish: limit output to bufSize

  if (validFileRef(ioDev))
    sprintf(ptr, "%s[%d] %ls",
            NodeDescr, (int)(FileRef - openFiles), FileRef->name);

  return buf;
}

//-----------------------------------------------------------------------------
