//----- file_io.c ----- 10/01/12 -----(xx/xx/xx)-----

//----- original code from SHDesigns -----

//--- modifications, additions, and reformatting by Mark Davis, NSCL ---

//-----------------------------------------------------------------------------
// NOTE:  Locking needed to avoid concurrency issues for many of the basic
//        functions (lke reading, writing, etc) is handled by the "wrapper"
//        flashFile.c module which provides the standard fcntl API
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//  Update Log (most recent first)
//
//  Entries marked with (*) indicates a change that addresses possible
//  stability/reliability issues or fixes a bug that can cause the system
//  to crash or otherwise malfunction.
//
//  2012-10-01 - Mark Davis (*)
//               - Fixes to avoid an error when the next free block happens
//                 to be block # 0.  Prior to this, loadBuffer() (used by the
//                 read and write functions) would return an error (with errno
//                 set to ENOSPC) if it allocated another block and got # 0.
//
//  2011-01-26 - Mark Davis
//               - getIPAddr() and setIPAddr() changed to readIPAddr() and
//                 writeIPAddr()
//
//  2010-12-17 - Mark Davis (*)
//               - fflush_file() now sets the FFLAG_UPDATED bit in the file's
//                 flags value if new data is written to the file.  Earlier
//                 changes prior to this one resulted in a file's date/time
//                 being updated only if writes to the file caused its size
//                 to increase.
//
//  2010-07-06 - Mark Davis (*)
//               - ffile_fsck() now checks to see if the # of sectors for the
//                 flash device is > MAX_BLK_NUM.
//               - References to MAX_BLK_NUM now replaced with references to
//                 to the actual # of blocks in the flash (which is stored
//                 in config_table->num_blocks).  Prior to this it appeared
//                 to be possible to use block #'s that exceeded the actual
//                 number of blocks in the file system.
//
//  2010-06-29 - Mark Davis
//               - Added log message for memory allocation calls
//
//  2010-06-07 - Mark Davis (*)
//               - Fixed problems caused by earlier simplification to how the
//                 sector cache (sector_buff) is handled for each open file.
//                 In particular, writes always started at the beginning of
//                 the file rather than the current positon (e.g. if the file
//                 was opened with O_APPEND or an lseek was done).
//
//  2010-03-31 - Mark Davis
//               - Removed the unnecessary critical section from write_block()
//                 and update_config()
//
//  2010-03-31 - Mark Davis
//               - fflush_file() no longer updates the date/time for the file
//                 as this is now handled in ffile_write() (because doing it
//                 in fflush_file() often sets the time to one that is later
//                 than the last change).
//
//  2010-03-30 - Mark Davis
//               - ffile_write() now modifies the date/time for the file if
//                 any bytes were processed.  Previously the date/time value
//                 was only updated when the file was closed.
//
//  2010-03-22 - Mark Davis
//               - ffile_create() now sets the date so a file that is created
//                 but not written to will have a valid date/time stamp.
//
//  2010-03-13 - Mark Davis (*)
//               - Added reset of watchdog timer before calls to WriteFlash().
//                 It appeared that multiple calls in rapid sequence was
//                 enough to cause a watchdog timer reset.
//               - The return status of a call to WriteFlash() is now checked
//                 and an error logged if there is a problem.
//               - Added either the use of the scheduler lock or a check for
//                 one in several functions where the lock should be in effect
//                 to avoid possible concurrency issues.
//
//  2009-03-16 - Mark Davis
//               - ffile_fstat() now checks for an invalid date/time
//               - expanded ipset critical area in write_block()
//               - added ipset critical area to update_config()
//
//  2009-03-12 - Mark Davis
//               - A file can no longer be deleted if it is already open
//
//  2008-04-30 - Mark Davis
//               - changes to make seek work properly and so all funcs can
//                 properly use the data in the sector buffer after EOF is
//                 reached.
//
//  2008-01-15 - Mark Davis
//               - some cleanup and commenting of code done while examining it
//                 for potential concurrency issues
//
//  2007-08-03 - Mark Davis
//               - added ffile_setName() and ffile_getName()
//
//  2006-xx-xx - Mark Davis
//               - original modifications
//
//-----------------------------------------------------------------------------

#include <common.h>

#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <utime.h>
#include <fcntl.h>

#include <stcpip.h>

#include "string.h"

#include <SHD_DLM.h>
#include <DLMiii.h>

#include <rabbit.h>
#include <dcdefs.h>

#include <sys/farprintf.h>


//#define DOPEN_FILES 4  // moved to SHD_DLM.h

//--- state of currently opened files ---
//--- The ffile "handles" are indexes in to this list ---
FILE_DATA  openFiles[DOPEN_FILES];

ulong  param_start = 0l;
ushort  sector_size;

//=============================================================================
//      ---------  !!!!! WARNING !!!!!    !!!!! WARNING !!!!!  ------------
//
//   The following table is used by both the Bootloader/Download Manager
//   and the application/OS!
//
//                !!!!! Do NOT change the structure at all !!!!!
//=============================================================================
//  This pointer is changed to a buffer allocated by init_file_data()
CONFIG_TABLE far *config_table = (CONFIG_TABLE far *)0l;



//-----------------------------------------------------------------------------
//  Initialize the flash file system
//-----------------------------------------------------------------------------
int init_file_data(void)
{
  static char file_init = 0;

  KLOCK();

  if (file_init)  {
    if (config_table->key != CONFIG_KEY)  {
      postMortem("Flash File driver data corrupted");  Exit(-666); }
    KUNLOCK();  return 1;
  }

  GetFlashDeviceEnd((FlDWord )0x0l, &sector_size);
  printf("Flash Sector Size: %u\r\n", sector_size);

  memset(openFiles, 0, sizeof(openFiles));

  if (!config_table)  {
    fprintf(DiagLog, "For flash file FAT:\r\n");
    config_table = (CONFIG_TABLE far *) xalloc(sizeof(CONFIG_TABLE));
  }

  if (!config_table)  {
    fprintf(DiagLog, "*** Unable to initialize Flash File driver ***\r\n");
    KUNLOCK();  return 0;
  }

  param_start = 0xc000l;  // where to start search from

  while (config_table AND !file_init)  {
    ReadFlash((FlByte const far *)param_start, (FlByte far *)config_table,
              sizeof(CONFIG_TABLE));
    if (config_table->key == CONFIG_KEY)  { file_init = 1;  break; }
    param_start += 0x1000;  // 4k alignment
    if (param_start >= 0x20000l)  break;  // if not found in 128k then exit
  }

  KUNLOCK();

  if (file_init)
    printf("in-flash CONFIG_TABLE found at: 0x%lX\r\n", param_start);

  return file_init;
}

//-----------------------------------------------------------------------------
//  Checks to see that the provided handle refers to a currently opened file.
//
//  Returns a pointer to the FILE_DATA structure for the open file on success,
//  or NULL if there is an error.
//
//  If the handle is valid, then the addr of the FILE_ENTRY structure for the
//  file is also returned in the file argument (if it is not NULL).
//-----------------------------------------------------------------------------
FILE_DATA *validHandle(int handle, FILE_ENTRY far **filePtr)
{
  FILE_DATA  *fileState;
  FILE_ENTRY far *file;
  unsigned  fatIdx;


  if (!init_file_data())  return NULL;

  if ((handle < 0) OR (handle >= DOPEN_FILES))  return NULL;

  fileState = openFiles + handle;

  if (!(fileState->flags & FFLAG_OPEN))  return NULL;

  fatIdx = fileState->fatIdx;

  if (fatIdx >= MAX_FILES)  return NULL;

  file = config_table->files + fatIdx;

  if (!(file->flags & FILE_ACTIVE))  return NULL;

  if (filePtr)  *filePtr = file;

  return fileState;
}


//-----------------------------------------------------------------------------
//  Return the next block in a linked list of blocks
//-----------------------------------------------------------------------------
static ushort getNextBlock(ushort curBlockNum)
{
  if (!KLOCKED())  klockErr(__func__);  //tdebug

  init_file_data();

  if (curBlockNum >= config_table->num_blocks)  return EOF_BLK;

  return config_table->nextBlock[curBlockNum];
}

//-----------------------------------------------------------------------------
//  Set the "next" block for the specified block
//-----------------------------------------------------------------------------
static void setNextBlock(ushort blockNum, ushort nxt)
{
  if (!KLOCKED())  klockErr(__func__);  //tdebug

  init_file_data();

  if (blockNum >= config_table->num_blocks)  return;
  if ((nxt >= config_table->num_blocks) AND (nxt != EOF_BLK))  nxt = EOF_BLK;

  config_table->nextBlock[blockNum] = nxt;
}

//-----------------------------------------------------------------------------
//  Return the index/# of the next unused block/sector in flash
//  Returns (uint)-1 if no unused blocks/sectors
//-----------------------------------------------------------------------------
static ushort get_freeblock(void)
{
  ushort far *blocks = config_table->nextBlock;
  ushort blockNum = 0;

  if (!KLOCKED())  klockErr(__func__);  //tdebug

  init_file_data();

  while (blockNum < config_table->num_blocks)  {
    if (*blocks == 0)  return blockNum;
    blocks++;  blockNum++;
  }

  return (uint)-1;
}

//-----------------------------------------------------------------------------
//  Read a sector from flash in to the provided buffer
//-----------------------------------------------------------------------------
static void read_block(unsigned block, char far * buff)
{
  ulong addr;

  if (!KLOCKED())  klockErr(__func__);  //tdebug

  init_file_data();

  addr = (ulong)block * sector_size;
//printf("ReadFlash(%lu, [%lu], %u)\r\n", (ulong)addr, (ulong)buff, sector_size);  //tdebug
//fflush(stdout);
  ReadFlash((FlByte const far *)(addr), (FlByte far *)buff, sector_size);
}

//-----------------------------------------------------------------------------
//  Copy an updated block of a file to flash
//-----------------------------------------------------------------------------
static void  write_block(unsigned blockNum, char far * buff)
{
  int  stat;
  ulong addr = blockNum;
  FlDWord  badAddr;

  // sector #'s are always > 0
  if (!blockNum OR (blockNum >= config_table->num_blocks))  return;

  addr *= sector_size;  // avoid "= blockNum * sector_size" as that looses bits!

  if (!KLOCKED())  klockErr(__func__);  //tdebug

  init_file_data();

  WDT_2S();

  _FlashUnlock = ~addr;

  stat = WriteFlash((FlByte const far *)addr, (FlByte far *)buff, sector_size);

  if (stat)  badAddr = FlashError();

  if (stat)
    fprintf(DiagLog, "***** Error writing to flash address: %06lX *****\r\n",
            badAddr);
}

//-----------------------------------------------------------------------------
//  copy the entire config_table to flash
//-----------------------------------------------------------------------------
static void  update_config(void)
{
  int  stat;
  FlDWord  badAddr;


  if (!KLOCKED())  klockErr(__func__);  //tdebug

  init_file_data();

  WDT_2S();

  _FlashUnlock = ~param_start;

  stat = WriteFlash((FlByte const far *)param_start,
                    (FlByte far *)config_table, sizeof(CONFIG_TABLE));

  if (stat)  badAddr = FlashError();

  if (stat)
    fprintf(DiagLog, "***** Error writing to flash address: %06lX *****\r\n",
            badAddr);
}


//-----------------------------------------------------------------------------
//  Return the number of bytes in unused sectors of flash
//-----------------------------------------------------------------------------
long int ffile_free(void)
{
  unsigned  i;
  long int  free = 0l;

  if (!init_file_data())  return FERR_NOXMEM;

  KLOCK();

  for (i=0; i<config_table->num_blocks; i++)
    if (config_table->nextBlock[i] == 0)  free += sector_size;

  KUNLOCK();

  return free;
}


//-----------------------------------------------------------------------------
//  Return the next unused file handle (an index in to the table of opened
//  files).  Returns < 0 if all handles are in use.
//-----------------------------------------------------------------------------
static int next_handle(void)
{
  if (!KLOCKED())  klockErr(__func__);  //tdebug

  init_file_data();

  for (int hdl=0; hdl<DOPEN_FILES; hdl++)
    if (! openFiles[hdl].flags) return hdl;

  return FERR_TOOMANYFILES;
}

//-----------------------------------------------------------------------------
//  Return the index for the next unused file entry in the FAT
//-----------------------------------------------------------------------------
static int next_file(void)
{
  FILE_ENTRY far *file = config_table->files;

  if (!KLOCKED())  klockErr(__func__);  //tdebug

  init_file_data();

  for (unsigned j=0; j<MAX_FILES; j++,++file)
    if ((file->flags & FILE_ACTIVE) == 0)  return j;

  return FERR_TOOMANYFILES;
}

//-----------------------------------------------------------------------------
//  Return the index in the FAT for the specified file name, or < 0 if the
//  file is not found.
//-----------------------------------------------------------------------------
int ffile_index(char * name)
{
  if (!KLOCKED())  klockErr(__func__);  //tdebug

  init_file_data();

  for (unsigned j=0; j<MAX_FILES; j++)
    if ((config_table->files[j].flags & (FILE_ACTIVE | FILE_SFLASH))
                == (FILE_ACTIVE)
        AND fstrcmp(config_table->files[j].name, name) == 0)
      return j;

  return FERR_NOTFOUND;
}

//-----------------------------------------------------------------------------
//  Allocate memory to manage reads/writes from/to the current sector for an
//  open flash file.
//-----------------------------------------------------------------------------
void allocBuf(FILE_DATA *fileState)
{
  fprintf(DiagLog, "For sector buf:\r\n");
  fileState->sector_buff = xalloc(sector_size);
}


//-----------------------------------------------------------------------------
//  If the file's R/W buffer has been altered, then flush it to flash
//-----------------------------------------------------------------------------
static void fflush_file(FILE_DATA *fileState)
{
  if (!KLOCKED())  klockErr(__func__);  //tdebug

  init_file_data();

  if (fileState->flags & FFLAG_WRITE)  {
    write_block(fileState->curBlkNum, fileState->sector_buff);
    fileState->flags &= ~FFLAG_WRITE;
    fileState->flags |= FFLAG_UPDATED;
  }
}

//-----------------------------------------------------------------------------
//  Load the sector buffer with a copy of the sector that corresponds to the
//  file's current position.  If the buffer already has data in it, and it
//  is newer than the flash, then it is flushed to flash first.
//
//  Returns the block number on success, (uint)-1 if an error
//-----------------------------------------------------------------------------
uint loadBuffer(FILE_ENTRY far *file, FILE_DATA *fileState, bool willWrite)
{
  uint  blk, prevBlk;
  ulong  sectorPos, curPos;


  blk = fileState->curBlkNum;  // 0 ==> nothing in the cache/buffer

//--- If the sector_buf currently has data in it, then flush it to flash ---
//--- if necessary ---
  if (blk)  fflush_file(fileState);

  curPos = fileState->pos;

//--- search forward from current sector, or start at the beginning? ---
  if (!blk OR (fileState->sectorPos + sector_size < curPos))  {
    blk = file->startblock;  sectorPos = 0;  }
  else
    sectorPos = fileState->sectorPos;

  prevBlk = blk;

//--- scan forward through the list of blocks assigned to the file until we ---
//--- locate the one that corresponds to the current file position ---
  while (blk AND (blk != EOF_BLK))  {
    if ((curPos >= sectorPos) AND (curPos < sectorPos + sector_size)) break;
    prevBlk = blk;
    blk = getNextBlock(blk);  sectorPos += sector_size;
  }
//NOTE:  If we reached the end of the file, then sectorPos should == the
//       current size of the file

//--- If the block we need is already part of the file, then read it in to ---
//--- the buffer ---
  if (blk AND (blk != EOF_BLK))  {
//    fprintf(DiagLog, "--- [%d] load sector %u for file %ls\r\n",   //tdebug
//           fileState - openFiles, blk, file->name);  //tdebug
    read_block(blk, fileState->sector_buff);
  }
//--- If we ran out of blocks, but the caller is about to write to the file ---
//--- and the we are currently positioned at the end of the file, then ---
//--- then allocate another block, zero it out, and append it to the file ---
  else  {
    if ((curPos != sectorPos) OR !willWrite)  { errno = EBADFD;  return 0; }
    if ((blk = get_freeblock()) == (uint)-1)  { errno = ENOSPC;  return 0; }
//    fprintf(DiagLog, "--- [%d] Append sector %u to file: %ls at pos %lu\r\n",  //tdebug
//            fileState - openFiles, blk, file->name, fileState->pos);  //tdebug
    if (sectorPos)
      setNextBlock(prevBlk, blk);
    else
      file->startblock = blk;
    setNextBlock(blk, EOF_BLK);  FBZERO(fileState->sector_buff);
    fileState->flags |= FFLAG_UPDATED;
  }

  fileState->buff_ptr = fileState->sector_buff + (curPos - sectorPos);
  fileState->curBlkNum = blk;  fileState->sectorPos = sectorPos;

  return blk;
}


//-----------------------------------------------------------------------------
//  Return a handle for the specified file.  Returns < 0 if the file cannot
//  be opened.
//-----------------------------------------------------------------------------
int ffile_open(char * name)
{
  int hdl, fatIdx;
  FILE_DATA  *fileState;  // state of currently opened file


  if (!KLOCKED())  klockErr(__func__);  //tdebug

  if (!init_file_data())  return FERR_NOXMEM;

//--- get a handle for opening a file ---
  hdl = next_handle();  if (hdl < 0)  return hdl;

//--- get the index in to the FAT for the specified file ---
  fatIdx = ffile_index(name);  if (fatIdx < 0)  return fatIdx;

  fileState = openFiles + hdl;

//--- Many (most?) flash memories require erasing an entire sector before ---
//--- writing to it, so we allocate a R/W buffer to hold a copy of the ---
//--- "current" sector (i.e. the one corresponding to the last one we read ---
//--- from flash).  The 'flags' value for each open file contains bits that ---
//--- are set if the buffer contains newer data than the flash (FFLAG_WRITE) ---
//--- and if the FAT has been modified (FFLAG_UPDATED) ---
  if (!fileState->sector_buff)  allocBuf(fileState);
  if (!fileState->sector_buff)  return FERR_NOXMEM;

//--- initialize the rest of the file handle ---
  fileState->buff_ptr = fileState->sector_buff;
  fileState->fatIdx = fatIdx;
  fileState->curBlkNum = 0;  // indicates no data in r/w buffer
  fileState->pos = 0;
  fileState->flags = FFLAG_OPEN;

  return hdl;
}

//-----------------------------------------------------------------------------
//  Close an open file, flushing any unwritten data from the write buffer
//  and updating the FAT table in flash if any info on the file has changed
//-----------------------------------------------------------------------------
void ffile_close(int handle)
{
  FILE_DATA  *fileState;

  if (!KLOCKED())  klockErr(__func__);  //tdebug

  if (!(fileState = validHandle(handle, NULL)))  { errno = EBADF;  return; }

  if (fileState->flags & FFLAG_WRITE)  fflush_file(fileState);
  if (fileState->flags & FFLAG_UPDATED)  update_config();

//  if (fileState->sector_buff)  {
//    farfree(fileState->sector_buff);  fileState->sector_buff = NULL;  }

  fileState->flags = 0;
}

//-----------------------------------------------------------------------------
//  Delete a flash file by name.
//
//  Returns 0 on success, < 0 if an error
//-----------------------------------------------------------------------------
int ffile_delete(char *name)
{
  int  index;

  if (!KLOCKED())  klockErr(__func__);  //tdebug

  init_file_data();

  if ((index = ffile_index(name)) < 0)  { ERRNO = ENOENT;  return -1; }

  if (ffile_delete_index(index) < 0)  return -1;

  return 0;
}

//-----------------------------------------------------------------------------
//  Delete the file for the specified file table index
//
//  Returns the index # on success (>= 0), or < 0 if an error;
//-----------------------------------------------------------------------------
int ffile_delete_index(int fatIdx)
{
  unsigned block;
  FILE_DATA  *fileState;  // state of currently opened file
  FILE_ENTRY far *file;   // ptr to FAT entry


  if (!KLOCKED())  klockErr(__func__);  //tdebug

  if (!init_file_data())  { ERRNO = ENOMEM;  return FERR_NOXMEM; }

  if ((fatIdx < 0) OR (fatIdx >= MAX_FILES)) { ERRNO = ENOENT;  return -1; }

//--- make sure the file is not in use ---
  for (uint ui=0; ui<DOPEN_FILES; ++ui)  {
    fileState = openFiles + ui;
    if (!fileState->flags)  continue;  // 0 means unused element in openFiles[]
    if (fileState->fatIdx == fatIdx) { ERRNO = EBUSY;  return -1; }
  }

  file = config_table->files + (uint)fatIdx;

  block = (uint)file->startblock;

  // FAT will have 0 size
  if (file->size)
    while (block AND (block != EOF_BLK))  {
      unsigned nxt = getNextBlock(block);
      setNextBlock(block, 0);  block = nxt;
    }

  fmemset(file, 0, sizeof(FILE_ENTRY));

  if (config_table->backup == fatIdx + 1)  config_table->backup = 0;
  if (config_table->primary == fatIdx + 1)  config_table->primary = 0;

  update_config();

  return fatIdx;
}

//-----------------------------------------------------------------------------
//  Create an entry in the file table for a new file, open it, and return the
//  handle for the open file
//-----------------------------------------------------------------------------
int ffile_create(char * name)
{
  int hdl, idx;
  FILE_ENTRY far  *file;  // ptr to FAT entry
  FILE_DATA  *fileState;  // state of currently opened file


//fprintf(DiagLog, "creating file: [%s]\n", name);  //tdebug

  if (!KLOCKED())  klockErr(__func__);  //tdebug

  if (!init_file_data())  return FERR_NOXMEM;

//--- get a handle for opening a file ---
  hdl = next_handle();  if (hdl < 0)  return hdl;

//fprintf(DiagLog, "Using handle: %d\n", hdl);  //tdebug

//--- make sure a file with the specified name doesn't already exist ---
  idx = ffile_index(name);  if (idx >= 0)  return FERR_FILE_EXISTS;

//--- get an unused entry in the file table ---
  idx = next_file();  if (idx < 0)  return idx;

//fprintf(DiagLog, "File Index: %d\n", idx);  //tdebug

  file = config_table->files + (uint) idx;
  fileState = openFiles + hdl;

//--- allocate a buffer for partial sector reads and writes ---
  if (!fileState->sector_buff)  allocBuf(fileState);
  if (!fileState->sector_buff)  return FERR_NOXMEM;

//--- initialize the new file table entry and the file handle ---
  fmemset(file, 0, sizeof(FILE_ENTRY));
  fstrcpy(file->name, name);
  file->flags = FILE_ACTIVE;
  file->date = time(NULL);  //readRTC(NULL);

  fileState->buff_ptr = fileState->sector_buff;
  fileState->fatIdx = idx;
  fileState->curBlkNum = 0;
  fileState->pos = 0;
  fileState->flags = FFLAG_OPEN | FFLAG_UPDATED;

  return hdl;
}

//-----------------------------------------------------------------------------
//  Create an entry in the file table for a new file and return its index
//  in the file table
//-----------------------------------------------------------------------------
int ffile_FATLink(char * name)
{
  int fatIdx;
  FILE_ENTRY far  *file;  // ptr to FAT entry


  if (!KLOCKED())  klockErr(__func__);  //tdebug

  if (!init_file_data())  return FERR_NOXMEM;

//--- make sure a file with the specified name doesn't already exist ---
  fatIdx = ffile_index(name);  if (fatIdx >= 0)  return FERR_FILE_EXISTS;

//--- get an unused entry in the file table ---
  fatIdx = next_file();  if (fatIdx < 0)  return fatIdx;

  file = config_table->files + (uint) fatIdx;

//--- initialize the entry ---
  fmemset(file, 0, sizeof(FILE_ENTRY));
  fstrcpy(file->name, name);
  file->flags = FILE_ACTIVE | FILE_SFLASH;

//--- save the update FAT to flash ---
  update_config();

  return fatIdx;
}

//-----------------------------------------------------------------------------
//  Return the current size of a currently open file
//
//  Returns >= 0 on success, < 0 if an error
//-----------------------------------------------------------------------------
long ffile_size(int handle)
{
  FILE_ENTRY far *file;  // FAT entry for the file

  if (!KLOCKED())  klockErr(__func__);  //tdebug

  if (!validHandle(handle, &file))  { errno = EBADF;  return -1; }

  return file->size;
}


//-----------------------------------------------------------------------------
//  Return the size of a file given its name
//
//  Returns >= 0 on success, < 0 if an error
//-----------------------------------------------------------------------------
long ffile_sizen(char * name)
{
  long index;

  if (!KLOCKED())  klockErr(__func__);  //tdebug

  init_file_data();

  index = ffile_index(name);

  if (index < 0)  return -1;

  return config_table->files[index].size;
}

//-----------------------------------------------------------------------------
//  Return a copy of the name of a file given its FAT index.
//
//  Returns the string "<invalid>" if the FAT index is not valid.
//-----------------------------------------------------------------------------
void ffile_name(unsigned int fatIdx, char *name)
{
  char far *ptr;
  FILE_ENTRY far *file = config_table->files + fatIdx;


  if (!KLOCKED())  klockErr(__func__);  //tdebug

  init_file_data();

  if (fatIdx >= MAX_FILES)  ptr = "<invalid>";  else
  if (! (file->flags & FILE_ACTIVE))  ptr = "<invalid>";  else
  ptr = file->name;

  fstrcpy(name, ptr);
}

//-----------------------------------------------------------------------------
//  Return current position for the specified open file
//-----------------------------------------------------------------------------
long ffile_tell(int handle)
{
  FILE_DATA *fileState;

  if (!KLOCKED())  klockErr(__func__);  //tdebug

  if (!(fileState = validHandle(handle, NULL)))  { errno = EBADF;  return -1; }

  return fileState->pos;
}

//-----------------------------------------------------------------------------
//  Change current file position relative to start, cur pos, or end of file
//  Returns 0 on success or < 0 if an error
//-----------------------------------------------------------------------------
int ffile_seek(int handle, long pos, int mode)
{
  ulong  newPos, sectorPos;
  FILE_DATA *fileState;  // state of currently opened file
  FILE_ENTRY far *file;  // ptr to FAT entry


  if (!KLOCKED())  klockErr(__func__);  //tdebug

  if (!(fileState = validHandle(handle, &file)))  {
    errno = EBADF;  return FERR_SEEK; }

//fprintf(DiagLog, "ffile_seek_(%d, %ld, %d), cur pos: %lu)\r\n",  //tdebug
//        handle, pos, mode, fileState->pos);

//--- Change the position value so it is an offset from the start ---
  switch (mode)  {
    case FSEEK_SET:  // pos value is already an offset from the start
      if (pos < 0)  return FERR_SEEK;
      newPos = (ulong) pos;  break;

    case FSEEK_END:  // passed value should be offset from end (i.e. < 0)
      if (pos > 0)  return FERR_SEEK;
      pos = -pos;
      newPos = file->size - (ulong)pos;  break;

    case FSEEK_CUR:  // passed value is an offset from current position
      if (!pos)  return 0; // no move, so just exit
      newPos = pos + fileState->pos;  break;

    default:
      errno = EINVAL;  return FERR_SEEK;
  }

//--- Is the resulting position past the end of the file? ---
  if (newPos > file->size)  return FERR_SEEK;

//--- If the R/W buffer doesn't have any data in it, then just update the ---
//--- position value ---
  if (fileState->curBlkNum == 0)  { fileState->pos = newPos;  return 0; }


//--- If the new position is outside the current buffer's contents, then ---
//--- flush the buffer to flash (if needed) and "clear" the buffer (i.e. ---
//--- we set its blockNum to 0 to indicate it does not contain a copy of ---
//--- the sector corresponding to the current file position) ---

  sectorPos = fileState->pos - (fileState->buff_ptr - fileState->sector_buff);
  if ((newPos < sectorPos) OR (newPos >= (sectorPos + sector_size)))  {
    fflush_file(fileState);
    fileState->curBlkNum = 0;  fileState->pos = newPos;  return 0;
  }

//--- The new position is within the buffer, so just update buff_ptr ---
  if (fileState->pos >= newPos)
    fileState->buff_ptr -= (fileState->pos - newPos);
  else
    fileState->buff_ptr += (newPos - fileState->pos);

  fileState->pos = newPos;

  return 0;
}

//-----------------------------------------------------------------------------
//  Read the specified # of bytes from an open file in to the provided buffer
//  Returns the # of bytes read.
//-----------------------------------------------------------------------------
int ffile_read(int handle, char far * buff, int bytes)
{
  unsigned  numleft, read;
  FILE_DATA  *fileState;  // state of currently opened file
  FILE_ENTRY  far *file;  // ptr to FAT entry


  if (!KLOCKED())  klockErr(__func__);  //tdebug

  if (!(fileState = validHandle(handle, &file)))  { errno = EBADF;  return -1; }

  if (bytes < 0)  { errno = EINVAL;  return -1; }

  if (bytes == 0)  return 0;

  read = 0;

//printf("ffile_read: pos: %lu  size: %lu  offset:%ld\r\n",
//       fileState->pos, file->size, (long)(fileState->buff_ptr - fileState->sector_buff));

  if (fileState->pos + bytes > file->size)
    bytes = (uint)(file->size - fileState->pos);
  if (!bytes)  return read;

//--- If the R/W buffer has no data in it, then read the 1st block of ---
//--- the file from the flash ---
  if (fileState->curBlkNum == 0)   // nothing in the R/W buffer yet
    if (loadBuffer(file, fileState, false) == (uint)-1)  return -1;

  while (bytes)  {
  //--- If we used up what was in the buffer, read the next sector ---
    if (fileState->buff_ptr >= fileState->sector_buff + sector_size)  {
      fflush_file(fileState);  // update flash if cached data newer than flash
      if (loadBuffer(file, fileState, false) == (uint)-1)  return read;
    }
  //--- how many of the unread bytes are in the current cache/buffer? ---
    numleft = sector_size - ((unsigned)fileState->pos & (sector_size-1));
    if (numleft > bytes)  numleft = bytes;
  //--- copy the bytes from the cache and update state of the open file ---
    fmemcpy(buff, fileState->buff_ptr, numleft);
    fileState->buff_ptr += numleft;  buff += numleft ;  //PSM
    bytes -= numleft;  fileState->pos += numleft;  read += numleft;
  }

  return read;
}

//-----------------------------------------------------------------------------
//  Write bytes to a file and return the # of bytes written
//-----------------------------------------------------------------------------
int ffile_write(int handle, char far *buff, int bytes)
{
  unsigned  numleft, wrote = 0;
  FILE_DATA  *fileState; // state of currently opened file
  FILE_ENTRY far *file;  // ptr to FAT entry


  if (!KLOCKED())  klockErr(__func__);  //tdebug

  if (!(fileState = validHandle(handle, &file)))  { errno = EBADF;  return -1; }

  if (bytes < 0)  { errno = EINVAL;  return -1; }

  if (bytes == 0)  return 0;

  file = config_table->files + fileState->fatIdx;

  if (!fileState->curBlkNum)   // nothing in R/W buffer at present
    if (loadBuffer(file, fileState, true) == (uint)-1)  return -1;

  while (bytes)  {
    // # of bytes in R/W buffer between the cur pos and the end of the buffer
    numleft = sector_size - (uint)(fileState->buff_ptr - fileState->sector_buff);

  //--- If some of the bytes belong in the current sector, then copy them ---
  //--- to the buffer ---
    if (numleft)  {
      if (numleft > bytes)  numleft = bytes;
      fmemcpy(fileState->buff_ptr, buff, numleft);
      fileState->buff_ptr += numleft;
      bytes -= numleft;  buff += numleft;
      fileState->pos += numleft;
      if (fileState->pos > file->size)  {
        file->size = fileState->pos;  fileState->flags |= FFLAG_UPDATED; }
      wrote += numleft;  file->date = time(NULL);
      fileState->flags |= FFLAG_WRITE;
    }

  //--- If not done writing, flush the current cache to flash and read the ---
  //--- next sector ---
    if (bytes)  {
      fflush_file(fileState);  // flush cache to flash if newer data than flash
      if (loadBuffer(file, fileState, true) == (uint)-1)
        return (wrote ? wrote : -1);
    }
  }

  return wrote;
}

//-----------------------------------------------------------------------------
//  Set the primary boot image to the file specified by its FAT index
//-----------------------------------------------------------------------------
int ffile_setPrimary(int index)
{
  if (!KLOCKED())  klockErr(__func__);  //tdebug

  if (!init_file_data())  return FERR_NOXMEM;

  config_table->primary = index + 1;
  if (config_table->backup == index + 1)  config_table->backup = 0;

  update_config();

  return 0;
}

//-----------------------------------------------------------------------------
//  Set the backup boot image to the file specified by its FAT index
//-----------------------------------------------------------------------------
int ffile_setBackup(int index)
{
  if (!KLOCKED())  klockErr(__func__);  //tdebug

  if (!init_file_data())  return FERR_NOXMEM;

  config_table->backup = index +1;
  if (config_table->primary == index +1)  config_table->primary = 0;

  update_config();

  return 0;
}

//-----------------------------------------------------------------------------
//  Return the FAT index of the primary boot image file (or < 0 if no primary
//  boot image file).
//-----------------------------------------------------------------------------
int ffile_getPrimary(void)
{
  int index;

  if (!KLOCKED())  klockErr(__func__);  //tdebug

  if (!init_file_data())  return FERR_NOXMEM;

  index = config_table->primary;

  return index - 1;
}

//-----------------------------------------------------------------------------
//  Return the FAT index of the backup boot image file (or < 0 if no backup
//  boot image file).
//-----------------------------------------------------------------------------
int ffile_getBackup(void)
{
  int index;

  if (!KLOCKED())  klockErr(__func__);  //tdebug

  if (!init_file_data())  return FERR_NOXMEM;

  index = config_table->backup;

  return index - 1;
}

//-----------------------------------------------------------------------------
//  Get the index of the binary image marked as the INIT program
//-----------------------------------------------------------------------------
int ffile_getInit(void)
{
  int index;

  if (!KLOCKED())  klockErr(__func__);  //tdebug

  if (!init_file_data())  return FERR_NOXMEM;

  index = config_table->init;  return index - 1;
}

//-----------------------------------------------------------------------------
//  Set the index of the binary image marked as the INIT program
//-----------------------------------------------------------------------------
int ffile_setInit(int index)
{
  if (!KLOCKED())  klockErr(__func__);  //tdebug

  if (!init_file_data())  return FERR_NOXMEM;

  config_table->init = index + 1;  update_config();

  return 0;
}


//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
void ffile_rename(unsigned int index, char * name)
{
  if (!KLOCKED())  klockErr(__func__);  //tdebug

  fstrcpy(config_table->files[index].name, name);  update_config();
}

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
int ffile_handle2Index(int handle)
{
  FILE_DATA * fileState = openFiles + handle;  // state of currently opened file

  if (!KLOCKED())  klockErr(__func__);  //tdebug

  return fileState->fatIdx;
}


//=============================================================================
//============================ NSCL additions =================================
//=============================================================================

#include <common.h>

//-----------------------------------------------------------------------------
//  Get info for file given its index in the flash file table.
//  Returns far ptr to file data if index is for a valid file, NULL if an error
//-----------------------------------------------------------------------------
FILE_ENTRY far *ffile_info(unsigned idx)
{
  FILE_ENTRY  far *file;  // ptr to FAT entry


  if (!KLOCKED())  klockErr(__func__);  //tdebug

  if (!init_file_data())  { ERRNO = ENOMEM;  return NULL; }

  if (idx >= MAX_FILES)  return NULL;

  file = config_table->files + idx;

  if ((file->flags & (FILE_ACTIVE | FILE_SFLASH)) != (FILE_ACTIVE))
    return NULL;

  return file;
}

//-----------------------------------------------------------------------------
//  Return info in standard fstat stat struct for an open file
//-----------------------------------------------------------------------------
int  ffile_fstat(int handle, struct stat *fsbuf)
{
  ulong  testTime;
  FILE_ENTRY  far *file;  // ptr to FAT entry


  if (!KLOCKED())  klockErr(__func__);  //tdebug

  if (!init_file_data())  { ERRNO = ENOMEM;  return -1; }

  if ((uint)handle >= MAX_FILES)  { ERRNO = ENOENT;  return -1; }

  file = config_table->files + openFiles[handle].fatIdx;

  testTime = file->date & 0xFFFFFF00L;
  if ((testTime == 0) OR (testTime == 0xFFFFFF00L))  {
    ERRNO = EBADFD;  return -1; }

  fsbuf->st_size  = file->size;  // total size in bytes
  fsbuf->st_atime = file->date;  // time of last access
  fsbuf->st_mtime = file->date;  // time of last modification
  fsbuf->st_ctime = file->date;  // time of last status change

  fsbuf->st_mode = S_IFREG | 00777;

  return 0;
}

//-----------------------------------------------------------------------------
//  Return the number of bytes the cur pos is from the end of the file
//-----------------------------------------------------------------------------
long ffile_unread(int handle)
{
  FILE_DATA  *fileState;   // state of currently opened file
  FILE_ENTRY  far *entry;  // ptr to FAT entry


  if (!KLOCKED())  klockErr(__func__);  //tdebug

  fileState = openFiles + handle;
  entry = config_table->files + fileState->fatIdx;

  return (entry->size - fileState->pos);
}


//-----------------------------------------------------------------------------
//  Return the IPv4 the SBC is configured for in host byte order
//-----------------------------------------------------------------------------
ulong readIPAddr(void)
{
  if (!init_file_data())  { ERRNO = ENOMEM;  return -1; }

  return ntohl(config_table->IPAddress);
}


//-----------------------------------------------------------------------------
//  Change the IPv4 stored in the configuration data.  The address must be in
//  host byte order.
//-----------------------------------------------------------------------------
int  writeIPAddr(ulong ipAddr)
{
  if (!init_file_data())  { ERRNO = ENOMEM;  return -1; }

  config_table->IPAddress = htonl(ipAddr);
  update_config();

  return 0;
}

//-----------------------------------------------------------------------------
//  Return the name of the primary boot image file (or NULL 0 if no primary
//  boot image file).
//-----------------------------------------------------------------------------
char *ffile_getPrimaryName(char *name, int nameLen)
{
  int index;

  if (!init_file_data())  { ERRNO = ENOMEM;  return NULL; }

  index = config_table->primary;
  if (index < 1)  return NULL;

  fstrcpy(name, config_table->files[(uint)index-1].name);

  return name;
}

//-----------------------------------------------------------------------------
//  Return the name of the backup boot image file (or NULL 0 if no backup
//  boot image file).
//-----------------------------------------------------------------------------
char *ffile_getBackupName(char *name, int nameLen)
{
  int index;

  if (!init_file_data())  { ERRNO = ENOMEM;  return NULL; }

  index = config_table->backup;
  if (index < 1)  return NULL;

  fstrcpy(name, config_table->files[(uint)index-1].name);

  return name;
}

//-----------------------------------------------------------------------------
//  change the name of a file
//-----------------------------------------------------------------------------
int ffile_rename_(const char *oldpath, const char *newpath)
{
  int index;

//printf("rename(%.16s, %.16s)\r\n", oldpath, newpath);  //tdebug

  if (!init_file_data())  { ERRNO = ENOMEM;  return -1; }

  KLOCK();

  index = ffile_index((char *)oldpath);
  if (index < 0)  { KUNLOCK();  ERRNO = ENOENT;  return -1; }

//printf("  delete(%.16s)\r\n", newpath);  //tdebug
  ffile_delete((char *)newpath);

//fprintf(DiagLog, "  change name of files[%d](%.16s) to %.16s\r\n",  //tdebug
//       index, config_table->files[(uint)index].name, newpath);  //tdebug
  fstrcpy(config_table->files[(uint)index].name, newpath);

  update_config();

  KUNLOCK();  return 0;
}


//-----------------------------------------------------------------------------
//  Set the primary boot image to the file specified by its name
//-----------------------------------------------------------------------------
int ffile_makePrimary(const char *fileName)
{
  int  index;


  if (!init_file_data())  { ERRNO = ENOMEM;  return -1; }

  KLOCK();

  index = ffile_index((char *)fileName);
  if (index < 0)  { KUNLOCK();  ERRNO = ENOENT;  return -1; }

  ffile_setPrimary(index);

  KUNLOCK();  return 0;
}

//-----------------------------------------------------------------------------
//  Set the backup boot image to the file specified by its name
//-----------------------------------------------------------------------------
int ffile_makeBackup(const char *fileName)
{
  int  index;

  if (!init_file_data())  { ERRNO = ENOMEM;  return -1; }

  KLOCK();

  index = ffile_index((char *)fileName);
  if (index < 0)  { KUNLOCK();  ERRNO = ENOENT;  return -1; }

  ffile_setBackup(index);

  KUNLOCK();  return 0;
}

//-----------------------------------------------------------------------------
//  Set date/time stamps on a file
//-----------------------------------------------------------------------------
int  ffile_utime(const char *fileName, const struct utimbuf *buf)
{
  int  index;
  FILE_ENTRY  far *file;  // ptr to FAT entry


  if (!init_file_data())  { ERRNO = ENOMEM;  return -1; }

  KLOCK();

  index = ffile_index((char *)fileName);
  if (index < 0)  { KUNLOCK();  ERRNO = ENOENT;  return -1; }

  file = config_table->files + (uint)index;

  file->date = (ulong) buf->modtime;

  update_config();

  KUNLOCK();  return 0;
}


//-----------------------------------------------------------------------------
//  set the board name
//-----------------------------------------------------------------------------
int  ffile_setName(char *newName)
{
  int  len;

  if ((len = strlen(newName)) >= BRD_NAME_LEN)  return -1;

  KLOCK();

  fmemcpy(config_table->board_name, newName, len);
  config_table->board_name[(uint)len] = 0;

  update_config();

  KUNLOCK();  return 0;
}


//-----------------------------------------------------------------------------
//  get the board name
//-----------------------------------------------------------------------------
char *ffile_getName(char *name, int maxLen)
{
  int  len;

  KLOCK();

  if ((len = (uint)fstrlen(config_table->board_name)) >= maxLen)  len = maxLen-1;

  fmemcpy(name, config_table->board_name, len);  name[len] = 0;

  KUNLOCK();  return name;
}


//-----------------------------------------------------------------------------
//  check the integrity of the file system
//
//  Note that if the file system is altered while this is running the results
//  may be incorrect.
//
//  Returns the # of errors found.
//
//  The repair parameter is currently ignored...
//-----------------------------------------------------------------------------
int ffile_fsck(bool repair)
{
  int  numErrs;
  uint  ui, numBlks, ttlBlks, usedBlks, freeBlks, rsvdBlks;
  ulong  spaceUsed, usedBytes, freeBytes, rsvdBytes;
  ushort  blkNum, nextBlk, sectorSize;
  uchar  mask;
  uchar *chkdFlag;
  FILE_ENTRY far *file;  // ptr to FAT entry
  uchar chkdBlocks[MAX_BLK_NUM/8];


  BZERO(chkdBlocks);  numErrs = usedBlks = freeBlks = rsvdBlks = 0;

  ttlBlks = config_table->num_blocks;  sectorSize = config_table->sector_size;

  printf("\r\n%u bytes/sector,  %u sectors,  Total %lu bytes\r\n",
         config_table->sector_size, config_table->num_blocks,
         (ulong) ttlBlks * sectorSize );

  if (config_table->num_blocks > MAX_BLK_NUM)  {
    printf("# sectors > MAX_BLK_NUM (%u)\r\n", MAX_BLK_NUM);  ++numErrs; }


//--- verify each file entry by following the linked list of blocks ---
  for (ui=0; ui<MAX_FILES; ++ui)  {
    file = config_table->files + ui;
    printf("\r\nentry %2u: ", ui);

    if ((file->flags & FILE_ACTIVE) == 0)  {
      printf("<unused>\r\n");  continue; }

    printf("%ls\r\n", file->name);
    spaceUsed = 0;  blkNum = file->startblock;  numBlks = 0;

    for (;;)  {
      if (blkNum == EOF_BLK)  break;

      if (!blkNum OR (blkNum >= config_table->num_blocks))  {
        printf("  Invalid sector # %u\r\n", blkNum);  ++numErrs; }

      ++numBlks;

      chkdFlag = chkdBlocks + (blkNum >> 3);  mask = 1 << (blkNum & 7);
      if (*chkdFlag & mask)  {
        printf("  Cross-linked sector # %u\r\n", blkNum);  ++numErrs; }

      *chkdFlag |= mask;  spaceUsed += sectorSize;
      blkNum = config_table->nextBlock[blkNum];
    }
    printf("  %3d sectors:  Size: %6lu   on disk: %6lu\r\n",
           numBlks, file->size, spaceUsed);

    usedBlks += numBlks;

    if ((spaceUsed >= file->size) AND ((spaceUsed - file->size) <= sectorSize))
      continue;

    printf("  Size error\r\n");  ++numErrs;
  }
  printf("\r\n");


//--- check all the unused blocks ---
  for (ui=0; ui<ttlBlks; ++ui)  {
    chkdFlag = chkdBlocks + (ui >> 3);  mask = 1 << (ui & 7);
    if (*chkdFlag & mask)  continue;
    nextBlk = config_table->nextBlock[ui];
    if (!nextBlk) { ++freeBlks;  continue; }
    if ((nextBlk & 0xfffe) == 0xfffe)  { ++rsvdBlks;  continue; }
    printf("Unused sector # %4u linked to sector # %4u\r\n", ui, nextBlk);
    ++numErrs;
  }


  usedBytes = (ulong) usedBlks * sectorSize;
  freeBytes = (ulong) freeBlks * sectorSize;
  rsvdBytes = (ulong) rsvdBlks * sectorSize;

  printf(" used: %6lu bytes (%4u blocks)\r\n"
         " rsvd: %6lu bytes (%4u blocks)\r\n"
         " free: %6lu bytes (%4u blocks)\r\n",
          usedBytes, usedBlks, rsvdBytes, rsvdBlks, freeBytes, freeBlks );

  printf("\r\n %d errors\r\n\n", numErrs);

  return numErrs;
}


//-----------------------------------------------------------------------------

