//--- common.h --- 12/05/12 ---(05/05/05)---

//-----------------------------------------------------------------------------
//  Misc functions and definitions used by multiple modules.
//-----------------------------------------------------------------------------
//  UPDATE LOG (most recent first)
//
//
//  2012-12-05 - Mark Davis
//               - Removed Force/Release/Dig/In/Out() functions
//               - Moved the following from fcntl to common:
//                    vt100 escape sequence definitions
//                    DispStr() and DispStrAttr()
//                    read_() and write_()
//                    redirectStdio() and logStdout()
//
//  2012-09-28 - Mark Davis
//               - Added #include <math.h> and changed FABS() macro to a call
//                 to fabs()
//
//  2012-06-11 - Mark Davis
//               - Added IncU32() macro.
//
//  2012-03-15 - Mark Davis
//               - Added DEFAULT_TZ (sets default value for timezone)
//
//  2012-03-12 - Mark Davis
//               - Added GetSeconds_() and SecSince_() for faster handling
//                 of time intervals are known to be < 64K (65536) seconds.
//
//  2012-02-13 - Mark Davis
//               - Added GetMS_() and msSince_() for faster handling of time
//                 intervals that are known to be < 64K (65536) ms.
//
//  2011-03-29 - Mark Davis
//               - showNetState() now takes the interface ID as an argument
//
//  2010-07-22 - Mark Davis
//               - Added the VALID() macro
//
//  2010-07-13 - Mark Davis
//               - Commented out the fmult() function and replaced it with a
//                 macro that just does a regular multiply.
//                 fmult() is no longer necessary because I created a
//                 modified version of fmult() that can be substituted for the
//                 Softools internal __fmlt() function so ALL floating point
//                 multiples will use the faster (but less accurate) version.
//
//  2010-06-15 - Mark Davis (*)
//               - renamed fWriteBytes() to fwrite_()
//               - read_() and write_() now check for a NULL buffer pointer
//               - added skipBytes() to use instead of calling new read_()
//                 with a NULL buffer pointer.
//
//  2010-03-08 - Mark Davis
//               - Added fGetShort(), fGetLong(), fAddShort(), fAddLong()
//
//  2010-03-03 - Mark Davis
//               - Added sendBCastPkt() (moved it here from configd.c)
//
//  2010-02-16 - Mark Davis
//               - Added memCRC16()
//
//  2010-02-15 - Mark Davis
//               - Moved readBufCh() from the "busybox" shared code in to this
//                 module
//
//-----------------------------------------------------------------------------

#ifndef  _CommonH_
#define  _CommonH_


#define  __inline__


#include <time.h>

//----- must be compatible with ZWorld's defs -----
typedef unsigned short  word;
typedef unsigned char   byte;
typedef unsigned long   longword;


typedef char            bool;

typedef unsigned char   uchar;
typedef signed char     schar;


#define  _512K   0x80000UL
#define  _256K   0x40000UL
#define  _128K   0x20000UL
#define   _64K   0x10000UL


#include <sys/types.h>

#include <string.h>

#include <stddef.h>

#include <math.h>


#define  DEFAULT_TZ  -4


#ifdef RABBIT
  #define DiagLog  (&diagLog)
  #define  GetMS()  MS_TIMER
  #define  GetMS_()  ( (uint) _TimerMilliSeconds[1] )
  #define  GetSeconds()  SEC_TIMER
  #define  GetSeconds_()  ((uint)_TimerSeconds)

      //  This is the fastest way to incrment a 32-bit value.
      //  Since the compiler doesn't always have the address
      //  and it might need the value again, it can't generate this.
  #define IncU32(u32Ptr)  \
      _asm( _HL = (unsigned) (u32Ptr),  \
                  "inc (hl)", "jr nz,..x",  \
        "inc hl", "inc (hl)", "jr nz,..x",  \
        "inc hl", "inc (hl)", "jr nz,..x",  \
        "inc hl", "inc (hl)", "..x:" );
#else
  #define DiagLog  (stderr)
  #define near
  #define far
  #define _task
  #define  ERRNO  errno
  #define sleepMS(ms)  sleep((ms+999UL)/1000)
  #define fmemcpy  memcpy
  #define  socklen_t  int
  ulong  GetMS(void);
  uint  GetMS_(void);
  ulong  GetSeconds(void);
  #define  fmemset(ptr, val, count)  memset(ptr, val, count)
#endif




//--- some macros to make pseudo-class definitions more C++ like ---
#define  class_    typedef struct
#define  virtual_
#define  public_
#define  protected_
#define  private_


#ifndef AND
  #define  AND  &&
  #define  OR   ||
  #define  BOOL(expr)  ((expr) ? 1 : 0)
#endif

#ifndef true
  #define TRUE            1
  #define true            TRUE
  #define FALSE           0
  #define false           FALSE
#endif

#ifndef NULL
#  define NULL  ((void *)(0))
#endif

#define  VFP(ptr)   ( (void far *)((char far *)(ptr)) )
//#define  VFPV(var)  ( (void far *) ((ulong)(&(var))) )  // req by older compiler
#define  VFPV(var)  ( (void far *) ((char far *)(&(var))) )

#define  VFPN(ptr,typ)   ( (void far *)((typ far *)(ptr)) )



int redirectStdio(void);
int logStdout(void);


int ValidatePtr(void *structPtr, void *copyOfPtr,
                const char *typeName, const char *context);

// used to validate ptrs to structures that contain a copy of their own addr
#define VALID(typ, ptr)  (((typ *)(ptr))->myAddr == (typ *)(ptr))

#define INVALID(typ, ptr)  \
    ValidatePtr(ptr, ((typ *)(ptr))->myAddr, #typ, __func__)



unsigned long clockTicks(void);  // read clock ticks (32768 ticks/sec)


#define  ONE_DAY  86400L   // # seconds in one day
#define  ONE_HOUR  3600    // # seconds in one hour

// diff in seconds between NTP UTC value (seconds from 1900-01-01 00:00:00)
// and ZWorld clock value (seconds from 1980-01-01 00:00:00)
#define  UTC_OFFSET  2524521600UL


char *dateTimeMS(char *buf, const time_t dateTime, uint ms);


#ifndef LINUX
  char *asctime_r(const struct tm *dts, char *str);
  char *ctime_r(const time_t *utcSecs, char *str);
  struct tm *gmtime_r(const time_t *secs, struct tm *dts);
  struct tm *localtime_r(const time_t *utcSecs, struct tm *dts);
#endif

#define ATOI(ptr)   strtoi((ptr), NULL, 0)
#define ATOU(ptr)   strtoui((ptr), NULL, 0)
#define ATOL(ptr)   strtol((ptr), NULL, 0)
#define ATOUL(ptr)  strtoul((ptr), NULL, 0)

#define StrEq(s1, s2)  (strcmp((s1), (s2)) == 0)
#define StriEq(s1, s2)  (stricmp((s1), (s2)) == 0)


#define ULDIF(ul1, ul2)  ( (ul1) > (ul2) ? (ul1)-(ul2) : (ul2)-(ul1) )
//#define FABS(ff)  (((ff) < 0.0) ? (-(ff)) : (ff))
#define FABS(ff)  fabs(ff)

#define SIGN(ff)  ((int) (((ff) < 0.0) ? -1 : 1))

#define  BZERO(l)  memset(l, 0, sizeof(l))
#define  FBZERO(l)  fmemset(l, 0, sizeof(l))


#define SetBits(target, bitMask)    ( target |= (bitMask) )
#define ClearBits(target, bitMask)  ( target &= (~(bitMask)) )


#define HexVal(ch)   ( ((ch) >= 'A') ? (((ch) - 55) & 0x0f) : ((ch) & 0x0f) )
#define NibToHex(n) ( (((n)&0x0f) > 9) ? (((n)&0x0f) + 55) : (((n)&0x0f) | 0x30) )
#define ValidHex(ch)  ( (((ch) >= '0') && ((ch) <= '9')) \
                     || (((ch) >= 'A') && ((ch) <= 'F')) )


char *unused(void *arrayOf, int elemSize, int numElements, int inUseOffset);

//--- macro to return ptr to 1st unused element in an array of a struct ---
//--- (assumes an inUse element and a Max<name> macro) ---
#define  New_(type)  ( (type *)  \
         unused(type ## s, sizeof(type), Max ## type ## s, \
                (int)((char *)(&type ## s[0].inUse) - (char *)(type ## s)) ) )

//--- macro to mark an array element not in use ---
#define  Free_(ptr, type)  { ((type *)ptr)->inUse = false;  }


#define  SetBit(buf,i)    ( (buf)[(i)>>3] |= (1 << ((i) & 7)) )
#define  ClearBit(buf,i)  ( (buf)[(i)>>3] &= ~(1 << ((i) & 7)) )
#define  GetBit(buf,i)   ( ((buf)[(i)>>3] & (1 << ((i) & 7))) ? 1 : 0 )


void  addShort(uchar **ptr, short val);
void  addLong(uchar **ptr, long val);

void  fAddShort(uchar far **ptr, short val);
void  fAddLong(uchar far **ptr, long val);


short getShort(uchar **ptr);
long  getLong(uchar **ptr);

short fGetShort(uchar far **ptr);
long  fGetLong(uchar far **ptr);

short GetShort_(uchar *buf, int idx);


void  addShortParam(uchar **ptr, short id, short val);
void  addLongParam(uchar **ptr, short id, long val);
void  addStrParam(uchar **ptr, short id, char *buf, short len);


#define UI_Interval(from, to)  ( (uint) ((uint)(to) - (uint)(from)) )
#define UL_Interval(from, to)  ( (ulong) ((ulong)(to) - (ulong)(from)) )

#define msInterval(from, to)  UL_Interval(from, to)

#define msSince(sinceTime)   UL_Interval(sinceTime, GetMS())
#define msSince_(sinceTime)  UI_Interval(sinceTime, GetMS_())

#define SecSince_(secStart)  UI_Interval(secStart, GetSeconds_())

#define  secSince(msTime)  (msSince(msTime) / 1000L)
#define  minSince(msTime)  (msSince(msTime) / 60000L)


char *mactoa(char *str, uchar far *macAddr);

/*
char *xstrcpy(char *dest, ulong xstrTable, int idx);


ulong ipv4_ntohl(uchar *ipAddr);
void ipv4_htonl(ulong ipAddrLong, uchar *ipAddrStr);

*/
uint fastCRC16(uchar *ucpBuf, int nSize, ushort accum);

uint memCRC16(ulong startAddr, ulong size);


float RelativeDif(float val, float def);

char *FmtStrCmd(char *cmdBuf, char *dispBuf, short dispBufSize);
char* ReverseStr(char *buf, char count);

char *padStr(char *str, int toLen);

char *strstri(char *inStr, char *lookfor);

int parseArgs(char *buf, int maxArgs, char **argv, int *argc);

uchar far *fstrncat(uchar far *dst, uchar far *src, size_t size);

long DecimalRep(long num);
long HexToNum(char *hexStr);
char *Engineering(float num, int SigDig, char *str);
char *fmtFloat(float num, int ttlLen, int decDigs, char *str);


;;float near fmult(float f1, float f2);
#define  fmult(f1, f2)  ( f1 * f2 )


char *fmtDHMS(char *str, ulong count);

void stdHeader(void);
void showNetState(int iface);


int bytesAvail(fd_t fd);


int CheckETX(void);


time_t  compileTime(char *dateStr, char *timeStr);

int  skipBytes(fd_t fd, int numBytes, ulong timeOut);

int fwrite_(fd_t fd, char far *from, int numBytes,
            char *buf, uint bufSize, ulong inactiveTimeout);

typedef struct {
  char *buf;
  fd_t  fd;
  uint  bufSize;
  uint  nextOut;
  uint  count;
} BufInfo;

int  readBufCh(BufInfo *bufInfo);


int write_(fd_t fd, void *buf, int numBytes, long maxIdleMS);
int read_(fd_t fd, void *buf, int numBytes, long maxIdleMS);


int copyFileByName(char *fromName, char *toName, bool showProgress,
                   char *buf, int bufSize, char **errContext);


typedef int (*ReadFunc)(void *conext, char *buf, int count);
typedef int (*WriteFunc)(void *context, char *buf, int count);

ulong copyBytes(ReadFunc readFunc, void *readContext,
                WriteFunc writeFunc, void *writeContext,
                char *buf, int bufSize, ulong copyCount, bool showProgress);


int sendBCastPkt(int *sockfd, int bcPort, ulong bcAddr,
                 char far *pktBuf, int pktLen);


int strListIdx(char **listPtr, char *testStr);


char *LongToBinMS1st(long num, char *buf, int numBits);
char *LongToBinLS1st(long num, char *buf, int numBits);


//=============================================================================
//  Definitions to simplify code for managing application EVENT msgs

#ifndef MaxMsgDataSize
  #define  MaxMsgDataSize  16
#endif

//--- structure common to start of all Event messages ----
typedef struct  {
  uint    dataLen;     // length of used portion of data[]
  void   *eventFunc;   // func to call to process event
  int     chanNum;     // for which channel?
  uchar   data[MaxMsgDataSize];  // override if needed in application code
}  EventMsg;

#define  MSG_HDR_SIZE  ( offsetof(EventMsg, data) )

typedef void (*EventHandler)(EventMsg *event);


int  putMsg(fd_t eventsfd, EventHandler eventFunc, int chan,
            EventMsg *eventMsg, void *data, int dataLen);

void processMsg(fd_t eventsfd, EventMsg *event, int bufSize);


#define  EVENT(name, chan, data, dataLen)   \
                 putMsg(eventsfd, EVENT_##name, chan, &eventMsg, data, dataLen)

#define  EVENT_DEF(name)  void EVENT_##name(EventMsg *event);

#define EVENT_VAL(type)  ( *((type *)(event->data)) )


//=============================================================================
// NOTE:  The list code assumes that the elements in the list are simply
//        numbers in the range of 1 to 255 (0 is reserved)

/*
This stuff "superceeded by Linux version of linked list funcs/macros
(see linux/list.h)


typedef struct {
  uchar  next;
  uchar  prev;
  void  *listHead;  // which list is this element a part of
}  ListElement;

typedef struct  {
  uchar  first;
  uchar  last;
  ListElement *list;  // points to array of ListElements
} ListHead;


#define  InitList(head, list_)  \
  { (head).first = (head).last = 0;  (head).list = list_;  }

void near insertAtHead(ListHead *head, uchar item);

void near insertAtTail(ListHead *head, uchar item);

void near removeFromList(ListHead *head, uchar item);

bool  itemInList(ListHead *head, uchar item);

void  addIfNew(ListHead *head, uchar item);

// WARNING!  When using the ForEachNext() macro to traverse a list head,
//           do NOT remove the next item pointed to by the current item.
//           Doing so will cause the loop to fail!
#define  ForEachNext(hd, i, nxt)  \
    for ( (i)=(hd)->first; ((nxt)=(hd)->list[i].next,(i)); (i)=(nxt) )

// WARNING!  When using the ForEachPrev() macro to traverse a list head,
//           do NOT remove the prev item pointed to by the current item.
//           Doing so will cause the loop to fail!
#define  ForEachPrev(hd, i, prv)  \
    for ( (i)=(hd)->last; ((prv)=(hd)->list[i].prev,(i)); (i)=(prv) )

#define  FirstInList(hd)  ((hd)->first)

#define  LastInList(hd)  ((hd)->last)

#define  NextInList(hd, i)  ((hd)->list[i].next)
*/

//-------------------------------------------------------------------------

const char *inet_htop(int af, const void *src, char *dst, socklen_t cnt);
int inet_ptoh(int af, const char *src, void *dst);


//-------------------------------------------------------------------------

//--- CLearScreen, CLearLine, CLearPage terminal commands ---
extern const char clsCmd[], cllCmd[], clpCmd[];

void DispStr(uchar x, uchar y, char *msg, bool blankToEnd);
void DispStrAttr(uchar x, uchar y, char *attrs, char *msg, bool blankToEnd);

#define CursorTo(x, y)   printf("\033[%02d;%02dH", (y)+1, (x)+1)

//--- ANSI/vt100 Display Attribute codes ---
#define  DA_RESET_ALL     0
#define  DA_BRIGHT        1
#define  DA_DIM           2
#define  DA_UNDERSCORE    4
#define  DA_BLINK         5
#define  DA_REVERSE       7
#define  DA_HIDDEN        8
//--- Foreground Colors ---
#define  DA_FG_BLACK     30
#define  DA_FG_RED       31
#define  DA_FG_GREEN     32
#define  DA_FG_YELLOW    33
#define  DA_FG_BLUE      34
#define  DA_FG_MAGENTA   35
#define  DA_FG_CYAN      36
#define  DA_FG_WHITE     37
//--- Background Colors ---
#define  DA_BG_BLACK     40
#define  DA_BG_RED       41
#define  DA_BG_GREEN     42
#define  DA_BG_YELLOW    43
#define  DA_BG_BLUE      44
#define  DA_BG_MAGENTA   45
#define  DA_BG_CYAN      46
#define  DA_BG_WHITE     47

#define  RED  "\037"  // DA_FG_RED
#define  BLUE "\042"  // DA_FG_BLUE

// DA_DIM causes black text on OS-X terminals
//#define  DARK_RED  "\002\037"  // DA_DIM, DA_FG_RED
//#define  DARK_BLUE "\002\042"  // DA_DIM, DA_FG_BLUE
#define  DARK_RED   RED
#define  DARK_BLUE  BLUE

#define  RedText  "\033[31m"
#define  BlueText "\033[34m"

//-----------------------------------------------------------------------------

#endif


