//----- linux/list.h ----- 07/26/10 -----(04/19/07)-----

//-----------------------------------------------------------------------------
//  Significantly modified version of standard linux header file
//-----------------------------------------------------------------------------
//  Change log (most recent first):
//
//
//  2010-07-26 - Mark Davis
//               - Added list_for_each_prev_safe() macro
//
//  2010-04-29 - Mark Davis
//               - Cleaned up a bit and changed __list_del() to a macro.
//
//  2007-04-19 - Mark Davis
//               - Original version
//
//-----------------------------------------------------------------------------



#ifndef _LINUX_LIST_H
#define _LINUX_LIST_H

#if defined(__KERNEL__) || defined(_LVM_H_INCLUDE)

//#include <linux/prefetch.h>


/*
 * Simple doubly linked list implementation.
 *
 * Some of the internal functions ("__xxx") are useful when
 * manipulating whole lists rather than single entries, as
 * sometimes we already know the next/prev entries and we can
 * generate better code by using them directly rather than
 * using the generic single-entry routines.
 */

struct list_head {
  struct list_head  *next, *prev;
};
typedef struct list_head  list_head_t;


#define LIST_HEAD_INIT(name) { &(name), &(name) }

#define LIST_HEAD(name) \
  struct list_head name = LIST_HEAD_INIT(name)

#define INIT_LIST_HEAD(ptr) do { \
  (ptr)->next = (ptr); (ptr)->prev = (ptr);  } while (0)


/*----------------------------------------------------------------------------*
 * Insert a new entry between two known consecutive entries.
 *
 * This is only for internal list manipulation where we know
 * the prev/next entries already!
 *----------------------------------------------------------------------------*/
void __list_add(list_head_t *new_, list_head_t *prev_, list_head_t *next_);

/*----------------------------------------------------------------------------*
 * list_add - add a new entry
 * @new: new entry to be added
 * @head: list head to add it after
 *
 * Insert a new entry after the specified head.
 * This is good for implementing stacks.
 *----------------------------------------------------------------------------*/
#define list_add(new_, head_)  \
{  \
  __list_add((new_), (head_), (head_)->next);  \
}

/*----------------------------------------------------------------------------*
 * list_add_tail - add a new entry
 * @new: new entry to be added
 * @head: list head to add it before
 *
 * Insert a new entry before the specified head.
 * This is useful for implementing queues.
 *----------------------------------------------------------------------------*/
#define list_add_tail(new_, head_)  \
{  \
  __list_add((new_), (head_)->prev, (head_));  \
}

/*----------------------------------------------------------------------------*
 * Delete a list entry by making the prev/next entries
 * point to each other.
 *
 * This is only for internal list manipulation where we know
 * the prev/next entries already!
 *----------------------------------------------------------------------------*/
#define __list_del(_prev, _next)  \
{  \
  (_next)->prev = (_prev);  \
  (_prev)->next = (_next);  \
}

/*----------------------------------------------------------------------------*
 * list_del - deletes entry from list.
 * @entry: the element to delete from the list.
 * Note: list_empty on entry does not return true after this,
 *       the entry is in an undefined state.
 *----------------------------------------------------------------------------*/
#define list_del(entry_)  \
{  \
  __list_del((entry_)->prev, (entry_)->next);  \
}

/*----------------------------------------------------------------------------*
 * list_del_init - deletes entry from list and reinitialize it.
 * @entry: the element to delete from the list.
 *----------------------------------------------------------------------------*/
#define  list_del_init(entry_)  \
{  \
  __list_del((entry_)->prev, (entry_)->next);  \
  INIT_LIST_HEAD(entry_);  \
}

/*----------------------------------------------------------------------------*
 * list_empty - tests whether a list is empty
 * @head: the list to test.
 *----------------------------------------------------------------------------*/
#define  list_empty(head_)  ( (int) ( (head_)->next == (head_) ) )


/*----------------------------------------------------------------------------*
 * list_empty_careful - tests whether a list is empty and not being modified
 * @head: the list to test
 *
 * Description:
 * tests whether a list is empty _and_ checks that no other CPU might be
 * in the process of modifying either member (next or prev)
 *
 * NOTE: using list_empty_careful() without synchronization
 * can only be safe if the only activity that can happen
 * to the list entry is list_del_init(). Eg. it cannot be used
 * if another CPU could re-list_add() it.
 *----------------------------------------------------------------------------*/
int list_empty_careful(struct list_head *head);


/*----------------------------------------------------------------------------*
 * list_splice - join two lists
 * @list: the new list to add.
 * @head: the place to add it in the first list.
 *----------------------------------------------------------------------------*/
void list_splice(struct list_head *list, struct list_head *head);


/*----------------------------------------------------------------------------*
 * list_entry - get the struct for this entry
 * @ptr:  the &struct list_head pointer.
 * @type: the type of the struct this is embedded in.
 * @member: the name of the list_struct within the struct.
 *----------------------------------------------------------------------------*/
#define list_entry(ptr, type, member) \
  ( (type *) ((char *)(ptr)-(unsigned)(&((type *)0)->member)) )

#define list_entry_(ptr, type, offset) \
  ( (type *) ((char *)(ptr)-(unsigned)(offset)) )

/*----------------------------------------------------------------------------*
 * list_for_each  - iterate over a list
 * @pos:  the &struct list_head to use as a loop counter.
 * @head: the head for your list.
 *----------------------------------------------------------------------------*/
#define list_for_each(pos, head) \
  for (pos = (head)->next; pos != (head); pos = pos->next  )

/*----------------------------------------------------------------------------*
 * list_for_each_safe - iterate over a list safe against removal of list entry
 * @pos:  the &struct list_head to use as a loop counter.
 * @n:    another &struct list_head to use as temporary storage
 * @head: the head for your list.
 *----------------------------------------------------------------------------*/
#define list_for_each_safe(pos, n, head) \
  for ( pos = (head)->next, n = pos->next; \
        pos != (head); \
        pos = n, n = pos->next )

/*----------------------------------------------------------------------------*
 * list_for_each_prev - iterate over a list in reverse order
 * @pos:  the &struct list_head to use as a loop counter.
 * @head: the head for your list.
 *----------------------------------------------------------------------------*/
#define list_for_each_prev(pos, head) \
  for (pos = (head)->prev;  pos != (head);  pos = pos->prev)

/*----------------------------------------------------------------------------*
 * list_for_each_safe - iterate over a list in reverse order and safe against
 * removal of list entry
 * @pos:  the &struct list_head to use as a loop counter.
 * @n:    another &struct list_head to use as temporary storage
 * @head: the head for your list.
 *----------------------------------------------------------------------------*/
#define list_for_each_prev_safe(pos, p, head) \
  for ( pos = (head)->prev, p = pos->prev; \
        pos != (head); \
        pos = p, p = pos->prev )


#endif /* __KERNEL__ || _LVM_H_INCLUDE */

#endif
