//----- semaphore.h ----- 04/23/12 -----(06/08/09)-----

//-----------------------------------------------------------------------------
//  DEBROS version of standard linux header file
//-----------------------------------------------------------------------------
//  Change log (most recent first):
//
//
//  2012-04-23 - Mark Davis
//               - Added semCounts() to get total and free counts for the
//                 global pool of named semaphores.
//
//  2010-07-22 - Mark Davis
//               - Added prototypes for standard POSIX semaphore functions and
//                 the DEBROS version of the sem_t datatype.
//
//  2009-06-08 - Mark Davis
//               - Original version
//
//-----------------------------------------------------------------------------

#ifndef _SEMAPHORE_H_
#define _SEMAPHORE_H_



#include <sys/types.h>
#include <linux/sched.h>

#include <linux/wait.h>


#define  MAX_NUM_SEMS   16  // Max # of named semaphores for all tasks

#define  SEM_NSEMS_MAX   MAX_NUM_SEMS // Max # of named semaphores per task

#define  SEM_VALUE_MAX   32767  // Max initial value for a semaphore


typedef struct {
  char far  *name;         // NULL for unnamed semaphores
  WaitQueue  opened;       // what tasks have opened this semaphore?
  WaitQueue  locked;       // list of tasks that have been granted a lock
  WaitQueue  blocked;      // list of tasks waiting for a lock
  bool       unlinked;     // has sem_unlink() been called?
  bool       inUse;        // is this semaphore in use/initialized?
  int        value;        // current value: 0 means all locks in use
  void      *myAddr;       // used to verify it is a sem_t struct
}  sem_t;


void  initSemaphores(void);  // called during startup

uint _nearcall semCounts(uint *ttl);


int sem_wait_(sem_t *sem, ulong timeoutTicks);


sem_t *sem_open(const char *name, int oflag, ulong mode, uint value);
#define  SEM_FAILED  NULL

int sem_init(sem_t *sem, int pshared, uint value);

int sem_close(sem_t *sem);

#define sem_wait(sem)  sem_wait_(sem, MAX_SCHEDULE_TIMEOUT)

int sem_timedwait(sem_t *sem, const struct timespec *abs_timeout);

int sem_trywait(sem_t *sem);

int sem_post(sem_t *sem);

int sem_destroy(sem_t *sem);

int sem_getvalue(sem_t *sem, int *sval);

int sem_unlink(const char *name);



/*---- stuff derived from std Linux semaphore.h that I don't use right now ----

typedef char  atomic_t;

#define atomic_dec_return(__byteVar) \
  (char) _asm ([ _HL, _A ], \
      _HL = (unsigned) ((char *) &(__byteVar)), \
      "ipset 2",   \
      "dec (hl)",  \
      "ld a,(hl)", \
      "ipres",     \
      _A ), _A

#define atomic_inc_return(__byteVar) \
  (char) _asm ([ _HL, _A ], \
      _HL = (unsigned) ((char *) &(__byteVar)), \
      "ipset 2",   \
      "inc (hl)",  \
      "ld a,(hl)", \
      "ipres",     \
      _A ), _A

#define atomic_set(__bytePtr, __val) \
  (void) _asm ([ _HL, _A ], \
      _HL = (unsigned) ((char *) (__bytePtr)), \
      _A = (char) (__val), \
      "ld  (hl),a",  \
     )

#define  might_sleep()


struct semaphore {
  atomic_t           count;
  int                sleepers;
  wait_queue_head_t  wait;
};

#define __SEMAPHORE_INITIALIZER(name, n)        \
{                 \
  (n),     \  // .count =
   0,      \  // .sleepers =
   __WAIT_QUEUE_HEAD_INITIALIZER((name).wait)  \  // .wait =
}

#define __DECLARE_SEMAPHORE_GENERIC(name,count) \
  struct semaphore name = __SEMAPHORE_INITIALIZER(name,count)

#define DECLARE_MUTEX(name) __DECLARE_SEMAPHORE_GENERIC(name,1)
#define DECLARE_MUTEX_LOCKED(name) __DECLARE_SEMAPHORE_GENERIC(name,0)

//static inline void sema_init (struct semaphore *sem, int val)
#define sema_init(sem, val)  \
{  \
  atomic_set(&(sem)->count, (val));  \
  (sem)->sleepers = 0;  \
  init_waitqueue_head(&(sem)->wait);  \
}

//static inline void init_MUTEX (struct semaphore *sem)
#define  init_MUTEX(sem)  \
{  \
  sema_init((sem), 1);  \
}

//static inline void init_MUTEX_LOCKED (struct semaphore *sem)
#define  init_MUTEX_LOCKED(sem)  \
{  \
  sema_init((sem), 0);  \
}


//extern spinlock_t semaphore_wake_lock;

//---------------------------------------------------------
static inline void down(struct semaphore * sem)
{
  might_sleep();
  if (atomic_dec_return(&sem->count) < 0)  __down(sem);
}

//---------------------------------------------------------
static inline int down_interruptible(struct semaphore * sem)
{
  int ret = 0;

  might_sleep();
  if (atomic_dec_return(&sem->count) < 0)  ret = __down_interruptible(sem);
  return ret;
}

//---------------------------------------------------------
static inline int down_trylock(struct semaphore * sem)
{
  int ret = 0;

  if (atomic_dec_return(&sem->count) < 0)  ret = __down_trylock(sem);
  return ret;
}

//--------------------------------------------------------------
// Note! This is subtle. We jump to wake people up only if
// the semaphore was negative (== somebody was waiting on it).
//---------------------------------------------------------------
static inline void up(struct semaphore * sem)
{
  if (atomic_inc_return(&sem->count) <= 0)  __up(sem);
}

//---------------------------------------------------------

*/

#endif /* __ASM_SH_SEMAPHORE_H */
