//----- sys/hdwDefs.h ----- 10/27/11 -----(05/18/06)-----

// definitions and prototype common to most I/O boards

//-----------------------------------------------------------------------------
//  Update Log (most recent first)
//
//  2011-10-27 - Mark Davis
//               - Added CurSP() macro to get the actual stack pointer value
//                 (as opposed to the one stored with the task info which is
//                 only updated during a call to schedule() ).
//
//  2011-06-24 - Mark Davis
//               - SetMin(), Setmax(), ReadMin(), ReadMax() macros now all do
//                 a range check on the index
//
//  2011-04-20 - Mark Davis
//               - Added getAvgSamplesADC()
//
//  2011-04-15 - Mark Davis
//               - Added resetAvgDataADC()
//
//  2011-04-12 - Mark Davis
//               - Added getAvgdRawIn() to get
//
//  2011-04-08 - Mark Davis
//               - Moved functions for tracking and managing averaged and noise
//                 values for analog inputs to the hdwDefs module.
//
//  2011-04-07 - Mark Davis
//               - Added adcNoisePeriod[], adcRawNoise[], adcNoiseTimeMS[] for
//                 tracking noise for each ADC input for configurable periods
//                 of time.
//
//  2011-01-07 - Mark Davis
//               - Added AIntoADC() macro
//
//  2010-12-06 - Mark Davis
//               - Added declaration for softReset() (implemented in DEBROS
//                 version of flash.c)
//
//  2010-05-17 - Mark Davis
//               - Increased IntVectDesc[] to 32 entries to match later versions
//                 of Rabbit 3000.
//
//  2010-03-31 - Mark Davis
//               - Added declaration of IntVectDesc[]
//
//  2009-12-07 - Mark Davis
//               - Changed names for forced ADC/DAC variables and func to
//                 avoid confusion (i.e. is it the ADC/DAC or read/set values
//                 that are affected?)
//
//  2009-10-27 - Mark Davis
//               - Added readValMin[] and readValMax[]
//               - Changed SetMin()/SetMax() and ReadMin()/ReadMax() macros
//                 to return setValMin/Max[] and readValMin/Max[]
//
//  2009-05-19 - Mark Davis
//               - added newCalibsADC, newCalibsDAC
//
//  2008-04-07 - Mark Davis
//               - added registerFileSystems()
//
//  2007-08-07 - Mark Davis
//               - fixed AinToRead() macro to use gain/offset correctly
//
//-----------------------------------------------------------------------------

#ifndef _sys_hdwDefsH_
#define _sys_hdwDefsH_


#include <linux/sched.h>


#define  MAX_LABEL_LEN  24


#define  ADCHistSize        100  // # of samples in circular history buffer
//#define  ADCSampleInterval   10  // # ms between samples
#define  ADCSampleInterval   SchedMS  // # ms between samples


extern unsigned char  __brdInitFlag;

extern ulong  ramSize, flashSize;

extern int  qdCount;

extern const char ArchName[];

extern const uchar  ActivityLED;


//--- bit values for MBxCR's (Memory Bank # Control Registers) ---
#define  _WS0     0xC0
#define  _WS1     0x80
#define  _WS2     0x40
#define  _WS4     0x00

#define  _IA19    0x20
#define  _IA18    0x10

#define _OE1_WP   0x0C
#define _OE0_WP   0x08
#define _OE1_WE1  0x04
#define _OE0_WE0  0x00

#define _CS0      0x00
#define _CS1      0x01
#define _CS2      0x02




//--- flags to let the applic code know that something about one of the ---
//--- analog inputs or outputs changed and that values need to be recomputed ---
extern bool  updateAIn[], updateAOut[];


//--- benchtop calibration data for each analog input and output ---
//--- restored from a file on startup ---
extern float  adcGain[], adcOffset[];
extern float  dacGain[], dacOffset[];


//--- Computed by converting min/max raw ADC/DAC values to engineering ---
//--- units using calibration data for each channel ---
extern float  ainMin[], ainMax[], aoutMin[], aoutMax[];
extern float far  setValMin[], setValMax[], readValMin[], readValMax[];


//--- real (or forced) state for each channel ---
extern float  anaIn[], anaOut[], lastAnaOut[];
extern uint  adcRaw[], dacRaw[], lastadcRaw[];
extern bool  newRdg[];


//--- keep track of which channels are using forced values ---
extern bool  adcForced[], dacForced[];
extern float  adcForcedVal[], dacForcedVal[];


//--- data for keeping a running average for ADC inputs ---
extern uint far  adcRawHist[][ADCHistSize];
typedef struct {
  ulong  sum;
  uint   oldest;
  uint   next;
  uint   curNumSamples;
  uint   useNumSamples;
}  ADCAvgData;

extern ADCAvgData  adcAvgData[];


//--- data for computing noise over set blocks of time for each input ---
typedef struct {
  uint  rawMin;       // min raw ADC values for specified period
  uint  rawMax;       // max raw ADC values for specified period
  uint  noisePeriod;  // # of ms period used for computing noise
  uint  rawNoise;     // max-min diff from prev period
  uint  noiseTimeMS;  // LS 16-bits of ms time the noise was updated
} ADCNoiseData;

extern ADCNoiseData  adcNoiseData[];


extern const uint  NumHdwADC, NumHdwDAC;
extern const uint  MaxDIN, MaxDOUT;


//--- conversion btwn ctlr analog engineering and appl engineering units ---
extern float  readGain[], readOffset[];
extern float  setGain[], setOffset[];

extern float  readVal[], setVal[], lastSetVal[];


#define  NumIntVectors  30

extern const char far *IntVectDesc[NumIntVectors];



#define ADCGain(chan)    ( adcGain[(uint)(chan)] )
#define ADCOffset(chan)  ( adcOffset[(uint)(chan)] )

#define DACGain(chan)    ( dacGain[(uint)(chan)] )
#define DACOffset(chan)  ( dacOffset[(uint)(chan)] )





//--- NOTE: must accept float values for rawADC ---
//--- (if avgd from raw values, we don't want to loose the extra accuracy) ---
#define ADCtoAIn(ui, rawADC)  \
    ( fmult(((float)(rawADC)-ADCOffset(ui)), ADCGain(ui)) )

#define AIntoADC(ui, anaVal)  \
    ( (anaVal) / ADCGain(ui) + ADCOffset(ui) )

#define AInToRead(ui, ai)  ( fmult(((ai)-readOffset[ui]), readGain[ui]) )


#define AOutToDAC(ui, ao)  \
           (uint) ( (ao) / DACGain(ui) + DACOffset(ui) )
#define DACtoAOut(ui, rawDAC)  \
           ( ((float)rawDAC - DACOffset(ui)) * DACGain(ui) )

#define SetToAOut(ui, egu)  ( (egu)/setGain[ui] + setOffset[ui] )
#define AOutToSet(ui, ao)   ( fmult(((ao)-setOffset[ui]), setGain[ui]) )


#define  ANAIN(ui)   ( (uint)(ui) >= NumHdwADC ? 0.0 : anaIn[ui] )
#define  ANAOUT(ui)  ( (uint)(ui) >= NumHdwDAC ? 0.0 : anaOut[ui] )

#define  ADCRAW(ui)  ( (uint)(ui) >= NumHdwADC ? 0 : adcRaw[ui] )
#define  DACRAW(ui)  ( (uint)(ui) >= NumHdwDAC ? 0 : dacRaw[ui] )


#define AInMin(ui)  ( (uint)(ui) >= NumHdwADC ? -1.0 : ainMin[(uint)(ui)] )
#define AInMax(ui)  ( (uint)(ui) >= NumHdwADC ?  1.0 : ainMax[(uint)(ui)] )

#define AOutMin(ui)  ( (uint)(ui) >= NumHdwDAC ? -1.0 : aoutMin[(uint)(ui)] )
#define AOutMax(ui)  ( (uint)(ui) >= NumHdwDAC ?  1.0 : aoutMax[(uint)(ui)] )


//#define SetMin(ui)  ( AOutToSet((ui), AOutMin(ui)) )
//#define SetMax(ui)  ( AOutToSet((ui), AOutMax(ui)) )

#define SetMin(ui)  ( (ushort)(ui) >= NumHdwDAC ?  1.0 : setValMin[(ushort)(ui)] )
#define SetMax(ui)  ( (ushort)(ui) >= NumHdwDAC ? -1.0 : setValMax[(ushort)(ui)] )

//#define  ReadMax(ui)  fmult( (readMaxRaw(ui) - readOffset[ui]), readGain[ui] )
//#define  ReadMin(ui)  fmult( (readMinRaw(ui) - readOffset[ui]), readGain[ui] )

#define ReadMin(ui)  ( (ushort)(ui) >= NumHdwADC ?  1.0 : readValMin[(ushort)(ui)] )
#define ReadMax(ui)  ( (ushort)(ui) >= NumHdwADC ? -1.0 : readValMax[(ushort)(ui)] )



extern ulong  digitalIn, digitalOut, nextDigitalOut, prevDigitalOut;
extern ulong  forcedDigInMask, forcedDigInBits,
              forcedDigOutMask, forcedDigOutBits;
extern ulong  extraIn;
extern ulong  forcedExtraInMask, forcedExtraInBits;


extern char far  ainLabel[][MAX_LABEL_LEN];
extern char far aoutLabel[][MAX_LABEL_LEN];
extern char far  dinLabel[][MAX_LABEL_LEN];
extern char far doutLabel[][MAX_LABEL_LEN];
extern char far extraInLabel[][MAX_LABEL_LEN];

#define EXTRA_IN_BITS  32



#define CurSP() (uint) ( _asm ( [ _HL ], "ld hl,0", "add hl,sp", _HL ) )


void initIOvalues(void);


//unsigned int anaIn(unsigned int channel);
//int anaOut(unsigned int channel, unsigned int rawcount);

ulong  ReadDigInputs(void);
void   WriteDigOutputs(ulong digOut);

ulong  ReadDeviceID(void);


void hdwInit(ulong digOutState);
void hdwInit2(void);

void netHdwInit(void);

void registerFileSystems(void);


void digOutConfig(ulong outputMode);


void _nearcall newDACCalib(uint dacIdx);
void _nearcall newADCCalib(uint adcIdx);

void _nearcall setStateDAC(uint uidx, float val, bool clearForced);


void writeAnalogOutputs(ulong mask);
void readAnalogInputs(ulong mask);


int saveAnalogCalib(void);
int loadAnalogCalib(void);


//--- functions for managing averaged version of ADC readings ---
uint setAvgPeriodADC(uint idx, uint numMS);
uint getAvgPeriodADC(uint idx);
uint getAvgSamplesADC(uint idx);

void resetAvgDataADC(uint idx);

void updateAvgsADC(void);

float getAvgdRawIn(uint idx);
float getAvgdAnaIn(uint idx);

#define  GetAvgdRead(idx)  ( AInToRead(idx, getAvgdAnaIn(idx)) )


//--- functions for managing the tracking of noise levels on ADC readings ---
uint setNoisePeriodADC(uint idx, uint numMS);
uint getNoisePeriodADC(uint ui);

void resetNoiseDataADC(uint idx);

uint getNoiseADC(uint idx, uint *ms16);



#endif
