//----- sys/socket.h ----- 06/06/12 -----(03/12/06)-----

//-----------------------------------------------------------------------------
//  Update Log (most recent first)
//
//
//  2012-06-06 - Mark Davis
//               - Added sockState() (which takes a SockInfo *) rather than
//                 sockstate() (which takes a sock_type *) for use in other
//                 modules.
//               - Changed SockInfo.closeCalled to noFDEVrefs to clarify the
//                 meaning of this flag.
//
//  2012-04-20 - Mark Davis
//               - Added tcpCounts() and udpCounts() to get total and free
//                 counts for the global pool of sockets.
//
//  2012-02-24 - Mark Davis
//               - Changed SockInfo.state to conState and expanded its use to
//                 better track the state of a connection.
//
//  2011-03-25 - Mark Davis (*)
//               - Added closeCalled to avoid freeing up a SockInfo struct when
//                 it is still in use by an FDEV
//
//  2011-01-26 - Mark Davis
//               - added gateway to IFConfig structure
//               - Added DEF_NETMASK and DEF_GATEWAY values
//               - Added showRoutes()
//
//-----------------------------------------------------------------------------

#ifndef _sys_socketH_
#define _sys_socketH_


#include <sys/types.h>
#include <netinet/in.h>

#include <stcpip.h>

#include <linux/limits.h>

#include <linux/wait.h>



#define DEF_NETMASK  "255.255.0.0"
#define DEF_GATEWAY  "192.168.0.1"


extern ulong  netmask, gateway;  // actually declared by each appl as a register


typedef struct  {
  sock_type  *sock;        // tcp_Socket or udp_Socket - MUST be 1st in struct!
  bool        inUse;       // this entry is in use by an FDEV
  bool        noFDEVrefs;  // OK to free this struct after *sock is closed
  char        type;        // SOCK_DGRAM or SOCK_STREAM
  int         ourPort;     // local port
  ulong       ourAddr;     // local addr (sets/indicates interface)
  bool        listener;    // passive open used to listen for connections
  short       conState;    // connection state (e.g. SS_UNCONNECTED)
  ulong       startTime;   // used for checking timeouts on calls
  bool        rdOK;        // open for reading
  bool        wrOK;        // open for writing
  short       closeState;  // state of closing process
  bool        bcastMode;   // use broadcast address when sending packets
  bool        opened;      // socket has been opened (UDP only)
  ulong       diagFlags;   // set by task that last accessed the socket
  ulong       lastClosed;  // last time this entry was closed
  WaitQueue   readQueue;   // processes waiting to read from this socket
  WaitQueue   writeQueue;  // processes waiting to write to this socket
  void       *myAddr;      // used to verify this is a valid SockInfo struct
} SockInfo;


extern SockInfo  tcpSockets[];
extern SockInfo  udpSockets[];


#define RemoteAddr(sp)  ((sp)->sock ? (sp)->sock->tcp.hisaddr : 0)
#define RemotePort(sp)  ((sp)->sock ? (sp)->sock->tcp.hisport : 0)

#define RemoteAddrUDP(sp)  ((sp)->sock->udp.hisaddr)
#define RemotePortUDP(sp)  ((sp)->sock->udp.hisport)

char *fmtAddrPort(char *buf, ulong addr, uint port);

const char *sockState(SockInfo *si);


//--- values for SockInfo.state ---
// !!!WARNING!!!  This refers to the state of processing by the functions in
// the socket.c module, NOT the state of the underlying socket!!!
#define  SS_UNCONNECTED   0    // not yet connected
#define  SS_CONNECTING    1    // connection being established
#define  SS_CONNECTED     2    // connection established
#define  SS_DISCONNECTED  3    // no longer connected

//--- values for SockInfo.closeState ---
// !!!WARNING!!!  This refers to the state of processing by the functions in
// the socket.c module, NOT the state of the underlying socket!!!
#define  CS_OPEN         0  // socket open
#define  CS_CALL_CLOSE   1  // need to initiate a close
#define  CS_CLOSE_WAIT   2  // waiting for close to finish
#define  CS_IS_CLOSED    3  // socket is closed
#define  CS_WAIT_FREE    4  // _socketClose_() not called yet (*)
/*
  (*) When shutting down the interface, the driver closes all the underlying
      open sockets, but the associated SockInfo structure must stay around
      until it is freed by a call to _socketClose_()
*/

//-----------------------------------------------------------------------------
//  stuff for which I haven't found or implemented a standard Unix/Linux func
//-----------------------------------------------------------------------------

typedef struct {
  ulong   ipAddr;     // in host byte order
  ulong   bcAddr;     // in host byte order
  ulong   netmask;    // in host byte order
  ulong   gateway;    // in host byte order
  uchar   macAddr[6]; // Ethernet MAC address
  uint    mtu;        // fixed at 1500 by SoftTools port of net libs (?)
  uchar   state;      // IF_DOWN, IF_COMING_UP, IF_UP, IF_COMING_DOWN
} IFConfig;


int getIFConfig(int iface, IFConfig *ifConfig);
int setIFConfig(int iface, IFConfig *ifConfig);

int startNetUp(int iface);
int startNetDown(int iface);

// use ifpending(iface) to get state of interface

void chgIPAddr(int iface, ulong newAddr);  // for convenience only...


void showRoutes(void);

//-----------------------------------------------------------------------
//--- prototypes for functions for use ONLY by other Kernel functions ---
//-----------------------------------------------------------------------

uint _nearcall tcpCounts(uint *ttl);
uint _nearcall udpCounts(uint *ttl);


void socketInit_(void);

int TCP_tick(void* s);

int numOpenSockets(void);

fd_t _socket_(int family, int type, int protocol);

int _listen_(fd_t sockfd, int backlog);

int _connect_(fd_t sockfd, const struct sockaddr *remoteAddr, int addrLen);


//=============================================================================
//               prototypes for standard BSD-style socket API
//=============================================================================

fd_t accept(fd_t listeningSock, struct sockaddr *newClient, int *addrLen);

int bind(fd_t sockfd, const struct sockaddr *name, int nameLen);

int connect(fd_t sockfd, const struct sockaddr *remoteAddr, int addrLen);

int listen(fd_t sockfd, int backlog);

int recv(fd_t sockfd, void far *buff, int numBytes, int flags);

int recvfrom(fd_t sockfd, void far *buff, int numBytes, int flags,
             struct sockaddr *from, int *addrLen);

int send(fd_t sockfd, const void far *buff, int numBytes, int flags);

int sendto(fd_t sockfd, const void far *buff, int numBytes, int flags,
           const struct sockaddr *to, int addrLen);

  //--- values for flags (mostly ignored) ---
  #define MSG_OOB         0x0001  // process out-of-band data
  #define MSG_PEEK        0x0002  // peek at incoming message
  #define MSG_DONTROUTE   0x0004  // send without using routing tables
  #define MSG_TRYHARD     0x0004  // Synonym for MSG_DONTROUTE for DECnet
  #define MSG_CTRUNC      0x0008
  #define MSG_PROBE       0x0010  // Do not send. Only probe path f.e. for MTU
  #define MSG_TRUNC       0x0020
  #define MSG_DONTWAIT    0x0040  // Nonblocking I/O
  #define MSG_EOR         0x0080  // End of record
  #define MSG_WAITALL     0x0100  // Wait for a full request
  #define MSG_FIN         0x0200
  #define MSG_SYN         0x0400
  #define MSG_CONFIRM     0x0800  // Confirm path validity
  #define MSG_RST         0x1000
  #define MSG_ERRQUEUE    0x2000  // Fetch message from error queue
  #define MSG_NOSIGNAL    0x4000  // Do not generate SIGPIPE
  #define MSG_MORE        0x8000  // Sender will send more

  #define MSG_EOF         MSG_FIN


fd_t socket(int family, int type, int protocol);

  //--- family values supported ---
  #define AF_INET         2               // internetwork: UDP, TCP, etc.
  #define PF_INET         2               // internetwork: UDP, TCP, etc.

  //--- type values supported ---
  #define SOCK_DGRAM      1               // datagram socket
  #define SOCK_STREAM     2               // stream socket

  //--- protocol type (currently ignored) ---
  #define IPPROTO_TCP      6              // tcp
  #define IPPROTO_UDP     17              // user datagram protocol


int setsockopt(int sockfd, int level, int opt,
               const void *val, socklen_t valLen);

int getsockopt(int sockfd, int level, int opt,
               void *val, socklen_t *valLen);

  //--- values for set/getsocktopt() level ---
  #define SOL_SOCKET    1

  //--- values for set/getsocktopt() opt ---
  #define SO_BROADCAST  6



/*----- not yet implemented -----
int getpeername(int, struct sockaddr *, Psocklen_t);

int getsockname(int, struct sockaddr *, Psocklen_t);

int recvmsg(int, struct msghdr *, int);

int socketpair(int, int, int, int *);

ssize_t sendmsg(int, const struct msghdr *, int);

int shutdown(int, int);
---------------------------------*/



#endif
