//----- fcntl.c ----- 12/05/12 -----(03/12/06)-----

//=============================================================================
//  Copyright (c) 2006-2012, NSCL and Mark Davis.  All rights reserved.
//=============================================================================
//  The generic or Virtual File System layer.  This layer provides the
//  functionality that is common to most of the file system drivers (such as
//  the handling of blocking vs non-blocking mode).  This avoids redundancy in
//  the individual drivers, helping to keep them small and minimize differences
//  in their behavior.
//=============================================================================
// TO-DO:
//
//  - Recent (re)reading leads me to believe that blocking mode for fds is
//    supposed to behave a bit differently than I thought.  While the man pages
//    do state that reads and writes on fds that have blocking mode set may
//    return without having read or written the requested # of bytes, my
//    impression was that only occured when there was an error or some signal
//    received that interrupted the process.
//
//    But now I believe that the intended behavior is that blocking mode simply
//    means that SOME data is read or written (or a non-blocked signal was
//    received).  The caller is supposed to take in to account the fact that
//    it is quite normal for a read or write on an fd with blocking mode set
//    to return having processed less than the requested # of bytes.
//
//    Even with this new interpretation, blocking mode is useful because it
//    allows a task to wait for SOME progress without using CPU cycles.  But
//    it means that is the ONLY difference between the two.  In both cases,
//    you are responsible for dealing with the possibility that only some (or
//    even NONE) of the specified # of bytes were read or written (a rather
//    strong possibility when using non-blocking mode).
//
//    But I THOUGHT it meant that, excluding errors or interruptions (which
//    you would be informed of), blocking mode insured no interference with
//    your reading or writing the specified # of bytes.  That no other task
//    or thread could write to or read from the connection until your read
//    or write completed (or was interrupted).
//
//    I thought, in addition to being critical to saving CPU cycles, that
//    blocking mode effectively created a temporary lock that meant no other
//    thread of execution could inject bytes into or take bytes from the
//    middle of the data stream you were writing to or reading from (with a
//    single call).
//
//    Could this be what locking mode is designed to handle?
//
//    A common enough circumstance where it would seem to me to be critical
//    is a service that forks off multiple child tasks that all read from
//    the same fds.  I don't remember any examples for such code that
//    mentions the possibility that multiple child tasks all reading from
//    the same fd could end up getting non-contiguous blocks of the byte
//    stream, or that blocking writes could return with only a portion of
//    the specified buffer written to the fd, and that you had to use locks
//    to insure other tasks did not inject data into the stream between
//    the multiple write calls that might be needed to write all the bytes
//    you need to.
//
//
//  NOTES:
//
//    The linux open() function defines O_DIRECT and O_ASYNC options for
//    the flags value.  O_DIRECT specifies synchronous I/O (the data is
//    guaranteed to have been transferred at the end of the call), while
//    O_ASYNC specifies asynchronous mode, which generates a signal when
//    input or output becomes possible on the descriptor (which is only
//    available for terminals, pseudo-terminals, and sockets).
//
//    So the question is, if neither of these is specified, what behavior
//    can you expect from read() and write() (and send() and recv()) calls?
//
//    Is it really normal that these return with less then the specified
//    number of bytes having been transferred?  Or, if they don't return
//    with having transferred all the bytes, do such operations on an fd
//    with blocking-mode set really not guard against concurrent operations
//    on the same I/O channel?
//
//    What if the fd represents a directory?  Surely you can't allow the
//    bytes from single reads and writes from multiple threads to be
//    mingled together without total chaos!
//
//    Could it be that some (maybe most) devices simply don't allow the
//    documented behavior to occur?  That most of them DO guarantee
//    all the bytes are transferred before the read or write returns?
//    And that each transferred block will be processed by the device
//    atomically?
//
//    I don't see how any other mode of operation make sense!
//
//    Also, while it is a bit old now, the book "UNIX Network Programming"
//    by W. Richard Stevens states (on page 163 of the 1990 edition):
//
//    "Unix I/O is termed synchronous.  A read or write system call does
//    not return to the calling process until the operation is complete
//    and the requested data has been copied by the kernel, either into
//    the buffer for a read, or from the buffer for a write."
//
//    While, as the book also explains, this is separate from the actual
//    reading from or writing to a disk, it means (at least) that the
//    call doesn't return with only SOME of the buffer transfered and
//    consider that a normal, error-free condition!
//
//    Of course, this doesn't address non-blocking mode calls, but you
//    only do that when your code IS designed to handle incomplete reads
//    and writes.
//
//=============================================================================
//  Update Log (most recent first):
//
//
//  2012-12-05 - Mark Davis
//               - Moved the following from fcntl to common:
//                    vt100 escape sequence definitions
//                    DispStr() and DispStrAttr()
//                    read_() and write_()
//                    redirectStdio() and logStdout()
//
//  2012-04-20 - Mark Davis
//               - Added gfdCounts() to get total and free counts for the
//                 global file descriptor table.
//
//  2011-12-29 - Mark Davis (*)
//               - read_(),readFDEV(), write_(), and writeFDEV() now set errno
//                 to ETIME if a timeout expires (regardless of what the return
//                 value is, which could be > 0 if any bytes were read/written).
//                 Prior to this, the caller of one of these functions had no
//                 way to tell when a failure to read/write the requested # of
//                 of bytes was due to a timeout (which is a normal, expected
//                 scenario in many cases).
//
//  2011-12-22 - Mark Davis (*)
//               - openCrashLog() now clears the redirStdio flag for the calling
//                 task so the data meant for the crash-log file is actually
//                 written to the file even if the task running at the time had
//                 the flag set (either because it had nothing open on pfd 1 or
//                 because it had been using that fd for something else).
//
//  2011-03-17 - Mark Davis
//               - Removed the nestedFDs[][] table as it is no longer needed.
//
//  2011-03-17 - Mark Davis (*)
//               - Fixed bug in fileSysCheck() where if it detected orphaned
//                 global fds and was asked to close them then it would
//                 attempt to do so using invalid (or incorrect) pointers.
//                 Luckily, the option to close them was never used, so this
//                 never actually happened.
//
//  2011-03-15 - Mark Davis
//               - Reduced stack usage of fileSysCheck() by using bit rather
//                 than byte arrays.
//               - added clear-to-end-of-line strings at end of each line of
//                 output for use with the -r option added to lsof
//
//  2011-02-04 - Mark Davis
//               - Added nestedFDs[][] for better supporting drivers like the
//                 revised SRFS one where one fd can be referenced by other fds
//
//  2011-01-10 - Mark Davis
//               - Elimiated extra blank lines in system log when the applic
//                 task writes to it in 2 conscecutive passes w/o another
//                 task doing so in between.
//
//  2010-12-14 - Mark Davis
//               - Changed _TimerSeconds + secTimerOffset to TIME()
//
//  2010-08-13 - Mark Davis
//               - Changed 'logStdio' to 'logStdout' in the task structure
//               - Added redirStdio flag to the task structure, which the
//                 redirectStdio() macro now sets in addition to the logStdout
//                 flag.  Setting both now causes redirectStdio() to work as
//                 originally intended:  Leave fd's 0,1,2 alone - just don't
//                 use them for stdin, stdout, stderr.
//               - Added the logStdout() macro that does what redirectStdio
//                 WAS doing: DUPLICATES stdout in the system log, rather than
//                 redirecting it.
//
//  2010-08-12 - Mark Davis
//               - Added OS and Appl date/time strings to the information
//                 written to the crash-log file.
//
//  2010-06-30 - Mark Davis (*)
//               - read_() and write_() now return -1 if a wait node cannot be
//                 allocated.
//
//  2010-06-15 - Mark Davis (*)
//               - read_() and write_() now check for a NULL buffer pointer
//
//  2010-06-02 - Mark Davis
//               - Added openCrashLog() to allow the stdio of the task that
//                 called postMortem() to change its stdout to a file to copy
//                 crash log data to.
//
//  2010-05-27 - Mark Davis
//               - Got rid of unnecessary modulus operation for milliseconds
//                 value in diagOut function (the one that writes to the system
//                 log).
//
//  2010-05-13 - Mark Davis
//               - Added read_() and changed write_() to they can be used as
//                 more efficient replacements for readBytes() and writeBytes()
//
//  2010-04-29 - Mark Davis
//               - Changes as needed to support the overhauled wait queue code.
//
//  2010-03-05 - Mark Davis (*)
//               - When not in use by a user shell, diagOut_() copies characters
//                 written to the system log to the console port (serial port A).
//                 Prior to this change, it would wait (if necessary) for space
//                 in the output buffer before returning.  Now it only does this
//                 if the WaitOnConsole bit is set in the calling task's
//                 diagFlags.
//
//                 NOTE that msgs are always written to the circular buffer,
//                 regardless of whether this bit is set, so the only downside
//                 to not having the bit set is some lost characters if you are
//                 monitoring things via the serial port (without a shell
//                 running on it).  The advantage of this change is to eliminiate
//                 the delays that could affect any task writing to the system
//                 log (since any task can fill the console's buffer, even a task
//                 that writes very little can be affected).
//
//                 Also NOTE that diagOut_() had KLOCK()/KUNLOCK() calls, but
//                 is only called by diagOut() (no underscore), which has
//                 KLOCK() invoked whenever diagOut_() is called, so the
//                 KUNLOCK()/sleep/KLOCK() sequence in diagOut_() would not
//                 suspend the process.  The effect was to waste CPU cycles by
//                 causing everything but interrupts to be disabled until space
//                 opened up in the serial output buffer.  Given that the sleep
//                 call was only for 5ms, and the serial output rate is 115000
//                 baud (11,500 cps), the delay would not be long in human terms,
//                 but it was probably one of the causes of measurable scheduler
//                 delays.
//
//  2009-12-18 - Mark Davis
//               - Made several near functions _nearcall to free up root space
//
//  2009-08-05 - Mark Davis
//               - _File_Swap_() now stops the descriptor's flags attributes
//                 from being swapped
//
//  2009-06-18 - Mark Davis
//               - The index used to copy data from the syslog buffer when
//                 writing to the remote backup copy of the log is now updated
//                 if the index for writing new chars "catches up" to it.  This
//                 minimizes the potential for lost data (the old way would drop
//                 an entire buffer's worth when this happened).
//
//  2009-06-12 - Mark Davis
//               - The clearing of the pid that is doing a read or write on a
//                 file is now done in a critical section to insure can't be
//                 preempted.
//
//  2009-06-11 - Mark Davis
//              - Corrected diagOut() so the date/time values in the system log
//                 entries have a millisecond value that is syncronized with the
//                 hh:mm:ss portion of the logged time.
//
//  2009-05-01 - Mark Davis
//               - fileSysCheck() (used by the lsof command) now takes a pointer
//                 to a buffer and its size as parameters (to reduce stack usage,
//                 which was overflowing sometimes...)
//
//  2009-04-15 - Mark Davis
//               - various calls that did not set errno for bad fds now set it
//                 to EBADF
//
//  2009-04-14 - Mark Davis
//               - diagIn() (which handles reads from FILE DiagLog/syslog) now
//                 sets errno to EAGAIN and returns -1 (so any task trying to
//                 read from it will not block and will not believe it has
//                 reached the end of the input).
//
//  2009-04-02 - Mark Davis
//               - The read/write mode of the flags value in an open are now
//                 enforced by read() and write()
//
//  2009-03-25 - Mark Davis
//               - Added unlink() and rename() to the FileSysFuncs API
//
//  2009-03-24 - Mark Davis
//               - Finished up fileSysChk() for checking on corrupt or orphaned
//                 file descriptors at the global and driver levels.
//
//  2009-03-20 - Mark Davis
//               - fixed bug in _closeFD_() that could result in a task closing
//                 a file just opened by another and leave the one it was
//                 supposed to close "orphaned".  This could be the cause of the
//                 "orphaned" UDP sockets that use up the small pool of available
//                 ones (resulting in a configd that can't do its job).
//
//  2008-04-30 - Mark Davis
//               - lseek now resets the dontCallRead flag if it doesn't return
//                 an error
//
//  2008-04-07 - Mark Davis
//               - changes to use dynamically registered file systems model
//
//  2007-10-27 - Mark Davis
//               - got basic wait queue stuff working
//
//  2007-10-05 - Mark Davis
//               - added support for unlink() calls for SRFS files
//
//  2006-03-12 - Mark Davis
//               - began work on original version
//
//-----------------------------------------------------------------------------
/*
  NOTES:


 Process, Global, and Driver File Descriptors

    The standard file system API (open, close, read, write, etc) provides
    a generic API for creating connections to files and devices.  The
    fcntl layer provides the functionality that is common to all the drivers,
    allowing the drivers to be smaller and concern themselves with just the
    device-specific functionality.

    Each driver implements a standard set of functions, allowing the fcntl
    layer to work with all the different types of devices supported by the
    drivers with little or no concern for diferrences between the different
    devices.

    Each open file normally has 3 different descriptors associated with it:

       - Process-specific File Descrptor (pfd, or just fd):  This is a
         number assigned to an open file.  Its value is usuable ONLY by
         the process that called the function that returned the pfd (e.g.
         open() or socket()).  The only exception to this is when a task
         spawns a child process.  In this case the child usually has a
         copy of its parent's pfds.  But while the fd values will be the
         same, they still represent a reference to the child's list of
         open files, not the parents.  pfd values currently represent an
         index in to the process' list of Global File Descriptors pointers
         (FD[pix][fd]).

       - Global File Descriptor (gfd):  A structure allocated by the fcntl
         layer to hold state information common to most drivers.  Multiple
         pfds (belonging to 1 or more tasks) can reference a given gfd.
         A gfd is usually expressed as a pointer to an FDEV structure.

       - Driver-specific File Descriptor (dfd):  A structure allocated by
         the driver responsible for accessing and making changes to the
         file.  In most cases, each dfd is referred to by just 1 gfd.
         But there are exceptions (such s the FIFO driver).  The structures
         are generally referred to generically as ioDev structures, and
         dfd's are usually expressed as a void pointer called ioDev.


    The FDEV structure is used by the fcntl layer to keep track of a single
    connection.  It contains the following information:

      - The data related to the state of the connection which is independent
        of what driver is used to access it.

      - A pointer to a structure (of pointers to) the driver-specifc versions
        of standard I/O functions (e.g. read, write, lseek) (*ioFuncs)

      - A pointer to a driver-specific structure that contains driver-specific
        state information for an open connection. (*ioDev)

    Each call to a driver-specific function from the fcntl layer includes the
    driver provided structure so the driver knows which device or connection
    to operate on.

    When a process opens a connection it gets back a pfd that refers to an
    FDEV structure for an open file.  NOTE that an FDEV structure can be
    referenced by more than one pfd.  This means that any changes to the
    state of a file by one process will be seen by all the other pfds that
    reference the same FDEV.

    Each FDEV entry has a counter that keeps track of how many pfds reference
    it.  Only when all pfd references to an FDEV are closed will the
    device represented by the FDEV actually be closed.


 Protecting shared resources:

    fcntl.c insures that only one task enters the _xxxxRead_() or _xxxxWrite()_
    driver functions for a given FDEV, so most drivers do not have to
    use locks in these functions to protect against concurrent use of an
    ioDev entry.

    However, if more than one FDEV can point to the same ioDev (as in
    FIFOs) or to ioDev's that share a resource, then the driver must use
    locks to protect the shared resource.  The SPI and SRFS drivers are
    examples of this:  The SPI driver because only one SPI transfer can
    be going on at any given time (even where there are multiple ports
    open in SPI mode).  And the SRFS driver because it uses a single
    connection to the server for all files and tasks and can handle just
    1 conversation for 1 operation at a time.


 Blocking - The fcntl layer vs the driver layer:

    From the point of view of the fcntl layer, drivers come in 2
    varieties:  Those that handle blocking on their own and those
    that don't.

    For the later group, the driver's read and write functions
    never block:  They simply read or write however bytes they can
    each time they are called.  It is up to the fcntl layer to handle
    the difference between blocking and non-blocking calls.

    For non-blocking reads and writes, the fcntl layer simply
    returns whatever the driver supplies.

    For blocking calls, it will call the driver's read or write
    function repeatedly until the specified # of bytes are processed
    or an error occurs.  Between calls to the driver, the task is
    suspended using the specified wait queue.  The more intelligent
    drivers provide their own wait queues which allows them to wake
    up tasks as soon as (and only when) there is a good chance the
    driver can process more bytes.  This improves response time and
    minimizes the number of CPU cycles wasted calling the driver
    when it cannot process more bytes.


-----//finish------------------------------------------------------------------

- Should implement 2 tables to manage all the drivers:
  - one with an entry for each "mounted" file system
  - one for each type of device

  The device table will have one entry for each driver (e.g. flash, srfs,
  async, spi, etc).  Each row will have pointers to generic func calls, some
  of which may be NULL (e.g. lseek, unlink, and rename would be null for
  char-mode devices like async).

  The file system table will have one entry for each file system.  Each entry
  in this table will have a device type that refers to an entry in the device
  table.

2008-04-07
  NOTE:  I now use a table of drivers, where each filesystem driver registers
         itself (usually at startup).  This allows the BSP (Board Support
         Package) or arch(itecture) hardware-specific layer to register
         drivers based on what the hardware supports.

-------------------------------------------------------
- Use a global (far?) array of file descriptors that are allocated as needed
  when processes open a new descriptor.

- Move the elements that should be per-process (like permissions, flags, etc)
  out of the global file descriptor and in to the per-process descriptor
  (actually I don't believe this is correct - I think the flags ARE supposed
  to be shared...)

- Create a global array of _IO_FILE structures that are allocated as needed
  when a process calls functions like fopen(), fdopen(), etc. that return
  a pointer to a FILE.  This structure will have to keep a pointer to the
  owning process, so it can be cleaned up when a process terminates (or
  keep a linked list of open FILE's in the process descriptor?)

-----------------------------------------------------------------------------*/

#include <time.h>
#include <common.h>

#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <utime.h>


#include <busybox.h>

#include <sys/hdwDefs.h>

#include <sys/srfs.h>

#include <dcdefs.h>

#include <ctype.h>
#include <limits.h>

#include <sys/SHD_headers.h>

#include <sys/modbus.h>

#include <tsh.h>

#include <applic.h>

#include <sys/serial.h>
#include <sys/fifo.h>
#include <sys/flashFile.h>
#include <sys/srfs.h>


#include <sys/farprintf.h>


#define __KERNEL__

#include <linux/wait.h>



//--- global list of open files ---
FDEV  far fileTbl[MaxOpenFiles];

//--- all references to open files by process ---
FDEV far *FD[NUM_TASKS][FD_SETSIZE];   // NULL ==> not in use

//--- table of registered file systems ---
FileSysFuncs *fileSysFuncs[MAX_FILE_SYS];

DiagLogInfo far  *diagLogInfo;
uint  DiagLogSize;
char far  *diagLogBuf;
bool  logDateTime = false;  // prefix each line with a date/time stamp

bool  noGettyOnA = true;



typedef struct  {
  char  *prefix;              // used to determine which driver to use
  char  *driverName;          // name of driver to use for prefix value
  const FileSysFuncs  *fsFuncs;     // table of funcs provided by each driver
}  FSTAB;

// fsFuncs value set when a driver's init func calls registerFileSys()
static FSTAB  fstab[] = {
  { "",          "flashFile", NULL },  // reserved for root file system
//  { "syslog",    "syslog",    NULL },  // dummy stdin, stdout/err ==> syslog
  { "fifo",      "fifo",      NULL },  // fifos and pipes
  { "/srfs/",    "srfs",      NULL },
  { "/dev/tty",  "serial",    NULL },
  { "/dev/ttyx", "serial",    NULL },
  { "/dev/spi",  "serial",    NULL }
};

#define NumPrefixes  (sizeof(fstab) / sizeof(FSTAB) )


#include <SHD_DLM.h>

#include <sys/farprintf.h>



//--- for now assume only one file system per device type/driver ---
//finish: Eventually arg should be index in to table of file systems
//        that allows multiple file systems per device type/driver
#define FileSysDev(fsidx)  (fsidx)

fd_t  corefd = -1;


//-----------------------------------------------------------------------------
//  for debugging - list names of registered file systems
//-----------------------------------------------------------------------------
/*
void listFileSystems(void)
{
  uint  u;
  FileSysFuncs *fsf;

//--- check each file system driver for orphaned entries ---
  printf("~~~~~~~ ");
  for (u=0; u<MAX_FILE_SYS; ++u)  {
    fsf = fileSysFuncs[u];  if (!fsf)  continue;
    printf("[%.16s] ", fsf->name);
  }
  printf("\r\n");
}
*/

//-----------------------------------------------------------------------------
//  Initialize file descriptor table and related structures
//-----------------------------------------------------------------------------
void fcntlInit()
{
  FBZERO(fileTbl);  FBZERO(FD);  BZERO(fileSysFuncs);

//--- initialize and register file systems that are NOT hardware dependent ---
  //syslogInit_();
  fifoInit_();

//--- initialize and regsiter file systems that ARE hardware dependent ---
  registerFileSystems();
}

//-----------------------------------------------------------------------------
//  function called by hardware-specific setup code (in the arch source
//  directory) to register a file system with the fcntl layer
//
//  On success, returns index in to list of file systems
//  On error, returns < 0
//-----------------------------------------------------------------------------
int registerFileSys(const FileSysFuncs *fsFuncs)
{
  uint  u, d;
  const FileSysFuncs  **fsf;


  printf("  Registering file system: %.32s\r\n", fsFuncs->name);

  for (u=0; u<MAX_FILE_SYS; ++u)  {
    fsf = fileSysFuncs + u;

  //--- return existing index if already registered ---
    if (*fsf)  {
      if ((*fsf)->open == fsFuncs->open)  return (int)u;
      continue;
    }

  //--- Found an unused entry, which means we checked all the registered ---
  //--- drivers for a match, so add the driver to the list ---
    *fsf = fsFuncs;
//    printf("Registered file system: %u, %.32s\r\n", u, fsFuncs->name);

  //--- Scan the list of file name prefixes and save a pointer to the ---
  //--- driver's functions for each entry with a matching driver name ---
    for (d=0; d<NumPrefixes; ++d)
      if (StrEq(fstab[d].driverName, fsFuncs->name))
        fstab[d].fsFuncs = fsFuncs;

    return (int)u;
  }


  return -1;
}

//-----------------------------------------------------------------------------
//  call the "tick" function for each registered file system that has one
//-----------------------------------------------------------------------------
void fileSysTick(void)
{
  uint  u;
  FileSysFuncs  *fsf;

  for (u=0; u<MAX_FILE_SYS; ++u)  {
    fsf = fileSysFuncs[u];
    if (fsf)  if (fsf->tick)  (fsf->tick)();
  }

}

//-----------------------------------------------------------------------------
//  Get the total and free counts for the global file descriptor table.
//  The *ttl is set to the total, and the free count is the return value.
//-----------------------------------------------------------------------------
uint gfdCounts(uint *ttl)
{
  uint  u, freeCount = 0;


//--- count the # unused descriptors ---
  KLOCK();
  for (u=0; u<MaxOpenFiles; ++u)  if (fileTbl[u].openCount == 0)  ++freeCount;
  KUNLOCK();

  *ttl = MaxOpenFiles;

  return freeCount;
}


//-----------------------------------------------------------------------------
//  check for "orphaned" global file descriptors (ones that are marked as
//  in-use but are not referenced by any process' local descriptors)
//-----------------------------------------------------------------------------
void fileSysCheck(bool closeOrphans, char *buf, int bufSize)
{
  uint  pfd, gfd, dfd, pix, u;
  FDEV far *file;
  void  *ioDev;
  const IODevFuncs  *ioFuncs;
  FileSysFuncs  *fsf;
  EntryState  entryState;
  bool  hasRef[MaxOpenFiles/8], orphaned[MaxOpenFiles/8],
        invalidFD[NUM_TASKS][FD_SETSIZE/8];


  BZERO(hasRef);  BZERO(orphaned);  BZERO(invalidFD);

  KLOCK();

//--- Check each process fd for each active task to make sure it is valid ---
//--- and generate a list of which global fds have active references ---
  for (pix=0; pix<NUM_TASKS; ++pix)  {
    if (! tasks[pix].inUse)  continue;
    for (pfd=0; pfd<FD_SETSIZE; ++pfd)    {
      file = FD[pix][pfd];  if (! file)  continue;
      gfd = (uint)(file - fileTbl);
      if ((file == fileTbl + gfd) AND (gfd < MaxOpenFiles))  {
         SetBit(hasRef, gfd);  continue; }
      SetBit(invalidFD[pix], pfd);
    }
  }

//--- Generate a list of global fds that are active but are not referenced ---
//--- by any process fds ---
  for (gfd=0; gfd<MaxOpenFiles; ++gfd)  {
    if (GetBit(hasRef, gfd))  continue;
    file = fileTbl + gfd;
    if (file->openCount)  SetBit(orphaned, gfd);
  }

  KUNLOCK();


//===== Display a list of the active driver fds for each driver =====
  for (u=0; u<MAX_FILE_SYS; ++u)  {
    fsf = fileSysFuncs[u];  if (!fsf)  continue;
    printf("checking fsf[%u] (%.32s)%s\r\n", u, fsf->name, cllCmd);
    for (dfd=0; ; ++dfd)  {
      if (! (ioDev = (fsf->entry)(dfd, &entryState)))  break;
      if (entryState.unused)  continue;
      printf("  [%2u]: %2u refs:  %.100s%s\r\n",
             dfd, entryState.openCount,
             (fsf->ioFuncs->descr)(ioDev, buf, bufSize-1), cllCmd );
    }
    printf("%s\r\n", cllCmd);
  }


//===== Display a list of the invalid process fds =====
//   Strictly speaking the task that had the corrupted fd(s) may be gone
//   by the time we print/log a message about it (since we release the
//   global lock) but we still want to know that it happened
  for (pix=0; pix<NUM_TASKS; ++pix)
    for (pfd=0; pfd<FD_SETSIZE; ++pfd)   if (GetBit(invalidFD[pix], pfd))
      printf("PID %5d:  fd[%u]:  Invalid fd%s\r\n",
             tasks[pix].pid, pfd, cllCmd);


//===== Display a list of (and optionally close) any orphaned global fds =====
//   NOTE that a lock isn't necessary here because the global FDs that
//   we are dealing with are already orphans
  for (gfd=0; gfd<MaxOpenFiles; ++gfd)  {
    if (!GetBit(orphaned, gfd))  continue;
    file = fileTbl + gfd;
    if (file->shared)  continue;
    ioDev = file->ioDev;  ioFuncs = file->ioFuncs;
    ioFuncs->descr(ioDev, buf, bufSize-1);
    printf("Orphaned global fd[%d]: %.128%s", gfd, buf, cllCmd);
    if (closeOrphans)  ioFuncs->close(ioDev);
  }

}

//-----------------------------------------------------------------------------
//  attempt to open a file to copy messages for a post-mortem dump to and
//  replace stdout with the fd for the file
//-----------------------------------------------------------------------------
void openCrashLog(void)
{
  if (corefd >= 0)  return;

  logStdout();  // so FILE *stdout is copied to the syslog buffer

  noGettyOnA = true;  // so syslog is copied to the console port

  corefd = open("crash-log", O_RDWR | O_CREAT | O_APPEND, 0x666);

//--- change pfd 1 so it also points to the crash-log file and make sure ---
//--- anything written to FILE *stdout is written to pfd 1 ---
  if (corefd >= 0)  { dup2(corefd, 1);  current->redirStdio = false;  }

  printf("\r\n"
    "=====================================================================\r\n");
  printf("\r\n"
    "  RTOS: %.40ls\r\n"
    "  APPL: %.40ls  %.40ls\r\n"
    "\r\n",
    osDTM, applDTM, applName);
}

//-----------------------------------------------------------------------------
//  attempt to open a file to copy messages for a post-mortem dump to and
//  replace stdout with the fd for the file
//-----------------------------------------------------------------------------
void closeCrashLog(void)
{
  if (corefd < 0)  return;

  close(corefd);  corefd = -1;

//--- The crash-log fd is duplicated on pfd 1 (stdout) by openCrashLog() ---
//--- which means we must close pfd 1 for the file to actually BE closed and ---
//--- cached data flushed to flash before we reboot ---
  fflush(stdout);  close(1);
}

//-----------------------------------------------------------------------------
//  function to handle reads from FILE diagLog
//-----------------------------------------------------------------------------
int _nearcall diagIn(void *userData)
{
  errno = EAGAIN;  return -1;
}

//-----------------------------------------------------------------------------
//  Write a character to the circular diagnostic log buffer (diagLogBuf[]).
//
//  KLOCK() must be active when this is called.  This should ONLY be called
//  by diagOut()!
//-----------------------------------------------------------------------------
void diagOut_(uchar ch)
{
  diagLogBuf[diagLogInfo->toIdx] = ch;
  if (++diagLogInfo->toIdx >= DiagLogSize)  {
    diagLogInfo->toIdx = 0;  diagLogInfo->wrapped = true; }

//--- if we "caught up" to the next char to be copied to the remote log, then ---
//--- bump the index for the log copying, because its better to loose a few ---
//--- chars then the entire buffer worth of messages ---
  if (diagLogInfo->toIdx == diagLogInfo->fromIdx)  {
    if (++diagLogInfo->fromIdx >= DiagLogSize)  diagLogInfo->fromIdx = 0;  }

//--- if there is no login shell running on the serial console/diag/prog ---
//--- port, then copy the char to the serial port's output buffer ---
  if (noGettyOnA)  {
    if (SerialSendCountA() < 250)
      SerialPutcA(ch);
    else  if (WaitOnConsole)  {
      while (SerialSendCountA() > 250);
      SerialPutcA(ch);
    }
  }

}

//-----------------------------------------------------------------------------
//  The output handler for the FILE * DiagLog.
//
//  This function is called whenever one of the stdio stream oriented output
//  functions is called for the FILE *DiagLog (e.g. fprintf, fputs, fwrite,
//  fputc, putc, vfprintf)
//
//  The FILE *DiagLog is used to write messages to the system log, which can
//  be viewed by the syslog command.  The system log is stored in a circular
//  buffer that is continuosly backed up to a remote file server by the backupd
//  daemon.
//
//  NOTES:
//
//    The FILE *DiagLog stream does not simply copy the data sent to it to the
//    buffer.  It keeps track of 3 things:
//
//     - What task is it currently writing a message for?
//     - Is the character a newline character?
//     - How long has it been since it was last called?
//
//    The purpose of this "filtering" is to make the log messages more useful
//    and readable as follows:
//
//     - Each non-blank line in the log has a date/time stamp and identifies
//       what task wrote the message
//
//     - Output from different tasks never shows up on the same line.  This
//       means messages can sometimes be split across lines when a task is
//       preempted while in the middle of writing a message to the log, but it
//       allows concurrent writes to the log without adding undue latency.
//
//     - If enough time has passed between messages, a blank line is added to
//       the log before writing the new message.  This keeps messages in
//       groups and makes the log MUCH easier to read.
//
//-----------------------------------------------------------------------------
int _nearcall diagOut(uchar ch, void *userData)
{
  uint  curTime;
  time_t  dateTime;
  uint  ms;
  char  *ptr;
  struct tm  tms;
  static bool  onNewLine, blankLine;
  static uint  lastTime, lastApplicTicks;
  static task_t  *lastTask;
  static char  temp[64];


  if (!(ch & 0x60) AND (ch != '\n'))  return 0;

  KLOCK();

  curTime = GetMS_();

//--- if char is from diff task than prev char, or it is the first char ---
//--- written to the log in several seconds, start a new group of lines ---
  if (logDateTime) if ((current != lastTask) OR (curTime - lastTime >  4000)) {
    if (!onNewLine)  { diagOut_('\r');  diagOut_('\n'); }
    if (!blankLine)  { diagOut_('\r');  diagOut_('\n'); }
    onNewLine = blankLine = true;  lastTask = current;
  }

  if (onNewLine AND (ch != '\n'))  {
    if (current == applTask)  if (applicTicks != lastApplicTicks)
      if (!blankLine)  { diagOut_('\r');  diagOut_('\n');  }
    if (logDateTime)  {
      _EnterCritical_();
      dateTime = TIME();  ms = milliTicks;
      _ExitCritical_();
      localtime_r(&dateTime, &tms);
      sprintf(temp, "%04d-%02d-%02d %02d:%02d:%02d.%03d %.24s[%d]: ",
              tms.tm_year+1900, tms.tm_mon+1, tms.tm_mday,
              tms.tm_hour, tms.tm_min, tms.tm_sec, ms,
              current->argv[0], current->pid );
      for (ptr=temp; *ptr; ++ptr)  diagOut_(*ptr);
    }
    onNewLine = blankLine = false;
  }

//--- if end of line char, prefix it with a CR and set the flag saying ---
//--- we need to start a new line ---
  if (ch != '\n')  {
    diagOut_(ch);  onNewLine = blankLine = false;  }
  else
    if (!blankLine)  {
      if (!onNewLine)  diagOut_('\r');
      diagOut_('\n');
      if (onNewLine)  blankLine = true;  else   onNewLine = true;
    }

  lastTime = curTime;  lastApplicTicks = applicTicks;

  KUNLOCK();

  return 0;
}


//-----------------------------------------------------------------------------
//  Handles input for FILE *stdin
//-----------------------------------------------------------------------------
int _nearcall rcvChar(void *userData)
{
  char  ch;


  fflush(stdout);

  if (!FD[MyTaskIdx][0])  return 0;

  read(0, &ch, 1);  putchar(ch);

//finish - handle cooked vs raw line discipline
  if (ch == 0x0d)  { ch = 0x0a;  putchar(ch);  }

//finish - do proper signal handling
  if (ch == 0x03)  { printf("\r\n");  fflush(stdout);  Exit(-1); }

  return (int) ch;
}

//-----------------------------------------------------------------------------
//  The output handler for the FILE * stdout.
//
//  This function is called whenever one of the stdio stream oriented output
//  functions is called for the FILE *stdout (e.g. fprintf, fputs, fwrite,
//  fputc, putc, vfprintf) and the equivalent ones that assume FILE *stdout
//  printf, puts, putchar)
//
//  By default, writes to FILE *stdout get buffered and eventually written to
//  fd_1.  And while DEBROS currently doesn't support freopen(), stdout CAN
//  be redirected by calling redirectStdio() which sets a flag that causes
//  this function to redirect the output to FILE *DiagLog.
//-----------------------------------------------------------------------------
int _nearcall xmitChar(uchar ch, void *userData)
{
  FDEV far *stdoutDev = FD[userPix][1];

//--- If the process has no fd_1, or the flag indicating that data written ---
//--- to stdout should be copied to the diagnostic log, then write the char ---
//--- to the diagnostic log.  NOTE: logStdout must be set if redirStdio is. ---
  if (!stdoutDev OR current->logStdout)  {
    diagOut(ch, userData);
    if (!stdoutDev OR current->redirStdio)  return 0;  // don't use fd_1 for stdout
  }

//--- add char to the stdout buffer for the current process ---
  *current->outPtr++ = ch;  ++current->outCount;

//--- if end of a "line" or the buffer is full, then flush it now ---
  if ((ch == '\n') OR (current->outCount == StdOutBufSize))  fflush(stdout);

  return 0;
}


//-----------------------------------------------------------------------------
//  Return the device funcs for the "file system" corresponding to a file name
//
//  Basically a simple stand-in for /dev files their major device #
//
//  NOTE:  This does not check to see if the file actually exists.  It mearly
//         determines which driver to use.
//-----------------------------------------------------------------------------
const FileSysFuncs *drvFuncs(const char *fileName)
{
  uint  u;

  for (u=1; u<NumPrefixes; ++u)
    if (strncmp(fstab[u].prefix, fileName, strlen(fstab[u].prefix)) == 0)
      return fstab[u].fsFuncs;

  return fstab[0].fsFuncs;
}


//-----------------------------------------------------------------------------
//  set the istty flag for the specified file descriptor
//-----------------------------------------------------------------------------
void setIstty(fd_t fd, bool istty)
{
  if ((uint)fd >= FD_SETSIZE) return;

printf("pid: %d: setIstty(%d, %d)\r\n", getpid(), fd, istty);  //tdebug
  FD[MyTaskIdx][fd]->istty = istty;
}

//-----------------------------------------------------------------------------
//  return 1 if fd is a valid descriptor for a terminal device
//-----------------------------------------------------------------------------
int isatty(fd_t fd)
{
  if ((uint)fd >= FD_SETSIZE) return 0;

  return (FD[MyTaskIdx][fd]->istty ? 1 : 0);
}


//-----------------------------------------------------------------------------
//  return first unused file descriptor for specified pid that is >= the
//  specified minimum fd #
//  returns -1 if no unused descriptors available
//
//  NOTE that the structure being used is specific to the task that called,
//  and the structure itself is not altered, so locking is not required to
//  protect against corruption (although any task that looks at ANOTHER task's
//  structures should use locking to make sure the structures don't change
//  while it is looking at them).
//-----------------------------------------------------------------------------
int nextfd(uint pix, fd_t minFD)
{
  fd_t  fd;
  FDEV far  **fdList;

  fdList = FD[pix];

//--- see if there is an unused descriptor for the calling process ---
  for (fd=minFD; fd<FD_SETSIZE; ++fd)  if (! fdList[fd])  break;

  return (fd < FD_SETSIZE) ? fd : -1;
}

//-----------------------------------------------------------------------------
//  Returns a non-zero value if the provided pointer refers to an in-use
//  global file descriptor.  Otherwise, it returns 0.
//-----------------------------------------------------------------------------
bool  _validFDEV_(FDEV far *file)
{
  uint  gfd;
  int  openCount;


  gfd = (uint)(file - fileTbl);

  if (gfd >= MaxOpenFiles) return false;

  if (file != fileTbl + gfd)  return false;

  _EnterCritical_();  openCount = file->openCount;  _ExitCritical_();

  if (openCount < 1)  return false;

  return true;
}

//-----------------------------------------------------------------------------
//  Return a user-space (task-specific) descriptor for the specified global
//  file table entry.
//
//  Normally called by individual device drivers
//
//  WARNING:  KLOCK() must be active before calling this function!
//-----------------------------------------------------------------------------
fd_t _addFileRef_(FDEV far *file)
{
  fd_t  fd;


  if (! KLOCKED())  { klockErr(__func__);  return -1; }

//--- Make sure the global file table pointer points to an in-use file ---
  if (! _validFDEV_(file))  { errno = EBADF;  return -1; }

//--- see if there is an unused descriptor for the calling process ---
  if ((fd = nextfd(userPix, 0)) < 0)  { ERRNO = EMFILE;  return -1; }

//--- point new user-space descriptor to global file table entry and ---
//--- increment the usage count of the global entry ---
  FD[(uint)userPix][(uint)fd] = file;
  ++file->openCount;

  return fd;
}


//-----------------------------------------------------------------------------
//  Allocate an entry in the global file table and a User-space File Descriptor
//  that points to it.
//
//  Normally called by individual device drivers
//
//  WARNING:  KLOCK() must be active before calling this function!
//-----------------------------------------------------------------------------
fd_t _allocFile_(fd_t minFD, const IODevFuncs *ioFuncs, void *ioDev)
{
  char  i;
  fd_t  fd;
  FDEV far  *file;
  FDEV far  **fdList;


  if (! KLOCKED())  { klockErr(__func__);  return -1; }

  fdList = FD[userPix];

//--- see if there is an unused descriptor for the calling process ---
  if ((fd = nextfd(userPix, minFD)) < 0)  { ERRNO = EMFILE;  return -1; }

//--- see if there is an unused entry in the global file table ---
  for (i=0; i<MaxOpenFiles; ++i)  if (fileTbl[i].openCount == 0)  break;

  if (i == MaxOpenFiles)  { ERRNO = ENFILE;  return -1; }

//--- point descriptor to global file table entry ---
  fdList[fd] = file = fileTbl + i;

//--- initialize global file table entry ---
  fmemset(file, 0, sizeof(FDEV));
  file->openCount = 1;
  file->ioFuncs = ioFuncs;
  file->ioDev = ioDev;
  file->readCount = file->writeCount = 0;
  file->dontCallRead = file->dontCallWrite = false;
  (file->ioFuncs->getQueues)(file->ioDev, &file->waitQueues);
  file->flags = O_RDWR;

  return fd;
}


//-----------------------------------------------------------------------------
//  Return a pointer to the entry in the global file table that a file
//  descriptor refers to.
//
//  Returns NULL if an error or the FD does not refer to an open file.
//  Sets errno if the fd is out of range.
//-----------------------------------------------------------------------------
FDEV far *_FileFromFD_(fd_t fd)
{
  if ((uint)fd >= FD_SETSIZE)  { ERRNO = EBADF;  return NULL; }

  return FD[(uint)userPix][fd];
}

//-----------------------------------------------------------------------------
//  Check the specified descriptor to see if it is valid and refers to an
//  open file.
//
//  Returns a pointer to the entry in the global file table if the descriptor
//  points to an open file.
//
//  Returns NULL if not a valid open file and sets errno.
//-----------------------------------------------------------------------------
FDEV far *_fdToOpenFile_(fd_t fd)
{
  FDEV far  *file;


  if ((fd < 0) OR (fd >= FD_SETSIZE))  { ERRNO = EBADF;  return NULL; }
  if (! (file = FD[userPix][fd]))  { ERRNO = EBADF;  return NULL; }

  if (! _validFDEV_(file))  { ERRNO = EBADF;  return NULL; }

  return file;
}

//-----------------------------------------------------------------------------
//  Swap everything in two file table entries EXCEPT the reference counts and
//  the fcntl-type (task-fd-specific) attributes.
//
//  WARNING:  KLOCK() must be in effect before calling this function!
//-----------------------------------------------------------------------------
int _File_Swap_(FDEV far *f1, FDEV far *f2)
{
  FDEV  origf1, origf2;


  if (! KLOCKED())  klockErr(__func__);

  origf1 = *f1;  origf2 = *f2;

  *f1 = origf2;
  f1->openCount = origf1.openCount;
  f1->flags = origf1.flags;

  *f2 = origf1;
  f2->openCount = origf2.openCount;
  f2->flags = origf2.flags;

  return 0;
}

//-----------------------------------------------------------------------------
//  Func for use by kernel-space functions to safely close a user-space
//  file descriptor
//-----------------------------------------------------------------------------
int _closeFD_(task_t *task, fd_t fd)
{
  FDEV far  *file;
  FDEV far  **fdPtr;
  const IODevFuncs  *ioFuncs;
  void  *ioDev;
  bool  doDevClose;


  if ((fd < 0) OR (fd >= FD_SETSIZE))  { ERRNO = EBADF;  return -1; }

  fdPtr = (FDEV far **)task->FDSet + fd;  // ptr to process' descriptor

  KLOCK();

//--- Validate descriptor and keep a pointer to it ---
  if (! (file = *fdPtr))  { KUNLOCK();  ERRNO = EBADF;  return -1; }

  if (ShowFileEvents)
    fprintf(DiagLog, "close fd %d for [%d](%s), openCounts: %d\n",
           fd, task->pid, task->argv[0], file->openCount);

//--- Free descriptor and decrement open count for referenced file. ---
//--- If no more references to the file, then prepare to close it ---
  *fdPtr = NULL;  doDevClose = false;
  if (file->openCount > 0)  doDevClose = (--file->openCount == 0);

//--- We can't call the driver's close func with KLOCK in place because ---
//--- the close func might need to block, but releasing the lock before ---
//--- calling the driver means the global descriptor could get snatched ---
//--- up by another task so we need to save refs to the driver and the ---
//--- device driver's structure BEFORE we release the lock ---
  if (doDevClose)  { ioFuncs = file->ioFuncs;  ioDev = file->ioDev; }

  KUNLOCK();

  if (doDevClose)  ioFuncs->close(ioDev);

  return 0;
}


//-----------------------------------------------------------------------------
//  Open a file/device
//-----------------------------------------------------------------------------
fd_t open(const char *path, int flags, int mode)
{
  int  fd;
  PID  pid;
  uint  pix;
  FDEV far  *file;
  const FileSysFuncs  *fsf;


//--- check for a request to open the terminal device ---
  if (StrEq(path, "/dev/tty"))  {
    for (pid=getpid(); pid>=0; pid=tasks[pix].ppid)  {
      KLOCK();
      pix = TaskIdx(pid);  if (!tasks[pix].inUse)  { KUNLOCK();  continue; }
      if ((file = FD[pix][0]) == NULL)  { KUNLOCK();  continue; }
      if (file->istty)  {
        if ((fd = nextfd(pix, 0)) < 0)  {
          KUNLOCK();  ERRNO = EMFILE;  return -1; }
        FD[MyTaskIdx][fd] = file;  ++file->openCount;
        KUNLOCK();  return fd;
      }
      KUNLOCK();
    }
  }

  if (ShowFileEvents)
    fprintf(DiagLog, "open('%.*s', 0x%04x, 0%03o)\n",
            PATH_MAX, path, flags, mode);


  fsf = drvFuncs(path);  if (!fsf)  { ERRNO = ENOENT;  return -1; }


  fd = fsf->open(path, flags, mode);

  if (fd < 0)  {
    if (ShowFileEvents)
      fprintf(DiagLog, "failed open [%.32s] using [%.32s]\r\n", path, fsf->name);
    return -1;
  }

//--- save parameter values passed in open call ---
//NOTE:  Most of the FDEV struct is initialized in _allocFile_(), which is
//       called by the open func in each driver to get an unused FDEV element
  KLOCK();
  file = _fdToOpenFile_(fd);
  file->flags = flags;  file->mode = mode;
  KUNLOCK();

/*
fprintf(DiagLog, "%.64s: %d: readQueue:[%04lx]  writeQueue:[%04lx]\n",  //tdebug
        path, fd,
        (ulong)file->waitQueues.readQueue,
        (ulong)file->waitQueues.writeQueue );
*/

  return fd;
}



//-----------------------------------------------------------------------------
//  Close the specified user-space file descriptor
//-----------------------------------------------------------------------------
int close(fd_t fd)
{
  if (ShowFileEvents)  fprintf(DiagLog, "close(%d)\n", fd);

  return  _closeFD_(current, fd);
}



//-----------------------------------------------------------------------------
//  Attempt to read specified # of bytes from a device
//
//  maxIdleMS determines blocking mode as follows:
//    < 0: indefinite timeout (i.e. block until done or error)
//      0: non-blocking:  read as much as possible and return
//    > 0: max # of ms to wait with no progress
//
//  Note that a positive value does NOT place an upper limit on how long this
//  function will wait to read the requested # of bytes, but rather the max
//  time it will wait while NO PROGRESS is being made at reading the bytes.
//
//  Note that if non blocking mode is set for the file descriptor then the
//  maxIdleMS value is ignored (treated as if it were 0)
//
//  If a timeout > 0 expires, errno is set to ETIME (regardless of what the
//  return value is, which could be > 0 if some bytes were read).
//-----------------------------------------------------------------------------
int readFDEV(FDEV far *file, void *buf, int numBytes, long maxIdleMS)
{
  int  stat, pid, bytesRead;
  bool  nonBlocking;
  char  *inPtr;
  ulong  maxIdleTicks = 0;
  WaitQueue  *readQueue;
  WaitQueueNode  *wqNode;


  if (ShowRWCalls)
    fprintf(DiagLog, "readFDEV([%05lX], 0x%04X, %d)\r\n",
            (ulong)file, (uint)buf, numBytes);

  if (!file)  {
    if (ShowRWCalls)  fprintf(DiagLog, "  read returned: %d\r\n", -1);
    return -1;
  }

  if (numBytes <= 0)  return 0;

  if (!buf)  { errno = EINVAL;  return -1; }

  pid = getpid();

//--- if driver won't return any new data, just return status from last ---
//--- call to driver ---
  if (file->dontCallRead)  {  errno = file->rdErr;  return file->rdStat; }

  if (! ((file->flags & O_RDONLY) OR (file->flags & O_RDWR)) )  {
    errno = EBADF;  return -1; }

  KLOCK();

    //finish: all the relavant "flags" values...
    //finish: lots of other details - see recv()/revfrom() man page

  if (file->flags & O_NONBLOCK)  maxIdleMS = 0;

  nonBlocking = (maxIdleMS == 0);

  if (maxIdleMS > 0)  maxIdleTicks = msToTicks((ulong)maxIdleMS);


//--- is another process in the middle of a read on this file/device? ---
  while (file->rdInProgress)  if (file->rdInProgress != pid)
    if (_activeTaskIdx_(file->rdInProgress) >= 0)  {
      if (nonBlocking)  { KUNLOCK();  ERRNO = EAGAIN;  return -1; }
      KUNLOCK();  sleepMS(5);  KLOCK();
    }

  if (file->eof)  {
    KUNLOCK();
    if (ShowRWCalls)  fprintf(DiagLog, "  read returned: %d\r\n", 0);
    return 0;
  }

  file->rdInProgress = pid;

  KUNLOCK();



  readQueue = file->waitQueues.readQueue;

//finish:  don't have driver set global errno?  Or push/pop global errno?

//--- If doing a non-blocking read, or the driver directly below this level ---
//--- does not provide wait queues for use by the fcntl layer, then simply ---
//--- return the result from the driver (if doing a blocking read, the ---
//--- blocking needs to be handled by the driver or whatever the driver is ---
//--- calling to do the work) ---
  if (nonBlocking OR !readQueue)  {
    stat = (file->ioFuncs->read)(file->ioDev, buf, numBytes, nonBlocking);
    if (stat > 0)  {
      _EnterCritical_();  file->readCount += (ulong)stat;  _ExitCritical_(); }
    else  if ((stat == 0) OR ((stat < 0) AND (errno != EAGAIN)))  {
      file->dontCallRead = true;
      file->rdStat = stat;  file->rdErr = errno;
    }
    if (ShowRWCalls)  fprintf(DiagLog, "  read returned: %d\r\n", stat);
    _EnterCritical_();  file->rdInProgress = 0;  _ExitCritical_();
    return stat;
  }


//--- do blocking read using queue specified by the driver ---
  bytesRead = 0;  inPtr = buf;

//  if (! list_empty(&WQNode.task_list))  {
//    fprintf(DiagLog, "Already in wait queue!\n");  //tdebug
//    errno = EDEADLK;  file->rdInProgress = false;  return -1; }

  if (!(wqNode = alloc_WaitQueueNode()))  return -1;

  for (;;)  {
    if (readQueue == &sleepQueue)  {
      current->sleepTicks = 1;  current->sleepStarted = GetMS(); }

    prepare_to_wait(readQueue, wqNode, TASK_UNINTERRUPTIBLE);

    if (errno == EAGAIN)  errno = 0;  //tdebug

    stat = (file->ioFuncs->read)(file->ioDev, inPtr, numBytes-bytesRead, true);

    if (stat > 0)  {
      _EnterCritical_();  file->readCount += (ulong)stat;  _ExitCritical_();
      inPtr += (uint)stat;  bytesRead += stat;
    }
    else  if ((stat == 0) OR ((stat == -1) AND (errno != EAGAIN)))  {
      file->dontCallRead = true;
      file->rdStat = stat;  file->rdErr = errno;
      break;
    }

    if (bytesRead >= numBytes)  break;

    if (maxIdleMS < 0)  // indefinite wait for possibility to read more bytes
      schedule();
    else  // block but with a limit on idle (no progress) time
      if (! schedule_timeout(maxIdleTicks))  { errno = ETIME;  break; }
  }

  finish_wait(readQueue, wqNode);  // set state to RUNNING & free the node

  file->rdInProgress = 0;

  if (bytesRead)  stat = bytesRead;

  if (ShowRWCalls)  fprintf(DiagLog, "  read returned: %d\r\n", stat);

  return stat;
}

//-----------------------------------------------------------------------------
//  Attempt to read the requested # of bytes from a device.
//-----------------------------------------------------------------------------
int read(fd_t fd, void *buf, int numBytes)
{
  return readFDEV(_fdToOpenFile_(fd), buf, numBytes, -1);
}


//-----------------------------------------------------------------------------
//  Attempt to write specified # of bytes to a device.
//
//  maxIdleMS determines blocking mode as follows:
//    < 0: indefinite timeout (i.e. block until done or error)
//      0: non-blocking:  write as much as possible and return
//    > 0: max # of ms to wait with no progress
//
//  Note that a positive value does NOT place an upper limit on how long this
//  function will wait to write the specified # of bytes, but rather the max
//  time it will wait while NO PROGRESS is being made at writing the bytes.
//
//  Note that if non blocking mode is set for the file descriptor then the
//  maxIdleMS value is ignored (treated as if it were 0)
//
//  If a timeout > 0 expires, errno is set to ETIME (regardless of what the
//  return value is, which could be > 0 if some bytes were written).
//----------------------------------------------------------------------------
int writeFDEV(FDEV far *file, void *buf, int numBytes, long maxIdleMS)
{
  int  stat, pid, bytesSent;
  bool  nonBlocking;
  char  *outPtr;
  ulong  maxIdleTicks = 0;
  WaitQueue  *writeQueue;
  WaitQueueNode  *wqNode;


  if (numBytes <= 0)  return 0;

  if (!buf)  { errno = EINVAL;  return -1; }

  pid = getpid();

  if (ShowRWCalls)
    fprintf(DiagLog, "writeFDEV([%05lX], 0x%04X, %d, %ld)\r\n",
            (ulong)file, buf, numBytes, maxIdleMS);

  if (!file)  {
    if (ShowRWCalls)  fprintf(DiagLog, "  writeFDEV returned: %d\r\n", -1);
    return -1;
  }

//--- if driver won't return any new data, just return status from last ---
//--- call to driver ---
  if (file->dontCallWrite)  {  errno = file->wrErr;  return file->wrStat; }

  if (! ((file->flags & O_WRONLY) OR (file->flags & O_RDWR)) )  {
    errno = EBADF;  return -1; }


  KLOCK();

  //finish: all the relavant "flags" values...
  //finish: lots of other details - see recv()/revfrom() man page

  if (file->flags & O_NONBLOCK)  maxIdleMS = 0;

  nonBlocking = (!maxIdleMS);

  if (maxIdleMS > 0)  maxIdleTicks = msToTicks((ulong)maxIdleMS);


//--- is another process in the middle of a write on this file/device? ---
  while (file->wrInProgress)  if (file->wrInProgress != pid)
    if (_activeTaskIdx_(file->wrInProgress) >= 0)  {
      if (nonBlocking)  { KUNLOCK();  ERRNO = EAGAIN;  return -1; }
      KUNLOCK();  sleepMS(5);  KLOCK();
    }

  if (file->eof)  {
    KUNLOCK();
    if (ShowRWCalls)  fprintf(DiagLog, "  writeFDEV returned: %d\r\n", 0);
    return 0;
  }

  file->wrInProgress = pid;

  KUNLOCK();


  writeQueue = file->waitQueues.writeQueue;

//finish:  don't have driver set global errno?  Or push/pop global errno?

//--- if doing a non-blocking write, or the driver directly below this ---
//--- level does not use wait queues, then simply return the result from ---
//--- the driver (if doing a blocking write, the blocking needs to be ---
//--- handled by whatever the driver is calling to do the work) ---
  if (nonBlocking OR !writeQueue)  {
    stat = (file->ioFuncs->write)(file->ioDev, buf, numBytes, nonBlocking);
    if ((stat == 0) OR ((stat < 0) AND (errno != EAGAIN)))  {
      file->dontCallWrite = true;
      file->wrStat = stat;  file->wrErr = errno;
    }
    if (stat > 0)  {
      _EnterCritical_();  file->writeCount += (ulong)stat;  _ExitCritical_(); }
    if (ShowRWCalls)  fprintf(DiagLog, "  writeFDEV returned: %d\r\n", stat);
    _EnterCritical_();  file->wrInProgress = 0;  _ExitCritical_();
    return stat;
  }

//--- do blocking write using queue specified by the driver ---
  bytesSent = 0;  outPtr = buf;

//  if (! list_empty(&WQNode.queueList))  {
//    fprintf(DiagLog, "Already in wait queue!\n");  //tdebug
//    errno = EDEADLK;  file->wrInProgress = false;  return -1; }


  if (!(wqNode = alloc_WaitQueueNode()))  return -1;

  for (;;)  {
    if (writeQueue == &sleepQueue)  {
      current->sleepTicks = 1;  current->sleepStarted = GetMS(); }

    prepare_to_wait(writeQueue, wqNode, TASK_UNINTERRUPTIBLE);

    if (errno == EAGAIN)  errno = 0;  //tdebug

    stat = (file->ioFuncs->write)
                (file->ioDev, outPtr, numBytes-bytesSent, true);

    if (stat > 0)  {
      _EnterCritical_();  file->writeCount += (ulong)stat;  _ExitCritical_();
      outPtr += (uint)stat;  bytesSent += stat;
    }
    else  {
      if ((stat == 0) OR ((stat == -1) AND (errno != EAGAIN)))  {
        file->dontCallWrite = true;
        file->wrStat = stat;  file->wrErr = errno;
        break;
      }
    }

    if (bytesSent >= numBytes)  break;

    if (maxIdleMS < 0)  // indefinite wait for possibility to write more bytes
      schedule();
    else  // block but with a limit on idle (no progress) time
      if (! schedule_timeout(maxIdleTicks))  {
        stat = -1;  errno = ETIME;  break; }
  }

  finish_wait(writeQueue, wqNode);  // set state to RUNNING & free the node

  file->wrInProgress = 0;

  if (bytesSent)  stat = bytesSent;

  if (ShowRWCalls)  fprintf(DiagLog, "  writeFDEV returned: %d\r\n", stat);

  return stat;
}


//-----------------------------------------------------------------------------
//  Attempt to write specified # of bytes to a device.
//-----------------------------------------------------------------------------
int write(fd_t fd, void *buf, int numBytes)
{
  return writeFDEV(_fdToOpenFile_(fd), buf, numBytes, -1);
}


//-----------------------------------------------------------------------------
//  Reposition a file offset
//-----------------------------------------------------------------------------
off_t lseek(fd_t fd, off_t offset, int whence)
{
  off_t  stat;
  FDEV far  *file;


  if (! (file = _fdToOpenFile_(fd)))  return -1;

  stat = (file->ioFuncs->lseek)(file->ioDev, offset, whence);

  if (stat != (off_t)-1)  file->dontCallRead = false;

  return stat;
}


//-----------------------------------------------------------------------------
//  return information about an open file descriptor
//-----------------------------------------------------------------------------
int fstat(int fd, struct stat *fsbuf)
{
  FDEV far  *file;
  int  stat;


  if (! fsbuf)  { ERRNO = EFAULT;  return -1; }

  memset(fsbuf, 0, sizeof(struct stat));

  if (! (file = _fdToOpenFile_(fd)))  return -1;

  stat = (file->ioFuncs->fstat)(file->ioDev, fsbuf);

  return stat;
}


//-----------------------------------------------------------------------------
//  return information about specified file name
//-----------------------------------------------------------------------------
int stat(const char *fileName, struct stat *fsbuf)
{
  fd_t  fd;
  int  status;


  if ((fd = open(fileName, O_RDONLY, 0)) < 0)  return -1;

  status = fstat(fd, fsbuf);  close(fd);

  return status;
}


//-----------------------------------------------------------------------------
//  delete a name and possibly the file it refers to
//-----------------------------------------------------------------------------
int unlink(char const *path)
{
  const FileSysFuncs *fsf;

  fsf = drvFuncs(path);  if (!fsf)  { ERRNO = ENOENT;  return -1; }

  return fsf->unlink((char *)path);
}

//-----------------------------------------------------------------------------
//  User-space func to rename a file
//-----------------------------------------------------------------------------
int rename(const char *oldPath, const char *newPath)
{
  const FileSysFuncs  *fsf;

  fsf = drvFuncs(oldPath);  if (!fsf)  { ERRNO = ENOENT;  return -1; }
  if (fsf != drvFuncs(newPath))  { ERRNO = EXDEV;  return -1; }

  return fsf->rename(oldPath, newPath);
}


//-----------------------------------------------------------------------------
//  Modify date/time stamps on a file
//-----------------------------------------------------------------------------
int utime(const char *fileName, const struct utimbuf *buf)
{
//finish - handle other names besides just flash file names

  return  ffile_utime(fileName, buf);
}

//-----------------------------------------------------------------------------
//  Set or read options for a file descriptor
//-----------------------------------------------------------------------------
int fcntl(fd_t fd, int cmd, int arg)
{
  int  stat;
  FDEV far  *file;


  if (! (file = _fdToOpenFile_(fd)))  return -1;

  KLOCK();

  switch (cmd)  {
    case F_GETFL:  // get flags
      stat = file->flags;  KUNLOCK();  return stat;

    case F_SETFL:  // set flags
      file->flags = arg;  KUNLOCK();  return 0;
  }

  KUNLOCK();  ERRNO = EINVAL;  return -1;  // unsupported cmd value
}


//-----------------------------------------------------------------------------
//  Set or read options for a file descriptor
//-----------------------------------------------------------------------------
int ioctl(int fd, int request, void *arg)
{
  FDEV far  *file;


  if (! (file = _fdToOpenFile_(fd)))  return -1;

  return (file->ioFuncs->ioctl)(file->ioDev, request, arg);
}


//-----------------------------------------------------------------------------
//  Duplicate a file descriptor
//
//  If fromFD does not refer to an open file, then an error is returned.
//  If toFD refers to an open file different than fromFD, then toFD is first
//  closed.
//  Returns fromFD on succcess, -1 if an error (and errno is set)
//-----------------------------------------------------------------------------
fd_t dup2(fd_t fromFD, fd_t toFD)
{
  FDEV far  *fromFile, *toFile;
  FDEV far  **myFDs;


//--- make sure fromFD and toFD are valid descriptor #'s ---
  if ((fromFD < 0) OR (fromFD >= FD_SETSIZE)
     OR (toFD < 0) OR (toFD >= FD_SETSIZE))  { ERRNO = EBADF;  return -1; }

  KLOCK();

  myFDs = FD[userPix];

  fromFile = myFDs[fromFD];  toFile = myFDs[toFD];

//--- make sure fromFD points to a an entry for an open device ---
  if (! fromFile)  { KUNLOCK();  ERRNO = EBADF;  return -1; }
  if (fromFile->openCount < 1)  { KUNLOCK();  ERRNO = EBADF;  return -1; }

//--- if toFD already points to the same file as fromFD, we are done ---
  if (toFile == fromFile)   { KUNLOCK();  return toFD; }

//--- if toFD points to an open file, close it first ---
  if (toFile)  _closeFD_(current, toFD);

//--- set toFD to point to the same file as fromFD and increment the ---
//--- file's openCount ---
  myFDs[toFD] = myFDs[fromFD];  ++fromFile->openCount;

  KUNLOCK();  return toFD;
}


//=============================================================================
//                       conventional stdio type functions
//=============================================================================

//-----------------------------------------------------------------------------
//  User-space func to flush the output buffer for a FILE stream
//-----------------------------------------------------------------------------
int fflush(FILE *stream)
{
  int  count, saveFlags;
  FDEV far *file;
  char  buf[StdOutBufSize];


  if (stream != _stdio)  { ERRNO = EBADF;  return EOF; }

  count = current->outCount;

  if (count < 1)  return 0;

  if (! (file = _fdToOpenFile_(1)))  return -1;

  fmemcpy(buf, outBuf[(uint)userPix], count);

  saveFlags = file->flags;  file->flags &= ~O_NONBLOCK;
  write_(1, buf, count, -1);
  file->flags = saveFlags;

  current->outCount = 0;  current->outPtr = outBuf[(uint)userPix];

  return 0;
}



//-----------------------------------------------------------------------------

