//----- signal.c ----- 08/03/09 -----(04/30/07)-----

//=============================================================================
//  sigsetops(3) functions that allow the manipulation of POSIX signal sets.
//=============================================================================
// NOTES
//
//   Signal handling in DEBROS is currently limited to setting which signals
//   are blocked (or not).  Setting a signal handler is not yet implemented,
//   so the only real use for this is to allow a process to block a signal
//   giving it the chance to periodically check for pending signals and then
//   deal with them within the normal program flow.
//
//   For example, simply killing all processes to do a shutdown doesn't give
//   processes a chance to clean up after themselves in an orderly manner.
//
//   To this end, each process will receive a terminate signal (SIGTERM)
//   several seconds before actually being killed.  This will cause all
//   blocked processes to be resumed with errno set to EINTR (interrupted
//   system call).
//
//   To exit cleanly, processes must check for errors and call sigpending()
//   to see if they have received a SIGTERM signal and then do what is needed
//   to exit cleanly on their own.  If they fail to do so within a few seconds
//   after receiving the SIGTERM signal, they will be killed and any files
//   they had opened will be closed.
//
//   Example:
//    sigset_t  sigSet;
//
//    // block Terminate signal so we can shut down cleanly if needed
//    sigSet = sigmask(SIGTERM);  sigprocmask(SIG_BLOCK, &sigSet, NULL);
//
//    for (;;)  {
//      ...
//      // check for pending Terminate signal
//      if (sigpending(&sigSet) == 0)  if (sigismember(&sigSet, SIGTERM))  {
//        ...;
//        Exit(0);
//      }
//    }
//
//=============================================================================
//
//
//  2009-08-03 - Mark Davis
//               - sigprocmask now ignores SIGKILL and SIGSTOP bits rather
//                 than returning -1 if either one is set.
//
//  2007-04-30 - Mark Davis
//               - original version
//
//=============================================================================

#include <common.h>

#include <linux/sched.h>

#include <sys/types.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>

#include <signal.h>


#define UnblockableSet  ( sigmask(SIGKILL) | sigmask(SIGSTOP) )

//-----------------------------------------------------------------------------
//  initialize the signal set given by set to empty, with all signals excluded
//  from the set.
//-----------------------------------------------------------------------------
int sigemptyset(sigset_t *set)
{
  *set = 0;  return 0;
}

//-----------------------------------------------------------------------------
//  initialize set to full, including all signals
//-----------------------------------------------------------------------------
int sigfillset(sigset_t *set)
{
  *set = 0xfffffffUL;  return 0;
}

//-----------------------------------------------------------------------------
//  add signal signum to a set of signals
//-----------------------------------------------------------------------------
int sigaddset(sigset_t *set, int signum)
{
  if ((signum < 0) OR (signum > 31))  return -1;

  *set |= sigmask(signum);  return 0;
}

//-----------------------------------------------------------------------------
//  delete signal signum from a set of signals
//-----------------------------------------------------------------------------
int sigdelset(sigset_t *set, int signum)
{
  if ((signum < 0) OR (signum > 31))  return -1;

  *set &= ~sigmask(signum);  return 0;
}

//-----------------------------------------------------------------------------
//  test whether signum is a member of a signal set
//-----------------------------------------------------------------------------
int sigismember(const sigset_t *set, int signum)
{
  if ((signum < 0) OR (signum > 31))  return -1;

  if (*set & sigmask(signum))  return 1;

  return 0;
}


//-----------------------------------------------------------------------------
//  install a new signal handler for the signal with number signum
//-----------------------------------------------------------------------------
sighandler_t signal(int signum, sighandler_t handler)
{
  //finish

  return (sighandler_t) 0;
}

//-----------------------------------------------------------------------------
//  change the action taken by a process on receipt of a specific signal
//-----------------------------------------------------------------------------
int sigaction(int signum, const struct sigaction *act,
              struct sigaction *oldact)
{
//finish

  return -1;
}

//-----------------------------------------------------------------------------
//  change the set of currently blocked signals
//-----------------------------------------------------------------------------
int sigprocmask(int how, const sigset_t *newSet, sigset_t *oldSet)
{
  KLOCK();

  if (oldSet)  *oldSet = current->sig_blocked;

  switch (how)  {

    case SIG_BLOCK:
      current->sig_blocked |= (*newSet & ~UnblockableSet);
      break;

    case SIG_UNBLOCK:
      current->sig_blocked &= ~(*newSet);
      break;

    default:
      current->sig_blocked = *newSet;
  }

  KUNLOCK();

  return 0;
}

//-----------------------------------------------------------------------------
//  get the mask/set of pending signals
//-----------------------------------------------------------------------------
int sigpending(sigset_t *set)
{
  *set = current->sig_pending;

  return 0;
}



//-----------------------------------------------------------------------------
//  temporarily replace the signal mask for the process with that given by
//  mask and then suspend the process until a signal is received
//-----------------------------------------------------------------------------
int sigsuspend(const sigset_t *mask)
{
//finish

  return -1;
}



//-----------------------------------------------------------------------------

