//----- unistd.c ----- 01/11/13 -----(03/05/06)----

//-----------------------------------------------------------------------------
//  Functions to mimic unix kernel functions for managing processes
//-----------------------------------------------------------------------------
// Update Log (most recent first)
//
//  Entries marked with (*) indicate ones that improve the stability and/or
//  reliability of the code.
//
//
//  2013-01-11 - Mark Davis
//               - Added support for a patched version of the RDLM reads and
//                 saves the value of GCSR (the register with the resetCode
//                 bits).  Prior to this, there was no way to get this value
//                 because the version of the RDLM works with DEBROS also
//                 reads this value before DEBROS starts (which clears it so
//                 it is gone before DEBROS or the cstart module can read it).
//
//  2012-12-17 - Mark Davis
//               - Began adding code to support a backup approach to the
//                 softReset concept:  To attempt to restore the state of the
//                 controller ASAP if it did an unexpected CPU reset (not
//                 fully implemented yet due to issues with how it could affect
//                 various appliations).
//
//  2012-12-12 - Mark Davis
//               - postMortem() now includes the # of ms since the kernel and
//                 applic tasks have called resetWDT(), and the value of the
//                 flags: doCleanShutdown, useSoftReset, allowSoftReset
//
//  2012-09-13 - Mark Davis (*)
//               - Fixed computeCRC() so it doesn't log ALL tested portions of
//                 memory as being corrupted after the first detected
//                 corruption.
//
//  2012-09-06 - Mark Davis
//               - Improved the efficiency of resetWDT() a bit.
//
//  2012-06-11 - Mark Davis
//               - Moved updating of _TimerMilliSeconds[1] and _TimerSeconds,
//                 and the check for the scheduler tick (a SchedMS millisecond
//                 long interval) to _TimerBISR() to avoid the overhead of
//                 calling timer_interrupt() every millisecond.
//               - Renamed timer_interrupt() to sched_tick().
//               - current->lastRunning is no longer set within the periodic
//                 timer interrupt (now sched_tick()) but at the beginning of
//                 schedule() (which has to be called to change which task
//                 has the CPU, so that is the only place it is needed).
//
//  2012-06-07 - Mark Davis
//               - Added schedCallsPrevSec (count of # of times schedule() was
//                 called in the last second).
//
//  2012-03-29 - Mark Davis
//               - The interrupt functions that service fault conditions now
//                 save the AF' regsiter on the stack.  Prior to this, the
//                 various functions called between when the fault occured
//                 and when the dump is done could have used the AF' register
//                 without saving and restoring its value, resulting in an
//                 incorrect value in the dump.
//
//  2012-03-21 - Mark Davis
//               - Finished changes to new logic to that checks for an
//                 unreasonable hdwResetTime value and attempts to correct it.
//                 Prior to this, if the RTC did not have a valid value at
//                 startup, this value would not be updated when the RTC was
//                 set, which the result that, following a soft reset, the
//                 counter for the # of seconds since the last hard reset could
//                 be wildly off (which could affect application logic which
//                 uses time-since-hard-reset extensively for managing timing.
//               - The IP (Interrupt Priority) register is now cleared a bit
//                 later so there are no values left over from the startup
//                 proceedure.
//
//  2012-03-15 - Mark Davis
//               - The hdwResetTime and rebootTime in the diagLogInfo structure
//                 are now corrected whenever the RTC is updated.
//
//  2012-03-07 - Mark Davis
//               - postMortem() now includes the IP (Interrupt Priority)
//                 register value for each task.
//               - IP reg now zeroed out at startup to insure there is no
//                 value left in it after a soft reboot.
//               - Added rtosVers and invInfoKey to DiagLogInfo struct
//
//  2012-03-02 - Mark Davis
//               - KUNLOCK() now increases/decreases preempt_count by 8 so
//                 other code (primarily postMortem() and ps cmd) can
//                 distinguish between a normal preemption due to a schedule
//                 timer tick and a delayed one from KUNLOCK().
//
//  2012-03-01 - Mark Davis
//               - More fixes to postMortem().  Include changes needed because
//                 timer_interrupt() is now a far function, and fixing some
//                 bugs that caused it to grab the wrong XPC value from the
//                 stack for preempted tasks.
//
//  2012-02-24 - Mark Davis
//               - tcp_lock and tcp_unlock pointers now refer directly to the
//                 KLOCK() and KUNLOCK() functions, although this is mostly a
//                 "backup" as lock macros in the Softools/Rabbit Stcpip/net.h
//                 header file have been modified to call KLOCK() and KUNLOCK()
//                 directly.
//
//  2012-02-15 - Mark Davis
//               - Made timer_interrupt() a far func to free up some root
//                 memory space.
//               - If sock_init() fails when starting up the primary boot image,
//                 it now sets the flag that tells the DLM to load the backup
//                 image.  Prior to this it would just keep running the DLM.
//
//  2012-01-03 - Mark Davis
//               - Added getPC() (returns Program Counter for cur task)
//               - postMortem() now shows the correct PC (Program Counter)
//                 value for preempted tasks.
//
//  2011-12-22 - Mark Davis (*)
//               - stackTrace() has been overhauled so it produces more
//                 accurate output.  While there are still lots of (unavoidable)
//                 false matches, it no longer misses the real matches.
//
//  2011-11-11 - Mark Davis (*)
//               - Missed the watchdog timer imminent-timeout in the last
//                 change.  This also does a softReset now.
//
//  2011-10-20 - Mark Davis (*)
//               - A soft rather than hard reset is now done whenever possible
//                 for any runtime exception.  This should minimize the
//                 disruption such errors can cause.  And since, to date,
//                 nearly all of these coincide with a memory corruption, it is
//                 unlikely that a software bug has anything to do with it.
//                 And on the chance that one WAS caused by a bug, then the
//                 crash-log will still be there and will provide the same
//                 information regardless of whether it was followed by a
//                 soft or a hard reset.
//
//  2011-03-29 - Mark Davis
//               - Changes to stop kernel task from blocking when changing
//                 the network configuration.
//
//  2011-03-11 - Mark Davis
//               - Added stack protection values for each task to the data
//                 shown by the showMemProtection() function.
//
//  2011-03-09 - Mark Davis
//               - Significant changes to how a stack trace is done to provide
//                 more accurate information in the crash-log.
//
//  2011-02-10 - Mark Davis
//               - The kernel task now forces a reboot 20 seconds after a
//                 shutdown was started regardless of the number of tasks
//                 still running.
//
//  2011-02-09 - Mark Davis
//               - Moved the code that gets the location and size of the
//                 memory segments to an earlier point in the startup sequence
//                 so shared, unnamed semaphores (ones with a static address
//                 accessable to all tasks) will work earlier.
//
//  2011-01-06 - Mark Davis
//               - Removed code that delayed rebooting when memory corruption
//                 is detected as there was never a case of a false positive
//                 the first time.  This reduces the chances of the corrupted
//                 memory will cause worse problems before the controller can
//                 do a soft restart to correct the problem.
//
//  2010-12-14 - Mark Davis
//               - Changed secTimerOffset to hdwResetTime and added rebootTime
//
//  2010-12-06 - Mark Davis (*)
//               - Added useSoftReset flag to support restarting the OS without
//                 resetting the CPU in the process.  This allows the OS to
//                 reload and restart itself without causing the entire CPU
//                 state to revert to its default reset values.  In particular,
//                 this makes it possible to reboot without affecting the
//                 state of the analog and digital outputs.
//
//  2010-12-02 - Mark Davis
//               - Began work on softReset() as a way to run the RDLM and
//                 reload and run the boot image.  This provides the means
//                 to do everything that happens for a CPU reset EXCEPT reset
//                 the CPU, which means the state of all the I/O bits will not
//                 change to their default modes and states.
//
//  2010-11-08 - Mark Davis
//               - On the off chance that a failure of a block of memory with
//                 unchanging values fails its CRC test due to a sporadic
//                 glitch in reading it, the failure must now occur twice in
//                 a row before the controller resorts to rebooting itself.
//                 Note that ANY CRC check failures are still recorded in the
//                 log.  While any glitch indicates a problem, the rebooting
//                 of a controller may be more disruptive than the chance that
//                 it happened at some time other than during the continuous
//                 background CRC test and caused a significant problem.  And
//                 while the new approach is statisically more liklely to allow
//                 such an occurance to cause ongoing problems, if the problem
//                 is real (i.e. memory HAS become corrupted), then the problem
//                 will still be detected (albiet a bit later than before) and
//                 the controller rebooted to correct it.
//
//  2010-09-22 - Mark Davis
//               - ramSize now set at beginning of main() using the value
//                 found by the CStart module for _xmemEnd (i.e. the end of
//                 unused extended memory BEFORE any of it has been allocated).
//
//  2010-08-12 - Mark Davis (*)
//               - Fixed a problem with the process that causes the controller
//                 to revert to the backup image when the primary version
//                 crashes unexpectedly.  Previous to this fix, the code made
//                 no distinction between a reboot that happened for no known
//                 reason and one that caused a post mortem dump and a
//                 controlled shutdown/reboot.  The result was that, depending
//                 on the timing, the backup image would mistakenly shut down
//                 and run the DLM.
//
//  2010-08-02 - Mark Davis
//               - Added memory protection values to data dumped by
//                 postMortem() when an error occurs.
//
//  2010-07-27 - Mark Davis (*)
//               - Since the RDLM uses a fixed location for its RUN_INFO
//                 structure, and that location is within the memory that
//                 xalloc() and bb_xalloc() allocate from, the value that
//                 the RDLM reads from there to determine what (if any)
//                 binary image should loaded and executed cannot updated
//                 except before it might be allocated for something else
//                 (which is necessary in case we crash during startup),
//                 or just before be do a controller reboot.
//                 To restore the feature that causes the fallback to the
//                 backup boot image or the RDLM in case of a problem, I
//                 now store the last boot time in the BB ram structure.
//                 Along with the flag that indicates an intentional
//                 reboot, this allows me to detect a crash shortly after
//                 startup and do a controlled fallback.
//               - The computeMemCRCs() function now displays all segments
//                 (not just the ones it computes a CRC for).
//
//  2010-07-22 - Mark Davis
//               - Added call to initSemaphores()
//
//  2010-07-19 - Mark Davis (*)
//               - Added check at startup to make sure the segments are in the
//                 correct order and have required alignments.
//               - The mask that controls which 64K segments are write
//                 protected is now reapplied after the task table is
//                 initialized (in case the mask is changed).
//
//  2010-07-15 - Mark Davis
//               - Changes to memory usage to allow ALL of the FARCODE and
//                 FARCONST segments to be write protected as well as all of
//                 the CODE segment and most of the CONST segment.
//
//  2010-07-08 - Mark Davis
//               - Simplified stack trace code.  The changes result in far more
//                 false "calls", but no longer misses actual calls.
//
//  2010-07-07 - Mark Davis (*)
//               - Extended write protection to unused extended memory blocks
//
//  2010-07-01 - Mark Davis
//               - The REBOOT() function argument now specifies if the RDLM
//                 should (re)load one of the boot images (rather than should
//                 it skip setting the value that tells the RDLM what to do
//                 before forcing a CPU reset).
//
//  2010-06-29 - Mark Davis
//               - Moved bb_xalloc() to modified xalloc.c module
//               - Modified log messages for memory allocations
//               - Ability to force the backup image to run using Ctl-B during
//                 startup now working again.
//
//  2010-06-24 - Mark Davis
//               - The idle task now continuously recomputes the CRC's for the
//                 memory segments that should never change (those with code
//                 and constant values) and does a dump/rebooot if any of them
//                 don't match the values computed at startup.
//
//  2010-06-09 - Mark Davis
//               - Basic stack and memory hdw protections implemented
//
//  2010-06-08 - Mark Davis
//               - Moved check for single-key console command to reboot using
//                 backup image or shutdown and run DLM to BEFORE the check
//                 that determines if main() restarted by some means other
//                 than the RDLM reloading the boot image.
//
//  2010-06-02 - Mark Davis
//               - Began adding support for limited memory protection mechanisms
//                 available on Rabbit 3000A processors.
//               - Added computeMemCRCs() and showIntVectors()
//
//  2010-05-28 - Mark Davis
//               - The size of the circular buffer used for the system log and
//                 the size of the far heap are both set dynamically based on
//                 the amount of free memory at startup.  This was done to free
//                 up more memory on the BL2200 devices which have only 512K of
//                 RAM (compared to the 768K on RCM3200 devices).
//
//  2010-05-27 - Mark Davis
//               - The flag that controls if syslog entries are copied to the
//                 console port (serial port A) is set when a shutdown is
//                 started so the console will show the process even if there
//                 was a login shell using it.
//
//  2010-05-14 - Mark Davis (*)
//               - timeoutTicks is now decremented only for tasks in a blocked
//                 state (i.e. INTERRUPTIBLE or UNINTERRUPTIBLE).  This leaves
//                 the timeoutTicks count intact when a task is woken up by
//                 a wait queue handler.  Prior to this schedule_timeout()
//                 would always return 0, which would cause problems (or at
//                 least wasted effort) when doing a blocking call with a
//                 timeout.
//               - The postMortem() dump of the interrupt table vectors now
//                 includes the name for each vector.
//
//  2010-05-13 - Mark Davis
//               - Replaced setBlockTimeout() with msToTicks()
//
//  2010-05-10 - Mark Davis
//               - Added recursiveDump flag and used it to avoid postMortem
//                 dumps from happening recursively.
//
//  2010-05-07 - Mark Davis (*)
//               - Changed code so DATASEG is always the same as the ROOT
//                 segment (0), rather than trying to follow the STACK segment.
//                 This is simpler, slightly faster. and avoids any possibility
//                 of there being a problem if the schedule() function is
//                 interrupted at just the wrong moment.
//
//  2010-05-03 - Mark Davis (*)
//               - Corrected bug in sleepMS() that could cause a watchdog
//                 timer reboot if sleepMS() was called with the scheduler
//                 locked.
//               - Also made sure that if the applic task calls sleepMS()
//                 that it will relinquish the CPU until the next scheduler
//                 tick.
//
//  2010-04-29 - Mark Davis
//               - Changes to accomodate more capable wait queue handling.
//
//  2010-04-23 - Mark Davis
//               - The WatchDog timer is now enabled first thing and left
//                 enabled except when initializing the network hardware and
//                 the IP stack (which takes longer than the max 2 seconds
//                 the WatchDog timer can be set for).  This helps avoid the
//                 possibility that the a problem during startup could cause
//                 the system to hang.
//
//  2010-04-22 - Mark Davis
//               - Now initializing the wq element rather than the waitq
//                 element for the kernel task (the unused waitq element has
//                 been removed).
//
//  2010-03-31 - Mark Davis
//               - Expanded on data provided by postMortem()
//
//  2010-03-22 - Mark Davis
//               - postMortem() now takes a string parameter so a description
//                 of the cause can be saved in the crash-log file.
//
//  2010-03-11 - Mark Davis (*)
//               - The startup code in main() now checks to see if this is the
//                 first time it has been run since the RDLM loaded it.  If
//                 not, then it forces a reboot using the watchdog timer.  This
//                 guards against the endless problems that can arise when
//                 a bug causes a jump to 0, causing my code to restart, but
//                 without the static variables having the proper values and
//                 (possibly) the code being corrupted.
//
//  2010-03-05 - Mark Davis (*)
//               - Added check for an unbalanced KUNLOCK() (i.e. that it
//                 hasn't been called more times than KLOCK()).  KUNLOCK()
//                 now logs a message if this happens.
//               - postMortem() now includes the state of the _disabled counter
//
//  2010-03-03 - Mark Davis (*)
//               - Completely redid how ISRs are protected from preemption.
//                 The 16-byte interrupt vector itself is now a short piece of
//                 code that disables the scheduler, calls the ISR, then
//                 reenables the scheduler and returns.  This drastically
//                 reduces the added overhead, the liklihood that an ISR will
//                 be preempted, and shrinks the possibility that it would be
//                 a problem if it should still happen to near 0.
//
//  2010-03-01 - Mark Davis
//               - Added check to see if the diagnostic/log info in battery
//                 backed RAM is already where it should be.  Previous to this
//                 change it always checked for a pre-shifted copy, but the
//                 newer version of the RDLM seems to have eliminated this.
//                 Without this change, if the board had been update from an
//                 older version of the RDLM to a newer one, the code could
//                 find the copy of the data structure in its older location
//                 which could be months out of date.
//
//  2010-02-27 - Mark Davis (*)
//               - Significant changes to timer_interrupt() to address
//                 recently recognized problem that can cause the stacks of
//                 one or more tasks to overflow.
//               - Added preempt_count to the info shown for each task in
//                 postMortem().
//
//  2010-02-26 - Mark Davis (*)
//               - The timer_interrupt() function now checks the scheduler
//                 disabled flag earlier and does not try to update anything
//                 related to the current task if it is disabled.  Before
//                 this change it always used the current pointer to update
//                 the lastRunning time and preempt_count.  If schedule()
//                 was interrupted after it had changed the current pointer
//                 but before it had actually switched tasks, then the values
//                 would be modified for the task about to be resumed, rather
//                 than for the one that was just preempted.
//
//  2010-02-18 - Mark Davis (*)
//               - The timer_interrupt() function now returns if it preempted
//                 the application task.
//
//  2010-02-13 - Mark Davis (*)
//               - The ISR shims now save and restore the Interrupt Priority
//                 (IP) register.
//               - The timer_interrupt() function now properly checks (and
//                 does not allow) recursive calls (for those once-in-a-great
//                 while instances where the timer interrupt is delayed
//                 enough to cause a recursive call).
//               - The updating of the time info in the diagLogInfo structure
//                 is now done only every SchedMS milliseconds, which reduces
//                 the overhead of the majority of the calls to this func.
//
//  2010-01-29 - Mark Davis (*)
//               - setBlockTimeout() now properly handles 0.  Prior to this,
//                 it ensured a timeout of at least once scheduler tick for
//                 any value it was passed (to avoid a 0 timeout for ms values
//                 that were not zero but smaller than one scheduler tick).
//                 This could explain lots of problems with socket operations
//                 that didn't make sense before.  Basically any task that
//                 called readBytes() or writeBytes() (directly or indirectly)
//                 would likely experience problems, especially due to calls
//                 that expected to block without a timeout.
//
//  2009-12-25 - Mark Davis
//               - restored resetCode msg now that DLM (supposedly) no longer
//                 clears it
//               - Added check and counter for recursive call to
//                 timer_interrupt()
//
//  2009-12-21 - Mark Davis
//               - resetCode msg removed for now as the value is apparently
//                 cleared from the CPU register before we can get to it.
//
//  2009-12-20 - Mark Davis
//               - Added syslog msg with value of resetCode at startup.  This
//                 is the value read by CStart from the ZWorld Global Control/
//                 Status register at startup (shifted 6 bits so it just has
//                 the 2 bits indicating the cause of the reset).
//
//  2009-12-18 - Mark Davis
//               - Initialized diaglogInfo->-Ticks before starting the
//                 watchdog timer.  This avoids a possible false triggering of
//                 the post-mortem dump code. (*)
//               - Expanded the post-mortem dump code to do the show the same
//                 info that the inttab userline command does.
//
//  2009-12-09 - Mark Davis (*)
//               - Fixed code for interrupt handler "shims" that wrap all
//                 interrupts but the timer one with a scheduler lock (which
//                 insures a task switch doesn't occur within an ISR).
//               - Added period check to kernel task to check for interrupt
//                 handlers that have been changed (and to change them to
//                 use the shim again with the new address).
//               - Added UnhandledInterruptHandler() function to override the
//                 default one that simply reboots.  The new function can't
//                 handle an unknown interrupt, but it should at least add
//                 and entry to the log so we know if one happens.
//
//  2009-12-07 - Mark Davis
//               - Added copy of timeZone value to the values stored in
//                 the battery-backed memory (and restored on bootup).  This
//                 allows the system log to have the correct time when it
//                 first boots up.
//
//  2009-11-05 - Mark Davis
//               - The configd task's announcementInterval is now reset
//                 whenever the DeviceID changes by calling the new configd
//                 function resetAnnounceInterval() (rather than having the
//                 configd module be responsible for detecting the change).
//               - The new function is also called repeatedly whenever the the
//                 OS is about to shutdown.
//
//  2009-09-18 - Mark Davis
//               - The "killing process ..." syslog message is no longer
//                 written for a task that called Exit()
//
//  2009-08-24 - Mark Davis
//               - Modified resetWDT() so it now works as follows:
//                  - BOTH the kernel and applic tasks must call resetWDT()
//                    at least once every 60000ms
//                  - EITHER the kernel or the applic task must call resetWDT()
//                    at least once every 1900ms.
//                 This change is to avoid the problem with the last method
//                 that allowed the kernel and/or applic task get stuck in a
//                 loop that did NOT call resetWDT().
//
//  2009-07-30 - Mark Davis
//               - sleepMS() no longer lets the kernel try to sleep.  Since it
//                 is the one responsible for waking up sleeping tasks this
//                 would effectively hang the OS.
//
//  2009-06-28 - Mark Davis
//               - Task state now shown as "Blocked*" if the task is blocked
//                 but has a timeout value.
//
//  2009-06-23 - Mark Davis
//               - Eliminated the redundant reloading of the param files.
//                 This was a waste of time and just added to the mass of
//                 log entries for each startup.
//
//  2009-06-21 - Mark Davis
//               - The current time is now saved to battery-backed RAM so
//                 the time the module was last running can be logged next
//                 time it boots up.
//
//  2009-06-19 - Mark Davis
//               - Added Module ID (last 3 bytes of MAC addr) to info logged
//                 at startup (independent of the messages showing the state
//                 of the network, which, on occasion, don't get logged if
//                 there is a problem).
//               - Added Module ID, appl vers, and RTOS vers to info logged
//                 by postMortem()
//               - Increased the length of time that postMortem() resets the
//                 watchdog timer for so it has more time to log its info.
//
//  2009-06-18 - Mark Davis
//               - The postMortem() function now dumps a list of the processes
//                 that were running and some information about each one that
//                 could help determine what caused the reboot.  In case the
//                 controller is unable to save the new log entries after
//                 rebooting, it writes the same information to a file on
//                 the flash file system.
//
//  2009-06-17 - Mark Davis
//               - Added a counter to keep track of how many milliseconds it
//                 has been since the watchdog timer was reset.  If the counter
//                 indicates the timer is about to reset the CPU, the new
//                 new postMortem() func is called to log the fact.
//
//  2009-06-16 - Mark Davis
//               - The system log from the previous reboot/reset is restored
//                 (if there is a recognizable one to restore).
//
//  2009-06-11 - Mark Davis
//               - Added gettimeofday() and settimeofday() functions (although
//                 the settimeofday() day currently ignores the microseconds).
//
//  2009-06-08 - Mark Davis
//               - The interval for scheduler ticks is now settable in the code.
//                 This allows applications that do a lot in the main applic
//                 task (which is supposed to run every tick) but don't need
//                 to do it every 10 ms to free up a lot of CPU cycles by
//                 doing scheduler ticks (and running the main applic task
//                 and the kernel task) less frequently.  This makes most of
//                 the other tasks more responsive since there are more cycles
//                 per second to divide up among them.
//
//  2009-05-13 - Mark Davis -
//               - Network init now done before the majority of the task table
//                 initialization is done.  The primary reason for this is to
//                 allocate space for the stacks AFTER the network init is
//                 done.  For some reason, the network init fails if the
//                 amount of free memory isn't significantly MORE than what
//                 it allocates (even though it never USES any more than what
//                 it allocates at init time ?!?!?!?!?)
//
//  2009-04-30 - Mark Davis -
//               - The OS is rebooted if we are running the primary boot image
//                 and the socket lib did not properly initialize
//
//  2009-04-15 - Mark Davis
//               - Added applicActive flag that must be set by the applic task
//                 to indicate when it is NOT safe to reboot the controller.
//                 (although it is not used yet...)
//
//  2009-04-13 - Mark Davis
//               - Increased max time for all tasks to exit after a shutdown
//                 starts from 5 to 10 seconds.
//
//  2009-03-19 - Mark Davis
//               - The runningPrimary flag is now set at startup using the
//                 return value of the modified DLM_Init() function.
//
//  2009-03-17 - Mark Davis -
//               - Changed handling of shutdown as follows:
//                  Previous method:
//                  - Send SIGTERM to all processes
//                  - If no tasks but the kernel and idle tasks are left or
//                    10 seconds elapse then shut down the network and reboot
//                  New method:
//                  - Send SIGTERM to all processes
//                  - If 5 seconds elapse, Send SIGKILL to all processes
//                  - When no tasks but the kernel and idle tasks remain, shut
//                    down the network and reboot
//                  The old way could result in the system being rebooted while
//                  tasks that had blocked the SIGTERM signal were still
//                  running.  The new approach ensures all but the kernel and
//                  idle tasks have been cleaned up by _killTask_(), which
//                  removes tasks from wait queues and closes any open
//                  descriptors.
//
//  2008-04-07 - Mark Davis -
//               - changes to support more generic "registered" file system
//                 model
//
//  2008-03-07 - Mark Davis -
//               - added applicTicks counter
//
//  2007-11-15 - Mark Davis -
//               - added support for SRFS keep-alive packets
//
//  2007-11-14 - Mark Davis -
//               - fixed first entry in process table that was not properly
//                 initialized (showed up as a problem in the Modular Power
//                 supply appl when trying to init the serial ports)
//
//  2007-07-14 - Mark Davis -
//               - added pipe() function
//
//  2007-02-13 - Mark Davis -
//               - started changes to switch from using TurboTask to using
//                 "really tiny" micro-kernel based on SHD's CoExec lib
//
//  2006-03-11 - Mark Davis -
//               - started work on first version based on pieces from earlier
//                 test program
//
//-----------------------------------------------------------------------------

#include <common.h>

#include <linux/sched.h>

#include <sys/types.h>
#include <errno.h>
#include <unistd.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <signal.h>
#include <sys/time.h>

#include <Rabbit.h>


#include <sys/backup.h>
#include <sys/configd.h>

#include <sys/spi.h>
#include <sys/hdwDefs.h>


#include <DLMiii.h>
#include <SHD_DLM.h>


#include <sys/serial.h>

#include <sys/hdwDefs.h>

#include <applic.h>

#include <TIMERB.h>

#include <sys/modbus.h>

//up above #include <sys/backup.h>


#include <linux/wait.h>


//--- buffer for improving printf/etc performance on stdout ---
char far  outBuf[NUM_TASKS][StdOutBufSize];


#include <stcpip.h>

#include <inetd.h>

#include <busybox.h>

#include <semaphore.h>

#include <sys/serial.h>


//void startMyTimer(void);


//#include <sys/RegVars.h>

#include <sys/farprintf.h>


#define  ValidTZ(tz)  (((tz) >= -12) AND ((tz) <= 13))



extern ulong  rtosVersTime;


extern void logAllocation(ulong start, ulong size);


void wdTimingOut(void);


#asm
    .extern __callde
    .extern __callhl
#endasm

#asm
  .defseg FARBSS
#endasm

#pragma seg (farbss, xmemstart)
char far xmemplaceholder[2];
#pragma seg (farbss)



#define _str_(s)  # s

#define  SegStart(seg)  ( \
  msw = _asm ([_DE], "ld de,(sgat " _str_(seg) ") >> 16", _DE ),  \
  lsw = _asm( [_BC], "ld bc,word (sgat " _str_(seg) ")", _BC ),  \
  (ulong)msw << 16 | lsw )

#define  SegSize(seg)  ( \
  msw = _asm ([_DE], "ld de,(sgsz " _str_(seg) ") >> 16", _DE ),  \
  lsw = _asm( [_BC], "ld bc,word (sgsz " _str_(seg) ")", _BC ),  \
  (ulong)msw << 16 | lsw )


#define  RootSegStart(seg)  \
  ( _asm( [_BC], "ld bc,word (sgat " _str_(seg) ")", _BC )  )

#define  RootSegSize(seg)  \
  ( _asm( [_BC], "ld bc,word (sgsz " _str_(seg) ")", _BC ) )


ulong  xmemAtStartup, xmemAfterBB;


//----- Rabbit 3000A interrupts for memory protection errors ---
#define  INT_WRITE_FAULT     0x09
#define  INT_MODE_VIOLATION  0x18
#define  INT_STACK_LIMIT     0x1B

//-----------------------------------------------------------------------------
uint  countersUpdated;

uint schedCallsPrevSec, taskSwPrevSec;

ulong  kernelLastCalled, applicLastCalled;

static int  lastTaskIdx;

static ulong  lastDeviceID;

char  hostName[HostNameLen];

FILE  taskIO, diagLog, *origIO;

char far osName[64], osDTM[32];

//uint   kernelTicks, kernelTicksPerSec;
ulong  shutdownStarted;

bool  checkWDT;

DiagLogInfo far  prevLogInfo;

int ii;

bool  gotTime, runningPrimary, runBackup, recursiveDump, attemptRestore;

bool  rebootMsg, doReboot, doShutdown, useSoftReset, memCorrupted,
      applicActive, restoredLog, doCleanShutdown;

int  klocks;

uint  milliTicks;
ulong  hdwResetTime, rebootTime;
uint  applicTicks;

char nestedTimerInt = -1;

ulong  deviceID;
uint  devType, hdwType, percentIdle;
int  timeZone, lastTimeZone;

int copyGCSR;


//--- values used by farmalloc()/etc --- must be init before use ---
char const far *_farheap;
unsigned long _farheapsize;

ulong bbAllocatedCount;

uint  codeStart, codeSize, codeEnd,
      constStart, constSize, constEnd,
      dataStart, dataSize, dataEnd,
      bssStart, bssSize, bssEnd;
ulong  farCodeStart, farCodeSize, farCodeEnd,
       farConstStart, farConstSize, farConstEnd,
       farBSSStart, farBSSSize, farBSSEnd;
uint  codeCRC, constCRC, farCodeCRC, farConstCRC;

uint  intVectStart, intVectEnd;

uint  roStart, roEnd, wpStart, wpEnd;
ulong  roStartFar, roEndFar, wpStartFar, wpEndFar;

uint  wpSegMask, wpRootMask;

uchar  cpuID, cpuRev;
bool  hdwProtection = false;

int sockStat;

char  moduleID[8];


//#define  MaxReturnAddrs  16

//ulong  far returnAddrs[MaxReturnAddrs], newReturnAddrs[MaxReturnAddrs];
//uint  numReturnAddrs, nextReturnIdx, numNewAddrs;

#asm
.cseg
  .extern __callhl_
  .extern __callde_
#endasm

#define  CallDE  ( _asm ([_DE], "ld de,__callde##", _DE ) )
#define  CallHL  ( _asm ([_DE], "ld de,__callhl##", _DE ) )


uchar *sp_;

const far char asterisks[] = {
"******************************************************************************"
};

const far char *resetCause[4] = {
  "Jump to 0?",
  "Watchdog Timer",
  "N/A",
  "Hdw Reset or Power Cycled"
};

const sigset_t  DefaultIgnoreSet =
   (sigmask(SIGCHLD) | sigmask(SIGURG) | sigmask(SIGCONT) | sigmask(SIGWINCH));

const sigset_t  UnblockableSet = ( sigmask(SIGKILL) | sigmask(SIGSTOP) );


//--- Code sections that can be tracked by toggling digital outputs ---
//--- The "diag" command can be used to set the bits to toggle using ---
//--- these names ---
ulong  TrackApplLoop = 0, TrackScheduler = 0,
       TrackKernelLock = 0, TrackTimerISR = 0;


DECLARE_WAIT_QUEUE_HEAD(sleepQueue);


//-----------------------------------------------------------------------------
//  Print a message regarding a function that was called w/o KLOCK() active.
//  The only real purpose of this function is to save code and data space in
//  the calling functions.
//-----------------------------------------------------------------------------
void klockErr(const char far *context)
{
  printf("***** %ls called without KLOCK()!!! *****\r\n", context);
}

//-----------------------------------------------------------------------------
//  assert "the big kernel lock"
//
//  NOTE:  This is also called directly via lock macros in Stcpip/net.h
//-----------------------------------------------------------------------------
void near KLOCK(void)
{
   SAFE_INC(_disabled);

   if (TrackKernelLock)  BITSON_(TrackKernelLock);
}

//-----------------------------------------------------------------------------
//  release "the big kernel lock"
//
//  NOTE:  This is also called directly via unlock macros in Stcpip/net.h
//-----------------------------------------------------------------------------
void near KUNLOCK(void)
{
  _EnterCritical_();

  if (!_disabled)  {
    _ExitCritical_();
    fprintf(DiagLog, "*** Unbalanced KUNLOCK() !!! ***\n");  return;
  }

  if (--_disabled)  {  _ExitCritical_();  return; }

  if (TrackKernelLock)  BITSOFF_(TrackKernelLock);

//--- If there is a pending task switch then we "preempt" ourselves ---
  if (_pending)  {
    current->preempt_count += 8;
    _ExitCritical_();  schedule();  _EnterCritical_();
    current->preempt_count -= 8;
  }

  _ExitCritical_();
}

//-----------------------------------------------------------------------------
//  Reboot the CPU and run the RDLM.
//
//  If restart is false, it tells the RDLM it should NOT (re)load one of the
//  boot images, which leaves the device in a state where the Update Manager
//  application can see it and communicate with it.
//
//  If restart is true, it tells the RDLM to load and run either the primary
//  or backup boot images, based on whether global variable runBackup is true
//  or false.
//-----------------------------------------------------------------------------
void REBOOT(bool restartOS)
{
  int  mode;


  ipset1();

//--- lets next instance know this was a controlled shutdown ---
  diagLogInfo->ctldShutdown = true;

//--- In case time values are goofy.  This helps insure a clean soft reset ---
//--- and lets us fix the RTC-based time values if needed. ---
  diagLogInfo->timerSeconds = _TimerSeconds;


  if (!restartOS)  {
    fprintf(DiagLog, "\n----- shutdown, run DLM ---\n\n");
    mode = UDPDNLD_MODE_RUN_DLM;
  }
  else  {
    if (runBackup)  {
      fprintf(DiagLog, "\n----- reboot and run Backup -----\n\n");
      mode = UDPDNLD_MODE_RUN_BACKUP;
    }
    else  {
      fprintf(DiagLog, "\n----- reboot and run Primary -----\n\n");
      mode = UDPDNLD_MODE_RUN_PRIMARY;
    }
  }
  if (useSoftReset AND allowSoftReset)  mode |= UDPDNLD_MODE_SOFT_RESET;

  flushConsole();


  #ifdef SHD_DLM
    DLM_Reboot(mode);
  #else
    ipset3();  WDT_250mS();  WDT_ENABLE();  for(;;);
  #endif
}

//-----------------------------------------------------------------------------
//  The purpose of these functions is to cause the boot image to reloaded and
//  execution restarted the same as if the CPU had been reset, but with
//  one important exception:  Avoid doing an actual reset.
//
//  This approach to restarting the OS and the application code provides
//  the means for restarting while avoiding one important aspect of a real
//  reset:  It does not cause the processor to return to its default state.
//
//  The advantage of this method is that it provides a less painful way to
//  recover from fatal runtime errors (such as the code becoming corrupted,
//  or some unrecoverable corruption of a data structure.
//
//  This "pseudo" reset is accomplished by allocating a block of memory
//  in a memory bank that is not needed by the RDLM/boot manager, copying
//  a small function to it, and then "calling" the function.  That function
//  then maps the first 256K bytes of flash to the first 256K memory bank
//  and jumps to address 0, causing the RDLM to execute.
//
//  NOTE that this mearly takes care of avoiding the consequences (and
//  advantages) of doing an actual CPU reset.  It still requires that
//  other modules (including cstart and hardware/platform specific ones)
//  skip steps they normally do at startup that may alter the state of
//  any outputs that may affect the devices being controlled.
//
//  Also keep in mind the limitations and downside to this approach:
//
//    - If some part of the hardware or the CPU is in a messed up state
//      and there is no way to insure against this during startup (without
//      changing the state of the outputs), then using this method to restart
//      can lead to erratic or undesired behavior that a true reset would
//      have avoided.
//
//    - A soft reset is still a significant disruption.  A clean shutdown of
//      the OS should still be done first (if possible), and the OS will still
//      be reloaded and restarted.  The only difference is that much of the
//      default state of the CPU caused by a real reset does not occur, so
//      it should be possible to avoid changes to the outputs used to control
//      any external devices.
//
//  If handled properly, a soft reset can leave external devices in a steady
//  state while the OS is shutting down and restarting.  Of course, it is up
//  to the application to make sure that the control signals are in a state
//  that is SAFE to leave them in during a reboot.
//
//-----------------------------------------------------------------------------
#define  PseudoReset_FuncSize  256   // must be >= size of pseudoReset() func

void _nostep pseudoReset(void)
{
//  ulong  startWait;

  fprintf(DiagLog, "\n***** Restarting w/o CPU reset *****\n\n\n");

  flushConsole();  KLOCK();

  // startWait = GetMS();  while (msSince(startWait) < 100);

  WDT_DISABLE();  ipset3();

  outi(MMIDR, 0x80);  // enable access to registers > 0x00FF

  outi(STKCR, 0x00);  // disable hdw stack limit checks
  outi(WPCR, 0);      // disable write protection for System Mode

  outi(WPSALR, 0);    // clear 4K write protection masks
  outi(WPSAHR, 0);    // clear 4K write protection masks
  outi(WPSBLR, 0);    // clear 4K write protection masks
  outi(WPSBHR, 0);    // clear 4K write protection masks

  outi(MMIDR, 0x00);  // disable access to registers > 0x00FF


// The next statement maps flash to the first memory bank (usually 256K), so
// don't call any functions after this point that could possibly lead to an
// attempt to access something in the 1st memory bank (which basically means
// NO function calls after this point!)

  outi(MB0CR, _CS0 | _OE0_WP | _WS4);

#asm
    ld     sp,0

    xor    a,a
    ld     eir,a
    ld     iir,a

    jp     0
#endasm

}

#pragma offset_labels on
#pragma warn -757

//---------------------------------------------------------------
void softReset(void)
{
  char far *bufAddr;
  uchar  memBank;
  uint  logBufAddr;
  uchar  bufXPC, _4kblock;



  if (!allowSoftReset)  {
    printf("Soft Reset disabled - doing normal reset\r\n");
    REBOOT(true);
  }

  saveApplState();

  printf("\nFor PseduoReset func:\r\n");
  bufAddr = bb_xalloc(PseudoReset_FuncSize);  // highest addr unused xmem block

  _4kblock = (uchar)((ulong)bufAddr >> 12);
  memBank = _4kblock >> 6;
  logBufAddr = ((uint)(ulong)bufAddr & 0x0FFF) | 0xE000;
  bufXPC = _4kblock - 0x0E;

//  printf("buffer for pseudoReset():\r\n"
//         "  bufAddr: 0x%05lX\r\n"
//         "  _4kblock: 0x%02X\r\n"
//         "  memBank: %u\r\n"
//         "  logBufAddr: 0x%04X\r\n"
//         "  bufXPC: 0x%02X\r\n",
//         (ulong)bufAddr, _4kblock, memBank, logBufAddr, bufXPC);

// If the buffer for pseudoReset() is in the same bank that the flash needs,
// then we can't use it
  if (memBank == 0)  {
    printf("Unable to perform softReset(), so doing normal reboot\r\n");
    REBOOT(true);
  }

  diagLogInfo->digOutState = digitalOut;
  diagLogInfo->didSoftReset = true;

//  ipset3();  ipset3();  ipset3();  ipset3();  // set IntPri "stack" to 0xFF
;
#asm
;; fmemcpy(bufAddr, (void far *)(ulong)pseudoReset, PseudoReset_FuncSize);

  ;; PseudoReset_FuncSize
    ld      de,0x64
    ld      c,d
    ld      b,d
    push    bc
    push    de

  ;; pseudoReset
    ld      de,word (phys $pseudoReset)
    ld      c,phys $pseudoReset >> 16
    push    bc
    push    de

  ;; bufAddr
    ld      hl,(ix+.bufAddr)
    ex      de,hl
    ld      hl,(ix+.bufAddr+2)
    ld      bc,hl
    push    bc
    push    de

    lcall   $fmemcpy

    add     sp,12
#endasm

;
#asm
;; (*bufAddr)();
    ld     hl,(ix+.logBufAddr)  ;; value for PC
    push   hl

    ld     hl,(ix+.bufXPC)      ;; value for XPC
    ld     a,l

    ld     xpc,a                ;; new XPC value invoked after next instr
    ret                         ;; pop's PC from stack.  Along with the new
                                ;; value of XPC, this causes a jump to the
                                ;; copy of pseudoReset() in the xmem buffer

#endasm

}
#pragma warn 757
#pragma offset_labels off



//-----------------------------------------------------------------------------
//  return pointer to a string description of the state of the specified task
//-----------------------------------------------------------------------------
char *statusDescr(PIX pix)
{
  uint  state;
  task_t  *task;
  char  *descr;


  task = tasks + pix;  state = task->state;

  do  {
    if (! task->inUse)                      { descr = "Dead    ";  break; }

    if (state == TASK_RUNNING)              { descr = "Ready   ";  break; }

    if ((state == TASK_UNINTERRUPTIBLE) OR (state == TASK_INTERRUPTIBLE))  {
      if (list_empty(&task->waitQs.listHead))  { descr = "Schedule";  break; }
      if (task->sleepTicks)                 { descr = "Sleeping";  break; }
      if (task->timeoutTicks)               { descr = "Blocked*";  break; }
                                              descr = "Blocked ";  break;
    }
                                              descr = "Stopped ";
  }  while (0);


  return descr;
}

//-----------------------------------------------------------------------------
//  kill oldest "tsh" process so a newer one can be created
//-----------------------------------------------------------------------------
void killOldestShell(void)
{
  PIX  pix;
  PID  pidOldest;
  ulong  earliestStarted;
  task_t  *task;
  void  *ioDev;
  FDEV  far *file;
  struct stat  fsbuf;


  earliestStarted = ULONG_MAX;  pidOldest = 0;

//--- find earliest started "tsh" process using a socket for stdio ---
  for (pix=0; pix<NUM_TASKS; ++pix)  {
    if (! (task = tasks + (uint)pix)->inUse)  continue;
    if (strcmp(task->argv[0], "tsh") != 0)  continue;
    if (! (file = FD[pix][0]))  continue;

    ioDev = file->ioDev;
    if (file->ioFuncs->fstat(ioDev, &fsbuf))  continue;

    if (! S_ISSOCK(fsbuf.st_mode))  continue;

    if (task->started < earliestStarted)  {
      pidOldest = task->pid;  earliestStarted = task->started;  }
  }


  if (pidOldest)  kill(pidOldest, SIGKILL);
}


//-----------------------------------------------------------------------------
//  Make a copy of the argument strings for a process
//  Returns 0 on success, < 0 if an error
//-----------------------------------------------------------------------------
int copyArgs(task_t *task, int argc, char **argv)
{
  int  i, argBytes;
  char  *fromPtr, *toPtr;

//--- copy cmd line parameters ---
  for (i=0,argBytes=0; i<argc; ++i)  argBytes += strlen(argv[i]) + 1;
  argBytes += (argc+1) * sizeof(char *);
  if (argBytes > MaxArgBytes)  {
    printf("Argument list too long\r\n");  return -1;  }

  toPtr = task->args;
  task->argv = (char **)toPtr;  toPtr += (argc+1) * sizeof(char *);
  for (i=0; i<argc; ++i)  {
    task->argv[i] = toPtr;  fromPtr = argv[i];
    while (*fromPtr)  *toPtr++ = *fromPtr++;
    *toPtr++ = 0;
//     printf("(%lu)(%lu)<%s>\r\n", (ulong)task->argv[i], (ulong)argv[i], task->argv[i]);
  }
  task->argv[argc] = NULL;  task->argc = argc;

  return 0;
}

//-----------------------------------------------------------------------------
//  create a new DEBROS process
//-----------------------------------------------------------------------------
PID createProc(ProcMain procMain, char priority, int argc, char **argv,
               fd_t infd, fd_t outfd, fd_t errfd)
{
  PID  pid;
  PIX  i, j, pix;
  task_t  *task;
  FDEV far  **fdPtr;
  uchar  pass;


  KLOCK();

//--- determine if there is an unused entry in the process table ---
//--- (use round-robin pattern to minimize reuse of the same slot) ---
  for (pass=0; pass<3; ++pass)  {
    for (pix=lastTaskIdx,j=0; j<=NUM_TASKS; ++j)  {
      if (++pix == NUM_TASKS)  pix = 0;
      if (! (task = tasks + (uint)pix)->claimed)  break;
    }
    if (task->claimed)  {
      KUNLOCK();
      if (strcmp(argv[0], "tsh") == 0)  {
        killOldestShell();  sleepMS(500);  KLOCK();  continue; }
      printf("\r\nUnable to create process for: %.20s\r\n", argv[0]);
      return -1;
    }
    break;
  }
  lastTaskIdx = pix;  task->claimed = true;

  KUNLOCK();

  memset(&task->zeroAfter, 0, sizeof(task_t) - offsetof(task_t, zeroAfter));

  if (copyArgs(task, argc, argv))  { task->claimed = false;  return -1; }

//  printf("createProc: ");
//  for (i=0; i<task->argc; ++i)  printf("[%s] ", task->argv[i]);
//  printf("\r\n");

//--- Create a pid for this process.  The pid is made up of the index into ---
//--- the process table and the # of times the slot has been used.  This ---
//--- avoids resuse of a pid too quickly, but the index in to the table ---
//--- can still be determined with a simple mask rather than a search ---
  pid = (int)pix + task->useCount * (TaskIdxMask+1);  ++task->useCount;

  if (priority >= NUM_PRIORITIES)  priority = NUM_PRIORITIES - 1;

// ***** IMPORTANT *****
//--- anything new or changed here must be done in main() for the kernel ---
//--- process as well ---

//--- initialize settings for the new process ---
  task->diagFlags = newTaskDiagFlags;
  task->pid = pid;
  task->ppid = current->pid;
  task->priority = priority;
  task->outPtr = outBuf[(uint)pix];
  task->procMain = procMain;
  task->FDSet = (void far **) FD[(uint)pix];

  KLOCK();

//--- copy the specified fd's for stdin, stdout, stderr ---
  FD[pix][0] = FD[userPix][infd];
  FD[pix][1] = FD[userPix][outfd];
  FD[pix][2] = FD[userPix][errfd];

//--- increment the openCount for each file referenced ---
  fdPtr = FD[pix];
  for (i=0; i<3; ++i)  if (fdPtr[i])  ++fdPtr[i]->openCount;

  KUNLOCK();


  task->started = readRTC(NULL);
  task->lastResumed = task->lastStopped = 0;
  task->resumedTicks = clockTicks();


  init_waitqueue_head(&task->waitQs);

  task->inUse = true;

  reset_task(pix);  // prep task for running

  fprintf(DiagLog, "Created process #%d (", pid);
  for (i=0; i<argc; ++i)  fprintf(DiagLog, " %s", argv[i]);
  fprintf(DiagLog, " )\n");

  _EnterCritical_();  task->state = TASK_RUNNING;  _ExitCritical_();

  return pid;
}


//-----------------------------------------------------------------------------
//  Determine if a given process id is for an active process
//
//  Returns the index if the process id is valid and belongs to a running
//  or blocked process.  Otherwise it sets errno and returns -1.
//-----------------------------------------------------------------------------
PIX _activeTaskIdx_(PID pid)
{
  task_t  *task;
  PIX  pix;


  if (pid < 0)  { ERRNO = ESRCH;  return -1; }

  pix = TaskIdx(pid);

  if (pix >= NUM_TASKS)  { ERRNO = ESRCH;  return -1; }

  task = tasks + (uint)pix;

  if (task->pid != pid)  { ERRNO = ESRCH;  return -1; }  // slot reused?

  if (! task->inUse)  { ERRNO = ESRCH;  return -1; }  // no longer a process?

  if (!ActiveTask(task))  { ERRNO = ESRCH;  return -1; }  // dead process?

  return pix;
}


//-----------------------------------------------------------------------------
//  Determine if a given process id is for an active process
//
//  Returns the index if the process id is valid and belongs to a running
//  or blocked process.  Otherwise it sets errno and returns -1.
//-----------------------------------------------------------------------------
PIX _inUseTaskIdx_(PID pid)
{
  task_t  *task;
  PIX  pix;


  if (pid < 0)  { ERRNO = ESRCH;  return -1; }

  pix = TaskIdx(pid);

  if (pix >= NUM_TASKS)  { ERRNO = ESRCH;  return -1; }

  task = tasks + (uint)pix;

  if (task->pid != pid)  { ERRNO = ESRCH;  return -1; }  // slot reused?

  if (! task->inUse)  { ERRNO = ESRCH;  return -1; }  // no longer a process?

  return pix;
}


//-----------------------------------------------------------------------------
//  Determine the number of unused bytes on an active process' stack
//  process id.
//
//  Returns amount of stack unused if the process id is for an active process,
//  or -1 if an error.
//-----------------------------------------------------------------------------
int _stackFree_(PID pid)
{
  PIX  pix;
  int  unused;


  if ((pix = _inUseTaskIdx_(pid)) < 0)  return -1;

  unused = check_stack(pix);

  return unused;
}


//-----------------------------------------------------------------------------
//  Get the Program Counter for the current task
//-----------------------------------------------------------------------------
uint getPC(void)
{
  uint  sp;

  sp = _asm( [_HL], "ld hl,ix", _HL ) + 2; // IX itself is pushed, using 2 bytes
  return *((uint *)sp);
}

//-----------------------------------------------------------------------------
//  return a count of the # of active tasks
//-----------------------------------------------------------------------------
int numActiveTasks(void)
{
  task_t  *task;
  int  count;

  for (count=0,task=tasks; task<endTasks; ++task)  if (task->inUse)  ++count;

  return count;
}


//-----------------------------------------------------------------------------
//  Kill a process and clean up after it
//
//  NOTE:  Will NOT kill the kernel or the idle task
//-----------------------------------------------------------------------------
void _killTask_(task_t *task)
{
  uint  i;
  int  pid;
  task_t  *childTask;


  if (! task->inUse)  return;

//--- kernel and idle tasks must NOT be killed ---
  if ((task == kernelTask) OR (task == idleTask))  return;

//  if (strcmp(task->argv[0], "syslog") == 0)  return;  //tdebug


  pid = task->pid;

  if (task != current)  {
    fprintf(DiagLog, "killing process %.20s[%d]\r\n", task->argv[0], pid);
    KLOCK();  task->preempt_count = 0;  task->state = EXIT_DEAD;  KUNLOCK();
  }

//  fprintf(DiagLog, "  pending: 0x%08lX  blocked: 0x%08lX\r\n",  //tdebug
//          (ulong)task->sig_pending, (ulong)task->sig_blocked);  //tdebug


//--- remove all the nodes from any queues the task might be in ---
  KLOCK();
  task->timeoutTicks = 0;
  removeFromQueues(task, true);
  KUNLOCK();

//--- close all of the process' descriptors ---
  for (i=0; i<FD_SETSIZE; ++i)  _closeFD_(task, i);

  KLOCK();

//--- if the process has any child processes, then change their parent ---
//--- process id's to -1 ---
  for (childTask=tasks; childTask<endTasks; ++childTask)  {
    if (! childTask->inUse)  continue;
    if (childTask->ppid == pid)  childTask->ppid = -1;
  }

//finish - need generic method for handling real-time scheduling
  if (current == applTask)  {
    _EnterCritical_();  applTask = NULL;  _ExitCritical_(); }

//--- cancel the task and mark the process table entry as free for use ---
  task->preempt_count = 0;  task->state = TASK_DEAD;
  task->inUse = task->claimed = false;

  KUNLOCK();

  if (task == current)  for (;;)  schedule();
}


//-----------------------------------------------------------------------------
//  exit the calling process
//-----------------------------------------------------------------------------
void Exit(int stat)
{
  int  stkFree;


  current->exit = stat;

  stkFree = _stackFree_(getpid());
  fprintf(DiagLog, "exiting with status: %d, %d stack free\n", stat, stkFree);

  if (stkFree < 1)
    fprintf(DiagLog, "***** !!! overflowed its stack !!! *****\n");

//finish: need to deal with informing parent process, zombie procs, etc...

  fflush(stdout);

  _killTask_(current);
}

//-----------------------------------------------------------------------------
//  Unix-compatible version of sethostname() (unfortunately, the Softools port
//  of the DC libs already uses the standard name...)
//-----------------------------------------------------------------------------
int SetHostName(char *name, int nameLen)
{
  if (!name)  { errno = EFAULT;  return -1; }
  if ((nameLen >= sizeof(hostName)) OR (nameLen < 1))  {
    errno = EINVAL;  return -1; }

  if (nameLen >= HostNameLen)  nameLen = HostNameLen - 1;

  KLOCK();

  if (name != hostName)  {  // skip copy if src addr same as dest
    memcpy(hostName, name, nameLen);  hostName[nameLen] = 0;  }

  resetWDT();

  sethostname(hostName);  ffile_setName(hostName);

  KUNLOCK();

  resetAnnounceInterval();  // announce the change ASAP

  return 0;
}

//-----------------------------------------------------------------------------
//  Kernel-space function to send a signal to a process
//-----------------------------------------------------------------------------
int kill(PID pid, int sig)
{
  PIX  pix;
  task_t  *task;
  sigset_t  sigSet;



  if ((sig < 0) OR (sig > 31))  { errno = EINVAL;  return -1; }

  sigSet = sigmask(sig);

  KLOCK();

//--- deliver signal to ALL tasks? ---
  if (pid == -1)  {
    for (task=tasks; task!=endTasks; ++task)
      if (task->inUse)  task->sig_pending |= sigSet;
    KUNLOCK();  return 0;
  }

//--- does the target process still exist? ---
  if ((pix = _inUseTaskIdx_(pid)) < 0)  {
    KUNLOCK();  errno = ESRCH;  return -1; }

  task = tasks + pix;

//finish: LOTS  more to do (check for blocked signals, signal handler, etc)

  task->sig_pending |= sigSet;


  KUNLOCK();  return 0;
}

//-----------------------------------------------------------------------------
//  kernel-space function to wait for a child process to exit
//-----------------------------------------------------------------------------
PID waitpid(PID pid, int *status, int options)
{
  int  pix;
  task_t  *childTask;


  KLOCK();

  pix = TaskIdx(pid);
  if ((pix < 0) OR (pix >= NUM_TASKS))  {
    ERRNO = ECHILD;  KUNLOCK();  return -1;  }

  childTask = tasks + (uint)pix;

  fprintf(DiagLog, "blocking on %.20s[%d]\n", childTask->argv[0], pid);

//finish:  return standard values

//--- check for errors ---
  if ( (childTask->ppid != current->pid) OR  // not a child process
       (childTask->pid != pid) )  {  // process no longer exists
    ERRNO = ECHILD;  KUNLOCK();  return -1;  }

//finish:  need to handle exiting more thoroughly (like what happens if the
//         task's entry is reused between our checking here.  But task
//         entries really shouldn't BE reused if another process is blocked
//         waiting for them to exit)

//--- if child not dead, wait for it to exit ---
  while (childTask->inUse
      AND (childTask->ppid == current->pid)
      AND (childTask->pid == pid) )  { KUNLOCK();  sleepMS(50);  KLOCK(); }

  KUNLOCK();

  fprintf(DiagLog, "unblocked\n");

  return pid;  // child exited
}



//-----------------------------------------------------------------------------
//  Kernel space function to get the priority of a running process
//-----------------------------------------------------------------------------
int getpriority(int which, int who)
{
  PIX  pix;
  int  pri;


  if (which != PRIO_PROCESS)  { ERRNO = EINVAL;  return -1; }

  KLOCK();

  if ((pix = _activeTaskIdx_(who)) < 0)  {
    ERRNO = ESRCH;  KUNLOCK();  return -1; }

  pri = tasks[(uint)pix].priority;

  KUNLOCK();  return pri;
}


//-----------------------------------------------------------------------------
//  Kernel space function to set the priority of a running process
//-----------------------------------------------------------------------------
int setpriority(int which, int who, int priority)
{
  PIX  pix;


  if (which != PRIO_PROCESS)  { ERRNO = EINVAL;  return -1; }

  KLOCK();

  if ((pix = _activeTaskIdx_(who)) < 0)  {
    ERRNO = ESRCH;  KUNLOCK();  return -1; }

  if (priority < 1)  priority = 1;
  if (priority > NUM_PRIORITIES-1)  priority = NUM_PRIORITIES - 1;

  tasks[(uint)pix].priority = priority;

  KUNLOCK();  return 0;
}



//-----------------------------------------------------------------------------
//  Set the Real Time Clock and update any values that are dependent on it
//-----------------------------------------------------------------------------
int stime(time_t *timePtr)
{
  time_t  oldTime, newTime = *timePtr;


  _EnterCritical_();

  oldTime = TIME();  writeRTC(newTime);

//--- _TimerSeconds is the # of seconds since the last hard (CPU) reset, so ---
//--- rather than apply a correction to hdwResetTime, we simply recompute ---
//--- it from the current time (which is safer, since making sure that the ---
//--- _TimerSeconds value is consistent across soft resets is important) ---
  diagLogInfo->hdwResetTime = hdwResetTime = newTime - _TimerSeconds;

  if (newTime > oldTime)
    rebootTime += newTime - oldTime;
  else
    rebootTime -= oldTime - newTime;

  diagLogInfo->rebootTime  = rebootTime;

  _ExitCritical_();

//  The problem is that _TimerSeconds and _TimerMilliSeconds CANNOT be
//  changed after we have booted because they are used in endless places
//  to test for how much time has passed.  Change them after we have
//  booted, and all kinds of logic can simply hang or react at the wrong
//  time or the wrong way.
//
//  Aside from checking for reasonable values for the clock and time values
//  from the previous run, about all I can do is save a current correction
//  factor computed from what the clock SHOULD have set at startup and what
//  it DID say.  This would allow me to update values like upSecs that are
//  seen by outside parties but not used internally.  Although that could
//  cause the same kind of problem for external logic that is using such
//  a value the same way.
//
//  NOTE:  Fixing the values I pass on to the next instance will not only
//  correct the values seen externally, but will cause the internal ones to
//  drop back to the correct values, which will cause havoc for everything
//  after a soft reset.
//
//  Then again, a controller that is restarted with a soft reset and comes
//  starts with a messed up value means that the values used by the code
//  for computing intervals all suddenly jump to a much LARGER value.  Is
//  that really any better than having a correct value to begin with?
//
//  There really isn't a good, practical solution for a controller that is
//  already running with a messed-up value (other than to reboot it).  The
//  only REAL solution is to update ALL the starting time values used in
//  ALL the code, anywhere!  And to force all code that checks for elapsed
//  time to call some function to compute that time (as opposed to the
//  current fast, efficient method of simply doing a subtraction).
//


  return 0;
}

//-----------------------------------------------------------------------------
//  standard linux/unix func to get the time with sub-second accuracy
//-----------------------------------------------------------------------------
int gettimeofday(struct timeval *tv, void *obsolete_tz)
{
  ulong  ms;

//finish:  Once settimeofday() uses the tv_usecs value, I will have to
//         set and use a milliTimerOffset as well and deal with
//         wrap-around (i.e. if ms + milliTimerOffset > 1000, ++secs, or
//         if millTimerOffset < 0, then ms < 0, etc...)
  _EnterCritical_();
  tv->tv_sec = TIME();  ms = milliTicks;
  _ExitCritical_();

  tv->tv_usec = ms * 1000;

  return 0;
}

//-----------------------------------------------------------------------------
//  standard linux/unix func to set the time with sub-second accuracy
//-----------------------------------------------------------------------------
int settimeofday(const struct timeval *tv , const void *obsolete_tz)
{
//finish:  Need to set RTC with sub-second accuracy and implement a
//         milliTimerOffset value
  writeRTC(tv->tv_sec);

  return 0;
}

//finish:  Assuming the existing Softools time() func reads the RTC, should
//         I replace it with one that returns _TimerSeconds + hdwResetTime?
//         that would be much faster than reading the RTC.  But I would have
//         make sure that I DO read the RTC when necessary (like at startup).
//time_t  far time (time_t *__timer);


//-----------------------------------------------------------------------------
//  Sleep for at least the specified # of milliseconds.
//
//  Returns the aprox # of milliseconds the process actually slept
//-----------------------------------------------------------------------------
ulong sleepMS(ulong ms)
{
  current->sleepStarted = MS_TIMER;

//--- return immediately if kernel task ---
  if (current == kernelTask)  return msSince(current->sleepStarted);

  if (_disabled)  // busy wait if scheduler disabled
    while (msSince(current->sleepStarted) < ms);
  else  {  // otherwise use the sleep wait queue
    current->sleepTicks = msToTicks(ms);
    __wait_event(sleepQueue, !current->sleepTicks);
  }

  return msSince(current->sleepStarted);
}


//-----------------------------------------------------------------------------
//  Return the # of scheduler ticks equivalent to the given # of milliseconds,
//  rounded up to the nearest # of ticks.
//-----------------------------------------------------------------------------
ulong msToTicks(ulong ms)
{
  ulong numTicks = (ulong)((ms + ((uint)SchedMS>>1)) / (uint)SchedMS);
  if (ms AND (numTicks < 1))  numTicks = 1;

  return numTicks;
}


//-----------------------------------------------------------------------------
//  Do nothing process set to lowest priority - runs only when no other
//  processes are ready
//-----------------------------------------------------------------------------
void _task idle(int argc, char **argv)
{
  for (;;)  computeMemCRCs(false, false);
}



extern ushort nonKernelLocks, duplicateLocks, tcpTickTime, failedLocks;

//sem_t  tcpLock;

//-----------------------------------------------------------------------------
//  Callback for network lib locking when used by an RTOS
//-----------------------------------------------------------------------------
// void  near lock_semaphore(void) { sem_wait(&tcpLock); }
// void  near unlock_semaphore( void )  { sem_post(&tcpLock); }

/* This results in much better response times than using a semaphore (as above)
   But they aren't actually needed because we point tcp_lock and tcp_unlock
   directly at KLOCK and KUNLOCK()
  ------------------------------------
void  near lock_semaphore(void)  {
  KLOCK();  // if (current != kernelTask)  ++nonKernelLocks;
}

void  near unlock_semaphore( void )  { KUNLOCK(); }
*/

/*
//-----------------------------------------------------------------------------
//  substitue for Softools tcp_tick(s) func
//-----------------------------------------------------------------------------
_tcp_nodebug int tcp_tick(void* s)
{
  int retVal;
  uint  msTaken;
  ulong  startTime;


  if (SEM_LOCK(_tcpBusy) == 1)  {
    startTime = GetMS();
    retVal = _tcp_tick_internal(s);
    msTaken = (uint) msSince(startTime);
  }
  else  {
    ++failedLocks;  //tdebug
    msTaken = 0;  retVal = 1;  }

  _EnterCritical_();  tcpTickTime = msTaken;  _ExitCritical_();

  SEM_UNLOCK(_tcpBusy);

  return retVal;
}
*/
//-----------------------------------------------------------------------------
//  substitue for Softools tcp_tick(s) func
//-----------------------------------------------------------------------------
_tcp_nodebug int tcp_tick(void *sock)
{
  int  retVal;
  uint  startTime;

  KLOCK();

  startTime = GetMS_();
  retVal = _tcp_tick_internal(sock);
  tcpTickTime = msSince_(startTime);

  KUNLOCK();

  return retVal;
}


//-----------------------------------------------------------------------------
//  Called every SchedMS milliseconds by a periodic timer interrupt to update
//  various values and call the scheduler if it is not disabled.
//
// *** WARNING ***
//   The addition of any local variables will require changes to the
//   postMortem() function!
// *** WARNING ***
//-----------------------------------------------------------------------------
void _nostep sched_tick(void)
{
  StartInterval(TrackTimerISR);  // for real-time profiling/diagnostics

//tdebug - generate a write fault to test dumpFault(), stackTrace(), etc ---
//  if (current->priority == 13)  { current->priority = 12;  (*(uchar *)0) = 0; }

//--- Check for an imminent Watch-Dog Timer timeout ---
  if (checkWDT)  if ((diagLogInfo->wdtTicks += SchedMS) >= 1900)
    if (!doShutdown)  wdTimingOut();

//--- wake up the application and kernel tasks ---
  if (applTask)  if (ActiveTask(applTask))  applTask->state = TASK_RUNNING;
  tasks->state = TASK_RUNNING;

//--- keep track of the last time the OS was running (so we can log it when ---
//--- we reboot) ---
  diagLogInfo->seconds = TIME();  diagLogInfo->milliSecs = milliTicks;
  diagLogInfo->msTime = _TimerMilliSeconds[1];


/*
  disabled  schedule  preempt     -------- action --------
    x           0        x        return
    0           1        0        call schedule
    x           1        1        set 'pending', then return
    1           1        x        set 'pending', then return

    If the scheduler is disabled, or the current task is already marked as
    preempted, then one of 4 things is true:

      1. The current task is using the scheduler lock to protect shared data.

      2. The current task was just about to call the scheduler.

      3. We interrupted the scheduler itself.

      4. The current task was just resumed by the scheduler.

    In the 1st case, we must wait to switch tasks to avoid possibly corrupting
    the data the current task is accessing.

    In the 2nd case, calling the scheduler here would just be a waste, since
    the scheduler will be called as soon as we allow the preempted task to
    resume execution.

    In the 3rd case, the scheduler will simply return to protect itself from
    concurrent calls.

    The 4th case is a bit tricky.  The issue we need to guard against is as
    follows:

      - Suppose a task that was preempted by this function is later resumed.
        Tasks that were suspended due to a call to schedule by this function
        have a bunch of stuff on their stack that was pushed there by the
        TimerBISR function, plus its call to this function, and this
        function's call to schedule.

      - If a just-resumed task is preempted again before execution returns to
        the point it was last preempted, we end up shoving all the same stuff
        on the task's stack again.  Since each time it happens increases the
        amount of time before execution DOES return to the point it was 1st
        preempted, the risk of this happening a 2nd (and 3rd, and 4th, ...)
        time increases.  Once it starts, this can snowball to the point where
        the task's stack overflows which can lead to all kinds of chaos
        (although on the Rabbit 3000's this is now trapped by a hardware
        fault, so at least it is obvious when it happens, rather than causing
        strange behavior with no obvious cause).

    The way we handle all these cases is to set a flag indicating that we have
    a "pending" call to the scheduler.  Any code that clears the scheduler
    lock (which is limited to the scheduler itself and the KUNLOCK() function)
    checks the pending flag and, if it is set, invokes the scheduler.  For
    the scheduler itself, this just means looping until the flag is clear.
    For the KUNLOCK() function, it means it increments the preempt flag for the
    task, calls schedule, then decrements the prempt flag.

    NOTE that, while the amount of stuff pushed on to the stack to call the
    KUNLOCK() function is probably much smaller (since it is not invoked by
    an interrupt that has to insure all the registers are protected), the
    risk still exists, so it also increments and decrements the task's
    preempt counter.  This is also useful in monitoring the tasks, as a call
    to schedule from KUNLOCK is really nothing more than a delayed preempt -
    It just happens to be the task doing so "voluntarily".
*/

  if (_disabled OR current->preempt_count)  {
    _pending = 1;  EndInterval(TrackTimerISR);  return; }

/*
    So, we finally eliminated all the special cases.  If we got here then
    we are due to call the scheduler and know of no reason not to, so we
    mark the task as preempted and call the scheduler.  The preempted
    counter serves 2 purposes:

      1. It lets the scheduler know that the task was running when it was
         preempted, regardless of what its run state is, so it should be
         treated like any other READY task.  This handles the case where
         the task changed its run state and was then preempted before it
         could call schedule() or at the start of schedule(), before it
         suspended the task.

      2. It allows this function to detect the cases described above that
         cause it to set the pending flag instead of calling schedule
         itself.  As noted above this is important because it allows us to
         avoid calling the scheduler recursively.
*/

  ++current->preempt_count;

  EndInterval(TrackTimerISR);  // for real-time profiling/diagnostics

//--- Remember that a call to schedule() does not return (for the task ---
//--- that was running when it was called) until the task is resumed by ---
//--- the scheduler ---
  ipres();  schedule();  ipset1();

//=== WARNING: schedule() usually changes "current" and the stack pointer, ===
//=== so do NOT use any local variables after the call to schedule() !!! ===

  --current->preempt_count;
}


/*=============================================================================
 Identifier      Vector - Supports
          --- IIR ---
 INT_TIMER         0x00 - Periodic timer
                   0x01 - Secondary WDT (*)
 INT_RST10         0x02 - RST 10
 INT_RST18         0x03 - RST 18
 INT_RST20         0x04 - RST 20
 INT_RST28         0x05 - RST 28 (used by WinIDE debugger)
                   0x06 - Syscall (*)
 INT_RST38         0x07 - RST 38
 INT_SLAVE_PORT    0x08 - Slave port
                   0x09 - Write Protect (*)
 INT_TIMER_A       0x0A - Timer A
 INT_TIMER_B       0x0B - Timer B
 INT_SERIAL_A      0x0C - Serial A
 INT_SERIAL_B      0x0D - Serial B
 INT_SERIAL_C      0x0E - Serial C
 INT_SERIAL_D      0x0F - Serial D
          --- EIR and rest of IIR overlapped ---
 INT_INT0          0x10 - INT0
 INT_INT1          0x11 - INT1
                   0x12 -
                   0x13 -
                   0x14 -
                   0x15 -
                   0x16 -
                   0x17 - PWM (*)
                   0x18 - Mode Violation (*)
 INT_QUAD_DECODER  0x19 - Quadrature decoder
 INT_INPUT_CAP     0x1A - Input capture
                   0x1B - Stack Limit (*)
 INT_SERIAL_E      0x1C - Serial E (*)
 INT_SERIAL_F      0x1D - Serial F (*)
                   0x1E - Network Port A (*)
                   0x1F - Timer C (*)

 (*) indicates interrupts not available on the Rabbit 2000 (some where not
     introduced until the Rabbit 3000A)
*/


extern void _near _TimerBISR(void);

//-----------------------------------------------------------------------------
//  Modify the entries in the interrupt vector table so ISRs are protected
//  against preemption.
//
//  An example of why this is needed in some cases:
//
//  Suppose an ISR starts by performing some time-critical function like
//  reading a character just received on a serial port, or disabling the
//  source of the interrupt.  It then restores the processor to the
//  interrupt priority level that existed before the interrupt occured so
//  other interrupts of the same level (or possibly even lower) could occur.
//
//  Once the ISR restores the interrupt priority level, a timer interrupt
//  could happen, and if the processing of that interrupt causes the
//  scheduler to be called, then the ISR that interrupted the current task
//  would be suspended along with the current task.  It would not get to
//  finish its processing until the task it interrupted was resumed.
//
//  Such a delay could be bad enough in its own right.  But what is likely
//  even worse is if the same interrupt happens again before the processing
//  for the previous instance finishes.  At the very least, the results
//  will probably not be correct.  Worse case would be that is causes
//  chaos.  And if it happened enough times, it might even overflow the
//  stack of one or more processes.
//
//  The method for dealing with this is as follows:
//    - The simple jump instruction that is normally put in each location of
//      the interrupt vector table is modified to do the following:
//         - disable the scheduler
//         - call the ISR
//         - enable the scheduler
//         - return control to the interrupted task
//
//  This can all be done (including saving and restoring the state of the
//  two registers affected) in 13 bytes, which will fit within the 16 bytes
//  available for each vector.
//
//  So, the added overhead is minimal, and does not require any additional
//  use of the root CODE space.
//
//  === NOTES ===
//  Substitutes for the Softools functions for modifying and accessing the
//  interrupt vector table should also be provided to minimize the chances of
//  some other code modifying the table in an incompatible manner.
//
//  For now, we rely on periodic scanning of the table to see if it has been
//  changed (and then change any that were back to the "shim" again).  This
//  should be left as a safeguard even after the substitute functions are
//  written to minimize the risk should some other code modify these later
//  on.
//
//
//  There is one problem with how the shims work which stems from the use
//  of a call instruction rather than a jump instruction:
//
//  Normally the interrupt vectors contain just a jump instruction to
//  transfer control to the ISR.  Because the return address was pushed
//  on the stack when the interrupt occured, the IRET (actually an IPRES
//  - a privileged instruction - followed by a RET) at the end of the
//  ISR insures that no interrupts of the same priority can happen until
//  control is transfered back to the originally interrupted code.
//
//  The method below, however, uses a call to transfer control to the
//  ISR so the IRET at the end of the ISR means that the shim itself
//  can be interrupted between the return from the ISR and the return
//  at the end of the shim.  Not REAL likely, but it can (and I believe
//  has) happen(ed) on at least one of our controllers with a digital
//  knob connected to an interrupt input.  And if (as in the one case
//  I have had it happen), you get a lengthy stream of interrupts, it can
//  cause a stack overflow.
//
//  There are, of course ways around this, but it would require more
//  instructions than will fit in the interrupt vector, which would
//  mean more overhead and higher latencies (because we would have to
//  call a function that would call the ISR, adding another call and
//  return, plus the additional code to save and restore the interrupt
//  priority register).
//
//  And last but not least, the case that has been known to cause a stack
//  overflow is a problem with how fast the interrupts occur.  While no
//  shim at all would eliminate the stack overflow issue, the high rate
//  of interrupts would still be a problem as the CPU would still be doing
//  nothing but servicing the same interrupt over and over.
//
  static uchar intVectInstrs[16] =  {
    0xF5,              //  push  AF
    0xE5,              //  push  HL
    0x21, 0x00, 0x00,  //  ld    HL,mn  (n,m: addr of _disable flag/counter)
    0x34,              //  inc  (HL)
    0xCD, 0x00, 0x00,  //  call  mn  (n,m:  addr of ISR)
    0x35,              //  dec  (HL)
    0xE1,              //  pop   HL
    0xF1,              //  pop   AF
    0xC9,              //  ret
    0x00,
    0x00,
    0x00
};
// # of bytes pushed on the stack (incl the return addr pushed by call to ISR)
//#define IntShim_PushCount  6   // moved to unistd.h

// Any changes to the shim may require udpating viableCodeAddr() (at least)


//--------------------------------------------------------------
void checkIntShims(void)
{
  static bool firstTime = true;
  uchar  IIR, EIR;
  uchar  *ptr;
  uint  u;


  IIR = _asm( [_A], "ld a,iir", _A );
  EIR = _asm( [_A], "ld a,eir", _A );

  if (firstTime)  {
    printf("\r\nIIR: %02X,  EIR: %02X\r\n", IIR, EIR);
    printf("_TimerBISR: %04X\r\n", (uint)_TimerBISR);
  }

//--- Replace the code for each vector that doesn't yet or no longer has the ---
//--- "shim" code to make it scheduler-safe ---
//   NOTE: This assumes the IIR and EIR tables are contiguous or overlap
  ptr = (uchar *)((uint)IIR << 8);
  for (u=0; u<NumIntVectors; ++u,ptr+=16)  {
    if (*ptr != 0xC3)  continue;  // verify existing JP instr
    *((uint *)(intVectInstrs + 3)) =
       (u == INT_TIMER_B) ? (uint)(&nestedTimerInt) : (uint)(&_disabled);
    if (!firstTime) fprintf(DiagLog, "*** INT %02X changed to: %04X ***\n",
                                     u, *((uint *)(ptr+1)) );
    *((uint *)(intVectInstrs + 7)) = *((uint *)(ptr + 1)); // addr from cur JP instr
    _EnterCritical_();  memcpy(ptr, intVectInstrs, 16);  _ExitCritical_();
  }

  firstTime = false;



/*
    wrappedHandler[i] = getInterruptHandler(i);

//    printf("INT %02X: %04X\r\n", i, (uint)wrappedHandler[i]);  //tdebug

    if (i == INT_TIMER_B)  continue;

    _EnterCritical_();  setInterruptHandler(i, shimFuncs[i]);  _ExitCritical_();
  }
*/

//  printf("\r\n");
//  for (i=0; i<16; ++i)
//    printf("INT %02X: %04X\r\n", i, (uint)getInterruptHandler(i) );
}

//-----------------------------------------------------------------------------
//  Called to dump the state of the machine just before doing a reboot due to
//  recover from a detected error.  This information is written to the circular
//  in-memory syslog buffer and to the crash-log file.
//
//  NOTE:  On Rabbit 3000 systems, the memory protection capability is used to
//  protect the CODE, CONST, FARCODE, and FARCONST segments from being written
//  to.  However, we have found that environmental conditions can still result
//  in these segments being corrupted without triggering a fault condition.
//  As it ONLY occurs while the cyclotron is in operation, it could be any
//  number of things: heat, EMI, power fluctuations, even energetic particles.
//
//  At present, the most common circumstance that triggers a dump and reboot
//  is when the idle task has detected corruption in one of the protected
//  segments, which it does by continuously recomputing checksums on them.
//  In most cases, the corruption does not appear to be in this code, or the
//  code it depends on, but there is no guarantee that will always be the case.
//-----------------------------------------------------------------------------
void postMortem(const char far *cause)
{
  PID  i;
  task_t  *task;
  ulong  absPC, idle, stackStart, stackEnd;
  uint  xpc, pc, sp, ui, log_sp, ip;
  uchar far *stackPtr;
  char  temp[32];


  if (recursiveDump)  return;
  recursiveDump = true;

  KLOCK();  WDT_2S();  checkWDT = false;

  useSoftReset = doShutdown = doReboot = true;

  logDateTime = false;

  openCrashLog();  // copy all diagnostic msgs to the crash log as well

//printf("stdoutDev:[%05lX], logStdout:%d, redirectStdio:%d\r\n",  //tdebug
//       (ulong)FD[userPix][1], current->logStdout, current->redirStdio);

  dateTimeMS(temp, TIME(), milliTicks);

  printf("\r\n"
         "\r\n"
         "---------- Post Mortem dump ----------\r\n"
         "%s %.24s[%d]: ", temp, current->argv[0], current->pid);

  printf("  Core Module ID: %.6s\r\n"
         "\n"
         " *** %.128ls ***\r\n"
         "\n"
         "  RTOS: %.40ls\r\n"
         "  APPL: %.40ls  %.40ls\r\n"
         "\n"
         "  kernel lock level: %d\r\n"
         "\n"
         "  Age of resetWDT() calls:  kernel: %ldms  applic: %ldms  any: %u\r\n"
         "\n",
         moduleID, cause, osDTM, applDTM, applName, _disabled,
         GetMS() - kernelLastCalled,
         GetMS() - applicLastCalled,
         diagLogInfo->wdtTicks
        );

/*-----------------------
 Using a combination of what is in the task table and what is stored in its
 stack, we can determine the state of each task at the time the error occured.

 When postMortem() is called, each task will be in one of 4 states:

   - The task called a system function that called schedule().  The task's
     preempt_count is not changed in this case.

   - The task was pre-empted (i.e sched_tick() called schedule(),
     incrementing the task's preempt_count by 1 just before doing so)

   - The task released the scheduler/kernel lock by calling KUNLOCK(),
     which found the pending flag set (which means that sched_tick()
     determined that a call to schedule() was due, but the scheduler was
     disabled by the current task).  In this case, KUNLOCK() effectively
     preempts the task (it increases the task's preempt_count by 8 and
     calls schedule()), allowing the pending call to schedule() to finally
     happen.

   - The task was the one running when the error occured (i.e. it was the one
     that called this function)


 In the first three cases, the task table contains the stack pointer value for
 the task when it was suspeneded.  In addition, the preempt_count value tells
 us which of the first three cases applies.  With these two values, and the
 knowledge of what was pushed on to the stack in each case, we can determine
 the location in memory where we can find the pieces of information that are
 NOT stored in the task table.

 For the last case, the data in the task table is not up to date (it is only
 updated when a task is suspended), so some of the values we need are copied
 from the live register values.


 For the current task, the task table value for SP (stack) is outof date, so we
 grab the current value for it (and values we normally get from the stack)
 right from the live registers.

        ---------------------------------------------------------------
 For every suspended task (i.e. every one BUT the current one), this is ---
 what the top of the task's stack looks like:

--- pushed by schedule() (see Schedule_PushCount in linux/sched.h)---
  stackPtr[0]:    IX (LSB)
  stackPtr[1]:    IX (MSB)

  stackPtr[2]:    IY (LSB)
  stackPtr[3]:    IY (MSB)

  stackPtr[4]:    IP

--- from call to schedule() (from sys call or sched_tick(): 3 bytes) ---
  stackPtr[5]:    ret_addr (LSB)
  stackPtr[6]:    ret_addr (MSB)

  stackPtr[7]:    XPC


 The ret_addr at stackPtr[5] provides the address from which schedule() was
 called.  But if the task was preempted (task->preempt_count > 0), then
 that call came either from sched_tick() or KUNLOCK(), which doesn't tell
 us anything about what the task was doing when it was preempted.


        ---------------------------------------------------------------
 If it was preempted by KUNLOCK(), then this is what follows the above:

--- pushed by task's call to KUNLOCK() ---
  stackPtr[8]:    ret_addr (LSB)  (addr following task's call to KUNLOCK())
  stackPtr[9]:    ret_addr (MSB)


        ---------------------------------------------------------------
If it was preempted by sched_tick(), then this is what follows the above:

 --- pushed by _TimerBISR() (see TimerBISR_PushCount in timerb.h) ---
  stackPtr[8]:    ret_addr (LSB)  (addr inside _TimerBISR())
  stackPtr[9]:    ret_addr (MSB)

  stackPtr[10]:   XPC

  stackPtr[10]:   AF' (LSB)
  stackPtr[11]:   AF' (MSB)

  stackPtr[12]:   BC' (LSB)
  stackPtr[13]:   BC' (MSB)

  stackPtr[14]:   DE' (LSB)
  stackPtr[15]:   DE' (MSB)

  stackPtr[16]:   HL' (LSB)
  stackPtr[17]:   HL' (MSB)

  stackPtr[18]:   BC (LSB)
  stackPtr[19]:   BC (MSB)

  stackPtr[20]:   IY (LSB)
  stackPtr[21]:   IY (MSB)

  stackPtr[22]:   IX (LSB)
  stackPtr[23]:   IX (MSB)

  stackPtr[24]:   HL (LSB)
  stackPtr[25]:   HL (MSB)

  stackPtr[26]:   DE (LSB)
  stackPtr[27]:   DE (MSB)

  stackPtr[28]:   AF (LSB)
  stackPtr[29]:   AF (MSB)

 --- pushed by ISR table "shim" (see IntShim_PushCount) ---
  stackPtr[30]:   ret_addr (LSB)  (addr within ISR table)
  stackPtr[31]:   ret_addr (MSB)

  stackPtr[32]:   HL (LSB == L)
  stackPtr[33]:   HL (MSB == H)

  stackPtr[34]:   DE (LSB == E)
  stackPtr[35]:   DE (MSB == D)

  stackPtr[36]:   AF (LSB == F)
  stackPtr[37]:   AF (MSB == A)

 --- pushed by CPU when the timerB interrupt occurs ---
  stackPtr[38]:   ret_addr (LSB)  (addr where task was interrupted)
  stackPtr[39]:   ret_addr (MSB)

NOTE:

        ---------------------------------------------------------------
 For the current task, this is what the in-use stack looks like:

--- pushed by postMortem() ---
  stackPtr[0]:    IX (LSB)
  stackPtr[1]:    IX (MSB)

--- pushed by call to postMortem() ---
  stackPtr[2]:    ret_addr (LSB)
  stackPtr[3]:    ret_addr (MSB)

  stackPtr[4]:    XPC

   Keep in mind that postMortem() is called from a number of places, including
   dumpFault().  Basically, any code that detects an unrecoverable error will
   call postMortem() (compared to dumpFault(), which is called ONLY for
   run-time fault interrupts generated by the CPU).

---------------------*/

  printf(
 "                                    ---------Stack--------- \r\n"
 "  PID  PRI  STATUS            IDLE  Start  End   SP  Unused  --PC--  --IP---  CMD\r\n");
//#####  ###  ######## #  ########ms  ##### ##### #####  ####  ######  #-#-#-#  ############

//--- log info about the state of each task ---
  for (i=0; i<NUM_TASKS; ++i)  if ((task = tasks + (uint)i)->inUse)  {
    WDT_2S();

    idle = msSince(task->lastRunning);

    if (task == current)  {
      sp = _asm( [_HL], "ld hl,ix", _HL );  // IX = SP after IX itself was pushed
      stackPtr = (uchar far *)( ((ulong)(task->sseg) << 12) + (ulong)sp );
      pc = *(uint far *)(stackPtr + 2);
      xpc = (uint) stackPtr[4];
      ip = _asm( [_HL], "push ip", "push ip", "pop hl", _HL ) & 0xFF;
    }
    else  {  // suspended tasks (i.e. all BUT the current one)
      sp = (uint)task->stack;
      stackPtr = (uchar far *)( ((ulong)(task->sseg) << 12) + (ulong)sp );
      if (!task->preempt_count)  {  // schedule() called by kernel func
        pc = *(uint far *)(stackPtr + Schedule_PushCount);
        xpc = (uint) stackPtr[Schedule_PushCount + 2];
      }
      else if (task->preempt_count >= 8)  {  // schedule() called by KUNLOCK()
        pc = *(uint far *)(stackPtr + Schedule_PushCount + 3);
        xpc = (uint) stackPtr[7];
      }
      else  {  // schedule() called by sched_tick()
        pc = *(uint far *)(stackPtr + Schedule_PushCount + 3 +
                           TimerBISR_PushCount + IntShim_PushCount);
        xpc = (uint) stackPtr[10];
      }
      ip = (uint)stackPtr[4];
    }

    absPC = (ulong)pc;
    if (pc >= 0xE000)  absPC += ((ulong)xpc << 12);

    stackStart = ((ulong)task->stack_start + (((ulong)task->sseg) << 12));
    stackEnd = stackStart + task->stack_size - 1;

    printf("%5d  %3d  %s%2d  %8lums  %05lX %05lX %05lX  %4d  %06lX  %X-%X-%X-%X  %.64s\r\n",
            task->pid, task->priority, statusDescr(i), task->preempt_count,
            idle, stackStart, stackEnd, (ulong)stackPtr, _stackFree_(task->pid),
            absPC, ip >> 6, (ip >> 4) & 3, (ip >> 2) & 3, ip & 3,
            task->argv[0] );

    if (task == current)  {
      printf(" Stack: ");
      for (ui=0,log_sp=sp;
            (ulong)stackPtr<=stackEnd; ++ui,++stackPtr,++log_sp)  {
        if ((ui & 0x0F) == 0)
          printf("%05lX (%04X): ", (ulong)stackPtr, log_sp);
        printf("%02X ", *stackPtr);
        if ((ui & 0x0F) == 0x0F)  printf("\r\n        ");
      }
      printf("\r\n");
    }

  }

  printf("\n\n");  WDT_2S();


  showMemProtection();

  showIntVectors();



//--- recompute and log the CRCs for the code and const segments ---
  computeMemCRCs(false, true);

  WDT_2S();

  printf("doCleanShutdown:%d  useSoftReset:%d  allowSoftReset:%d\r\n",
         doCleanShutdown, useSoftReset, allowSoftReset);

  closeCrashLog();


  logDateTime = true;


  if (doCleanShutdown AND (current != kernelTask))  for (;;)  {
    while (KLOCKED())  KUNLOCK();
    if (current == applTask)  {
      resetWDT();  current->state = TASK_UNINTERRUPTIBLE;  }
    else  {
      current->state = TASK_STOPPED;
      current->sig_pending |= sigmask(SIGKILL);
    }
    schedule();
  }


  REBOOT(true);
}



//-----------------------------------------------------------------------------
//  Called when the watchdog timer is about to reboot the system.  This gives
//  us a chance to copy important diagnostic information to the log, which
//  will be accessable after we reboot.
//-----------------------------------------------------------------------------
void wdTimingOut(void)
{
  KLOCK();  WDT_2S();  openCrashLog();

  printf("\r\n***** WatchDog Timer reboot imminent! *****\r\n");

  doCleanShutdown = false;

  postMortem("WatchDog Timeout");
}


//-----------------------------------------------------------------------------
//  Test an address to see if it is a viable return address
//-----------------------------------------------------------------------------
bool  viableCodeAddr(ulong addr)
{
  if (IN_CODE(addr))  return true;

  if (IN_FARCODE(addr))  return true;

//--- The return addr for all ISR "shims" is 9 bytes into the table entry ---
  if (IN_ISR_TBL(addr))  if (((uint)addr & 0x000F) == 9)  return true;

  return false;
}



//--- values for flags argument ---
#define  GuessedXPC    0x0001
//-----------------------------------------------------------------------------
//  Generate a msg showing the following info on an (alleged) function call:
//    - The address of the called function (if not an indirect call)
//    - The address where the call was made from
//    - The (logical) address on the stack of the (alleged) return address
//-----------------------------------------------------------------------------
void showMatch(ulong toAddr, ulong fromAddr, uint flags, uchar *sp)
{
  static uchar  *prevSP;
  static ulong  prevTo, prevFrom;
  char *str;


  if (!toAddr)  return;

  if ((sp == prevSP) AND (toAddr == prevTo) AND (fromAddr == prevFrom))  return;

  str = (flags & GuessedXPC) ? "(?)" : "";

  if (sp != prevSP)  printf("\r\n");

  if ((toAddr == (ulong)CallDE) OR (toAddr == (ulong)CallHL))
    printf("  ?????");
  else
    printf("  %05lX", toAddr);

  printf(" called from %05lX [%04X] %s\r\n", fromAddr, (uint)sp, str);

  prevTo = toAddr;  prevFrom = fromAddr;  prevSP = sp;
}

//-----------------------------------------------------------------------------
//  Attempt to provide a reasonably accurate stack trace using heuristic
//  methods to deduce the sequence of function calls that preceeded a hardware
//  protection fault or an unhandled interrupt.
//
//  Arguments:
//    sp:
//       Logical addr within the current stack to begin looking.
//    endStack:
//       The logical addr of the end of the current stack
//  -------------------------------------------------------------------------
//
//  There is no information on the stack that makes a stack trace easy and
//  reliable.  However, using the following facts and some assumptions based
//  on the code generated by the compiler, we should still be able to reproduce
//  the sequence of calls that got us here.
//
//  NOTE:  There are many false positives in the results, which I expect is due
//         to the following fact:  Any local variable that a function doesn't
//         use means the location on the stack corresponding to that variable
//         will retain whatever value it had before.  For variables like
//         strings, there is a good chance that a significant size chunk of
//         the data on the stack has values from past function call events
//         that are not along the path leading to the present call to this
//         function.
//
//         The bytes that DO contain the return addresses for the calls along
//         the path should still be there, which means we should be able to
//         find them.  But experience has shown that the majority of the
//         candidates we find do NOT actually represent calls that are part
//         of the sequence that lead to this function being called, which means
//         we have to manually illiminate all the false positives to make
//         sense out of the result.
//
//
//    - Return addresses for function/subroutine calls will point to an address
//      that follows a CALL, LCALL, or some equivalent set of instructions
//      (like one or more PUSH instructions followed by a JP or LJP).
//
//    - A call from (or interrupt of) a far function will push a 2-byte address
//      that is >= 0xE000.  For a near function, the address will be < 0xD000.
//
//    - Return addresses will always be within one of the CODE segments or the
//      interrupt vector table.
//
//    - The presence (or lack of) an XPC value on the stack is dependent upon
//      whether the call was TO a near function or a far function.  It has
//      nothing to do with what type of function the call was FROM.  An LCALL
//      in a near func still pushes the current XPC value, and a CALL in a far
//      function does NOT.
//
//      Since we don't know with any certainty if the call sequence we are
//      trying to reconstruct is accurate, we cannot know for certain if we
//      are looking for a CALL or LCALL (or their equivalent) to preceed a
//      potential return address.  And the use of indirect calls (particularly
//      to a redirector for a far function) just makes any attempt to determine
//      such information even more difficult.
//
//      However, there is still one use we can get out of the difference
//      between CALL and LCALL:  When testing a potential return address, the
//      presence of an LCALL at the appropriate address can be further verified
//      by the fact that a real LCALL pushes the XPC value on the stack.  Which
//      means that if the next byte on the stack isn't a viable XPC value, then
//      the value being tested is isn't really a return address.
//
//    - DEBROS uses the interrupt vectors to disable the scheduler while an
//      interrupt is being serviced, and it calls the ISR from the vector.
//      This means the return address (from the point of view of the ISR) will
//      always be at a specific offset from the start of the vector.  This
//      makes it easy to determine when an interrupt has happened and makes it
//      easy to find the return address that was pushed when the interrupt
//      occured (since it will be at a fixed offset from the one the vector
//      pushed).
//
//      This is important because the address pushed on the stack when the
//      interrupt occured can be for any address in the CODE or FARCODE
//      segments, and there won't be any CALL or LCALL (or equiv instructions)
//      preceeding the address.
//
//
//  Since we can't, as stated, have any real confidence in our reconstruction
//  of the sequence of calls, here is how we use the above facts:
//
//    - For each pair of bytes on the stack, check to see if they form an
//      address within the CODE or FARCODE segment, or at the appropriate
//      offset within one of the entries of the interrupt vector table.  If the
//      value doesn't meet one of these criteria, advance to the next byte in
//      the stack.
//
//    - For each potential return address in the CODE or FARCODE segments,
//      check for a preceeding CALL or LCALL instruction at the appropriate
//      address.  If neither is found, then advance to the next byte in the
//      stack.
//
//    - If a CALL is found, make sure the called address is in the CODE segment.
//      If an LCALL is found, make sure the called address is in the FARCODE
//      segment, and that the next byte on the stack contains an XPC value that
//      matches the return address currently being tested.
//
//      Remember that, if the LCALL being tested is in the CODE segment, the
//      only verification we can do on the potential XPC value on the stack is
//      to determine whether or not it can point to addresses within the
//      FARCODE segment.
//
//    - For potential return addresses that point to an interrupt vector, check
//      the appropriate bytes on the stack to make sure they form an address
//      that passes the first test (i.e. the address is within one of the CODE
//      segments or the interrupt vector table).
//  -------------------------------------------------------------------------
//
//  NOTES/Thoughts:
//
//     The stack trace is not MISSING calls, so all the false matches are
//     coming from values on the stack that represent either arguments pushed
//     on the stack before a call or local variables for the called functions.
//
//     The only plausible reason I can think of for there being so many false
//     positives is that they are leftover from earlier calls along different
//     branches that happen to reside in a memory location for a local
//     variable that was not used by the sequence of functions along the path
//     that lead to the stack trace.
//
//  False vs real matches:
//
//     The existence of a "add sp,nn" instruction at a return address could
//     tell me how many bytes following a return address could be ignored.
//     But if the call is a false match, then that could easily cause me to
//     ignore real matches.
//
//     Then again, it is likely that the pushed argument values will overwrite
//     leftover return addresses.  What is MORE likely is that the leftover
//     values are occupying local variable space for variables that were not
//     written to before the function they belong to called another function.
//
//     In that case, I would want to look for a "sub sp,nn" (or whatever) at
//     the start of a function.  But I don't know where to look for that until
//     AFTER I have already processed the bytes I would most likely want to
//     skip.  And the effect of using this information for false matches is
//     still the same.
//
//  XPC values, near vs far calls:
//
//     The redirector method used by the Softools compiler means that near
//     calls can lead to far functions.  And the fact that the most common
//     (possibly the only) use of this approach is for indirect calls (i.e.
//     a pointer refers to the redirector) means there is no practical way
//     to determine WHAT was called by the redirector.
//
//     However, it is worth remembering that an XPC value can be useful
//     in tracing backwards beyond a single call.  If the potential return
//     address matches a call from a near function to a far function (which
//     could be a near call to a redirector), then the XPC value doesn't
//     tell us anything about the call that (potentialy) put it on the
//     stack, but it does help filter out false matches for an earlier call.
//     Specifically, as we trace backwards along the sequence of calls, we
//     can throw out any calls to near funcs from far funcs that don't fit
//     the XPC value until we find one that does.
//
//
//  Build a tree of possible call sequences:
//
//     The more I think about it, the only way I can think of to filter out
//     the false positives is to build a tree of possible call sequences.
//
//     Like I did with my original (less than accurate) attempt at a stack
//     trace, I need to take in to account the possible matches preceeding
//     the ones currently under consideration to help eliminate false
//     positives.  However, rather than merely keeping a fixed-size list of
//     recent possible called addresses, I need to make it a tree.
//
//     Using the address I am tracing backwards from as the root, what I
//     do is test each potential return address against every node in the
//     tree.  Each one that could be call leading to that node is added
//     as a child of that node.
//
//     When I am done, I print out all the sequences that lead back to
//     an address within the taskStart() function.
//
//     (Actually, that last part doesn't really work, since the call from
//     taskStart() is an indirect one, so it would match ALL the leaves on
//     the tree, so they would ALL pass that test!)
//
//     But, while there may be many trees, they should each be reasonably
//     short, so perhaps printing them all out is not a big deal.
//
//     However, to keep the dump process as quick as possible (and to avoid
//     using up even MORE memory on the controller for the additional code
//     and the data), this is probably best left as something to do in the
//     application that currently parses a dump file and a map file and adds
//     the symbolic information.
//
//-----------------------------------------------------------------------------
void stackTrace(uchar *sp, uchar *stackEnd)
{
  uint  retAddr,  prevAddr, intNum;
  ulong  testAddr, toAddr, xpcWindow;
  uchar far  *instrAddr;


#define  VIABLE_RET_ADDR(addr)  \
  (IN_CODE(retAddr) OR IN_FARCODE(retAddr) OR IN_ISR_TBL(retAddr))


  prevAddr = 0;

  for  (; sp < stackEnd; ++sp)  {
    retAddr = *(uint *)sp;  WDT_2S();

    if ((retAddr == 0xFFFF) AND (prevAddr == 0xFFFF))  continue;

    prevAddr = retAddr;

//    printf("    %04X: %04X\r\n", (uint)sp, retAddr);  //tdebug

    if (!viableCodeAddr(retAddr))  continue;


  //--- test for a call from an interrupt vector ---
    if (IN_ISR_TBL(retAddr))  {
      intNum = (retAddr - intVectStart) >> 4;
      retAddr = *(uint *)(sp + IntShim_PushCount);
      if (!viableCodeAddr(retAddr))  continue;
      printf("\r\n*** Interrupt 0x%02X occured at ", intNum);
      if (retAddr < 0xE000)
        printf("%05X", retAddr);
      else
        printf("??:%04X", retAddr);
      printf(" ***\r\n");  continue;
    }


    toAddr = 0;

  //===== NEAR call from NEAR func =====
  //---  For this case, there is no XPC value on the stack and the retAddr ---
  //--- tells us precisely where to look for a CALL instruction ---
    if (IN_CODE(retAddr))  {
      instrAddr = (uchar far *)retAddr - 3;
      if (*instrAddr == 0xCD)  {   // CALL mn (n m)
        toAddr = *(uint far *)(instrAddr+1);
        if (IN_CODE(toAddr))
          showMatch(toAddr, (ulong)instrAddr, 0, sp);
      }
    }


  //==== FAR call from FAR func =====
  //--- For this case, the return addr on the stack includes an XPC value ---
  //--- and it tells us precisely where to look for an LCALL instr ---
    if (retAddr >= 0xE000)  {
      testAddr = (ulong)retAddr + ((ulong)sp[2] << 12);
      instrAddr = (uchar far *)testAddr - 4;
      if (*instrAddr == 0xCF)  {  // LCALL x,mn  (n m x)
        toAddr = (ulong) *((uint far *)(instrAddr+1))
                   + ((ulong)instrAddr[3] << 12);
        if (IN_FARCODE(toAddr))
          showMatch(toAddr, (ulong)instrAddr, 0, sp);
      }
    }


  //===== FAR call from NEAR func =====
  //--- For this case, the return addr on the stack contains an XPC but it ---
  //--- has nothing to do with where we look for the LCALL instr.  However, ---
  //--- the XPC value DOES have to be one that at least starts in FARCODE ---
    if (IN_CODE(retAddr))  {
      xpcWindow = 0x0E000L + ((ulong)sp[2] << 12);
      if (IN_FARCODE(xpcWindow))  {
        instrAddr = (uchar far *)retAddr - 4;
        if (*instrAddr == 0xCF)  {  // LCALL x,mn  (n m x)
          toAddr = (ulong) *((uint far *)(instrAddr+1))
                     + ((ulong)instrAddr[3] << 12);
          if (IN_FARCODE(toAddr))
            showMatch(toAddr, (ulong)instrAddr, 0, sp);
        }
      }
    }


  //===== INDIRECT calls to a FAR func ====
  //--- Such calls are accomplished by doing a near call to a redirector ---
  //--- function using __callde() or __callhl().  The redirector pops the ---
  //--- logical return addr and does an LCALL to a stub preceeding the far ---
  //--- func which replaces the 2-byte logical return addr (from the ---
  //--- redirector) with the popped addr from the original near call to the ---
  //--- redirector.  The result is the logical return addr + the XPC value ---
  //--- the far func needs to do an LRET to the orig caller.  The retAddr ---
  //--- being tested tells us whether the call came from a near or far ---
  //--- function, which tells us whether the XPC value on the stack is ---
  //--- part of the return address ---

  //===== INDIRECT call to a FAR from a FAR func ====
  //--- This is very similar to a "FAR call from FAR func" except that it ---
  //--- uses a CALL to an indirection func rather than a direct LCALL to ---
  //--- the target function ---
    if (retAddr >= 0xE000)  {
      testAddr = (ulong)retAddr + ((ulong)sp[2] << 12);
      if (IN_FARCODE(testAddr))  {
        instrAddr = (uchar far *)testAddr - 3;
        if (*instrAddr == 0xCD)  {   // CALL mn (n m)
          toAddr = *(uint far *)(instrAddr+1);
          if ((toAddr == (ulong)CallDE) OR (toAddr == (ulong)CallHL))
            showMatch(toAddr, (ulong)instrAddr, 0, sp);
        }
      }
    }

  //===== INDIRECT call to a FAR from a NEAR func ====
  //--- This is very similar to a "FAR call from NEAR func" except that it ---
  //--- uses a CALL to an indirection func rather than a direct LCALL to ---
  //--- the target function ---
    if (IN_CODE(retAddr))  {
      xpcWindow = 0x0E000L + ((ulong)sp[2] << 12);
      if (IN_FARCODE(xpcWindow))  {
        instrAddr = (uchar far *)retAddr - 3;
        if (*instrAddr == 0xCD)  {   // CALL mn (n m)
          toAddr = *(uint far *)(instrAddr+1);
          if ((toAddr == (ulong)CallDE) OR (toAddr == (ulong)CallHL))
            showMatch(toAddr, (ulong)instrAddr, 0, sp);
        }
      }
    }


  //===== NEAR call from FAR func =====
  //--- For this case, there is no XPC value on the stack, but we DO care ---
  //--- what it was since such a call would have come from memory mapped by ---
  //--- the XPC reg.  Since we don't have any way to know what it was, all ---
  //--- we can do is try every viable XPC value and look for a valid-looking ---
  //--- CALL instruction ---
  //  NOTE:  I STILL cannot believe how many legitamate looking calls this
  //         finds!  Anywhere from 1 to 4 of the tested XPC values finds
  //         the value CD at the same offset from the 4K XPC boundaries,
  //         followed by a 2-byte value that is within the range of addrs
  //         in the CODE segment).   Even MORE unbelievable is that the
  //         addresses actually ARE valid function pointers in almost
  //         every case.  Which suggests that these really ARE function
  //         calls being found, and just by chance happen to be at an
  //         address whose least-significant 12-bits just happens to be
  //         the same as 1, 2, even 3 other function calls AND matches
  //         the least 12 significant bits of the value from the stack
  //         that is being tested.
    if (retAddr >= 0xE000)  {
      testAddr = (ulong)retAddr;
      for (; testAddr<=farCodeEnd; testAddr+=0x1000L)  {
        instrAddr = (uchar far *)testAddr - 3;
        if (*instrAddr == 0xCD)  {   // CALL mn (n m)
          toAddr = *(uint far *)(instrAddr+1);
          if (IN_CODE(toAddr))
            showMatch(toAddr, (ulong)instrAddr, GuessedXPC, sp);
        }
      }
    }

  }

}


//-----------------------------------------------------------------------------
//  The current task caused a hardware protection fault or invoked an interrupt
//  for which no handler was set.
//
//  This function dumps the state the processor had at the time the fault
//  occured.  It should ONLY be called by the ISR functions that handle one
//  of the hardware detected run-time faults.
//-----------------------------------------------------------------------------
void dumpFault(const char far *context)
{
  uint  ui;
  uchar  *sp, *ptr, *stackEnd;
  ulong  faultAddr, absAddr;
  char  temp[32];


  KLOCK();  WDT_2S();  openCrashLog();  ipset1();  logDateTime = false;

  useSoftReset = doShutdown = doReboot = true;

/*====================================================================*
  ===== This is what the stack should look like at this point: =====

      SP ==> (logical addr of last byte pushed on to stack)

    ---- pushed by this function (2 bytes) ---
              (?) local variables
    sp[-36]   (2) IX

    --- pushed by interrupt/fault handler (28 bytes) ---
    --- (e.g. WriteFault(), StackLimit(), ...)
    sp[-34]   (2) logical return address
    sp[-32]   (1) XPC of return address
    sp[-31]   (4) far pointer to context str
    sp[-27]   (2) AF'
    sp[-25]   (2) IX
    sp[-23]   (2) IY
    sp[-21]   (2) BC'
    sp[-19]   (2) DE'
    sp[-17]   (2) HL'
    sp[-15]   (2) AF  (F already altered by int vector shim)
    sp[-13]   (2) BC
    sp[-11]   (2) DE
    sp[-9]    (2) HL  (set to addr of _disable by int vector)
    sp[-7]    (1) IP

    --- pushed by interrupt vector "shim" (IntShim_PushCount bytes)---
    sp[-6]    (2) return addr within int vector
    sp[-4]    (2) HL
    sp[-2]    (2) AF

    --- pushed when fault occured ---
    sp[0]     (2) Addr of instr following the one that caused the fault

    --- everything that was on the stack before the fault occured ---
    sp[2]     (?) ...
 *====================================================================*/

//--- Assumes above is correct and sp = value of SP at time fault occured ---
#define  ReturnAddr  (*(uint *)(sp-6))

#define  REG_AF   (*(uint *)(sp-2))
#define  REG_HL   (*(uint *)(sp-4))
#define  REG_DE   (*(uint *)(sp-11))
#define  REG_BC   (*(uint *)(sp-13))

//#define  REG_AF_  _asm( [_HL], "exx", "push af", "pop hl", "exx", _HL )
#define  REG_AF_  (*(uint *)(sp-27))
#define  REG_HL_  (*(uint *)(sp-17))
#define  REG_DE_  (*(uint *)(sp-19))
#define  REG_BC_  (*(uint *)(sp-21))

#define  REG_IX   (*(uint *)(sp-25))
#define  REG_IY   (*(uint *)(sp-23))
#define  REG_SP   (sp)
#define  REG_PC   (*(uint *)(sp))

#define  REG_IP   (*(uchar *)(sp-7))
#define  REG_IIR  _asm( [_A], "ld a,iir", _A )
#define  REG_EIR  _asm( [_A], "ld a,eir", _A )
#define  REG_XPC  (*(uchar *)(sp-32))


  // When the fault occurs, the logical address at which it happened is pushed
  // on to the stack.  The interrupt vector "shim" then pushes IntShimCount
  // bytes (currently 6), the fault-handler pushes 26 bytes, and this function
  // pushes IX (2 more) before copying the value of SP to IX.  Here we get the
  // value SP had at that time from IX, then add to it so it points to the
  // logical address in the currently mapped in stack where the occurance of
  // a fault caused the value of PC to be saved.
  sp = (uchar *) _asm( [_HL], "ld hl,ix", _HL ) + 2 + 28 + IntShim_PushCount;

  faultAddr = REG_PC;
  if (faultAddr >= 0xE000)  faultAddr += (ulong)REG_XPC << 12;

  dateTimeMS(temp, TIME(), milliTicks);

  printf("\r\n\r\n%s %.24s[%d]:\r\n"
         " ***** %ls fault occured at: 0x%05lX *****\r\n"
         "\n"
         "Interrupt Vector: 0x%04X\r\n"
         "\n",
         temp, current->argv[0], current->pid, context, faultAddr,
         ReturnAddr & 0xFFF0 );

  faultAddr -= 16;
  printf("Code preceeding fault:\r\n  %06lX: ", faultAddr);
  for (ui=0; ui<16; ++ui,++faultAddr)
    printf("%02X ", *(uchar far *)faultAddr);
  printf("\r\n\n");

  printf("Registers:\r\n"
         "  AF: %04X  AF': %04X    IX: %04X    IP:  %02X\r\n"
         "  HL: %04X  HL': %04X    IY: %04X    IIR: %02X\r\n"
         "  DE: %04X  DE': %04X    SP: %04X    EIR: %02X\r\n"
         "  BC: %04X  BC': %04X    PC: %04X    XPC: %02X\r\n\n",
         REG_AF, REG_AF_, REG_IX, REG_IP,
         REG_HL, REG_HL_, REG_IY, REG_IIR,
         REG_DE, REG_DE_, REG_SP, REG_EIR,
         REG_BC, REG_BC_, REG_PC, REG_XPC );


  stackEnd = (uchar *)((uint)current->stack_start + current->stack_size - 1);

  printf("Stack:");
  absAddr = ((ulong)current->sseg << 12) + (uint)sp;
  for (ui=0,ptr=sp; ptr<=stackEnd; ++ui,++absAddr)  {
    if (!(ui & 0x0f))  printf("\r\n  %05lX (%04X): ", absAddr, (uint)ptr);
    printf("%02X ", *ptr++);  WDT_2S();
    if (ui & 0xFFC0)  { printf("...\r\n");  break; }
  }

  printf("\r\n\n"
         "Stack Trace:\r\n"
         "  %ls fault occured at: %05lX\r\n", context, faultAddr);

  stackTrace(sp, stackEnd-4);

  postMortem(context);
}

//-----------------------------------------------------------------------------
//  Override the standard UnhandledInterruptHandler (which just reboots).  Even
//  if this can't handle the unexpected interrupt, at least we will know that
//  it happened.
//----------------------------------
//  WARNING:  Do NOT add any local variables as this will change the contents
//            of the stack, causing dumpFault() to produce invalid information!
//-----------------------------------------------------------------------------
void _int UnhandledInterruptHandler( void )
{
//  save AF' on the stack
;
#asm
    exx
    push    af
    exx
#endasm
;
  outi(MMIDR, 0x80);  // enable access to registers > 0x00FF

  outi(STKCR, 0x00);  // disable hdw stack limit checks
  outi(WPCR, 0);  // disable write protection for System Mode

  outi(MMIDR, 0x00);  // disable access to registers > 0x00FF

  KLOCK();

  logDateTime = true;  fprintf(DiagLog, "Unhandled Interrupt error\r\n");

  dumpFault("Unhandled Interrupt");
}


//-----------------------------------------------------------------------------
//  The current task (or an interrupt routine) attempted to write to a memory
//  location in a block of write-protected memory.
//----------------------------------
//  WARNING:  Do NOT add any local variables as this will change the contents
//            of the stack, causing dumpFault() to produce invalid information!
//-----------------------------------------------------------------------------
void _int WriteFault( void )
{
//  save AF' on the stack
;
#asm
    exx
    push    af
    exx
#endasm
;
  outi(MMIDR, 0x80);  // enable access to registers > 0x00FF

  outi(STKCR, 0x00);  // disable hdw stack limit checks
  outi(WPCR, 0);  // disable write protection for System Mode

  outi(MMIDR, 0x00);  // disable access to registers > 0x00FF

  KLOCK();

  logDateTime = true;  fprintf(DiagLog, "Write Protection error\r\n");

  dumpFault("Write Protection");
}


//-----------------------------------------------------------------------------
//  A user/system mode violation occured
//----------------------------------
//  WARNING:  Do NOT add any local variables as this will change the contents
//            of the stack, causing dumpFault() to produce invalid information!
//-----------------------------------------------------------------------------
void _int ModeViolation( void )
{
//  save AF' on the stack
;
#asm
    exx
    push    af
    exx
#endasm
;
  dumpFault("User/Sys Mode Violation");
}


//-----------------------------------------------------------------------------
//  The stack pointer for the current task exceeded the set limits
//----------------------------------
//  WARNING:  Do NOT add any local variables as this will change the contents
//            of the stack, causing dumpFault() to produce invalid information!
//-----------------------------------------------------------------------------
void _int StackLimit( void )
{
//  save AF' on the stack
;
#asm
    exx
    push    af
    exx
#endasm
;
  outi(MMIDR, 0x80);  // enable access to registers > 0x00FF

  outi(STKCR, 0x00);  // disable hdw stack limit checks
  outi(WPCR, 0);  // disable write protection for System Mode

  outi(MMIDR, 0x00);  // disable access to registers > 0x00FF

  KLOCK();

  logDateTime = true;  fprintf(DiagLog, "Stack Limit error\r\n");

  doCleanShutdown = false;

  dumpFault("Stack Limit");
}



//-----------------------------------------------------------------------------
//  As the two truly critical tasks, the kernel and applic tasks must each call
//  this function at least once every 60 seconds.  In addition, this function
//  must be called by SOME task at least once every 1900ms to avoid a forced
//  reboot (the sched_tick() function increments and checks the wdtTicks value
//  that is set to 0 here each time this function is called).
//-----------------------------------------------------------------------------
void resetWDT(void)
{
  bool  bothActive = false;


  KLOCK();

  if (current == kernelTask)  {
    kernelLastCalled = GetMS();
    if (msSince(applicLastCalled) < 60000L)  bothActive = true;
  }
  else  if (current == applTask)  {
    applicLastCalled = GetMS();
    if (msSince(kernelLastCalled) < 60000L)  bothActive = true;
  }

  if (bothActive)  {
    WDT_2S();  // reset 2-second countdown Watchdog timer
    _EnterCritical_();  diagLogInfo->wdtTicks = 0;  _ExitCritical_();
  }

  KUNLOCK();
}

//-----------------------------------------------------------------------------
typedef struct {
  const char far *segName;
  ulong  *start;
  ulong  *end;
  ulong  *size;
  uint   *origCRC;
  uint   *errCount;
} CRC_Params;

//-----------------------------------------------------------------------------
//  Compute, display, validate, etc a CRC for a portion of memory.  Called at
//  startup to compute the CRC values for memory segments whose data should
//  not change after startup (i.e. CODE, CONST, FARCODE, FARCONST), and
//  periodically after startup to check for corruption to these segments.
//
//  NOTE that for Rabbit 3000 processors, nearly all the memory occupied by
//  these segments is write protected, so an attempt to write to them due to
//  a software bug causes a runtime fault interrupt which is caught and
//  logged (without changing the contents of the protected memory), so any
//  changes to these segments indicates an external (enviornmental) cause.
//
//  Parameters:
//     firstTime:
//         != 0:  Set *origCRC to computed CRC value
//         == 0:  Compare computed CRC value to *origCRC value
//     display:
//         != 0:  Show result regardless of whether there is an error
//         == 0:  Don't show result unless there is an error
//     *name:
//         Name of the memory segment
//     start:
//         The address of the first byte in the memory segment
//     end:
//         The address of the last byte in the memory segment
//     size:
//         The length of the memory segment
//     *origCRC:
//         if firstTime != 0: The location to store the computed CRC value
//         if firstTime == 0: The CRC value computed for the segment at startup
//-----------------------------------------------------------------------------
void computeCRC(bool firstTime, bool display, const far char *name,
                ulong start, ulong end, ulong size, uint *origCRC)
{
  bool  match;
  uint  curCRC;


  if (display)  {
    printf("%ls  %05lX  %05lX  ", name, start, end);
    fflush(stdout);
  }

  curCRC = memCRC16(start, size);
  if (firstTime)  {
    *origCRC = curCRC;  match = true; }
  else
    match = (curCRC == *origCRC);

  if (!match)  memCorrupted = true;

  if (display OR !match)
    printf("%04X %s\r\n", curCRC, firstTime ? "" : (match ? "OK" : "*BAD*") );

  if (!match)  {
    printf("***** %ls failed CRC test *****\r\n", name);
    useSoftReset = doShutdown = doReboot = true;
  }

}

//-----------------------------------------------------------------------------
//  Recompute the CRCs for the areas of memory containing code and constants
//  and compare them to the values that were computed at startup.
//-----------------------------------------------------------------------------
void computeMemCRCs(bool firstTime, bool display)
{
  if (display)
    printf("\r\n\n"
           "Segment   Start  End    CRC  \r\n"
           "--------  -----  -----  ---- \r\n");

  computeCRC(firstTime, display, "CODE    ",
             codeStart, codeEnd, codeSize, &codeCRC);

  computeCRC(firstTime, display, "CONST   ",
             constStart, constEnd, constSize, &constCRC);

  if (display)  {
    printf("%ls  %05X  %05X  ----\r\n",
           (const char far *)"DATA    ", dataStart, dataEnd);
    printf("%ls  %05X  %05X  ----\r\n",
           (const char far *)"BSS     ", bssStart, bssEnd);
    printf("%ls  %05X  %05X  ----\r\n",
           (const char far *)"STACK   ", 0xD000, 0xDFFF);
  }

  computeCRC(firstTime, display, "FARCODE ",
             farCodeStart, farCodeEnd, farCodeSize, &farCodeCRC);

  computeCRC(firstTime, display, "FARCONST",
             farConstStart, farConstEnd, farConstSize, &farConstCRC);

  if (display)
    printf("%ls  %05lX  %05lX  ----\r\n",
           (const char far *)"FARBSS  ", farBSSStart, farBSSEnd);

  if (display)  printf("\n\n");
}


//-----------------------------------------------------------------------------
//  Show the current contents of the Interrupt Vectors
//-----------------------------------------------------------------------------
void showIntVectors(void)
{
  char *intvec;
  uchar  IIR, EIR;
  uint  ui, uj;

//--- show the contents of the Interrupt vector table ---
  IIR = _asm( [_A], "ld a,iir", _A );
  EIR = _asm( [_A], "ld a,eir", _A );

  printf(" IIR: 0x%02X,  EIR: 0x%02X\r\n", IIR, EIR);
  printf(" _TimerBISR: 0x%04X\r\n\n", (uint)_TimerBISR);

  intvec = (uchar *)((uint)IIR << 8);
  for (ui=0; ui<NumIntVectors; ++ui)  {
    printf(" %04X: ", (uint)intvec);
    for (uj=0; uj<16; ++uj,intvec)  {
      printf("%02X ", *intvec++);
      if ((uj == 6) OR (uj == 8) OR (uj == 15))  printf(" ");
    }
    printf("%ls\r\n", IntVectDesc[ui]);
    if (ui == 15)  printf("\n");
  }

  printf("\n");
}


//-----------------------------------------------------------------------------
//  Enable write protection for the largest portion of root and extended
//  memory we can to cover as much as possible of the CODE, CONST, FARCODE,
//  and FARCONST segments.
//-----------------------------------------------------------------------------
void initWriteProtect(void)
{
  uchar  seg, page, firstPage, lastPage, firstSeg, lastSeg;
  uint  mask;


  if (!hdwProtection)  return;

  setInterruptHandler(INT_WRITE_FAULT, WriteFault);
  setInterruptHandler(INT_MODE_VIOLATION, ModeViolation);
  setInterruptHandler(INT_STACK_LIMIT, StackLimit);

  checkIntShims();


//--- Determine the start and end addr of the ReadOnly portion of root mem ---
  if (codeStart < constStart)  {
    roStart = codeStart;  roEnd = constEnd; }
  else  {
    roStart = constStart;  roEnd =  codeEnd; }

//--- Determine the start and end addr of the ReadOnly portion of far mem ---
  if (farCodeStart < farConstStart)  {
    roStartFar = farCodeStart;  roEndFar = farConstEnd; }
  else  {
    roStartFar = farConstStart;  roEndFar = farCodeEnd; }


//--- determine the start and end addrs in root that can be write protected ---
  wpStart = (roStart + 0xFFF) & 0xF000;
  wpEnd = (roEnd + 1) & 0xF000;

//--- determine the start and end addrs in xmem that can be write protected ---
  wpStartFar = (roStartFar + 0xFFFFL) & 0xF0000L;
  wpEndFar = (roEndFar + 1) & 0xF0000L;



//--- compute write protection for 4K ROOT memory blocks ---
  firstPage = (uchar)(wpStart >> 12);  lastPage = (uchar)(wpEnd >> 12);

//  wpRootMask = 0x2000;  // protect the kernel's stack as well (at 0xD000)
  mask = 1 << firstPage;
  for (page=firstPage; page<lastPage; ++page,mask<<=1)  wpRootMask |= mask;

//--- compute write protection for 64K extended memory blocks ---
  firstSeg = (uchar) MSWL(wpStartFar);  lastSeg = (uchar) MSWL(wpEndFar);

  wpSegMask = 0;  mask = 1 << firstSeg;
  for (seg=firstSeg; seg<lastSeg; ++seg)  { wpSegMask |= mask;  mask <<= 1; }
//NOTE:  wpSegMask MAY be added to later by code in sched.c


  outi(MMIDR, 0x80);  // enable access to registers > 0x00FF

//--- apply mask that sets which of the 16 64K pages are write protected ---
  outi(WPHR, MSBI(wpSegMask));  outi(WPLR, LSBI(wpSegMask));
  outi(WPCR, 1);  // enable Write Protection in System and User modes

//--- enable hdw stack limit checking ---
  outi(STKLLR, current->stkllr);
  outi(STKHLR, current->stkhlr);
  outi(STKCR, 0x01);

  outi(MMIDR, 0x00);  // disable access to registers > 0x00FF
}

//-----------------------------------------------------------------------------
//  Show the address ranges write protected by a given segment and 4k-write
//  protection bitmap
//-----------------------------------------------------------------------------
void showRanges(ulong addr, uint wpMask, ulong addrInc)
{
  uint  mask, prevBit, curBit;


  prevBit = curBit = 0;

  for (mask=1; mask; mask<<=1)  {
    curBit = wpMask & mask;
    if (!prevBit AND curBit)  printf("  %05lX-", addr);
    if (prevBit AND !curBit)  printf("%05lX", addr-1);
    prevBit = curBit;  addr += addrInc;
  }
  if (curBit)  printf("%05lX", addr-1);

}

//-----------------------------------------------------------------------------
//  Show the areas of memory with write protection enabled
//-----------------------------------------------------------------------------
void showMemProtection(void)
{
  uint  u;
  ulong  addr = 0;
  task_t  *task;


  if (!hdwProtection)  return;

  printf("64K protections for all tasks:\r\n");
  showRanges(0, wpSegMask, 0x10000L);  printf("\r\n\n");

  printf("4K protections by task:                       STKxLR\r\n");
  for (u=0,task=tasks; u<NUM_TASKS; ++u,++task)  {
    printf("  %2d:", u);
    MSWL(addr) = task->wpsar;
    showRanges(addr, task->wpsalhr, 0x1000L);
    MSWL(addr) = task->wpsbr;
    showRanges(addr, task->wpsblhr, 0x1000L);
    printf("   %02X-%02X\r\n", task->stkllr, task->stkhlr );
  }

  printf("\r\n");
}

//-----------------------------------------------------------------------------
//  Give some CPU time to various subsystems that need to do periodic
//  processing in order to function.
//
//  This process can only be called by the kernel task
//-----------------------------------------------------------------------------
void backgroundIO(void)
{
  static uint  ioTicks;
  task_t  *task;
  list_head_t  *cur, *next;
  WaitQueueNode  *node;
  sigset_t  notBlockedPending;
  IFConfig  ifConfig;
  ulong  ipAddr;


  if (current != kernelTask)  return;


  if (msSince_(countersUpdated) >= 1000)  {
    _EnterCritical_();

    schedCallsPrevSec = schedCalls;  schedCalls = 0;
    taskSwPrevSec = taskSwitches;  taskSwitches = 0;

    _ExitCritical_();
    countersUpdated = GetMS_();
  }



  resetWDT();  ++ioTicks;

//===== give file system drivers some CPU time =====
  fileSysTick();


  diagLogInfo->tz = timeZone;

  if (timeZone != lastTimeZone)  {
    resetAnnounceInterval();  // announce the change ASAP
    lastTimeZone = timeZone;
  }

//--- periodically check for a change in the interrupt vectors ---
  if ((ioTicks & 0x007) == 2)  checkIntShims();


//===== Periodically check the device ID and make sure the current =====
//===== configuration is correct for the current device ID =====
  if ((ioTicks & 0x01f) == 0)  {

  //--- check for a change in the device ID switches ---
    deviceID = ReadDeviceID();
    if (deviceID != lastDeviceID)  {
      printf("deviceID changed: from:[%04lX]  to:[%04lX]\r\n",
             lastDeviceID, deviceID);
      lastDeviceID = deviceID;  loadAllParamsMsg();
      resetAnnounceInterval();
    }

  //--- if deviceID is in range 0x00E0 to 0x00FF, then use the last nibble ---
  //--- as an index in to the range of rsvd addrs (192.168.250.224 - .255) ---
    if (FixedAddr())  {
      ipAddr = 0xC0A8FA00L | (deviceID & 0x00FF);
      getIFConfig(IF_ETH0, &ifConfig);
      if (ifConfig.ipAddr != ipAddr)  {
        if ((ifConfig.state == IF_UP) OR (ifConfig.state == IF_COMING_UP))
          startNetDown(IF_ETH0);
        if (ifConfig.state == IF_DOWN)  chgIPAddr(IF_ETH0, ipAddr);
      }
    }
  }


//===== Check to see if we should initiate a shutdown or reboot =====
  if (doShutdown)  do {
    resetAnnounceInterval();

    if (!shutdownStarted)  {  // send terminate signal to all processes
      noGettyOnA = true;
      printf("----- starting shutdown ----\r\n\n");
      kill(-1, SIGTERM);  shutdownStarted = GetMS();  break;
    }

    //--- if all but the kernel and idle tasks are gone, shutdown now ---
    if ((numActiveTasks() <= 2) AND (numOpenSockets() < 1))  {
      if (! rebootMsg)  { printf("bye bye...\r\n\n\n\n");  rebootMsg = true; }
      if (msSince(shutdownStarted) > 2000)  {
        startNetDown(IF_ETH0);  sleepMS(1000);  REBOOT(doReboot);  }
    }

  //--- kill any tasks still running 10 seconds after getting SIGTERM ---
    if (msSince(shutdownStarted) > 10000)  kill(-1, SIGKILL);

  //--- Reboot after 15 seconds no matter WHAT!!! ---
    if (msSince(shutdownStarted) > 15000)  REBOOT(doReboot);

  } while(0);



//===== Check to see if any sleeping tasks should be woken up or have =====
//===== pending signals to be processed =====
  KLOCK();

//--- check for expired sleep timers ---
  list_for_each_safe(cur, next, &sleepQueue.listHead)  {
    node = list_entry(cur, WaitQueueNode, taskList);
    task = node->task;

    if (_activeTaskIdx_(task->pid) < 0)  {
      fprintf(DiagLog, "bad task in sleepQueue: %d/%d\r\n",   //tdebug
               (uint)(task-tasks), task->pid);
      __removeWaitQueue(&sleepQueue, node);  free_WaitQueueNode(node);
      continue;
    }
    if (task->preempt_count)  continue;

    if (task->sleepTicks)
      --task->sleepTicks;
    else  {
      __removeWaitQueue(&sleepQueue, node);  task->state = TASK_RUNNING; }
  }


  for (task=tasks; task!=endTasks; ++task)  {
    if (! task->inUse)  continue;

  //--- check for an expired timeout count ---
      if (task->timeoutTicks AND BlockedTask(task))
        if (--task->timeoutTicks == 0)  task->state = TASK_RUNNING;

  //--- check for pending signals ---
    if (! task->sig_pending)  continue;

  //finish:  clear ignored signals without a handler

  //finish
  // until I implement signal handlers, just clear any pending signals
  // for ones that are normally ignored
  // (use another mask for ones that have an associated handler?)
    task->sig_pending &= ~DefaultIgnoreSet;

  //--- Remove blocked pending signals from the set being considered ---
  //--- NOTE that they have to be left in the pending list for the task ---
  //--- so it can check for ones it blocked and deal with them when and ---
  //--- how it wants. ---
    notBlockedPending = task->sig_pending & ~task->sig_blocked;
    if (! notBlockedPending)  continue;

  //--- process has pending signals that are not blocked and not ignored ---
  //finish: until I implement signal handlers, just terminate it
    _killTask_(task);
  }

  KUNLOCK();
}


//-----------------------------------------------------------------------------
//  Do remaining hardware initialization then process kernel/system calls
//
//  This process must be run at the 2nd highest priority (below the applic
//  task priority, but above everything else).
//-----------------------------------------------------------------------------
void _task kernel(int argc, char **argv)
{
//  PIX  pix;
  int  pid, ch;
  uint  startTime;
  char  *argv_[2];
  uchar  *macAddr;
  uint  u;
  char  temp[64];


//--- initialize IP "stack" and allow all levels of interrupts
  ipset0();  ipset0();  ipset0();  ipset0();  // initialize IP "stack"

  printf("\r\nkernel process started\r\n");

//--- set up buffer and FILE "device" to handle diagnostic msgs ---
  fassign(DiagLog, diagIn, diagOut);


//--- redirect Softools stdin/stdout ---
  fassign(&taskIO, rcvChar, xmitChar);
  origIO = _stdio;  _stdio = &taskIO;


  setApplDTM();

  logDateTime = false;

  macAddr = idBlock->macAddr[0];
  sprintf(moduleID, "%02X%02X%02X", macAddr[3], macAddr[4], macAddr[5]);


  if (!restoredLog)
    fprintf(DiagLog, "\r\n----- Starting new log -----\r\n");
  else  {
    fprintf(DiagLog, "\r\n***** Restored previous log *****\r\n"
                     "\n"
                     "Module %s was last running at: %s\r\n"
                     "\n"
                     "This reboot was %sexpected\r\n\n",
            moduleID,
            dateTimeMS(temp, prevLogInfo.seconds, prevLogInfo.milliSecs),
            (prevLogInfo.ctldShutdown ? "" : "NOT ")
          );
    if (!prevLogInfo.ctldShutdown AND (time(NULL) - prevLogInfo.seconds < 90))
      attemptRestore = true;
  }

//--- If using hdw protection make sure the order and alignment of the ---
//--- segments are correct ---
  if (hdwProtection)
    if (!( (codeStart == 0) AND (constStart > codeStart)
         AND (farCodeStart > constStart) AND (farConstStart > farCodeStart)
         AND (farBSSStart > farConstStart)
         AND ((farCodeStart & 0x0FFF) == 0) AND ((farBSSStart & 0x0FFF) == 0)
       ) )
      printf("***** Segment order or alignment not correct *****\r\n"
             "Must be as follows:\r\n"
             "  CODE, CONST, DATA, BSS, FARCODE, FARCONST, FARBSS, XMEMSTART\r\n"
             "  with FARCODE and FARBSS using an alignment of 0x1000\r\n");

//--- compute CRCs for each segment and save the result for comparison ---
//--- should a detectable crash occur ---
  computeMemCRCs(true, true);


//--- check for reboot or shutdown command before continuing ---
  ch = getcConsole() & 0x7F;
  for (u=0; u<20; ++u)  if (((ch = getcConsole()) & 0x7F) < 0x7F)  break;
//  printf("\r\nConsole char: 0x%02X\r\n\n", ch);  //tdebug
  if (ch == 0x02)  {  // Ctl-B:  reboot and run backup bootimg
    runBackup = true;  REBOOT(true); }
  if ((ch == 0x03) OR (ch == 0x1B)) REBOOT(false); // Ctl-C or ESC: restart DLM

//--- If we rebooted unexpectedly shortly after the previous startup, then ---
//--- revert to the backup image or the DLM ---
  if (restoredLog)  // we have info from previous reboot/run
    if (!prevLogInfo.ctldShutdown)  // last reboot was not intentional
      // prev run crashed after a short time
      if (diagLogInfo->rebootTime - prevLogInfo.rebootTime < 90)  {
        if (runningPrimary)  { runBackup = true;  REBOOT(true); }
        REBOOT(false);  // no good boot image - revert to DLM
      }


  fprintf(DiagLog,
         "%ls\r\n"
         " RTOS: %.40ls  %.40ls\r\n"
         " APPL: %.40ls  %.40ls\r\n"
         "%ls\r\n",
         asterisks, osDTM, osName, applDTM, applName, asterisks);


  logDateTime = true;

  if (prevLogInfo.didSoftReset)
    printf("*** Restarted from a Soft Reset ***\r\n"
           "===== Last hard/CPU reset occured at: %.19s ====\r\n\n",
           dateTimeMS(temp, hdwResetTime, 0) );
//  else
//    if (attemptRestore)  printf("*** Attempt to restore prev state ***\r\n");

  if (copyGCSR < 0)
    printf("*** GCSR was not saved ***\r\n");
  else  {
    resetCode = (copyGCSR >> 6) & 0x03;
    printf("GCSR at restart: [0x%02X] (%d)(%ls)\r\n",
           copyGCSR, resetCode, resetCause[resetCode] );
  }

  printf("\r\n"
         "free Root: %u\r\n"
         "free xmem: %lu  (%05lX-%05lX)\r\n"
         "\n",
         (uint)0xD000 - bssStart - bssSize,
         xavail(), _xmemHeap, _xmemEnd-1 );


  printf("for CStart, BB log & info (earlier, during startup):\r\n");
  logAllocation(_xmemEnd, xmemAtStartup - xmemAfterBB);


  printf("Initializing... (%.6s)\r\n", moduleID);


  initWriteProtect();  // must come BEFORE init_tasks()


//NOTE:  sock_init() cannot be called until after the periodic timer interrupt
//       (the one that updates the milliseconds and seconds counters) is
//       running.  For DEBROS, that means the timerB ISR that calls the
//       sched_tick() function in this module.

//finish:  The various fcntl drivers are "registered" much earlier than this,
//         (during the fcntlInit() call).  But for drivers like the socket one
//         that need additional initialization that cannot be done at the time
//         they are registered, I should add an init function to the std fcntl
//         driver interface and call it for each driver at this time...

//--- initialize network interface ---
//  tcp_lock = lock_semaphore;  tcp_unlock = unlock_semaphore;
  tcp_lock = KLOCK;  tcp_unlock = KUNLOCK;  // just in case I missed any...
  printf("Initializing network hdw...\r\n"
         "  tcp_Socket size: %d   udp_Socket size: %d\r\n",
         sizeof(tcp_Socket), sizeof(udp_Socket) );
  WDT_DISABLE();
  netHdwInit();
  printf("Initializing network stack (xavail():%ld)...\r\n", xavail());
  sockStat = sock_init();
  if (sockStat)  printf("***** sock_init() error: %d *****\r\n", sockStat);
  WDT_2S();  WDT_ENABLE();
  printf("\n");

  if (init_tasks())  REBOOT(runningPrimary);

  printf("Task table initialized\r\n\n");


  if (hdwProtection)  {
//    showMemProtection();

  //--- in case the 64K WP segment mask was changed ---
    outi(MMIDR, 0x80);  // enable access to registers > 0x00FF
    outi(WPHR, MSBI(wpSegMask));  outi(WPLR, LSBI(wpSegMask));
    outi(MMIDR, 0x00);  // disable access to registers > 0x00FF
  }

  printf("For far heap:\r\n");
  if (_farheapsize > xavail())  _farheapsize = xavail();
  if (! (_farheap = xalloc(_farheapsize)))
    printf("***** Unable to init far heap *****\r\n");

  fmemset((void far *)_farheap, 0, _farheapsize);


  deviceID = lastDeviceID = ReadDeviceID();
  printf("\r\nDevice ID: 0x%04lX\r\n\n", deviceID);

  flushConsole();

#ifdef SHD_DLM
  DLM_Init(hostName);    // initializes flash file system
  if (sockStat AND runningPrimary)  doShutdown = doReboot = true;
#endif


  lastReadTime = 0;

  hdwInit2();


  checkIntShims();  // make ISR vectors "scheduler-safe"

  countersUpdated = GetMS_();

  schedCallsPrevSec = 0;


  diagLogInfo->wdtTicks = 0;

  WDT_ENABLE();  WDT_2S();  checkWDT = true;  // start 2-second countdown Watchdog timer


  applicActive = false;

  argv_[0] = "idle";  argv_[1] = NULL;
  pid = createProc(idle, NUM_PRIORITIES-1, 1, argv_, 0, 1, 2);
  idleTask = tasks + TaskIdx(pid);

  argv_[0] = "applic";  argv_[1] = NULL;
  pid = createProc(applic, 0, 1, argv_, 0, 1, 2);
  applTask = tasks + TaskIdx(pid);

  resetWDT();

  srand(GetMS_());

  KUNLOCK();  // and away we go!!!


  for (;;)  {
    KLOCK();

    startTime = GetMS_();

    backgroundIO();

    kernelLoopTime = msSince_(startTime);

    current->state = TASK_UNINTERRUPTIBLE;

    KUNLOCK();  schedule();  // woken by sched_tick() each scheduler tick
  }

}

//-----------------------------------------------------------------------------
//  Check for a copy of the GCSR register value saved to RAM by a patched
//  version of the RDLM.  This value tells us if the CPU was reset by the
//  Reset signal (or a power failure), or by the watchdog timer.
//
//  The patched copy of the RDLM does the following:
//
//    - Reads the value of GCSR and creates a 16-bit value with the value in the
//      LSB and a 1's complement of the value in the MSB
//
//    - Stores a copy of the 16-bit value at the following offsets from the
//      start of the RAM activated by the /CS1 signal:
//          0x28000, 0x28080, 0x78000, 0x78080
//      Which means the following:
//        For RCM3200:  The value will be at 0xA8000 and 0xA8080
//        For RCM22xx:  The value will be at 0x78000 and 0x78080
//
//  Returns the value if found, or -1 if the value was not found at the
//  specified locations.
//-----------------------------------------------------------------------------
int readGCSR(ulong testAddr)
{
  uint  copy1, copy2;


  copy1 = *((uint far *)testAddr);  copy2 = *((uint far *)(testAddr + 0x80L));

  if ((copy1 == copy2) AND (LSBI(copy1) == ~MSBI(copy1)))  return LSBI(copy1);

  return -1;
}


//-----------------------------------------------------------------------------
//  Start of execution:  Initialize hardware, the timer interrupt, and create
//  the kernel process
//-----------------------------------------------------------------------------
void main( void )
{
  uint  i, msw, lsw;
  int  runKey;
  time_t  curTime;
  DiagLogInfo far *prevInfo;
  char far *fromPtr, *toPtr;
  ulong  len, downSecs;
  char  *argv[2], *stack_low, *stack_high;
#define LocalVarBytes  64  // must be at LEAST as large as above local vars!!!


//--- Get the value GCSR had after a restart, if available ---
  copyGCSR = readGCSR(0xA8000L);
  if (copyGCSR < 0)  copyGCSR = readGCSR(0x78000L);


  KLOCK();  // make sure the scheduler can't start until we are ready

  WDT_ENABLE();  WDT_2S();  kernelLastCalled = applicLastCalled = 0;

  xmemAtStartup = xavail();  ramSize = _xmemEnd;

  cpuID = ini(GCPU) & 0x1F;  cpuRev = ini(GREV) & 0x1F;

  hdwProtection = (cpuID > 1) OR ((cpuID == 1) AND (cpuRev >= 1));


  logDateTime = recursiveDump = useSoftReset = memCorrupted = false;
  attemptRestore = false;
  runBackup = doCleanShutdown = true;

//NOTE:  Interrupts are not yet enabled, so no data would actually be sent
//       out the console serial port until initialization finishes unless
//       we temporarily enable interrupts

  ipset0();  initConsole();


  if (xmemAtStartup < _256K)  {
    DiagLogSize = 10 * 1024;  _farheapsize = 12 * 1024;  }
  else  {
    DiagLogSize = 16 * 1024;  _farheapsize = 24 * 1024;  }

  diagLog = *_stdio;  // temporary (until syslog buffer is ready)


  timeZone = DEFAULT_TZ;  strcpy(hostName, "unknown");

//--- get location and size of the memory segments that should not change ---
  codeStart = RootSegStart(CODE);    codeSize = RootSegSize(CODE);
  codeEnd = codeStart + codeSize - 1;

  constStart = RootSegStart(CONST);  constSize = RootSegSize(CONST);
  constEnd = constStart + constSize - 1;

  dataStart = RootSegStart(DATA);  dataSize = RootSegSize(DATA);
  dataEnd = dataStart + dataSize - 1;

  bssStart = RootSegStart(BSS);  bssSize = RootSegSize(BSS);
  bssEnd = bssStart + bssSize - 1;

  farCodeStart = SegStart(FARCODE);     farCodeSize = SegSize(FARCODE);
  farCodeEnd = farCodeStart + farCodeSize - 1;

  farConstStart = SegStart(FARCONST);   farConstSize = SegSize(FARCONST);
  farConstEnd = farConstStart + farConstSize - 1;

  farBSSStart = SegStart(FARBSS);   farBSSSize = SegSize(FARBSS);
  farBSSEnd = farBSSStart + farBSSSize - 1;

//--- Softools places the Internal (IIR) and External (EIR) vector tables ---
//--- in contiguous memory, so we can treat it as one long table ---
  intVectStart = (uint)_asm( [_A], "ld a,iir", _A ) << 8;
  intVectEnd = intVectStart + 512;



  initWaitQueues();  initSemaphores();

//  sem_init(&tcpLock, 1, 1);

//--- Detect flash - also initializes lowest-level of flash driver ---
  DetectFlash();

//--- bb_xalloc() does a one-time (permanant) allocation of memory from the ---
//--- end of memory normally managed by xalloc().  For RCM3200's, this ---
//--- provides a way to allocate blocks of memory in the upper (slower) 256K ---
//--- that is battery-backed.  For BL2105's, all memory is battery-backed ---
//--- so for them it doesn't really matter ---

  printf("\nFor CStart:\r\n");
  bb_xalloc(256);  // reserved space at end of RAM used by CStart (and/or RDLM?)

  printf("For syslog:\r\n");
  diagLogBuf = bb_xalloc(DiagLogSize);

  printf("For DiagLogInfo:\r\n");
  diagLogInfo = (DiagLogInfo far *) bb_xalloc(sizeof(DiagLogInfo));

  if (!diagLogInfo OR !diagLogBuf)  {
    printf("\r\n***** Unable to alloc buffer for system log *****\r\n");
    REBOOT(true);
  }

  printf("For saving Application State:\r\n");
  stateDataBuf = bb_xalloc(ApplStateDataSize + sizeof(ApplStateDataHdr));


  flushConsole();

  xmemAfterBB = xavail();

/*
  The SHDesigns RDLM copies (or did - I think this may no longer be true of
  the newer versions) ALL of regular RAM (the 1st 512K)  "down" by 0x6000 (24K)
  bytes as part of the sequence that loads the code from flash.  For boards
  without fast RAM, this will include the buffer we used for the log.  For
  boards with fast RAM, the log will be in the (slower) 256K battery-backed
  RAM above the 512K mark, and is not moved during the "boot" process.

  Check for the case where the allocated memory is in the lower 512K that got
  shifted 24K bytes by cstart.  If this case applies, then copy the memory
  previously allocated by bb_xalloc() back to where it was.  NOTE:  Newer
  version of RDLM seems to have eliminated this process, loading the code in
  to the same place even in the cases where it previously moved it by 24K
  after loading it in to RAM.
*/
  if (diagLogInfo->infoKey != LogInfoKey)  {
    prevInfo = (DiagLogInfo far *)((ulong)diagLogInfo - 0x6000L);
    if (prevInfo->infoKey == LogInfoKey)  {
      fromPtr = (char far *)prevInfo;  toPtr = (char far *)diagLogInfo;
      len = prevInfo->bbCount;   //DiagLogSize + sizeof(DiagLogInfo);
      fmemcpy(toPtr, fromPtr, len);
    }
  }


/*
  When the RDLM loads the primary of backup boot image file, it saves some
  info in the RUN_INFO struct (located at 0x7FF80-0x7FFFF).  One of these
  values is a Key that indicates which boot image file (the primary or the
  backup) was loaded.  DLM_RunKey() returns this value and sets it to 0, so
  if it called more than once, the value will not be one of the special
  values, which would indicate we somehow got here by some path other than
  a restart that included the RDLM (re)loading one of the boot image files.
*/
  runKey = DLM_RunKey();

  runningPrimary = (runKey == UDPDNLD_MODE_RUN_PRIMARY);

/*
  If we don't have any indication that we got here on purpose (i.e. the DLM ran
  and (re)loaded the OS from flash), then we need to force a reboot.  This
  insures that all the static variables have the right starting values and the
  copy of the code we are running is not corrupted.
*/
  if ((runKey != UDPDNLD_MODE_RUN_PRIMARY)
      AND (runKey != UDPDNLD_MODE_RUN_BACKUP))  {
    printf("\r\n\n***** Restarted w/o RDLM - Forcing a reboot now *****\r\n\n");
    flushConsole();
    DLM_Reboot(UDPDNLD_MODE_RUN_OK);  runBackup = false;  REBOOT(true);
  }

/*
  Seems we got here via a true reboot (i.e. the RDLM reloaded the image from                                                                ---
  the flash file), so all static variables should be initialized and the CODE                                                               ---
  and CONST segments (near and far) should be intact.
*/
  printf("\r\n ===== Clean bootup verified =====\r\n\n");  flushConsole();


  fstrcpy(osName, "Davis Embedded Basic Real-time OS (DEBROS)");

  sprintf(osDTM, "%s %s", __DATE__, __TIME__);

  printf("\r\n RTOS: %ls  %ls\r\n\n", osDTM, osName);  flushConsole();

  WDT_2S();


//--- Now we check to see if have a filled in DiagLogInfo struct from the ---
//--- previous boot, and deal with any version differences. ---
#define  DLI  diagLogInfo

  do  {
    restoredLog = false;
    if (DLI->infoKey != LogInfoKey)  break;  // no data from last time?

  //--- If the log size is now smaller, it is better to loose some of the ---
  //--- data (and even risk repeating part of it) than to ignore it all. ---
    if (DLI->toIdx > DiagLogSize)  DLI->toIdx = DiagLogSize - 1;
    if (DLI->fromIdx > DiagLogSize)  DLI->fromIdx = DiagLogSize - 1;

  //--- Check for a valid timezone value ---
    if (!ValidTZ(DLI->tz))  DLI->tz = DEFAULT_TZ;

    downSecs = time(NULL) - DLI->seconds;

    if (DLI->didSoftReset AND (downSecs > 120))  {
      printf("*** Down for %lu seconds: Soft Reset ignored ***\r\n", downSecs);
      DLI->didSoftReset = false;
    }

    restoredLog = true;
  } while(0);


  if (restoredLog)  {
    prevLogInfo = *diagLogInfo;  // save a copy of the header data
    timeZone = prevLogInfo.tz;
  }
  else  {
    fmemset(diagLogInfo, 0, sizeof(DiagLogInfo));
    fmemset(&prevLogInfo, 0, sizeof(DiagLogInfo));
    timeZone = DEFAULT_TZ;
  }


//--- now we can init/update values based on the current vers of the struct ---
  diagLogInfo->infoKey = LogInfoKey;
  diagLogInfo->bbCount = bbAllocatedCount;

  diagLogInfo->ctldShutdown = diagLogInfo->didSoftReset = false;

  FBZERO(diagLogInfo->filler1);
  FBZERO(diagLogInfo->filler2);

  diagLogInfo->rtosVers = rtosVersTime;

  diagLogInfo->invInfoKey = ~LogInfoKey;

  // hdwResetTime, rebootTime, and others are handled later...


  ipset3();  // disable all normal interrupts (doesn't include the WDT)

  if ((time(NULL) - prevLogInfo.seconds) > 60L)
    prevLogInfo.didSoftReset = false;


//--- initialize hardware ---
  digitalOut =
    (prevLogInfo.didSoftReset ? prevLogInfo.digOutState : DigOutInitState);

  hdwInit(digitalOut);  flushConsole();


//--- initialize I/O related state info ---
  for (i=0; i<NumHdwADC; ++i)  adcForced[i] = false;
  for (i=0; i<NumHdwDAC; ++i)  dacForced[i] = false;


//--- initialize basic kernel data structures ---
  applTask = (task_t *)0;  // gets assigned to actual task later

  kernelTask = current = tasks;  userPix = 0;


  memset(tasks, 0, sizeof(task_t)*NUM_TASKS);

  TrackApplLoop = TrackScheduler = TrackKernelLock = TrackTimerISR = 0;

  lastTaskIdx = 0;  klocks = 0;

  rebootMsg = doShutdown = doReboot = gotTime = false;  shutdownStarted = 0;


//
;
#asm
;--- DATASEG starts at 0x0000, STACKSEG starts at 0xD000 ---
    ld      a,0xD0
    ioi ld  (SEGSIZE),a

;--- set BC = current SP ---
    bool    hl             ; zero HL
    ld      l,h
    add     hl,sp          ; copy SP in to BC
    push    hl
    pop     bc
;--- set HL = addr of start/bottom of stack + 4 bytes for a marker ---
    ld      hl,_stackMarker##
    ld      de,4
    add     hl,de
#endasm
;
  stack_high = (char *)_BC;
  stack_low = (char *)_HL; // skip over stack marker

//--- fill portion of stack not used yet with 0xff ---
  stack_high -= LocalVarBytes;  // shouldn't that already be taken care of?
  memset(stack_low, 0xff, stack_high - stack_low);
  stack_high += LocalVarBytes;

  tasks[0].stack_start = (uint *)stack_low;
  tasks[0].stack_size = stack_high - stack_low;
  tasks[0].stack = (uint *)stack_high;

  tasks[0].sseg = ini(STACKSEG);


//--- set values for this process ---
  argv[0] = "[kernel]";  argv[1] = NULL;

  copyArgs(tasks, 1, argv);

  tasks[0].pid = 0;
  tasks[0].ppid = -1;
  tasks[0].priority = 1;
  tasks[0].claimed = true;

  tasks[0].outPtr = outBuf[0];
  tasks[0].procMain = kernel;
  tasks[0].FDSet = (void far **) FD[0];

  tasks[0].started = readRTC(NULL);
  tasks[0].resumedTicks = clockTicks();

  init_waitqueue_head(&tasks[0].waitQs);

  tasks[0].state = TASK_RUNNING;

  tasks[0].inUse = true;

  tasks[0].stkllr = 0xD0;  // stack limits:  0xD000 - 0xDF00
  tasks[0].stkhlr = 0xDF;


  rtosVersTime = compileTime(__DATE__, __TIME__);

  devType = 0;

//--- initialize file system drivers ---
  WDT_2S();  fcntlInit();  flushConsole();

/*
  If this was a soft reset, then save the hdwResetTime and set the seconds
  and milliseconds since hdwReset so they pick up where left off. Otherwise,
  set hdwResetTime to the rebootTime and initialize the counters to 0.

  The rebootTime always indicates the time of the most recent restart, for
  both soft and hard (CPU) resets.  NOTE that, if the RTC doesn't have a
  valid value at startup (never set or power cycle w/o battery backup),
  then the rebootTime and/or the hdwResetTime will be invalid.

  The newer versions of this code update these values whenever the RTC is
  set, so any controller that has done a hard reset using the newer code
  will have hdwResetTime and rebootTime values that are in sync with the
  seconds value (they may ALL be wrong, but they will all be wrong by a
  consistent amount: RTC - seconds).

  But older versions did NOT correct the rebootTime and/or hdwResetTime times
  when the RTC value was changed.  Because the seconds value has always
  mirrored the RTC value, this means the the hdwResetTime and rebootTime
  values might be based on different values than seconds is, which leaves
  us with no way to determine if (and how much) the hdwResetTime and
  rebootTime might be off.

  The best we can do is check for unreasonable values and try our best to
  correct them based on the only bit of information we have, which is that
  msTime should be the # of ms since the last hard reset.  But of course,
  it only has a range of around 49 days, so there is no guarantee that using
  it will produce the correct result either.

  Out one saving grace is that most the application logic that tests for how
  much time has passed uses the millisecond counter rather than the second
  counter, which at least reduces the risks of having an invalid seconds count.

  IMPORTANT POINT:  The seconds value is always set from the RTC value, so it
  is always in agreement with the RTC.  And since a soft reset cannot occur
  unless power was maintained during the reboot, the current RTC value must
  still be counting from the last value it had when the seconds value was
  last updated (just prior to shutting down and restarting the OS).
*/

#define  TenYearSecs  315360000L

  curTime = time(NULL);

  hdwResetTime = rebootTime = curTime;  // from real-time clock

  _TimerMilliSeconds[0] = _TimerMilliSeconds[1] = _TimerSeconds = 0;
  milliTicks = 0;

  if (prevLogInfo.didSoftReset)  {
    _TimerMilliSeconds[1] = prevLogInfo.msTime + downSecs * 1000L;

    hdwResetTime = prevLogInfo.hdwResetTime;
    _TimerSeconds = curTime - hdwResetTime;

    // In case the hdwResetTime and seconds values are from an older version
    // of code and are based on before and after RTC-was-set values

    if (_TimerSeconds > TenYearSecs)  {  // unlikely time since last hard reset?
      _TimerSeconds = prevLogInfo.seconds - prevLogInfo.rebootTime
                    - prevLogInfo.hdwResetTime;

      if (_TimerSeconds > TenYearSecs)  // fallback: secs since prev reboot
        _TimerSeconds = _TimerMilliSeconds[1] / 1000;

      hdwResetTime = curTime - _TimerSeconds;

      printf("\r\n"
             "*** Invalid hdwResetTime: ***\r\n"
             "  curTime: 0x%08lX   seconds: 0x%08lX  downSecs: %ld\r\n"
             "  rebootTime: 0x%08lX   hdwResetTime: 0x%08lX  msTime: 0x%08lX\r\n"
             "  Corrected hdwResetTime: 0x%08lX\r\n\n",
             curTime, prevLogInfo.seconds, downSecs,
             prevLogInfo.rebootTime,  prevLogInfo.hdwResetTime,
             prevLogInfo.msTime, hdwResetTime);

      flushConsole();
    }

  }

  diagLogInfo->rebootTime = rebootTime;
  diagLogInfo->hdwResetTime = hdwResetTime;


  checkWDT = false;  //  Don't check for imminent WatchDog timer reboot yet

//--- start periodic timer ---
  select_timer(1000, 1);  // calls sched_tick() every SchedMS ms; use IP 1

//  startTimer(1000, timer_interrupt, 1);

  kernel(1, argv);  // doesn't return
}


//----------------------------------------------------------------------------


