//---- nsclSerialProto.c ----- 06/23/06 -----(11/xx/95)-----

//-----------------------------------------------------------------------------
//  Utility functions used to encode, decode and verify messages using the
//  NSCL Serial Protocol Specification.
//-----------------------------------------------------------------------------
//
//  2006-06-23 - Mark Davis - ported for use with new RTOS-based base code
//
//  1995-11-xx - Ben DeBiase - original version
//
//-----------------------------------------------------------------------------

#include <sys/types.h>
#include <Common.h>
#include <unistd.h>
#include <errno.h>

#include <sys/nsclSerialProto.h>

#include <sys/farprintf.h>



bool validHexal(char X);
bool validCMD( char X );
char dehexalize( char*, char*, char* );
void hexalize( char*, char*, char* );


#define IDX_DEST       1
#define IDX_SRC        2
#define IDX_COUNT      3
#define IDX_CMDSTAT    4
#define IDX_DATA       5

const char HexChar[] = { "0123456789ABCDEF" };


//-----------------------------------------------------------------------------
//  Reverse the order of bytes in a string of given length
//-----------------------------------------------------------------------------
char* reverseString( char* buffer, char dataCount)
{
  char temp, *left, *right;

  left = buffer;
  right = buffer + dataCount - 1;
  while ( left < right )  {
    temp = *left;
    *left = *right;
    *right = temp;
    ++left;  --right;
  }

  return buffer;
}


//-----------------------------------------------------------------------------
//  Given a valid message string pointed to by "buffer" the
//  command/status byte and any data are extracted out of the string.
//  The number of data bytes extracted are written to *count.
//
//  Returns true for success, false if an error occured
//-----------------------------------------------------------------------------
bool unpackMessage(char* buffer, char *commStat, char* data, char *count)
{
  char *ptr, *dataBuf, *dataEnd;


  if (!buffer OR !data)  return false;

  *count = 0;
  ptr = buffer + IDX_COUNT;  // get byte count
  dataBuf = data;
  dataEnd = data + ((*ptr - 0x40 - 1) >> 1);
  ++ptr;
  *commStat = *ptr;  // get command-status char
  ++ptr;

  *dataEnd = 0;   // append null to data string
  for (dataBuf=data; dataBuf<dataEnd; ++dataBuf)  {
    dehexalize(ptr, ptr+1, dataBuf);
    ptr += 2;  ++(*count);
  }

  return true;
}


//-----------------------------------------------------------------------------
//  Create a message string using the given command, "numBibbles" nibbles from
//  "rawData" and the sendors ID. The msg is stored in the memory pointed to
//  by "buffer".
//
//  Returns true for success, false if an error occured
//-----------------------------------------------------------------------------
bool  packMessage( char* buffer, char ID, char command,
                   char* rawData, char numNibbles)
{
  uchar   *msg, check, nib;
  int  i, sum, numNibs;


  if (! buffer)  return false;

  sum = 0;

  msg = buffer;
  *msg = '\n';  // LF
  ++msg;

  sum += (*msg = 0x40 + MASTER_ID );  // destination addr
  ++msg;
  sum += (*msg = 0x40 + (uint)ID);  // source address
  ++msg;

  if (rawData)  numNibs = numNibbles;  else  numNibs = 0;
  if (numNibs > 62)  numNibs = 62;
  sum += ( *msg = 1 + numNibs + 0x40 );  // byte count
  ++msg;

  sum += ( *msg = command  );  // command-status byte
  ++msg;

  for (i=0; i<numNibs; ++i)  {   /* DATA */
    nib = rawData[i/2];  if (! (i & 1)) nib >>= 4;
    sum += (*msg = HexChar[(int)nib & 0x0f]);  ++msg;
  }

  check = sum % 0x100;
  *msg = HexChar[(check >> 4) & 0x0f];  ++msg;
  *msg = HexChar[check & 0x0f];  ++msg;

  *msg = '\r';  // CR
  ++msg;
  *msg = 0;

  return true;
}


//-----------------------------------------------------------------------------
//  Given a message pointed to by "buffer" this function verifies that the
//  message format is OK. Also, the destination address is extract and passed
//  to ID.
//
// RETURN VALUES:
//   STAT_IGNORE_MESSAGE         Ignore the message in question
//   STAT_OK                     All is ok
//   STAT_BADFORMAT              Bad message format
//   STAT_CHECKSUM_ERROR         Checksum error
//   STAT_BAD_FIELD_CHAR         Data/Info field characters was bad
//   STAT_MESSAGE_LONG           Message too long - byte count too small
//   STAT_NOT_FROM_MASTER        Message not from the master
//   STAT_MESSAGE_SHORT          Message too short - byte count too long
//   STAT_UNDEFINED_COMMAND      Undefined command
//-----------------------------------------------------------------------------
char checkMessage(char *cmdBuf, char *ID)
{
  char  byteCount;
  uchar  *msg;
  int  msgSum, checkSum, diff;


  msgSum = checkSum = 0;  msg = cmdBuf;

//--- cmdBuf[0] ---
  if (!msg)  return STAT_IGNORE_MESSAGE;
  if (!*msg)  return STAT_IGNORE_MESSAGE;
  ++msg;

//--- cmdBuf[1] ---
  if (!*msg)  return STAT_IGNORE_MESSAGE;
  *ID = *msg - 0x40;  msgSum += *msg;  ++msg;

//--- cmdBuf[2] ---
  if (!(*msg & 0x40))  return STAT_BADFORMAT;
  if ((*msg-0x40) != MASTER_ID)  return STAT_NOT_FROM_MASTER;
  msgSum += *msg;  ++msg;

//--- cmdBuf[3] ---
  if ((*msg < 0x40) OR (*msg > 0x7F))  return STAT_BADFORMAT;
  byteCount = strlen(cmdBuf) - 7 ;
  diff = (int)byteCount - ((int)*msg - 0x40);
  if (diff > 0)  return STAT_MESSAGE_LONG;
  if (diff < 0)  return STAT_MESSAGE_SHORT;
  msgSum += *msg;  ++msg;

//--- cmdBuf[4] ---
  if (!validCMD(*msg))  return STAT_UNDEFINED_COMMAND;
  msgSum += *msg;  ++msg;  --byteCount;

  while (byteCount)  {
    if (!validHexal(*msg))  return STAT_BAD_FIELD_CHAR;
    msgSum += *msg;  ++msg;  --byteCount;
  }

  if (dehexalize(msg, msg+1, (char *)&checkSum ))
    return STAT_BAD_FIELD_CHAR;

  if (checkSum != (msgSum & 0xff))  return STAT_CHECKSUM_ERROR;

  return STAT_OK;
}

//-----------------------------------------------------------------------------
bool validCMD(char X)
{
  switch (X)  {
    case CMD_CLEAR_ERR:
    case CMD_DEBUG:
    case CMD_ECHO:
    case CMD_INIT:
    case CMD_CALIBRATE:
    case CMD_MODBUSREGS:
    case CMD_POLL:
    case CMD_SET_AND_POLL:
    case CMD_REPORT_ERRS:
    case CMD_SET:
    case CMD_WHATARESET:
    case CMD_XCAL:
    case CMD_REBOOT:
      return true;
  }

  return false;
}


//-----------------------------------------------------------------------------
bool validHexal(char X)
{
  return (((X >= 0x30) AND (X <= 0x39)) OR ((X >= 0x41) AND (X <= 0x46)));
}

//-----------------------------------------------------------------------------
void hexalize( char* H, char* L, char* Value )
{
 *H = (*Value >> 4) & 0x0F;
 *L = *Value & 0x0F;
 *H += ( (*H < 0x0A) ? 0x30 : 0x37 ) ;
 *L += ( (*L < 0x0A) ? 0x30 : 0x37 ) ;
}


//-----------------------------------------------------------------------------
char dehexalize(char* H, char* L, char* Value)
{
  bool dataError;


  dataError = 0;  *Value = 0;

  if ((*H >= 0x30) AND (*H <= 0x39))  *Value = (*H - 0x30)<<4;
  else if ((*H >= 0x41) AND (*H <= 0x46))  *Value = (*H - 0x37)<<4;
  else  dataError |= 1;

  if ((*L >= 0x30) AND (*L <= 0x39))  *Value = *Value | (*L - 0x30);
  else if ((*L >= 0x41) AND (*L <= 0x46))  *Value = *Value | (*L - 0x37);
  else  dataError |= 2;

  return dataError;
}

//-----------------------------------------------------------------------------
