//----- diag.c ----- 06/29/10 -----(03/14/06)-----

//-----------------------------------------------------------------------------
//
//  2010-06-29 - Mark Davis
//               - Changed new "display" argument t computeMemCRCs() to "true"
//                 so it shows the results of the computed CRCs again.
//
//  2010-06-02 - Mark Davis
//               - Moved computeMemCRCs() to unistd.c
//
//  2010-03-10 - Mark Davis
//               - Added ability to change the value of SchedMS
//
//  2010-03-05 - Mark Davis
//               - Added WaitOnConsoleBits to list of standard diag msg settings
//
//  2010-02-17 - Mark Davis
//               - Added crc option to recompute and display the CRCs for the
//                 code and const areas of RAM, and compare them with the ones
//                 computed during startup.
//
//  2009-xx-xx - Mark Davis
//               - original version of ...
//
//-----------------------------------------------------------------------------

#include <time.h>
#include <common.h>

#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <errno.h>
#include <signal.h>
#include <utime.h>


#include <busybox.h>

#include <sys/hdwDefs.h>

#include <dcdefs.h>

#include <ctype.h>
#include <limits.h>

#include <sys/SHD_headers.h>

#include <applic.h>

#include <sys/farprintf.h>



//-----------------------------------------------------------------------------
// diag msgs ...
//
// show/change diagnostic message bit settings for one or all functions
//-----------------------------------------------------------------------------
int setDiagMsgs(int argc, char **argv)
{
  uint  i;
  int  pix, setPid;
  ulong  bits, clearBits, setBits;
  bool  newSetting;
  task_t  *task;
  char  *ptr;
  char  ch;
  char  temp[80];

  typedef struct  {
    const far char *name;
    const ulong     bits;
  }  DiagFlagBits;

  static const DiagFlagBits far diagFlagBits[] = {
    { "all",     0xffffffffL },
    { "debug",   DebugBits },
    { "err",     ErrorBits },
    { "event",   EventMsgBits },
    { "file",    FileEventBits },
    { "rw",      RWCallBits },
    { "sock",    SocketEventBits },
    { "msgwait", WaitOnConsoleBits }
  };
  static uint  NumFlagNames = sizeof(diagFlagBits) / sizeof(diagFlagBits[0]);


  if (showHelp(3, argc, argv))  {
    printf("\nusage: diag msgs pid|all|name|new [#|+#|-#|+name|-name] ...\r\n\n");
    printf("  Symbolic names for diagnostic flag bits:\r\n");
    for (i=0; i<NumFlagNames; ++i)  {
      fstrcpy(temp, diagFlagBits[i].name);
      printf("    %-10.8s 0x%08lx\r\n", temp, diagFlagBits[i].bits);
    }
    return -1;
  }

  newSetting = (argc > 3);


//--- parse bits argument and determine which bits to set and clear ---
  if (! newSetting)
    printf("  PID BITS       CMD\r\n");
  else  {
    setBits = clearBits = 0;
    for (int a=3; a<argc; ++a)  {
      ch = argv[a][0];  ptr = argv[a];
      if ((ch == '-') OR (ch == '+'))  ++ptr;
      if (isdigit(*ptr))
        bits = ATOUL(argv[a]);
      else  {
        bits = 0;
        for (i=0; i<NumFlagNames; ++i)  {
          fstrcpy(temp, diagFlagBits[i].name);
          if (strncmp(temp, ptr, strlen(ptr)) == 0)  {
            bits = diagFlagBits[i].bits;  break; }
        }
        if (! bits)  {
          printf("Unknown bit name: %s\r\n", ptr);  return -1; }
      }
      if (ch == '-')  clearBits |= bits;  else
      if (ch == '+')  setBits |= bits;  else
        { setBits = bits;  clearBits = ~bits; }
    }
  }


//--- view (and optionally set) diag msg flag value for new tasks ---
  if (StriEq(argv[2], "new"))  {
    if (newSetting)  {
      SetBits(newTaskDiagFlags, setBits);
      ClearBits(newTaskDiagFlags, clearBits);
    }
    printf("  new task diag bits: 0x%08lx\r\n", newTaskDiagFlags);
    return 0;
  }


  if (isdigit(argv[2][0]))
    setPid = ATOI(argv[2]);
  else
    setPid = -1;

//--- show or alter diagnostic bits for specified process(es) ---
  for (pix=0; pix<NUM_TASKS; ++pix)  {
    task = tasks + pix;
    if ((task->pid != setPid) AND stricmp(argv[2], "all")
      AND stricmp(argv[2], task->argv[0]) )  continue;
    KLOCK();
    if (! task->inUse)  { KUNLOCK();  continue; }
    if (newSetting)  {
      SetBits(task->diagFlags, setBits);
      ClearBits(task->diagFlags, clearBits);
    }
    KUNLOCK();
    printf("%5d 0x%08lx %.20s\r\n",
              task->pid, task->diagFlags, task->argv[0]);
  }

  return 0;
}

//-----------------------------------------------------------------------------
//  diag events ...
//
//  show or set the bits to toggle for various code events
//-----------------------------------------------------------------------------
int setDiagBits(int argc, char **argv)
{
  uint  i;
  int  idx;
  ulong  setBits;
  ulong  *bitsPtr;
  bool  newSetting;
  char  temp[80];

  typedef struct  {
    const far char *name;
    ulong    *mask;
  }  EventFlag;

  static const EventFlag far eventFlags[] = {
    { "ApplLoop",   &TrackApplLoop   },
    { "Scheduler",  &TrackScheduler  },
    { "KernelLock", &TrackKernelLock },
    { "TimerISR",   &TrackTimerISR   }
  };
  static uint  NumEventFlags = sizeof(eventFlags) / sizeof(eventFlags[0]);


  if (showHelp(3, argc, argv))  {
    printf("\nusage:  diag events name|all bits\r\n\n");
    printf("      Supported events:\r\n");
    for (i=0; i<NumEventFlags; ++i)  printf("%20.12ls\r\n", eventFlags[i].name);
    return -1;
  }

//--- figure out which event name they entered ---
  idx = -1;
  if (stricmp(argv[2], "all"))  {
    for (i=0; i<NumEventFlags; ++i)  {
      fstrcpy(temp, eventFlags[i].name);
      if (StriEq(temp, argv[2]))  idx = i;
    }
    if (idx == -1)  {
      printf("Unknown event name: %s\r\n", argv[2]);  return -1; }
  }


  newSetting = (argc > 3);

//--- parse bits argument and determine which bits to set and clear ---
  if (newSetting)
    setBits = ATOUL(argv[3]);
  else
    printf("  BITS       EVENT\r\n");

  for (i=0; i<NumEventFlags; ++i)  {
    if ((idx != -1) AND (idx != i))  continue;
    bitsPtr = eventFlags[i].mask;
    if (newSetting)  *bitsPtr = setBits;
    printf("  0x%08lx %12ls\r\n", *bitsPtr, eventFlags[i].name);
  }
  printf("\n");


  return 0;
}

//-----------------------------------------------------------------------------
//  show or set diagnostic settings for processes or code events
//-----------------------------------------------------------------------------
void _task _diag(int argc, char **argv)
{
  uint ms;


  if (showHelp(2, argc, argv))  {
    printf("\n"
           "usage: diag crc|sched|msgs|events ...\r\n");
    Exit(-1);
  }

  if (StriEq(argv[1], "msgs"))    Exit( setDiagMsgs(argc, argv) );
  if (StriEq(argv[1], "events"))  Exit( setDiagBits(argc, argv) );
  if (StriEq(argv[1], "crc"))     { computeMemCRCs(false, true);  Exit(0);  }

  if (StriEq(argv[1], "sched"))  {
    if (argc > 2)  {
      ms = atoi(argv[2]);
      if (ms < 2)  ms = 2;  if (ms > 100) ms = 100;
      SchedMS = ms;
    }
    printf("SchedMS set to %u ms\r\n", SchedMS);
  }

  Exit(-1);
}

//-----------------------------------------------------------------------------
