//----- reg.c ----- 05/23/11 -----(03/14/06)-----

//-----------------------------------------------------------------------------
//
//  2011-05-23 - Mark Davis
//               - The user is no longer required to provide a channel number
//                 for registers that share one value for all channels.
//
//  2011-05-09 - Mark Davis
//               - Added support for a channel subscript on register names
//                 (e.g. regName[2])
//
//  2011-04-04 - Mark Davis
//               - Fixed interpretation of OnePerChan bit in the flags value
//                 for each register.
//
//  2010-05-27 - Mark Davis
//               - Added -d option to change the deadband (minChg) value for
//                 a register.
//
//  2009-10-26 - Mark Davis
//               - changes to work with register names ref via far ptr
//
//  2009-xx-xx - Mark Davis
//               - original version of ...
//
//-----------------------------------------------------------------------------

#include <time.h>
#include <common.h>

#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <errno.h>
#include <signal.h>
#include <utime.h>


#include <busybox.h>

#include <sys/hdwDefs.h>

#include <dcdefs.h>

#include <ctype.h>
#include <limits.h>

#include <sys/SHD_headers.h>

#include <sys/modbus.h>

#include <tsh.h>

#include <applic.h>

#include <sys/farprintf.h>


#define  MaxNameExprLen  64

//-----------------------------------------------------------------------------
//  process command for one of the groups of coils
//-----------------------------------------------------------------------------
bool checkCoils(CoilInfo **coils, int argc, char **argv,
                int numChans, fd_t eventsfd)
{

  CoilInfo  *coil, *lastCoil;
  uchar  *valPtr;
  uint  i, c, chan;
  bool  nameGiven, valueGiven, chanGiven, accepted, foundMatches, onePerChan;
  char *ptr;
  char  temp[64], regName[64], nameExpr[MaxNameExprLen];


  lastCoil = NULL;

  nameGiven = (argc > 1);
  valueGiven = (argc > 2);

//--- copy expr for matching reg names and extract chan index (if provided) ---
  chan = 0;  chanGiven = false;  *nameExpr = 0;
  if (nameGiven)  {
    strcpy(nameExpr, argv[1]);
    if ((ptr = strchr(nameExpr, '[')) != NULL)  {
      chan = (uint) atoi(ptr+1);  *ptr = 0;  chanGiven = true;
      if (chan >= numChans)  return false;
    }
  }

//----- coils (0xxxx) -----
  foundMatches = false;
  for (i=0,coil=c0Coils[0]; coil; ++i,coil=c0Coils[i])  {
    if (coil == lastCoil)  continue;

    strcpyfn(regName, coil->name);
    onePerChan = BOOL(coil->flags & OnePerChan);

  //--- if name given, skip non-matching coil names ---
    if (nameGiven)  {

    //--- if a new value was given then the entire name must also be given ---
      if (valueGiven)  {
        if (stricmp(regName, nameExpr))  continue;  // not an exact match

      //--- got exact name match - do we have a valid chan # ? ---
        if (onePerChan AND (numChans > 1) AND !chanGiven)  {
          printf("  Must supply chan # (e.g. regName[%d]\r\n", numChans-1);
          return false;
        }
        if (chan >= numChans)  {
          printf("  invalid chan #\r\n");  return false; }

        accepted = (coil->setCoilFunc)(coil, chan, atoi(argv[2]), eventsfd);
        printf("set %.32s[%d] = %s: %s\r\n",
               regName, chan, argv[2], (accepted ? "OK" : "*REJECTED*") );
        foundMatches = true;  break;
      }

    //--- otherwise check for a partial match ---
      if (!strstri(regName, nameExpr))  continue;
    }

  //--- if we got here, we have a partial match (or no name given at all) ---
    if (valueGiven)  continue;  // ignore partial matches if a value was given

  //--- no name filter given or got at least a partial match so show the ---
  //--- current value(s) ---
    sprintf(temp, "%05u     %s", coil->coilNum+1, regName);
    for (ptr=temp; *ptr; ++ptr);
    if (chanGiven)  sprintf(ptr, "[%u]", chan);
    printf("%-29s", temp);

    for (c=chan; c<NumC0Chans; ++c)  {
      valPtr = coil->coilVal + c;
      printf("  %4d", *valPtr);
      if (!onePerChan OR chanGiven) break;
    }
    printf("\r\n");

    lastCoil = coil;  foundMatches = true;
  }

  return foundMatches;
}


//-----------------------------------------------------------------------------
//  process command for one of the groups of registers
//-----------------------------------------------------------------------------
bool checkRegisters(RegInfo **regs, int argc, char **argv,
                    int numChans, fd_t eventsfd)
{
  RegInfo  *reg, *lastReg;
  uint  i, c, chan;
  bool  nameGiven, valueGiven, chanGiven, accepted, foundMatches,
        setDeadband, rwRegs, chgValue, prefix, onePerChan;
  char  *ptr;
  char  temp[64], regName[64], nameExpr[MaxNameExprLen];


  lastReg = NULL;

  nameGiven = (argc > 1);
  valueGiven = (argc > 2);
  setDeadband = (argc > 3) AND StriEq(argv[3], "-d");

//--- copy expr for matching reg names and extract chan index (if provided) ---
  chan = 0;  chanGiven = false;  *nameExpr = 0;
  if (nameGiven)  {
    strcpy(nameExpr, argv[1]);
    if ((ptr = strchr(nameExpr, '[')) != NULL)  {
      chan = (uint) atoi(ptr+1);  *ptr = 0;  chanGiven = true;
      if (chan >= numChans)  return false;
    }
  }

  rwRegs = (regs == r4Regs);

  if (valueGiven AND !rwRegs AND !setDeadband)  return false;

  chgValue = valueGiven AND (setDeadband OR rwRegs);

  foundMatches = false;


  for (i=0,reg=regs[0]; reg; ++i,reg=regs[i])  {
    if (reg == lastReg)  continue;

    strcpyfn(regName, reg->name);
    onePerChan = BOOL(reg->flags & OnePerChan);

  //--- if name given, skip non-matching register names ---
    if (nameGiven)  {

    //--- if a new value was given then the entire name must also be given ---
      if (chgValue)  {
        if (stricmp(regName, nameExpr))  continue;  // skip if not an exact match

      //--- got exact name match - do we have a valid chan # ? ---
        if (onePerChan AND (numChans > 1) AND !chanGiven)  {
          printf("  Must supply chan # (e.g. %s[%d])\r\n", regName, numChans-1);
          return false;
        }

        parseValue(reg, argv[2], temp);  // get new value from cmd string

        if (setDeadband)
          memcpy(reg->minChg, temp, reg->numRegs << 1);
        else  {
          if (!reg->setRegFunc)  {
            printf("  %s is Read Only\r\n", regName);
            foundMatches = true;  break;
          }
          accepted = (reg->setRegFunc)(reg, chan, temp, eventsfd);
          printf("set %.32s[%d] = %.20s: %s\r\n",
                 regName, chan, argv[2], (accepted ? "OK" : "*REJECTED*") );
        }
        foundMatches = true;  break;
      }

    //--- otherwise check for a partial match ---
      if (!strstri(regName, nameExpr))  continue;
    }


  //--- no name filter given or got at least a partial match ---
    foundMatches = true;

    for (prefix=true,c=chan; c<numChans; ++c,prefix=false)  {
      if (prefix)
        printf("%c%04u#%s ",
               rwRegs ? '4' : '3', reg->regNum+1, typeDesc[reg->dataType]);
      else
        printf("          ");
      strcpy(temp, regName);  for (ptr=temp; *ptr; ++ptr);
      if (onePerChan)  sprintf(ptr, "[%u]", c);
      printf("%-19s", temp);
      printf("  %s", fmtRegVal(reg, NULL, c, temp));
      printf("  %s\r\n", fmtVal(temp, reg->dataType, 12, reg->minChg));
      if (!onePerChan OR chanGiven)  break;
    }

    lastReg = reg;
  }

  return foundMatches;
}


//-----------------------------------------------------------------------------
//  list or modify ModBus register values
//-----------------------------------------------------------------------------
void _task _reg(int argc, char **argv)
{
  bool  foundMatches;
  fd_t  eventsfd = -1;


  if (showHelp(1, argc, argv))  {
    printf("\nusage:  reg [name [newValue [-d]]]\r\n");  Exit(-1); }


  if (argc > 2)  eventsfd = open(EventsFIFO, O_WRONLY, 0);

  printf("\r\n");

  if (argc > 1)  if (strlen(argv[1]) >= MaxNameExprLen)  {
    printf("reg name expr too long\r\n\n");  Exit(-1);  }

//--- read/write coils (0xxxx) ---
  if (checkCoils(c0Coils, argc, argv, NumC0Chans, eventsfd))  {
    foundMatches = true;  printf("\r\n");  }

//--- read-only registers (3xxxx) ---
  if (checkRegisters(r3Regs, argc, argv, NumR3Chans, eventsfd))  {
    foundMatches = true;  printf("\r\n");  }

//--- read/write registers (4xxxx) ---
  if (checkRegisters(r4Regs, argc, argv, NumR4Chans, eventsfd))  {
    foundMatches = true;  printf("\r\n");  }


  if (! foundMatches)  printf("No matching registers\r\n");

  Exit(0);
}


//-----------------------------------------------------------------------------
