//----- sysinfo.c ----- 06/07/12 -----(03/14/06)-----

//-----------------------------------------------------------------------------
//  Update Log (most recent first):
//
//
//  2012-06-07 - Mark Davis
//               - Added call to taskCounts()
//
//  2012-06-06 - Mark Davis
//               - Changed headers for the -r(esource) option to indicate # of
//                 socket connections rather than # of sockets (since the # of
//                 established connections is the more important # for TCP).
//
//  2012-04-20 - Mark Davis
//               - Added -r option to monitor resource usage.
//
//  2010-07-19 - Mark Davis
//               - Memory protection information is now shown.
//
//  2010-05-19 - Mark Davis
//               - Added current state of MMU Memory Bank Control Registers
//
//  2010-05-17 - Mark Davis
//               - Added CPU Type and Revision to information shown
//
//  2009-xx-xx - Mark Davis
//               - original version of ...
//
//-----------------------------------------------------------------------------

#include <time.h>
#include <common.h>

#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <errno.h>
#include <signal.h>
#include <utime.h>


#include <busybox.h>

#include <sys/hdwDefs.h>

#include <sys/srfs.h>

#include <dcdefs.h>

#include <ctype.h>
#include <limits.h>

#include <sys/SHD_headers.h>

#include <sys/modbus.h>

#include <tsh.h>

#include <applic.h>

#include <semaphore.h>


#include <linux/wait.h>


#include <sys/farprintf.h>



static uchar WaitStates[4] = { 4, 2, 1, 0 };

static char *EnableStates[4] = {
  "OE0 WE0",  "OE1 WE1",  "OE0 WP ",  "OE1 WP " };


typedef uint (*FreeCount)(uint *ttl);


typedef struct {
  const char far *descr;
  FreeCount  freeCount;
} ResCount;


#pragma seg (CONST,FARCONST)

const far ResCount resCount[] = {
  { " WaitQueues",  wqCounts   },
  { "  Named sem",  semCounts  },
  { " Global fds",  gfdCounts  },
  { "   TCP cons",  tcpCounts  },
  { "   UDP cons",  udpCounts  },
  { "  Processes",  taskCounts }
};

#pragma seg (CONST)

int const NumRes = ( sizeof(resCount) / sizeof(ResCount) );


//-----------------------------------------------------------------------------
//  Show a continuous display of counts for various system resources
//-----------------------------------------------------------------------------
void resourceMonitor(void)
{
  uint  u, total, used, free_, schedCalls, taskSws;


  printf(clsCmd);

  for (;;)  {
    stdHeader();  printf("%s", cllCmd);

    printf("\r\n"
           "Resources:           MAX       USED       FREE   USE%% \r\n");
    //        xxxxxxxxxxx:  ########   ########   ########   ###%

    for (u=0; u<NumRes; ++u)  {
      free_ = (*resCount[u].freeCount)(&total);  used = total - free_;
      printf("  %ls:  %8u   %8u   %8u   %3u%% \r\n",
             resCount[u].descr, total, used, free_, used * 100 / total);
    }


    _EnterCritical_();

    schedCalls = schedCallsPrevSec;
    taskSws = taskSwPrevSec;

    _ExitCritical_();

    printf("\r\n"
           "Event counters:\r\n"
           "  Sched Calls:  %8u\r\n"
           "  Task Switch:  %8u\r\n"
           , schedCalls, taskSws);

    printf(clpCmd);  fflush(stdout);  sleepMS(2000);  CheckETX();
  }

}


//-----------------------------------------------------------------------------
float adctoAin(uchar chan, uint rawADC)
{
  return (ADCOffset(chan) - (float)(rawADC)) * ADCGain(chan);
}

//-----------------------------------------------------------------------------
void showMBxCR(uint ui)
{
  uchar  mbxcr;

  mbxcr = *(&srMB0CR + ui);

  printf(" MB%uCR: %02X:  %u ws  A19:%c  A18:%c  %s  CS%u\r\n",
         ui, mbxcr, WaitStates[mbxcr >> 6],
         mbxcr & 0x20 ? 'I' : 'N',
         mbxcr & 0x10 ? 'I' : 'N',
         EnableStates[(mbxcr >> 2) & 0x03],
         mbxcr & 0x03 );
}


//-----------------------------------------------------------------------------
//  show information about the current system
//-----------------------------------------------------------------------------
void _task _sysinfo(int argc, char **argv)
{
  uchar  ui;
  FlashDevice  *fi;
  uchar  msg[80];


  if (showHelp(1, argc, argv))  {
    printf("\n usage:  %s [-r]\r\n", argv[0]);  Exit(-1); }


  if (argc > 1)  {
    if (StrEq(argv[1], "-r"))  resourceMonitor();
  }

//--- get hardware-related info ---
  printf("\r\n CPU ID: %u   Rev: %u\r\n\n",
         ini(GCPU) & 0x1F, ini(GREV) & 0x1F);

  for (ui=0; ui<4; ++ui)  showMBxCR(ui);
  printf("\r\n");

  if (hdwProtection)  showMemProtection();

  printf("\r\n ID Block vers: %d\r\n", idBlock->versionNumber);
  printf(" ID Block size: %d\r\n", idBlock->blockSize);
  printf(" Board ID: 0x%04X\r\n", idBlock->productID);
  sprintf(msg, " MAC Addr: ");
  for (ui=0; ui<6; ++ui)
    sprintf(msg+10+ui*3, "%02x:", idBlock->macAddr[0][ui]);
  strcpy(msg+strlen(msg)-1, "\r\n\n");  printf(msg);

  for (ui=0; ui<2; ++ui)  {
    fi = idBlock->flash + ui;  if (! fi->flashID)  continue;
    printf(" Flash%u:\r\n", ui);
    printf("   Part ID: 0x%04X\r\n", fi->flashID);
    printf("   Write method: %s\r\n", (fi->flashType == 1) ? "Byte" : "Sector");
    printf("   Size: %uK\r\n", fi->flashSize);
    printf("   Sectors: %u\r\n\n", fi->numSectors);
  }

//--- test floating point speed ---
  {  //tdebug
    uint  ui;
    float  x, y, z;
    ulong  startTime, ttlms;

    y = 1.2345;  z = -9.8765;
    KLOCK();
    startTime = MS_TIMER;
    for (ui=0; ui<1000; ++ui)  x = y * z;
    ttlms = msSince(startTime);
    KUNLOCK();
    y = x;  // to get rid of compiler warning...
    printf("took %d microsecs / float mult\r\n", (uint)ttlms);
  }


//--- test old-method analog conversion speed ---
  {  //tdebug
    uint  ui;
    ulong  startTime, ttlms;

    KLOCK();
    startTime = MS_TIMER;
    for (ui=0; ui<1000; ++ui)  adctoAin(0, 0x0123);
    ttlms = msSince(startTime);
    KUNLOCK();
    printf("using floats: took %d microsecs / ADC conversion\r\n", (uint)ttlms);
  }

/*
//--- test analog value conversion speed ---
  {  //tdebug
    uint  ui;
    ulong  startTime, ttlms;
    float  fv;

    KLOCK();
    startTime = MS_TIMER;
    for (ui=0; ui<1000; ++ui)  fv = ADCtoAin(0, 0x0123);
    ttlms = msSince(startTime);
    KUNLOCK();
    fv = fv * fv;  // to eliminate compiler warning...
    printf("using fixed: took %d microsecs / ADC conversion\r\n", (uint)ttlms);
  }
*/

  Exit(0);
}
//-----------------------------------------------------------------------------
