//----- test_pv.c ----- 02/24/12 -----(05/25/10)-----

//-----------------------------------------------------------------------------
//  process to test EPICS CA/PV interface to ModBus register values
//-----------------------------------------------------------------------------
// Update Log (most recent first)
//
//  Entries marked with (*) indicate ones that improve the stability and/or
//  reliability of the code.
//
//
//  2012-02-24 - Mark Davis
//               - Added elmSize and numElems to output.
//
//  2012-02-14 - Mark Davis
//               - Name change for function call to pvGetPID (now pvGetID)
//
//  2012-01-20 - Mark Davis
//               - Changed path for poll.h
//
//  2010-06-10 - Mark Davis
//               - Modified to accomodate the change of PID values from long
//                 to int.
//
//  2010-05-25 - Mark Davis
//               - first version started
//
//-----------------------------------------------------------------------------
#include <common.h>

#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <poll.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <errno.h>
#include <signal.h>
#include <utime.h>
#include <linux/sched.h>
#include <errno.h>

#include <string.h>

#include <stcpip.h>
#include <inetd.h>


#include <sys/farprintf.h>

#include <tsh.h>

#include <busybox.h>

#include <sys/modbus.h>


//-----------------------------------------------------------------------------
//  Ask for the value of a PV in each format and show the result
//-----------------------------------------------------------------------------
void pvShowVal(int pid)
{
  char  temp[64];


  pvGetVal(pid, DT_S16, sizeof(S16), temp, CUR_VALUE);
  printf("  converted to S16: %12d (0x%04x)\r\n", *(S16 *)temp, *(S16 *)temp);

  pvGetVal(pid, DT_U16, sizeof(U16), temp, CUR_VALUE);
  printf("  converted to U16: %12u (0x%04x)\r\n", *(U16 *)temp, *(U16 *)temp);

  pvGetVal(pid, DT_S32, sizeof(S32), temp, CUR_VALUE);
  printf("  converted to S32: %12ld (0x%08lx)\r\n", *(S32 *)temp, *(S32 *)temp);

  pvGetVal(pid, DT_U32, sizeof(U32), temp, CUR_VALUE);
  printf("  converted to U32: %12lu (0x%08lx)\r\n", *(U32 *)temp, *(U32 *)temp);

  pvGetVal(pid, DT_F32, sizeof(F32), temp, CUR_VALUE);
  printf("  converted to F32: %s (0x%08lx)\r\n",
         Engineering(*(F32 *)temp, 6, temp+16), *(F32 *)temp);

  pvGetVal(pid, DT_STR, sizeof(temp), temp, CUR_VALUE);
  printf("  converted to STR: '%.64s'\r\n", temp);

  printf("\r\n");
}


//-----------------------------------------------------------------------------
//  process to listen on specified ports and create processes to service
//  incoming connections on those ports
//-----------------------------------------------------------------------------
void _task test_pv(int argc, char **argv)
{
  PV_Info  pvInfo;
  int  eventsOccured;
  int  pid;
  uchar  useType;
  time_t  dateTime;
  uint  ms;
  char  temp[128], origVal[16];


  if (showHelp(2, argc, argv))  {
    printf("\nusage: test_pv pvName [setValue type]\r\n\n");  Exit(-1); }

  if ((pid = pvGetID(argv[1])) < 0)  {
    printf("No PV named: [%.32s]\r\n", argv[1]);  Exit(-2);  }


  pvGetInfo(pid, &pvInfo);

  printf("\r\n"
         " PID: %d  dataType: %d(%.8s)  flags: 0x%04X \r\n"
         " size: %d  elemSize: %u  numElems: %u \r\n\n",
         pid, pvInfo.dataType, typeDesc[pvInfo.dataType], pvInfo.flags,
         pvInfo.size, pvInfo.elemSize, pvInfo.numElems);

//--- save the current value ---
  pvGetVal(pid, pvInfo.dataType, sizeof(origVal), origVal, CUR_VALUE);

  printf(" Original value:\r\n");  pvShowVal(pid);


  if (argc < 4)  {
    pvCheckEvents(pid, DBE_VALUE);
    printf("Testing value change events...\r\n"
           "  press Ctl-C to exit\r\n");
    for (;;)  {
      CheckETX();  sleepMS(200);

      pvCheckAllEvents(&dateTime, &ms, false);

      eventsOccured = pvCheckEvents(pid, DBE_VALUE);
      if (eventsOccured & DBE_VALUE)  {
        pvGetVal(pid, DT_STR, sizeof(temp), temp, DBE_VALUE);
        printf("  %010lu.%03u: %s\r\n", dateTime, ms, temp);
      }
    }
  }

  if (! (pvInfo.flags & PV_WR))  { printf("Read-Only value\r\n\n");  Exit(-3); }


  for (useType=0; useType<NUM_MB_TYPES; ++useType)
    if (StriEq(typeDesc[useType], argv[3]))  break;

  if (useType >= NUM_MB_TYPES)  {
    printf("Invalid data type: '%.8s'\r\n\nValid types are: ", argv[3]);
    for (useType=0; useType<NUM_MB_TYPES; ++useType)
      printf("%s ", typeDesc[useType]);
    printf("\r\n\n");  Exit(-4);
  }


  printf(" Convert '%.32s' to %s and write: ", argv[2], typeDesc[useType]);

  switch (useType)  {

    case DT_S16:
      *(S16 *)temp = ATOI(argv[2]);
      pvSetVal(pid, DT_S16, sizeof(S16), temp);
      printf("%d\r\n", *(S16 *)temp);
      pvShowVal(pid);  break;

    case DT_U16:
      *(U16 *)temp = ATOU(argv[2]);
      pvSetVal(pid, DT_U16, sizeof(U16), temp);
      printf("%u\r\n", *(U16 *)temp);
      pvShowVal(pid);  break;

    case DT_S32:
      *(S32 *)temp = ATOL(argv[2]);
      pvSetVal(pid, DT_S32, sizeof(S32), temp);
      printf("%ld\r\n", *(S32 *)temp);
      pvShowVal(pid);  break;

    case DT_U32:
      *(U32 *)temp = ATOUL(argv[2]);
      pvSetVal(pid, DT_U32, sizeof(U32), temp);
      printf("%lu\r\n", *(U32 *)temp);
      pvShowVal(pid);  break;

    case DT_F32:
      *(F32 *)temp = atof(argv[2]);
      pvSetVal(pid, DT_F32, sizeof(F32), temp);
      printf("%s\r\n", Engineering(*(F32 *)temp, 6, temp+32));
      pvShowVal(pid);  break;

    case DT_STR:
      pvSetVal(pid, DT_STR, strlen(argv[2]), argv[2]);
      printf("\r\n");
      pvShowVal(pid);  break;
  }

//--- restore the original value ---
  pvSetVal(pid, pvInfo.dataType, sizeof(origVal), origVal);

  Exit(0);
}

//-----------------------------------------------------------------------------
